(*===========================================================================*)
(* 8250 hardware constants                                                   *)
(*                                                                           *)
(*   Copyright 1988, 1989 by H. Roy Engehausen.  All rights reserved.        *)
(*   This software may be freely distributed and used, but it may not        *)
(*   under any circumstances be sold by anyone other than the author.        *)
(*   It may be distributed by a commercial company as long as it is          *)
(*   for no cost.                                                            *)
(*                                                                           *)
(*===========================================================================*)

(*---------------------------------------------------------------------------*)
(* Address offsets                                                           *)
(*---------------------------------------------------------------------------*)

CONST
     rbr_8250      = $00   ; (* xF8   Receive Buffer Register                *)
     thr_8250      = $00   ; (* xF8   Transmitter Holding Register           *)
     ier_8250      = $01   ; (* xF9   Interrupt Enable Register              *)
     iir_8250      = $02   ; (* xFA   Interrupt Identification Register      *)
     lcr_8250      = $03   ; (* xFB   Line Control Register                  *)
     mcr_8250      = $04   ; (* xFC   Modem Control Register                 *)
     lsr_8250      = $05   ; (* xFD   Line Status Register                   *)
     msr_8250      = $06   ; (* xFE   Modem Status Register                  *)
     dll_8250      = $00   ; (* xF8   Divisor Latch Least Significant        *)
     dlm_8250      = $01   ; (* xF9   Divisor Latch Most  Significant        *)

(*---------------------------------------------------------------------------*)
(* Interrupt enable register definiton                                       *)
(*---------------------------------------------------------------------------*)

     (*            = $F0   ;          Not used                               *)
     ier_8250_ms   = $08   ; (* Modem status interrupt enable                *)
     ier_8250_rls  = $04   ; (* Receive line status interrupt enable         *)
     ier_8250_thre = $02   ; (* Transmit holding register interrupt enable   *)
     ier_8250_rda  = $01   ; (* Receive data available interrupt enable      *)

(*---------------------------------------------------------------------------*)
(* Interrupt identification register definition                              *)
(*---------------------------------------------------------------------------*)

     (*            = $F8   ;    Not used -- Always zero                      *)
     (*            = .....xx.B  These bites determine the interrupt type     *)
     iir_8250_ip   = $01   ; (* If this bit = 0 then interrupt pending       *)
     iir_8250_ms   = $00   ; (* Modem status interrupt                       *)
     iir_8250_thre = $02   ; (* Transmit holding register empty interrupt    *)
     iir_8250_rda  = $04   ; (* Receive data available interrupt             *)
     iir_8250_rls  = $08   ; (* Receive line status interrupt                *)

(*---------------------------------------------------------------------------*)
(* Line control register definition                                          *)
(*---------------------------------------------------------------------------*)

     (*            = ......xxB  Word size select                             *)
     lcr_8250_ws5  = $00   ; (* Word select -- 5 bits                        *)
     lcr_8250_ws6  = $01   ; (* Word select -- 6 bits                        *)
     lcr_8250_ws7  = $02   ; (* Word select -- 7 bits                        *)
     lcr_8250_ws8  = $03   ; (* Word select -- 8 bits                        *)

     lcr_8250_s2   = $04   ; (* 2 stop bits wanted instead of 1              *)
     lcr_8250_pe   = $08   ; (* Parity enable -- 1 = parity bit wanted       *)
     lcr_8250_eps  = $10   ; (* Even parity select -- 1 = even parity        *)
     lcr_8250_sp   = $20   ; (* Parity stick -- 1 = stick parity on          *)
     lcr_8250_sb   = $40   ; (* Set Break -- Latches output at 0 state       *)
     lcr_8250_dlab = $80   ; (* Divisior Latch Access Bit -- 1 = divisor     *)

(*---------------------------------------------------------------------------*)
(* Modem control register definition                                         *)
(*---------------------------------------------------------------------------*)

     (*            = 000.....B ;   Not used                                  *)
     mcr_8250_loop = $10   ; (* Loop back mode                               *)
     mcr_8250_out2 = $08   ; (* OUT2 on -- This enables interrupts           *)
     mcr_8250_out1 = $04   ; (* OUT1 on                                      *)
     mcr_8250_rts  = $02   ; (* Request To Sent (RTS)                        *)
     mcr_8250_dtr  = $01   ; (* Data Terminal Ready (DTR)                    *)

     hon_8250      = $0B   ; (* 8250 Initialization and handshaking const    *)
                             (*   Gives DTR, RTS, and OUT2.  OUT2 control    *)
                             (*   the IBM ASYNC card interrupt enable        *)

     hoff_8250     = $08   ; (* 8250 Handshaking off.  Gives only OUT2.      *)
                             (*    TNC1 uses RTS, TNC2 uses DTR.             *)

(*---------------------------------------------------------------------------*)
(* Line status register definition                                           *)
(*---------------------------------------------------------------------------*)

     (*            = $80   ;    Not used                                     *)
     lsr_8250_tsre = $40   ; (* Transmit shift register empty                *)
     lsr_8250_thre = $20   ; (* Transmit holding register empty (THRE)       *)
     lsr_8250_bi   = $10   ; (* Break interrupt                              *)
     lsr_8250_fe   = $08   ; (* Framing error                                *)
     lsr_8250_pe   = $04   ; (* Parity error                                 *)
     lsr_8250_or   = $02   ; (* Overrun error                                *)
     lsr_8250_dr   = $01   ; (* Receive data ready (RDA)                     *)

(*---------------------------------------------------------------------------*)
(* Modem status register definition                                          *)
(*---------------------------------------------------------------------------*)

     msr_8250_rlsd = $80   ; (* Receive line signal detect (DCD)             *)
     msr_8250_ri   = $40   ; (* Ring indicator (RI)                          *)
     msr_8250_dsr  = $20   ; (* Dataset ready (DSR)                          *)
     msr_8250_cts  = $10   ; (* Clear to send (CTS)                          *)

     (* These are deltas.. They turn on if the signal has changed since      *)
     (* last time you read this register                                     *)

     msr_8250_drlsd= $08   ; (* Delta Receive line signal detect (DCD)       *)
     msr_8250_dri  = $04   ; (* Delta Ring indicator (RI)                    *)
     msr_8250_ddsr = $02   ; (* Delta Dataset ready (DSR)                    *)
     msr_8250_dcts = $01   ; (* Delta Clear to send (CTS)                    *)

