/* ARP packet tracing routines
 * Copyright 1991 Phil Karn, KA9Q
 */
#include <stdio.h>
#include "global.h"
#include "mbuf.h"
#include "arp.h"
#include "netuser.h"
#include "trace.h"

void
arp_dump(ip,bpp)
struct iface *ip;
struct mbuf **bpp;
{
	struct arp arp;
	struct arp_type *at;
	int is_ip = 0;
	char tmp[25];

	if(bpp == NULLBUFP || *bpp == NULLBUF)
		return;
	prtTrace(ip,"ARP: len %d",len_p(*bpp));
	if(ntoharp(&arp,bpp) == -1){
		prtTrace(ip," bad packet\n");
		return;
	}
	if(arp.hardware < NHWTYPES)
		at = &Arp_type[arp.hardware];
	else
		at = NULLATYPE;

	/* Print hardware type in Ascii if known, numerically if not */
	prtTrace(ip," hwtype %s",smsg(Arptypes,NHWTYPES,arp.hardware));

	/* Print hardware length only if unknown type, or if it doesn't match
	 * the length in the known types table
	 */
	if(at == NULLATYPE || arp.hwalen != at->hwalen)
		prtTrace(ip," hwlen %u",arp.hwalen);

	/* Check for most common case -- upper level protocol is IP */
	if(at != NULLATYPE && arp.protocol == at->iptype){
		prtTrace(ip," prot IP");
		is_ip = 1;
	} else {
		prtTrace(ip," prot 0x%x prlen %u",arp.protocol,arp.pralen);
	}
	switch(arp.opcode){
	case ARP_REQUEST:
		prtTrace(ip," op REQUEST");
		break;
	case ARP_REPLY:
		prtTrace(ip," op REPLY");
		break;
	case REVARP_REQUEST:
		prtTrace(ip," op REVERSE REQUEST");
		break;
	case REVARP_REPLY:
		prtTrace(ip," op REVERSE REPLY");
		break;
	default:
		prtTrace(ip," op %u",arp.opcode);
		break;
	}
	prtTrace(ip,"\n");
	prtTrace(ip,"sender");
	if(is_ip)
		prtTrace(ip," IPaddr %s",inet_ntoa(arp.sprotaddr));
	prtTrace(ip," hwaddr %s\n",at->format(tmp,arp.shwaddr));

	prtTrace(ip,"target");
	if(is_ip)
		prtTrace(ip," IPaddr %s",inet_ntoa(arp.tprotaddr));
	prtTrace(ip," hwaddr %s\n",at->format(tmp,arp.thwaddr));
}
