/* AX25 header tracing
 * Copyright 1991 Phil Karn, KA9Q
 */
 /* Mods by PA0GRI */
#include <stdio.h>
#include "global.h"
#include "mbuf.h"
#include "ax25.h"
#include "lapb.h"
#include "trace.h"
#include "socket.h"

static char *decode_type __ARGS((int16 type));

/* Dump an AX.25 packet header */
void
ax25_dump(ip,bpp,check)
struct iface *ip;
struct mbuf **bpp;
int check;	/* Not used */
{
	char tmp[AXBUF];
	char frmr[3];
	int control,pid,seg;
	int16 type;
	int unsegmented;
	struct ax25 hdr;
	char *hp;

	prtTrace(ip,"AX25: ");
	/* Extract the address header */
	if(ntohax25(&hdr,bpp) < 0){
		/* Something wrong with the header */
		prtTrace(ip," bad header!\n");
		return;
	}
	prtTrace(ip,"%s",pax25(tmp,hdr.source));
	prtTrace(ip,"->%s",pax25(tmp,hdr.dest));
	if(hdr.ndigis > 0){
		prtTrace(ip," v");
		for(hp = hdr.digis[0]; hp < &hdr.digis[hdr.ndigis][0];
		 hp += AXALEN){
			/* Print digi string */
			prtTrace(ip," %s%s",pax25(tmp,hp),
			 (hp[ALEN] & REPEATED) ? "*":"");
		}
	}
	if((control = PULLCHAR(bpp)) == -1)
		return;

	prtTrace(ip, " ");
	type = ftype(control);
	prtTrace(ip,"%s",decode_type(type));
	/* Dump poll/final bit */
	if(control & PF){
		switch(hdr.cmdrsp){
		case LAPB_COMMAND:
			prtTrace(ip,"(P)");
			break;
		case LAPB_RESPONSE:
			prtTrace(ip,"(F)");
			break;
		default:
			prtTrace(ip,"(P/F)");
			break;
		}
	}
	/* Dump sequence numbers */
	if((type & 0x3) != U)	/* I or S frame? */
		prtTrace(ip," NR=%d",(control>>5)&7);
	if(type == I || type == UI){	
		if(type == I)
			prtTrace(ip," NS=%d",(control>>1)&7);
		/* Decode I field */
		if((pid = PULLCHAR(bpp)) != -1){	/* Get pid */
			if(pid == PID_SEGMENT){
				unsegmented = 0;
				seg = PULLCHAR(bpp);
				prtTrace(ip,"%s remain %u",seg & SEG_FIRST ?
				 " First seg;" : "",seg & SEG_REM);
				if(seg & SEG_FIRST)
					pid = PULLCHAR(bpp);
			} else
				unsegmented = 1;

			switch(pid){
			case PID_SEGMENT:
				prtTrace(ip, "\n");
				break;	/* Already displayed */
			case PID_ARP:
				prtTrace(ip," pid=ARP\n");
				arp_dump(ip,bpp);
				break;
			case PID_RARP:
				prtTrace(ip," pid=RARP\n");
				arp_dump(ip,bpp);
				break;
			case PID_NETROM:
				prtTrace(ip," pid=NET/ROM\n");
				/* Don't verify checksums unless unsegmented */
				netrom_dump(ip,bpp,unsegmented);
				break;
			case PID_IP:
				prtTrace(ip," pid=IP\n");
				/* Don't verify checksums unless unsegmented */
				ip_dump(ip,bpp,unsegmented);
				break;
			case PID_X25:
				prtTrace(ip," pid=X.25\n");
				break;
			case PID_TEXNET:
				prtTrace(ip," pid=TEXNET\n");
				break;
			case PID_NO_L3:
				prtTrace(ip," pid=Text\n");
				break;
			default:
				prtTrace(ip," pid=0x%x\n",pid);
			}
		}
	} else if(type == FRMR && pullup(bpp,frmr,3) == 3){
		prtTrace(ip,": %s",decode_type(ftype(frmr[0])));
		prtTrace(ip," Vr = %d Vs = %d",(frmr[1] >> 5) & MMASK,
			(frmr[1] >> 1) & MMASK);
		if(frmr[2] & W)
			prtTrace(ip," Invalid control field");
		if(frmr[2] & X)
			prtTrace(ip," Illegal I-field");
		if(frmr[2] & Y)
			prtTrace(ip," Too-long I-field");
		if(frmr[2] & Z)
			prtTrace(ip," Invalid seq number");
		prtTrace(ip, "\n");
	} else
		prtTrace(ip, "\n");

}
static char *
decode_type(int16 type)
{
	switch(type){
	case I:
		return "I";
	case SABM:
		return "SABM";
	case DISC:
		return "DISC";
	case DM:
		return "DM";
	case UA:
		return "UA";
	case RR:
		return "RR";
	case RNR:
		return "RNR";
	case REJ:
		return "REJ";
	case FRMR:
		return "FRMR";
	case UI:
		return "UI";
	default:
		return "[invalid]";
	}
}

/* Return 1 if this packet is directed to us, 0 otherwise. Note that
 * this checks only the ultimate destination, not the digipeater field
 */
int
ax_forus(iface,bp)
struct iface *iface;
struct mbuf *bp;
{
	struct mbuf *bpp;
	char dest[AXALEN];

	/* Duplicate the destination address */
	if(dup_p(&bpp,bp,0,AXALEN) != AXALEN){
		free_p(bpp);
		return 0;
	}
	if(pullup(&bpp,dest,AXALEN) < AXALEN)
		return 0;
	if(addreq(dest,iface->hwaddr))
		return 1;
	else
		return 0;
}

