(*===========================================================================*)
(* Monitor                                                                   *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1991 by H. Roy Engehausen.  All rights reserved.  *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

UNIT BBAUXM;

INTERFACE

  USES
    bbdummy;

  PROCEDURE monitor_all;

  PROCEDURE add_mon_call(to_port     : port_block_ptr;
                         port_letter : CHAR;
                         call_add    : bb_addr_str);

IMPLEMENTATION

  USES
    bbbug,
    bblog,
    bblstr,
    bbmisc,
    bbsrt,
    bbstr,
    bbtime,
    bbwin;

{$I BBADDM.PAS}   (* Add data to monitor list *)

(*===========================================================================*)
(* Monitor a channel                                                         *)
(*===========================================================================*)

PROCEDURE monitor_all;

  VAR
    add_port    : port_block_ptr;
    netrom      : BOOLEAN;
    netrom_node : BOOLEAN;
    i           : BYTE;
    in_1        : str_mixed;
    m_call      : bb_addr_str;
    m_color     : BYTE;
    to_pos      : BYTE;
    w           : WORD;
    t_str       : STRING[4];

  {$I BBAUXMBF.PAS} (* Bad monitor frame     *)
  {$I BBAUXMW.PAS}  (* Monitor window write  *)
  {$I BBAUXMNR.PAS} (* Monitor NETROM frames *)

  BEGIN;

    WITH active_port^, active_tcb^ DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Get ready to see what we got                                      *)
        (*-------------------------------------------------------------------*)

        add_port := active_port;

        t_str := port_char + ':';
        in_1  := tnc_data;

        to_pos := 4;

        (*-------------------------------------------------------------------*)
        (* Switch ports as needed for the PC*PA card and BPQ hosts           *)
        (*-------------------------------------------------------------------*)

        IF (port_type = port_pcpa) OR (port_type = port_bpqhost) THEN
          BEGIN;

            IF port_num <> tnc_data.str_data[1] THEN
              BEGIN;

                (*-----------------------------------------------------------*)
                (* Find the correct port                                     *)
                (*-----------------------------------------------------------*)

                add_port := rel_port;
                WHILE (add_port <> NIL)
                        AND (add_port^.port_num <> tnc_data.str_data[1]) DO
                  add_port := add_port^.rel_port;

                (*-----------------------------------------------------------*)
                (* If a sub port could not be found, handle it               *)
                (*-----------------------------------------------------------*)

                IF add_port = NIL THEN
                  BEGIN;

                    (*-------------------------------------------------------*)
                    (* In the case of a BPQ host mode, we could have a       *)
                    (* default port.  See if there is one                    *)
                    (*-------------------------------------------------------*)

                    IF active_port^.port_type = port_bpqhost THEN
                      BEGIN;
                        add_port := active_port;
                        WHILE (add_port <> NIL)
                                            AND (add_port^.port_num <> '0') DO
                          add_port := add_port^.rel_port;
                      END;

                    (*-------------------------------------------------------*)
                    (* If no port when we get here, quit                     *)
                    (*-------------------------------------------------------*)

                    IF add_port = NIL THEN
                      BEGIN;
                        window_write_critical(
                                'Incoming monitor on undefined port -- def = ',
                                active_port^.port_num);
                        window_write_critical('Data = ', tnc_data.str_data);
                        EXIT;
                      END;

                  END; (*----- End of port not found code -------------------*)

                (*-----------------------------------------------------------*)
                (* Set the correct character                                 *)
                (*-----------------------------------------------------------*)

                t_str[1] := add_port^.port_char;

              END;

            (*---------------------------------------------------------------*)
            (* Remove port data from front of message.                       *)
            (*---------------------------------------------------------------*)

            IF port_type = port_pcpa THEN
              BEGIN;
                tnc_data.str_data := substr(tnc_data.str_data, 4, 0);
                to_pos := 5;
              END
            ELSE
              IF add_port^.port_num <> '0' THEN
                BEGIN;
                  tnc_data.str_data := substr(tnc_data.str_data, 3, 0);
                  in_1              := l_substr(@in_1, 3, 0)^;
                END;

          END; (*----- End port switch --------------------------------------*)

        (*-------------------------------------------------------------------*)
        (* Set the color to use                                              *)
        (*-------------------------------------------------------------------*)

        m_color := add_port^.port_color;

        (*-------------------------------------------------------------------*)
        (* Get the call sign of the monitored station.                       *)
        (*-------------------------------------------------------------------*)

        CASE active_port^.port_type OF
          port_g8bpq, port_aeapk232:
            WITH tnc_data DO
              BEGIN;
                i := POS('>', str_data);
                IF (i > 1) AND (str_data[i-1] = '*') THEN
                  DEC(i);
                DEC(i);
                IF (i >= SIZEOF(bb_addr_str)) THEN
                  i := SIZEOF(bb_addr_str) - 1;
                IF i <> 0 THEN
                  m_call := substr(str_data, 1, i)
                ELSE
                  m_call := '';
              END;
          ELSE
            m_call := subwordl(tnc_data.str_data, 2, SIZEOF(bb_addr_str));
        END;

        (*-------------------------------------------------------------------*)
        (* Decode the monitor type                                           *)
        (*-------------------------------------------------------------------*)

        CASE tnc_type OF

          (*-----------------------------------------------------------------*)
          (* Link status change (should only be failed connect requests      *)
          (*-----------------------------------------------------------------*)

          t_to_h_links  : monitor_window_write(@in_1, TRUE,
                                                 opt_block.opt_mon_time_stamp);

          (*-----------------------------------------------------------------*)
          (* Monitor with no data frame following                            *)
          (*    DED mode -- This is all there is                             *)
          (*    PK232    -- All info is in this frame                        *)
          (*    BPQHOST  -- All info is in this frame                        *)
          (*-----------------------------------------------------------------*)

          t_to_h_mh_noi :
                          BEGIN;

                            IF active_port^.port_type = port_aeapk232 THEN
                              netrom_monitor_232;

                            IF active_port^.port_type = port_bpqhost THEN
                              BEGIN;
                           {$IFNDEF SLICE}
                                w := l_pos(@in_1, cr) + 1;

                                netrom := (w > 3) AND
                                        (subword(@in_1.str_data, 8, 1) = 'CF');

                                IF netrom THEN
                                  BEGIN;
                                    netrom_node :=
                                          subword(@in_1.str_data, to_pos, 1)
                                                                     = 'NODES';
                                    netrom_monitor(@in_1,
                                                   w,
                                                   netrom_node);
                                  END;
                           {$ENDIF}

                              END;

                            monitor_window_write(@in_1,
                                                 TRUE,
                                                 opt_block.opt_mon_time_stamp);
                                                 add_mon_call(add_port,
                                         add_port^.port_char,
                                         m_call);
                          END;

          (*-----------------------------------------------------------------*)
          (* Monitor with data frame following (DED mode only)               *)
          (*-----------------------------------------------------------------*)

          t_to_h_mh_i   :
            BEGIN;

              (*-------------------------------------------------------------*)
              (* Add call to the list                                        *)
              (*-------------------------------------------------------------*)

              add_mon_call(add_port, add_port^.port_char, m_call);

              (*-------------------------------------------------------------*)
              (* Pull the monitor info frame                                 *)
              (*-------------------------------------------------------------*)

              send_recv_tnc(2);

              (*-------------------------------------------------------------*)
              (* Set NETROM switches                                         *)
              (*-------------------------------------------------------------*)

              IF NOT tnc_null THEN
                BEGIN;
                  i := LENGTH(in_1.str_data);
                  netrom := (i > 3) AND
                                       (substr(in_1.str_data, i-1, 2) = 'CF');
                  IF netrom THEN
                    netrom_node := subword(@in_1.str_data, to_pos, 1) = 'NODES';
                END;

              (*-------------------------------------------------------------*)
              (* Write the info stuff out.  We do this now to ensure the     *)
              (* display is synchronized                                     *)
              (*-------------------------------------------------------------*)

              monitor_window_write(@in_1, TRUE, opt_block.opt_mon_time_stamp);

              (*-------------------------------------------------------------*)
              (* If null we are done                                         *)
              (*-------------------------------------------------------------*)

              IF tnc_null THEN
                EXIT;

              (*-------------------------------------------------------------*)
              (* Display the data                                            *)
              (*-------------------------------------------------------------*)

              IF netrom THEN
                netrom_monitor(@tnc_data, 1, netrom_node);
              monitor_window_write(@tnc_data, FALSE, FALSE);

              (*-------------------------------------------------------------*)
              (* If not right response type, give error message              *)
              (*-------------------------------------------------------------*)

              IF tnc_type <> t_to_h_mif THEN
                bad_mon_frame;

            END; (*----- End monitor frame with data frame following

          (*-----------------------------------------------------------------*)
          (* Invalid frame response                                          *)
          (*-----------------------------------------------------------------*)

          ELSE
            bad_mon_frame;

        END;

      END; (*----- End non-null response from TNC -----------------------*)

  END;

END.
