(*===========================================================================*)
(* Broadcasts                                                                *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

UNIT BBBCST;

INTERFACE

PROCEDURE send_any_broadcasts;

IMPLEMENTATION

  USES
    bbdummy,
    bbmdata,
    bbmess,
    bbmf,
    bbmisc,
    bbmisci,
    bbsdata,
    bbsess,
    bbsrt,
    bbstr,
    bbtask,
    bbtime,
    bbuser,
    bbwin;

  PROCEDURE send_any_broadcasts;

    VAR
      bcst_data    : STRING;
      save_max_pac : WORD;
      save_uf      : WORD;
      this_port    : port_block_ptr;

    (*=======================================================================*)
    (* This sends the broadcast                                              *)
    (*=======================================================================*)

    PROCEDURE send_bcst;

      VAR
        conn_str  : STRING;
        save_port : port_block_ptr;

      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Don't broadcast if not needed                                     *)
        (*-------------------------------------------------------------------*)

        IF NOT this_port^.port_bcst THEN EXIT;

        (*-------------------------------------------------------------------*)
        (* Save things                                                       *)
        (*-------------------------------------------------------------------*)

        save_port               := active_port;
        active_port             := this_port;
        active_tcb^.tcb_port    := this_port;
        active_tcb^.port_chan_s := this_port^.port_char + '0';

        (*-------------------------------------------------------------------*)
        (* If multiple ports overlayed then issue the UNPROTO path every time*)
        (*-------------------------------------------------------------------*)

        IF ((this_port^.rel_port <> NIL)
                            OR this_port^.port_sub_port)
                            AND (this_port^.port_type <> port_bpqhost)
                            AND (this_port^.bcst_path <> '') THEN
          BEGIN;
            conn_str := 'C ' + this_port^.bcst_path;
            cmd_tnc(@conn_str, TRUE);
          END;

        (*-------------------------------------------------------------------*)
        (* Send fixed message                                                *)
        (*-------------------------------------------------------------------*)

        send_message(message_bcst);

        (*-------------------------------------------------------------------*)
        (* If any variable data, send it                                     *)
        (*-------------------------------------------------------------------*)

        IF bcst_data <> '' THEN
          send_tnc_data_str(bcst_data);

        (*-------------------------------------------------------------------*)
        (* Flush the buffer                                                  *)
        (*-------------------------------------------------------------------*)

        send_flush;

        (*-------------------------------------------------------------------*)
        (* Restore things                                                    *)
        (*-------------------------------------------------------------------*)

        active_port             := save_port;
        active_tcb^.tcb_port    := save_port;
        active_tcb^.port_chan_s := save_port^.port_char + '0';

      END;

    (*=======================================================================*)
    (* This builds the broadcast                                             *)
    (*=======================================================================*)

    PROCEDURE build_bcst;

      VAR
        search_block : search_block_type;

      BEGIN;

        WITH search_block DO
          BEGIN;

            (*-----------------------------------------------------------------*)
            (* Initialize things                                               *)
            (*-----------------------------------------------------------------*)

            bcst_data := '';

            FILLCHAR(search_block, SIZEOF(search_block), 0);

            search_above  := TRUE;
            search_type   := 'N';
            search_msg_no := 0;
            search_mf     := mf_hold OR mf_read OR mf_fwd
                                                  OR mf_fwd_select OR mf_kill;
            search_last   := NIL;

            (*---------------------------------------------------------------*)
            (* Conduct first search                                          *)
            (*---------------------------------------------------------------*)

            search_msg(@search_block);

            (*---------------------------------------------------------------*)
            (* Loop while the search is working                              *)
            (*---------------------------------------------------------------*)

            WHILE (search_last <> NIL) AND (LENGTH(bcst_data) <= 230) DO

              WITH search_last^.msg_i_mb DO
                BEGIN;

                  (*---------------------------------------------------------*)
                  (* Switch tasks to allow breathing room                    *)
                  (*---------------------------------------------------------*)

                  task_switch;

                  (*---------------------------------------------------------*)
                  (* Add this call to the list if it is a local user and     *)
                  (* the call is not already there.   Don't put own call in  *)
                  (* list unless requested                                   *)
                  (*---------------------------------------------------------*)

                  IF ((msg_to_at = '') OR (msg_to_at = opt_block.this_bb_sign))
                        AND (opt_block.opt_own_in_bcst
                                         OR (msg_to <> opt_block.this_bb_sign))
                        AND (find(@bcst_data, @msg_to) = 0)
                        AND ((msg_flag AND mf_fwd_list) = 0) THEN
                    bcst_data := bcst_data + ' ' + msg_to;

                  (*---------------------------------------------------------*)
                  (* Conduct next search                                     *)
                  (*---------------------------------------------------------*)

                  search_msg(@search_block);

                END;

            (*---------------------------------------------------------------*)
            (* Search complete.  Add mail prefix                             *)
            (*---------------------------------------------------------------*)

             IF bcst_data <> '' THEN
               bcst_data := get_message(message_mail_for) + bcst_data + cr;

          END;

      END;

    (*=======================================================================*)
    (* Main line                                                             *)
    (*=======================================================================*)

    BEGIN;

      (*---------------------------------------------------------------------*)
      (* Build the broadcast                                                 *)
      (*---------------------------------------------------------------------*)

      build_bcst;

      (*---------------------------------------------------------------------*)
      (* Switch to non-transparent mode                                      *)
      (*---------------------------------------------------------------------*)

      save_uf    := active_tcb^.uid_data.user_flag;
      active_tcb^.uid_data.user_flag  :=
                         active_tcb^.uid_data.user_flag AND (NOT user_f_trans);

      (*---------------------------------------------------------------------*)
      (* Maximum packet size                                                 *)
      (*---------------------------------------------------------------------*)

      save_max_pac        := active_tcb^.max_pac;
      active_tcb^.max_pac := 255;

      (*---------------------------------------------------------------------*)
      (* Send broadcasts                                                     *)
      (*---------------------------------------------------------------------*)

      this_port := active_port;
      WHILE this_port <> NIL DO
        BEGIN;
          send_bcst;
          this_port := this_port^.rel_port;
        END;

      (*---------------------------------------------------------------------*)
      (* Restore things                                                      *)
      (*---------------------------------------------------------------------*)

      active_tcb^.max_pac            := save_max_pac;
      active_tcb^.uid_data.user_flag := save_uf;

    END;

END.
