(*===========================================================================*)
(* Handle bids                                                               *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

{$UNDEF  DEBUG_1}
{$DEFINE POINT_CHK}

UNIT BBBID;

INTERFACE

  USES
    bbdummy;

FUNCTION  bid_build_add (this_msg : msg_block_ptr; VAR dupe : BOOLEAN) :bid_str;
FUNCTION  bid_build_test(this_msg : msg_block_ptr) : BOOLEAN;
FUNCTION  bid_check : BOOLEAN;
FUNCTION  bid_add(t_bid : bid_str) : BOOLEAN;
PROCEDURE add_a_bid(t_bid : bid_str);
FUNCTION  bid_test(t_bid : bid_str) : BOOLEAN;

(*===========================================================================*)
(* Globals                                                                   *)
(*===========================================================================*)

(*---------------------------------------------------------------------------*)
(* Bid variables                                                             *)
(*---------------------------------------------------------------------------*)

TYPE
  bid_list_type    = ARRAY[1..5000] OF LONGINT;

VAR

  bid_next         : WORD;

  bid_file         : TEXT;

  bid_init_sw      : BOOLEAN;

  bid_list         : ^bid_list_type;

(*===========================================================================*)
(* Implementation                                                            *)
(*===========================================================================*)

IMPLEMENTATION

  USES
    bbbug,
    bbmisc,
    bbmisc3,
    bbsema2,
    bbstr,
    bbtask,
    bbtime,
    bbtrace;

(*===========================================================================*)
(* Build a 32-bit CRC                                                        *)
(*===========================================================================*)

FUNCTION build_crc_32(bin : BYTE; crc : LONGINT) : LONGINT;

  VAR j : BYTE;

  BEGIN;

    crc := (LONGINT(bin) SHL 24) XOR crc;

    FOR j := 1 TO 8 DO
      IF (crc AND $80000000) <> 0 THEN
        crc := (($7FFFFFFF AND crc) SHL 1) XOR $04C11DB7
      ELSE
        crc := ($7FFFFFFF AND crc) SHL 1;

    build_crc_32 := crc;

  END;

(*===========================================================================*)
(* Build a 32-bit CRC of a BID                                               *)
(*===========================================================================*)

FUNCTION make_bid_crc (s : bid_str) : LONGINT;

  VAR i   : BYTE;
      crc : LONGINT;

  BEGIN;

    crc := $FFFFFFFF;
    i := 0;

    WHILE i < LENGTH(s) DO
      BEGIN;
        INC(i);
        IF s[i] <> ' ' THEN
          crc := build_crc_32(ORD(s[i]), crc);
      END;

    make_bid_crc := crc;

  END;

(*===========================================================================*)
(* Add a bid to the table and discard oldest if too many                     *)
(*     Make sure it is not a dupe                                            *)
(*===========================================================================*)

FUNCTION bid_add(t_bid : bid_str) : BOOLEAN;

  VAR
    b : BOOLEAN;
    i : BYTE;

  BEGIN;

    b := bid_test(t_bid);

    bid_add := b;

    IF NOT b THEN
      add_a_bid(t_bid);

  END;

(*===========================================================================*)
(* Add a bid to the table and discard oldest if too many                     *)
(*===========================================================================*)

PROCEDURE add_a_bid(t_bid : bid_str);

  VAR
    my_sema : BOOLEAN;

  BEGIN;

    IF bid_next < opt_block.n_bid THEN
      INC(bid_next)
    ELSE
      bid_next := 1;

    {$IFDEF POINT_CHK}
      test_pointer(bid_list);
    {$ENDIF}

    bid_list^[bid_next] := make_bid_crc(t_bid);

    {$IFDEF DEBUG_1}
      trace_data('Bid add ', bid_next , NIL, t_bid);
      trace_data('Bid add ', bid_next , bid_list, l2c(bid_list^[bid_next]));
    {$ENDIF}

    IF NOT bid_init_sw THEN
      BEGIN;

        my_sema := NOT test_semaphore_tcb(semaphore_interrupts, sem_exclusive);
        IF my_sema THEN
          get_semaphore(semaphore_interrupts, sem_exclusive, FALSE);

        APPEND(bid_file);
        WRITELN(bid_file, todays_date_time, t_bid);
        CLOSE(bid_file);

        IF my_sema THEN
          free_semaphore(semaphore_interrupts);

      END;

  END;

(*===========================================================================*)
(* Build a bid and add it to the system                                      *)
(*===========================================================================*)

FUNCTION bid_build_add (this_msg : msg_block_ptr; VAR dupe : BOOLEAN) :bid_str;

  VAR
    m        : WORD;
    t_bid    : bid_str;
    use_call : call_sign_str;

  BEGIN;

    WITH this_msg^ DO
      BEGIN;

        IF (LENGTH(msg_bid) = 1) AND (msg_bid[1] <= CHR(1)) THEN
          BEGIN;
            m        := msg_number;
            use_call := opt_block.this_bb_bid;
          END
        ELSE
          BEGIN;
            m        := msg_no_orig;
            use_call := strip_ssid(msg_from_at);
            IF use_call = opt_block.this_bb_sign THEN
              use_call := opt_block.this_bb_bid;
          END;

        STR(m, t_bid);

        t_bid := t_bid + '_' + use_call;

        bid_build_add := t_bid;

        IF use_call = opt_block.this_bb_bid THEN
          BEGIN;
            dupe := FALSE;
            EXIT;
          END;

      END;

    dupe := bid_test(t_bid);
    IF dupe THEN
      EXIT;

    add_a_bid(t_bid);

  END;

(*===========================================================================*)
(* Build a bid and test it                                                   *)
(*===========================================================================*)

FUNCTION bid_build_test (this_msg : msg_block_ptr) : BOOLEAN;

  VAR
    t_bid    : bid_str;

  BEGIN;

    WITH this_msg^ DO
      BEGIN;
        STR(msg_no_orig, t_bid);
        t_bid := t_bid + '_' + strip_ssid(msg_from_at);

        IF t_bid = msg_bid THEN
          BEGIN;
            bid_build_test := FALSE;
            EXIT;
          END;

      END;

    bid_build_test := bid_test(t_bid);

  END;

(*===========================================================================*)
(* Test a bid                                                                *)
(*===========================================================================*)

FUNCTION bid_test(t_bid : bid_str) : BOOLEAN;

  VAR
    i : WORD;
    l : LONGINT;

  BEGIN;

    l := make_bid_crc(t_bid);

    {$IFDEF POINT_CHK}
      test_pointer(bid_list);
    {$ENDIF}

    FOR i := 1 TO opt_block.n_bid DO
      BEGIN;
        IF bid_list^[i] = l THEN
          BEGIN;
            bid_test := TRUE;
            EXIT;
          END;
        IF (i AND $3F) = 0 THEN
          task_switch;
      END;

    bid_test := FALSE;

  END;

(*===========================================================================*)
(* Test a bid in a message                                                   *)
(*===========================================================================*)

FUNCTION bid_check : BOOLEAN;

  VAR
    i  : BYTE;

  BEGIN;

    WITH active_tcb^.curr_msg.msg_i_mb DO
      BEGIN;

        IF msg_bid = '' THEN
          BEGIN;
            bid_check := FALSE;
            EXIT;
          END;

        i := LENGTH(msg_bid);
        WHILE (i > 0) AND (msg_bid[i] <> '_') DO
          DEC(i);

        IF (i > 0) AND (opt_block.this_bb_bid = substr(msg_bid, i+1, 0)) THEN
          BEGIN;
            bid_check := TRUE;
            EXIT;
          END;

        bid_check := bid_test(msg_bid);

      END;

  END;

END.
