(*===========================================================================*)
(* Bid auxiliary routines                                                    *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

{$UNDEF  DEBUG_1}
{$UNDEF  DEBUG_2}
{$DEFINE POINT_CHK}

UNIT BBBIDAUX;

INTERFACE

PROCEDURE bid_init;
PROCEDURE bid_clean;

IMPLEMENTATION

  USES
    CRT,
    DOS,
    bbbid,
    bbbug,
    bbdummy,
    bbdump,
    bbmisc,
    bbmisc3,
    bbsdata,
    bbsema2,
    bbstr,
    bbtime;

(*===========================================================================*)
(* Open bid file and read.                                                   *)
(*===========================================================================*)

PROCEDURE bid_init;

  VAR
    bid_buf  : POINTER;
    bid_size : LONGINT;
    i        : BYTE;
    t_str    : STRING;
    add_this : bid_str;

  BEGIN;

    {$IFDEF DEBUG_1}
      WRITELN('Bid start -- ', opt_block.n_bid);
      DELAY(1000);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Initialize                                                            *)
    (*-----------------------------------------------------------------------*)

    bid_next := 0;

    bid_size := LONGINT(opt_block.n_bid) * SIZEOF(LONGINT);

    GETMEM(bid_list, bid_size);

    {$IFDEF POINT_CHK}
      test_pointer(bid_list);
    {$ENDIF}

    FILLCHAR(bid_list^,  bid_size, 0);

    {$IFDEF DEBUG_1}
      WRITELN('Bid size = ', bid_size);
      WRITELN('Bid ptr  = ', p2x(bid_list));
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Obtain the interrupt lock                                             *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG_1}
      WRITELN('Lock');
      DELAY(1000);
    {$ENDIF}

    get_semaphore(semaphore_interrupts, sem_exclusive, FALSE);

    (*-----------------------------------------------------------------------*)
    (* Get the name                                                          *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG_1}
      WRITELN('Assign');
      DELAY(1000);
    {$ENDIF}

    ASSIGN(bid_file, opt_block.bid_fn);

    (*-----------------------------------------------------------------------*)
    (* Open the file                                                         *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG_1}
      WRITELN('Open');
      DELAY(1000);
    {$ENDIF}

    {$I-}
    RESET(bid_file);
    {$I+}
    i := IORESULT;

    IF i <> 0 THEN
      BEGIN;
        REWRITE(bid_file);
        CLOSE(bid_file);
        RESET(bid_file);
      END;

    (*-----------------------------------------------------------------------*)
    (* Free the interrupt lock                                               *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG_1}
      WRITELN('Free');
      DELAY(1000);
    {$ENDIF}

    free_semaphore(semaphore_interrupts);

    (*-----------------------------------------------------------------------*)
    (* Loop thru the file                                                    *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG_1}
      WRITELN('Loop');
      DELAY(1000);
    {$ENDIF}

    bid_init_sw := TRUE;

    WHILE NOT EOF(bid_file) DO
      BEGIN;

        READLN(bid_file, t_str);
        add_this := COPY(t_str, 12, bid_len);

        IF add_this <> '' THEN
          add_a_bid(add_this);

      END;

    bid_init_sw := FALSE;

    (*-----------------------------------------------------------------------*)
    (* Close things                                                          *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG_1}
      WRITELN('Close -- ', bid_next);
      DELAY(1000);
    {$ENDIF}

    CLOSE(bid_file);

    (*-----------------------------------------------------------------------*)
    (* All done                                                              *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG_1}
      WRITELN('Done');
      DELAY(1000);
    {$ENDIF}

  END;

(*===========================================================================*)
(* Clean out bid file                                                        *)
(*===========================================================================*)

PROCEDURE bid_clean;

  VAR
    bid_data  : STRING;
    bid_out   : TEXT;
    found     : BYTE;
    i         : INTEGER;
    temp_str  : STRING[8];
    test_this : bid_str;


  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Tell user what is happening                                           *)
    (*-----------------------------------------------------------------------*)

    send_tnc_data_str('Bid cleanup in progress' + cr);

    (*-----------------------------------------------------------------------*)
    (* Obtain the interrupt lock                                             *)
    (*-----------------------------------------------------------------------*)

    get_semaphore(semaphore_interrupts, sem_exclusive, TRUE);

    (*-----------------------------------------------------------------------*)
    (* Open input file                                                       *)
    (*-----------------------------------------------------------------------*)

    ASSIGN(bid_file, opt_block.bid_fn);

    {$I-}
    RESET(bid_file);
    {$I+}
    i := IORESULT;

    IF i <> 0 THEN
      BEGIN;
        REWRITE(bid_file);
        CLOSE(bid_file);
        RESET(bid_file);
      END;

    (*-----------------------------------------------------------------------*)
    (* Build name of output file                                             *)
    (*-----------------------------------------------------------------------*)

    FSPLIT(opt_block.bid_fn, bid_data, temp_str, test_this);
    bid_data := bid_data + 'BIDCLEAN.TMP';

    (*-----------------------------------------------------------------------*)
    (* Open output file                                                      *)
    (*-----------------------------------------------------------------------*)

    ASSIGN(bid_out, bid_data);

    REWRITE(bid_out);

    {$IFDEF DEBUG_2}
      WRITELN('Bid clean start -- ', opt_block.n_bid, ' / ', bid_next);
      DELAY(1000);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Loop thru input file                                                  *)
    (*-----------------------------------------------------------------------*)

    found := 0;

    WHILE NOT EOF(bid_file) DO
      BEGIN;

        READLN(bid_file, bid_data);

        IF found < 10 THEN
          BEGIN;
            test_this := COPY(bid_data, 12, bid_len);
            IF NOT bid_test(test_this) THEN
              found := 0
            ELSE
              BEGIN;
                INC(found);
                WRITELN(bid_out, bid_data);

                {$IFDEF DEBUG_2}
                  WRITELN('Bid clean found -- ', bid_data);
                  DELAY(1000);
                {$ENDIF}
              END;
          END
        ELSE
          WRITELN(bid_out, bid_data);

      END;

    (*-----------------------------------------------------------------------*)
    (* Close files                                                           *)
    (*-----------------------------------------------------------------------*)

    CLOSE(bid_file);
    CLOSE(bid_out);

    (*-----------------------------------------------------------------------*)
    (* Erase old file and rename new file to proper name                     *)
    (*-----------------------------------------------------------------------*)

    ERASE(bid_file);
    RENAME(bid_out, opt_block.bid_fn);

    (*-----------------------------------------------------------------------*)
    (* Release the interrupt lock                                            *)
    (*-----------------------------------------------------------------------*)

    free_semaphore(semaphore_interrupts);

    (*-----------------------------------------------------------------------*)
    (* Tell user what is happening                                           *)
    (*-----------------------------------------------------------------------*)

    send_tnc_data_str('Bid cleanup complete' + cr);

  END;

END.
