(*===========================================================================*)
(* Control Block -- Msgs                                                     *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Msg Blocks - Contains all the information about the msgs                  *)
(*===========================================================================*)

CONST
     mt_private = 'P';                  (* Message is private                *)
     mt_nts     = 'T';                  (* Message is NTS Traffic            *)

     msg_dist_max = 20;                 (* Max number of distribution list   *)
     msg_path_max_msgs = 254;           (* Max messages per path.  Don't     *)
                                        (* exceed 254.                       *)
     msg_path_max_route = 30;           (* Max routes per path.  Don't exceed*)
                                        (* 254                               *)

(*---------------------------------------------------------------------------*)
(* These define bits in msg_d_flag                                           *)
(*---------------------------------------------------------------------------*)

CONST
  df_fwd         = $01;         (* Message has been forwarded                *)
  df_fwd_select  = $02;         (* Message has been selected for forward     *)
  df_fwd_process = $04;         (* Forward in process now                    *)
  df_fwd_reject  = $08;         (* Bid was rejected                          *)
  df_fwd_cancel  = $10;         (* Forward was cancelled by operator         *)
  df_fwd_unknown = $20;         (* Unknown route                             *)
  df_fwd_spare1  = $40;         (* Spare                                     *)
  df_fwd_spare2  = $80;         (* Spare                                     *)

(*---------------------------------------------------------------------------*)
(* These define bits in msg_flag                                             *)
(*---------------------------------------------------------------------------*)

CONST

  mf_hold         = $0001;      (* Message is in hold                        *)
  mf_read         = $0002;      (* Message has been read                     *)
  mf_fwd          = $0004;      (* Message has been forwarded                *)
  mf_fwd_select   = $0008;      (* Message has been selected for fwd         *)
  mf_fwd_list     = $0010;      (* Distribution list attached                *)
  mf_old          = $0020;      (* Message is old                            *)
  mf_kill         = $0040;      (* Message has been killed                   *)
  mf_fwd_process  = $0080;      (* Forward in process now                    *)
  mf_review       = $0100;      (* Message for review by SYSOP               *)
  mf_h_receive    = $0200;      (* Hierarchial route received                *)
  mf_landline     = $0400;      (* Msg received from landline                *)
  mf_bid_override = $0800;      (* No bid override allowed                   *)
  mf_bid_change   = $1000;      (* Bid has been changed                      *)
  mf_archive      = $2000;      (* Msg has been archived                     *)
  mf_disrout      = $4000;      (* Distribution routing block present        *)
  mf_unknown      = $8000;      (* Routed to ?                               *)

(*---------------------------------------------------------------------------*)
(* Types                                                                     *)
(*---------------------------------------------------------------------------*)

TYPE
    msg_flag_type = WORD;

    msg_block_ptr = ^msg_block;

    msg_d_ptr     = ^msg_dist_block_type;

    msg_dr_ptr    = ^msg_dr_route_array;

    msg_index_ptr = ^msg_index_block;

    msg_p_ptr     = ^msg_path_block;

    msg_pi_ptr    = ^msg_path_array_item;

    msg_r_ptr     = ^msg_route_block_type;

    (* ----- This is the individual distribution list entries ------------- *)


    msg_dist_entry_type = RECORD
              msg_d_flag    : BYTE;         (* Flags for this item          *)
              msg_d_info    : bb_addr_str;  (* Where to forward to          *)
            END;

    msg_dist_block_type = RECORD
              msg_d_no      : BYTE;         (* No of items                  *)
              msg_d_array   : ARRAY[1..msg_dist_max] OF msg_dist_entry_type;
            END;

    (* ---- This holds the routing information for a distribution block --- *)
    (* It is inserted in the chain from the message block to the            *)
    (* distribution block.  Presence is indicated by mf_disrout             *)

    msg_dr_route_item  = msg_r_ptr;

    msg_dr_route_array = RECORD
              msg_dr_dblk : msg_d_ptr;
              msg_dr_data : ARRAY[1..msg_dist_max] OF msg_dr_route_item;
            END;

    (* ----- This is the file format -------------------------------------- *)


    msg_block = RECORD
              msg_number  : WORD;           (* This msg number               *)
              msg_type    : CHAR;           (* Type of msg -- list is below  *)
              msg_flag    : msg_flag_type;  (* Flags                         *)
              msg_size    : LONGINT;        (* Message size                  *)
              msg_to      : call_sign_str;  (* To addressee                  *)
              msg_to_at   : bb_addr_str;    (* To mailbox address            *)
              msg_to_h    : h_addr_str;     (* To hierarchial route          *)
              msg_from    : call_sign_str;  (* From                          *)
              msg_from_at : bb_addr_str;    (* From mailbox                  *)
              msg_from_h  : h_addr_str;     (* From hierarchial route        *)
              msg_dt_in   : LONGINT;        (* Message's date time at here   *)
              msg_dt_orig : LONGINT;        (* Message's date time at origin *)
              msg_no_orig : WORD;           (* Message number at origin      *)
              msg_tread   : BYTE;           (* Times read (except sysops)    *)
              msg_bid     : bid_str;        (* Message bid                   *)
              msg_r_call  : call_sign_str;  (* Call sign of reject station   *)
              msg_reason  : BYTE;           (* Reason for hold/review        *)
              msg_subj    : subj_str;       (* Subject                       *)
            END;

    (* ----- This is the in-storage format --------------------------------- *)


    msg_index_block = RECORD

              msg_i_next   : msg_index_ptr; (* Points to next msg index      *)
              msg_i_last   : msg_index_ptr; (* Points to previous msg index  *)
              msg_i_fwd_l  : msg_index_ptr; (* Points to next msg in fwd lst *)
              msg_i_fnm_l  : msg_index_ptr; (* Points to next msg in bnm lst *)
              msg_i_record : WORD;          (* Record number of message      *)
              CASE BYTE OF
                0 : (msg_i_dis : msg_d_ptr);  (* Ptr to distribution block   *)
                1 : (msg_i_dr  : msg_dr_ptr); (* Ptr to dist. route block    *)
                2 : (msg_i_rou : msg_r_ptr;   (* Ptr to route                *)
                     msg_i_mb  : msg_block);  (* Rest of message block       *)

            END;

    (*----- One item of this type for route that has messages ---------------*)


    msg_route_block_type = RECORD
              msg_r_next   : msg_r_ptr;     (* Pointer to next route block   *)
              msg_r_routes : BYTE;          (* Number of possible routes     *)
              msg_r_nroute : BYTE;          (* Next route to try             *)
              msg_r_scan   : BOOLEAN;       (* This route was scanned by fwd *)
              msg_r_info   : STRING;        (* The route statement itself    *)
            END;

    (*----- One item of this type for each path that has messages -----------*)
    (* This is variable length                                               *)

    msg_path_array_item = RECORD
              msg_p_i     : msg_index_ptr;
              msg_p_item  : BYTE;
            END;

    msg_path_block = ARRAY[1..msg_path_max_msgs] OF msg_path_array_item;

