(*===========================================================================*)
(* Control Blocks -- Port and Thread                                         *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

TYPE
  port_block_ptr  = ^port_block_type;

  tcb_ptr = ^tcb;

(*---------------------------------------------------------------------------*)
(* Port Block                                                                *)
(*---------------------------------------------------------------------------*)

    port_call_item = RECORD
                       port_call_date : STRING[11];
                       port_call_port : CHAR;
                       port_call_sign : bb_addr_str;
                     END;

    port_call_array = ARRAY[1..500] OF port_call_item;

    port_call_ptr   = ^port_call_array;

    port_connect_array = ARRAY[0..max_connect_port] OF tcb_ptr;

    port_connect_ptr   = ^port_connect_array;

    port_type_type  = (port_dedhost1,    (* DED host mode -- TNC1            *)
                       port_dedhost2,    (* DED host mode -- TNC2            *)
                       port_dedhost87,   (* DED host mode -- PK-87           *)
                       port_bpqhost,     (* DED host mode -- BPQ switch      *)
                       port_pcpa,        (* DRSI PC*PA                       *)
                       port_aeapk232,    (* AEA PK-232 mode                  *)
                       port_g8bpq,       (* G8BPQ switch                     *)
                       port_pc1xx,       (* Paccom PC-1xx                    *)
                       port_modem,       (* Modem with AT command set        *)
                       port_null_modem,  (* Cross connect to another system  *)
                       port_serial,      (* Dumb serial port                 *)
                       port_console,     (* Local terminal                   *)
                       port_unknown);    (* Not set                          *)

    port_block_type = RECORD
                        next_port   : port_block_ptr;   (* -> next block     *)
                        main_port   : port_block_ptr;   (* -> main port      *)
                        rel_port    : port_block_ptr;   (* -> related port   *)
                        aux_thread  : tcb_ptr;          (* Monitor thread    *)
                        com_number  : BYTE;             (* COMx number       *)
                        port_char   : CHAR;             (* Character for port*)
                        port_name   : STRING[20];       (* Name of port      *)
                        port_type   : port_type_type;   (* Port type         *)

                        port_host_only      : BOOLEAN;  (* Host mode only tnc*)
                        port_sub_port       : BOOLEAN;  (* Sub port (PC*PA)  *)
                        port_monitor        : BOOLEAN;  (* Monitor this port?*)
                        port_r_sysop        : BOOLEAN;  (* Remote sysop ok   *)
                        port_up_down        : BOOLEAN;  (* Upload/Download ok*)
                        port_bcst           : BOOLEAN;  (* Bcst on this port *)
                        port_pk232_data_ack : BOOLEAN;  (* PK232 dat ack wait*)
                        port_no_out_fwd     : BOOLEAN;  (* No outgoing fwd   *)
                        port_no_busy_fwd    : BOOLEAN;  (* No outgoing fwd   *)
                                                        (* when busy         *)
                        port_no_binary      : BOOLEAN;  (* No binary allowed *)
                        port_suppress_ssid  : BOOLEAN;  (* Suppress SSID addr*)
                        port_use_user_chan  : BOOLEAN;  (* User user channel *)
                                                        (* for forwarding    *)
                        port_force_register : BOOLEAN;  (* Force registration*)
                        port_bpq_port_detect: BOOLEAN;  (* Detect BPQ port   *)
                        port_dflt_trans     : BOOLEAN;  (* Default transpart *)

                        port_operate_mode : op_mode;    (* Operation mode    *)
                        port_num    : CHAR;             (* PCPA port #       *)
                        port_color  : BYTE;             (* Display color     *)
                        port_sema   : BYTE;             (* Semaphore number  *)
                        port_no_mail_bcst : BYTE;       (* Bdcst with no mail*)
                        port_no_mail_cnt  : BYTE;       (* Count of no mail  *)
                        port_allow    : user_class_type; (* Allow this class *)
                        port_restrict : user_class_type; (* Restrict class   *)
                        dflt_pac    : WORD;             (* Default packet siz*)
                        max_pac     : WORD;             (* Max packet size   *)
                        data_rate   : WORD;             (* Data rate to use  *)
                        time_out    : WORD;             (* Time out in second*)
                        fwd_min     : WORD;             (* Forward minute    *)
                        sked_fwd    : LONGINT;          (* Next time to fwd  *)
                        max_conn    : BYTE;             (* Num of connects   *)
                        max_chan    : BYTE;             (* Max capacity TNC  *)
                        port_pend   : BYTE;             (* Maximum port pend *)
                        dflt_scrl   : BYTE;             (* Default screen len*)
                        new_display : BYTE;             (* New user disp fmt *)
                        reject_act  : BYTE;             (* Reject action     *)
                        reject_bul  : BYTE;             (* Reject bultns act *)
                        mon_filter  : BYTE;             (* Monitor filter    *)
                        u_restrict  : BYTE;             (* User restrictions *)
                        bpq_appl    : BYTE;             (* BPQ appl number   *)
                        disc_delay  : BYTE;             (* Disconnect delay  *)
                        dflt_lang   : CHAR;             (* Default language  *)
                        dflt_access : access_block_type;(* Default access    *)
                        first_load  : file_name_str;    (* File to load      *)
                        dflt_order  : STRING[3];        (* Default fwd order *)
                        connected   : port_connect_ptr; (* List of connects  *)
                        call_list   : port_call_ptr;    (* list of calls hrd *)
                        CASE BYTE OF

                          0 :                           (* All ports but modm*)
                              (call_set  : STRING[9];   (* Set this call     *)
                               bcst_path : STRING[45]); (* Bcst this path    *)

                          1:                            (* Modem ports       *)
                              (modem_optns : BYTE;      (* Signify modem opt *)
                               answer_ring : BYTE;      (* Answer on this rin*)
                               cr_timeout  : BYTE;      (* CR time out       *)
                               cur_rate    : WORD;      (* Current data rate *)
                               modem_echo  : BOOLEAN;   (* Echo on modem     *)
                               modem_e_now : BOOLEAN;   (* Echo on modem NOW *)
                               modem_crlf  : BOOLEAN;   (* CR=CR/LF for modem*)
                               modem_dial  : BOOLEAN;   (* Modem dialing     *)
                               modem_conn  : BOOLEAN;   (* Modem connect msg *)
                               modem_freez : BOOLEAN;   (* Freeze modem speed*)
                               modem_reg   : BYTE;      (* Allow register    *)
                               port_modem_dcd : BOOLEAN); (* DCD on modem    *)

                      END;

(*---------------------------------------------------------------------------*)
(* Thread Block                                                              *)
(*---------------------------------------------------------------------------*)

     thread_type = (th_dead,                  (* Dead thread                 *)
                    th_main,                  (* Main thread                 *)
                    th_operator,              (* Operator thread             *)
                    th_fwd_control,           (* Forward control thread      *)
                    th_aux,                   (* Auxiliary (monitor) thread  *)
                    th_fwd_slave,             (* Forward slave thread        *)
                    th_opr_terminal,          (* Operator terminal thread    *)
                    th_user,                  (* User thread                 *)
                    th_answer);               (* Answer modem thread         *)

     tcb = RECORD

           (* Task information                                               *)
             next_tcb    : tcb_ptr;     (* Pointer to next TCB in chain      *)
             tcb_number  : BYTE;        (* TCB number                        *)
             window      : BYTE;        (* Window this TCB corresponds to    *)
             w_color     : BYTE;        (* Color to use                      *)
             tcb_name    : bb_addr_str; (* Name for tcb to show in window    *)
             tcb_type    : thread_type; (* Thread type                       *)

           (* Who am I attached to                                           *)
             tcb_port    : port_block_ptr;    (* Port info                   *)
             channel     : BYTE;        (* Channel for this task             *)
             port_chan_s : STRING[2];   (* Port/channel in character format  *)
             max_pac     : WORD;        (* Maximum packet len                *)

           (* User information                                               *)
             uid_data    : user_record_type; (* User's data                  *)
             last_l_time : LONGINT;     (* Time "L" was done in this session *)

           (* State information                                              *)
           (*      The first variable must be tcb_transmit_idle              *)
           (*      and the last one is bid_level.  Everything inbetween is   *)
           (*      zeroed during initialization and other places             *)
             tcb_transmit_idle    : BOOLEAN; (* Transmitter is idle          *)
             tcb_console          : BOOLEAN; (* Redirect TNC to console      *)
             tcb_connect          : BOOLEAN; (* This TCB is connected        *)
             tcb_abbs             : BOOLEAN; (* Advanced BBS mode on         *)
             tcb_mids_ok          : BOOLEAN; (* MIDS are ok                  *)
             tcb_h_ok             : BOOLEAN; (* Hierarchial address ok       *)
             tcb_ignore_lc        : BOOLEAN; (* Ignore all the LC            *)
             tcb_opr_talk         : BOOLEAN; (* Talking to operator          *)
             tcb_ignore_port_chan : BOOLEAN; (* No port/channel dedicatd     *)
             tcb_never_kill       : BOOLEAN; (* Never kill this task         *)
             tcb_dead_in_progress : BOOLEAN; (* This tcb is being killed     *)
             tcb_dead             : BOOLEAN; (* This tcb is killed           *)
             tcb_stop_tnc_io      : BOOLEAN; (* Stop TNC I/O                 *)
             tcb_make_cc          : BOOLEAN; (* Make multiple msg is act     *)
             tcb_no_show_sdata    : BOOLEAN; (* Don't show sending data      *)
             tcb_binary           : BOOLEAN; (* This thread is doing binary  *)
             tcb_rcv_msg          : BOOLEAN; (* This thread is receiving msg *)
             tcb_rev_fwd          : BOOLEAN; (* Reverse forward in progress  *)
             tcb_sysop_pw_ok      : BOOLEAN; (* A sysop password was valid   *)
             tcb_access_ok        : BOOLEAN; (* User access is ok            *)
             tcb_no_overlay       : BOOLEAN; (* Turn off overlay save/restore*)
             tcb_no_search_rel    : BOOLEAN; (* Don't release search memory  *)
             error_sw             : BOOLEAN; (* Error occurred on the command*)

           (* error reason                                                   *)
             tcb_error_reason     : BYTE;    (* Reason error occurred        *)

           (* option info                                                    *)
             tcb_bid_level        : 0..2;    (* Bid levels supported         *)
                                             (*   0 = no bids                *)
                                             (*   1 = bid but just NO/OK     *)
                                             (*   2 = bid with OK/NO/LATE/REJ*)

           (* Swapping info                                                  *)
             sseg_value  : WORD;        (* SS register value                 *)
             sptr_value  : WORD;        (* SP register value                 *)
             bptr_value  : WORD;        (* BP register value                 *)
             sseg_init   : WORD;        (* SS register initial setting       *)
             sseg_bot    : WORD;        (* Offset of segment allocation      *)
             sptr_init   : WORD;        (* SP register initial setting       *)
             sseg_size   : WORD;        (* Size of stack                     *)

           (* Swapping info                                                  *)
             stack_usage : WORD;        (* Stack usage                       *)
             stack_temp  : ARRAY[1..10] OF WORD;
             stack_cnt   : BYTE;

           (* Overlay control                                                *)
             tcb_ovr_cnt : BYTE;        (* Number of overlay items in stack  *)

           (* Access control area                                            *)
             tcb_access_mode : access_block_type; (* User access is ok       *)

           (* TNC buffers / interactions                                     *)
             tnc_b_size  : WORD;        (* Size of the following buffer      *)

             tnc_htt     : ^host_to_tnc;(* These actual point to one buffer  *)
             tnc_tth     : ^tnc_to_host;(* used for I/O to the tnc.          *)

             tnc_type    : BYTE;        (* Type of response from TNC         *)
             tnc_data    : str_mixed;   (* Data buffer from/to tnc           *)

             tnc_in_chn  : str_m_chain;  (* Chain of waiting data. PK232 and *)
                                         (* modem only                       *)

             tnc_null    : BOOLEAN;     (* Null response received            *)

           (* Files in use                                                   *)
             io_fe       : fe_ptr;      (* File element for I/O              *)
             path_fe     : fe_ptr;      (* File element for path tables      *)

           (* Msg information                                                *)
             curr_msg    : msg_index_block;  (* Block for current msg        *)
             curr_fwd    : msg_path_array_item; (* Forward msg pointer       *)

           (* Buffers                                                        *)
             i_data      : str_mixed;   (* Temporary holding spot for input  *)
             o_data      : str_mixed;   (* Temporary holding spot for output *)

           (* Converse into                                                  *)
             c_input     : str_m_chain; (* Incoming converse packets         *)
             conv_tcb    : tcb_ptr;     (* Converse TCB                      *)

           (* Storage used                                                   *)
             stor_list   : POINTER;     (* Pointer to storage list           *)

           (* Output line counter                                            *)
             out_force   : BOOLEAN;     (* Force output counting             *)
             out_line    : BYTE;        (* Counts output lines for more      *)
             out_char    : BYTE;        (* Counts output chars for more      *)

           END;
