(*===========================================================================*)
(* Control Block -- Users                                                    *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991 by H. Roy Engehausen.  All rights      *)
(*   reserved.                                                               *)
(*                                                                           *)
(*===========================================================================*)

(*---------------------------------------------------------------------------*)
(* User block                                                                *)
(*---------------------------------------------------------------------------*)

CONST
    user_class_string : STRING[7] = 'NUOEBRL';

(*---------------------------------------------------------------------------*)
(* These define bits in user_flag                                            *)
(*---------------------------------------------------------------------------*)

CONST

  user_f_trans     = $0001;     (* User is transparent                       *)
  user_f_bbs       = $0002;     (* User is a BBS                             *)
  user_f_abbs      = $0004;     (* User is an advanced BBS                   *)
  user_f_sysop     = $0008;     (* User can be a remote sysop                *)
  user_f_adrchg    = $0010;     (* User has updated his home BBS             *)
  user_f_exclude   = $0020;     (* User is excluded                          *)
  user_f_delete    = $0040;     (* User record is empty                      *)
  user_f_local     = $0080;     (* User is a local                           *)
  user_f_restrict  = $0100;     (* User is restricted                        *)
  user_f_reg_modem = $0200;     (* User is registering via modem             *)
  user_f_spare1    = $0400;     (* Not used                                  *)
  user_f_spare2    = $0800;     (* Not used                                  *)
  user_f_spare3    = $1000;     (* Not used                                  *)
  user_f_spare4    = $2000;     (* Not used                                  *)
  user_f_emerg     = $4000;     (* User is allowed on during emergencies     *)
  user_f_pbbs      = $8000;     (* User is a personal BBS                    *)

(*---------------------------------------------------------------------------*)
(* User class type                                                           *)
(*---------------------------------------------------------------------------*)

TYPE
    user_class_type =
                     (user_c_nu,         (* New user                         *)
                      user_c_uu,         (* Unregistered user                *)
                      user_c_ou,         (* Old user                         *)
                      user_c_eu,         (* Expert user                      *)
                      user_c_bu,         (* BBS user                         *)
                      user_c_rsu,        (* Remote sysop user                *)
                      user_c_lsu);       (* Local sysop user                 *)

(*---------------------------------------------------------------------------*)
(* User index block                                                          *)
(*---------------------------------------------------------------------------*)

    user_index_ptr = ^user_index_block;

    user_index_block = RECORD
                         user_next   : user_index_ptr;  (* Ptr to next block *)
                         user_id     : call_sign_str;   (* Call sign of user *)
                         user_recno  : WORD;            (* Rec # for user    *)
                       END;

(*---------------------------------------------------------------------------*)
(* User data                                                                 *)
(*---------------------------------------------------------------------------*)

    user_record_ptr = ^user_record_type;

    user_record_type = RECORD
                         user_i_ptr  : user_index_ptr;  (* Pointer to index  *)
                         user_id     : call_sign_str;   (* Call sign of user *)
                         user_ssid   : ssid_sign_str;   (* SSID of user      *)
                         user_name   : STRING[20];      (* User's name       *)
                         user_bbs    : bb_addr_str;     (* Mail address      *)
                         user_zip    : STRING[10];      (* Zip code          *)
                         user_port   : CHAR;            (* User's port       *)
                         user_class  : user_class_type; (* User class        *)
                         user_flag   : WORD;            (* User flags        *)
                         user_fmt    : BYTE;            (* Format of "L"     *)
                         max_pac     : WORD;            (* Maximum packet siz*)
                         user_last   : LONGINT;         (* Last time on      *)
                         user_l_time : LONGINT;         (* Last time "L" done*)
                         user_l_cnt  : WORD;            (* Number of logons  *)
                         user_n_time : LONGINT;         (* Last time "N" done*)
                         user_pw     : password_str;    (* Password          *)
                         user_sysop_p: BYTE;            (* SYSOP password #  *)
                         user_lang   : CHAR;            (* User language     *)
                         user_scr_len: BYTE;            (* Length of user's  *)
                                                        (*      screen       *)
                         user_scr_wid: BYTE;            (* Width of user's   *)
                                                        (*      screen       *)
                         user_access : access_block_type;(* User access  lvl *)
                         padding     : ARRAY[1..4] OF BYTE;
                       END;
