(*===========================================================================*)
(* Look up ham name & address via SAMAPI                                     *)
(*                                                                           *)
(*   Copyright 1991, by Tom Thompson.                                        *)
(*   Copyright 1991 by H. Roy Engehausen.  All rights reserved.              *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

UNIT BBDB1;

INTERFACE

  PROCEDURE sam_ulookup_call(cmd_string : STRING);

IMPLEMENTATION

  USES
    DOS,
    bbdummy,
    bbmdata,
    bbmess,
    bbmisc,
    bbsdata,
    bbstr;

  {$I SAMAPI.PAS}

  VAR
   sammuxid : BYTE;

(*===========================================================================*)
(* locatesam                                                                 *)
(*                                                                           *)
(* Finds the resident SAMAPI code.  This MUST be called before other         *)
(* functions are used!.  It checks for the environment string SAMAPI=number  *)
(* to see if there is an override for the muxid (SAMAPI.EXE does the same    *)
(* thing).                                                                   *)
(*                                                                           *)
(* returns: true if resident code found and can be used, else false          *)
(*===========================================================================*)

FUNCTION locatesam : BOOLEAN;
  VAR
    envstr : STRING[80];
    v      : INTEGER;
    code   : INTEGER;
    r      : REGISTERS;

  BEGIN;

    sammuxid := DEFAULT_SAMMUX;

    envstr := getenv('SAMAPI');
    IF LENGTH(envstr) <> 0 THEN
      BEGIN;
        VAL(envstr, v, code);
        IF (code = 0) and (v > 0) THEN
           sammuxid := v;
      END;

    r.ah := sammuxid;
    r.al := 0;
    INTR($2F, r);

    locatesam := (r.al = 1);

  END;

(*===========================================================================*)
(* callsam                                                                   *)
(*                                                                           *)
(* Call the resident SAMAPI code.                                            *)
(*                                                                           *)
(* This expects a filled-in (all but header) command buffer and a            *)
(* to-be-filled-in response buffer. cmd is stuffed into cmdbuf and the       *)
(* resident code is called (via int $2f).                                    *)
(*===========================================================================*)

PROCEDURE callsam(cmdcode : INTEGER; VAR cmd, rsp);

  VAR
    r : REGISTERS;
    i : INTEGER;

  BEGIN;

    chdr_t (cmd).cmd := cmdcode;
    FOR i := 0 TO 2 DO
      chdr_t (cmd).fill[i] := 0;

    r.di := OFS(rsp);
    r.es := SEG(rsp);
    r.si := OFS(cmd);
    r.ds := SEG(cmd);
    r.al := 1;
    r.ah := sammuxid;

    INTR($2F, r);

  END;

(*===========================================================================*)
(* Converts "C" style zero terminated string into Pascal string              *)
(*===========================================================================*)

PROCEDURE fromasciz(VAR dst : STRING;
                    VAR src;
                    maxlen  : BYTE);

  TYPE
    bary = ARRAY [0..256] OF CHAR;

  VAR
    i : BYTE;

  BEGIN;

    i := 0;

    WHILE (bary(src)[i] <> #0) AND (maxlen > 0) DO
      BEGIN;
        dst[i+1] := bary(src)[i];
        INC(i);
        DEC(maxlen);
      END;

    dst[0] := CHAR(i);

  END;

(*===========================================================================*)
(* Converts pascal type string into "C" style string                         *)
(*===========================================================================*)

PROCEDURE toasciz(VAR dst;
                  VAR src : STRING;
                  maxlen  : BYTE);

  TYPE
    bary = ARRAY [0..256] OF BYTE;

  VAR
    i : BYTE;

  BEGIN;

    i := LENGTH(src);

    IF i > maxlen THEN i := maxlen;

    IF i > 0 THEN
      MOVE(src[1], dst, i);

    bary(dst)[i] := 0;

  END;

(*===========================================================================*)
(* Converts the data from the interface record from on type string to another*)
(*===========================================================================*)

PROCEDURE getsamstrings(VAR p : callstrings_t;
                        VAR c : datarec_t);

  BEGIN;

    fromasciz(p.call,       c.call,        6);
    fromasciz(p.class,      c.class,       1);
    fromasciz(p.firstname,  c.firstname,  11);
    fromasciz(p.midinitial, c.midinitial,  1);
    fromasciz(p.lastname,   c.lastname,   20);
    fromasciz(p.address,    c.address,    35);
    fromasciz(p.city,       c.city,       20);
    fromasciz(p.state,      c.state,       2);
    fromasciz(p.zip,        c.zip,         5);
    fromasciz(p.dob,        c.dob,         2);

  END;

(*===========================================================================*)
(* Display call sign data                                                    *)
(*===========================================================================*)

PROCEDURE display_call_4lines(rec : callstrings_t);

  BEGIN;

    send_tnc_data_str(cr);

    strip_var(rec.call, 'B');
    send_tnc_data_str(rec.call + ' ' + rec.class + cr);

    send_tnc_data_str(rec.firstname + ' ');
    IF (rec.midinitial[1] <> ' ') THEN
      send_tnc_data_str(rec.midinitial + ' ');
    send_tnc_data_str(rec.lastname+cr);

    send_tnc_data_str(rec.address+cr);

    send_tnc_data_str(rec.city + ', ' + rec.state + ' ' + rec.zip+cr);

  END;

(*===========================================================================*)
(* Look up a call                                                            *)
(*===========================================================================*)

FUNCTION lookup_call(call : STRING; VAR rec : callstrings_t) : INTEGER;

  VAR
    cmd : cmdfindcall_t;
    rsp : rspdatarec_t;
    len : INTEGER;
    i   : INTEGER;

  BEGIN;
    toasciz(cmd.call, call, 6);
    cmd.packflags := 0;
    callsam(samfindcall, cmd, rsp);
    getsamstrings(rec, rsp.d);
    lookup_call := rsp.h.err;
  END;

(*===========================================================================*)
(* Man command processor                                                     *)
(*===========================================================================*)

PROCEDURE sam_ulookup_call(cmd_string : STRING);
  VAR
    drec     : callstrings_t;
    err      : INTEGER;
    i        : INTEGER;
    numwords : INTEGER;
    s        : STRING[50];
    w        : STRING[10];

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Prep things                                                           *)
    (*-----------------------------------------------------------------------*)

    upcase_str_var(cmd_string);
    numwords := words(cmd_string);

    (*-----------------------------------------------------------------------*)
    (* Check for errors                                                      *)
    (*-----------------------------------------------------------------------*)

    IF numwords < 2 THEN
      BEGIN;
        send_message(message_not_en);   {not enough operands message}
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Database around?                                                      *)
    (*-----------------------------------------------------------------------*)

    IF NOT locatesam THEN
      BEGIN;
        send_message(message_no_db);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* See what kind of search to be done?                                   *)
    (*-----------------------------------------------------------------------*)

    IF subword(@cmd_string, 1, 1) <> 'FIND' THEN
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Search list of calls                                              *)
        (*-------------------------------------------------------------------*)

        FOR i := 2 TO numwords DO
          BEGIN;

            w := subword(@cmd_string, i, 1);
            upcase_str_var(w);

            err := lookup_call(w, drec);

            CASE err OF
              0 : display_call_4lines(drec);
              SerrNotFound :
                BEGIN;
                  set_dollar1_parm(@w);
                  send_message(message_no_db_record);
                  active_tcb^.error_sw := TRUE;
                END;
              ELSE
                BEGIN;
                  send_tnc_data_str('*** SAMAPI error' + cr);
                  active_tcb^.error_sw := TRUE;
                END;
            END;

          END; (*----- End call sign loop -----------------------------------*)

        (*-------------------------------------------------------------------*)
        (* Done with call signs                                              *)
        (*-------------------------------------------------------------------*)

        send_message(message_op_complete);

        EXIT;

      END; (*----- End call sign processor ----------------------------------*)

    (*-----------------------------------------------------------------------*)
    (* Going to run a search.  Make sure user asked for something            *)
    (*-----------------------------------------------------------------------*)

    IF numwords < 2 THEN
      BEGIN;
        set_dollar1_parm(@cmd_string);
        send_message(message_bad_search);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* See what kind of search to be done?                                   *)
    (*-----------------------------------------------------------------------*)

    cmd_string := subword(@cmd_string, 3, 255);
    numwords := words(cmd_string);

    REPEAT

      numwords   := words(cmd_string);

      IF numwords < 2 THEN
        BEGIN;
          set_dollar1_parm(@cmd_string);
          send_message(message_bad_search);
          active_tcb^.error_sw := TRUE;
          EXIT;
        END;

      w          := subword(@cmd_string, 1, 1);
      s          := subword(@cmd_string, 2, 1);
      cmd_string := subword(@cmd_string, 3, 255);

      i := 999;
      IF w = 'LASTNAME'  THEN i := DnoLastName;
      IF w = 'FIRSTNAME' THEN i := DnoFirstName;
      IF w = 'CITY'      THEN i := DnoAddress;
      IF w = 'STATE'     THEN i := DnoState;

      IF i = 999 THEN
        BEGIN;
          set_dollar1_parm(@w);
          send_message(message_bad_search);
          active_tcb^.error_sw := TRUE;
          EXIT;
        END;

    UNTIL cmd_string = '';

  END;

END.
