(*===========================================================================*)
(* Edit a file -- subroutines overlayed                                      *)
(*                                                                           *)
(*   Copyright (c) 1985, 87 by Borland International, Inc.                   *)
(*   Copyright 1988, 1990 by H. Roy Engehausen.  All rights reserved.        *)
(*                                                                           *)
(*   This code cannot be used in other programs except in conjunction        *)
(*   with a licensed copy of the Turbo Editor Toolbox                        *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* DRAWBOX -- Draw a box around a window                                     *)
(*===========================================================================*)

PROCEDURE drawbox(border : borderchars; x1, y1, x2, y2 : BYTE);

  VAR
    i      : WORD;
    bar    : STRING;
    barlen : BYTE ABSOLUTE bar;
    s      : STRING;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Build horizontal bar                                                  *)
    (*-----------------------------------------------------------------------*)

    barlen := 3 + x2 - x1;
    FILLCHAR(bar[1], barlen, border[horiz]);

    (*-----------------------------------------------------------------------*)
    (* Draw top bar                                                          *)
    (*-----------------------------------------------------------------------*)


    IF edit_operation = 0 THEN
      BEGIN;
        bar[1]      := border[topleft];
        bar[barlen] := border[topright];
        crtputfast(x1, y1, bar);
      END
    ELSE
      BEGIN;
        s := header_msg_block(edit_parm, 3);
        s := border[topleft] + LEFT(s, 78) + border[topright];
        crtputfast(x1, y1, s);
      END;

    (*-----------------------------------------------------------------------*)
    (* Draw bottom bar                                                       *)
    (*-----------------------------------------------------------------------*)

    bar[1]      := border[botleft];
    bar[barlen] := border[botright];
    crtputfast(x1, y2+2, bar);

    (*-----------------------------------------------------------------------*)
    (* Build vertical bars                                                   *)
    (*-----------------------------------------------------------------------*)

    FOR i := SUCC(y1) TO SUCC(y2) DO
      BEGIN;
        crtputfast(x1,   i, border[vert]);
        crtputfast(x2+2, i, border[vert]);
      END;

  END;

(*===========================================================================*)
(* Write a status message                                                    *)
(*===========================================================================*)

PROCEDURE writestatus(msg : STRING);

  VAR
    msglen : BYTE ABSOLUTE msg;

  BEGIN;

    FILLCHAR(msg[SUCC(msglen)], 80-msglen, #32);
    msglen := 80;
    crtputfast(1, 25, caerr+msg);

  END;

(*===========================================================================*)
(* Check results of editor load                                              *)
(*===========================================================================*)

FUNCTION checkinitbinary(exitcode : WORD) : BOOLEAN;

  BEGIN;

    checkinitbinary := exitcode <> 0;

    IF exitcode <> 0 THEN
      BEGIN;

        GOTOXY(1, 1);

        CASE exitcode OF
          1 : WRITELN('Insufficient heap space for text buffer');
        ELSE
          WRITELN('Unknown load error');
        END;

        DELAY(1000);

      END;

  END;

(*===========================================================================*)
(* Check results of editor read                                              *)
(*===========================================================================*)

FUNCTION checkreadfile(exitcode : WORD; fname : string) : BOOLEAN;

  VAR
    f : file;
    i : INTEGER;

  BEGIN;

    checkreadfile := exitcode <> 0;

    IF exitcode <> 0 THEN
      BEGIN;

        CASE exitcode OF

          1 : BEGIN;
                {New file, assure valid file name}
                {$I-}
                ASSIGN(f, fname);
                REWRITE(f);
                IF ioresult <> 0 THEN
                  BEGIN;
                    CLOSE(f);
                    i := IORESULT;
                    WRITESTATUS('Illegal file name '+ fname);
                  END
                ELSE
                  BEGIN;
                    CLOSE(f);
                    ERASE(f);
                    WRITE('New File');
                    DELAY(1000);
                    WRITE(^M);
                    CLREOL;
                    GOTOXY(1, 1);
                    CLREOL;
                    checkreadfile := FALSE;
                    EXIT;
                  END;
                {$I+}
              END;

          2 : writestatus('Insufficient text buffer size');

        ELSE
          writestatus('Unknown read error');
      END;

      DELAY(1000);

    END;

    GOTOXY(1, 1);
    CLREOL;

  END;

(*===========================================================================*)
(* Check results of editor save                                              *)
(*===========================================================================*)

FUNCTION checksavefile(exitcode : WORD; fname : STRING) : BOOLEAN;

  BEGIN;

    checksavefile := exitcode <> 0;

    IF exitcode <> 0 THEN
      BEGIN;
        CASE exitcode OF
          1 : WriteStatus('Unable to create output file ' + fname);
          2 : WriteStatus('Error while writing output to '+ fname);
          3 : WriteStatus('Unable to close output file '  + fname);
        ELSE
          writestatus('Unknown write error');
        END;
        DELAY(1000);
      END;

  END;

(*===========================================================================*)
(* initstatusline -- Draw status/prompt line                                 *)
(*===========================================================================*)

PROCEDURE initstatusline;
  VAR
    junkroy : BYTE;

  BEGIN;
    IF edit_operation = 0 THEN
      writestatus(' ^K^D saves file, ^K^Q abandons')
    ELSE
      writestatus(
          'Alt Keys: (K)ill (R)elease (H)old (L)eave (Q)uit r(E)ply (C)hange');
  END;

(*===========================================================================*)
(* INITWINDOW -- Draw screen frame around window                             *)
(*===========================================================================*)

PROCEDURE initwindow(VAR eddata : edcb);

  VAR
    msgpos : BYTE;
    winmsg : String;

  BEGIN;                      (*InitWindow*)
    (*Draw a frame around the editor window*)
    WITH eddata DO
      BEGIN
        drawbox(border, x1, y1, x2, y2);
        winmsg := ' Borland Binary Editor ';
        msgpos := 2 + ((x2 + x1 - LENGTH(winmsg)) shr 1);
        CRTputFast(msgpos, Y2 + 2, caerr + winmsg);
      END;
  END;

(*===========================================================================*)
(* exitbinaryeditor -- Handle exit command                                   *)
(*===========================================================================*)

FUNCTION exitbinaryeditor(VAR eddata : edcb; exitcom : INTEGER) : BOOLEAN;
  VAR
    b        : BOOLEAN;
    exitcode : WORD;

    FUNCTION yesanswer(prompt : STRING) : BOOLEAN;
    VAR
      ch : Char;

    BEGIN;
      {$IFDEF DEBUG_OVER}
        WRITELN('1');
        DELAY(2000);
        WRITELN('1.1');
        DELAY(2000);
        stack_depth;
        WRITELN('1.2');
        DELAY(2000);
        dump_all;
      {$ENDIF}
      writestatus(prompt);
      {$IFDEF DEBUG_OVER}
        WRITELN('1.3');
        DELAY(2000);
        stack_depth;
        WRITELN('1.4');
        DELAY(2000);
        dump_all;
      {$ENDIF}
      REPEAT
        {$IFDEF DEBUG_OVER}
          WRITELN('2.1');
          DELAY(2000);
        {$ENDIF}
        REPEAT
          task_switch;
          {$IFDEF DEBUG_OVER}
            WRITELN('2.15');
          {$ENDIF}
        UNTIL KeyPressed;
        {$IFDEF DEBUG_OVER}
          WRITELN('2.15');
          DELAY(2000);
        {$ENDIF}
        ch:= READKEY;
        {$IFDEF DEBUG_OVER}
          WRITELN('2.2');
          DELAY(2000);
        {$ENDIF}
        ch:= UPCASE(ch);
        {$IFDEF DEBUG_OVER}
          WRITELN('2.3');
          DELAY(2000);
        {$ENDIF}
      UNTIL ch IN ['Y', 'N'];
      yesanswer := (ch = 'Y');
      {$IFDEF DEBUG_OVER}
        WRITELN('3');
        DELAY(2000);
      {$ENDIF}
    END;

  FUNCTION save_the_file : BOOLEAN;
    BEGIN;
      get_semaphore(semaphore_interrupts, sem_exclusive, FALSE);
      edit_size := eddata.EOtext;
      exitcode := savefilebinaryeditor(eddata, makebackup);
      save_the_file := checksavefile(exitcode, filenamebinaryeditor(eddata));
      free_semaphore(semaphore_interrupts);
    END;

  BEGIN;

    {$IFDEF DEBUG_OVER}
      dump_string('Exit binary editor');
      dump_overlay;
      dump_trace;
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Handle exit other than regular                                        *)
    (*-----------------------------------------------------------------------*)

    IF edit_operation <> 0 THEN
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* If one of the old codes, ignore it                                *)
        (*-------------------------------------------------------------------*)

        IF exitcom < 2 THEN
          BEGIN;
            exitbinaryeditor := FALSE;
            EXIT;
          END;

        (*-------------------------------------------------------------------*)
        (* Its a new code.  Save it and switch to ^KQ                        *)
        (*-------------------------------------------------------------------*)

        edit_exitcode := exitcom;
        exitcom       := 0;

      END;

    (*-----------------------------------------------------------------------*)
    (* If any code other than the first three reach here, just ignore them   *)
    (*-----------------------------------------------------------------------*)

    IF exitcom > 1 THEN
      BEGIN;
        exitbinaryeditor := FALSE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Assume that the exit is true;                                         *)
    (*-----------------------------------------------------------------------*)

    exitbinaryeditor := TRUE;

    (*-----------------------------------------------------------------------*)
    (* Do the correct exit                                                   *)
    (*-----------------------------------------------------------------------*)

    CASE exitcom OF

      (*---------------------------------------------------------------------*)
      (* ^KD                                                                 *)
      (*---------------------------------------------------------------------*)

      -1:
        b := save_the_file;

      (*---------------------------------------------------------------------*)
      (* ^KQ, F10                                                            *)
      (*---------------------------------------------------------------------*)

      0, + 1 :                     (*^K^Q*)(* F10 *)
        BEGIN;

          {$IFDEF DEBUG_OVER}
            WRITELN('A');
            DELAY(2000);
            dump_string('KQ');
            dump_overlay;
            dump_trace;
          {$ENDIF}

         (*------------------------------------------------------------------*)
         (* See if file is modified                                          *)
         (*------------------------------------------------------------------*)

         IF modifiedfilebinaryeditor(eddata) THEN
           BEGIN;

             (*--------------------------------------------------------------*)
             (* File has been modified                                       *)
             (*--------------------------------------------------------------*)

             (*--------------------------------------------------------------*)
             (* Debugging stuff                                              *)
             (*--------------------------------------------------------------*)

             {$IFDEF DEBUG_OVER}
               WRITELN('B');
               DELAY(2000);
               dump_overlay;
               dump_trace;
             {$ENDIF}

             (*--------------------------------------------------------------*)
             (* Ask user what he wants done?                                 *)
             (*--------------------------------------------------------------*)

             IF yesanswer('File modified. Save it? (Y/N) ') THEN
               BEGIN

                 (*----------------------------------------------------------*)
                 (* User wants file saved                                    *)
                 (*----------------------------------------------------------*)

                 {$IFDEF DEBUG_OVER}
                   WRITELN('C');
                   DELAY(2000);
                   dump_overlay;
                   dump_trace;
                 {$ENDIF}

                 (*----------------------------------------------------------*)
                 (* Save the file                                            *)
                 (*----------------------------------------------------------*)

                 b := save_the_file;
                 exitbinaryeditor := TRUE;

               END;
           END;

           GOTOXY(1, 1);

        END; (*----- End handling ^KQ ---------------------------------------*)

    END; (*----- End case statement on editor exit code ---------------------*)

  END; (*----- ExitBinaryEditor ---------------------------------------------*)
