(*===========================================================================*)
(* Get a file in from the TNC                                                *)
(*                                                                           *)
(*   Copyright 1990, 1991 by H. Roy Engehausen.  All rights reserved.        *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Begin to receive a text file from the TNC                                 *)
(*===========================================================================*)

PROCEDURE in_text_file(fileid    : file_name_str;
                       head_scan : BOOLEAN);

  VAR
    calls_scanned : STRING;
    cr_sw         : BOOLEAN;
    dupe_cnt      : WORD;
    eof_str       : STRING[5];
    eof_sw        : BOOLEAN;
    first_line    : BOOLEAN;
    header        : BOOLEAN;
    i             : INTEGER;
    in_str        : STRING;
    p             : POINTER;

{$UNDEF DEBUG}

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Clear calls scanned                                                   *)
    (*-----------------------------------------------------------------------*)

    free_task_mem(memid_call_scanned, TRUE);

    (*-----------------------------------------------------------------------*)
    (* Open the output file                                                  *)
    (*-----------------------------------------------------------------------*)

    in_str := open_text_file(fileid, FALSE);

    IF in_str <> '' THEN
      BEGIN;
        send_tnc_data_str(in_str);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    WITH active_tcb^.io_fe^ DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Initialize the loop                                               *)
        (*-------------------------------------------------------------------*)

        cr_sw      := FALSE;
        eof_sw     := FALSE;
        first_line := head_scan;
        header     := head_scan;
        dupe_cnt   := 0;

        (*-------------------------------------------------------------------*)
        (* Loop around reading lines                                         *)
        (*-------------------------------------------------------------------*)

        REPEAT

          (*-----------------------------------------------------------------*)
          (* Get a line                                                      *)
          (*-----------------------------------------------------------------*)

          in_str := read_tnc_data_str;

          (*-----------------------------------------------------------------*)
          (* See if EOF character is present.                                *)
          (*-----------------------------------------------------------------*)

          i := POS(eof_chr, in_str);

          (*-----------------------------------------------------------------*)
          (* Calculate line length and leave it in the variable "i"          *)
          (*-----------------------------------------------------------------*)

          IF i > 0 THEN
            BEGIN;
              eof_sw := TRUE;
              IF i > 1 THEN
                in_str := COPY(in_str, 1, i-1)
              ELSE
                in_str := '';
            END;

          (*-----------------------------------------------------------------*)
          (* See if MBL eof                                                  *)
          (*-----------------------------------------------------------------*)

          IF (in_str[1] = '/') AND (NOT eof_sw) THEN
            BEGIN;
              eof_str := upcase_str(in_str);
              strip_crlf(eof_str);
              IF (eof_str = eof_chars1) OR (eof_str = eof_chars2) THEN
                BEGIN;
                  eof_sw := TRUE;
                  in_str := '';
                END;
            END;

          (*-----------------------------------------------------------------*)
          (* See if we should parse this as a header                         *)
          (*-----------------------------------------------------------------*)

          IF header THEN
            BEGIN;

              {$IFDEF DEBUG}
                WRITELN('HEADER1 = ', header);
              {$ENDIF}

              header := scan_head(in_str, calls_scanned, dupe_cnt);

              {$IFDEF DEBUG}
                WRITELN('HEADER2 = ', header);
              {$ENDIF}

            END;

          (*-----------------------------------------------------------------*)
          (* See if we need to insert a blank line.  All messages must either*)
          (* start with a header or with a blank line                        *)
          (*-----------------------------------------------------------------*)

          IF first_line THEN
            BEGIN;
              first_line := FALSE;
              IF NOT header THEN
                BEGIN;
                  eof_str := in_str;
                  strip_crlf(eof_str);
                  IF eof_str <> '' THEN
                    WRITELN(fe_text);
                END;
            END;

          (*-----------------------------------------------------------------*)
          (* If there is a line then write it.                               *)
          (*-----------------------------------------------------------------*)

          IF LENGTH(in_str) <> 0 THEN
            BEGIN;

              (*-------------------------------------------------------------*)
              (* Loop until no more data to write                            *)
              (*-------------------------------------------------------------*)

              WHILE in_str <> '' DO
                BEGIN;

                  (*---------------------------------------------------------*)
                  (* If pending CR and no matching LF, add one               *)
                  (*---------------------------------------------------------*)

                  IF cr_sw AND (in_str[1] <> lf) THEN
                    BEGIN;
                      fe_size := fe_size + 1;
                      WRITE(fe_text, lf);
                    END;

                  (*---------------------------------------------------------*)
                  (* See if we have a CR.  Leave location in "i"             *)
                  (*---------------------------------------------------------*)

                  i     := POS(cr, in_str);
                  cr_sw := i <> 0;

                  (*---------------------------------------------------------*)
                  (* Write the line up to and including the CR               *)
                  (*---------------------------------------------------------*)

                  WRITE(fe_text, substr(in_str, 1, i));

                  (*---------------------------------------------------------*)
                  (* Separate on the CR ("i" contains it location)           *)
                  (*---------------------------------------------------------*)

                  IF NOT cr_sw THEN
                    BEGIN;
                      i := LENGTH(in_str);
                      in_str := '';
                    END
                  ELSE
                    IF i <> 255 THEN
                      in_str := COPY(in_str, i+1, 255)
                    ELSE
                      in_str := '';

                  (*---------------------------------------------------------*)
                  (* Bump the size                                           *)
                  (*---------------------------------------------------------*)

                  fe_size := fe_size + i;

                END; (*----- End the write loop -----------------------------*)

            END; (*----- End write data if statement ------------------------*)

          (*-----------------------------------------------------------------*)
          (* Loop until EOF                                                  *)
          (*-----------------------------------------------------------------*)

        UNTIL eof_sw;

        (*-------------------------------------------------------------------*)
        (* If we are missing an LF, then add it                              *)
        (*-------------------------------------------------------------------*)

        IF cr_sw THEN
          BEGIN;
            WRITE(fe_text, lf);
            fe_size := fe_size + 1;
          END;

        (*-------------------------------------------------------------------*)
        (* Save the size                                                     *)
        (*-------------------------------------------------------------------*)

        io_file_size := fe_size;

      END;

    (*-----------------------------------------------------------------------*)
    (* Close the output file                                                 *)
    (*-----------------------------------------------------------------------*)

    in_str := close_text_file;
    IF in_str <> '' THEN
      BEGIN;
        send_tnc_data_str(in_str);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Push calls scanned into the memory stack                              *)
    (*-----------------------------------------------------------------------*)

    i := LENGTH(calls_scanned) + 1;
    IF i > 1 THEN
      BEGIN;
        p := get_task_mem(memid_call_scanned, i);
        MOVE(calls_scanned, p^, i);
      END;

  END;
