(*===========================================================================*)
(* Forward task -- Find next route, find_next_msg                            *)
(*                                                                           *)
(*   Copyright 1990, 1991 by H. Roy Engehausen.  All rights reserved.        *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

{$DEFINE POINT_CHK}
{$UNDEF  DEBUG_TIME} (* Debug forwarding time *)

UNIT BBFNR;

INTERFACE

  USES
    bbdummy,
    bbfwdd;

  FUNCTION find_next_route(path_common : path_block_ptr;
                           this_route  : msg_r_ptr) : msg_r_ptr;

  FUNCTION find_next_msg (this_route   : msg_r_ptr;
                          this_msg     : msg_index_ptr;
                          VAR this_inx : BYTE) : msg_index_ptr;

  VAR
    fnm_linked_list_valid : BOOLEAN;

IMPLEMENTATION

  USES
    bbstr,
    bbmisc3,
    bbtime;

  VAR
    fnm_linked_list_start : msg_index_ptr;

(*===========================================================================*)
(* Find next route to forward to this target                                 *)
(*===========================================================================*)

FUNCTION find_next_route(path_common : path_block_ptr;
                         this_route  : msg_r_ptr) : msg_r_ptr;

  VAR
    code       : INTEGER;
    i          : BYTE;
    j          : INTEGER;

  LABEL
    find_next,
    find_none;

  (*=========================================================================*)
  (* Subprocedure to set the minimum age of a message to be forwarded        *)
  (*=========================================================================*)

  PROCEDURE set_path_age;

    VAR
      code       : INTEGER;
      hours      : INTEGER;
      hour_sum   : WORD;
      word_max   : INTEGER;
      word_no    : BYTE;
      word_to_do : STRING[4];

    LABEL
      iterate;

    BEGIN;

      (*---------------------------------------------------------------------*)
      (* Initialize counters                                                 *)
      (*---------------------------------------------------------------------*)

      hour_sum := 0;

      (*---------------------------------------------------------------------*)
      (* Get the number of words to test.  If none, we are done              *)
      (*---------------------------------------------------------------------*)

      word_max := path_common^.path_route_inx - 1;
      IF word_max < 1 THEN EXIT;

      (*---------------------------------------------------------------------*)
      (* Loop thru the words                                                 *)
      (*---------------------------------------------------------------------*)

      word_no := word_max;
      WHILE word_no > 0 DO
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* Get the word to be evaulated and see if it is short enough      *)
          (*-----------------------------------------------------------------*)

          word_to_do := SUBWORD(@this_route^.msg_r_info, word_no, 1);
          IF LENGTH(word_to_do) > 3 THEN
            GOTO iterate;

          (*-----------------------------------------------------------------*)
          (* Evaluate the word                                               *)
          (*-----------------------------------------------------------------*)

          VAL(word_to_do, hours, code);

          (*-----------------------------------------------------------------*)
          (* Reject certain things                                           *)
          (*-----------------------------------------------------------------*)

          IF code <> 0 THEN GOTO iterate;
          IF hours < 0 THEN GOTO iterate;

          (*-----------------------------------------------------------------*)
          (* The data is valid                                               *)
          (*-----------------------------------------------------------------*)

          hour_sum := hour_sum + hours;

          (*-----------------------------------------------------------------*)
          (* If there is no plus sign then data is absolute.  Compute        *)
          (* minimum path age and we are done                                *)
          (*-----------------------------------------------------------------*)

          IF word_to_do[1] <> '+' THEN
            BEGIN;
              path_common^.path_age := current_day_time
                                             - LONGINT(hours) * ticks_per_hour;
              EXIT;
            END;

          (*-----------------------------------------------------------------*)
          (* Loop end for loop thru each word                                *)
          (*-----------------------------------------------------------------*)

iterate:

          DEC(word_no);

        END; (*----- End of loop thru words ---------------------------------*)

      (*---------------------------------------------------------------------*)
      (* Compute the maximum path age and we are done                        *)
      (*---------------------------------------------------------------------*)

      path_common^.path_age := current_day_time
                                          - LONGINT(hour_sum) * ticks_per_hour;

    END;

  (*=========================================================================*)
  (* Main line                                                               *)
  (*=========================================================================*)

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Chain forward to next block.  Note that the inital call has a         *)
    (* fake pointer to the top of the list                                   *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF POINT_CHK}
       test_pointer(this_route);
    {$ENDIF}

    this_route := this_route^.msg_r_next;

    (*-----------------------------------------------------------------------*)
    (* If nothing is next then we are done                                   *)
    (*-----------------------------------------------------------------------*)

    IF this_route = NIL THEN
      GOTO find_none;

    (*-----------------------------------------------------------------------*)
    (* If this is a normal forward then the current path must be the         *)
    (* the next entry in the route list for the route under test             *)
    (*                                                                       *)
    (* For abnormal forwards (forced or reverse) then we will                *)
    (* select a route if the current path is anywhere in the route list      *)
    (*-----------------------------------------------------------------------*)

    IF (NOT path_common^.path_rev)
                                   AND (NOT path_common^.path_force_sw)
                                   AND (NOT path_common^.path_force_this) THEN
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* This code is for normal searches                                  *)
        (*-------------------------------------------------------------------*)

        WHILE (this_route <> NIL)
            AND (SUBWORD(@this_route^.msg_r_info, this_route^.msg_r_nroute, 1)
                                               <> path_common^.path_target) DO
          BEGIN;

            {$IFDEF POINT_CHK}
              test_pointer(this_route);
            {$ENDIF}

            this_route := this_route^.msg_r_next;
          END;

        IF this_route = NIL THEN
          GOTO find_none;

        path_common^.path_route_inx := this_route^.msg_r_nroute;

      END
    ELSE
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* This code is for the abnormal search                              *)
        (*-------------------------------------------------------------------*)

        GOTO find_next;
        REPEAT

          this_route := this_route^.msg_r_next;

find_next:

          IF this_route = NIL THEN
            GOTO find_none;

          {$IFDEF POINT_CHK}
            test_pointer(this_route);
          {$ENDIF}

          i := FIND(@this_route^.msg_r_info, @path_common^.path_target);

        UNTIL i <> 0;

        path_common^.path_route_inx := i;

      END;

    (*-----------------------------------------------------------------------*)
    (* Set the result                                                        *)
    (*-----------------------------------------------------------------------*)

find_none:

    find_next_route := this_route;

    (*-----------------------------------------------------------------------*)
    (* Set the forward time limit                                            *)
    (*-----------------------------------------------------------------------*)

    path_common^.path_age := max_time;
    set_path_age;

  END; (*----- End find next route procedure --------------------------------*)

(*===========================================================================*)
(* Find next msg                                                             *)
(*===========================================================================*)

FUNCTION  find_next_msg (this_route   : msg_r_ptr;
                         this_msg     : msg_index_ptr;
                         VAR this_inx : BYTE) : msg_index_ptr;

  VAR
    i           : BYTE;
    inx         : BYTE;
    temp_flag   : msg_flag_type;

  CONST
    dont_fwd_flags  : msg_flag_type = mf_hold
                                      OR mf_read
                                      OR mf_fwd
                                      OR mf_fwd_select
                                      OR mf_kill
                                      OR mf_review;

  (*=========================================================================*)
  (* Procedure to build the list of messages to search                       *)
  (*=========================================================================*)

  PROCEDURE build_fnr_list;

    VAR
      last_msg     : msg_index_ptr;
      this_msg     : msg_index_ptr;

    BEGIN;

      (*---------------------------------------------------------------------*)
      (* Initialize                                                          *)
      (*---------------------------------------------------------------------*)

      fnm_linked_list_valid := TRUE;

      fnm_linked_list_start := NIL;

      this_msg := msg_index_start;

      (*---------------------------------------------------------------------*)
      (* Loop thru all the messages                                          *)
      (*---------------------------------------------------------------------*)

      WHILE this_msg <> NIL DO
        BEGIN;

          {$IFDEF POINT_CHK}
            test_pointer(this_msg);
          {$ENDIF}

          (*-----------------------------------------------------------------*)
          (* See if this is a message we want                                *)
          (*-----------------------------------------------------------------*)

          IF (this_msg^.msg_i_mb.msg_flag AND dont_fwd_flags) = mf_fwd_select
                                                                           THEN
            BEGIN;

              (*-------------------------------------------------------------*)
              (* Yes it is.  Add to chain                                    *)
              (*-------------------------------------------------------------*)

              IF fnm_linked_list_start = NIL THEN
                fnm_linked_list_start := this_msg
              ELSE
                last_msg^.msg_i_fnm_l := this_msg;

              last_msg := this_msg;

            END;

          (*-----------------------------------------------------------------*)
          (* Chain to next message                                           *)
          (*-----------------------------------------------------------------*)

          this_msg := this_msg^.msg_i_next;

        END;

      (*---------------------------------------------------------------------*)
      (* Make sure last message in chain has a null followon pointer         *)
      (*---------------------------------------------------------------------*)

      IF fnm_linked_list_start <> NIL THEN
        BEGIN;

          {$IFDEF POINT_CHK}
            test_pointer(last_msg);
          {$ENDIF}

          last_msg^.msg_i_fnm_l := NIL;

        END;

    END;


  BEGIN;

    IF NOT fnm_linked_list_valid THEN
      build_fnr_list;

    (*-----------------------------------------------------------------------*)
    (* Get ready for the loop                                                *)
    (*-----------------------------------------------------------------------*)

    IF this_msg = NIL THEN
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Start the search at the beginning                                 *)
        (*-------------------------------------------------------------------*)

        this_msg := fnm_linked_list_start;
        inx      := 0;

      END
    ELSE
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Start the search at the next message or destination in the list   *)
        (*-------------------------------------------------------------------*)

        inx := this_inx;
        IF inx = 0 THEN
          this_msg := this_msg^.msg_i_fnm_l;

      END;

    WHILE this_msg <> NIL DO
      BEGIN;

        {$IFDEF POINT_CHK}
          test_pointer(this_msg);
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Get the flag in a handy place                                     *)
        (*-------------------------------------------------------------------*)

        temp_flag := this_msg^.msg_i_mb.msg_flag;

        (*-------------------------------------------------------------------*)
        (* Only do something if forward select is on and others are off      *)
        (*-------------------------------------------------------------------*)

        IF (temp_flag AND dont_fwd_flags) = mf_fwd_select THEN
          BEGIN;

            (*---------------------------------------------------------------*)
            (* Have to handle distribution lists separately                  *)
            (*---------------------------------------------------------------*)

            IF (temp_flag AND mf_fwd_list) <> 0 THEN
              BEGIN;

                (*-----------------------------------------------------------*)
                (* If there is no distribution routing block then the code   *)
                (* has gone haywire                                          *)
                (*-----------------------------------------------------------*)

                IF (temp_flag AND mf_disrout) = 0 THEN
                  WRITELN('Select but no dblk -- #=',
                                                 this_msg^.msg_i_mb.msg_number)
                ELSE
                  WITH this_msg^.msg_i_dr^ DO
                    BEGIN;

                     {$IFDEF POINT_CHK}
                       test_pointer(this_msg^.msg_i_dr);
                     {$ENDIF}

                     {$IFDEF POINT_CHK}
                       test_pointer(msg_dr_dblk);
                     {$ENDIF}

                      (*-----------------------------------------------------*)
                      (* Search all members of the distribution routing      *)
                      (* block looking for a hit                             *)
                      (*-----------------------------------------------------*)

                      i := msg_dr_dblk^.msg_d_no;
                      WHILE inx < i DO
                        BEGIN;
                          INC(inx);
                          IF msg_dr_data[inx] = this_route THEN
                            BEGIN;
                              this_inx := inx;
                              find_next_msg := this_msg;
                              EXIT;
                            END;
                        END;

                    END;
              END
            ELSE

              (*-------------------------------------------------------------*)
              (* Check none distribution list items                          *)
              (*-------------------------------------------------------------*)

              IF (this_msg^.msg_i_rou = this_route) THEN
                BEGIN;
                  this_inx := 0;
                  find_next_msg := this_msg;
                  EXIT;
                END;

          END; (*---- End msg forward select test ---------------------------*)

        this_msg := this_msg^.msg_i_fnm_l;
        inx := 0;

      END; (*----- End loop thru all messages -------------------------------*)

    find_next_msg := NIL;
    this_inx      := 0;

  END;

END.
