(*===========================================================================*)
(* File subsystem -- Make file from message                                  *)
(*                                                                           *)
(*   Copyright 1988, 1990, 1991 by H. Roy Engehausen.  All rights reserved.  *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

UNIT BBFSF;

INTERFACE

  PROCEDURE make_file_cmd(cmd_string : STRING);

IMPLEMENTATION

  USES
    DOS,
    bbcopy,
    bbdummy,
    bbfsm,
    bbmdata,
    bbmess,
    bbmf,
    bbmisc2,
    bbrmc,
    bbsdata,
    bbstr,
    match;

PROCEDURE make_file_cmd(cmd_string : STRING);

  VAR
    code          : INTEGER;
    dir_to_search : fsb_name_str;
    i             : WORD;
    look          : SEARCHREC;
    m_no          : LONGINT;
    p_sw          : BYTE;
    search_arg    : file_name_str;
    t_str         : STRING[6];
    this_fsb      : fsb_ptr;
    this_msg      : msg_index_ptr;
    word_count    : BYTE;

  {$I BBFSI.PAS}

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Ready the input string                                                *)
    (*-----------------------------------------------------------------------*)

    upcase_str_var(cmd_string);
    word_count := words(cmd_string);

    (*-----------------------------------------------------------------------*)
    (* Check for special pointer to printer                                  *)
    (*-----------------------------------------------------------------------*)

    IF (word_count = 3)
                 AND match_str(subword(@cmd_string, 3, 1), printer_match) THEN
      BEGIN;
        p_sw := 1;
        cmd_string := cmd_string +  ' ' + subword(@cmd_string, 3, 1);
        word_count := 4;
      END
    ELSE
      p_sw := 0;

    (*-----------------------------------------------------------------------*)
    (* Check command format                                                  *)
    (*-----------------------------------------------------------------------*)

    IF word_count <> 4 THEN
      BEGIN;
        IF word_count < 4 THEN
          send_message(message_not_en)
        ELSE
          send_message(message_err_wrd);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Parse                                                                 *)
    (*-----------------------------------------------------------------------*)

    t_str := subword(@cmd_string, 2, 1);
    VAL(t_str, m_no, code);
    IF (code <> 0) OR (m_no < 1) OR (m_no > 65535) THEN
      BEGIN;
        send_message(message_err_ivm);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    dir_to_search := subwordl(cmd_string, 3, SIZEOF(fsb_name_str) - 1);

    search_arg    := subwordl(cmd_string, 4, SIZEOF(search_arg)   - 1);

    (*-----------------------------------------------------------------------*)
    (* Find the directory                                                    *)
    (*-----------------------------------------------------------------------*)

    IF p_sw = 0 THEN
      BEGIN;

        this_fsb := find_fsb(dir_to_search);

        (*-------------------------------------------------------------------*)
        (* If directory not found, give error message                        *)
        (*-------------------------------------------------------------------*)

        IF (this_fsb = NIL) OR
                   (active_tcb^.uid_data.user_class < this_fsb^.fsb_up) THEN
          BEGIN;
            send_message(message_no_files_one);
            active_tcb^.error_sw := TRUE;
            EXIT;
          END;

      END;

    (*-----------------------------------------------------------------------*)
    (* Check for subdirectory                                                *)
    (*-----------------------------------------------------------------------*)

    IF (p_sw = 0) AND (POS('\', search_arg) > 0)
                                        AND NOT this_fsb^.fsb_f_subdir_ok THEN
      BEGIN;
        send_message(message_no_slash);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Check for wildcards                                                   *)
    (*-----------------------------------------------------------------------*)

    IF (POS('*', search_arg) > 0) THEN
      BEGIN;
        send_message(message_no_wild);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Find the message                                                      *)
    (*-----------------------------------------------------------------------*)

    this_msg:= find_msg(m_no);
    IF this_msg = NIL THEN
      BEGIN;
        send_message(message_rmc_nf);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Build output file name                                                *)
    (*-----------------------------------------------------------------------*)

    IF p_sw = 0 THEN
      search_arg := this_fsb^.fsb_path + search_arg;

    (*-----------------------------------------------------------------------*)
    (* Execute proper subcommand                                             *)
    (*-----------------------------------------------------------------------*)

    CASE cmd_string[2] OF

      (*---------------------------------------------------------------------*)
      (* Handle " " and "H" -- X400 header to be added to front of file      *)
      (*                       If "H", keep message headers                  *)
      (*---------------------------------------------------------------------*)

      ' ', 'H' :
                 BEGIN;
                   read_a_msg(this_msg, search_arg, p_sw,
                                                    cmd_string[2] = 'H', TRUE);
                   IF NOT active_tcb^.error_sw THEN
                     send_message(message_file_saved);
                 END;

      (*---------------------------------------------------------------------*)
      (* "N" -- Copy message as is                                           *)
      (*---------------------------------------------------------------------*)

      'N' : BEGIN;

              (*-------------------------------------------------------------*)
              (* Build input file name                                       *)
              (*-------------------------------------------------------------*)

              STR(this_msg^.msg_i_mb.msg_number, cmd_string);
              cmd_string := opt_block.msg_file_dir + 'BB' + cmd_string + '.MSG';

              (*-------------------------------------------------------------*)
              (* Copy file                                                   *)
              (*-------------------------------------------------------------*)

              IF p_sw = 1 THEN
                cmd_string := copy_file_ascii(cmd_string, search_arg, TRUE)
              ELSE
                cmd_string := copy_file_binary(cmd_string, search_arg, FALSE);

              (*-------------------------------------------------------------*)
              (* Report any errors                                           *)
              (*-------------------------------------------------------------*)

              IF cmd_string <> '' THEN
                BEGIN;
                  send_tnc_data_str(cmd_string + cr);
                  active_tcb^.error_sw := TRUE;
                  EXIT;
                END
              ELSE
                send_message(message_file_saved);

            END;

      (*---------------------------------------------------------------------*)
      (* The second character is NOT valid                                   *)
      (*---------------------------------------------------------------------*)

      ELSE
        BEGIN;
          send_message(message_err_2nd);
          active_tcb^.error_sw := TRUE;
          EXIT;
        END;

  END; (*----- End select statement -----------------------------------------*)

END;

END.
