(*===========================================================================*)
(* File subsystem -- list a directory (subroutines)                          *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Convert a date/time stamp to characters                                   *)
(*===========================================================================*)

FUNCTION  dos_time_str(in_time : LONGINT) : STRING;

  VAR
    i        : BYTE;
    t_str    : STRING;
    t_time   : DATETIME;
    work_str : STRING[4];

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Break time into pieces.                                               *)
    (*-----------------------------------------------------------------------*)

    UNPACKTIME(in_time, t_time);

    WITH t_time DO
      BEGIN;

        STR(year:2, work_str);
        t_str[0] := CHR(2);
        IF year > 99 THEN
          i := 3
        ELSE
          i := 1;
        t_str[1] := work_str[i];
        t_str[2] := work_str[i+1];

        STR(month:2, work_str);
        IF work_str[1] = ' ' THEN work_str[1] := '0';
        t_str := t_str + work_str;

        STR(day:2, work_str);
        IF work_str[1] = ' ' THEN work_str[1] := '0';
        t_str := t_str + work_str +  '/';

        STR(hour:2, work_str);
        IF work_str[1] = ' ' THEN work_str[1] := '0';
        t_str := t_str + work_str;

        STR(min:2, work_str);
        IF work_str[1] = ' ' THEN work_str[1] := '0';
        t_str := t_str + work_str;

      END;

    dos_time_str := t_str;

  END;

(*===========================================================================*)
(* Compare two data items                                                    *)
(*===========================================================================*)

FUNCTION compare_items( a : WORD; b : WORD) : BOOLEAN;
  BEGIN;

    WITH dir_data^[a] DO
      CASE sort_fr OF
        name_sort:
            compare_items := fr_name <= dir_data^[b].fr_name;
        date_sort:
            compare_items := fr_time <= dir_data^[b].fr_time;
        size_sort:
            compare_items := fr_size <= dir_data^[b].fr_size;
      END;

  END;

(*===========================================================================*)
(* This does a straight bubble sort.  Call when the sort partition           *)
(* is small                                                                  *)
(*===========================================================================*)

PROCEDURE simple_sort(i : WORD; j : WORD);

  VAR
    k : WORD;
    l : WORD;
    t : fr_type;
  BEGIN;

    k := i + 1;

    WHILE k <= j DO
      BEGIN;

        IF NOT compare_items(k-1, k) THEN
          BEGIN;
            t := dir_data^[k];
            dir_data^[k] := dir_data^[k-1];
            dir_data^[k-1] := t;
            k := i;
          END;

        INC(k);

      END;

  END;

(*===========================================================================*)
(* Function returns the median of the data                                  *)
(*===========================================================================*)

FUNCTION med3i(q,r,s: word) : word;

  BEGIN;
    IF (compare_items(q, r) AND (compare_items(r, s))) OR
           (compare_items(s, r) AND (compare_items(r, q))) THEN
      med3i := r
    ELSE
      IF (compare_items(q, s) AND (compare_items(s, r))) OR
           (compare_items(r, s) AND (compare_items(s, q))) THEN
        med3i := q
      ELSE
        med3i := q;
  END;

(*===========================================================================*)
(* Partitions the data into two equal sections                               *)
(*===========================================================================*)

FUNCTION partition (i : WORD; j : WORD): WORD;

  VAR
    mid : WORD;

  BEGIN;

    mid := (i+j) SHR 1;

    partition := med3i(i, mid, j);

  END;

(*===========================================================================*)
(* Quick sort                                                                *)
(*===========================================================================*)

PROCEDURE quicksort(i : WORD; j : WORD);

  VAR
    k      : WORD;
    border : WORD;
    l      : WORD;
    r      : WORD;
    t      : fr_type;

  BEGIN
    IF i >= j THEN
      EXIT;


    IF (j - i) <= 10 THEN
      simple_sort(i, j)
    ELSE
        BEGIN;
          border := partition(I, J);
          l := i; r := j;
          WHILE (l < r) do
            BEGIN;
              WHILE (l < j) AND (NOT compare_items(border, l)) DO
                INC(l);
              WHILE (r > i) AND compare_items(border, r) DO
                DEC(r);
              IF l < r THEN
                BEGIN;
                  t := dir_data^[l];
                  dir_data^[l] := dir_data^[r];
                  dir_data^[r] := t;
                  INC(l);
                  DEC(r);
                END;
            END;

          quicksort(I,   R);
          quicksort(R+1, J);
        END;

  END;

(*===========================================================================*)
(* Display item                                                              *)
(*===========================================================================*)

FUNCTION display_item(itno : WORD) : item_string;
  VAR
    c  : CHAR;
    j  : BYTE;
    t  : STRING[28];
    w  : WORD;

  BEGIN;

    IF itno > dir_cnt THEN
      BEGIN;
        display_item := '';
        EXIT;
      END;

    WITH dir_data^[itno] DO
      BEGIN;

        IF (display_fr = full) OR (display_fr = short_size) THEN
          BEGIN;

            j := column_no;

            IF fr_size < 0 THEN
              IF j <> 2 THEN
                t := '  <DIR>'
              ELSE
                t := '     <DIR>'
            ELSE
              BEGIN;
                IF j = 2 THEN
                  STR(fr_size:10, t)
                ELSE
                  BEGIN;
                    IF fr_size < 10000000 THEN
                      BEGIN;
                        w := (fr_size + 512) DIV 1000;
                        c := 'K';
                      END
                    ELSE
                      BEGIN;
                        w := (fr_size + 500000) DIV 1000000;
                        c := 'M';
                      END;

                    IF w = 0 THEN
                      w := 1;

                    STR(w:5, t);
                    t := ' ' + t + c;
                  END;
              END;
          END
        ELSE
          t := '';

        IF (display_fr = full) OR (display_fr = short_date) THEN
          BEGIN;
            t := ' ' + dos_time_str(fr_time) + t;
            IF display_fr = short_date THEN
              IF fr_size < 0 THEN
                t := ' <DIR> '
              ELSE
                t := substr(t, 1, 7);
          END;

        display_item := LEFT(fr_name, (SIZEOF(fr_name) - 1)) + t + '     ';

      END;

  END;

(*===========================================================================*)
(* Display the directory                                                     *)
(*===========================================================================*)

PROCEDURE display_dir;
  VAR
    i : WORD;
    j : WORD;
    k : WORD;
    t : STRING[80];

  BEGIN;

    quicksort(1, dir_cnt);

    j := (dir_cnt + column_no - 1) div column_no;

    {$IFDEF DEBUG_DMORE}
      WRITELN('Directory display start');
    {$ENDIF}

    FOR i := 1 TO j DO
      BEGIN;

        {$IFDEF DEBUG_DMORE}
          WRITELN('Loop ', i);
        {$ENDIF}

        t := '';

        FOR k := 0 TO column_no-1 DO
          t := t + display_item (i + k * j);

        strip_var(t, 'T');
        send_tnc_data_str(t + cr);

        IF more_check THEN
          BEGIN;

            more_kill := TRUE;

            {$IFDEF DEBUG_DMORE}
              WRITELN('Exit ', i);
            {$ENDIF}

            EXIT;
          END;

      END;

  END;

(*===========================================================================*)
(* Display the directory                                                     *)
(*===========================================================================*)

PROCEDURE display_space;

  TYPE
    str10 = STRING[10];

  VAR
    ext  : EXTSTR;
    i    : BYTE;
    l1   : LONGINT;
    l2   : LONGINT;
    l3   : LONGINT;
    nam  : NAMESTR;
    s    : STRING;

  (*=========================================================================*)
  (* Subroutine to display a number in K                                     *)
  (*=========================================================================*)

  FUNCTION d_num( l : LONGINT) : str10;

    VAR
      l2 : LONGINT;
      s1 : STRING[15];
      s2 : STRING[15];

    BEGIN;

      (*---------------------------------------------------------------------*)
      (* Initialize                                                          *)
      (*---------------------------------------------------------------------*)

      s1 := '';
      s2 := '';

      (*---------------------------------------------------------------------*)
      (* Convert number to K                                                 *)
      (*---------------------------------------------------------------------*)

      l := (l + 500) DIV 1000;

      (*---------------------------------------------------------------------*)
      (* Loop converting each section of the number                          *)
      (*---------------------------------------------------------------------*)

      REPEAT

        (*-------------------------------------------------------------------*)
        (* Get a section of the number                                       *)
        (*-------------------------------------------------------------------*)

        l2 := l MOD 1000;
        l  := l DIV 1000;

        (*-------------------------------------------------------------------*)
        (* Convert to characters                                             *)
        (*-------------------------------------------------------------------*)

        STR(l2, s2);

        (*-------------------------------------------------------------------*)
        (* If more of the number remains then add leading zeros to what      *)
        (* we have and add the leading comma                                 *)
        (*-------------------------------------------------------------------*)

        IF l > 0 THEN
          BEGIN;
            WHILE LENGTH(s2) < 3 DO
              s2 := '0' + s2;
            s2 := ',' + s2;
          END;

        (*-------------------------------------------------------------------*)
        (* Put this section in front of the others                           *)
        (*-------------------------------------------------------------------*)

        s1 := s2 + s1;

      UNTIL l = 0; (*----- End loop thru all the numbers --------------------*)

      (*---------------------------------------------------------------------*)
      (* Add the "K" symbol                                                  *)
      (*---------------------------------------------------------------------*)

      d_num := s1 + 'K';

    END;

  (*=========================================================================*)
  (* Main line of display space                                              *)
  (*=========================================================================*)

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Get the drive information from the search arguement                   *)
    (*-----------------------------------------------------------------------*)

    s := FEXPAND(this_fsb^.fsb_path + search_arg);
    FSPLIT(s, s, nam, ext);

    (*-----------------------------------------------------------------------*)
    (* Convert the character back to a drive number                          *)
    (*-----------------------------------------------------------------------*)

    i := ORD(s[1]) - ORD('A') + 1;

    (*-----------------------------------------------------------------------*)
    (* Get the space information                                             *)
    (*-----------------------------------------------------------------------*)

    l1 := DISKFREE(i);
    l2 := DISKSIZE(i);

    (*-----------------------------------------------------------------------*)
    (* Get the space message header and save its length                      *)
    (*-----------------------------------------------------------------------*)

    s := get_message(message_disk_space) + ': ';
    i := LENGTH(s);

    (*-----------------------------------------------------------------------*)
    (* Put the whole message together                                        *)
    (*-----------------------------------------------------------------------*)

    s := s + d_num(space_sum) + ' / ' + d_num(l1) + ' / ' + d_num(l2) + cr;

    (*-----------------------------------------------------------------------*)
    (* If the line is too long then split it at the end of the header        *)
    (*-----------------------------------------------------------------------*)

    IF LENGTH(s) > 78 THEN
      BEGIN;
        send_tnc_data_str(copy(s, 1, i) + cr);
        IF more_check THEN
          EXIT;
        s := copy (s, i+1, 255);
      END;

    (*-----------------------------------------------------------------------*)
    (* Display remainder of message                                          *)
    (*-----------------------------------------------------------------------*)

    send_tnc_data_str(s);
    IF more_check THEN
      EXIT;

  END;
