(*===========================================================================*)
(* File subsystem -- Erase file                                              *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1991 by H. Roy Engehausen.  All rights reserved.  *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

UNIT BBFSZ;

INTERFACE

  PROCEDURE erase_file_cmd(cmd_string : STRING);

IMPLEMENTATION

  USES
    DOS,
    bbdummy,
    bbfsm,
    bbfssf,
    bbmdata,
    bbmess,
    bbmisc5,
    bbsema2,
    bbsdata,
    bbstr,
    bbtask,
    bbtime,
    bbwin,
    match;

  PROCEDURE erase_file_cmd(cmd_string : STRING);

  VAR
    dir_to_search : fsb_name_str;
    erase_me      : TEXT;
    i             : WORD;
    search_arg    : file_name_str;
    this_fsb      : fsb_ptr;
    word_count    : BYTE;

  {$I BBFSI.PAS}

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Parse command and execute right routine                               *)
    (*-----------------------------------------------------------------------*)

    upcase_str_var(cmd_string);

    word_count := words(cmd_string);

    IF word_count <> 3 THEN
      BEGIN;
        send_message(message_not_en);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Parse                                                                 *)
    (*-----------------------------------------------------------------------*)

    dir_to_search := subwordl(cmd_string, 2, SIZEOF(fsb_name_str) - 1);

    search_arg    := subwordl(cmd_string, 3, SIZEOF(search_arg)   - 1);

    (*-----------------------------------------------------------------------*)
    (* Find the directory                                                    *)
    (*-----------------------------------------------------------------------*)

    this_fsb := find_fsb(dir_to_search);

    IF (this_fsb = NIL) OR
                   (active_tcb^.uid_data.user_class < this_fsb^.fsb_up) THEN
      BEGIN;
        send_message(message_no_files_one);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Check for subdirectory                                                *)
    (*-----------------------------------------------------------------------*)

    IF (POS('\', search_arg) > 0) AND NOT this_fsb^.fsb_f_subdir_ok THEN
      BEGIN;
        send_message(message_no_slash);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Obtain the interrupt lock                                             *)
    (*-----------------------------------------------------------------------*)

    get_semaphore(semaphore_interrupts, sem_exclusive, FALSE);

    (*-----------------------------------------------------------------------*)
    (* Erase the file                                                        *)
    (*-----------------------------------------------------------------------*)

    ASSIGN(erase_me, this_fsb^.fsb_path + search_arg);
    {$I-}
    ERASE(erase_me);
    i := IORESULT;
    {$I+}

    (*-----------------------------------------------------------------------*)
    (* Free the interrupt lock                                               *)
    (*-----------------------------------------------------------------------*)

    free_semaphore(semaphore_interrupts);

    (*-----------------------------------------------------------------------*)
    (* Output error message as needed                                        *)
    (*-----------------------------------------------------------------------*)

    IF i <> 0 THEN
      send_tnc_data_str(cr + dos_err_message(i) + cr)
    ELSE
      send_message(message_action_complete);

  END;

END.
