(*===========================================================================*)
(* Forward task -- Routing data                                              *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991 by H. Roy Engehausen.  All rights      *)
(*   reserved.                                                               *)
(*                                                                           *)
(*===========================================================================*)

UNIT BBFWDD;

INTERFACE

  USES
    bbdummy;

(*---------------------------------------------------------------------------*)
(* Global definitions                                                        *)
(*---------------------------------------------------------------------------*)

CONST
  path_sort_none  = ' ';
  path_sort_type  = 'T';
  path_sort_age   = 'A';
  path_sort_date  = 'D';
  path_sort_rdate = 'R';
  path_sort_size  = 'S';

TYPE

  path_block = RECORD
            path_pattern    : bb_addr_str; (* Pattern to forward to           *)
            path_target     : bb_addr_str; (* Target for the current path     *)
            path_call       : call_sign_str; (* Callsign for the current path *)
            path_info       : STRING;      (* Place to read path info into    *)
            path_route      : msg_r_ptr;   (* Ptr to route block under test   *)
            path_msg_count  : BYTE;        (* Number of messages on this path *)
            path_msg_list   : msg_p_ptr;   (* Ptr to msg list for this path   *)
            path_port       : str5;        (* Port the path is on             *)
            path_port_p     : str8;        (* Pattern to match on port        *)
            path_write_p    : STRING[4];   (* Prefix to use when writing msgs *)
            path_depth      : BYTE;        (* Number of times around the loop *)
            path_line       : WORD;        (* Place in path file              *)
            path_age        : LONGINT;     (* Min time for a msg before fwd   *)
            path_size       : LONGINT;     (* Max size for a msg to fwd       *)
            path_route_inx  : BYTE;        (* Path route index.  Number of    *)
                                           (* word in route statement that    *)
                                           (* has been selected               *)
            fwd_slave_tcb   : tcb_ptr;     (* Ptr to sub-task's TCB           *)
            path_fe         : fe_ptr;      (* Ptr to path files' fe           *)
            path_sub_sw     : BOOLEAN;     (* If on, forward via sub-process  *)
            path_force_sw   : BOOLEAN;     (* If on, FORCE specified on cmd   *)
            path_force_this : BOOLEAN;     (* If on, FORCE this path          *)
            path_rev        : BOOLEAN;     (* If on, this is reverse forward  *)
            path_r_nochange : BOOLEAN;     (* If on, route list not changed   *)
            path_delay      : BOOLEAN;     (* If on, delay before path start  *)
            path_t_check    : BOOLEAN;     (* If on, time slot was checked    *)
            path_t_ok       : BOOLEAN;     (* If on, time slot is ok          *)
            path_did_msg    : BOOLEAN;     (* If on, atleast one msg was sent *)
            path_did_all    : BOOLEAN;     (* If on, did all the messages     *)
            path_type_rej   : BOOLEAN;     (* At least one message wrong type *)
            path_size_rej   : BOOLEAN;     (* At least one message wrong size *)
            path_date_rej   : BOOLEAN;     (* At least one message wrong date *)
            path_type_send  : STRING[10];  (* List of types to send           *)
            path_sort       : ARRAY[1..3] OF CHAR;
                                           (* Path sort algorithm             *)
          END;

  path_block_ptr = ^path_block;

  alias_ptr      = ^alias_type;

  alias_type     = RECORD
                     next_alias  : alias_ptr;
                     alias_name  : bb_addr_str;
                     alias_value : STRING;
                   END;

(*---------------------------------------------------------------------------*)
(* Global variables                                                          *)
(*---------------------------------------------------------------------------*)

VAR

  alias_list      : alias_ptr;

(*---------------------------------------------------------------------------*)
(* Maximum number of ports to schedule for a forward.                        *)
(*                                                                           *)
(* 1) Get length of field "path_port_p".                                     *)
(* 2) String to look like (xxxx)* so remove 3!                               *)
(*                                                                           *)
(*---------------------------------------------------------------------------*)

CONST
  max_ports_to_sked = 5;

IMPLEMENTATION

END.
