(*===========================================================================*)
(* Forward task -- Pathing                                                   *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991 by H. Roy Engehausen.  All rights      *)
(*   reserved.                                                               *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

{$UNDEF  DEBUG}
{$UNDEF  DEBUG_REMAIN}    (* Debug the path left stuff *)
{$UNDEF  DEBUG_REV}       (* Debug reverse forwarding *)
{$UNDEF  DEBUG_PATHMATCH} (* Debug path matching code *)
{$DEFINE DEBUG_PORT}      (* Debug port switch  *)
{$DEFINE POINT_CHK}

UNIT BBFWDP;

INTERFACE

USES
  bbfwdd;

PROCEDURE do_path(path_common : path_block_ptr);

IMPLEMENTATION

USES
  CRT,
  bbbpa,
  bbbug,
  bbdummy,
  bbexport,
  bbfwdc,
  bbfwdn,
  bbfwds,
  bbmem,
  bbmf,
  bbmisc,
  bbmisc3,
  bbmisc4,
  bbmisc5,
  bbstr,
  bbtask,
  bbtime,
  bbtrace,
  bbuf,
  bbwin,
  match;

(*===========================================================================*)
(* Find the paths                                                            *)
(*===========================================================================*)

PROCEDURE do_path(path_common : path_block_ptr);

  VAR
    i          : INTEGER;
    path_check : msg_r_ptr;
    t_str      : STRING[12];

  PROCEDURE do_this_path_part_0(path_common : path_block_ptr); FORWARD;
  PROCEDURE do_this_path_part_1(path_common : path_block_ptr); FORWARD;
  PROCEDURE do_this_path_part_2(path_common : path_block_ptr); FORWARD;

  {$I BBFWDP0.PAS}
  {$I BBFWDP1.PAS}
  {$I BBFWDP2.PAS}

  BEGIN;

    {$IFDEF DEBUG_PORT}
      trace_data('P  1', 0, active_port, '');
    {$ENDIF}

    WITH active_tcb^ DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Get proper console write prefix                                   *)
        (*-------------------------------------------------------------------*)

        IF path_common^.path_rev THEN
          path_common^.path_write_p := active_tcb^.port_chan_s + 'F:'
        ELSE
          path_common^.path_write_p := 'FO::';

        (*-------------------------------------------------------------------*)
        (* See if path fileblock is here already.  If not make one.  If so,  *)
        (* clean things up                                                   *)
        (*-------------------------------------------------------------------*)

        IF path_fe = NIL THEN
          NEW(path_fe)
        ELSE
          WITH path_fe^ DO
            BEGIN;

              IF NOT fe_type THEN
                BEGIN;
                  CLOSE(fe_text);
                  i := IORESULT;
                  {$I+}
                END
              ELSE
                BEGIN;
                  {$I-}
                  CLOSE(fe_bin);
                  i := IORESULT;
                  {$I+}
                END;

            END;

        (*-------------------------------------------------------------------*)
        (* Clean up the file block                                           *)
        (*-------------------------------------------------------------------*)

        FILLCHAR(path_fe^, SIZEOF(fe), CHR(0));

      END;

    (*-----------------------------------------------------------------------*)
    (* Chain everyting together                                              *)
    (*-----------------------------------------------------------------------*)

    path_common^.path_fe := active_tcb^.path_fe;

    (*-----------------------------------------------------------------------*)
    (* Now get ready to process the file                                     *)
    (*-----------------------------------------------------------------------*)

    WITH active_tcb^.path_fe^ DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Get the path file open                                            *)
        (*-------------------------------------------------------------------*)

        ASSIGN(fe_text, opt_block.path_fn);

        (*-------------------------------------------------------------------*)
        (* Loop for all route/path combos.  We are done when no more changes *)
        (* or if this was a reverse forward                                  *)
        (*-------------------------------------------------------------------*)

        path_common^.path_depth := 1;

        REPEAT

          (*-----------------------------------------------------------------*)
          (* Tell user what is happening                                     *)
          (*-----------------------------------------------------------------*)

          STR(path_common^.path_depth, t_str);
          window_write(path_common^.path_write_p,
                                        'Pass # ' + t_str + ' thru path file');

          (*-----------------------------------------------------------------*)
          (* Signal which routes will have been processed                    *)
          (*-----------------------------------------------------------------*)

          IF NOT path_common^.path_rev THEN
            BEGIN;

              path_check := msg_route_list;

              WHILE path_check <> NIL DO
                BEGIN;

                  {$IFDEF POINT_CHK}
                    test_pointer(path_check);
                  {$ENDIF}

                  path_check^.msg_r_scan := TRUE;
                  path_check             := path_check^.msg_r_next

                END;

            END;

          (*-----------------------------------------------------------------*)
          (* Ready the change flag                                           *)
          (*-----------------------------------------------------------------*)

          path_common^.path_r_nochange := TRUE;

          (*-----------------------------------------------------------------*)
          (* Process things                                                  *)
          (*-----------------------------------------------------------------*)

          {$IFDEF DEBUG_PORT}
            trace_data('P  2', 0, active_port, '');
          {$ENDIF}

          do_this_path_part_0(path_common);

          {$IFDEF DEBUG_PORT}
            trace_data('P  3', 0, active_port, '');
          {$ENDIF}

          (*-----------------------------------------------------------------*)
          (* Bump the count                                                  *)
          (*-----------------------------------------------------------------*)

          INC(path_common^.path_depth);

          (*-----------------------------------------------------------------*)
          (* Rest to allow anything waiting to occur                         *)
          (*-----------------------------------------------------------------*)

          task_switch;

          {$IFDEF DEBUG_PORT}
            trace_data('P  4', 0, active_port, '');
          {$ENDIF}

          (*-----------------------------------------------------------------*)
          (* Loop completes when no more changes or if this is reverse fwd   *)
          (*-----------------------------------------------------------------*)

        UNTIL path_common^.path_rev
                                    OR path_common^.path_force_sw
                                    OR path_common^.path_r_nochange
                                    OR (fwd_command = 'XSTOP');

      END;

    (*-----------------------------------------------------------------------*)
    (* All done so we can drop the path file                                 *)
    (*-----------------------------------------------------------------------*)

    WITH active_tcb^ DO
      BEGIN;
        DISPOSE(path_fe);
        path_common^.path_fe := NIL;
        path_fe              := NIL;
      END;

    (*-----------------------------------------------------------------------*)
    (* If reverse forwarding, force forwarding or command to stop then stop  *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG_PORT}
      trace_data('P  5', 0, active_port, '');
    {$ENDIF}

    IF path_common^.path_rev OR path_common^.path_force_sw
                                                OR (fwd_command = 'XSTOP') THEN
      EXIT;

    {$IFDEF DEBUG_PORT}
      trace_data('P  6', 0, active_port, '');
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* This code checks for routes without a path                            *)
    (*-----------------------------------------------------------------------*)

    fwd_route_clean_empty_route;

    path_check := msg_route_list;

    WHILE path_check <> NIL DO
      BEGIN;

        {$IFDEF POINT_CHK}
          test_pointer(path_check);
        {$ENDIF}

        IF (path_check^.msg_r_nroute <= path_check^.msg_r_routes)
                                                AND path_check^.msg_r_scan THEN
          BEGIN;

            t_str :=
                 SUBWORD(@path_check^.msg_r_info, path_check^.msg_r_nroute, 1);

            IF (t_str <> leave)
                                AND (t_str <> done)
                                AND (t_str <> qmark) THEN
              window_write(path_common^.path_write_p, '====> Route to '
                                    + t_str
                                    + ' does not seem to have a path defined');

            {$IFDEF DEBUG_REMAIN}
              WRITELN('Remainder -- ', path_check^.msg_r_info,
                               ' -- ', path_check^.msg_r_nroute);
            {$ENDIF}

          END;

        path_check := path_check^.msg_r_next

      END;

  END; (*----- End DO_PATH procedure ----------------------------------------*)


END.
