(*===========================================================================*)
(* Forward task -- Pathing                                                   *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991 by H. Roy Engehausen.  All rights      *)
(*   reserved.                                                               *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Find the paths                                                            *)
(*===========================================================================*)

PROCEDURE do_this_path_part_1(path_common : path_block_ptr);

  VAR
    this_path_selected : BOOLEAN;
    time_reject        : BOOLEAN;

  PROCEDURE get_port_defaults;

    VAR
      look_port : port_block_ptr;

    BEGIN;

      (*---------------------------------------------------------------------*)
      (* Find the port wanted                                                *)
      (*---------------------------------------------------------------------*)

      look_port := find_port_addr(path_common^.path_port[1]);

      IF look_port = NIL THEN
        EXIT;

      (*---------------------------------------------------------------------*)
      (* Set default path order                                              *)
      (*---------------------------------------------------------------------*)

      MOVE(look_port^.dflt_order[1],
           path_common^.path_sort,
           SIZEOF(path_common^.path_sort));

    END;

  BEGIN;

    WITH path_common^ DO
      BEGIN;

        {$IFDEF DEBUG_PORT}
          trace_data('P1 1', 0, active_port, '');
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Get the destination and port                                      *)
        (*-------------------------------------------------------------------*)

        path_target := 'FORCE';
        IF (path_depth = 1) AND (FIND(@path_info, @path_target) <> 0) THEN
          path_force_this := TRUE
        ELSE
          path_force_this := FALSE;

        path_target := subword(@path_info, 2, 1);

        path_port   := subword(@path_info, 3, 1);

        path_call   := subword(@path_info, 4, 1);
        IF path_call = '' THEN
          path_call := path_target;

        (*-------------------------------------------------------------------*)
        (* Complete initialization of path block                             *)
        (*-------------------------------------------------------------------*)

        path_age       := max_time;
        path_size      := 10000000;
        path_type_send := '';
        path_t_check   := FALSE;
        path_t_ok      := FALSE;

        get_port_defaults;

        (*-------------------------------------------------------------------*)
        (* Process all the "T" commands                                      *)
        (*-------------------------------------------------------------------*)

        this_path_selected := match_str(path_port, path_port_p)
                              AND (match_str(path_target, path_pattern)
                                        OR match_str(path_call, path_pattern));

        (*-------------------------------------------------------------------*)
        (* Process all the "T" commands                                      *)
        (*-------------------------------------------------------------------*)

        WITH path_common^.path_fe^ DO
          REPEAT

            IF EOF(fe_text) THEN EXIT;

            READLN(fe_text, path_info);

            strip_var(path_info, 'B');

            upcase_str_var(path_info);

        UNTIL NOT forward_time_script(path_common);
                                      (*----- Read until connect line loop --*)

        (*-------------------------------------------------------------------*)
        (* Did we have a time reject?                                        *)
        (*-------------------------------------------------------------------*)

        time_reject := path_t_check AND (NOT path_t_ok);

        {$IFDEF DEBUG_PORT}
          trace_data('P1 2', 0, active_port, '');
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Now find a route and messages that go down this path.             *)
        (*-------------------------------------------------------------------*)

        path_type_rej  := FALSE;
        path_size_rej  := FALSE;
        path_date_rej  := FALSE;
        path_did_all   := FALSE;

        build_path_array(path_common, NOT path_rev);

        (*-------------------------------------------------------------------*)
        (* Give rejected messages                                            *)
        (*-------------------------------------------------------------------*)

        IF (path_msg_list = NIL)
              AND this_path_selected
              AND (path_size_rej OR path_type_rej OR path_date_rej)
              AND (NOT time_reject) THEN
          BEGIN;
            IF path_size_rej THEN
              BEGIN;
                STR(path_size, path_info);
                window_write(path_write_p,
                              'At least one msg rejected because of size'
                              + ' -- Max was ' + path_info);
              END;
            IF path_type_rej THEN
              window_write(path_write_p,
                              'At least one msg rejected because of type'
                              + ' -- Filter was ' + path_type_send);
            IF path_date_rej THEN
              window_write(path_write_p,
                              'At least one message rejected because of age');

            window_write(path_write_p,
                            'Message transfer skipped because'
                            + ' all messages rejected on path '
                            + path_target);

          END;

        (*-------------------------------------------------------------------*)
        (* Debugging                                                         *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG}
          IF path_common^.path_msg_count > 0 THEN
            WRITELN('Path nonzero -- ', path_msg_count, ' / ',
                                                           p2x(path_msg_list));
        {$ENDIF}

        {$IFDEF DEBUG_REV}
          IF path_rev THEN
            trace_data('FWDP1-BPA', path_common^.path_msg_count,
                                    path_common^.path_msg_list,
                                    path_common^.path_target);
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* If no msgs were found, we can quit (but only if force is off)     *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_PORT}
          trace_data('P1 3', 0, active_port, '');
        {$ENDIF}

        IF path_msg_list = NIL THEN
          BEGIN;
            IF (NOT path_force_sw) AND (NOT path_force_this) THEN EXIT;
            IF path_rev THEN EXIT;
          END;

        {$IFDEF DEBUG}
          WRITELN('Path has something -- ', path_target, ' / ', path_call,
                                                         ' / ', path_pattern);
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Say we did something                                              *)
        (*-------------------------------------------------------------------*)

        IF (NOT path_rev) THEN
          path_common^.path_did_msg := TRUE;

        (*-------------------------------------------------------------------*)
        (* Now see if the pattern and port allow us to forward               *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_PATHMATCH}
          trace_data('path match1', 0, NIL, path_port + '/' + path_port_p);
          trace_data('path match2', 0, NIL, path_target + '/' + path_pattern);
          trace_data('path match3', 0, NIL, path_call + '/' + path_pattern);
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Do the path as necessary                                          *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_PORT}
          trace_data('P1 4', 0, active_port, '');
        {$ENDIF}

        IF this_path_selected THEN
          BEGIN;
            IF NOT time_reject THEN
              do_this_path_part_2(path_common)
            ELSE
              BEGIN;
                STR(path_msg_count, path_info);
                window_write(path_write_p,
                           path_info + ' outbound msgs for path '
                           + path_target + ' skipped because not proper time');
              END;
          END;

        {$IFDEF DEBUG_PORT}
          trace_data('P1 5', 0, active_port, '');
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* We are now done with forwarding -- free list                      *)
        (*-------------------------------------------------------------------*)

        free_task_mem(path_block_mem_id, TRUE);

        (*-------------------------------------------------------------------*)
        (* If we finished the path then set the pointer so we show the       *)
        (* route is exhausted                                                *)
        (*-------------------------------------------------------------------*)

        IF path_did_all THEN
          fwd_route_mark_done(path_common)
        ELSE
          free_task_mem(path_block_lst_id, TRUE);

      END; (*----- End of loop cleaning things up ---------------------------*)

  END;
