(*===========================================================================*)
(* Forward task -- Routing -- Alias                                          *)
(*                                                                           *)
(*   Copyright 1988, 1989 by H. Roy Engehausen.  All rights reserved.        *)
(*                                                                           *)
(*===========================================================================*)

{$UNDEF  DEBUG_ALIAS_DUPE1}
{$UNDEF  DEBUG_ALIAS_DUPE2}

(*===========================================================================*)
(* Does the actual aliasing                                                  *)
(*     target = the string to be updated                                     *)
(*     a_ptr  = pointer to the alias block                                   *)
(*     w_index = the # of word in the target to be replaced by the alias     *)
(*===========================================================================*)

PROCEDURE alias_insert(VAR target : STRING; a_ptr : alias_ptr; w_index : BYTE);

  VAR
    i        : WORD;
    temp_str : STRING;

  BEGIN;

    WITH a_ptr^ DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Get the remaining piece of the target                             *)
        (*-------------------------------------------------------------------*)

        temp_str := subword(@target, w_index + 1, 0);

        (*-------------------------------------------------------------------*)
        (* Truncate the target at the word before the one to be replaced     *)
        (*-------------------------------------------------------------------*)

        IF w_index = 1 THEN
          target := ''
        ELSE
          target := subword(@target, 1, w_index - 1);

        (*-------------------------------------------------------------------*)
        (* Verify that this will all fit when we are done                    *)
        (*-------------------------------------------------------------------*)

        i := 255 - (LENGTH(target) + LENGTH(temp_str));

        IF i <= LENGTH(alias_value) THEN
          BEGIN;
            window_write(prefix_str , 'Too long of a route list for '
                                       + route_for + ' during alias processing.');
            target := target + ' ' + alias_value;
            EXIT;
          END;

        (*-------------------------------------------------------------------*)
        (* Put the alias in then put the target remainder back in            *)
        (*-------------------------------------------------------------------*)

        target := target + ' ' + alias_value + ' ' + temp_str;

      END;

  END;

(*===========================================================================*)
(* See if we need to do an alias substitution                                *)
(*===========================================================================*)

PROCEDURE do_alias_sub;

  VAR
    done : STRING;
    i    : BYTE;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Loop thru all the aliases looking for a match                         *)
    (*-----------------------------------------------------------------------*)

    work_alias := alias_list;
    done       := '';

    WHILE work_alias <> NIL DO
      WITH work_alias^ DO
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* See if we have a match and we haven't already done it.  If so   *)
          (* substitute it                                                   *)
          (*-----------------------------------------------------------------*)

          i := find(@route_info, @alias_name);
          IF (i <> 0) AND (find(@done, @alias_name) = 0) THEN
            BEGIN;
              done := done + ' ' + alias_name;
              alias_insert(route_info, work_alias, i);
            END;

          (*-----------------------------------------------------------------*)
          (* On to next alias                                                *)
          (*-----------------------------------------------------------------*)

          work_alias := next_alias;

        END;

  END;

(*===========================================================================*)
(* Called when the input line shows that an alias is to be done              *)
(*===========================================================================*)

PROCEDURE do_alias_def;

  VAR

    code       : INTEGER;
    i          : BYTE;
    j          : INTEGER;
    new_alias  : alias_ptr;
    w          : STRING[15];

  LABEL
    dupe_found,
    loop_iterate,
    loop_leave;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* First check that the alias name we are defining is not too long!      *)
    (*-----------------------------------------------------------------------*)

    IF LENGTH(route_for) >= SIZEOF(bb_addr_str) THEN
      BEGIN;
        window_write(prefix_str, 'Alias name of ' + route_for
                                                          + ' is too long.');
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Now grab the alias info                                               *)
    (*-----------------------------------------------------------------------*)

    route_info := subword(@route_info, 2, 0);

    (*-----------------------------------------------------------------------*)
    (* Make a pass thru any already defined aliases                          *)
    (*-----------------------------------------------------------------------*)

    work_alias := alias_list;

    WHILE work_alias <> NIL DO
      WITH work_alias^ DO
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* Check for a duplicate alias                                     *)
          (*-----------------------------------------------------------------*)

          IF alias_name = route_for THEN
            BEGIN;
              window_write(prefix_str, 'Duplicate alias for ' + route_for
                                                         +  ' ignored!');
              EXIT;
            END;

          (*-----------------------------------------------------------------*)
          (* Now see if we need to substitute an existing alias into the     *)
          (* alias that we are building                                      *)
          (*-----------------------------------------------------------------*)

          i := find(@route_info, @alias_name);
          IF i <> 0 THEN
            alias_insert(route_info, work_alias, i);

          (*-----------------------------------------------------------------*)
          (* All done, chain to next alias                                   *)
          (*-----------------------------------------------------------------*)

          work_alias := next_alias

        END; (*----- End loop thru alias file -------------------------------*)

    (*-----------------------------------------------------------------------*)
    (* Remove duplicate routes                                               *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG_ALIAS_DUPE1}
      WRITELN('Dupe check start -- ', route_info);
    {$ENDIF}

    i := 1;

    REPEAT

      (*---------------------------------------------------------------------*)
      (* Iterate the loop counter                                            *)
      (*---------------------------------------------------------------------*)

loop_iterate:

      INC(i);
      IF i > WORDS(route_info) THEN
        GOTO loop_leave;

      (*---------------------------------------------------------------------*)
      (* Get word to check                                                   *)
      (*---------------------------------------------------------------------*)

      w := subword(@route_info, i, 1);

      (*---------------------------------------------------------------------*)
      (* Skip numbers                                                        *)
      (*---------------------------------------------------------------------*)

      IF LENGTH(w) <= 4 THEN
        BEGIN;
          VAL(w, j, code);
          IF (code = 0) AND (j >= 0) AND (j <= 255) THEN
            GOTO loop_iterate;
        END;

      (*---------------------------------------------------------------------*)
      (* See if it is a dupe.  If not then iterate                           *)
      (*---------------------------------------------------------------------*)

      FOR j := 1 TO i-1 DO
        BEGIN;
          {$IFDEF DEBUG_ALIAS_DUPE2}
            WRITELN('Check -- ', j, '/', i, '/', w,
                                    '/', subword(@route_info, j, 1),
                                    '/', subword(@route_info, j, 1) = w);
            DELAY(500);
          {$ENDIF}

          IF subword(@route_info, j, 1) = w THEN
            GOTO dupe_found;

        END;

      GOTO loop_iterate;

      (*---------------------------------------------------------------------*)
      (* Remove dupe                                                         *)
      (*---------------------------------------------------------------------*)

dupe_found:

      route_info := subword(@route_info, 1, i-1) + ' '
                                                + subword(@route_info, i+1, 0);

      (*---------------------------------------------------------------------*)
      (* Decrement the word position pointer so we test the same position    *)
      (* again.  This is because we are now one word shorter                 *)
      (*---------------------------------------------------------------------*)

      DEC(i); (* This forces a test of same word pos again *)

    UNTIL FALSE; (*----- End dupe check loop --------------------------------*)

loop_leave:

    {$IFDEF DEBUG_ALIAS_DUPE1}
      WRITELN('Dupe check end -- ', route_info);
      DELAY(2000);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Create a new alias, move the data in and add it to the chain          *)
    (*-----------------------------------------------------------------------*)

    NEW(new_alias);

    WITH new_alias^ DO
      BEGIN;
        next_alias  := alias_list;
        alias_name  := route_for;
        alias_value := route_info;
      END;

    alias_list := new_alias;

  END;
