(*===========================================================================*)
(* Forward task -- Cleanup                                                   *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991 by H. Roy Engehausen.  All rights      *)
(*   reserved.                                                               *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Clean up aliases                                                          *)
(*===========================================================================*)

PROCEDURE fwd_clean_aliases;

  VAR
    work_alias   : alias_ptr;

  BEGIN;

    WHILE alias_list <> NIL DO
      BEGIN;

        {$IFDEF POINT_CHK}
          test_pointer(alias_list);
        {$ENDIF}

        work_alias := alias_list;
        alias_list := work_alias^.next_alias;
        DISPOSE(work_alias);

        {$IFDEF FREE_CHECK}
          test_free_list;
        {$ENDIF}

      END;

  END;

(*===========================================================================*)
(* Clean up all messages                                                     *)
(*===========================================================================*)

PROCEDURE fwd_route_clean_all_msg;

  VAR
    this_msg     : msg_index_ptr;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Loop down the msg list cleaning them all                              *)
    (*-----------------------------------------------------------------------*)

    this_msg := msg_index_start;
    WHILE this_msg <> NIL DO
      BEGIN;

        {$IFDEF POINT_CHK}
          test_pointer(this_msg);
        {$ENDIF}

        fwd_route_clean_msg(this_msg);
        this_msg := this_msg^.msg_i_next;

      END; (*----- End loop resetting forwarding flags ------------------*)

    (*-----------------------------------------------------------------------*)
    (* Ensure we process the whole message database                          *)
    (*-----------------------------------------------------------------------*)

    msg_route_num := 0;

  END;

(*===========================================================================*)
(* Clean up routes.  This destroys all routes!                               *)
(*===========================================================================*)

PROCEDURE fwd_route_clean;

  VAR
    next_route   : msg_r_ptr;
    this_route   : msg_r_ptr;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Loop down the route list and dispose of them all                      *)
    (*-----------------------------------------------------------------------*)

    next_route := msg_route_list;

    WHILE next_route <> NIL DO
      BEGIN;

        {$IFDEF POINT_CHK}
          test_pointer(next_route);
        {$ENDIF}

        this_route := next_route;

        {$IFDEF ROUTE_CLEAN}
          WRITELN('Route clean -- ', this_route^.msg_r_info);
        {$ENDIF}

        next_route := this_route^.msg_r_next;
        DISPOSE(this_route);

        {$IFDEF FREE_CHECK}
          test_free_list;
        {$ENDIF}

      END; (*----- End of free route list loop ------------------------------*)

    (*-----------------------------------------------------------------------*)
    (* Clean the chain pointers                                              *)
    (*-----------------------------------------------------------------------*)

    msg_route_list := NIL;

    (*-----------------------------------------------------------------------*)
    (* Clean the alias list as needed                                        *)
    (*-----------------------------------------------------------------------*)

    fwd_clean_aliases;

    (*-----------------------------------------------------------------------*)
    (* Clean all the messages                                                *)
    (*-----------------------------------------------------------------------*)

    fwd_route_clean_all_msg;

  END;
