(*===========================================================================*)
(* Forward task -- Routing -- Calculate routes                               *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Figure out routes                                                         *)
(*===========================================================================*)

PROCEDURE do_route_build;

  VAR
    b             : BOOLEAN;
    d_block       : msg_d_ptr;
    d_cnt         : BYTE;
    i             : WORD;
    new_route     : BOOLEAN;
    m_cnt         : WORD;

  BEGIN;

    WITH search_block DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Search for anyone wanting to go to the destination                *)
        (*-------------------------------------------------------------------*)

        search_last := NIL;
        search_str  := route_for;

        forward_search_msg(@search_block);

        (*-------------------------------------------------------------------*)
        (* If we were successful in finding some one, we better build the    *)
        (* proper route control blocks                                       *)
        (*-------------------------------------------------------------------*)

        IF search_last <> NIL THEN
          BEGIN;

            {$IFDEF POINT_CHK}
              test_pointer(search_last);
            {$ENDIF}

            (*---------------------------------------------------------------*)
            (* Do any alias substitution necessary                           *)
            (*---------------------------------------------------------------*)

            do_alias_sub;

            (*---------------------------------------------------------------*)
            (* See if we can find an existing route block                    *)
            (*---------------------------------------------------------------*)

            this_route    := msg_route_list;

            WHILE (this_route <> NIL)
                                 AND (this_route^.msg_r_info <> route_info) DO
              BEGIN;

                {$IFDEF POINT_CHK}
                  test_pointer(this_route);
                {$ENDIF}

                this_route    := this_route^.msg_r_next;

              END;

            (*---------------------------------------------------------------*)
            (* Create a new route block if necessary or say we didn't        *)
            (*---------------------------------------------------------------*)

            IF this_route <> NIL THEN
              BEGIN;
                {$IFDEF POINT_CHK}
                  test_pointer(this_route);
                {$ENDIF}
                new_route := FALSE
              END
            ELSE
              BEGIN;

                new_route := TRUE;

                (*-----------------------------------------------------------*)
                (* Get a route block                                         *)
                (*-----------------------------------------------------------*)

                NEW(this_route);

                (*-----------------------------------------------------------*)
                (* Initialize the route block                                *)
                (*-----------------------------------------------------------*)

                FILLCHAR(this_route^, SIZEOF(msg_route_block_type), 0);
                WITH this_route^ DO
                  BEGIN;
                    msg_r_routes := words(route_info);
                    msg_r_info   := route_info;
                    fwd_route_reset(this_route);
                  END;

                (*-----------------------------------------------------------*)
                (* Chain route on list of routes                             *)
                (*-----------------------------------------------------------*)

                IF msg_route_list = NIL THEN
                  msg_route_list := this_route
                ELSE
                  BEGIN;
                    {$IFDEF POINT_CHK}
                      test_pointer(msg_route_end);
                    {$ENDIF}
                    msg_route_end^.msg_r_next := this_route;
                  END;

                msg_route_end := this_route;

              END;

            (*---------------------------------------------------------------*)
            (* Address the route block                                       *)
            (*---------------------------------------------------------------*)

            WITH this_route^ DO
              BEGIN;

                {$IFDEF POINT_CHK}
                  test_pointer(this_route);
                {$ENDIF}

                (*-----------------------------------------------------------*)
                (* Now loop for any other messages for this route and save   *)
                (*-----------------------------------------------------------*)

                m_cnt := 0;

                REPEAT

                  {$IFDEF POINT_CHK}
                    test_pointer(search_last);
                  {$ENDIF}

                  WITH search_last^ DO
                    BEGIN;

                      (*-----------------------------------------------------*)
                      (* Save the data for the msg we just found.  Test      *)
                      (* before increment to prevent overflow                *)
                      (*-----------------------------------------------------*)

                      INC(m_cnt);

                      (*-----------------------------------------------------*)
                      (* Assume the select flag should be turned on          *)
                      (*-----------------------------------------------------*)

                      b := TRUE;

                      (*-----------------------------------------------------*)
                      (* If no distribution list, things are simple.  If not *)
                      (* then lots of work to do                             *)
                      (*-----------------------------------------------------*)

                      IF search_fi = 0 THEN
                        msg_i_rou := this_route
                      ELSE
                        BEGIN;

                          (*-------------------------------------------------*)
                          (* This is a select of a message with distribution *)
                          (*-------------------------------------------------*)

                          (*-------------------------------------------------*)
                          (* Build the extra control block if not already    *)
                          (* in the chain.  If we already have one, just     *)
                          (* get the pointers we need                        *)
                          (*-------------------------------------------------*)

                          d_block := find_dist_list(search_last);
                          d_cnt   := d_block^.msg_d_no;

                          {$IFDEF POINT_CHK}
                            test_pointer(d_block);
                          {$ENDIF}

                          IF (msg_i_mb.msg_flag AND mf_disrout) = 0 THEN
                            BEGIN;
                              i := WORD(d_cnt) * SIZEOF(msg_dr_route_item)
                                                         + SIZEOF(msg_d_ptr);
                              GETMEM(msg_i_dr, i);
                              FILLCHAR(msg_i_dr^, i, #0);
                              msg_i_dr^.msg_dr_dblk := d_block;
                              msg_i_mb.msg_flag :=
                                               msg_i_mb.msg_flag OR mf_disrout;
                            END;

                          (*-------------------------------------------------*)
                          (* Bug check                                       *)
                          (*-------------------------------------------------*)

                          IF (search_fi > d_cnt)
                                            OR (search_fi > msg_dist_max) THEN
                            BEGIN;
                              WRITELN('FWDRD Search fail in dist -- ',
                                       search_fi, ' -- ', d_cnt);
                              dump_trace;
                              dump_msg(search_last);
                              HALT;
                            END;

                          (*-------------------------------------------------*)
                          (* Set the pointer to the route block              *)
                          (*-------------------------------------------------*)

                          msg_i_dr^.msg_dr_data[search_fi] := this_route;

                          (*-------------------------------------------------*)
                          (* Set the select flag in the distribution list    *)
                          (*-------------------------------------------------*)

                          WITH d_block^.msg_d_array[search_fi] DO
                            msg_d_flag := msg_d_flag OR df_fwd_select;

                          (*-------------------------------------------------*)
                          (* Loop thru and see if all are selected           *)
                          (*-------------------------------------------------*)

                          i := 1;
                          REPEAT
                            b := b AND
                                     (d_block^.msg_d_array[i].msg_d_flag <> 0);
                            INC(i);
                          UNTIL (NOT b) OR (i > d_cnt);

                        END; (*----- End of handling distribution list ------*)

                      (*-----------------------------------------------------*)
                      (* Turn on select flag if we should                    *)
                      (*-----------------------------------------------------*)

                      IF b THEN
                        msg_i_mb.msg_flag :=
                                            msg_i_mb.msg_flag OR mf_fwd_select;

                    END;

                  (*---------------------------------------------------------*)
                  (* Conduct another search                                  *)
                  (*---------------------------------------------------------*)

                  forward_search_msg(@search_block);

                UNTIL search_last = NIL; (*----- Loop until search fails ----*)

                (*-----------------------------------------------------------*)
                (* Tell what we are doing                                    *)
                (*-----------------------------------------------------------*)

                STR(m_cnt, n_str);
                window_write(prefix_str, '**** Routed ' + n_str
                                              + ' message(s) to ' + route_for);

                (*-----------------------------------------------------------*)
                (* Signal that at least one message has been routed          *)
                (*-----------------------------------------------------------*)

                fnm_linked_list_valid := FALSE;

              END; (*----- End handling this route --------------------------*)


          END; (*----- End successful find of message for route -------------*)

        (*-------------------------------------------------------------------*)
        (* Now that we have completed one search, the "n_list" is built so   *)
        (* use it for all the rest of the searched.                          *)
        (*-------------------------------------------------------------------*)

        use_n_list := TRUE;

    END; (*----- End WITH statement -----------------------------------------*)

  END;
