(*===========================================================================*)
(* Forward task -- Search for message                                        *)
(*                                                                           *)
(*   Copyright 1991 by H. Roy Engehausen.  All rights reserved.              *)
(*                                                                           *)
(*===========================================================================*)

VAR
  n_list            : msg_index_ptr;

(*===========================================================================*)
(* Search for a msg to be forwarded                                          *)
(*===========================================================================*)

PROCEDURE forward_search_msg(search : search_block_ptr);

  CONST
    dont_fwd_flags  : msg_flag_type = mf_hold
                                   OR mf_read
                                   OR mf_fwd
                                   OR mf_fwd_select
                                   OR mf_kill
                                   OR mf_review;

  VAR
    b                    : BOOLEAN;
    found                : BOOLEAN;
    i                    : BYTE;
    msg_index_current    : msg_index_ptr;
    our_search           : search_block_type;
    save_last            : msg_index_ptr;

  {$I BBMF3A.PAS}

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Save search block                                                     *)
    (*-----------------------------------------------------------------------*)

    our_search        := search^;

    (*-----------------------------------------------------------------------*)
    (* Debugging                                                             *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG3B}
       trace_data('M31', this_search_block^.search_msg_no,
                         our_search.search_last,
                         '');
    {$ENDIF}

    {$IFDEF DEBUG3C}
       trace_data('MF3 1', ORD(our_search.search_type),
                         our_search.search_last,
                         '');
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Initialize for the search loop.  The start of the search should       *)
    (* be the next msg from our last hit.  If this is the start of a         *)
    (* search then use one of the ends of the chain.  Note that there        *)
    (* are two chains: the main chain of all messages and the "n_list"       *)
    (* chain.  The latter contains messages that may be awaiting forward.    *)
    (* This chain is a lot shorter and speeds up forwarding                  *)
    (*-----------------------------------------------------------------------*)

    IF our_search.search_last <> NIL THEN
      BEGIN;
        IF use_n_list THEN
          msg_index_current   := our_search.search_last^.msg_i_fwd_l
        ELSE
          msg_index_current := our_search.search_last^.msg_i_next;
      END
    ELSE
      IF use_n_list THEN
        msg_index_current := n_list
      ELSE
        BEGIN;
          n_list := NIL;
          msg_index_current := msg_index_start;
        END;

    (*-----------------------------------------------------------------------*)
    (* Debugging                                                             *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG3B}
       trace_data('M32', this_search_block^.search_msg_no,
                         our_search.search_last,
                         '');
    {$ENDIF}

    {$IFDEF DEBUG3C}
       trace_data('MF3 2', ORD(our_search.search_type),
                         our_search.search_last,
                         '');
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Set our switches off and prep the save pointer.                       *)
    (*-----------------------------------------------------------------------*)

    found     := FALSE;
    save_last := our_search.search_last;

    (*-----------------------------------------------------------------------*)
    (* This loop is the search loop                                          *)
    (*-----------------------------------------------------------------------*)

    WHILE (NOT found) AND (msg_index_current <> NIL) DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* If eligible then do it                                            *)
        (*-------------------------------------------------------------------*)

        IF (msg_index_current^.msg_i_mb.msg_flag AND dont_fwd_flags) = 0 THEN
          BEGIN;

            (*---------------------------------------------------------------*)
            (* If we are not doing a search using the n_list                 *)
            (* the we are going to build one                                 *)
            (*---------------------------------------------------------------*)

            IF NOT use_n_list THEN
              BEGIN;
                IF save_last <> NIL THEN
                  save_last^.msg_i_fwd_l := msg_index_current
                ELSE
                  n_list                 := msg_index_current;

                msg_index_current^.msg_i_fwd_l := NIL;
              END;

            save_last := msg_index_current;

            (*---------------------------------------------------------------*)
            (* Now search the various fields                                 *)
            (*---------------------------------------------------------------*)

            do_fwd_search;

          END (*----- End bad flag test -------------------------------------*)

        ELSE

            (*---------------------------------------------------------------*)
            (* If we are using the "n_list" then remove this                 *)
            (* guy from the list                                             *)
            (*---------------------------------------------------------------*)

            IF use_n_list THEN
              BEGIN;
                IF save_last <> NIL THEN
                  save_last^.msg_i_fwd_l := msg_index_current^.msg_i_fwd_l
                ELSE
                  n_list := msg_index_current^.msg_i_fwd_l;
              END;

        (*-------------------------------------------------------------------*)
        (* If we didn't find the right one and there are more to search      *)
        (* then point to the next message to be tested                       *)
        (*-------------------------------------------------------------------*)

        IF (NOT found) AND (msg_index_current <> NIL) THEN
          BEGIN;
            IF use_n_list THEN
              msg_index_current := msg_index_current^.msg_i_fwd_l
            ELSE
              msg_index_current := msg_index_current^.msg_i_next;
          END;

      END; (*----- End search loop ------------------------------------------*)

    (*-----------------------------------------------------------------------*)
    (* Set the search results                                                *)
    (*-----------------------------------------------------------------------*)

    IF found THEN
      search^.search_last := msg_index_current
    ELSE
      search^.search_last := NIL;

    search^.search_fi := our_search.search_fi;

  END;
