(*===========================================================================*)
(* Forward task -- Routing -- Main line                                      *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

{$UNDEF  DEBUGLOOP} (* Show where in loop we are *)
{$UNDEF  DEBUG_RMR} (* Debug initial route/alias table *)

(*===========================================================================*)
(* Figure out routes                                                         *)
(*===========================================================================*)

PROCEDURE do_route;

  VAR
    b             : BOOLEAN;
    i             : BYTE;
    j             : BYTE;
    high_msg      : WORD;
    l_count       : WORD;
    msg_route_end : msg_r_ptr;
    n_str         : STRING[10];
    next_route    : msg_r_ptr;
    prefix_str    : STRING[4];
    route_file    : TEXT;
    route_for     : STRING[30];
    route_info    : STRING;
    search_block  : search_block_type;
    switch_count  : BYTE;
    this_route    : msg_r_ptr;
    use_n_list    : BOOLEAN;
    w_cnt         : BYTE;
    work_alias    : alias_ptr;

  LABEL read_loop;

  {$I BBFWDRF.PAS}                        (* Find message to route           *)
  {$I BBFWDRC.PAS}                        (* Handle various resettings       *)
  {$I BBFWDRA.PAS}                        (* Handle Aliases                  *)
  {$I BBFWDRD.PAS}                        (* Do route definitions            *)
  {$I BBFWDRL.PAS}                        (* Handle LEAVE                    *)

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Set up the prefix for messages                                        *)
    (*-----------------------------------------------------------------------*)

    prefix_str := active_tcb^.port_chan_s + '::';

    (*-----------------------------------------------------------------------*)
    (* This code is serially reusable!                                       *)
    (*-----------------------------------------------------------------------*)

    get_semaphore(semaphore_fwd_route_build, sem_exclusive, TRUE);

    (*-----------------------------------------------------------------------*)
    (* See if routes are active.                                             *)
    (*-----------------------------------------------------------------------*)

    b := test_semaphore(semaphore_fwd_route_use) <> sem_clear;

    (*-----------------------------------------------------------------------*)
    (* Lock routes as used.  This is unlocked much later after the path      *)
    (* code runs                                                             *)
    (*-----------------------------------------------------------------------*)

    get_semaphore(semaphore_fwd_route_use, sem_shared, TRUE);

    (*-----------------------------------------------------------------------*)
    (* Handle forced reset of all routes                                     *)
    (*-----------------------------------------------------------------------*)

    IF msg_route_force THEN
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* If routes are active then skip doing anything                     *)
        (*-------------------------------------------------------------------*)

        IF b THEN
          BEGIN;
            window_write(prefix_str,
    'Fwding aborted because routes are in use and were changed by GF command');
            free_semaphore(semaphore_fwd_route_build);
            EXIT;
          END;

        (*-------------------------------------------------------------------*)
        (* Reset the message number and force flag                           *)
        (*-------------------------------------------------------------------*)

        msg_route_num  := 0;
        msg_route_force := FALSE;

      END; (*----- End handling forced msg route reset ----------------------*)

    (*-----------------------------------------------------------------------*)
    (* Is the route active or is this only partial                           *)
    (*-----------------------------------------------------------------------*)

    IF b OR (msg_route_num <> 0) THEN
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Routes active and we are going to use them                        *)
        (*-------------------------------------------------------------------*)

        fwd_route_clean_empty_route;

        (*-------------------------------------------------------------------*)
        (* If master forward task then reset the routes                      *)
        (*-------------------------------------------------------------------*)

        IF active_tcb^.tcb_type = th_fwd_control THEN
          BEGIN;

            window_write(prefix_str, 'Route reset in progress');
            this_route := msg_route_list;
            WHILE this_route <> NIL DO
              BEGIN;
                fwd_route_reset(this_route);
                {$IFDEF DEBUG_RMR}
                  WRITELN('Route ',
                          this_route^.msg_r_nroute,
                          '/',
                          this_route^.msg_r_routes,
                          ' -- ',
                          this_route^.msg_r_info);
                {$ENDIF}
                this_route := this_route^.msg_r_next;
              END;

          END;

      END
    ELSE

      (*---------------------------------------------------------------------*)
      (* Routes not active and we want to start from scratch                 *)
      (*---------------------------------------------------------------------*)

      fwd_route_clean;

    (*-----------------------------------------------------------------------*)
    (* IF no messages to be routed then lets leave now                       *)
    (*-----------------------------------------------------------------------*)

    IF msg_route_num = next_msg_no THEN
      BEGIN;
        window_write(prefix_str, 'No messages awaiting routing');
        free_semaphore(semaphore_fwd_route_build);
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Tell user what is happening                                           *)
    (*-----------------------------------------------------------------------*)

    IF msg_route_num <> 0 THEN
      BEGIN;
        STR(msg_route_num, n_str);
        window_write(prefix_str,
                              'First message number to be routed is ' + n_str);
      END
    ELSE
      window_write(prefix_str, 'Routing all messages');

    (*-----------------------------------------------------------------------*)
    (* Throw away the alias list                                             *)
    (*-----------------------------------------------------------------------*)

    fwd_clean_aliases; (* Not sure this works right *)

    (*-----------------------------------------------------------------------*)
    (* Save the number of the last message processed                         *)
    (*-----------------------------------------------------------------------*)

    high_msg := next_msg_no;

    (*-----------------------------------------------------------------------*)
    (* Find the end of the route blocks                                      *)
    (*-----------------------------------------------------------------------*)

    msg_route_end := NIL;

    this_route := msg_route_list;

    WHILE this_route <> NIL DO
      BEGIN;

        {$IFDEF POINT_CHK}
          test_pointer(this_route);
        {$ENDIF}

        msg_route_end := this_route;
        this_route    := this_route^.msg_r_next;
      END;

    (*-----------------------------------------------------------------------*)
    (* Initialize the search block                                           *)
    (*-----------------------------------------------------------------------*)

    FILLCHAR(search_block, SIZEOF(search_block), #0);

    WITH search_block DO
      BEGIN;
        search_nok    := TRUE;
        search_ascend := TRUE;
        search_type   := '!';
      END;

    (*-----------------------------------------------------------------------*)
    (* Set to build n-list                                                   *)
    (*-----------------------------------------------------------------------*)

    use_n_list := FALSE;

    (*-----------------------------------------------------------------------*)
    (* Get the route file open                                               *)
    (*-----------------------------------------------------------------------*)

    ASSIGN(route_file, opt_block.route_fn);

    {$I-}
    RESET(route_file);
    {$I+}

    i := IORESULT;

    IF i <> 0 THEN
      BEGIN;
        window_write(prefix_str, 'Error opening forward file (' +
                                  opt_block.route_fn + ')');

        window_write(prefix_str, dos_err_message(i));

        free_semaphore(semaphore_fwd_route_use);
        free_semaphore(semaphore_fwd_route_build);

        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Loop around reading the route file                                    *)
    (*-----------------------------------------------------------------------*)

    l_count := 0;
    switch_count := 0;

read_loop:

    WHILE NOT EOF(route_file) DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Switch tasks                                                      *)
        (*-------------------------------------------------------------------*)

        task_switch;

        (*-------------------------------------------------------------------*)
        (* If it as been a while, check for pending tnc things               *)
        (*-------------------------------------------------------------------*)

        IF switch_count > 5 THEN
          BEGIN;
            IF active_tcb^.tcb_type <> th_fwd_control THEN
              i := send_pending(FALSE);
            switch_count := 0;
          END
        ELSE
          INC(switch_count);

        (*-------------------------------------------------------------------*)
        (* Bump line counter                                                 *)
        (*-------------------------------------------------------------------*)

        INC(l_count);

        (*-------------------------------------------------------------------*)
        (* Read a line from the route file and process it                    *)
        (*-------------------------------------------------------------------*)

        READLN(route_file, route_info);
        strip_var(route_info, 'B');

        IF (route_info = '') OR (route_info[1] = ';') THEN
          GOTO read_loop;

        upcase_str_var(route_info);

        w_cnt     := WORDS(route_info);

        (*-------------------------------------------------------------------*)
        (* Process line according to number of words in it                   *)
        (*     1 - Error                                                     *)
        (*    >1 - OK!                                                       *)
        (*-------------------------------------------------------------------*)

        IF w_cnt = 1 THEN
          BEGIN;
            STR(l_count, n_str);
            window_write(prefix_str,
                        'Probable error in forward file on line ' + n_str);
            GOTO read_loop;
          END;

        (*-------------------------------------------------------------------*)
        (* Valid line .. Figure out what to do with it                       *)
        (*-------------------------------------------------------------------*)

        route_for  := subword(@route_info, 1, 1);
        route_info := subword(@route_info, 2, 0);

        (*-------------------------------------------------------------------*)
        (* If there is an '=', then its an alias                             *)
        (*-------------------------------------------------------------------*)

        IF subword(@route_info, 1, 1) = '=' THEN
          do_alias_def
        ELSE
          do_route_build;

      END; (*----- End loop thru all routes ---------------------------------*)

    (*-----------------------------------------------------------------------*)
    (* All done with route file.  We can destroy it                          *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUGLOOP}
      WRITELN('Close route file');
    {$ENDIF}

    CLOSE(route_file);

    (*-----------------------------------------------------------------------*)
    (* Throw away the alias list                                             *)
    (*-----------------------------------------------------------------------*)

    fwd_clean_aliases;

    (*-----------------------------------------------------------------------*)
    (* Show messages processed to high message                               *)
    (*-----------------------------------------------------------------------*)

    msg_route_num := high_msg;

    (*-----------------------------------------------------------------------*)
    (* See if any of the routes are to LEAVE.  If so, reset the messages,    *)
    (* and drop the route.                                                   *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUGLOOP}
      WRITELN('Leave loop free');
    {$ENDIF}

    do_leave;

    (*-----------------------------------------------------------------------*)
    (* Free the route semaphore.  This releases any other forwards           *)
    (* that are pending                                                      *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUGLOOP}
      WRITELN('Sema free');
    {$ENDIF}

    free_semaphore(semaphore_fwd_route_build);

    {$IFDEF DEBUGLOOP}
      WRITELN('Route build exit');
    {$ENDIF}

  END;
