(*===========================================================================*)
(* Forward task -- Forward a message                                         *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Forward a message                                                         *)
(*===========================================================================*)

PROCEDURE forward_msg(path_common : path_block_ptr; mi_ptr : msg_index_ptr);

  VAR
    rejected : BOOLEAN;
    temp_str : STRING;

  (*=========================================================================*)
  (* This subroutine figures out if a bid should be sent                     *)
  (*=========================================================================*)

  FUNCTION send_bid : BOOLEAN;
    BEGIN;

      WITH mi_ptr^.msg_i_mb, active_tcb^ DO
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* Assume true                                                     *)
          (*-----------------------------------------------------------------*)

          send_bid := TRUE;

          (*-----------------------------------------------------------------*)
          (* IF MIDS then send bid                                           *)
          (*-----------------------------------------------------------------*)

          IF tcb_mids_ok THEN EXIT;

          CASE msg_type OF

            (*---------------------------------------------------------------*)
            (* If NTS then fall thru                                         *)
            (*---------------------------------------------------------------*)

            mt_nts: ;

            (*---------------------------------------------------------------*)
            (* If PRIVATE and there is a distribution list, send bid         *)
            (* Otherwise continue                                            *)
            (*---------------------------------------------------------------*)

            mt_private : IF (msg_flag AND mf_fwd_list) <> 0 THEN EXIT;

            (*---------------------------------------------------------------*)
            (* Type of blank.  Send a bid if there is a list or if a bid     *)
            (* exists                                                        *)
            (*---------------------------------------------------------------*)

            ' ': IF ((msg_flag AND mf_fwd_list) <> 0) OR (msg_bid <> '') THEN
                     EXIT;

            (*---------------------------------------------------------------*)
            (* Send bids for all other message types                         *)
            (*---------------------------------------------------------------*)

            ELSE
              EXIT;

          END; (*----- End CASE on message_type -----------------------------*)

          (*-----------------------------------------------------------------*)
          (* If we get here, don't send a BID                                *)
          (*-----------------------------------------------------------------*)

          send_bid := FALSE;

        END; (*----- End SEND_BID Subroutine --------------------------------*)

    END;

  (*=========================================================================*)
  (* Subroutine to send the send command                                     *)
  (*=========================================================================*)

  PROCEDURE send_send_command;

    VAR
      t_str : STRING;

    BEGIN;

      rejected := FALSE;

      WITH mi_ptr^.msg_i_mb DO
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* Build the send command                                          *)
          (*-----------------------------------------------------------------*)

          t_str := 'S' + msg_type + ' ' + msg_to;

          IF msg_to_at <> '' THEN
            BEGIN;
              t_str := t_str + ' @ ' + msg_to_at;
              IF (msg_to_h <> '') AND active_tcb^.tcb_h_ok THEN
                t_str := t_str + '.' + msg_to_h;
            END;

          t_str := t_str + ' < ' + msg_from;

          (*-----------------------------------------------------------------*)
          (* Add bid as needed                                               *)
          (*-----------------------------------------------------------------*)

          IF active_tcb^.tcb_abbs AND (msg_bid <> '') AND send_bid THEN
              t_str := t_str + ' $' + msg_bid;

          (*-----------------------------------------------------------------*)
          (* Send the send command                                           *)
          (*-----------------------------------------------------------------*)

          send_tnc_data_str(t_str + cr);

          (*-----------------------------------------------------------------*)
          (* If this is advanced BBS mode (BIDS) then wait for the OK        *)
          (*-----------------------------------------------------------------*)

          IF active_tcb^.tcb_bid_level > 0 THEN
            BEGIN;

              (*-------------------------------------------------------------*)
              (* Get the response                                            *)
              (*-------------------------------------------------------------*)

              t_str := read_tnc_data_str;
              strip_crlf(t_str);
              t_str := upcase_str(subword(@t_str, 1, 1));

              (*-------------------------------------------------------------*)
              (* Strip a trailing comma                                      *)
              (*-------------------------------------------------------------*)

              IF (LENGTH(t_str) > 0) AND (t_str[LENGTH(t_str)] = ',') THEN
                t_str[0] := CHR(ORD(t_str[0]) - 1);

              (*-------------------------------------------------------------*)
              (* Get just the single response character                      *)
              (*-------------------------------------------------------------*)

              response_char := t_str[1];

              (*-------------------------------------------------------------*)
              (* If rejected, wait for a prompt then we are done             *)
              (*-------------------------------------------------------------*)

              IF (t_str = 'NO') OR (t_str = 'N')
                      OR (t_str = 'LATER') OR (t_str = 'L')
                      OR (t_str = 'REJECT') OR (t_str = 'R') THEN
                BEGIN;
                  forward_wait_prompt(path_common, FALSE);
                  rejected := TRUE;
                  EXIT;
                END;

              (*-------------------------------------------------------------*)
              (* Maybe bad bid answer?                                       *)
              (*-------------------------------------------------------------*)

              IF (t_str <> 'OK') AND (t_str <> 'O') THEN
                BEGIN;
                  window_write(active_tcb^.port_chan_s + 'F:',
                                           'Improper BID answer');
                  end_session(TRUE);
                END;

            END;

          (*-----------------------------------------------------------------*)
          (* Send the subject                                                *)
          (*-----------------------------------------------------------------*)

          send_tnc_data_str(msg_subj + cr);

          (*-----------------------------------------------------------------*)
          (* Wait for any two responses.                                     *)
          (*-----------------------------------------------------------------*)

          IF NOT active_tcb^.tcb_abbs THEN
            BEGIN;
              t_str := read_tnc_data_str;
              t_str := read_tnc_data_str;
              response_char := 'O';
            END
          END;

    END;

  (*=========================================================================*)
  (* Main line                                                               *)
  (*=========================================================================*)

  BEGIN;

    WITH mi_ptr^.msg_i_mb DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Move the info to the current msg location                         *)
        (*-------------------------------------------------------------------*)

        active_tcb^.curr_msg := mi_ptr^;

        (*-------------------------------------------------------------------*)
        (* Send the first stuff                                              *)
        (*-------------------------------------------------------------------*)

        send_send_command;
        IF rejected THEN EXIT;

        (*-------------------------------------------------------------------*)
        (* Time zone correction to header is send in proper time             *)
        (*-------------------------------------------------------------------*)

        INC(active_tcb^.curr_msg.msg_i_mb.msg_dt_in,   opt_block.z_time_fwd);
        INC(active_tcb^.curr_msg.msg_i_mb.msg_dt_orig, opt_block.z_time_fwd);

        (*-------------------------------------------------------------------*)
        (* Send the new forward header line.                                 *)
        (*-------------------------------------------------------------------*)

        IF NOT opt_block.opt_personal_bbs THEN
          BEGIN;

IF ((active_tcb^.curr_msg.msg_i_mb.msg_from_at = opt_block.this_bb_addr)
             OR ((active_tcb^.curr_msg.msg_i_mb.msg_flag AND mf_bid_change)<>0))
       AND NOT opt_block.opt_no_alt_header THEN
  temp_str := get_message(message_alt_header)
ELSE
  temp_str := get_message(message_fwd_head);

            send_tnc_data_str(temp_str + cr);

          END;

        (*-------------------------------------------------------------------*)
        (* Open the file                                                     *)
        (*-------------------------------------------------------------------*)

        STR(msg_number, temp_str);
        temp_str := opt_block.msg_file_dir + 'BB' + temp_str + '.MSG';

        (*-------------------------------------------------------------------*)
        (* Use send a file                                                   *)
        (*-------------------------------------------------------------------*)

        active_tcb^.error_sw := FALSE;
        send_file(temp_str, FALSE);

        (*-------------------------------------------------------------------*)
        (* Handle error on OPEN                                              *)
        (*-------------------------------------------------------------------*)

        IF active_tcb^.error_sw THEN
          BEGIN;
            end_session(TRUE);
            EXIT;
          END;

        (*-------------------------------------------------------------------*)
        (* Send the end of the file                                          *)
        (*-------------------------------------------------------------------*)

        send_tnc_data_str(eof_chr + cr);

        (*-------------------------------------------------------------------*)
        (* Wait for a prompt                                                 *)
        (*-------------------------------------------------------------------*)

        forward_wait_prompt(path_common, FALSE);

      END;

  END;
