(*===========================================================================*)
(* Forward task -- Main route                                                *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Main routine to forward                                                   *)
(*   At this point, the path file is positioned right after the line         *)
(*   describing the path definition.                                         *)
(*                                                                           *)
(* This does the connect script....                                          *)
(*===========================================================================*)

{$UNDEF DBGSPD}

PROCEDURE forward_main(path_common : path_block_ptr);

  VAR
    i        : BYTE;
    l_err    : BYTE;
    modem_sw : BOOLEAN;
    p        : STRING[4];

  LABEL loop;

  (*=========================================================================*)
  (* Subroutine to destroy a task                                            *)
  (*=========================================================================*)

  PROCEDURE destroy_me;
    BEGIN;
      cmd_tnc(@disc_cmd, TRUE);
      task_destroy_active;
    END;

  (*=========================================================================*)
  (* Process incoming data line from a modem                                 *)
  (*=========================================================================*)

  PROCEDURE process_modem_data;

    VAR
      s : STRING[20];

    BEGIN;

      WITH active_tcb^, active_port^ DO
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* Move the data into the input buffer.  This is so we can use     *)
          (* read_tnc_data_str to glkue it back together                     *)
          (*-----------------------------------------------------------------*)

          {$IFDEF DBGSPD}
            trace_data('MDB0', tnc_data.long_length, NIL, tnc_data.str_data);
            trace_data('MDB1', i_data.long_length, NIL, i_data.str_data);
            trace_data('MDB2', ORD(modem_sw), NIL, '');
          {$ENDIF}

          MOVE(tnc_data, i_data, tnc_data.long_length + str_m_oh);

          {$IFDEF DBGSPD}
            trace_data('MDS0', tnc_data.long_length, NIL, tnc_data.str_data);
            trace_data('MDS1', i_data.long_length, NIL, i_data.str_data);
            trace_data('MDS2', ORD(modem_sw), NIL, '');
          {$ENDIF}

          (*-----------------------------------------------------------------*)
          (* Loop until we have processed all the input                      *)
          (*-----------------------------------------------------------------*)

          REPEAT

            (*---------------------------------------------------------------*)
            (* Get some input                                                *)
            (*---------------------------------------------------------------*)

            cr_timeout := 2;
            s := read_tnc_data_str;
            cr_timeout := 0;

            {$IFDEF DBGSPD}
              trace_data('MDT1', LENGTH(s), NIL, s);
            {$ENDIF}

            (*---------------------------------------------------------------*)
            (* Upcase it without any garbage                                 *)
            (*---------------------------------------------------------------*)

            strip_crlf(s);
            upcase_str_var(s);

            {$IFDEF DBGSPD}
              trace_data('MDT2', LENGTH(s), NIL, s);
            {$ENDIF}

            (*---------------------------------------------------------------*)
            (* NO CARRIER?                                                   *)
            (*---------------------------------------------------------------*)

            IF s = no_carrier THEN
              destroy_me;

            (*---------------------------------------------------------------*)
            (* CONNECT XXXX?                                                 *)
            (*---------------------------------------------------------------*)

            IF subword(@s, 1, 1) = connect THEN
              BEGIN;

                s := subword(@s, 2, 1);
                connect_speed(s);

                {$IFDEF DBGSPD}
                  trace_data('MDU1', LENGTH(s), NIL, s);
                {$ENDIF}

                i_data.long_length := 0;

              END;

          UNTIL i_data.long_length <= 1; (*----- Loop thru input data -------*)

        END;

    END;

  (*=========================================================================*)
  (* Main line                                                               *)
  (*=========================================================================*)

  BEGIN;

    {$IFDEF DEBUG_PORT}
      trace_data('SMP0', 0, active_port, '');
    {$ENDIF}

    WITH path_common^, path_common^.path_fe^ DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Initialize a few things                                           *)
        (*-------------------------------------------------------------------*)

        p        := active_tcb^.port_chan_s + 'M:';
        modem_sw := FALSE;

        (*-------------------------------------------------------------------*)
        (* Read in lines until we reach connect                              *)
        (*-------------------------------------------------------------------*)

        WHILE (path_info[1] <> 'C') AND (path_info[1] <> ' ') DO
          BEGIN;

            forward_script(path_common, TRUE, modem_sw);

            IF EOF(fe_text) THEN EXIT;

            READLN(fe_text, path_info);

            strip_var(path_info, 'B');

            upcase_str_var(path_info);

          END; (*----- end read until connect line loop --*)

        (*-------------------------------------------------------------------*)
        (* If no sub task then do it!                                        *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_PORT}
          trace_data('SMP1', 0, active_port, '');
        {$ENDIF}

        IF NOT path_common^.path_sub_sw THEN
          BEGIN;
            forward_loop(path_common);
            {$IFDEF DEBUG_PORT}
              trace_data('SMP2', 0, active_port, '');
            {$ENDIF}
            EXIT;
          END;

        {$IFDEF DEBUG_PORT}
          trace_data('SMP3', 0, active_port, '');
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Set modem switch                                                  *)
        (*-------------------------------------------------------------------*)

        modem_sw := (active_port^.port_type = port_modem)
                                 OR (active_port^.port_type = port_null_modem);

        (*-------------------------------------------------------------------*)
        (* If this is a telephone modem, get rid of any left overs           *)
        (* If this is a telephone modem, set to tolerate some garbage before *)
        (* link start appears                                                *)
        (*-------------------------------------------------------------------*)

        IF modem_sw THEN
          BEGIN;
            gc;
            l_err := 2;
            active_port^.modem_dial := TRUE;
            active_port^.modem_conn := FALSE;
          END
        ELSE
          l_err := 0;

        (*-------------------------------------------------------------------*)
        (* Issue the connect.  It must be one to exit that loop              *)
        (*-------------------------------------------------------------------*)

        connect_format(@path_info, active_port);

        cmd_tnc(@path_info, TRUE);
        IF NOT active_tcb^.tnc_null THEN
          task_destroy_active;

        (*-------------------------------------------------------------------*)
        (* If this is a telephone modem, set to tolerate some garbage before *)
        (* link start appears                                                *)
        (*-------------------------------------------------------------------*)

        IF modem_sw THEN
          BEGIN;
            l_err := 2;
            active_port^.modem_crlf := FALSE;
          END
        ELSE
          l_err := 0;

        (*-------------------------------------------------------------------*)
        (* This label is used for recovery from the PK-232 and its           *)
        (* insistance of sending two disconnect status messages under        *)
        (* certain conditions (e.g. busy)                                    *)
        (* This label is also used for telephone modems to recover from      *)
        (* non-link status messages before connect                           *)
        (*-------------------------------------------------------------------*)

loop:

        (*-------------------------------------------------------------------*)
        (* Wait for some word from the TNC                                   *)
        (*-------------------------------------------------------------------*)

        active_tcb^.tnc_null := TRUE;

        WHILE active_tcb^.tnc_null
                              AND NOT (modem_sw AND active_port^.modem_conn) DO
          BEGIN;
            task_switch;
            send_recv_tnc(2);
          END;

        (*-------------------------------------------------------------------*)
        (* Get the addressing                                                *)
        (*-------------------------------------------------------------------*)

        WITH active_tcb^ DO
          BEGIN;

            {$IFDEF DBGSPD}
              trace_data('MDL0', tnc_data.long_length, NIL, tnc_data.str_data);
              trace_data('MDL1', i_data.long_length, NIL, i_data.str_data);
              trace_data('MDL2', ORD(modem_sw), NIL, '');
              trace_data('MDL3', ORD(active_port^.modem_conn), NIL, '');
            {$ENDIF}

            (*---------------------------------------------------------------*)
            (* It must be a LINK start                                       *)
            (*---------------------------------------------------------------*)

            IF (tnc_type = t_to_h_links)
                                 OR (modem_sw AND active_port^.modem_conn) THEN
              BEGIN;

                (*-----------------------------------------------------------*)
                (* Process link status change                                *)
                (*-----------------------------------------------------------*)

                IF tnc_type <> t_to_h_links THEN
                  BEGIN;
                    tnc_data.str_data    := modem_conn;
                    tnc_data.long_length := LENGTH(tnc_data.str_data);
                    tnc_type := t_to_h_links;
                    tnc_null := FALSE;
                  END;

                IF modem_sw THEN
                  active_port^.modem_dial := FALSE;

                IF tnc_data.str_data[1] = '(' THEN
                  tnc_data.str_data := subword(@tnc_data.str_data, 2, 0);

                IF (subword(@tnc_data.str_data, 1, 2) + ' ') = t_to_h_ct THEN
                  link_start
                ELSE
                  IF find(@tnc_data.str_data, @low_busy) > 0 THEN
                    GOTO loop
                  ELSE
                    BEGIN;
                      window_write(p, tnc_data.str_data);
                      destroy_me;
                    END;

              END
            ELSE
              BEGIN;

                (*-----------------------------------------------------------*)
                (* Handle the non linkstart stuff                            *)
                (*-----------------------------------------------------------*)

                (*-----------------------------------------------------------*)
                (* Do any display needed                                     *)
                (*-----------------------------------------------------------*)

                IF (l_err = 0) OR NOT modem_sw THEN
                  BEGIN;
                    window_write(p, 'First thing from TNC is not link status! -- '
                                                     + byte_to_char[tnc_type]);
                    window_write(p, tnc_data.str_data);
                  END;

                (*-----------------------------------------------------------*)
                (* If we can't tolerate errors then quit;                    *)
                (*-----------------------------------------------------------*)

                IF l_err = 0 THEN
                  destroy_me;

                (*-----------------------------------------------------------*)
                (* Decrement error count                                     *)
                (*-----------------------------------------------------------*)

                DEC(l_err);

                (*-----------------------------------------------------------*)
                (* If this is not a modem, loop now                          *)
                (*-----------------------------------------------------------*)

                IF NOT modem_sw THEN
                  GOTO loop;

                (*-----------------------------------------------------------*)
                (* This is a modem.  Reconstruct things so we can glue       *)
                (* lines back together                                       *)
                (*-----------------------------------------------------------*)

                process_modem_data;
                GOTO loop;

              END;

          END;

        IF modem_sw THEN
          active_port^.modem_dial := FALSE;

        (*-------------------------------------------------------------------*)
        (* Signal we will need a delay before next path starts               *)
        (*-------------------------------------------------------------------*)

        path_delay := TRUE;

        (*-------------------------------------------------------------------*)
        (* Now loop thru any more connect script.                            *)
        (*-------------------------------------------------------------------*)

        WHILE NOT EOF(fe_text) AND (path_info <> '') DO
          BEGIN;
            READLN(fe_text, path_info);
            strip_var(path_info, 'B');
            IF path_info <> '' THEN
              forward_script(path_common, FALSE, modem_sw);
          END;

        (*-------------------------------------------------------------------*)
        (* Wait for the initial prompt                                       *)
        (*-------------------------------------------------------------------*)

        forward_wait_prompt(path_common, TRUE);

        (*-------------------------------------------------------------------*)
        (* Tell advanced mode as necessary                                   *)
        (*-------------------------------------------------------------------*)

        IF active_tcb^.tcb_abbs THEN
          BEGIN;
            send_tnc_data_str(this_bbs_handshake + cr);
            forward_wait_prompt(path_common, FALSE);
          END;

        (*-------------------------------------------------------------------*)
        (* All connected -- Loop thru all the routes via this path           *)
        (*-------------------------------------------------------------------*)

        forward_loop(path_common);

    END;

  END;
