(*===========================================================================*)
(* Forward task -- Handle script                                             *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Handle script lines                                                       *)
(*===========================================================================*)

PROCEDURE forward_script(path_common : path_block_ptr;
                         b4_connect  : BOOLEAN;
                     VAR modem_drop  : BOOLEAN);

  VAR
    i          : INTEGER;
    j          : INTEGER;
    modem_port : BOOLEAN;
    p          : STRING[4];
    t          : LONGINT;
    t_ok       : BOOLEAN;
    w_cnt      : BYTE;
    work       : STRING;

  (*=========================================================================*)
  (* Clean up and leave                                                      *)
  (*=========================================================================*)

  PROCEDURE end_it;
    BEGIN;
      IF b4_connect THEN
        task_destroy_active
      ELSE
        end_session(FALSE);
    END;

  (*=========================================================================*)
  (* Main line of forward script                                             *)
  (*=========================================================================*)

  BEGIN;

    modem_port := (active_port^.port_type = port_modem)
                                 OR (active_port^.port_type = port_null_modem);

    p := active_tcb^.port_chan_s + 'L:';

    WITH path_common^ DO
      BEGIN;

        {$IFDEF DEBUG_SCR}
          WRITELN('Script line -- ', path_info);
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Handle statements that can occur anytime                          *)
        (*-------------------------------------------------------------------*)

        IF comment_line(path_info) THEN
          EXIT;

        (*-------------------------------------------------------------------*)
        (* Statements with a command word                                    *)
        (*-------------------------------------------------------------------*)

        (*-------------------------------------------------------------------*)
        (* One letter commands                                               *)
        (*-------------------------------------------------------------------*)

        CASE path_info[1] OF

          (*-----------------------------------------------------------------*)
          (* Comment                                                         *)
          (*-----------------------------------------------------------------*)

          ';': EXIT;

          (*-----------------------------------------------------------------*)
          (* Command to port                                                 *)
          (*-----------------------------------------------------------------*)

          'P': BEGIN;

                 (*----------------------------------------------------------*)
                 (* Don't do port command if time is not right               *)
                 (*----------------------------------------------------------*)

                 IF path_t_check
                                 AND (NOT path_t_ok)
                                 AND (NOT path_force_sw) THEN
                   EXIT;

                 (*----------------------------------------------------------*)
                 (* Send port command to TNC                                 *)
                 (*----------------------------------------------------------*)

                 path_info := COPY(path_info, 2, 255);
                 cmd_tnc(@path_info, TRUE);

                 EXIT;

               END;

         (*------------------------------------------------------------------*)
         (* Handle sort order                                                *)
         (*------------------------------------------------------------------*)

         'O': BEGIN;

                work := subword(@path_common^.path_info, 2, 1) + '   ';
                MOVE(work[1], path_common^.path_sort,
                                               SIZEOF(path_common^.path_sort));
                EXIT;

              END;

          (*-----------------------------------------------------------------*)
          (* Start and end time                                              *)
          (*-----------------------------------------------------------------*)

          'T': BEGIN;

                 (*----------------------------------------------------------*)
                 (* Cannot happen here                                       *)
                 (*----------------------------------------------------------*)

                 window_write(p,  '"T" not in proper sequence: '
                                                                 + path_info);
                     end_it;


               END; (*----- End "T" script ----------------------------------*)

        END; (*----- End before connect CASE statement ----------------------*)

        (*-------------------------------------------------------------------*)
        (* If we reach here, the before-connect statements are all done      *)
        (*-------------------------------------------------------------------*)

        IF b4_connect THEN
          BEGIN;
            window_write(p,
                     'Illegal line before connect in path file: ' + path_info);
            task_destroy_active;
          END;

        (*-------------------------------------------------------------------*)
        (* Handle statements that can occur after connect                    *)
        (*-------------------------------------------------------------------*)

        CASE path_info[1] OF

          (*-----------------------------------------------------------------*)
          (* Send data                                                       *)
          (*-----------------------------------------------------------------*)

          'S': send_tnc_data_str(SUBSTR(path_info, 2, 0) + cr);

          (*-----------------------------------------------------------------*)
          (* Wait for a work                                             *)
          (*-----------------------------------------------------------------*)

          'R': BEGIN;

                 (*----------------------------------------------------------*)
                 (* This loop gets the data from the TNC and handles special *)
                 (* stuff from modems                                        *)
                 (*----------------------------------------------------------*)

                 REPEAT
                   IF modem_port THEN
                     active_port^.cr_timeout := 5;
                   work := read_tnc_data_str;
                   IF modem_port THEN
                     active_port^.cr_timeout := 0;
                   strip_crlf(work);
                   IF modem_drop AND (work <> '') THEN
                     BEGIN;
                       IF NOT match_str(work, 'CONNECT*') THEN
                         modem_drop := FALSE
                       ELSE
                         BEGIN;
                           work := subword(@work, 2, 1);
                           connect_speed(work);
                         END;
                     END;
                 UNTIL NOT modem_drop;

                 (*----------------------------------------------------------*)
                 (* Handle mismatches                                        *)
                 (*----------------------------------------------------------*)

                 IF NOT match_str(work, substr(path_info, 2, 0)) THEN
                   BEGIN;

                     IF opt_block.opt_show_mismatch THEN
                       BEGIN;
                         window_write(p, 'MISMATCH');
                         p[4] := '<';
                         window_write(p, work);
                         p[4] := '>';
                         window_write(p, path_info);
                       END;

                     path_delay := FALSE;
                     end_session(TRUE);

                   END;
               END;

          (*-----------------------------------------------------------------*)
          (* Delay                                                           *)
          (*-----------------------------------------------------------------*)

          'D': BEGIN;

                 work := subword(@path_info, 2, 1);
                 VAL(work, i, j);
                 IF (work = '') OR (j <> 0) OR (i < 1) OR (i > 255) THEN
                   BEGIN;
                     window_write(p, 'Illegal delay in path file: '
                                      + path_info);
                     path_delay := FALSE;
                     end_session(FALSE);
                   END;

                 task_wait(i, FALSE);

               END;


          (*-----------------------------------------------------------------*)
          (* All others                                                      *)
          (*-----------------------------------------------------------------*)

          ELSE
               BEGIN;
                 window_write(p, 'Illegal line in path file: ' + path_info);
                 path_delay := FALSE;
                 end_session(FALSE);
               END;

        END; (*----- End after connect CASE statement -----------------------*)

      END;

  END; (*----- End forward script -------------------------------------------*)
