(*===========================================================================*)
(* Forward task -- Handle script lines for time only!                        *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Handle script lines for time only.  Return True if handled                *)
(*===========================================================================*)

FUNCTION forward_time_script(path_common : path_block_ptr) : BOOLEAN;

  VAR
    err        : BOOLEAN;
    i          : INTEGER;
    j          : INTEGER;
    response   : STRING;
    p          : STRING[4];
    t          : LONGINT;
    t_ok       : BOOLEAN;
    w_index    : BYTE;
    w_cnt      : BYTE;

  (*=========================================================================*)
  (* Variable part of "T" command                                            *)
  (*=========================================================================*)

  PROCEDURE vary_t;
    VAR
      hour_div : INTEGER;
      hour_rem : INTEGER;

    BEGIN;

      (*---------------------------------------------------------------------*)
      (* Break out the word and increment the counter                        *)
      (*---------------------------------------------------------------------*)

      response := subword(@path_common^.path_info, w_index, 1);
      INC(w_index);

      {$IFDEF DEBUG_SIZE}
        WRITELN('Response = ', response);
      {$ENDIF}

      {$IFDEF DEBUG_TYPE}
        WRITELN('Response = ', response);
      {$ENDIF}

      (*---------------------------------------------------------------------*)
      (* Process REVERSE NOREVERSE                                           *)
      (*---------------------------------------------------------------------*)

      IF response = 'REVERSE' THEN
        BEGIN;
          t_ok := path_common^.path_rev;
          EXIT;
        END;

      IF response = 'NOREVERSE' THEN
        BEGIN;
          t_ok := NOT path_common^.path_rev;
          EXIT;
        END;

      (*---------------------------------------------------------------------*)
      (* Process FORCE/NOFORCE                                               *)
      (*---------------------------------------------------------------------*)

      IF response = 'FORCE' THEN
        BEGIN;
          t_ok := path_common^.path_force_sw;
          EXIT;
        END;

      IF response = 'NOFORCE' THEN
        BEGIN;
          t_ok := NOT path_common^.path_force_sw;
          EXIT;
        END;

      (*---------------------------------------------------------------------*)
      (* Things beyond this point require two words                          *)
      (*---------------------------------------------------------------------*)

      IF w_index > w_cnt THEN
        BEGIN;
          window_write(p, 'Bad or missing word in script: '
                           + path_common^.path_info);
          err := TRUE;
          EXIT;
        END;

      (*---------------------------------------------------------------------*)
      (* Save the word pointer for later and then increment it               *)
      (*---------------------------------------------------------------------*)

      i := w_index;
      INC(w_index);

      (*---------------------------------------------------------------------*)
      (* Process hour                                                        *)
      (*---------------------------------------------------------------------*)

      IF response = 'HOUR' THEN
        BEGIN;

          response := subword(@path_common^.path_info, i, 1);

          hour_div := ORD(response[1]) - ORD('0');
          hour_rem := ORD(response[3]) - ORD('0');

          IF (NOT match_str(response, '#/#')) OR (hour_rem >= hour_div) THEN
            BEGIN;
              window_write(p, 'Bad HOUR divider on "T" command: '
                                                     + path_common^.path_info);

              err := TRUE;
              EXIT;
            END;

          t := current_day_time MOD ticks_per_day;
          j := t DIV ticks_per_hour;
          IF (j MOD hour_div) <> hour_rem THEN
            t_ok := FALSE;

          EXIT;

        END;

      (*---------------------------------------------------------------------*)
      (* Process order                                                       *)
      (*---------------------------------------------------------------------*)

      IF response = 'ORDER' THEN
        BEGIN;

          response := subword(@path_common^.path_info, i, 1) + '   ';
          MOVE(response[1], path_common^.path_sort,
                                               SIZEOF(path_common^.path_sort));

          EXIT;

        END;

      (*---------------------------------------------------------------------*)
      (* Process type                                                        *)
      (*---------------------------------------------------------------------*)

      IF response = 'TYPE' THEN
        BEGIN;

          path_common^.path_type_send :=
                                      subword(@path_common^.path_info, i, 1);

          {$IFDEF DEBUG_SIZE}
            WRITELN('Type = ', path_common^.path_type_send);
          {$ENDIF}

          t := POS('_', path_common^.path_type_send);
          IF t > 0 THEN
            path_common^.path_type_send[t] := ' ';
          EXIT;

        END;

      (*---------------------------------------------------------------------*)
      (* Process size                                                        *)
      (*---------------------------------------------------------------------*)

      IF response = 'SIZE' THEN
        BEGIN;

          response := subword(@path_common^.path_info, i, 1);

          {$IFDEF DEBUG_SIZE}
            WRITELN('Size # string =', response);
          {$ENDIF}

          VAL(response, t, hour_div);
          IF (hour_div <> 0) OR (t < 1) OR (t > 1000000) THEN
            BEGIN;
              window_write(p, 'Bad SIZE on "T" command: '
                                                 + path_common^.path_info);
              err := TRUE;
              EXIT;
            END;

          {$IFDEF DEBUG_SIZE}
            WRITELN('Size # value = ', t);
          {$ENDIF}

          path_common^.path_size := t;

          EXIT;

        END;

      (*---------------------------------------------------------------------*)
      (* If we reach here we had a bad word                                  *)
      (*---------------------------------------------------------------------*)

      window_write(p, 'Bad word on "T" command: ' + path_common^.path_info);

      err := TRUE;

    END;

  (*=========================================================================*)
  (* Main line of forward time script                                        *)
  (*=========================================================================*)

  BEGIN;

    p := active_tcb^.port_chan_s + 'L:';

    WITH path_common^ DO
      BEGIN;

        forward_time_script := TRUE;

        (*-------------------------------------------------------------------*)
        (* Look for comments                                                 *)
        (*-------------------------------------------------------------------*)

        IF comment_line(path_info) THEN
          EXIT;

        IF path_info[1] <> 'T' THEN
          BEGIN;
            forward_time_script := FALSE;
            EXIT;
          END;

        {$IFDEF DEBUG_T}
          WRITELN('T process 1 -- ', path_info);
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* We are running a time check                                       *)
        (*-------------------------------------------------------------------*)

        path_t_check := TRUE;

        (*-------------------------------------------------------------------*)
        (* If a previous time check was OK, skip this one                    *)
        (*-------------------------------------------------------------------*)

        IF path_t_ok THEN
          EXIT;

        {$IFDEF DEBUG_T}
          WRITELN('T process 2');
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Check statement format                                            *)
        (*-------------------------------------------------------------------*)

        w_cnt := words(path_info);
        IF w_cnt < 2 THEN
          BEGIN;
            window_write(p, 'Bad word count in script: ' + path_info);
            EXIT;
          END;

        (*-------------------------------------------------------------------*)
        (* Start time                                                        *)
        (*-------------------------------------------------------------------*)

        response := substr(todays_date_time, 1, 6) + subword(@path_info, 2, 1);

        t := time_unstr(response, err);
        IF err THEN
          BEGIN;
            window_write(p, 'Bad start time in path file: ' + path_info);
            EXIT;
          END;

        IF current_day_time < t THEN
          EXIT;

        {$IFDEF DEBUG_T}
          WRITELN('T process 4');
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* End time                                                          *)
        (*-------------------------------------------------------------------*)

        IF w_cnt = 2 THEN
          BEGIN;
            path_t_ok := TRUE;
            EXIT;
          END;

        response := substr(todays_date_time, 1, 6)
                                     + subword(@path_info, 3, 1);

        t := time_unstr(response, err);
        IF err THEN
          BEGIN;
            window_write(p, 'Bad end time in path file: ' + path_info);
            EXIT;
          END;

        IF current_day_time > t THEN
          EXIT;

        {$IFDEF DEBUG_T}
          WRITELN('T process 5');
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Start DOW or only DOW                                             *)
        (*-------------------------------------------------------------------*)

        IF w_cnt = 3 THEN
          BEGIN;
            path_t_ok := TRUE;
            EXIT;
          END;

        {$IFDEF DEBUG_T}
          WRITELN('T process 6');
        {$ENDIF}

        VAL(subword(@path_info, 4, 1), i, j);
        IF (j <> 0) OR (i < 0) OR (i > 6) THEN
          BEGIN;
            window_write(p, 'Bad start day of week in pathfile: ' + path_info);
            EXIT;
          END;

        IF ((w_cnt = 4) AND (todays_dow <> i))
                                     OR ((w_cnt = 5) AND (todays_dow < i)) THEN
          EXIT;

        {$IFDEF DEBUG_T}
          WRITELN('T process 7');
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* End DOW                                                           *)
        (*-------------------------------------------------------------------*)

        IF w_cnt = 4 THEN
          BEGIN;
            path_t_ok := TRUE;
            EXIT;
          END;

        {$IFDEF DEBUG_T}
          WRITELN('T process 8');
        {$ENDIF}

        VAL(subword(@path_info, 5, 1), i, j);
        IF (j <> 0) OR (i < 0) OR (i > 6) THEN
          BEGIN;
            window_write(p, 'Bad end day of week in pathfile: ' + path_info);
            EXIT;
          END;

        IF todays_dow > i THEN
          EXIT;

        {$IFDEF DEBUG_T}
          WRITELN('T process 9');
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Now the variable part                                             *)
        (*-------------------------------------------------------------------*)

        path_common^.path_type_send := '';

        err  := FALSE;
        t_ok := TRUE;

        w_index := 6;

        WHILE (w_index <= w_cnt) DO
          BEGIN;

            {$IFDEF DEBUG_T}
              WRITELN('T process 10 - ', i, '/', w_cnt);
            {$ENDIF}

            vary_t;

            IF err OR NOT t_ok THEN
              BEGIN;

                {$IFDEF DEBUG_T}
                  WRITELN('T process 11');
                {$ENDIF}

                EXIT;

              END;

          END;

        (*-------------------------------------------------------------------*)
        (* If we get here, everything is AOK                                 *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_T}
          WRITELN('T process 12');
        {$ENDIF}

        path_t_ok := TRUE;

      END;

  END; (*----- End forward time script --------------------------------------*)
