(*===========================================================================*)
(* Forward task -- Forward a specific message                                *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

{$UNDEF DEBUG_TH}

(*===========================================================================*)
(* Forward this guy from a route                                             *)
(*===========================================================================*)

PROCEDURE forward_this_one   (path_common : path_block_ptr;
                              this_one    : msg_pi_ptr);

  VAR
    d_block       : msg_d_ptr;
    response_char : CHAR;
    this_msg      : msg_path_array_item;

  {$IFDEF DEBUG_TH}
  VAR
    numstr : STRING[11];
  {$ENDIF}

  {$I BBFWDSF.PAS}           (* Forward a message                            *)

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* If the pointer has been cleared, skip it                              *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF POINT_CHK}
      test_pointer(this_one);
    {$ENDIF}

    IF this_one^.msg_p_i = NIL THEN
      EXIT;

    {$IFDEF POINT_CHK}
      test_pointer(this_one^.msg_p_i);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Pointer is OK.  Snarf the item                                        *)
    (*-----------------------------------------------------------------------*)

    this_msg := this_one^;

    WITH this_msg.msg_p_i^ DO
      BEGIN;

        {$IFDEF DEBUG_TH}
          STR(msg_i_mb.msg_number, numstr);
          window_write('', 'FWD MSG SELECT ' + numstr);
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* If its state has changed then skip it                             *)
        (*-------------------------------------------------------------------*)

        IF (msg_i_mb.msg_flag AND
                           (mf_hold OR mf_read OR mf_fwd OR mf_kill)) <> 0 THEN
          EXIT;

        {$IFDEF DEBUG_TH}
          window_write('', 'FWD Test 1 OK');
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* If forward flag on, don't do it twice                             *)
        (*-------------------------------------------------------------------*)

        IF (this_msg.msg_p_item = 0)
                         AND ((msg_i_mb.msg_flag AND mf_fwd_process) <> 0) THEN
          EXIT;

        {$IFDEF DEBUG_TH}
          window_write('', 'FWD Test 2 OK');
        {$ENDIF}

        IF (this_msg.msg_p_item <> 0) AND
          ((msg_i_dr^.msg_dr_dblk^.msg_d_array[this_msg.msg_p_item].msg_d_flag
                                     AND (df_fwd OR df_fwd_process)) <> 0) THEN
          EXIT;

        {$IFDEF DEBUG_TH}
          window_write('', 'FWD Test 3 OK');
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* If size is wrong then skip it                                     *)
        (*-------------------------------------------------------------------*)

        IF msg_i_mb.msg_size > path_common^.path_size THEN
          EXIT;

        {$IFDEF DEBUG_TH}
          window_write('', 'FWD Test 4 OK');
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Turn on in process bit (both main bit and list bit)               *)
        (*-------------------------------------------------------------------*)

        IF this_msg.msg_p_item <> 0 THEN
          BEGIN;

            d_block := find_dist_list(this_msg.msg_p_i);

            WITH d_block^.msg_d_array[this_msg.msg_p_item] DO
              msg_d_flag := msg_d_flag OR df_fwd_process;

          END;

        msg_i_mb.msg_flag := msg_i_mb.msg_flag OR mf_fwd_process;

        (*-------------------------------------------------------------------*)
        (* Set data in TCB                                                   *)
        (*-------------------------------------------------------------------*)

        active_tcb^.curr_fwd := this_msg;

        (*-------------------------------------------------------------------*)
        (* Send this message                                                 *)
        (*-------------------------------------------------------------------*)

        forward_msg(path_common, this_msg.msg_p_i);

        (*-------------------------------------------------------------------*)
        (* Some validity checks                                              *)
        (*-------------------------------------------------------------------*)

        {$IFDEF POINT_CHK}
          test_pointer(this_one^.msg_p_i);
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* If not later then log it                                          *)
        (*-------------------------------------------------------------------*)

        IF response_char <> 'L' THEN
          log_data_is(msg_i_mb.msg_number,
                                          'F ' + path_common^.path_target);

        (*-------------------------------------------------------------------*)
        (* Set the proper flags to indicate success, etc                     *)
        (*-------------------------------------------------------------------*)

        mark_as_forwarded(this_msg.msg_p_i,
                          this_msg.msg_p_item,
                          response_char);

        (*-------------------------------------------------------------------*)
        (* Remove data from TCB                                              *)
        (*-------------------------------------------------------------------*)

        active_tcb^.curr_fwd.msg_p_i := NIL;

        (*-------------------------------------------------------------------*)
        (* Update the msg file and remove sent message from list             *)
        (*-------------------------------------------------------------------*)

        {$IFDEF POINT_CHK}
          test_pointer(this_msg.msg_p_i);
        {$ENDIF}

        update_msg(this_msg.msg_p_i);
        this_one^.msg_p_i := NIL;

      END; (*----- End WITH statement for msg block pointer -----------------*)

  END;
