(*===========================================================================*)
(* Look up "H" addresses in a control file                                   *)
(*                                                                           *)
(*   Copyright 1989, 1990, 1991 by H. Roy Engehausen.  All rights reserved.  *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

UNIT BBHLOOK;

INTERFACE

  PROCEDURE h_look_up(cmd_string : STRING);

IMPLEMENTATION

USES
  bbdummy,
  bbmdata,
  bbmess,
  bbmf,
  bbmisc,
  bbmisc5,
  bbsdata,
  bbsema2,
  bbstr,
  bbtask,
  match;

(*===========================================================================*)
(* Look up "H" addresses                                                     *)
(*===========================================================================*)

PROCEDURE h_look_up(cmd_string : STRING);

  CONST
    dont_look_flags : msg_flag_type = mf_hold
                                     OR mf_read
                                     OR mf_fwd
                                     OR mf_fwd_list
                                     OR mf_fwd_select
                                     OR mf_kill
                                     OR mf_review;

  VAR
    h_file_name  : file_name_str;
    h_file       : TEXT;
    i            : INTEGER;
    j            : WORD;
    l_cnt        : WORD;
    loop_count   : BYTE;
    s_buff       : STRING;
    search_str   : STRING;
    start_msg    : msg_index_ptr;
    this_msg     : msg_index_ptr;

  LABEL
    read_loop;

  (*=========================================================================*)
  (* Subroute to look up the messages                                        *)
  (*         1) must be above a certain number                               *)
  (*         2) TO @ is not blank                                            *)
  (*         3) OK for forwarding and no list                                *)
  (*         4) TO .haddress = blank                                         *)
  (*         5) String match                                                 *)
  (*=========================================================================*)

  FUNCTION find_a_msg (last_find : msg_index_ptr;
                       s_str : STRING) : msg_index_ptr;

    BEGIN;

      (*---------------------------------------------------------------------*)
      (* Point to msg to start search at                                     *)
      (*---------------------------------------------------------------------*)

      IF last_find = NIL THEN
        last_find := start_msg
      ELSE
        last_find := last_find^.msg_i_next;

      (*---------------------------------------------------------------------*)
      (* Loop thru search.  Exit if found                                    *)
      (*---------------------------------------------------------------------*)

      WHILE last_find <> NIL DO
        WITH last_find^.msg_i_mb DO
          BEGIN;
            IF ((msg_flag AND dont_look_flags) = 0)
                       AND (LENGTH(msg_to_h) = 0)
                       AND (LENGTH(msg_to_at) <> 0)
                       AND match_str(msg_to_at, s_str) THEN
              BEGIN;
                find_a_msg := last_find;
                EXIT;
              END
            ELSE
              last_find := last_find^.msg_i_next;
          END;


      (*---------------------------------------------------------------------*)
      (* Search failed!                                                      *)
      (*---------------------------------------------------------------------*)

      find_a_msg := NIL;

    END;

  (*=========================================================================*)
  (* Main line                                                               *)
  (*=========================================================================*)

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Any file to search?                                                   *)
    (*-----------------------------------------------------------------------*)

    IF opt_block.hlook_fn = '' THEN
      EXIT;

    (*-----------------------------------------------------------------------*)
    (* Process parm                                                          *)
    (*-----------------------------------------------------------------------*)

    i := words(cmd_string);
    IF i > 2 THEN
      BEGIN;
        send_message(message_err_wrd);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    IF i = 2 THEN
      BEGIN;

        VAL(subword(@cmd_string, 2, 1), j, i);

        IF (i <> 0) OR (j > 65000) OR (j < 0) THEN
          BEGIN;
            send_message(message_err_ivm);
            active_tcb^.error_sw := TRUE;
            EXIT;
          END;

        IF j = 0 THEN
          INC(j);
        last_h_msg := j - 1;

      END;

    (*-----------------------------------------------------------------------*)
    (* Tell user started                                                     *)
    (*-----------------------------------------------------------------------*)

    IF last_h_msg >= next_msg_no THEN
      BEGIN;
        send_tnc_data_str('No work for address lookup' + cr);
        last_h_msg := next_msg_no;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Tell user started                                                     *)
    (*-----------------------------------------------------------------------*)

    STR(last_h_msg, s_buff);
    send_tnc_data_str('Address lookup started at ' + s_buff + cr);

    (*-----------------------------------------------------------------------*)
    (* Find place to start                                                   *)
    (*-----------------------------------------------------------------------*)

    this_msg := msg_index_start;
    WHILE (this_msg <> NIL)
                    AND (this_msg^.msg_i_mb.msg_number < last_h_msg) DO
      this_msg := this_msg^.msg_i_next;

    start_msg := this_msg;

    (*-----------------------------------------------------------------------*)
    (* Do a search to see if we actually have any                            *)
    (*-----------------------------------------------------------------------*)

    this_msg := find_a_msg(NIL, '*');

    IF this_msg = NIL THEN
      BEGIN;
        send_tnc_data_str('No messages to be done' + cr);
        last_h_msg := next_msg_no;
        EXIT;
      END;

    start_msg := this_msg;

    (*-----------------------------------------------------------------------*)
    (* Obtain the interrupt lock                                             *)
    (*-----------------------------------------------------------------------*)

    get_semaphore(semaphore_interrupts, sem_exclusive, TRUE);

    (*-----------------------------------------------------------------------*)
    (* Open file                                                             *)
    (*-----------------------------------------------------------------------*)

    ASSIGN(h_file, opt_block.hlook_fn);

    {$I-}
      RESET(h_file);
    {$I+}

    i := IORESULT;

    (*-----------------------------------------------------------------------*)
    (* Free the lock                                                         *)
    (*-----------------------------------------------------------------------*)

    free_semaphore(semaphore_interrupts);

    (*-----------------------------------------------------------------------*)
    (* File doesn't exist.                                                   *)
    (*-----------------------------------------------------------------------*)

    IF i = 2 THEN
      BEGIN;
        send_tnc_data_str('Can''t find ' + opt_block.hlook_fn + cr);
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Different open error                                                  *)
    (*-----------------------------------------------------------------------*)

    IF i <> 0 THEN
      BEGIN;
        send_tnc_data_str('Unable to open ' + opt_block.hlook_fn + cr);
        send_tnc_data_str(dos_err_message(i) + cr);
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Read loop                                                             *)
    (*-----------------------------------------------------------------------*)

    l_cnt      := 0;
    loop_count := 0;

read_loop:

    WHILE NOT EOF(h_file) AND NOT shutdown_switch DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Bump counters                                                     *)
        (*-------------------------------------------------------------------*)

        INC(l_cnt);

        task_switch;

        (*-------------------------------------------------------------------*)
        (* Get a line                                                        *)
        (*-------------------------------------------------------------------*)

        READLN(h_file, s_buff);


        (*-------------------------------------------------------------------*)
        (* Skip comments                                                     *)
        (*-------------------------------------------------------------------*)

        i := 1;
        WHILE (i <=  LENGTH(s_buff)) AND (s_buff[i] = ' ') DO
          INC(i);

        IF (i > LENGTH(s_buff)) OR (s_buff[i] = ';') THEN
          GOTO read_loop;

        (*-------------------------------------------------------------------*)
        (* Check word count                                                  *)
        (*-------------------------------------------------------------------*)

        IF WORDS(s_buff) = 1 THEN
          BEGIN;
            STR(l_cnt, s_buff);
            send_tnc_data_str('Bad word count at line # ' + s_buff + cr);
            GOTO read_loop;
          END;

        (*-------------------------------------------------------------------*)
        (* Upper case the line                                               *)
        (*-------------------------------------------------------------------*)

        upcase_str_var(s_buff);

        (*-------------------------------------------------------------------*)
        (* Get the data out and validate                                     *)
        (*-------------------------------------------------------------------*)

        search_str := strip_ssid(subword(@s_buff, 1, 1));
        s_buff := subword(@s_buff, 2, 1);

        IF s_buff[1] <> '.' THEN
          BEGIN;
            STR(l_cnt, s_buff);
            send_tnc_data_str('Bad address format at line # ' + s_buff + cr);
            GOTO read_loop;
          END;

        (*-------------------------------------------------------------------*)
        (* Do search                                                         *)
        (*-------------------------------------------------------------------*)

        this_msg := find_a_msg(NIL, search_str);

        IF this_msg <> NIL THEN
          BEGIN;

            (*---------------------------------------------------------------*)
            (* We found somebody!                                            *)
            (*---------------------------------------------------------------*)

            send_tnc_data_str('Address update for ' + search_str
                               + s_buff + cr);

            (*---------------------------------------------------------------*)
            (* Loop thru them all updating the list as we go                 *)
            (*---------------------------------------------------------------*)

            REPEAT

              this_msg^.msg_i_mb.msg_to_h := COPY(s_buff, 2, 255);

              update_msg(this_msg);

              j := this_msg^.msg_i_mb.msg_number;
              IF msg_route_num > j THEN
                msg_route_num := j;

              this_msg := find_a_msg(this_msg, search_str);

            UNTIL this_msg = NIL;

          END;

      END; (*----- End read loop --------------------------------------------*)

    (*-----------------------------------------------------------------------*)
    (* Close things up                                                       *)
    (*-----------------------------------------------------------------------*)

    CLOSE(h_file);

    (*-----------------------------------------------------------------------*)
    (* Reset highest message number processed                                *)
    (*-----------------------------------------------------------------------*)

    last_h_msg := next_msg_no;

    (*-----------------------------------------------------------------------*)
    (* Tell user ended                                                       *)
    (*-----------------------------------------------------------------------*)

    send_tnc_data_str('Address lookup ended' + cr);

  END;

END.
