(*===========================================================================*)
(* Import from a file                                                        *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991 by H. Roy Engehausen.  All rights      *)
(*   reserved.                                                               *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

UNIT BBIMPORT;

INTERFACE

PROCEDURE import_cmd(cmd_string : STRING);

IMPLEMENTATION

USES
CRT,
  bbbid,
  bbcmsg,
  bbdummy,
  bbfin,
  bblog,
  bbmdata,
  bbmess,
  bbmem,
  bbmf,
  bbmisc5,
  bbsdata,
  bbsema2,
  bbsto,
  bbstr;

{$UNDEF DEBUG}
{$UNDEF DEBUG_CS}

(*===========================================================================*)
(* IMPORT                                                                    *)
(*===========================================================================*)

PROCEDURE import_cmd(cmd_string : STRING);

  VAR
    erase_sw    : BOOLEAN;
    i           : INTEGER;
    import_in   : TEXT;
    import_name : file_name_str;
    line_count  : LONGINT;
    server_sw   : BOOLEAN;
    temp_name   : file_name_str;

  PROCEDURE import_clean;
    BEGIN;
      {$I-}
      CLOSE(import_in);
      {$I+}
    END;

  PROCEDURE import_error;
    BEGIN;
      STR(line_count, cmd_string);
      send_tnc_data_str('Line #' + cmd_string + cr);
    END;

  (*=========================================================================*)
  (* Execute for each subcommand                                             *)
  (*=========================================================================*)

  PROCEDURE import_send;

    VAR
      b             : BOOLEAN;
      calls_scanned : STRING;
      dupe_cnt      : WORD;
      header        : BOOLEAN;
      i             : WORD;
      import_out    : TEXT;
      io_count      : LONGINT;
      p             : POINTER;
      t_str         : STRING;

    BEGIN;

      (*---------------------------------------------------------------------*)
      (* Clear calls scanned                                                 *)
      (*---------------------------------------------------------------------*)

      free_task_mem('CS', TRUE);

      (*---------------------------------------------------------------------*)
      (* Read in the send command                                            *)
      (*---------------------------------------------------------------------*)

      REPEAT
        IF EOF(import_in) THEN EXIT;
        INC(line_count);
        READLN(import_in, cmd_string);
        upcase_str_var(cmd_string);
      UNTIL cmd_string <> '';

      {$IFDEF DEBUG}
        WRITELN(cmd_string);
      {$ENDIF}

      IF (cmd_string[1] <> 'S') OR (cmd_string[3] <> ' ') THEN
        BEGIN;
          send_message(message_err_2nd);
          active_tcb^.error_sw := TRUE;
          import_error;
          EXIT;
        END;

      (*---------------------------------------------------------------------*)
      (* Initialize                                                          *)
      (*---------------------------------------------------------------------*)

      FILLCHAR(active_tcb^.curr_msg, SIZEOF(msg_index_block), CHR(0));

      header        := TRUE;
      calls_scanned := '';

      (*---------------------------------------------------------------------*)
      (* Set the msg type and flag                                           *)
      (*---------------------------------------------------------------------*)

      WITH active_tcb^.curr_msg.msg_i_mb DO
        BEGIN;

          msg_type := UPCASE(cmd_string[2]);

          msg_flag := 0;

        END;

      (*---------------------------------------------------------------------*)
      (* Process 'TO' data if present.  If not get some and process it       *)
      (*---------------------------------------------------------------------*)

      send_msg_to_process(subword(@cmd_string, 2, 0));

      IF active_tcb^.error_sw THEN
        BEGIN;
          send_message(message_op_halted);
          import_error;
          EXIT;
        END;

      (*---------------------------------------------------------------------*)
      (* Read subject                                                        *)
      (*---------------------------------------------------------------------*)

      INC(line_count);
      READLN(import_in, t_str);

      strip_crlf(t_str);

      {$IFDEF DEBUG}
        WRITELN(t_str);
      {$ENDIF}

      active_tcb^.curr_msg.msg_i_mb.msg_subj := t_str;

      (*---------------------------------------------------------------------*)
      (* Obtain the interrupt lock                                           *)
      (*---------------------------------------------------------------------*)

      get_semaphore(semaphore_interrupts, sem_exclusive, FALSE);

      (*---------------------------------------------------------------------*)
      (* Open the output file                                                *)
      (*---------------------------------------------------------------------*)

      FILLCHAR(import_out, SIZEOF(import_out), #0);

      ASSIGN(import_out, temp_name);
      REWRITE(import_out);

      (*---------------------------------------------------------------------*)
      (* Release the interrupt lock                                          *)
      (*---------------------------------------------------------------------*)

      free_semaphore(semaphore_interrupts);

      (*---------------------------------------------------------------------*)
      (* Read loop                                                           *)
      (*---------------------------------------------------------------------*)

      io_count := 0;
      b := TRUE;

      WHILE b AND (NOT EOF(import_in)) DO
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* Is there an end of line?                                        *)
          (*-----------------------------------------------------------------*)

          IF EOLN(import_in) THEN
            BEGIN;

              (*-------------------------------------------------------------*)
              (* Yes, transfer the end of line from input to output          *)
              (*-------------------------------------------------------------*)

              INC(line_count);

              READLN(import_in, t_str);
              WRITELN(import_out, t_str);

              {$IFDEF DEBUG}
                WRITELN(t_str);
              {$ENDIF}

              INC(io_count);
              INC(io_count, LENGTH(t_str));

            END (*----- End handling the EOL character ----------------------*)
          ELSE
            BEGIN;

              (*-------------------------------------------------------------*)
              (* No end of line..  Get the data                              *)
              (*-------------------------------------------------------------*)

              READ(import_in, t_str);

              {$IFDEF DEBUG}
                WRITELN(t_str);
              {$ENDIF}

              (*-------------------------------------------------------------*)
              (* See if we have the special /EX ending                       *)
              (*-------------------------------------------------------------*)

              IF t_str <> '/EX' THEN
                BEGIN;

                  (*---------------------------------------------------------*)
                  (* Regular data, write it out and header scan it           *)
                  (*---------------------------------------------------------*)

                  WRITE(import_out, t_str);

                  INC(io_count);
                  INC(io_count, LENGTH(t_str));

                  IF header THEN
                    BEGIN;
                      header := scan_head(t_str, calls_scanned, dupe_cnt);
                      {$IFDEF DEBUG_CS}
                        WRITELN('Loop CS -- ', header, ' = ', calls_scanned);
                      {$ENDIF}
                    END;

                END
              ELSE
                BEGIN;

                  (*---------------------------------------------------------*)
                  (* Special /EX                                             *)
                  (*---------------------------------------------------------*)

                  b := FALSE;
                  READLN(import_in, t_str);
                  t_str := '';

                END;

            END; (*----- End handling the non-EOLN state --------------------*)

        END; (*----- End loop thru the file ---------------------------------*)

      (*---------------------------------------------------------------------*)
      (* Obtain the interrupt lock                                           *)
      (*---------------------------------------------------------------------*)

      get_semaphore(semaphore_interrupts, sem_exclusive, FALSE);

      (*---------------------------------------------------------------------*)
      (* Close the file                                                      *)
      (*---------------------------------------------------------------------*)

      CLOSE(import_out);

      (*---------------------------------------------------------------------*)
      (* Release the interrupt lock                                          *)
      (*---------------------------------------------------------------------*)

      free_semaphore(semaphore_interrupts);

      (*---------------------------------------------------------------------*)
      (* Push calls scanned into the memory stack                            *)
      (*---------------------------------------------------------------------*)

      i := LENGTH(calls_scanned) + 1;
      IF i > 1 THEN
        BEGIN;
          p := get_task_mem('CS', i);
          MOVE(calls_scanned, p^, i);
        END;

      {$IFDEF DEBUG_CS}
        WRITELN('Push CS -- ', calls_scanned);
      {$ENDIF}

      (*---------------------------------------------------------------------*)
      (* Text is loaded.                                                     *)
      (*---------------------------------------------------------------------*)

      WITH active_tcb^.curr_msg.msg_i_mb DO
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* Test BID                                                        *)
          (*-----------------------------------------------------------------*)

          IF (LENGTH(msg_bid) = 1) AND (msg_bid[1] = CHR(0)) THEN
            b := FALSE
          ELSE
            b := bid_check;

          msg_size := io_count;

          add_msg(temp_name, FALSE);
          make_cc;

          log_data_is(msg_number, cmd_string);

        END;

      send_message(message_added_msg);
      IF b THEN
        send_message(message_dupe_bid);

    END;

  (*=========================================================================*)
  (* Main line                                                               *)
  (*=========================================================================*)

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Log it                                                                *)
    (*-----------------------------------------------------------------------*)

    log_data_s('L' + cmd_string);

    {$IFDEF DEBUG}
      WRITELN(cmd_string);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Check command format                                                  *)
    (*-----------------------------------------------------------------------*)

    IF words(cmd_string) < 2 THEN
      BEGIN;
        send_message(message_not_en);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    IF words(cmd_string) > 2 THEN
      BEGIN;
        send_message(message_err_wrd);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Erase?  Server?                                                       *)
    (*-----------------------------------------------------------------------*)

    import_name := subword(@cmd_string, 1, 1);
    upcase_str_var(import_name);
    erase_sw  := POS('E', import_name) > 0;
    server_sw := POS('S', import_name) > 0;

    {$IFDEF DEBUG}
      WRITELN(import_name);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Obtain the interrupt lock                                             *)
    (*-----------------------------------------------------------------------*)

    get_semaphore(semaphore_interrupts, sem_exclusive, FALSE);

    (*-----------------------------------------------------------------------*)
    (* Get input file and open it                                            *)
    (*-----------------------------------------------------------------------*)

    import_name := subword(@cmd_string, 2, 1);

    {$IFDEF DEBUG}
      WRITELN(import_name);
    {$ENDIF}

    FILLCHAR(import_in, SIZEOF(import_in), #0);
    ASSIGN(import_in, import_name);

    {$I-}
    RESET(import_in);
    {$I+}
    i := IORESULT;

    (*-----------------------------------------------------------------------*)
    (* Release the interrupt lock                                            *)
    (*-----------------------------------------------------------------------*)

    free_semaphore(semaphore_interrupts);

    (*-----------------------------------------------------------------------*)
    (* Handle I/O error                                                      *)
    (*-----------------------------------------------------------------------*)

    IF i <> 0 THEN
      BEGIN;

        IF server_sw AND (i = 2) THEN
          BEGIN;
            send_tnc_data_str('No output from server' + cr);
            EXIT;
          END;

        send_tnc_data_str(dos_err_message(i) + cr);
        active_tcb^.error_sw := TRUE;
        EXIT;

      END;

    (*-----------------------------------------------------------------------*)
    (* Build temp name                                                       *)
    (*-----------------------------------------------------------------------*)

    temp_name := opt_block.msg_file_dir + active_tcb^.port_chan_s + '.IN';

    {$IFDEF DEBUG}
      WRITELN(temp_name);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Loop around reading the IMPORT file                                   *)
    (*-----------------------------------------------------------------------*)

    line_count := 0;

    WHILE (NOT EOF(import_in)) AND (NOT active_tcb^.error_sw) DO
      import_send;

    (*-----------------------------------------------------------------------*)
    (* Import file done                                                      *)
    (*-----------------------------------------------------------------------*)

    import_clean;

    (*-----------------------------------------------------------------------*)
    (* Erase?                                                                *)
    (*-----------------------------------------------------------------------*)

    IF erase_sw AND NOT active_tcb^.error_sw THEN
      ERASE(import_in);

    (*-----------------------------------------------------------------------*)
    (* Tell user                                                             *)
    (*-----------------------------------------------------------------------*)

    send_message(message_action_complete);

  END;

END.
