(*===========================================================================*)
(* Kill msg command -- Entered by command processor                          *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990 by H. Roy Engehausen.  All rights reserved.  *)
(*                                                                           *)
(*===========================================================================*)

{$UNDEF debug}

{$O+}

UNIT BBKMC;

INTERFACE

  USES
    bbdummy;

  PROCEDURE kill_msg_cmd(cmd_string : STRING);
  PROCEDURE kill_a_msg(m_ptr : msg_index_ptr; tell_sw : BOOLEAN);

IMPLEMENTATION

  USES
    CRT,
    bblog,
    bbmdata,
    bbmess,
    bbmisc,
    bbmf,
    bbmmsgn,
    bbrdata,
    bbsearch,
    bbstr;

(*===========================================================================*)
(* Kill msg command                                                          *)
(*===========================================================================*)

PROCEDURE kill_msg_cmd(cmd_string : STRING);

  VAR
    b            : BOOLEAN;
    i_ptr        : msg_index_ptr;
    search_block : search_block_type;
    kill_char    : CHAR;
    k_msg_no     : LONGINT;
    ok_to_kill   : BOOLEAN;
    word_count   : BYTE;

  (*=========================================================================*)
  (* Put kill error message out                                              *)
  (*=========================================================================*)


  PROCEDURE put_kill_error(err_num : BYTE);

    VAR
      err_string : STRING[6];

    BEGIN;

      WITH active_tcb^ DO
        BEGIN;

          STR(k_msg_no, err_string);

          set_dollar1_parm(@err_string);
          send_message(err_num);

          error_sw   := TRUE;
          ok_to_kill := FALSE;

        END;

    END;

  (*=========================================================================*)
  (* Check kill -- See if user authorized                                    *)
  (*=========================================================================*)

  PROCEDURE check_kill(m_ptr : msg_index_ptr; kill_type : CHAR);

    BEGIN;

      WITH m_ptr^.msg_i_mb, active_tcb^.uid_data DO
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* If this is an NTS kill, then it must be an NTS message!         *)
          (*-----------------------------------------------------------------*)

          IF (kill_type = mt_nts) THEN
            IF (msg_type = mt_nts) THEN
              EXIT
            ELSE
              BEGIN;
                put_kill_error(message_kill_not_a);
                EXIT;
              END;

          (*-----------------------------------------------------------------*)
          (* SYSOPs are ok                                                   *)
          (*-----------------------------------------------------------------*)

          IF user_class >= user_c_rsu THEN EXIT;
          IF (user_f_sysop AND user_flag) <> 0 THEN EXIT;

          (*-----------------------------------------------------------------*)
          (* To or from this user is ok                                      *)
          (*-----------------------------------------------------------------*)

          IF (msg_to = user_id) OR (msg_from = user_id) THEN
            EXIT;

        END;

      (*---------------------------------------------------------------------*)
      (* If we get here, it is not ok so tell user that                      *)
      (*---------------------------------------------------------------------*)

      put_kill_error(message_kill_not_a);
      active_tcb^.error_sw := TRUE;

    END; (*----- End CHECK KILL Procedure -----------------------------------*)

  (*=========================================================================*)
  (* Kill command main line                                                  *)
  (*=========================================================================*)

  BEGIN;

    log_data_s('L' + cmd_string);

    word_count := WORDS(cmd_string);

    IF LENGTH(cmd_string) = 1 THEN
      kill_char := ' '
    ELSE
      kill_char := UPCASE(cmd_string[2]);

    (*-----------------------------------------------------------------------*)
    (* Handle Kill                                                           *)
    (*-----------------------------------------------------------------------*)

    CASE kill_char OF

      (*---------------------------------------------------------------------*)
      (* KK is dumb                                                          *)
      (*---------------------------------------------------------------------*)

      'K': BEGIN;
              send_message(message_err_2nd);
              active_tcb^.error_sw := TRUE;
              EXIT;
            END;

      (*---------------------------------------------------------------------*)
      (* Kill ' '  or 'T'                                                    *)
      (*---------------------------------------------------------------------*)

      ' ',
      'T': BEGIN;

             (*--------------------------------------------------------------*)
             (* Handle kill without a number                                 *)
             (*--------------------------------------------------------------*)

             IF word_count = 1 THEN
               BEGIN;

                 (*----------------------------------------------------------*)
                 (* Ask for a number and get answer                          *)
                 (*----------------------------------------------------------*)

                 send_message(message_kmc_num);
                 cmd_string := read_tnc_data_str;
                 strip_crlf(cmd_string);

                 (*----------------------------------------------------------*)
                 (* If no answer, we quit                                    *)
                 (*----------------------------------------------------------*)

                 IF LENGTH(cmd_string) = 0 THEN
                   BEGIN;
                     send_message(message_nodata_can);
                     EXIT;
                   END;

                 (*----------------------------------------------------------*)
                 (* If too long, we quit                                     *)
                 (*----------------------------------------------------------*)

                 IF LENGTH(cmd_string) > 253 THEN
                   BEGIN;
                     send_message(message_err_wrd);
                     active_tcb^.error_sw := TRUE;
                     EXIT;
                   END;

                 (*----------------------------------------------------------*)
                 (* Fake input line                                          *)
                 (*----------------------------------------------------------*)

                 cmd_string := 'K ' + cmd_string;
                 word_count := WORDS(cmd_string);

               END;

             (*--------------------------------------------------------------*)
             (* Validate number                                              *)
             (*--------------------------------------------------------------*)

             upcase_str_var(cmd_string);

             check_multiple_msg(@cmd_string, 2, word_count);
             IF active_tcb^.error_sw THEN
               EXIT;

             (*--------------------------------------------------------------*)
             (* Loop thru all the messages                                   *)
             (*--------------------------------------------------------------*)

             k_msg_no := get_next_multiple_msg;

             WHILE k_msg_no <> 0 DO
               BEGIN;

                 (*----------------------------------------------------------*)
                 (* Ready to kill                                            *)
                 (*----------------------------------------------------------*)

                 ok_to_kill := TRUE;

                 (*----------------------------------------------------------*)
                 (* Find message,  Give error if cannot be found             *)
                 (*----------------------------------------------------------*)

                 i_ptr := find_msg(k_msg_no);

                 IF i_ptr = NIL THEN
                   BEGIN;
                     put_kill_error(message_kmc_nf);
                     IF NOT in_multiple_msg_range THEN
                       EXIT;
                   END;

                 (*----------------------------------------------------------*)
                 (* If so far so good then see if allowed to kill            *)
                 (*----------------------------------------------------------*)

                 IF ok_to_kill THEN
                   BEGIN;
                     check_kill(i_ptr, kill_char);
                     IF (NOT ok_to_kill) AND (NOT in_multiple_msg_range) THEN
                       EXIT;
                   END;

                 (*----------------------------------------------------------*)
                 (* If so far so good then check for double kill             *)
                 (*----------------------------------------------------------*)

                 IF ok_to_kill
                        AND ((i_ptr^.msg_i_mb.msg_flag AND mf_kill) <> 0) THEN
                   BEGIN;
                     put_kill_error(message_k_already);
                     IF NOT in_multiple_msg_range THEN
                       EXIT;
                   END;

                 (*----------------------------------------------------------*)
                 (* If so far so good then kill it                           *)
                 (*----------------------------------------------------------*)

                 IF ok_to_kill THEN
                   kill_a_msg(i_ptr, TRUE);

                 (*----------------------------------------------------------*)
                 (* Get next message to kill                                 *)
                 (*----------------------------------------------------------*)

                 k_msg_no := get_next_multiple_msg;

               END; (*----- End loop thru all messages ----------------------*)

           END;

      (*---------------------------------------------------------------------*)
      (* All other letters                                                   *)
      (*---------------------------------------------------------------------*)

      ELSE
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* KM is a special case                                            *)
          (*-----------------------------------------------------------------*)

          IF kill_char = 'M' THEN
            BEGIN;
              IF word_count > 1 THEN
                BEGIN;
                  send_message(message_err_wrd);
                  active_tcb^.error_sw := TRUE;
                  EXIT;
                END;

              cmd_string[2] := 'Q';
              cmd_string := cmd_string + ' ' + active_tcb^.uid_data.user_id;

            END;

          (*-----------------------------------------------------------------*)
          (* Build search block                                              *)
          (*-----------------------------------------------------------------*)

          cmd_string := COPY(cmd_string, 2, 255);
          set_search(cmd_string, @search_block);

          IF active_tcb^.error_sw THEN
            EXIT;

          (*-----------------------------------------------------------------*)
          (* Find first one                                                  *)
          (*-----------------------------------------------------------------*)

          search_msg(@search_block);

          IF search_block.search_last = NIL THEN
            BEGIN;
              send_message(message_lmc_nf);
              active_tcb^.error_sw := TRUE;
              EXIT;
            END;

          (*-----------------------------------------------------------------*)
          (* Check for dumbness                                              *)
          (*-----------------------------------------------------------------*)

          {$IFDEF debug}
            WRITELN('K1=', ORD(active_tcb^.tcb_type));
            DELAY(2000);
          {$ENDIF}

          IF (kill_char <> 'M')
                 AND ((active_tcb^.tcb_type = th_operator)
                      OR (active_tcb^.tcb_type = th_user)) THEN
            BEGIN;

              {$IFDEF debug}
                WRITELN('K2=', ORD(active_tcb^.tcb_type));
                DELAY(2000);
              {$ENDIF}

              (*-------------------------------------------------------------*)
              (* Prepare the loop                                            *)
              (*-------------------------------------------------------------*)

               b := FALSE;

               REPEAT

                (*-----------------------------------------------------------*)
                (* Ask user to confirm                                       *)
                (*-----------------------------------------------------------*)

                 send_message(message_confirm_pls);

                (*-----------------------------------------------------------*)
                (* Get a response                                            *)
                (*-----------------------------------------------------------*)

                 cmd_string := read_tnc_data_str;

                (*-----------------------------------------------------------*)
                (* Parse response                                            *)
                (*-----------------------------------------------------------*)

                 strip_crlf(cmd_string);
                 upcase_str_var(cmd_string);

                (*-----------------------------------------------------------*)
                (* Halt operation?                                           *)
                (*-----------------------------------------------------------*)

                 IF (cmd_string = '0')
                               OR (cmd_string = 'N')
                               OR (cmd_string = 'NO') THEN
                   BEGIN;
                     send_message(message_op_halted);
                     EXIT;
                   END;

                (*-----------------------------------------------------------*)
                (* Continue with the operation?                              *)
                (*-----------------------------------------------------------*)

                 IF (cmd_string = '1')
                               OR (cmd_string = 'Y')
                               OR (cmd_string = 'YES') THEN
                   b := TRUE;

              UNTIL b; (*----- Loop until user answers question -------------*)

            END;

          (*-----------------------------------------------------------------*)
          (* Loop killing things                                             *)
          (*-----------------------------------------------------------------*)

          {$IFDEF debug}
            WRITELN('K3=', ORD(active_tcb^.tcb_type));
            DELAY(2000);
          {$ENDIF}

          REPEAT
            check_kill(search_block.search_last, kill_char);
            IF active_tcb^.error_sw THEN
              EXIT;
            kill_a_msg(search_block.search_last, TRUE);
            search_msg(@search_block);
          UNTIL search_block.search_last = NIL;

        END;

    END;

  END;

(*===========================================================================*)
(* Kill a specific message (with or without telling user)                    *)
(*===========================================================================*)

PROCEDURE kill_a_msg(m_ptr : msg_index_ptr; tell_sw : BOOLEAN);
  BEGIN;

    WITH m_ptr^.msg_i_mb DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Log message as killed                                             *)
        (*-------------------------------------------------------------------*)

        log_data_is(msg_number, 'K');

        (*-------------------------------------------------------------------*)
        (* Set flag and update the message file                              *)
        (*-------------------------------------------------------------------*)

        msg_flag := msg_flag OR mf_kill;

        update_msg(m_ptr);

        (*-------------------------------------------------------------------*)
        (* Tell user if he wants to know about it                            *)
        (*-------------------------------------------------------------------*)

        IF NOT tell_sw THEN
          EXIT;

        active_tcb^.curr_msg := m_ptr^;

        send_message(message_killed_msg);

      END;

  END;

END.
