(*===========================================================================*)
(* Display                                                                   *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Create a header string for a message block                                *)
(*    Formats are 0, 1, 2, 3, 4                                              *)
(*===========================================================================*)

FUNCTION header_msg_block(i_ptr : msg_index_ptr; type_fmt : INTEGER) : STRING;

  VAR
    d_str      : STRING[2];
    i          : WORD;
    j          : BYTE;
    k          : BYTE;
    msg_dis    : msg_d_ptr;
    result_str : STRING;
    t_str      : STRING[12];

  CONST
    dont_show_dist_list = mf_review
                          OR mf_hold
                          OR mf_read
                          OR mf_fwd
                          OR mf_kill;

  BEGIN;

    IF type_fmt < 0 THEN
      type_fmt := active_tcb^.uid_data.user_fmt;

    IF type_fmt = 4 THEN
      BEGIN;
        active_tcb^.curr_msg := i_ptr^;
        send_message(message_output_fmt_msg);
        header_msg_block := '';
        EXIT;
      END;

    WITH i_ptr^.msg_i_mb DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Output message number                                             *)
        (*-------------------------------------------------------------------*)

        IF type_fmt <> 3 THEN
          BEGIN;
            STR(msg_number:5, t_str);
            result_str := t_str + ' ';
          END
        ELSE
          result_str := '';

        (*-------------------------------------------------------------------*)
        (* Output type and flags                                             *)
        (*-------------------------------------------------------------------*)

        IF type_fmt <> 2 THEN
          result_str := result_str + msg_type
                              + msg_flag_char(i_ptr^.msg_i_mb.msg_flag) + ' ';

        (*-------------------------------------------------------------------*)
        (* Output size                                                       *)
        (*-------------------------------------------------------------------*)

        IF msg_size < 100000 THEN
          STR(msg_size:5, t_str)
        ELSE
          BEGIN;
            STR((msg_size div 1024):4, t_str);
            t_str[0] := CHR(5);
            t_str[5] := 'K';
          END;

        IF type_fmt = 3 THEN
          strip_var(t_str, 'B');

        result_str := result_str + t_str + ' ';

        (*-------------------------------------------------------------------*)
        (* To addressee length                                               *)
        (*-------------------------------------------------------------------*)

        CASE type_fmt OF

          1 : i := call_sign_len + 1 + bb_addr_len + 1;

          2 : i := call_sign_len + 2;

          ELSE
            BEGIN;

              (*-------------------------------------------------------------*)
              (* Format 0,3 -- Short from display.                           *)
              (*-------------------------------------------------------------*)

              (*-------------------------------------------------------------*)
              (* i = size of display field   XXXXXX@XXXXXX. plus a blank     *)
              (*-------------------------------------------------------------*)

              i := call_sign_len + call_sign_len + 3;

              (*-------------------------------------------------------------*)
              (* j = length of the to field plus the "@"                     *)
              (*-------------------------------------------------------------*)

              j := LENGTH(msg_to) + 1;

              (*-------------------------------------------------------------*)
              (* k = length of the to @ field.  Include the "." hierarchical *)
              (* address and the trailing blank                              *)
              (*-------------------------------------------------------------*)

              k := LENGTH(msg_to_at) + 1;

              IF LENGTH(msg_to_h) > 0 THEN
                INC(k);

              (*-------------------------------------------------------------*)
              (* If we overflow the display field then expand it             *)
              (*-------------------------------------------------------------*)

              IF i < (j + k) THEN
                i := j + k;

            END; (*----- End type 0/3 display -------------------------------*)

        END; (*----- End case statement on TO addressee ---------------------*)

        (*-------------------------------------------------------------------*)
        (* j = Size of address area                                          *)
        (* i = length of resultant string                                    *)
        (*-------------------------------------------------------------------*)

        j := i;
        i := LENGTH(result_str) + i;

        (*-------------------------------------------------------------------*)
        (* Tack on the "to"                                                  *)
        (*-------------------------------------------------------------------*)

        result_str := result_str + msg_to;

        (*-------------------------------------------------------------------*)
        (* Add the "@" and hierarchical address as needed                    *)
        (*-------------------------------------------------------------------*)

        IF (LENGTH(msg_to_at) > 0) AND (type_fmt <> 2) THEN
          BEGIN;
            result_str := result_str + '@' + msg_to_at;
            IF LENGTH(msg_to_h) > 0 THEN
              result_str := result_str + '.';
          END;

        (*-------------------------------------------------------------------*)
        (* Make the resultant string the right size.                         *)
        (*-------------------------------------------------------------------*)

        IF type_fmt <> 3 THEN
          result_str := left(result_str, i)
        ELSE
          result_str := result_str + ' ';

        (*-------------------------------------------------------------------*)
        (* From addressee                                                    *)
        (*-------------------------------------------------------------------*)

        IF (type_fmt = 1) OR (type_fmt = 3) THEN
          BEGIN;
            IF LENGTH(msg_from_at) = 0 THEN
              result_str := result_str + left(msg_from, j)
            ELSE
              BEGIN;
                IF LENGTH(msg_from_h) > 0 THEN
                  t_str := '.'
                ELSE
                  t_str := '';
                result_str := result_str +
                              left(msg_from + '@' + msg_from_at + t_str, j);
              END;
            IF type_fmt = 3 THEN
              result_str := strip(result_str, 'B') + ' ';
          END
        ELSE
          result_str := result_str + left(msg_from, SIZEOF(msg_from));

        (*-------------------------------------------------------------------*)
        (* Type 3 format throws away trailing blanks                         *)
        (*-------------------------------------------------------------------*)

        IF type_fmt = 3 THEN
          result_str := strip(result_str, 'T') + ' ';

        (*-------------------------------------------------------------------*)
        (* Date/Time                                                         *)
        (*-------------------------------------------------------------------*)

        result_str := result_str + time_str(msg_dt_in, FALSE) + ' ';

        (*-------------------------------------------------------------------*)
        (* Title                                                             *)
        (*-------------------------------------------------------------------*)

        i := SIZEOF(result_str) - 1 - LENGTH(result_str);
        result_str := result_str + SUBSTR(msg_subj, 1, i);

        (*-------------------------------------------------------------------*)
        (* If the right type of user, show distribution list                 *)
        (*-------------------------------------------------------------------*)

        IF ((msg_flag AND mf_fwd_list) <> 0)
                AND (check_priv(user_c_bu)
                                 OR (msg_from = active_tcb^.uid_data.user_id))
                AND ((msg_flag AND dont_show_dist_list) = 0) THEN
          BEGIN;
            result_str := result_str + cr + ' ';

            msg_dis := find_dist_list(i_ptr);

            {$IFDEF POINT_CHK}
              test_pointer(msg_dis);
            {$ENDIF}

            FOR i := 1 TO msg_dis^.msg_d_no DO
              WITH msg_dis^.msg_d_array[i] DO
                BEGIN;
                  d_str := '  ';
                  IF (df_fwd_select AND msg_d_flag) <> 0 THEN
                    d_str[2] := '+';
                  IF (df_fwd_process AND msg_d_flag) <> 0 THEN
                    d_str[2] := '-';
                  IF (df_fwd_unknown AND msg_d_flag) <> 0 THEN
                    d_str[2] := '?';
                  IF (df_fwd AND msg_d_flag) <> 0 THEN
                    d_str[2] := '*';

                  result_str := result_str + d_str + msg_d_info;
                END;

          END;

      END; (*----- End WITH message index -----------------------------------*)

    header_msg_block := result_str;

  END;

(*===========================================================================*)
(* Get the character that represents the msg flag state.                     *)
(*===========================================================================*)

FUNCTION msg_flag_char(msg_flag : msg_flag_type) : CHAR;

  BEGIN;

    msg_flag_char := 'N';
    IF (msg_flag AND mf_old)         <> 0  THEN msg_flag_char := 'O';
    IF (msg_flag AND mf_unknown)     <> 0  THEN msg_flag_char := '?';
    IF (msg_flag AND mf_fwd_select)  <> 0  THEN msg_flag_char := '+';
    IF (msg_flag AND mf_fwd_process) <> 0  THEN msg_flag_char := '-';
    IF (msg_flag AND mf_fwd)         <> 0  THEN msg_flag_char := 'F';
    IF (msg_flag AND mf_read)        <> 0  THEN msg_flag_char := 'Y';
    IF (msg_flag AND mf_hold)        <> 0  THEN msg_flag_char := 'H';
    IF (msg_flag AND mf_review)      <> 0  THEN msg_flag_char := 'V';
    IF (msg_flag AND mf_kill)        <> 0  THEN msg_flag_char := 'K';

  END;
