(*===========================================================================*)
(* Misc                                                                      *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1991 by H. Roy Engehausen.  All rights reserved.  *)
(*                                                                           *)
(*===========================================================================*)

{$UNDEF DEBUG_CS} (* Debug call scan in distribution processing *)

(*===========================================================================*)
(* Checks to see if a message is hidden  -- Respond TRUE if NOT hidden       *)
(*===========================================================================*)

FUNCTION check_hidden (i_ptr  : msg_index_ptr): BOOLEAN;

  CONST no_can_see = mf_kill OR mf_hold OR mf_review;

  BEGIN;

    check_hidden := TRUE;

    WITH active_tcb^.uid_data DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* If the user is a BBS remote sysop or higher, he can see all       *)
        (* If the user is a BBS then OPTION must be on to see all            *)
        (*-------------------------------------------------------------------*)

        IF (user_class >= user_c_rsu) OR
                    ((user_class = user_c_bu) AND opt_block.opt_bbs_see_p) THEN
          EXIT;

        (*-------------------------------------------------------------------*)
        (* If the user can be a SYSOP, then he can see all                   *)
        (*-------------------------------------------------------------------*)

        IF active_tcb^.tcb_sysop_pw_ok
                                   AND ((user_flag AND user_f_sysop) <> 0) THEN
          EXIT;

        (*-------------------------------------------------------------------*)
        (* If not held or killed, then NO                                    *)
        (*-------------------------------------------------------------------*)

        IF (i_ptr^.msg_i_mb.msg_flag AND no_can_see) <> 0 THEN
          BEGIN;
            check_hidden := FALSE;
            EXIT;
          END;

        (*-------------------------------------------------------------------*)
        (* If private and not addressed to/from this person then NO          *)
        (*-------------------------------------------------------------------*)

        IF (i_ptr^.msg_i_mb.msg_type = mt_private)
                              AND (i_ptr^.msg_i_mb.msg_to <> user_id)
                              AND (i_ptr^.msg_i_mb.msg_from <> user_id) THEN
          BEGIN;
            check_hidden := FALSE;
            EXIT;
          END;

        (*-------------------------------------------------------------------*)
        (* Everything else is true                                           *)
        (*-------------------------------------------------------------------*)

      END;

  END;

(*===========================================================================*)
(* Verify authorization list                                                 *)
(*===========================================================================*)

FUNCTION verify_auth_list(in_str : str_ptr; who : call_sign_str) : BOOLEAN;

  VAR
    b : BOOLEAN;
    i : BYTE;
    w : BYTE;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Get the word count                                                    *)
    (*-----------------------------------------------------------------------*)

    w := words(in_str^);

    (*-----------------------------------------------------------------------*)
    (* Initialize for loop                                                   *)
    (*-----------------------------------------------------------------------*)

    i := 1;
    b := FALSE;

    (*-----------------------------------------------------------------------*)
    (* Loop until tests are done or we have a TRUE                           *)
    (*-----------------------------------------------------------------------*)

    WHILE (NOT b) AND (i < w) DO
      BEGIN;
        INC(i);
        b := match_str(who, subword(in_str, i, 1))
      END;

    (*-----------------------------------------------------------------------*)
    (* If this is a NEGATIVE test then invert the sense of the test          *)
    (*-----------------------------------------------------------------------*)

    IF substr(in_str^, 1, 3) = 'NOT' THEN
      b := NOT b;

    (*-----------------------------------------------------------------------*)
    (* Set result and leave                                                  *)
    (*-----------------------------------------------------------------------*)

    verify_auth_list := b

  END;

(*===========================================================================*)
(* Build a distribution list                                                 *)
(*===========================================================================*)

PROCEDURE build_dis(msg_to_do : msg_index_ptr;
                    dis_file_name : file_name_str);

  VAR
    b             : BOOLEAN;
    calls_scanned : str_ptr;
    d_size        : WORD;
    dis_cnt       : BYTE;
    dis_data      : STRING;
    dis_file      : TEXT;
    i             : BYTE;
    wrd_count     : BYTE;
    same_dis      : str_chain_ptr;
    scan_ok       : BOOLEAN;
    this_call     : str_ptr;
    word_data     : bb_addr_str;

  LABEL
    read_loop;

  BEGIN;

    {$IFDEF DEBUG_DIST}
      WRITELN('Distribution call ', dis_file_name);
      DELAY(1000);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Get ready to do distribution                                          *)
    (*-----------------------------------------------------------------------*)

    msg_to_do^.msg_i_dis := NIL;

    IF dis_file_name = '' THEN
      dis_file_name := msg_to_do^.msg_i_mb.msg_to_at;

    {$IFDEF DEBUG_DIST}
      WRITELN('Verify no override -- ', dis_file_name);
      DELAY(1000);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Build the file name.  Truncate the address if necessary               *)
    (*-----------------------------------------------------------------------*)

    IF POS('.', dis_file_name) = 0 THEN
      BEGIN;
        IF LENGTH(dis_file_name) > 8 THEN
          dis_file_name := SUBSTR(dis_file_name, 1, 8);

        dis_file_name := opt_block.msg_file_dir + dis_file_name + '.DIS';
      END;

    {$IFDEF DEBUG_DIST}
      WRITELN('Assign file name -- ', dis_file_name);
      DELAY(1000);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Open the file for input                                               *)
    (*-----------------------------------------------------------------------*)

    ASSIGN(dis_file, dis_file_name);
    {$I-}
    RESET(dis_file);
    {$I+}

    (*-----------------------------------------------------------------------*)
    (* If the open fails then nothing is necessary                           *)
    (*-----------------------------------------------------------------------*)

    IF IORESULT <> 0 THEN
      EXIT;

    (*-----------------------------------------------------------------------*)
    (* Get a list of calls scanned in headers                                *)
    (*-----------------------------------------------------------------------*)

    calls_scanned := find_task_mem_addr('CS');
    scan_ok       := calls_scanned <> NIL;

    {$IFDEF DEBUG_CS}
      WRITELN('CScanSW=', scan_ok);
      IF scan_ok THEN
        WRITELN('CScanD=', calls_scanned^);
      DELAY(2000);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Get a distribution area if one isn't handy                            *)
    (*-----------------------------------------------------------------------*)

    IF dis_area = NIL THEN
      NEW(dis_area);

    dis_cnt := 0;

    (*-----------------------------------------------------------------------*)
    (* Loop around reading the distribution list into the area.              *)
    (*-----------------------------------------------------------------------*)

    WITH dis_area^ DO
      BEGIN;

read_loop:

        WHILE (NOT EOF(dis_file)) DO
          BEGIN;

            (*---------------------------------------------------------------*)
            (* Read a line and strip blanks                                  *)
            (*---------------------------------------------------------------*)

            READLN(dis_file, dis_data);

            strip_var(dis_data, 'B');

            upcase_str_var(dis_data);

            IF comment_line(dis_data) THEN
              GOTO read_loop;

            (*---------------------------------------------------------------*)
            (* Get the data                                                  *)
            (*---------------------------------------------------------------*)

            word_data := subword(@dis_data, 1, 1);

            (*---------------------------------------------------------------*)
            (* See if a special statement.  If so handle                     *)
            (*---------------------------------------------------------------*)

            IF (word_data = 'OK') OR (word_data = 'NOT_OK') THEN
              IF NOT verify_auth_list(@dis_data,
                                      active_tcb^.uid_data.user_id) THEN
                BEGIN;
                  CLOSE(dis_file);
                  EXIT;
                END
              ELSE
                word_data := '';

            (*---------------------------------------------------------------*)
            (* If data exists and we have room for it, put it in the array   *)
            (*---------------------------------------------------------------*)

            IF (word_data <> '') AND (dis_cnt < msg_dist_max) THEN
              BEGIN;

                dis_cnt := dis_cnt + 1;

                WITH active_tcb^.uid_data, msg_d_array[dis_cnt] DO
                  BEGIN;

                    msg_d_info := word_data;
                    msg_d_flag := 0;

                    (*-------------------------------------------------------*)
                    (* This little section of code checks to see if this     *)
                    (* the message is coming from a station on the           *)
                    (* distribution list.  If it is, then we mark it as      *)
                    (* already forwarded to.   If not, check all the         *)
                    (* headers to see if we already went thru someone        *)
                    (*-------------------------------------------------------*)

                    IF find(@dis_data, @user_id) <> 0 THEN
                      msg_d_flag := df_fwd
                    ELSE
                      IF scan_ok THEN
                        BEGIN

                          wrd_count := words(dis_data);
                          i         := 0;
                          b         := FALSE;

                          REPEAT

                            INC(i);
                            word_data := subword(@dis_data, i, 1);

                            {$IFDEF DEBUG_CS}
                              WRITELN('SCANNING=', i, '/', wrd_count, '=',
                                                                    word_data);
                              DELAY(2000);
                            {$ENDIF}

                            IF find(calls_scanned, @word_data) <> 0 THEN
                              BEGIN;

                                {$IFDEF DEBUG_CS}
                                  WRITELN('Found it');
                                {$ENDIF}

                                msg_d_flag := df_fwd;
                                b := TRUE;

                              END;

                          UNTIL b OR (i > wrd_count);

                        END; (*----- End check thru all stations on scan ----*)

                  END; (*----- End addressing into distribution list --------*)

              END; (*----- End putting route into distribution list ---------*)

          END; (*----- End read distribution list loop ----------------------*)

        msg_d_no := dis_cnt;

      END;

    (*-----------------------------------------------------------------------*)
    (* Close things up                                                       *)
    (*-----------------------------------------------------------------------*)

    CLOSE(dis_file);

    (*-----------------------------------------------------------------------*)
    (* If the distribution list is complete, mark message as done            *)
    (*-----------------------------------------------------------------------*)

    i := dis_cnt;
    WHILE (i > 0)
                AND ((dis_area^.msg_d_array[i].msg_d_flag AND df_fwd) <> 0) DO
      DEC(i);

    IF i < 1 THEN
      msg_to_do^.msg_i_mb.msg_flag := msg_to_do^.msg_i_mb.msg_flag OR mf_fwd;

    (*-----------------------------------------------------------------------*)
    (* Move the distribution list to the right spot and set flags.           *)
    (*-----------------------------------------------------------------------*)

    d_size := 1 + dis_area^.msg_d_no * SIZEOF(msg_dist_entry_type);

    GETMEM(msg_to_do^.msg_i_dis, d_size);
    MOVE(dis_area^, msg_to_do^.msg_i_dis^, d_size);

    msg_to_do^.msg_i_mb.msg_flag :=
                                   msg_to_do^.msg_i_mb.msg_flag OR mf_fwd_list;

  END;

(*===========================================================================*)
(* Count msg list                                                            *)
(*===========================================================================*)

PROCEDURE count_msg_list;

  VAR
    i_ptr : msg_index_ptr;

  BEGIN;

    msg_counter_ok    := 0;
    msg_counter_kill  := 0;

    i_ptr := msg_index_start;

    WHILE i_ptr <> NIL DO
      BEGIN;

        IF (i_ptr^.msg_i_mb.msg_flag AND mf_kill) <> 0 THEN
          INC(msg_counter_kill)
        ELSE
          INC(msg_counter_ok);

        i_ptr := i_ptr^.msg_i_next;

      END;

    END;
