(*===========================================================================*)
(* Misc #2                                                                   *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1991 by H. Roy Engehausen.  All rights reserved.  *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Marks a message as send, etc                                              *)
(*===========================================================================*)

PROCEDURE mark_as_forwarded (i_ptr         : msg_index_ptr;
                             dis_number    : BYTE;
                             response_char : CHAR);

  VAR
    b_allsent : BOOLEAN;
    b_inprog  : BOOLEAN;
    b_select  : BOOLEAN;
    d_block   : msg_d_ptr;
    f1        : msg_flag_type;
    f2        : msg_flag_type;
    i         : BYTE;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Assume whole message sent if we didn't get a later                    *)
    (*-----------------------------------------------------------------------*)

    b_allsent  := response_char <> 'L';

    (*-----------------------------------------------------------------------*)
    (* Assume whole message not in progress                                  *)
    (*-----------------------------------------------------------------------*)

    b_inprog   := FALSE;
    b_select   := FALSE;

    (*-----------------------------------------------------------------------*)
    (* Distribution list?                                                    *)
    (*-----------------------------------------------------------------------*)

    IF dis_number <> 0 THEN
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Handle distribution list message sent.                            *)
        (*-------------------------------------------------------------------*)

        (*-------------------------------------------------------------------*)
        (* If we have a distribution route block then reset it and           *)
        (* get the distribution block pointer from it.  If there is          *)
        (* no route bock then we just get the pointer.                       *)
        (*-------------------------------------------------------------------*)

        IF (i_ptr^.msg_i_mb.msg_flag AND mf_disrout) <> 0 THEN
          BEGIN;

           {$IFDEF POINT_CHK}
             test_pointer(i_ptr^.msg_i_dr);
           {$ENDIF}

            i_ptr^.msg_i_dr^.msg_dr_data[dis_number] := NIL;

          END;

        (*-------------------------------------------------------------------*)
        (* Mark item as ether sent or waiting                                *)
        (*-------------------------------------------------------------------*)

        d_block := find_dist_list(i_ptr);

        {$IFDEF POINT_CHK}
          test_pointer(d_block);
        {$ENDIF}

        WITH d_block^.msg_d_array[dis_number] DO
          IF (response_char = 'L')
                  OR ((response_char = 'R')
                                    AND (active_port^.reject_bul = 1)) THEN
            msg_d_flag := msg_d_flag AND NOT df_fwd_process
          ELSE
            msg_d_flag := (msg_d_flag OR df_fwd) AND
                                     NOT (df_fwd_select OR df_fwd_process);

        (*-------------------------------------------------------------------*)
        (* Check to see if whole list is sent, any still in process,         *)
        (* and any still selected.                                           *)
        (*-------------------------------------------------------------------*)

        FOR i := 1 TO d_block^.msg_d_no DO
          WITH d_block^.msg_d_array[i] DO
            BEGIN;
              b_allsent := b_allsent AND ((df_fwd AND msg_d_flag) <> 0);
              b_inprog  := b_inprog
                                 OR ((df_fwd_process AND msg_d_flag) <> 0);
              b_select  := b_select
                                 OR ((df_fwd_select AND msg_d_flag) <> 0);
            END;

        (*-------------------------------------------------------------------*)
        (* If reject then put in hold if requested                           *)
        (*-------------------------------------------------------------------*)

        IF (response_char = 'R') AND (active_port^.reject_bul = 2) THEN
          BEGIN;

            response_char       := 'O';

            b_allsent           := FALSE;
            b_inprog            := FALSE;
            b_select            := FALSE;

            i_ptr^.msg_i_mb.msg_flag   := i_ptr^.msg_i_mb.msg_flag OR mf_hold;
            i_ptr^.msg_i_mb.msg_reason := message_reason_reject;

          END;

      END  (*----- End handling a distribution list send --------------------*)
    ELSE
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Handle non-distribution list message sent.  Remove route if done  *)
        (*-------------------------------------------------------------------*)

        IF b_allsent THEN
          i_ptr^.msg_i_rou := NIL
        ELSE
          b_select := TRUE;

      END;

    (*-----------------------------------------------------------------------*)
    (* Either a single item or whole list sent.                              *)
    (*-----------------------------------------------------------------------*)

    IF b_allsent THEN
      BEGIN;

        IF response_char = 'R' THEN
          BEGIN;

            (*---------------------------------------------------------------*)
            (* Message was rejected.  Put in hold                            *)
            (*---------------------------------------------------------------*)

            f1 := mf_hold;
            i_ptr^.msg_i_mb.msg_reason := message_reason_reject;

          END
        ELSE
          BEGIN;

            (*---------------------------------------------------------------*)
            (* Set forward flag.  KILL if right type and NOT a               *)
            (* distribution list                                             *)
            (*---------------------------------------------------------------*)

            IF (POS(i_ptr^.msg_i_mb.msg_type, opt_block.nofwd_kill) = 0)
                                                     AND (dis_number = 0) THEN
               f1 := mf_fwd OR mf_kill
            ELSE
               f1 := mf_fwd;

          END;

        i_ptr^.msg_i_mb.msg_flag := i_ptr^.msg_i_mb.msg_flag OR f1;
      END;

    (*-----------------------------------------------------------------------*)
    (* If not still in progress then reset master progress bit               *)
    (*-----------------------------------------------------------------------*)

    IF NOT b_inprog THEN
      i_ptr^.msg_i_mb.msg_flag := i_ptr^.msg_i_mb.msg_flag
                                                      AND (NOT mf_fwd_process);

    (*-----------------------------------------------------------------------*)
    (* If not still selected then reset master select bit                    *)
    (*-----------------------------------------------------------------------*)

    IF NOT b_select THEN

      i_ptr^.msg_i_mb.msg_flag := i_ptr^.msg_i_mb.msg_flag
                                                       AND (NOT mf_fwd_select);

  END;
