(*===========================================================================*)
(* Handle msg files -- Initialization, termination, and cleanup              *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

{$DEFINE POINT_CHK}
{$DEFINE FREE_CHK}

{$UNDEF  DEBUG_FREE}
{$UNDEF  DEBUG_GM} (* Debug new GM code *)

{$O+}

UNIT BBMFI;

INTERFACE

  PROCEDURE open_msg;
  PROCEDURE oper_gm(cmd_string : STRING);

IMPLEMENTATION

USES
  DOS,
  bbdummy,
  bbdump,
  bbfnr,
  bbmdata,
  bbmess,
  bbmisc2,
  bbmisc3,
  bbmf,
  bbrunerr,
  bbsema2,
  bbstr,
  bbtask,
  bbtime,
  bbtrace,
  bbwin;

PROCEDURE oper_gl;       FORWARD;
PROCEDURE free_msg_list; FORWARD;

CONST stars = '***************************';

(*===========================================================================*)
(* Open msg file                                                             *)
(*===========================================================================*)

PROCEDURE open_msg;

  VAR
    i                    : WORD;
    j                    : WORD;
    k                    : WORD;
    m_overhead1          : WORD;
    m_overhead2          : WORD;
    m_size               : WORD;
    msg_buffer           : msg_block;
    msg_index_current    : msg_index_ptr;

  CONST
    dont_get_dist_flags : msg_flag_type = mf_hold
                                          OR mf_fwd
                                          OR mf_read
                                          OR mf_kill
                                          OR mf_review;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Open msg file for read                                                *)
    (*-----------------------------------------------------------------------*)

    ASSIGN(msg_file, opt_block.msg_file_name);
    {$I-}
    RESET(msg_file);
    {$I+}
    i := IORESULT;

    (*-----------------------------------------------------------------------*)
    (* Error opening file                                                    *)
    (*-----------------------------------------------------------------------*)

    IF (i <> 0) AND (i <> 2) THEN
      BEGIN;
        WRITELN('Unable to open ', opt_block.msg_file_name);
        WRITELN('DOS Error # ', i);
      END;

    (*-----------------------------------------------------------------------*)
    (* File doesn't exist.  Initialize it                                    *)
    (*-----------------------------------------------------------------------*)

    IF i <> 0 THEN
      BEGIN;

        REWRITE(msg_file);

        FILLCHAR(msg_buffer, SIZEOF(msg_buffer), CHR(0));

        msg_buffer.msg_number := sys_version;
        WRITE(msg_file, msg_buffer);

        CLOSE(msg_file);

        RESET(msg_file);

      END;

    (*-----------------------------------------------------------------------*)
    (* Check the version of the file                                         *)
    (*-----------------------------------------------------------------------*)

    READ(msg_file, msg_buffer);
    i := msg_buffer.msg_number;
    IF i <> sys_version THEN
      BEGIN;
        WRITELN;
        WRITELN(stars);
        WRITELN('Wanted version ', sys_version, ' message file');
        WRITELN('Got version ', i, ' message file');
        WRITELN(stars);
        WRITELN;
        RUNERROR(msg_runerr);
      END;

    (*-----------------------------------------------------------------------*)
    (* Read in the message file                                              *)
    (*-----------------------------------------------------------------------*)

    msg_index_start   := NIL;
    msg_index_current := NIL;
    msg_index_end     := NIL;
    next_msg_no       := 1;
    next_record_no    := 1;

    m_overhead1 := OFS(msg_index_start^.msg_i_mb.msg_subj[1])
                                                       - OFS(msg_index_start^);

    m_overhead2 := OFS(msg_index_start^.msg_i_mb.msg_subj[1])
                                              - OFS(msg_index_start^.msg_i_mb);

    i := 0;
    k := FILESIZE(msg_file) - 1;

    WHILE i < k DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Read in the next record                                           *)
        (*-------------------------------------------------------------------*)

        INC(i);

        READ(msg_file, msg_buffer);
        INC(next_record_no);

        (*-------------------------------------------------------------------*)
        (* Reset any wayward things                                          *)
        (*-------------------------------------------------------------------*)

        WITH msg_buffer DO
          BEGIN;

            msg_flag := msg_flag
                                 AND NOT (mf_fwd_select OR mf_fwd_process
                                                  OR mf_disrout OR mf_unknown);

            IF msg_dt_in > current_day_time THEN
              msg_dt_in := current_day_time;

            IF msg_dt_orig > current_day_time THEN
              msg_dt_orig := current_day_time;

          END;

        (*-------------------------------------------------------------------*)
        (* Save number as being the highest so far                           *)
        (*-------------------------------------------------------------------*)

        WITH msg_buffer DO
          IF next_msg_no > msg_number THEN
            BEGIN;
              WRITELN('**** Warning -- MSG file out of sequence');
              WRITELN('Record number # ', i, ' is msg # ', msg_number);
              WRITELN('High   number # ', next_msg_no);
            END
          ELSE
            next_msg_no := msg_number + 1;

        (*----------------------------------------------------------------*)
        (* Store record in the chain                                         *)
        (*-------------------------------------------------------------------*)

        m_size := m_overhead1 + LENGTH(msg_buffer.msg_subj);
        GETMEM(msg_index_current, m_size);

        WITH msg_index_current^ DO
          BEGIN

            IF msg_index_start = NIL THEN
              msg_index_start := msg_index_current
            ELSE
              msg_index_end^.msg_i_next := msg_index_current;

            msg_i_last  := msg_index_end;
            msg_i_next  := NIL;
            msg_i_dis   := NIL;

            msg_i_record := i;

            m_size := m_overhead2 + LENGTH(msg_buffer.msg_subj);
            MOVE(msg_buffer, msg_i_mb, m_size);

          END;

        (*-------------------------------------------------------------------*)
        (* Read in distribution list                                         *)
        (*-------------------------------------------------------------------*)

        WITH msg_index_current^ DO
          BEGIN;

            (*---------------------------------------------------------------*)
            (* Test for distribution list exists                             *)
            (*---------------------------------------------------------------*)

            IF (mf_fwd_list AND msg_i_mb.msg_flag) <> 0 THEN
              BEGIN;

                i := i + 1;

                READ(msg_file, msg_buffer);
                INC(next_record_no);

                (*-----------------------------------------------------------*)
                (* Should the record be loaded?                              *)
                (*-----------------------------------------------------------*)

                IF (msg_i_mb.msg_flag AND dont_get_dist_flags) <> 0 THEN
                  msg_i_dis := NIL
                ELSE
                  BEGIN;

                    (*-------------------------------------------------------*)
                    (* Load the dirstribution record into storage            *)
                    (*-------------------------------------------------------*)

                    msg_i_dis := @msg_buffer;
                    j := msg_i_dis^.msg_d_no;

                    IF j > msg_dist_max THEN
                      BEGIN;
                        WRITELN;
                        WRITELN(stars);
                        WRITELN('Invalid distribution # -- ', j);
                        WRITELN('I = ', i);
                        WRITELN(stars);
                        WRITELN;
                        dump_trace;
                        dump_msg(msg_index_current);
                        RUNERROR(msg_runerr);
                      END;

                    j := 1 + j * SIZEOF(msg_dist_entry_type);

                    GETMEM(msg_i_dis, j);

                    MOVE(msg_buffer, msg_i_dis^, j);

                    WITH msg_i_dis^ DO
                      FOR j := 1 TO msg_d_no DO
                        WITH msg_d_array[j] DO
                          msg_d_flag := msg_d_flag
                                        AND NOT (df_fwd_select
                                          OR df_fwd_process OR df_fwd_unknown);

                  END; (*----- End of load distribution into storgae --------*)

              END; (*----- End of test for distribution list ----------------*)

          END; (*----- End of WITH statement --------------------------------*)

        (*-------------------------------------------------------------------*)
        (* Save pointer to last item in chain.  This may be it!              *)
        (*-------------------------------------------------------------------*)

        msg_index_end := msg_index_current;

      END; (*----- End of loop reading records ------------------------------*)

    (*-----------------------------------------------------------------------*)
    (* Cleanup                                                               *)
    (*-----------------------------------------------------------------------*)

    CLOSE(msg_file);

    dis_area := NIL;

    msg_route_num := 0;

    count_msg_list;

    fnm_linked_list_valid := FALSE;

  END;

(*===========================================================================*)
(* Operator reorganize msgs                                                  *)
(*===========================================================================*)

PROCEDURE oper_gm(cmd_string : STRING);

  VAR

    b                 : BOOLEAN;
    dis_buffer        : msg_block;
    buff_add          : msg_block_ptr;
    fn_1              : file_name_str;
    fn_2              : file_name_str;
    m_fn_old          : file_name_str;
    m_fn_new          : file_name_str;
    m_fn_dir          : file_name_str;
    msg_buffer        : msg_block;
    msg_index_current : msg_index_ptr;
    msg_loop          : msg_index_ptr;
    i                 : INTEGER;
    j                 : LONGINT;
    k                 : INTEGER;
    new_num           : WORD;
    new_file          : FILE OF msg_block;
    renum             : BOOLEAN;
    rec_cnt           : WORD;
    rec_max           : WORD;
    re_file           : TEXT;
    type_gm           : CHAR;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Get type of operation                                                 *)
    (*-----------------------------------------------------------------------*)

    type_gm := UPCASE(cmd_string[2]);

    (*-----------------------------------------------------------------------*)
    (* Only GR can have more than one word.                                  *)
    (*-----------------------------------------------------------------------*)

    i := WORDS(cmd_string);
    IF (i > 2) OR ((i > 1) AND (type_gm <> 'R')) THEN
      BEGIN;
        send_message(message_err_wrd);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Run type L                                                            *)
    (*-----------------------------------------------------------------------*)

    IF type_gm = 'L' THEN
      BEGIN;
        oper_gl;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Set type M/R switch                                                   *)
    (*-----------------------------------------------------------------------*)

    renum := type_gm = 'R';

    (*-----------------------------------------------------------------------*)
    (* Process optional operand on GR                                        *)
    (*-----------------------------------------------------------------------*)

    IF renum THEN
      BEGIN;

        IF i = 1 THEN
          new_num := 0
        ELSE
          BEGIN;

            VAL(subword(@cmd_string, 2, 1), j, i);

            IF (i <> 0) OR (j > 65000) OR (j < 1) THEN
              BEGIN;
                send_message(message_err_ivm);
                active_tcb^.error_sw := TRUE;
                EXIT;
              END;

            new_num := j - 1;

          END;

      END;

    (*-----------------------------------------------------------------------*)
    (* See if we can run.                                                    *)
    (*-----------------------------------------------------------------------*)

    IF bbs_busy THEN
      BEGIN;
        cannot_do_this(message_other_active);
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Obtain the interrupt lock                                             *)
    (*-----------------------------------------------------------------------*)

    get_semaphore(semaphore_interrupts, sem_exclusive, TRUE);

    (*-----------------------------------------------------------------------*)
    (* Tell user started                                                     *)
    (*-----------------------------------------------------------------------*)

    window_write(active_tcb^.port_chan_s + '>:',
                    'Reorg is started -- All tasks are locked.  Please wait.');

    (*-----------------------------------------------------------------------*)
    (* Isolate the directory and/or drive of the msg_file_name.  Third and   *)
    (* fourth parameters are place holders.  The output is not used.         *)
    (*-----------------------------------------------------------------------*)

    FSPLIT(opt_block.msg_file_name, m_fn_dir, m_fn_new, m_fn_old);

    (*-----------------------------------------------------------------------*)
    (* Generate some file names for work info                                *)
    (*-----------------------------------------------------------------------*)

    m_fn_new := m_fn_dir + 'BBNEWMSG.BB';
    m_fn_old := m_fn_dir + 'BBOLDMSG.BB';

    (*-----------------------------------------------------------------------*)
    (* Destroy all inside information                                        *)
    (*-----------------------------------------------------------------------*)

    free_msg_list;

    (*-----------------------------------------------------------------------*)
    (* Cause a reload so we can firm file is valid!                          *)
    (*-----------------------------------------------------------------------*)

    open_msg;

    (*-----------------------------------------------------------------------*)
    (* Destroy all inside information                                        *)
    (*-----------------------------------------------------------------------*)

    free_msg_list;

    (*-----------------------------------------------------------------------*)
    (* Open msg file for read                                                *)
    (*-----------------------------------------------------------------------*)

    ASSIGN(msg_file, opt_block.msg_file_name);
    {$I-}
    RESET(msg_file);
    {$I+}
    i := IORESULT;

    (*-----------------------------------------------------------------------*)
    (* Error opening file                                                    *)
    (*-----------------------------------------------------------------------*)

    IF i <> 0 THEN
      BEGIN;
        WRITELN;
        WRITELN(stars);
        WRITELN('Unable to open ', opt_block.msg_file_name);
        WRITELN('DOS Error # ', i);
        WRITELN(stars);
        WRITELN;
        RUNERROR(msg_runerr);
      END;

    (*-----------------------------------------------------------------------*)
    (* Check the version of the file                                         *)
    (*-----------------------------------------------------------------------*)

    READ(msg_file, msg_buffer);
    i := msg_buffer.msg_number;
    IF i <> sys_version THEN
      BEGIN;
        WRITELN;
        WRITELN(stars);
        WRITELN('Wanted version ', sys_version, ' message file');
        WRITELN('Got version ', i, ' message file');
        WRITELN(stars);
        WRITELN;
        RUNERROR(msg_runerr);
      END;

    (*-----------------------------------------------------------------------*)
    (* Open a new file for output                                            *)
    (*-----------------------------------------------------------------------*)

    ASSIGN(new_file, m_fn_new);
    REWRITE(new_file);

    FILLCHAR(msg_buffer, SIZEOF(msg_buffer), #0);

    msg_buffer.msg_number := sys_version;
    WRITE(new_file, msg_buffer);

    (*-----------------------------------------------------------------------*)
    (* Loop thru writing the data to a new file                              *)
    (*-----------------------------------------------------------------------*)

    rec_cnt := 0;
    rec_max := FILESIZE(msg_file) - 1;

    WHILE rec_cnt < rec_max DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Read in the next record                                           *)
        (*-------------------------------------------------------------------*)

        INC(rec_cnt);

        READ(msg_file, msg_buffer);

        b := (msg_buffer.msg_flag AND mf_fwd_list) <> 0;

        IF b THEN
          BEGIN;
            INC(rec_cnt);
            READ(msg_file, dis_buffer);
          END;

        {$IFDEF DEBUG_GM}
          WRITELN('Read msg # ', msg_buffer.msg_number, ' as record ', rec_cnt,
                  ' -- dist = ', b);
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Different actions based on KILL flag, etc.  We kill all           *)
        (* messages but the last one.  This is needed to hold the number     *)
        (* of the last message                                               *)
        (*-------------------------------------------------------------------*)

        IF ((msg_buffer.msg_flag AND mf_kill) = 0)
                                    OR ((rec_cnt = rec_max) AND NOT renum) THEN
          BEGIN;

            {$IFDEF DEBUG_GM}
              WRITELN('Msg # ', msg_buffer.msg_number, ' to be passed on');
            {$ENDIF}

            (*---------------------------------------------------------------*)
            (* KILL flag off....  See if we want to renumber                 *)
            (*---------------------------------------------------------------*)

            IF renum THEN
              BEGIN;

                (*-----------------------------------------------------------*)
                (* Renumber                                                  *)
                (*-----------------------------------------------------------*)

                new_num := new_num + 1;

                (*-----------------------------------------------------------*)
                (* Generate file names                                       *)
                (*-----------------------------------------------------------*)

                STR(msg_buffer.msg_number, fn_1);
                fn_1 := opt_block.msg_file_dir + 'BB' + fn_1 + '.MSG';
                STR(new_num, fn_2);
                fn_2 := opt_block.msg_file_dir + 'BB' + fn_2 + '.MSG';

                (*-----------------------------------------------------------*)
                (* Don't rename if the same                                  *)
                (*-----------------------------------------------------------*)

                IF fn_1 <> fn_2 THEN
                  BEGIN;

                    (*-------------------------------------------------------*)
                    (* Erase any existing file under new name                *)
                    (*-------------------------------------------------------*)

                    ASSIGN(re_file, fn_2);
                    {$I-}
                    ERASE(re_file);
                    i := IORESULT;
                    {$I+}

                    (*-------------------------------------------------------*)
                    (* Rename the message                                    *)
                    (*-------------------------------------------------------*)

                    ASSIGN(re_file, fn_1);
                    {$I-}
                    RENAME(re_file, fn_2);
                    i := IORESULT;
                    {$I+}

                    IF i <> 0 THEN
                      window_write(active_tcb^.port_chan_s + '::',
                             'Error on rename of ' + fn_1 + ' to ' + fn_2);

                  END;

                (*-----------------------------------------------------------*)
                (* Set the new number                                        *)
                (*-----------------------------------------------------------*)

                msg_buffer.msg_number := new_num;

              END;

            (*---------------------------------------------------------------*)
            (* Write of the new block.  If necessary, write out dist.        *)
            (*---------------------------------------------------------------*)

            {$IFDEF DEBUG_GM}
              WRITELN('Writing msg # ', msg_buffer.msg_number);
            {$ENDIF}

            WRITE(new_file, msg_buffer);

            IF b THEN
              WRITE(new_file, dis_buffer);

          END (*----- End steps when KILL flag is off -----------------------*)

        ELSE
          BEGIN;

            (*---------------------------------------------------------------*)
            (* KILL flag on.  Handle killed msg                              *)
            (*---------------------------------------------------------------*)

            {$IFDEF DEBUG_GM}
              WRITELN('Discard msg # ', msg_buffer.msg_number);
            {$ENDIF}

            STR(msg_buffer.msg_number, fn_2);
            fn_1 := opt_block.msg_file_dir + 'BB' + fn_2 + '.MSG';

            IF opt_block.opt_rename_k THEN
              BEGIN;

                (*-----------------------------------------------------------*)
                (* Rename the msg                                            *)
                (*-----------------------------------------------------------*)

                fn_2 := opt_block.msg_file_dir + 'BB' + fn_2 + '.KIL';

                ASSIGN(re_file, fn_2);
                {$I-}
                ERASE(re_file);
                {$I+}
                i := IORESULT;

                ASSIGN(re_file, fn_1);
                {$I-}
                RENAME(re_file, fn_2);
                {$I+}
                i := IORESULT;

                IF i <> 0 THEN
                  window_write(active_tcb^.port_chan_s + '::',
                                             'Error on rename of ' + fn_1);

              END  (*----- End KILL rename ----------------------------------*)

            ELSE

              BEGIN;

                (*-----------------------------------------------------------*)
                (* Erase the msg                                             *)
                (*-----------------------------------------------------------*)


                ASSIGN(re_file, fn_1);
                {$I-}
                ERASE(re_file);
                {$I+}
                i := IORESULT;

                IF i <> 0 THEN
                  window_write(active_tcb^.port_chan_s + '::',
                                             'Error on erase of ' + fn_1);

              END; (*----- End KILL erase -----------------------------------*)

          END; (*----- End KILL flag on actions -----------------------------*)

      END; (*----- End loop for writing out the new file --------------------*)

    (*-----------------------------------------------------------------------*)
    (* Close things up                                                       *)
    (*-----------------------------------------------------------------------*)

    CLOSE(msg_file);
    CLOSE(new_file);

    (*-----------------------------------------------------------------------*)
    (* Erase any old file                                                    *)
    (*-----------------------------------------------------------------------*)

    ASSIGN(re_file, m_fn_old);
    {$I-}
    ERASE(re_file);
    {$I+}
    i := IORESULT;

    (*-----------------------------------------------------------------------*)
    (* Rename current file to old file                                       *)
    (*-----------------------------------------------------------------------*)

    ASSIGN(re_file, opt_block.msg_file_name);
    RENAME(re_file, m_fn_old);

    (*-----------------------------------------------------------------------*)
    (* Rename new file to current file                                       *)
    (*-----------------------------------------------------------------------*)

    RENAME(new_file, opt_block.msg_file_name);

    (*-----------------------------------------------------------------------*)
    (* Reload data                                                           *)
    (*-----------------------------------------------------------------------*)

    open_msg;

    (*-----------------------------------------------------------------------*)
    (* Do a task switch here to allow unblocked tasks to update              *)
    (* Also verify time                                                      *)
    (*-----------------------------------------------------------------------*)

    task_switch;
    time_check;

    (*-----------------------------------------------------------------------*)
    (* Release the interrupt lock                                            *)
    (*-----------------------------------------------------------------------*)

    free_semaphore(semaphore_interrupts);

    (*-----------------------------------------------------------------------*)
    (* Set H lookup switch                                                   *)
    (*-----------------------------------------------------------------------*)

    IF renum THEN
      last_h_msg := 0;

    (*-----------------------------------------------------------------------*)
    (* Tell user done                                                        *)
    (*-----------------------------------------------------------------------*)

    window_write(active_tcb^.port_chan_s + '>:', 'Reorg is done');

  END;

(*===========================================================================*)
(* Reload messages                                                           *)
(*===========================================================================*)

PROCEDURE oper_gl;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* See if we can run.                                                    *)
    (*-----------------------------------------------------------------------*)

    IF bbs_busy THEN
      BEGIN;
        cannot_do_this(message_other_active);
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Obtain the interrupt lock                                             *)
    (*-----------------------------------------------------------------------*)

    get_semaphore(semaphore_interrupts, sem_exclusive, TRUE);

    (*-----------------------------------------------------------------------*)
    (* Tell user started                                                     *)
    (*-----------------------------------------------------------------------*)

    window_write(active_tcb^.port_chan_s + '>:',
                   'Reload is started -- All tasks are locked.  Please wait.');

    (*-----------------------------------------------------------------------*)
    (* Destroy all inside information                                        *)
    (*-----------------------------------------------------------------------*)

    free_msg_list;

    (*-----------------------------------------------------------------------*)
    (* Get a new set                                                         *)
    (*-----------------------------------------------------------------------*)

    open_msg;

    (*-----------------------------------------------------------------------*)
    (* Do a task switch here to allow unblocked tasks to update              *)
    (* Also verify time                                                      *)
    (*-----------------------------------------------------------------------*)

    task_switch;
    time_check;

    (*-----------------------------------------------------------------------*)
    (* Release the interrupt lock                                            *)
    (*-----------------------------------------------------------------------*)

    free_semaphore(semaphore_interrupts);

    (*-----------------------------------------------------------------------*)
    (* Tell user done                                                        *)
    (*-----------------------------------------------------------------------*)

    window_write(active_tcb^.port_chan_s + '>:', 'Reload is done');

  END;

(*===========================================================================*)
(* Subroutine to free msgs                                                   *)
(*===========================================================================*)

PROCEDURE free_msg_list;

  VAR
    d_size               : WORD;
    m_i_dis              : msg_d_ptr;
    m_i_dr               : msg_dr_ptr;
    m_overhead           : WORD;
    m_size               : WORD;
    msg_index_current    : msg_index_ptr;
    msg_index_next       : msg_index_ptr;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Get ready to start                                                    *)
    (*-----------------------------------------------------------------------*)

    msg_index_next := msg_index_start;

    m_overhead := OFS(msg_index_start^.msg_i_mb.msg_subj[1])
                                                       - OFS(msg_index_start^);

    (*-----------------------------------------------------------------------*)
    (* Loop thru all messages                                                *)
    (*-----------------------------------------------------------------------*)

    WHILE msg_index_next <> NIL DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Chain to next message                                             *)
        (*-------------------------------------------------------------------*)

        msg_index_current := msg_index_next;

        {$IFDEF POINT_CHK}
          test_pointer(msg_index_current);
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Look at message                                                   *)
        (*-------------------------------------------------------------------*)

        WITH msg_index_current^ DO
          BEGIN;

            (*---------------------------------------------------------------*)
            (* Distribution list present?                                    *)
            (*---------------------------------------------------------------*)

            IF (msg_i_mb.msg_flag AND mf_fwd_list) <> 0 THEN
              BEGIN;

                (*-----------------------------------------------------------*)
                (* Handle distribution list                                  *)
                (*-----------------------------------------------------------*)

                (*-----------------------------------------------------------*)
                (* Check for distribution route list                         *)
                (*-----------------------------------------------------------*)

                IF ((msg_i_mb.msg_flag AND mf_disrout) <> 0)
                                                    AND (msg_i_dr <> NIL) THEN
                  BEGIN;

                    (*-------------------------------------------------------*)
                    (* Handle route list                                     *)
                    (*-------------------------------------------------------*)

                    {$IFDEF POINT_CHK}
                      test_pointer(msg_i_dr);
                    {$ENDIF}

                    (*-------------------------------------------------------*)
                    (* Get pointers to route list and to distribution list   *)
                    (*-------------------------------------------------------*)

                    m_i_dr  := msg_i_dr;

                    m_i_dis := m_i_dr^.msg_dr_dblk;

                    {$IFDEF POINT_CHK}
                      test_pointer(m_i_dis);
                    {$ENDIF}

                    {$IFDEF POINT_CHK}
                      test_pointer(m_i_dr);
                    {$ENDIF}

                    (*-------------------------------------------------------*)
                    (* Verify distribution number                            *)
                    (*-------------------------------------------------------*)

                    IF m_i_dis^.msg_d_no > msg_dist_max THEN
                      BEGIN;
                        WRITELN;
                        WRITELN(stars);
                        WRITELN('Invalid distribution # -- ',
                                                          m_i_dis^.msg_d_no);
                        WRITELN(stars);
                        WRITELN;
                        dump_trace;
                        dump_msg(msg_index_current);
                        RUNERROR(msg_runerr);
                        HALT;
                      END;

                    (*-------------------------------------------------------*)
                    (* Get size of route list and free it                    *)
                    (*-------------------------------------------------------*)

                    d_size := SIZEOF(msg_d_ptr)
                               + m_i_dis^.msg_d_no * SIZEOF(msg_dr_route_item);

                    FREEMEM(msg_i_dr, d_size);

                    {$IFDEF  DEBUG_FREE}
                      trace_data('FREE DR', d_size, m_i_dr, '');
                    {$ENDIF}

                    {$IFDEF FREE_CHECK}
                      test_free_list;
                    {$ENDIF}

                  END
                ELSE

                  (*---------------------------------------------------------*)
                  (* If no route list, just get dis list                     *)
                  (*---------------------------------------------------------*)

                  m_i_dis := msg_i_dis;

                (*-----------------------------------------------------------*)
                (* Is the distribution list in memory?                       *)
                (*-----------------------------------------------------------*)

                IF m_i_dis <> NIL THEN
                  BEGIN;

                    (*-------------------------------------------------------*)
                    (* Distribution list present                             *)
                    (*-------------------------------------------------------*)

                    {$IFDEF POINT_CHK}
                      test_pointer(m_i_dis);
                    {$ENDIF}

                    (*-------------------------------------------------------*)
                    (* Verify distribution number                            *)
                    (*-------------------------------------------------------*)

                    IF m_i_dis^.msg_d_no > msg_dist_max THEN
                      BEGIN;
                        WRITELN;
                        WRITELN(stars);
                        WRITELN('Invalid distribution # -- ',
                                                          m_i_dis^.msg_d_no);
                        WRITELN(stars);
                        WRITELN;
                        dump_trace;
                        dump_msg(msg_index_current);
                        RUNERROR(msg_runerr);
                      END;

                    (*-------------------------------------------------------*)
                    (* Calculate space to free and do it                     *)
                    (*-------------------------------------------------------*)

                    d_size := 1
                             + m_i_dis^.msg_d_no * SIZEOF(msg_dist_entry_type);

                    {$IFDEF DEBUG_FREE}
                      trace_data('FREE DIS', d_size, m_i_dis, '');
                    {$ENDIF}

                    FREEMEM(m_i_dis, d_size);

                    {$IFDEF FREE_CHECK}
                      test_free_list;
                    {$ENDIF}

                  END; (*----- End handling of in core distribution list ----*)

              END; (*----- End handling of distribution list ----------------*)

            (*---------------------------------------------------------------*)
            (* Chain to next message                                         *)
            (*---------------------------------------------------------------*)

            msg_index_next := msg_i_next;

          END;

        (*-------------------------------------------------------------------*)
        (* Calculate size of message and free it                             *)
        (*-------------------------------------------------------------------*)

        m_size := m_overhead + LENGTH(msg_index_current^.msg_i_mb.msg_subj);

        {$IFDEF DEBUG_FREE}
          trace_data('FREE MSG', SIZEOF(msg_index_current^),
                                 msg_index_current, '');
        {$ENDIF}

        FREEMEM(msg_index_current, m_size);

        {$IFDEF FREE_CHECK}
          test_free_list;
        {$ENDIF}

      END; (*----- End of loop thru all messages ----------------------------*)

    (*-----------------------------------------------------------------------*)
    (* Clear master pointer                                                  *)
    (*-----------------------------------------------------------------------*)

    msg_index_start := NIL;

  END;

END.
