(*===========================================================================*)
(* Miscellaneous things - 3 - Non overlay                                    *)
(*                                                                           *)
(*   Copyright 1990, 1991 by H. Roy Engehausen.  All rights reserved.        *)
(*                                                                           *)
(*===========================================================================*)

UNIT BBMISC3;

INTERFACE

  FUNCTION  called_by (depth : BYTE) : POINTER;
  PROCEDURE test_pointer(p : POINTER);
  PROCEDURE test_free_list;

  VAR
    pointer_bottom_l : LONGINT;
    pointer_top_l    : LONGINT;

    pointer_bottom_p : POINTER;
    pointer_top_p    : POINTER;

    heaporg_l        : LONGINT;

IMPLEMENTATION

USES
  DOS,
  bbbug,
  bbdummy,
  bbdump,
  bbrunerr;

(*===========================================================================*)
(* Routine that returns who called me                                        *)
(*===========================================================================*)

FUNCTION called_by (depth : BYTE) : POINTER;

  TYPE stack_info = RECORD
                      next_bp   : WORD;
                      link_off  : WORD;
                      link_seg  : WORD;
                    END;

  VAR
    var_bp            : WORD;
    stack_ptr     : ^stack_info;
    displaced_seg : WORD;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Get the current setting of BP                                         *)
    (*-----------------------------------------------------------------------*)

    ASM;
      MOV var_bp,BP
    END;

    (*-----------------------------------------------------------------------*)
    (* Set the depth to account for ourselves                                *)
    (*-----------------------------------------------------------------------*)

    INC(depth);

    (*-----------------------------------------------------------------------*)
    (* Loop back thru the stack                                              *)
    (*-----------------------------------------------------------------------*)

    REPEAT

      stack_ptr := PTR(SSEG, var_bp);
      var_bp    := stack_ptr^.next_bp;
      DEC(depth);

    UNTIL (depth < 1) OR (var_bp = 0); (*----- End loop thru stack --------------*)

    (*-----------------------------------------------------------------------*)
    (* Get segment we were called from                                       *)
    (*-----------------------------------------------------------------------*)

    displaced_seg := stack_ptr^.link_seg;

    (*-----------------------------------------------------------------------*)
    (* Compute difference from first entry CSEG                              *)
    (*-----------------------------------------------------------------------*)

    IF displaced_seg >= main_cs THEN
      displaced_seg := displaced_seg - main_cs
    ELSE
      displaced_seg := NOT(main_cs - displaced_seg);

    (*-----------------------------------------------------------------------*)
    (* Set the result                                                        *)
    (*-----------------------------------------------------------------------*)

    called_by := PTR(displaced_seg, stack_ptr^.link_off);

    (*-----------------------------------------------------------------------*)
    (* All done                                                              *)
    (*-----------------------------------------------------------------------*)

  END;

(*===========================================================================*)
(* Make sure a pointer is within the bounds                                  *)
(*===========================================================================*)

PROCEDURE test_pointer(p : POINTER);

  TYPE

    PFreeRec = ^TFreeRec;
    TFreeRec = RECORD
                 Next : PFreeRec;
                 Size : POINTER;
               END;

    p_type   = RECORD
                 CASE BYTE OF
                   0: (pf  : PFreeRec);
                   1: (off : WORD;
                        sg : WORD);
                 END;

  VAR
    i  : BYTE;
    l  : LONGINT;
    m  : LONGINT;
    p1 : p_type;
    p2 : p_type;

  PROCEDURE crash;
    BEGIN;

      dump_state;
      dump_trace;
      dump_all_thread;
      dump_free_list;
      dump_window_all;
      HALT;

    END;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Get the address in memory                                             *)
    (*-----------------------------------------------------------------------*)

    l := WORD(p) + LONGINT(16) * SEG(p^);

    (*-----------------------------------------------------------------------*)
    (* Within the program?     If not, blow up                               *)
    (*-----------------------------------------------------------------------*)

    IF (l < pointer_bottom_l) OR (l > pointer_top_l) THEN
      BEGIN;

        dump_reason('Pointer range failure -- ' + p2x(p)
                + ' <> ' + p2x(pointer_bottom_p) + '/' + p2x(pointer_top_p));

        crash;

      END;

    (*-----------------------------------------------------------------------*)
    (* Get ready to chain down the list                                      *)
    (*-----------------------------------------------------------------------*)

    p1.pf := FreeList;

    (*-----------------------------------------------------------------------*)
    (* Make sure its below the heap pointer                                  *)
    (*-----------------------------------------------------------------------*)

    m := WORD(HeapPtr) + LONGINT(16) * SEG(HeapPtr^);
    IF l > m THEN
      BEGIN;
        dump_reason('Pointer in free heap -- ' + p2x(p)
                                                       + ' > ' + p2x(HeapPtr));
        crash;
      END;

    (*-----------------------------------------------------------------------*)
    (* Loop down the chain                                                   *)
    (*-----------------------------------------------------------------------*)

    WHILE p1.pf <> HeapPtr DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Get the start of the block we are looking at                      *)
        (*-------------------------------------------------------------------*)

        m := WORD(p1.pf) + LONGINT(16) * SEG(p1.pf^);

        (*-------------------------------------------------------------------*)
        (* Address we want is below so search is done                        *)
        (*-------------------------------------------------------------------*)

        IF l < m THEN
          EXIT;

        (*-------------------------------------------------------------------*)
        (* Calculate the end of the free area                                *)
        (*-------------------------------------------------------------------*)

        p2.pf := p1.pf^.SIZE;
        m := m - 1 + WORD(p2.pf) + LONGINT(16) * SEG(p2.pf^);

        (*-------------------------------------------------------------------*)
        (* Address we want is below so things are bad                        *)
        (*-------------------------------------------------------------------*)

        IF l < m THEN
          BEGIN;
            dump_reason('Pointer in free area -- ' + p2x(p) + ' <> '
                                              + p2x(p1.pf) + '/' + p2x(p2.pf));
            crash;
          END;

        (*-------------------------------------------------------------------*)
        (* Get start of next free area                                       *)
        (*-------------------------------------------------------------------*)

        p1.pf := p1.pf^.Next;

      END;

  END;

(*===========================================================================*)
(* Test the free list for proper chaining                                    *)
(*===========================================================================*)

PROCEDURE test_free_list;

  TYPE

    PFreeRec = ^TFreeRec;
    TFreeRec = RECORD
                 Next : PFreeRec;
                 Size : POINTER;
               END;

    p = RECORD
          CASE BYTE OF
            0: (pf : PFreeRec);
            1: (off : WORD;
                sg  : WORD);
          END;

  VAR
    p1    : p;
    p2    : p;
    p3    : p;
    hi_l  : LONGINT;
    l     : LONGINT;
    o     : WORD;
    s     : WORD;

  LABEL iterate;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Get high water mark to check                                          *)
    (*-----------------------------------------------------------------------*)

    hi_l  := WORD(HeapPtr) + LONGINT(16) * SEG(HeapPtr^);

    (*-----------------------------------------------------------------------*)
    (* Verify the water mark are ok                                          *)
    (*-----------------------------------------------------------------------*)

    IF hi_l < heaporg_l THEN
      BEGIN;
        dump_reason('HeapPtr failure LOW -- ' + p2x(HeapOrg)
                                                         + '/' + p2x(HeapPtr));
        dump_all;
        HALT;
      END;

    IF hi_l > pointer_top_l THEN
      BEGIN;
        dump_reason('HeapPtr failure HI -- ' + p2x(HeapOrg)
                                                   + '/' + p2x(pointer_top_p));
        dump_all;
        HALT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Get ready to chain down the list                                      *)
    (*-----------------------------------------------------------------------*)

    p1.pf := FreeList;

    (*-----------------------------------------------------------------------*)
    (* Loop down the chain                                                   *)
    (*-----------------------------------------------------------------------*)

iterate:

    WHILE p1.pf <> HeapPtr DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Make sure we are withing the ball park                            *)
        (*-------------------------------------------------------------------*)

        l := WORD(p1.pf) + LONGINT(16) * SEG(p1.pf^);

        IF (l < heaporg_l) OR (l > hi_l) THEN
          BEGIN;

            dump_reason('FreeList Pointer range failure -- ' + p2x(p1.pf)
                         + ' <> ' + p2x(HeapOrg) + '/' + p2x(HeapPtr));

            dump_all;

            HALT;

          END;

        (*-------------------------------------------------------------------*)
        (* Calculate the end of the free area                                *)
        (*-------------------------------------------------------------------*)

        p3    := p1;
        p2.pf := p1.pf^.SIZE;
        o     := p2.off + p1.off;
        s     := p2.sg + p1.sg;
        WHILE o >= 16 DO
          BEGIN;
            DEC(o,16);
            INC(s);
          END;

        (*-------------------------------------------------------------------*)
        (* Get start of next free area                                       *)
        (*-------------------------------------------------------------------*)

        p1.pf := p1.pf^.Next;

        (*-------------------------------------------------------------------*)
        (* Verify there is no overlap                                        *)
        (*-------------------------------------------------------------------*)

        IF s < p1.sg THEN GOTO ITERATE;
        IF (s = p1.sg) AND (o < p1.off) THEN GOTO ITERATE;

        (*-------------------------------------------------------------------*)
        (* Overlap...  Blow up                                               *)
        (*-------------------------------------------------------------------*)

        dump_reason('Memory allocation error at ' + p2x(p3.pf));
        dump_free_list;
        dump_trace;
        RUNERROR(free_list_error);

      END;

  END;

END.
