(*===========================================================================*)
(* Miscellaneous things - 6                                                  *)
(*                                                                           *)
(*   Copyright 1991 by H. Roy Engehausen.  All rights reserved.              *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

UNIT BBMISC6;

INTERFACE

  USES
    bbax25,
    bbdummy;

  PROCEDURE build_ax25_header(    in_s      : STRING;
                                  p_addr    : address_area_ptr_type;
                              VAR out_size  : BYTE);

  FUNCTION  find_forward_port : str8;

  PROCEDURE sked_forward_port(in_str : str8);

IMPLEMENTATION

USES
  bbfwdd,
  bbstr,
  bbtime;

(*=========================================================================*)
(* This routine is change a call sign to internal format.                  *)
(*=========================================================================*)

PROCEDURE call_change_to_packet(VAR out_call : address_field;
                                    in_call  : bb_addr_str);

  VAR
    char_temp   : CHAR;
    i           : INTEGER;
    j           : INTEGER;
    out_temp    : address_field;
    ssid_temp   : BYTE;
    ssid_switch : BOOLEAN;

  LABEL iterate;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Blank it out to start with                                            *)
    (*-----------------------------------------------------------------------*)

    out_temp.name[1] := CHR(ORD('?') * 2);

    FILLCHAR(out_temp.name[2], 5, CHR(ORD(' ') * 2));

    (*-----------------------------------------------------------------------*)
    (* Set the reserved bits                                                 *)
    (*-----------------------------------------------------------------------*)

    out_temp.addr_flag := flag_reser;

    (*-----------------------------------------------------------------------*)
    (* Initialize the output area                                            *)
    (*-----------------------------------------------------------------------*)

    out_call := out_temp;

    (*-----------------------------------------------------------------------*)
    (* Handle dumb cases                                                     *)
    (*-----------------------------------------------------------------------*)

    strip_var(in_call, 'B');
    upcase_str_var(in_call);

    (*-----------------------------------------------------------------------*)
    (* Initialize for loop                                                   *)
    (*-----------------------------------------------------------------------*)

    i           := 0;
    ssid_switch := FALSE;
    ssid_temp   := 0;

    (*-----------------------------------------------------------------------*)
    (* Loop thru the call                                                    *)
    (*-----------------------------------------------------------------------*)

    WHILE i < LENGTH(in_call) DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Move pointer to next character                                    *)
        (*-------------------------------------------------------------------*)

        INC(i);

        (*-------------------------------------------------------------------*)
        (* Get character                                                     *)
        (*-------------------------------------------------------------------*)

        char_temp := in_call[i];

        (*-------------------------------------------------------------------*)
        (* Check for start of SSID                                           *)
        (*-------------------------------------------------------------------*)

        IF char_temp = '-' THEN
          BEGIN;
            ssid_switch := TRUE;
            GOTO iterate;
          END;

        (*-------------------------------------------------------------------*)
        (* Build SSID as we go                                               *)
        (*-------------------------------------------------------------------*)

        IF ssid_switch THEN
          BEGIN;
            IF (char_temp >= '0') AND (char_temp <= '9') THEN
              ssid_temp := 10 * ssid_temp + (ORD(char_temp) - ORD('0'));
            GOTO iterate;
          END;

        (*-------------------------------------------------------------------*)
        (* Put character in its place with the shift                         *)
        (*-------------------------------------------------------------------*)

        IF i <= 6 THEN
          out_temp.name[i] := CHR(ORD(char_temp) * 2);

        (*-------------------------------------------------------------------*)
        (* Come here to process next character                               *)
        (*-------------------------------------------------------------------*)

iterate:

      END; (*----- End loop thru name ---------------------------------------*)


    (*-----------------------------------------------------------------------*)
    (* Set the SSID                                                          *)
    (*-----------------------------------------------------------------------*)

    IF ssid_temp <= ssid_max THEN
      out_temp.addr_flag := ((ssid_temp SHL flag_ssid_shift) AND flag_ssid);

    (*-----------------------------------------------------------------------*)
    (* Move data to output area                                              *)
    (*-----------------------------------------------------------------------*)

    out_call := out_temp;

  END;

(*===========================================================================*)
(* Build an AX25 packet header from a string.                                *)
(*===========================================================================*)

PROCEDURE build_ax25_header(    in_s      : STRING;
                                p_addr    : address_area_ptr_type;
                            VAR out_size  : BYTE);

  VAR
    i          : WORD;
    j          : WORD;
    w          : bb_addr_str;
    word_count : BYTE;
    p_packet   : ^packet_data;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Get the source call sign.  Assume one if necessary                    *)
    (*-----------------------------------------------------------------------*)

    w := active_port^.call_set;
    IF w = '' THEN
      w := opt_block.this_bb_addr;

    (*-----------------------------------------------------------------------*)
    (* Set the source into the address area                                  *)
    (*-----------------------------------------------------------------------*)

    call_change_to_packet(p_addr^.source, w);

    (*-----------------------------------------------------------------------*)
    (* Prep the input string                                                 *)
    (*-----------------------------------------------------------------------*)

    strip_var(in_s, 'B');
    upcase_str_var(in_s);

    (*-----------------------------------------------------------------------*)
    (* Handle default destination if none given                              *)
    (*-----------------------------------------------------------------------*)

    IF in_s = '' THEN
      in_s := 'MAIL';

    (*-----------------------------------------------------------------------*)
    (* Get the first word as the destination                                 *)
    (*-----------------------------------------------------------------------*)

    w := subword(@in_s, 1, 1);
    call_change_to_packet(p_addr^.dest, w);

    (*-----------------------------------------------------------------------*)
    (* Get read to loop                                                      *)
    (*-----------------------------------------------------------------------*)

    word_count := words(in_s);

    j := 0;

    (*-----------------------------------------------------------------------*)
    (* Loop thru all the words in the address                                *)
    (*-----------------------------------------------------------------------*)

    FOR i := 2 TO word_count DO
      BEGIN;
        w := subword(@in_s, i, 1);
        IF (w <> 'V') AND (w <> 'VIA') THEN
          BEGIN;
            INC(j);
            call_change_to_packet(p_addr^.repeaters[j], w);
          END;
      END;

    (*-----------------------------------------------------------------------*)
    (* Set the end of the address sequence                                   *)
    (*-----------------------------------------------------------------------*)

    IF j > 0 THEN
      p_addr^.repeaters[j].addr_flag :=
                                      p_addr^.repeaters[j].addr_flag OR flag_e
    ELSE
      p_addr^.source.addr_flag := p_addr^.source.addr_flag OR flag_e;

    (*-----------------------------------------------------------------------*)
    (* Add the rest of the protocol                                          *)
    (*-----------------------------------------------------------------------*)

    p_addr^.dest.addr_flag := p_addr^.dest.addr_flag OR flag_c;

    p_packet := @p_addr^.repeaters[j+1];

    p_packet^.control_packet := control_ui;
    p_packet^.pid_packet     := protocol_no3;

    (*-----------------------------------------------------------------------*)
    (* Compute length                                                        *)
    (*-----------------------------------------------------------------------*)

    i := OFS(p_addr^);
    p_addr := @p_packet^.pid_packet;
    j := OFS(p_addr^);
    out_size := j - i + 1;

  END;

(*===========================================================================*)
(* Find ports to forward                                                     *)
(*===========================================================================*)

FUNCTION find_forward_port : str8;

  VAR
    work_port : port_block_ptr;
    work_str  : str8;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Get ready for loop                                                    *)
    (*-----------------------------------------------------------------------*)

    work_str := '';

    work_port := ring_port;

    (*-----------------------------------------------------------------------*)
    (* Now loop                                                              *)
    (*-----------------------------------------------------------------------*)

    REPEAT

      WITH work_port^ DO
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* Is it time for this port?  If so, save the character            *)
          (*-----------------------------------------------------------------*)

          IF sked_fwd < current_day_time THEN
            work_str := work_str + port_char;

          (*-----------------------------------------------------------------*)
          (* Chain to next port                                              *)
          (*-----------------------------------------------------------------*)

          work_port := next_port;

        END;

      (*---------------------------------------------------------------------*)
      (* Loop until all ports are checked or we are out of room              *)
      (*---------------------------------------------------------------------*)

    UNTIL (work_port = ring_port) OR (LENGTH(work_str) >= max_ports_to_sked);

    (*-----------------------------------------------------------------------*)
    (* Build the result and we are done                                      *)
    (*-----------------------------------------------------------------------*)

    CASE LENGTH(work_str) OF

      0: find_forward_port := '';

      1: find_forward_port := work_str + '*';

      ELSE
        find_forward_port := '(' + work_str + ')*';

    END;

  END;

(*===========================================================================*)
(* Schedule port forwarding                                                  *)
(*===========================================================================*)

PROCEDURE sked_forward_port(in_str : str8);

  VAR
    work_port : port_block_ptr;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Get ready for loop                                                    *)
    (*-----------------------------------------------------------------------*)

    work_port := ring_port;

    (*-----------------------------------------------------------------------*)
    (* Now loop                                                              *)
    (*-----------------------------------------------------------------------*)

    REPEAT

      WITH work_port^ DO
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* Reschedule this port?                                           *)
          (*-----------------------------------------------------------------*)

          IF POS(port_char, in_str) <> 0 THEN
            BEGIN;

              (*-------------------------------------------------------------*)
              (* Schedule next forward for this port                         *)
              (*-------------------------------------------------------------*)

              time_check;
              sked_fwd := time_next_hour(fwd_min);
            END;

          (*-----------------------------------------------------------------*)
          (* Chain to next port                                              *)
          (*-----------------------------------------------------------------*)

          work_port := next_port;

        END;

      (*---------------------------------------------------------------------*)
      (* Loop until all ports are checked                                    *)
      (*---------------------------------------------------------------------*)

    UNTIL (work_port = ring_port);

  END;

END.
