(*===========================================================================*)
(* Multiple message number handler                                           *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1991 by H. Roy Engehausen.  All rights reserved.  *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

UNIT BBMMSGN;

INTERFACE

USES
  bbdummy;

PROCEDURE check_multiple_msg(in_str        : str_ptr;
                             word_to_start : BYTE;
                             word_cnt      : BYTE);

FUNCTION  get_next_multiple_msg : WORD;

FUNCTION  in_multiple_msg_range : BOOLEAN;

IMPLEMENTATION

USES
  bbmem,
  bbmdata,
  bbmess,
  bbmisc,
  bbstr;

TYPE
  mmsg_data_type = RECORD
                     this_number      : WORD;
                     end_number       : WORD;
                     this_string      : str_ptr;
                     word_to_do       : STRING[6];
                     range_switch     : BOOLEAN;
                     next_word_number : BYTE;
                     word_count       : BYTE;
                   END;

(*===========================================================================*)
(* Check multiple message string                                             *)
(*===========================================================================*)

PROCEDURE check_multiple_msg(in_str        : str_ptr;
                             word_to_start : BYTE;
                             word_cnt      : BYTE);

  VAR
    b                    : BOOLEAN;
    code                 : INTEGER;
    i                    : BYTE;
    last_num             : WORD;
    mmsg_data            : ^mmsg_data_type;
    range_word           : BOOLEAN;
    this_num             : LONGINT;
    word_number_to_check : BYTE;
    word_to_check        : STRING[15];

  BEGIN;

    range_word := FALSE;
    last_num   := 0;

    (*-----------------------------------------------------------------------*)
    (* Loop thru all words                                                   *)
    (*-----------------------------------------------------------------------*)

    FOR word_number_to_check := word_to_start TO word_cnt DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Get the word to check                                             *)
        (*-------------------------------------------------------------------*)

        word_to_check := subwordl(in_str^, word_number_to_check,
                                                       SIZEOF(word_to_check)-1);

        (*-------------------------------------------------------------------*)
        (* Make sure it isn't too long                                       *)
        (*-------------------------------------------------------------------*)

        IF LENGTH(word_to_check) > 5 THEN
          BEGIN;
            set_dollar1_parm(@word_to_check);
            send_message(message_err_ivm_wp);
            active_tcb^.error_sw := TRUE;
            EXIT;
          END;

        (*-------------------------------------------------------------------*)
        (* Handle special range word                                         *)
        (*-------------------------------------------------------------------*)

        IF word_to_check = 'TO' THEN
          BEGIN;

            (*---------------------------------------------------------------*)
            (* Prevent two range words in a row or a range word at           *)
            (* either beginning or the end                                   *)
            (*---------------------------------------------------------------*)

            IF range_word
                           OR (last_num <> 0)
                           OR (word_number_to_check = 1)
                           OR (word_number_to_check = word_cnt) THEN
              BEGIN;
                set_dollar1_parm(@word_to_check);
                send_message(message_err_ivm_wp);
                active_tcb^.error_sw := TRUE;
                EXIT;
              END;

            range_word := TRUE;
            last_num   := this_num;

          END

        ELSE

          BEGIN;

            (*---------------------------------------------------------------*)
            (* It must be a number so evaluate it                            *)
            (*---------------------------------------------------------------*)

            VAL(word_to_check, this_num, code);

            (*---------------------------------------------------------------*)
            (* For a range, the number must be above the range start.        *)
            (*---------------------------------------------------------------*)

            b := range_word AND (this_num <= last_num);

            (*---------------------------------------------------------------*)
            (* Give error message if needed                                  *)
            (*---------------------------------------------------------------*)

            IF (code <> 0)
                           OR (this_num > 65535)
                           OR (this_num < 1)
                           OR b THEN
              BEGIN;

                IF b THEN
                  FOR i := (word_number_to_check - 1)
                                          DOWNTO (word_number_to_check - 2) DO
                    word_to_check := subword(in_str, i, 1)
                                                        + ' ' + word_to_check;

                set_dollar1_parm(@word_to_check);
                send_message(message_err_ivm_wp);

                active_tcb^.error_sw := TRUE;
                EXIT;

              END;

            (*---------------------------------------------------------------*)
            (* Turn off range switch                                         *)
            (*---------------------------------------------------------------*)

            IF NOT range_word THEN
              last_num := 0;

            range_word := FALSE;

          END;

      END; (*----- End loop thru all words ----------------------------------*)

    (*-----------------------------------------------------------------------*)
    (* Record the data for the next guy...                                   *)
    (*-----------------------------------------------------------------------*)

    mmsg_data := get_task_mem('MMS', SIZEOF(mmsg_data_type));

    WITH mmsg_data^ DO
      BEGIN;
        this_string      := in_str;
        next_word_number := word_to_start;
        word_count       := word_cnt;
        range_switch     := FALSE;
        word_to_do       := subword(in_str, next_word_number, 1);
      END;

  END;

(*===========================================================================*)
(* Get next message number of a multiple string                              *)
(*===========================================================================*)

FUNCTION  get_next_multiple_msg: WORD;

  VAR
    code       : INTEGER;
    mmsg_data  : ^mmsg_data_type;
    val_n      : LONGINT;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Retrieve the pointer and do a sanity check                            *)
    (*-----------------------------------------------------------------------*)

    mmsg_data := find_task_mem_addr('MMS');

    IF mmsg_data = NIL THEN
      BEGIN;
        WRITELN('Internal error in multiple message');
        HALT;
      END;

    WITH mmsg_data^ DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Are we in a range or not                                          *)
        (*-------------------------------------------------------------------*)

        IF range_switch THEN
          BEGIN;

            (*---------------------------------------------------------------*)
            (* Inside a range, see if we can give the next number.  If so,   *)
            (* do it!                                                        *)
            (*---------------------------------------------------------------*)

            INC(this_number);
            IF this_number <= end_number THEN
              BEGIN;
                get_next_multiple_msg := this_number;
                EXIT;
              END;

            range_switch := FALSE;

          END;

        (*-------------------------------------------------------------------*)
        (* See if we have run out of words.  If so we are done               *)
        (*-------------------------------------------------------------------*)

        IF next_word_number > word_count THEN
          BEGIN;
            get_next_multiple_msg := 0;
            free_task_mem('MMS', TRUE);
            EXIT;
          END;

        (*-------------------------------------------------------------------*)
        (* Convert the number to what we need                                *)
        (*-------------------------------------------------------------------*)

        VAL(word_to_do, val_n, code);
        this_number := val_n;

        (*-------------------------------------------------------------------*)
        (* Set output                                                        *)
        (*-------------------------------------------------------------------*)

        get_next_multiple_msg := this_number;

        (*-------------------------------------------------------------------*)
        (* Bump word number to next                                          *)
        (*-------------------------------------------------------------------*)

        INC(next_word_number);
        word_to_do := subword(this_string, next_word_number, 1);

        (*-------------------------------------------------------------------*)
        (* If not range start then we are done                               *)
        (*-------------------------------------------------------------------*)

        IF word_to_do <> 'TO' THEN
          EXIT;

        (*-------------------------------------------------------------------*)
        (* Range start.  Get end of range and convert.                       *)
        (*-------------------------------------------------------------------*)

        range_switch := TRUE;

        word_to_do := subword(this_string, next_word_number + 1, 1);
        VAL(word_to_do, val_n, code);
        end_number := val_n;

        (*-------------------------------------------------------------------*)
        (* Correct word counters, etc                                        *)
        (*-------------------------------------------------------------------*)

        next_word_number := next_word_number + 2;
        word_to_do := subword(this_string, next_word_number, 1);

      END;

  END;

FUNCTION  in_multiple_msg_range : BOOLEAN;

  VAR
    mmsg_data  : ^mmsg_data_type;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Get info                                                              *)
    (*-----------------------------------------------------------------------*)

    mmsg_data := find_task_mem_addr('MMS');

    (*-----------------------------------------------------------------------*)
    (* Set result                                                            *)
    (*-----------------------------------------------------------------------*)

    IF mmsg_data = NIL THEN
      in_multiple_msg_range := FALSE
    ELSE
      in_multiple_msg_range := mmsg_data^.range_switch;

  END;

END.
