(*===========================================================================*)
(* Operator command processor                                                *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

UNIT BBOCMD;

INTERFACE

USES
  bbdummy;

FUNCTION op_cmd(operator_line_in : str_ptr) : BOOLEAN;

PROCEDURE execute_opr_command(operator_line_in : str_ptr);

IMPLEMENTATION

  USES
    CRT,
    bbact,
    bbbidaux,
    bbcaptur,
    bbcmf,
    bbconsl,
    bbconv,
    bbcopy,
    bbedit,
    bbexdos,
    bbexport,
    bbfsf,
    bbfsz,
    bbfwdr,
    bbhlook,
    bbimport,
    bbmark,
    bbmdata,
    bbmess,
    bbmfi,
    bbminit,
    bbmisc,
    bbmisci,
    bbmisc2,
    bbmisc4,
    bboprd,
    bboprm,
    bbopro,
    bbopru,
    bbover,
    bbrmc,
    bbstr,
    bbtcmd,
    bbucmd,
    bbwakeup,
    bbwin,
    bbwp;

(*===========================================================================*)
(* Execute command                                                           *)
(*===========================================================================*)

PROCEDURE execute_opr_command(operator_line_in : str_ptr);

  VAR
    i : WORD;

  BEGIN;

    strip_crlf(operator_line_in^);

    active_tcb^.error_sw := FALSE;

    op_sw := op_cmd(operator_line_in);

    IF op_sw THEN EXIT;

    user_command(operator_line_in^);

  END;

(*===========================================================================*)
(* Execute an operator command                                               *)
(*===========================================================================*)

FUNCTION op_cmd(operator_line_in : str_ptr) : BOOLEAN;

  TYPE
    b_ptr = ^BOOLEAN;
    p_ptr = PROCEDURE(str_parm : STRING);

  VAR
    i         : WORD;
    op_word   : STRING[20];
    op_word_l : BYTE;

  (*=========================================================================*)
  (* Set a switch                                                            *)
  (*=========================================================================*)

  PROCEDURE set_switch(this_switch : b_ptr);

    BEGIN;

      op_word := subword(operator_line_in, 2, 1);

      IF LENGTH(op_word) = 0 THEN
        this_switch^ := NOT this_switch^
      ELSE
        this_switch^ := (ORD(op_word[1]) AND 1) = 1;

      switch_show(this_switch^);

    END;

  (*=========================================================================*)
  (* Main line of operator command                                           *)
  (*=========================================================================*)

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Set the switch to assume we found a command                           *)
    (*-----------------------------------------------------------------------*)

    op_cmd   := TRUE;

    (*-----------------------------------------------------------------------*)
    (* Extract the word and process it                                       *)
    (*-----------------------------------------------------------------------*)

    op_word := subword(operator_line_in, 1, 1);

    op_word_l := LENGTH(op_word);

    upcase_str_var(op_word);

    (*-----------------------------------------------------------------------*)
    (* If there is no command then leave.  This will give a false positive   *)
    (* test for a blank line but we want to ignore them anyway               *)
    (*-----------------------------------------------------------------------*)

    IF op_word_l = 0 THEN
      EXIT;

    (*-----------------------------------------------------------------------*)
    (* This catches a "B" command issued in any mode by local operator       *)
    (*-----------------------------------------------------------------------*)

    IF (op_word = 'B')
           AND ((active_tcb^.tcb_type = th_operator)
                              OR (active_tcb^.tcb_type = th_fwd_control)) THEN
      BEGIN;
        close_operator_session;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* This section is local operator commands only                          *)
    (*-----------------------------------------------------------------------*)

    IF (active_tcb^.uid_data.user_class = user_c_lsu)
                                OR (active_tcb^.tcb_type = th_fwd_control) THEN
      BEGIN;

        IF op_word = 'EF' THEN
          BEGIN;
            edit_file_main(operator_line_in^);
            EXIT;
          END;

        IF op_word = 'ET' THEN
          BEGIN;
            oper_et(operator_line_in^);
            EXIT;
          END;

        IF op_word = 'GB' THEN
          BEGIN;
            set_switch(@opt_block.opt_sysop_bell);
            EXIT;
          END;

        IF op_word = 'GS' THEN
          BEGIN;
            set_switch(@opt_block.opt_sysop_is_in);
            EXIT;
          END;

        IF op_word = 'GP' THEN
          BEGIN;

            set_switch(@opt_block.opt_suppress_pfx);

            IF opt_block.opt_suppress_pfx THEN
              active_tcb^.max_pac := special_max_pac_wop
            ELSE
              active_tcb^.max_pac := special_max_pac_wp;

            EXIT;

          END;

        IF op_word = 'GZ' THEN
          BEGIN;
            set_switch(@opt_block.opt_show_mismatch);
            EXIT;
          END;

        IF op_word = 'QUIT' THEN HALT;

        IF op_word = 'T' THEN
          BEGIN;
            oper_t(operator_line_in^);
            IF NOT active_tcb^.error_sw THEN
              operator_talk_action;

            EXIT;
          END;

      END;

    (*-----------------------------------------------------------------------*)
    (* This section is various operator change status commands               *)
    (*-----------------------------------------------------------------------*)

    IF FIND(@change_status_list, @op_word) > 0 THEN
      BEGIN;
        change_status_cmd(operator_line_in^);
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* This section is operator commands with more than 2 characters         *)
    (*-----------------------------------------------------------------------*)

    IF (op_word = 'OPEN') OR (op_word = 'CLOSE') THEN
      BEGIN;
        oper_ep(operator_line_in^);
        EXIT;
      END;

    IF (op_word = 'EXPORT') OR (op_word = 'EXPORTK') THEN
      BEGIN;
        export_cmd(operator_line_in^, NIL);
        EXIT;
      END;

    IF (op_word = 'FORCE') OR (op_word = 'KILL') THEN
      BEGIN;
        oper_force(operator_line_in^);
        EXIT;
      END;

    IF (op_word = 'IMPORT') OR (op_word = 'IMPORTE') THEN
      BEGIN;
        import_cmd(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'SHUT' THEN
      BEGIN;

        IF WORDS(operator_line_in^) > 1 THEN
          BEGIN;
            send_message(message_err_wrd);
            active_tcb^.error_sw := TRUE;
            EXIT;
          END;

        non_busy_shut(operator_line_in^);
        IF active_tcb^.error_sw THEN EXIT;

        op_word := 'SHUTNOW';
      END;

    IF op_word = 'SHUTNOW' THEN
      BEGIN;
        IF active_tcb^.tcb_type <> th_fwd_control THEN
          shutdown_bbs;
        wakeup_shut_sw := TRUE;
        EXIT;
      END;

    IF (op_word = 'X') OR (op_word = 'XSTOP') THEN
      BEGIN;
        fwd_command := operator_line_in^;
        upcase_str_var(fwd_command);
        send_message(message_op_complete);
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* This section is operator commands with either 1 or 2 characters       *)
    (*-----------------------------------------------------------------------*)

    IF op_word_l > 2 THEN
      BEGIN;
        op_cmd := FALSE;
        EXIT;
      END;

    IF op_word[1] = 'A' THEN
      BEGIN;
        action_cmd(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'C' THEN
      BEGIN;
        set_clock(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'CF' THEN
      BEGIN;
        oper_cf(operator_line_in);
        EXIT;
      END;

    IF op_word = 'CM' THEN
      BEGIN;
        oper_cm(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'DM' THEN
      BEGIN;
        oper_dm(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'DR' THEN
      BEGIN;
        disp_route(operator_line_in^);
        EXIT;
      END;

    (* DU Command is in UCMD *)

    IF op_word = 'EC' THEN
      BEGIN;
        oper_ec(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'EP' THEN
      BEGIN;
        oper_ep(operator_line_in^);
        EXIT;
      END;

    IF (op_word = 'EL') OR (op_word = 'ES') OR (op_word = 'EX') THEN
      BEGIN;
        exec_dos(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'EM' THEN
      BEGIN;
        oper_em(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'EU' THEN
      BEGIN;
        oper_eu(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'EW' THEN
      BEGIN;

        IF NOT wp_write THEN EXIT;

        operator_line_in^ := 'IMPORTES ' + 'WP.OUT';

        import_cmd(operator_line_in^);

        EXIT;

      END;

    IF op_word[1] = 'F' THEN
      BEGIN;
        make_file_cmd(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'GC' THEN
      BEGIN;
        capture_session(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'GE' THEN
      BEGIN;
        message_reload;
        EXIT;
      END;

    IF op_word = 'GF' THEN
      BEGIN;
        msg_route_force := TRUE;
        send_message(message_op_complete);
        EXIT;
      END;

    IF op_word = 'GH' THEN
      BEGIN;
        h_look_up(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'GI' THEN
      BEGIN;
        bid_clean;
        EXIT;
      END;

    IF op_word = 'GK' THEN
      BEGIN;
        wakeup_force := TRUE;
        EXIT;
      END;

    IF (op_word = 'GL') OR (op_word = 'GM')  OR (op_word = 'GR') THEN
      BEGIN;
        oper_gm(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'GN' THEN
      BEGIN;
        file_name_change(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'GO' THEN
      BEGIN;
        mark_msgs_old(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'GW' THEN
      BEGIN;
        capture_window(operator_line_in^);
        EXIT;
      END;

    IF op_word[1] = 'M' THEN
      BEGIN;
        oper_mm(operator_line_in^);
        EXIT;
      END;

    IF op_word = 'Z' THEN
      BEGIN;
        erase_file_cmd(operator_line_in^);
        EXIT;
      END;

    op_cmd := FALSE;

  END;

END.
