(*===========================================================================*)
(* Operator task -- Work with message file                                   *)
(*    Does the following commands:                                           *)
(*           CM -- copy message                                              *)
(*           EM -- edit message                                              *)
(*           M. -- make message from file                                    *)
(*           IMPORT -- make message from a file containing a message         *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

{$DEFINE POINT_CHK}
{$DEFINE FREE_CHK}

{$O+}

UNIT BBOPRM;

INTERFACE

PROCEDURE oper_em(cmd_in : STRING);
PROCEDURE oper_cm(cmd_in : STRING);
PROCEDURE oper_mm(cmd_in : STRING);

IMPLEMENTATION

  USES
    CRT,
    bbcmsg,
    bbcopy,
    bbdummy,
    bbfin,
    bbfwdc,
    bbfwdr,
    bbmdata,
    bbmess,
    bbmisc2,
    bbmisc3,
    bbmf,
    bbrdata,
    bbsdata,
    bbsto,
    bbstr,
    bbtask,
    bbwin;

(*===========================================================================*)
(* Operator edit message                                                     *)
(*===========================================================================*)

PROCEDURE oper_em(cmd_in : STRING);

  VAR
    b              : BOOLEAN;
    code           : INTEGER;
    e_msg_no       : LONGINT;
    i              : WORD;
    msg_ptr        : msg_index_ptr;
    new_mf         : msg_flag_type;
    t_char         : CHAR;
    t_str          : STRING;
    t_str_l        : STRING;

  {$I BBOPRM2.PAS}

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Check for valid syntax                                                *)
    (*-----------------------------------------------------------------------*)

    i := WORDS(cmd_in);

    IF i <> 2 THEN
      BEGIN;
        IF i > 2 THEN
          send_message(message_err_wrd)
        ELSE
          send_message(message_not_en);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Convert message number to something reasonable                        *)
    (*-----------------------------------------------------------------------*)

    VAL(subword(@cmd_in, 2, 1), e_msg_no, code);

    IF (code <> 0) OR (e_msg_no < 1) OR (e_msg_no > 65535) THEN
      BEGIN;
        send_message(message_err_ivm);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Find the message number                                               *)
    (*-----------------------------------------------------------------------*)

    msg_ptr := find_msg(e_msg_no);

    IF msg_ptr = NIL THEN
      BEGIN;
        send_message(message_rmc_nf);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    {$IFDEF POINT_CHK}
      test_pointer(msg_ptr);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Send the message header title in format #1                            *)
    (*-----------------------------------------------------------------------*)

    send_msg_header(1);

    (*-----------------------------------------------------------------------*)
    (* Loop while message is being edited                                    *)
    (*-----------------------------------------------------------------------*)

    WHILE TRUE DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Copy the data area to the active area                             *)
        (*-------------------------------------------------------------------*)

        active_tcb^.curr_msg := msg_ptr^;

        (*-------------------------------------------------------------------*)
        (* Send the message header in format #1                              *)
        (*-------------------------------------------------------------------*)

        send_tnc_data_str(header_msg_block(msg_ptr, 1) + cr);

        (*-------------------------------------------------------------------*)
        (* Send edit message                                                 *)
        (*-------------------------------------------------------------------*)

        send_message(message_em);

        (*-------------------------------------------------------------------*)
        (* Get operator response                                             *)
        (*-------------------------------------------------------------------*)

        t_str := read_tnc_data_str;

        (*-------------------------------------------------------------------*)
        (* Strip extra stuff off                                             *)
        (*-------------------------------------------------------------------*)

        strip_crlf(t_str);

        (*-------------------------------------------------------------------*)
        (* If a null line then we are done                                   *)
        (*-------------------------------------------------------------------*)

        IF t_str = '' THEN
          BEGIN;
            update_msg(msg_ptr);
            EXIT;
          END;

        (*-------------------------------------------------------------------*)
        (* Get the command character.  Make sure it is a single character    *)
        (* If it isn't, we will force garbage and an error will occur later  *)
        (*-------------------------------------------------------------------*)

        t_char := UPCASE(t_str[1]);
        IF (t_char <> '.') AND (LENGTH(t_str) > 1) AND (t_str[2] <> ' ') THEN
          t_char := '?';

        (*-------------------------------------------------------------------*)
        (* Get the operand.  It must be at least one character long for the  *)
        (* rest of the code to work                                          *)
        (*-------------------------------------------------------------------*)

        IF t_char = '.' THEN
          t_str := '. ' + COPY(t_str, 2, 255);

        t_str := subword(@t_str, 2, 0);
        IF t_str = '' THEN
          t_str[1] := ' ';

        (*-------------------------------------------------------------------*)
        (* Have both an untouched and an upper-cased copy.                   *)
        (*-------------------------------------------------------------------*)

        t_str_l := t_str;
        upcase_str_var(t_str);

        (*-------------------------------------------------------------------*)
        (* Execute the right command                                         *)
        (*-------------------------------------------------------------------*)

        CASE t_char OF

          'T' : BEGIN;
                  msg_ptr^.msg_i_mb.msg_to      := t_str;
                  IF msg_route_num > e_msg_no THEN
                    msg_route_num := e_msg_no;
                  fwd_route_clean_msg(msg_ptr);
                END;

          'B' : BEGIN;
                  msg_ptr^.msg_i_mb.msg_to_at   := t_str;
                  IF msg_route_num > e_msg_no THEN
                    msg_route_num := e_msg_no;
                  IF last_h_msg > e_msg_no THEN
                    last_h_msg := e_msg_no;
                  fwd_route_clean_msg(msg_ptr);
                END;

          'F' : msg_ptr^.msg_i_mb.msg_from    := t_str;
          'A' : msg_ptr^.msg_i_mb.msg_from_at := t_str;
          'R' : msg_ptr^.msg_i_mb.msg_from_h  := t_str;
          'D' : msg_ptr^.msg_i_mb.msg_type    := t_str[1];
          'X' : msg_ptr^.msg_i_mb.msg_from_h  := t_str;

          'I' : BEGIN;
                  msg_ptr^.msg_i_mb.msg_bid := t_str;
                  msg_ptr^.msg_i_mb.msg_flag :=
                                   msg_ptr^.msg_i_mb.msg_flag OR mf_bid_change;
                END;

          'J' : BEGIN;

                  i := LENGTH(msg_ptr^.msg_i_mb.msg_subj);
                  IF i = LENGTH(t_str_l) THEN
                    MOVE(t_str_l[1], msg_ptr^.msg_i_mb.msg_subj[1], i)
                  ELSE
                    BEGIN;
                      msg_resize(msg_ptr, LENGTH(t_str_l));
                      msg_ptr^.msg_i_mb.msg_subj := t_str_l;
                    END;
                END;

          'H',
          '.' : BEGIN;
                  msg_ptr^.msg_i_mb.msg_to_h := t_str;
                  IF msg_route_num > e_msg_no THEN
                    msg_route_num := e_msg_no;
                  IF last_h_msg > e_msg_no THEN
                    last_h_msg := e_msg_no;
                  fwd_route_clean_msg(msg_ptr);
                END;

          'S' : BEGIN;

                  msg_ptr^.msg_i_mb.msg_flag :=
                               msg_ptr^.msg_i_mb.msg_flag
                                               AND (mf_fwd_list OR mf_disrout);

                  CASE t_str[1] OF
                    'F' : new_mf := mf_fwd;
                    'H' : BEGIN;
                            new_mf := mf_hold;
                            IF msg_ptr^.msg_i_mb.msg_reason = 0 THEN
                              msg_ptr^.msg_i_mb.msg_reason :=
                                                           message_reason_opr;
                          END;
                    'K' : new_mf := mf_kill;
                    'O' : new_mf := mf_old;
                    'V' : new_mf := mf_review;
                    'Y' : new_mf := mf_read;
                    ELSE
                      new_mf := 0;
                  END;

                  msg_ptr^.msg_i_mb.msg_flag :=
                                          msg_ptr^.msg_i_mb.msg_flag OR new_mf;

                  IF msg_route_num > e_msg_no THEN
                    msg_route_num := e_msg_no;

                  fwd_route_clean_msg(msg_ptr);

                END;
          ELSE

            (*---------------------------------------------------------------*)
            (* If we don't understand it then beep the console               *)
            (*---------------------------------------------------------------*)

            IF active_tcb^.tcb_console THEN
              BEGIN;
                SOUND(440);
                DELAY(100);
                NOSOUND;
              END;

        END; (*----- End case statement on command type ---------------------*)

      END;

  END;

(*===========================================================================*)
(* Copy message                                                              *)
(*===========================================================================*)

PROCEDURE oper_cm(cmd_in : STRING);

  VAR
    code           : INTEGER;
    e_msg_no       : LONGINT;
    i              : BYTE;
    msg_ptr        : msg_index_ptr;
    t_str          : STRING[6];

  BEGIN

    active_tcb^.error_sw := FALSE;

    (*-----------------------------------------------------------------------*)
    (* Make sure things are valid                                            *)
    (*-----------------------------------------------------------------------*)

    i := WORDS(cmd_in);

    IF i < 3 THEN
      BEGIN;
        send_message(message_not_en);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Get the message number                                                *)
    (*-----------------------------------------------------------------------*)

    t_str := subword(@cmd_in, 2, 1);
    VAL(t_str, e_msg_no, code);

    IF (code <> 0) OR (e_msg_no < 1) OR (e_msg_no > 65535) THEN
      BEGIN;
        send_message(message_err_ivm);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Find the message                                                      *)
    (*-----------------------------------------------------------------------*)

    msg_ptr := find_msg(e_msg_no);

    IF msg_ptr = NIL THEN
      BEGIN;
        send_message(message_rmc_nf);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    {$IFDEF POINT_CHK}
      test_pointer(msg_ptr);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Copy the message                                                      *)
    (*-----------------------------------------------------------------------*)

    copy_msg(e_msg_no, subword(@cmd_in, 3, 0), TRUE, FALSE);

  END;

(*===========================================================================*)
(* Make message from file                                                    *)
(*===========================================================================*)

PROCEDURE oper_mm(cmd_in : STRING);

  VAR
    f_str : file_name_str;
    i     : INTEGER;
    t_str : file_name_str;
    w_str : STRING;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Clean out the current message info and set the message type from      *)
    (* the incoming command                                                  *)
    (*-----------------------------------------------------------------------*)

    FILLCHAR(active_tcb^.curr_msg, SIZEOF(msg_index_block), CHR(0));

    WITH active_tcb^.curr_msg.msg_i_mb DO
      BEGIN;

        msg_type := UPCASE(cmd_in[2]);

        msg_flag := 0;

      END;

    (*-----------------------------------------------------------------------*)
    (* Make sure things are valid                                            *)
    (*-----------------------------------------------------------------------*)

    i := WORDS(cmd_in);

    IF i < 3 THEN
      BEGIN;
        send_message(message_not_en);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Get the file name                                                     *)
    (*-----------------------------------------------------------------------*)

    IF LENGTH(subword(@cmd_in, 2, 1)) >= SIZEOF(file_name_str) THEN
      BEGIN;
        send_message(message_fn_bad);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    f_str := subword(@cmd_in, 2, 1);

    (*-----------------------------------------------------------------------*)
    (* Process the "TO" data                                                 *)
    (*-----------------------------------------------------------------------*)

    send_msg_to_process(subword(@cmd_in, 3, 0));
    IF active_tcb^.error_sw THEN
      EXIT;

    (*-----------------------------------------------------------------------*)
    (* Get the file name to copy to                                          *)
    (*-----------------------------------------------------------------------*)

    t_str := opt_block.msg_file_dir + active_tcb^.port_chan_s + '.IN';

    (*-----------------------------------------------------------------------*)
    (* Copy the file in                                                      *)
    (*-----------------------------------------------------------------------*)

    cmd_in := copy_file_binary(f_str, t_str, TRUE);
    IF cmd_in <> '' THEN
      BEGIN;
        send_tnc_data_str(cmd_in + cr);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Set the file size                                                     *)
    (*-----------------------------------------------------------------------*)

    active_tcb^.curr_msg.msg_i_mb.msg_size := io_file_size;

    (*-----------------------------------------------------------------------*)
    (* Get the subject                                                       *)
    (*-----------------------------------------------------------------------*)

    send_message(message_smc_subj);
    w_str := read_tnc_data_str;

    WITH active_tcb^.curr_msg.msg_i_mb DO
      BEGIN;

        strip_crlf(w_str);
        i := LENGTH(w_str);

        msg_subj := COPY(w_str, 1, i);

      END;

    (*-----------------------------------------------------------------------*)
    (* Add the message                                                       *)
    (*-----------------------------------------------------------------------*)

    add_msg(t_str, FALSE);
    make_cc;

    send_message(message_added_msg);

  END;

END.
