(*===========================================================================*)
(* Operator task -- Edit ports                                               *)
(*                                                                           *)
(*   Copyright 1991 by H. Roy Engehausen.  All rights reserved.              *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Operator edit control                                                     *)
(*===========================================================================*)

PROCEDURE oper_ep(cmd_in : STRING);

  TYPE
    op_mode_ptr = ^op_mode;

  VAR
    c               : CHAR;
    cmd_word        : str8;
    edit_spot       : op_mode_ptr;
    i               : BYTE;
    interactive     : BOOLEAN;
    control_to_edit : CHAR;
    s               : STRING[2];
    this_port       : port_block_ptr;

  (*=========================================================================*)
  (* Subroutine to display control info                                      *)
  (*=========================================================================*)

  PROCEDURE display_control(edit_spot : op_mode_ptr);
    BEGIN

      IF control_to_edit <> 'S' THEN
        send_tnc_data_str('Port '+ control_to_edit + cr)
      ELSE
        send_tnc_data_str('Master system settings' + cr);

      send_tnc_data_str('1: Lock out connects:          '
                                   + on_or_off(edit_spot^.mode_stop_connect));

      send_tnc_data_str('2: Don''t forward on this port: '
                                   + on_or_off(edit_spot^.mode_stop_fwd));

    END;

  (*=========================================================================*)
  (* Turn a switch on or off                                                 *)
  (*=========================================================================*)

  PROCEDURE set_switch(VAR this_switch : BOOLEAN; invert : BOOLEAN);

    BEGIN;

      IF LENGTH(cmd_in) = 0 THEN
        this_switch := NOT this_switch
      ELSE
        this_switch := invert XOR ((ORD(cmd_in[1]) AND 1) = 1);

      switch_show(invert XOR this_switch);

    END;

  (*=========================================================================*)
  (* Process a line                                                          *)
  (*=========================================================================*)

  PROCEDURE process_line;

    BEGIN;


      s      := subword(@cmd_in, 1, 1);
      cmd_in := subword(@cmd_in, 2, 2);

      c := s[1];

      IF (LENGTH(s) > 1) OR (words(cmd_in) > 1) THEN
        c := #0;

      CASE c OF

        '?' : display_control(edit_spot);
        '1' : set_switch(edit_spot^.mode_stop_connect,    FALSE);
        '2' : set_switch(edit_spot^.mode_stop_fwd,        FALSE);

        ELSE
          BEGIN;
            send_message(message_unknown_edit);
            IF active_tcb^.tcb_console THEN
              BEGIN;
                SOUND(440);
                DELAY(100);
                NOSOUND;
              END;
          END;

      END;

    END;

  (*=========================================================================*)
  (* Main line of edit control                                               *)
  (*=========================================================================*)

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Check the command for validity                                        *)
    (*-----------------------------------------------------------------------*)

    upcase_str_var(cmd_in);

    i := WORDS(cmd_in);

    interactive := (cmd_in[1] = 'E') AND (cmd_in[2] = 'P');

    IF i <> 2 THEN
      BEGIN;
        IF i < 2 THEN
          send_message(message_not_en)
        ELSE
          send_message(message_err_wrd);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Get the control word                                                  *)
    (*-----------------------------------------------------------------------*)

    cmd_word := SUBWORD(@cmd_in, 1, 1);

    (*-----------------------------------------------------------------------*)
    (* Get the control letter                                                *)
    (*-----------------------------------------------------------------------*)

    cmd_in := SUBWORD(@cmd_in, 2, 1);
    IF LENGTH(cmd_in) <> 1 THEN
      BEGIN;
        send_message(message_bad_task);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    control_to_edit := UPCASE(cmd_in[1]);

    (*-----------------------------------------------------------------------*)
    (* Find the control                                                      *)
    (*-----------------------------------------------------------------------*)

    IF control_to_edit = 'S' THEN
      edit_spot := @opt_block.operate_mode
    ELSE
      BEGIN;

        this_port := ring_port^.next_port;

        WHILE (this_port <> NIL)
                                 AND (this_port^.port_char <> control_to_edit)
                                 AND (this_port <> ring_port) DO
          this_port := this_port^.next_port;

        IF (this_port = NIL) OR (this_port^.port_char <> control_to_edit) THEN
          BEGIN;
            send_message(message_bad_task);
            active_tcb^.error_sw := TRUE;
            EXIT;
          END;

        edit_spot := @this_port^.port_operate_mode;

      END;

    (*-----------------------------------------------------------------------*)
    (* Handle 'CLOSE' command                                                *)
    (*-----------------------------------------------------------------------*)

    IF cmd_word = 'CLOSE' THEN
      BEGIN;
        cmd_in := '1 1';
        process_line;
        cmd_in := '2 1';
        process_line;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Handle 'OPEN' command                                                 *)
    (*-----------------------------------------------------------------------*)

    IF cmd_word = 'OPEN' THEN
      BEGIN;
        cmd_in := '1 0';
        process_line;
        cmd_in := '2 0';
        process_line;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Prepare to start edit loop                                            *)
    (*-----------------------------------------------------------------------*)

    display_control(edit_spot);

    REPEAT
      cmd_in := read_tnc_data_str;

      strip_crlf(cmd_in);
      upcase_str_var(cmd_in);

      IF LENGTH(cmd_in) = 0 THEN
        EXIT;

      process_line;

    UNTIL FALSE;

  END;
