(*===========================================================================*)
(* Operator task -- Work with user file                                      *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

UNIT BBOPRU;

INTERFACE

PROCEDURE oper_du(cmd_in : STRING);
PROCEDURE oper_eu(cmd_in : STRING);

IMPLEMENTATION

  USES
    CRT,
    bbdummy,
    bbmdata,
    bbmess,
    bbmisc,
    bbmore,
    bbrdata,
    bbsdata,
    bbstr,
    bbtime,
    bbuf,
    bbwin,
    match;

(*===========================================================================*)
(* Display a userid                                                          *)
(*===========================================================================*)

FUNCTION display_uid(uid_data : user_record_ptr) : STRING;

  VAR
    i     : WORD;
    n_str  : STRING[7];
    t_char : CHAR;
    t_str  : STRING;

  BEGIN;

    WITH uid_data^ DO
      BEGIN;

        t_str := LEFT(user_id, 6) + ' ' + LEFT(user_ssid, 2) + ' ' +
                                               time_str(user_last, TRUE) + ' ';

        IF user_l_cnt > 999 THEN
          i := 999
        ELSE
          i := user_l_cnt;
        STR(i:3, n_str);

        CASE user_class OF
          user_c_nu  : t_char := 'N';
          user_c_uu  : t_char := 'U';
          user_c_ou  : t_char := 'O';
          user_c_eu  : t_char := 'E';
          user_c_bu  : t_char := 'B';
          user_c_rsu : t_char := 'R';
          user_c_lsu : t_char := 'L';
        END;

        t_str := t_str + n_str + ' ' + user_port + t_char + ' ';

        n_str := 'TARXDLE';

        IF (user_flag AND user_f_trans  ) = 0 THEN n_str[1] := ' ';
        IF (user_flag AND user_f_exclude) = 0 THEN n_str[4] := ' ';
        IF (user_flag AND user_f_delete ) = 0 THEN n_str[5] := ' ';
        IF (user_flag AND user_f_local  ) = 0 THEN n_str[6] := ' ';
        IF (user_flag AND user_f_emerg  ) = 0 THEN n_str[7] := ' ';

        IF (user_flag AND user_f_abbs   ) = 0 THEN n_str[2] := 'B';
        IF (user_flag AND user_f_bbs    ) = 0 THEN n_str[2] := ' ';
        IF ((user_flag AND user_f_pbbs) <> 0) AND (n_str[2] = ' ') THEN
          n_str[2] := 'P';

        IF (user_flag AND user_f_sysop  ) = 0 THEN
          n_str[3] := ' ';

        t_str := t_str + n_str + ' ' + LEFT(user_bbs, 10) + ' ' +
                       LEFT(user_zip, 10) + ' ' + user_name;

      END;

    display_uid := t_str;

  END;

(*===========================================================================*)
(* Operator display users                                                    *)
(*===========================================================================*)

PROCEDURE oper_du(cmd_in : STRING);

  VAR
    find_this : call_sign_str;
    head_sent : BOOLEAN;
    this_uid  : user_index_ptr;
    w_cnt     : WORD;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Validate command                                                      *)
    (*-----------------------------------------------------------------------*)

    w_cnt := WORDS(cmd_in);

    IF w_cnt > 2 THEN
      BEGIN;
        send_message(message_err_wrd);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Get search arguement if there is one                                  *)
    (*-----------------------------------------------------------------------*)

    IF w_cnt = 2 THEN
      BEGIN;
        find_this := subword(@cmd_in, 2, 1);
        upcase_str_var(find_this);
      END;

    (*-----------------------------------------------------------------------*)
    (* Handle two word command for users                                     *)
    (*-----------------------------------------------------------------------*)

    IF (w_cnt = 2) AND (active_tcb^.uid_data.user_class < user_c_rsu) THEN
      BEGIN;

        this_uid := find_uid(find_this);

        IF this_uid = NIL THEN
          BEGIN;
            send_message(message_u_not_fnd);
            active_tcb^.error_sw := TRUE;
            EXIT;
          END;

        send_message(message_user_head);
        send_tnc_data_str(display_uid(get_uid(this_uid)) + cr);

        EXIT;

      END;

    (*-----------------------------------------------------------------------*)
    (* User must specify a user to display                                   *)
    (*-----------------------------------------------------------------------*)

    IF active_tcb^.uid_data.user_class < user_c_rsu THEN
      BEGIN;
        send_message(message_not_en);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Handle command to list all or a match                                 *)
    (*-----------------------------------------------------------------------*)

    head_sent := FALSE;

    this_uid := uid_chain;
    WHILE this_uid <> NIL DO
      BEGIN;

        IF (w_cnt = 1) OR match_str(this_uid^.user_id, find_this) THEN
          BEGIN;
            IF NOT head_sent THEN
              BEGIN;
                send_message(message_user_head);
                head_sent := TRUE;
              END;
            send_tnc_data_str(display_uid(get_uid(this_uid)) + cr);
          END;

        this_uid := this_uid^.user_next;
        IF more_check THEN
          BEGIN;
            this_uid := NIL;
            send_message(message_op_halted);
          END;

      END;

    (*-----------------------------------------------------------------------*)
    (* If the header was not shown then we never found anything              *)
    (*-----------------------------------------------------------------------*)

    IF NOT head_sent THEN
      BEGIN;
        send_message(message_u_not_fnd);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

  END;

(*===========================================================================*)
(* Operator edit users                                                       *)
(*===========================================================================*)

PROCEDURE oper_eu(cmd_in : STRING);

  CONST
    no_operands = 'ABDELPTX';

  VAR
    code           : INTEGER;
    i              : WORD;
    j              : LONGINT;
    k              : BYTE;
    look_tcb       : tcb_ptr;
    modem_line     : BOOLEAN;
    no_save        : BOOLEAN;
    t_char         : CHAR;
    t_str          : STRING[3];
    save_uid       : user_record_type;
    this_uid       : user_record_ptr;
    this_uid_place : user_record_type;
    this_uid_index : user_index_ptr;
    uid_to_find    : call_sign_str;

  LABEL
    edit_loop;

  PROCEDURE do_a_beep;
    BEGIN;
      IF active_tcb^.tcb_console THEN
        BEGIN;
          SOUND(440);
          DELAY(100);
          NOSOUND;
        END;
    END;

  BEGIN;

    modem_line := (active_port^. port_type = port_modem)
                                OR (active_port^. port_type = port_null_modem);

    (*-----------------------------------------------------------------------*)
    (* Check the command for validity                                        *)
    (*-----------------------------------------------------------------------*)

    i := WORDS(cmd_in);

    IF i <> 2 THEN
      BEGIN;
        IF i < 2 THEN
          send_message(message_not_en)
        ELSE
          send_message(message_err_wrd);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Extract the user and find him/her/it                                  *)
    (*-----------------------------------------------------------------------*)

    uid_to_find    := subword(@cmd_in, 2, 1);
    {$V-}
    upcase_str_var(uid_to_find);
    {$V+}

    this_uid_index := find_uid(uid_to_find);

    (*-----------------------------------------------------------------------*)
    (* User not found, add one                                               *)
    (*-----------------------------------------------------------------------*)

    IF this_uid_index = NIL THEN
      BEGIN;

        uid_init(@this_uid_place);
        WITH this_uid_place DO
          BEGIN;
            user_id   := uid_to_find;
            user_ssid := '';
          END;

        add_uid(@this_uid_place);
        this_uid_index := find_uid(uid_to_find);

      END;

    (*-----------------------------------------------------------------------*)
    (* If we are editing ourselves, that is cool                             *)
    (*-----------------------------------------------------------------------*)

    IF this_uid_index = active_tcb^.uid_data.user_i_ptr THEN
      BEGIN;
        no_save  := TRUE;
        this_uid := @active_tcb^.uid_data
      END
    ELSE
      BEGIN;
        no_save        := FALSE;
        this_uid       := @this_uid_place;
        this_uid_place := get_uid(this_uid_index)^;
      END;

    (*-----------------------------------------------------------------------*)
    (* Send heading                                                          *)
    (*-----------------------------------------------------------------------*)

    send_message(message_user_head);

    (*-----------------------------------------------------------------------*)
    (* Loop editing this user                                                *)
    (*-----------------------------------------------------------------------*)

    WHILE TRUE DO
      BEGIN;

edit_loop:

        (*-------------------------------------------------------------------*)
        (* Display the user data                                             *)
        (*-------------------------------------------------------------------*)

        send_tnc_data_str(display_uid(this_uid)+ cr);
        str(this_uid^.max_pac, t_str);

        (*-------------------------------------------------------------------*)
        (* If the user has a password and this is the local operator,        *)
        (* display the password.  Display password to remote sysops          *)
        (* logged on via modem                                               *)
        (*-------------------------------------------------------------------*)

        IF this_uid^.user_pw <> '' THEN
          IF (active_tcb^.tcb_type = th_operator)
               OR ((active_tcb^.uid_data.user_class >= user_c_rsu)
                                                          AND modem_line) THEN
            set_dollar1_parm(@this_uid^.user_pw)
          ELSE
            BEGIN;
              cmd_in := '????????';
              set_dollar1_parm(@cmd_in);
            END;

        (*-------------------------------------------------------------------*)
        (* If this isn't ourselves, save our user data and put in who        *)
        (* we want to edit.  This makes messages work better                 *)
        (*-------------------------------------------------------------------*)

        IF NOT no_save THEN
          BEGIN;
            save_uid := active_tcb^.uid_data;
            active_tcb^.uid_data := this_uid^;
          END;

        (*-------------------------------------------------------------------*)
        (* Send the edit user message                                        *)
        (*-------------------------------------------------------------------*)

        send_message(message_eu);

        (*-------------------------------------------------------------------*)
        (* Swap the data back so everything is AOK                           *)
        (*-------------------------------------------------------------------*)

        IF NOT no_save THEN
          active_tcb^.uid_data := save_uid;

        (*-------------------------------------------------------------------*)
        (* Get the user's request                                            *)
        (*-------------------------------------------------------------------*)

        cmd_in := read_tnc_data_str;
        strip_crlf(cmd_in);

        (*-------------------------------------------------------------------*)
        (* If it is blank, we are done                                       *)
        (*-------------------------------------------------------------------*)

        IF cmd_in = '' THEN
          BEGIN;

            (*---------------------------------------------------------------*)
            (* Update the UID                                                *)
            (*---------------------------------------------------------------*)

            update_uid(this_uid);

            (*---------------------------------------------------------------*)
            (* Check to see if this is the local SYSOP.  If so, replace      *)
            (* all occurances of this user                                   *)
            (*---------------------------------------------------------------*)

            IF uid_to_find = main_tcb^.uid_data.user_id THEN
              BEGIN;

                look_tcb := main_tcb;
                REPEAT
                  IF uid_to_find = look_tcb^.uid_data.user_id THEN
                    look_tcb^.uid_data := this_uid^;
                  look_tcb := look_tcb^.next_tcb;
                UNTIL look_tcb = main_tcb;

              END;

            (*---------------------------------------------------------------*)
            (* Done                                                          *)
            (*---------------------------------------------------------------*)

            EXIT;

          END;

        (*-------------------------------------------------------------------*)
        (* All the edit commands are one character so here we check to       *)
        (* see if the command had more than one character.  If so, we        *)
        (* force an abort for later                                          *)
        (*-------------------------------------------------------------------*)

        IF (LENGTH(cmd_in) > 1) AND (cmd_in[2] <> ' ') THEN
          cmd_in[1] := #0;

        (*-------------------------------------------------------------------*)
        (* Get the command and operands seperated                            *)
        (*-------------------------------------------------------------------*)

        t_char := UPCASE(cmd_in[1]);
        cmd_in  := COPY(cmd_in, 3, 255);
        strip_var(cmd_in, 'B');

        (*-------------------------------------------------------------------*)
        (* Check operand count                                               *)
        (*-------------------------------------------------------------------*)

        IF (POS(t_char, no_operands) <> 0) AND (cmd_in <> '') THEN
          BEGIN;
            send_message(message_err_wrd);
            do_a_beep;
            GOTO edit_loop;
          END;

        (*-------------------------------------------------------------------*)
        (* Fill things out                                                   *)
        (*-------------------------------------------------------------------*)

        IF cmd_in = '' THEN
          cmd_in[1] := ' ';

        (*-------------------------------------------------------------------*)
        (* Now we execute the command                                        *)
        (*-------------------------------------------------------------------*)

        WITH this_uid^ DO
          CASE t_char OF

            (*---------------------------------------------------------------*)
            (* ABBS flag.  If ABBS is off turn on ABBS and BBS else turn     *)
            (* off both ABBS and BBS                                         *)
            (*---------------------------------------------------------------*)

            'A' : BEGIN;
                    user_flag := user_flag AND NOT user_f_pbbs;
                    IF (user_flag AND user_f_abbs) = 0 THEN
                      user_flag := user_flag OR (user_f_abbs OR user_f_bbs)
                    ELSE
                      user_flag := user_flag
                                           AND NOT(user_f_bbs OR user_f_abbs);
                  END;

            (*---------------------------------------------------------------*)
            (* BBS flag.  Invert flag and remove any ABBS or PBBS flags      *)
            (*---------------------------------------------------------------*)

            'B' : BEGIN;

                    user_flag := (user_flag XOR user_f_bbs)
                                        AND (NOT (user_f_abbs OR user_f_pbbs));

                  END;


            (*---------------------------------------------------------------*)
            (* Set user class                                                *)
            (*---------------------------------------------------------------*)

            'C' : BEGIN;
                    t_char := UPCASE(cmd_in[1]);
                    i := POS(t_char, 'NUOEBRL');
                    IF i > 0 THEN
                      i := i - 1;
                    user_class := user_class_type(i);
                  END;

            (*---------------------------------------------------------------*)
            (* Simple flag inverts                                           *)
            (*---------------------------------------------------------------*)

            'D' : user_flag := user_flag XOR user_f_delete;
            'E' : user_flag := user_flag XOR user_f_emerg;

            (*---------------------------------------------------------------*)
            (* Set display format code                                       *)
            (*---------------------------------------------------------------*)

            'F' : BEGIN;
                    cmd_in := subword(@cmd_in, 1, 1);
                    VAL(cmd_in, j, code);
                    IF (code <> 0) OR (j < 0) OR (j > 2) THEN
                      BEGIN;
                        send_tnc_data_str('Invalid format'+ cr);
                        do_a_beep;
                      END
                    ELSE
                      user_fmt := j;
                  END;

            (*---------------------------------------------------------------*)
            (* Set home BBS                                                  *)
            (*---------------------------------------------------------------*)

            'H' : BEGIN;
                    user_bbs    := upcase_str(cmd_in);
                    user_flag   := user_flag OR user_f_adrchg;
                    user_n_time := current_day_time;
                  END;

            (*---------------------------------------------------------------*)
            (* Invert local flag                                             *)
            (*---------------------------------------------------------------*)

            'L' : user_flag := user_flag XOR user_f_local;

            (*---------------------------------------------------------------*)
            (* Set maximum packet size                                       *)
            (*---------------------------------------------------------------*)

            'M' : BEGIN;
                    cmd_in := subword(@cmd_in, 1, 1);
                    VAL(cmd_in, j, code);
                    IF (code <> 0) OR ((j < 20) AND (j <> 0)) OR (j > 255) THEN
                      BEGIN;
                        send_tnc_data_str('Invalid packet size'+ cr);
                        do_a_beep;
                      END
                    ELSE
                      max_pac := j;
                  END;

            (*---------------------------------------------------------------*)
            (* Set name                                                      *)
            (*---------------------------------------------------------------*)

            'N' : user_name := cmd_in;

            (*---------------------------------------------------------------*)
            (* Set language                                                  *)
            (*---------------------------------------------------------------*)

            'Q' : BEGIN;
                    user_lang := UPCASE(cmd_in[1]);
                    IF user_lang = ' ' THEN
                      user_lang := '?';
                  END;

            (*---------------------------------------------------------------*)
            (* PBBS flag.  Invert flag and remove any BBS flags              *)
            (*---------------------------------------------------------------*)

            'P' : BEGIN;

                    user_flag := (user_flag XOR user_f_pbbs)
                                         AND (NOT (user_f_abbs OR user_f_bbs));

                  END;

            (*---------------------------------------------------------------*)
            (* Set remote sysop flag and password number                     *)
            (*---------------------------------------------------------------*)

            'R' : IF cmd_in = '' THEN
                    BEGIN;
                      user_flag := user_flag XOR user_f_sysop;
                      IF (user_flag AND user_f_sysop) = 0 THEN
                        user_sysop_p := 0;
                    END
                  ELSE
                    BEGIN;
                      user_flag := user_flag OR user_f_sysop;
                      VAL(cmd_in, j, code);
                      IF (code <> 0) OR (j < 0) OR (j > 35) THEN
                        BEGIN;
                          send_tnc_data_str('Invalid format'+ cr);
                          do_a_beep;
                        END
                      ELSE
                        user_sysop_p := j;
                    END;

            (*---------------------------------------------------------------*)
            (* Set screen length                                             *)
            (*---------------------------------------------------------------*)

            'S' : BEGIN;
                    cmd_in := subword(@cmd_in, 1, 1);
                    VAL(cmd_in, j, code);
                    IF (code <> 0) OR (j < 0) OR (j > 255) THEN
                      BEGIN;
                        send_tnc_data_str('Invalid screen length'+ cr);
                        do_a_beep;
                      END
                    ELSE
                      user_scr_len := j;
                  END;

            (*---------------------------------------------------------------*)
            (* Invert transparent flag                                       *)
            (*---------------------------------------------------------------*)

            'T' : user_flag := user_flag XOR user_f_trans;

            (*---------------------------------------------------------------*)
            (* Access control                                                *)
            (*---------------------------------------------------------------*)

            'U': BEGIN;

                   cmd_in := upcase_str(cmd_in);
                   user_access.access_flags := 0;

                   FOR i := 1 TO LENGTH(cmd_in) DO
                     BEGIN;

                       k := POS(cmd_in[i], access_letter_group);
                       IF k = 0 THEN
                         BEGIN;
                           send_tnc_data_str(
                                      'Invalid character in access list -- '
                                      + cmd_in[i]
                                      + cr);
                           do_a_beep;
                           GOTO edit_loop;
                         END;

                       IF k > 1 THEN
                         k := $80 SHR (k-1)
                       ELSE
                         k := $80;

                         user_access.access_flags :=
                                                 user_access.access_flags OR k;

                     END;

                   END;

            (*---------------------------------------------------------------*)
            (* Set modem password                                            *)
            (*---------------------------------------------------------------*)

            'W' : user_pw   := upcase_str(cmd_in);

            (*---------------------------------------------------------------*)
            (* Invert exclusion flag                                         *)
            (*---------------------------------------------------------------*)

            'X' : user_flag := user_flag XOR user_f_exclude;

            (*---------------------------------------------------------------*)
            (* Set screen width                                              *)
            (*---------------------------------------------------------------*)

            'Y' : BEGIN;
                    cmd_in := subword(@cmd_in, 1, 1);
                    VAL(cmd_in, j, code);
                    IF (code <> 0) OR (j < 0) OR (j > 255) THEN
                      BEGIN;
                        send_tnc_data_str('Invalid screen length'+ cr);
                        do_a_beep;
                      END
                    ELSE
                      user_scr_wid := j;
                  END;

            (*---------------------------------------------------------------*)
            (* Set home zip code                                             *)
            (*---------------------------------------------------------------*)

            'Z' : user_zip  := upcase_str(cmd_in);

            (*---------------------------------------------------------------*)
            (* Handle errors                                                 *)
            (*---------------------------------------------------------------*)

            ELSE
              BEGIN;
                send_tnc_data_str('Bad command'+ cr);
                do_a_beep;
              END;

          END; (*----- End CASE statement to execute the command ------------*)

      END; (*----- End forever loop -----------------------------------------*)

  END;

(*===========================================================================*)
(* Operator edit ports                                                       *)
(*===========================================================================*)

PROCEDURE oper_ep(cmd_in : STRING);

  TYPE
    op_mode_ptr = ^op_mode;

  VAR
    c            : CHAR;
    edit_spot    : op_mode_ptr;
    i            : BYTE;
    port_to_edit : CHAR;
    s            : STRING[2];
    this_port    : port_block_ptr;

  (*=========================================================================*)
  (* Return ON or OFF                                                        *)
  (*=========================================================================*)

  FUNCTION  on_or_off(b : BOOLEAN) : str8;
    BEGIN;
      IF b THEN
        on_or_off := 'ON' + cr
      ELSE
        on_or_off := 'OFF' + cr;
    END;

  (*=========================================================================*)
  (* Subroutine to display port info                                         *)
  (*=========================================================================*)

  PROCEDURE display_port(edit_spot : op_mode_ptr);
    BEGIN

      IF port_to_edit <> 'S' THEN
        send_tnc_data_str('Port '+ port_to_edit + cr)
      ELSE
        send_tnc_data_str('Master system settings' + cr);

      send_tnc_data_str('1: Accept Special Traffic Only:  '
                                   + on_or_off(edit_spot^.mode_acc_emer));

      send_tnc_data_str('2: Accept NTS Traffic:           '
                                   + on_or_off(NOT edit_spot^.mode_acc_nonts));

      send_tnc_data_str('3: Forward Special Traffic Only: '
                                   + on_or_off(edit_spot^.mode_fwd_emer));

      send_tnc_data_str('4: Forward NTS Traffic:          '
                                   + on_or_off(NOT edit_spot^.mode_fwd_nonts));

      send_tnc_data_str('5: "E" Users Only:               '
                                   + on_or_off(edit_spot^.mode_e_users));

      send_tnc_data_str('6: Allow Callsign Changes        '
                                   + on_or_off(edit_spot^.mode_user_change));

    END;

  (*=========================================================================*)
  (* Turn a switch on or off                                                 *)
  (*=========================================================================*)

  PROCEDURE set_switch(VAR this_switch : BOOLEAN; invert : BOOLEAN);

    BEGIN;

      IF cmd_in = 'ON' THEN
        cmd_in := '1';

      IF cmd_in = 'OFF' THEN
        cmd_in := '0';

      IF LENGTH(cmd_in) = 0 THEN
        this_switch := NOT this_switch
      ELSE
        this_switch := invert XOR ((ORD(cmd_in[1]) AND 1) = 1);

      switch_show(invert XOR this_switch);

    END;

  (*=========================================================================*)
  (* Main line of edit port                                                  *)
  (*=========================================================================*)

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Check the command for validity                                        *)
    (*-----------------------------------------------------------------------*)

    i := WORDS(cmd_in);

    IF i <> 2 THEN
      BEGIN;
        IF i < 2 THEN
          send_message(message_not_en)
        ELSE
          send_message(message_err_wrd);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Get the port letter                                                   *)
    (*-----------------------------------------------------------------------*)

    cmd_in := SUBWORD(@cmd_in, 2, 1);
    IF LENGTH(cmd_in) <> 1 THEN
      BEGIN;
        send_message(message_bad_task);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    port_to_edit := UPCASE(cmd_in[1]);

    (*-----------------------------------------------------------------------*)
    (* Find the port                                                         *)
    (*-----------------------------------------------------------------------*)

    IF port_to_edit = 'S' THEN
      edit_spot := @opt_block.operate_mode
    ELSE
      BEGIN;

        this_port := ring_port^.next_port;

        WHILE (this_port <> NIL)
                                 AND (this_port^.port_char <> port_to_edit)
                                 AND (this_port <> ring_port) DO
          this_port := this_port^.next_port;

        IF (this_port = NIL) OR (this_port^.port_char <> port_to_edit) THEN
          BEGIN;
            send_message(message_bad_task);
            active_tcb^.error_sw := TRUE;
            EXIT;
          END;

        edit_spot := @this_port^.port_operate_mode;

      END;

    (*-----------------------------------------------------------------------*)
    (* Prepare to start edit loop                                            *)
    (*-----------------------------------------------------------------------*)

    display_port(edit_spot);

    REPEAT
      cmd_in := read_tnc_data_str;

      strip_crlf(cmd_in);
      upcase_str_var(cmd_in);

      IF LENGTH(cmd_in) = 0 THEN
        EXIT;

      s      := subword(@cmd_in, 1, 1);
      cmd_in := subword(@cmd_in, 2, 2);

      c := s[1];

      IF (LENGTH(s) > 1) OR (words(cmd_in) > 1) THEN
        c := #0;

      CASE c OF

        '?' : display_port(edit_spot);
        '1' : set_switch(edit_spot^.mode_acc_emer,    FALSE);
        '2' : set_switch(edit_spot^.mode_acc_nonts,   TRUE);
        '3' : set_switch(edit_spot^.mode_fwd_emer,    FALSE);
        '4' : set_switch(edit_spot^.mode_fwd_nonts,   TRUE);
        '5' : set_switch(edit_spot^.mode_e_users,     FALSE);
        '6' : set_switch(edit_spot^.mode_user_change, FALSE);
        '?' : set_switch(edit_spot^.mode_user_change, FALSE);

        ELSE
          BEGIN;
            send_message(message_unknown_edit);
            IF active_tcb^.tcb_console THEN
              BEGIN;
                SOUND(440);
                DELAY(100);
                NOSOUND;
              END;
          END;

      END;

    UNTIL FALSE;

  END;

END.
