(*===========================================================================*)
(* Register user command                                                     *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

UNIT BBREG;

INTERFACE

  USES
    bbdummy,
    bbmdata,
    bbmess,
    bbrdata,
    bbsess,
    bbstr,
    bbuf;

PROCEDURE register_cmd(cmd_string : STRING);

IMPLEMENTATION

(*===========================================================================*)
(* REGISTER command                                                          *)
(*      Entered when 'N' received by command processor                       *)
(*===========================================================================*)

PROCEDURE register_cmd(cmd_string : STRING);

  (*=========================================================================*)
  (* Register command processor for each letter                              *)
  (*=========================================================================*)

  PROCEDURE register_letter(cmd_string : STRING);

    VAR
      cmd_letter    : CHAR;
      code          : INTEGER;
      i             : INTEGER;
      scr_save      : STRING[5];
      w_cnt         : BYTE;

    (*=======================================================================*)
    (* Subprocedure to get an operand                                        *)
    (*=======================================================================*)

    PROCEDURE register_get(m_no : BYTE; m_size : BYTE);
      BEGIN;

        IF w_cnt = 0 THEN
          BEGIN;
            send_message(m_no);
            cmd_string := read_tnc_data_str;
            strip_crlf(cmd_string);
            w_cnt := words(cmd_string);
          END;

        IF w_cnt > 1 THEN
          BEGIN;
            send_message(message_err_wrd);
            active_tcb^.error_sw := TRUE;
            EXIT;
          END;

        IF w_cnt < 1 THEN
          BEGIN;
            send_message(message_nodata_can);
            active_tcb^.error_sw := TRUE;
            EXIT;
          END;

        IF LENGTH(cmd_string) > m_size THEN
          BEGIN;
            send_message(message_too_long);
            active_tcb^.error_sw := TRUE;
            EXIT;
          END;

      END;

    (*=======================================================================*)
    (* Main line                                                             *)
    (*=======================================================================*)

    BEGIN;

      (*---------------------------------------------------------------------*)
      (* Parse command                                                       *)
      (*---------------------------------------------------------------------*)

      cmd_letter := UPCASE(cmd_string[2]);

      cmd_string := COPY(cmd_string, 3, 255);
      strip_crlf(cmd_string);
      w_cnt      := words(cmd_string);

      (*---------------------------------------------------------------------*)
      (* Don't permit password changes except on land line                   *)
      (*---------------------------------------------------------------------*)

      IF (cmd_letter = 'P') AND (active_port^.port_type <> port_modem) THEN
        cmd_letter := '?';

      (*---------------------------------------------------------------------*)
      (* No language change if no list                                       *)
      (*---------------------------------------------------------------------*)

      IF (cmd_letter = 'L') AND (opt_block.language_list = '') THEN
        cmd_letter := '?';

      (*---------------------------------------------------------------------*)
      (* Go to proper routine                                                *)
      (*---------------------------------------------------------------------*)

      WITH active_tcb^.uid_data DO
        BEGIN;

          CASE cmd_letter OF

            (*---------------------------------------------------------------*)
            (* Change home BBS                                               *)
            (*---------------------------------------------------------------*)

            'H' : BEGIN;

                    register_get(message_user_home, SIZEOF(user_bbs));
                    IF active_tcb^.error_sw THEN
                      EXIT;

                    user_bbs    := upcase_str(cmd_string);
                    user_flag   := user_flag OR user_f_adrchg;
                    user_n_time := current_day_time;

                  END;

            (*---------------------------------------------------------------*)
            (* Change password                                               *)
            (*---------------------------------------------------------------*)

            'P' : BEGIN;

                    register_get(message_enter_password, SIZEOF(user_pw));
                    IF active_tcb^.error_sw THEN
                      EXIT;

                    user_pw := upcase_str(subword(@cmd_string, 1, 1));

                  END;

            (*---------------------------------------------------------------*)
            (* Change name                                                   *)
            (*---------------------------------------------------------------*)

            'N' : BEGIN;

                    register_get(message_user_name, SIZEOF(user_name));
                    IF active_tcb^.error_sw THEN
                      EXIT;

                    user_name := cmd_string;
                    user_name[1] := UPCASE(user_name[1]);

                  END;

            (*---------------------------------------------------------------*)
            (* Change zipcode                                                *)
            (*---------------------------------------------------------------*)

            'Z' : BEGIN;

                    register_get(message_user_zip, SIZEOF(user_zip));
                    IF active_tcb^.error_sw THEN
                      EXIT;

                    user_zip := upcase_str(cmd_string);
                    user_flag := user_flag OR user_f_adrchg;

                  END;

            (*---------------------------------------------------------------*)
            (* Change expert                                                 *)
            (*---------------------------------------------------------------*)

            'E' : BEGIN;

                    IF w_cnt > 0 THEN
                      BEGIN;
                        send_message(message_err_wrd);
                        active_tcb^.error_sw := TRUE;
                        EXIT;
                      END;

                    IF user_class = user_c_ou THEN
                      user_class := user_c_eu
                    ELSE
                      IF user_class = user_c_eu THEN
                        user_class := user_c_ou;

                    send_message(message_mode_switch);

                  END;

            (*---------------------------------------------------------------*)
            (* Change message format                                         *)
            (*---------------------------------------------------------------*)

            'F' : BEGIN;

                    register_get(message_user_msg_fmt, 2);
                    IF active_tcb^.error_sw THEN
                      EXIT;

                    i := ORD(cmd_string[1]) - ORD('0');
                    IF (i >= 0) AND (i <= 4) THEN
                      user_fmt := i
                    ELSE
                      BEGIN;
                        send_message(message_bad_number);
                        active_tcb^.error_sw := TRUE;
                        EXIT;
                      END;

                  END;

            (*---------------------------------------------------------------*)
            (* Change language                                               *)
            (*---------------------------------------------------------------*)

            'L' : BEGIN;

                    REPEAT

                      active_tcb^.error_sw := FALSE;

                      register_get(message_user_lang, 1);

                      upcase_str_var(cmd_string);

                      IF active_tcb^.error_sw THEN
                        i := 1
                      ELSE
                        IF cmd_string[1] <> '?' THEN
                          i := POS(cmd_string[1], opt_block.language_list)
                        ELSE
                          i := 256;

                      IF i > 0 THEN
                        user_lang := cmd_string[1]
                      ELSE
                        BEGIN;
                          send_message(message_invalid_select);
                          w_cnt := 0;
                        END;

                    UNTIL i > 0;

                  END;

            (*---------------------------------------------------------------*)
            (* Transparent                                                   *)
            (*---------------------------------------------------------------*)

            'T' : BEGIN;

                    IF w_cnt > 0 THEN
                      BEGIN;
                        send_message(message_err_wrd);
                        active_tcb^.error_sw := TRUE;
                        EXIT;
                      END;

                    user_flag := user_flag XOR user_f_trans;
                    IF (user_flag AND user_f_trans) <> 0 THEN
                      send_message(message_sw_on)
                    ELSE
                      send_message(message_sw_off);

                  END;

            (*---------------------------------------------------------------*)
            (* Change screen info                                            *)
            (*---------------------------------------------------------------*)

            'S' : BEGIN;

                    (*-------------------------------------------------------*)
                    (* Because this letter can have up to two operands, it   *)
                    (* has to be checked special                             *)
                    (*-------------------------------------------------------*)

                    IF w_cnt > 2 THEN
                      BEGIN;
                        send_message(message_err_wrd);
                        active_tcb^.error_sw := TRUE;
                        EXIT;
                      END;

                    (*-------------------------------------------------------*)
                    (* Tuck the second operand away                          *)
                    (*-------------------------------------------------------*)

                    scr_save   := SUBWORD(@cmd_string, 2, 1);
                    cmd_string := SUBWORD(@cmd_string, 1, 1);
                    w_cnt      := WORDS(cmd_string);

                    (*-------------------------------------------------------*)
                    (* Get the screen length                                 *)
                    (*-------------------------------------------------------*)

                    register_get(message_user_scrl, 3);
                    IF active_tcb^.error_sw THEN
                      EXIT;

                    VAL(cmd_string, i, code);
                    IF (code = 0) AND (i >= 0) AND (i <= 255) THEN
                      user_scr_len := i
                    ELSE
                      BEGIN;
                        send_message(message_bad_number);
                        active_tcb^.error_sw := TRUE;
                        EXIT;
                      END;

                    (*-------------------------------------------------------*)
                    (* Restore the saved operand                             *)
                    (*-------------------------------------------------------*)

                    cmd_string := scr_save;
                    w_cnt      := WORDS(cmd_string);

                    (*-------------------------------------------------------*)
                    (* Get the screen width                                  *)
                    (*-------------------------------------------------------*)

                    register_get(message_user_scrw, 3);
                    IF active_tcb^.error_sw THEN
                      EXIT;

                    VAL(cmd_string, i, code);
                    IF (code = 0) AND (i >= 0) AND (i <= 255) THEN
                      user_scr_wid := i
                    ELSE
                      BEGIN;
                        send_message(message_bad_number);
                        active_tcb^.error_sw := TRUE;
                        EXIT;
                      END;

                    (*---------------------------------------------------------*)
                    (* Show current settings to confirm                        *)
                    (*---------------------------------------------------------*)

                    send_message(message_user_scr_disp);

                  END;

            (*-----------------------------------------------------------------*)
            (* Errors                                                          *)
            (*-----------------------------------------------------------------*)

            ELSE

              BEGIN;
                send_message(message_err_2nd);
                active_tcb^.error_sw := TRUE;
                EXIT;
              END;

          END; (*----- End case statement on type -----------------------------*)


        END; (*----- End user info addressability -----------------------------*)

      (*-----------------------------------------------------------------------*)
      (* Update the user file                                                  *)
      (*-----------------------------------------------------------------------*)

      update_uid(@active_tcb^.uid_data);

    END;

  (*=========================================================================*)
  (* Main line of register command processor                                 *)
  (*=========================================================================*)

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Get ready to process the command                                      *)
    (*-----------------------------------------------------------------------*)

    strip_var(cmd_string, 'B');

    (*-----------------------------------------------------------------------*)
    (* See if user wants whole registration process or just one piece        *)
    (*-----------------------------------------------------------------------*)

    IF (words(cmd_string) > 1)
                              OR (LENGTH(subword(@cmd_string, 1, 1)) <> 1) THEN
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Execute the single letter                                         *)
        (*-------------------------------------------------------------------*)

        register_letter(cmd_string);

        (*-------------------------------------------------------------------*)
        (* Leave if it fails                                                 *)
        (*-------------------------------------------------------------------*)

        IF active_tcb^.error_sw THEN
          EXIT;

        (*-------------------------------------------------------------------*)
        (* If user has completed registration piecemeal then upgrade his     *)
        (* status                                                            *)
        (*-------------------------------------------------------------------*)

        IF (active_tcb^.uid_data.user_class < user_c_ou)
                                AND (active_tcb^.uid_data.user_name <> '')
                                AND (active_tcb^.uid_data.user_bbs  <> '')
                                AND (active_tcb^.uid_data.user_zip  <> '') THEN
          BEGIN;
            active_tcb^.uid_data.user_class    := user_c_ou;
            update_uid(@active_tcb^.uid_data);
            send_message(message_reg_ok);
          END;

        (*-------------------------------------------------------------------*)
        (* Done with command.  Tell user and leave                           *)
        (*-------------------------------------------------------------------*)

        send_message(message_op_complete);

        EXIT;

      END; (*----- End single register command process ----------------------*)

    (*-----------------------------------------------------------------------*)
    (* Whole registration process follows                                    *)
    (*-----------------------------------------------------------------------*)

    (*-----------------------------------------------------------------------*)
    (* Name                                                                  *)
    (*-----------------------------------------------------------------------*)

    register_letter('NN');
    IF active_tcb^.error_sw THEN
      EXIT;

    (*-----------------------------------------------------------------------*)
    (* Zip                                                                   *)
    (*-----------------------------------------------------------------------*)

    register_letter('NZ');
    IF active_tcb^.error_sw THEN
      EXIT;

    (*-----------------------------------------------------------------------*)
    (* Home BBS                                                              *)
    (*-----------------------------------------------------------------------*)

    register_letter('NH');
    IF active_tcb^.error_sw THEN
      EXIT;

    (*-----------------------------------------------------------------------*)
    (* Screen size                                                           *)
    (*-----------------------------------------------------------------------*)

    register_letter('NS');
    IF active_tcb^.error_sw THEN
      EXIT;

    (*-----------------------------------------------------------------------*)
    (* Language (if needed)                                                  *)
    (*-----------------------------------------------------------------------*)

    IF (NOT opt_block.opt_no_lang_prompt)
                                      AND (opt_block.language_list <> '')  THEN
      BEGIN;
        register_letter('NL');
        IF active_tcb^.error_sw THEN
          EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Password (if needed)                                                  *)
    (*-----------------------------------------------------------------------*)

    IF (active_port^.port_type = port_modem)
                                  AND (active_tcb^.uid_data.user_pw = '')  THEN
      BEGIN;
        register_letter('NP');
        IF active_tcb^.error_sw THEN
          end_session(FALSE);
      END;

    (*-----------------------------------------------------------------------*)
    (* Upgrade user status if needed                                         *)
    (*-----------------------------------------------------------------------*)

    WITH active_tcb^.uid_data DO
      IF user_class < user_c_ou THEN
        user_class := user_c_ou;

    (*-----------------------------------------------------------------------*)
    (* Update user file                                                      *)
    (*-----------------------------------------------------------------------*)

    update_uid(@active_tcb^.uid_data);

    (*-----------------------------------------------------------------------*)
    (* Tell user AOK                                                         *)
    (*-----------------------------------------------------------------------*)

    send_message(message_reg_ok);

  END;

END.
