(*===========================================================================*)
(* Read msg command                                                          *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991 by H. Roy Engehausen.  All rights      *)
(*   reserved.                                                               *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

{$DEFINE POINT_CHK}

UNIT BBRMC;

INTERFACE

USES
  bbdummy;

PROCEDURE read_msg_cmd(cmd_string : STRING);
PROCEDURE read_a_msg (msg_ptr : msg_index_ptr;
                      output_dest : file_name_str;
                      write_mode  : BYTE;
                      header_sw   : BOOLEAN;
                      x400        : BOOLEAN);

IMPLEMENTATION

USES
  DOS,
  bbedit,
  bbfu,
  bblog,
  bbkmc,
  bbmdata,
  bbmess,
  bbmf,
  bbmisc,
  bbmisc2,
  bbmisc3,
  bbmisc5,
  bbmmsgn,
  bbmore,
  bboprm,
  bbrdata,
  bbsdata,
  bbsearch,
  bbsema2,
  bbsess,
  bbsmc,
  bbstr,
  bbtask,
  bbtime,
  bbwin,
  match;

(*===========================================================================*)
(* Read msg command                                                          *)
(*===========================================================================*)

PROCEDURE read_msg_cmd(cmd_string : STRING);

  VAR
    err_string   : STRING[6];
    exec_char    : CHAR;
    r_msg_no     : LONGINT;
    read_sw      : BOOLEAN;
    review_cmd   : BOOLEAN;
    msg_ptr      : msg_index_ptr;
    search_block : search_block_type;
    send_header  : BOOLEAN;
    send_x400    : BOOLEAN;
    word_count   : BYTE;

  (*=========================================================================*)
  (* Subroutine to do the actual read                                        *)
  (*=========================================================================*)

  PROCEDURE read_this_one;

    VAR
      action_code : BYTE;
      fl          : msg_flag_type;
      i           : BYTE;
      save_msg_no : LONGINT;

    CONST
      action_letters : STRING[7] = 'LEQKRHC'; (* The number of the letter pos *)
                                             (* in this string is the action *)
                                             (* number.  Must be coordinated *)
                                             (* with BBEDITS quit keys       *)

      action_words : STRING[41] = 'LEAVE REPLY QUIT KILL RELEASE HOLD CHANGE';
                                (* 12345678901234567890123456789012345678901 *)
                                             (* Words in same order as letter*)

    LABEL msg_action_loop;

    (*=======================================================================*)
    (* Issue the reply action prompt                                         *)
    (*=======================================================================*)

    PROCEDURE line_prompt;

      LABEL do_action;

      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Prompt SYSOP for what he wants done with message.  Stay in        *)
        (* loop until he gets it right                                       *)
        (*-------------------------------------------------------------------*)

        REPEAT

          (*-----------------------------------------------------------------*)
          (* Send prompt message                                             *)
          (*-----------------------------------------------------------------*)

          send_message(message_review_prompt);

          (*-----------------------------------------------------------------*)
          (* Get response and clean it up                                    *)
          (*-----------------------------------------------------------------*)

          err_string := read_tnc_data_str;
          strip_crlf(err_string);
          upcase_str_var(err_string);

          (*-----------------------------------------------------------------*)
          (* Make sure too many words will give an error                     *)
          (*-----------------------------------------------------------------*)

          IF WORDS(err_string) > 1 THEN
            err_string := '?';

          (*-----------------------------------------------------------------*)
          (* Handle special aliases                                          *)
          (*-----------------------------------------------------------------*)

          IF err_string = 'EDIT' THEN
            BEGIN;
              err_string := 'C';
              GOTO do_action;
            END;

          action_code := FIND(@action_words, @err_string);
          IF action_code <> 0 THEN
            err_string := CHR(ORD('0') + action_code);

          (*-----------------------------------------------------------------*)
          (* Come here after alias processing                                *)
          (*-----------------------------------------------------------------*)

do_action:

          (*-----------------------------------------------------------------*)
          (* Validate the single letter for action                           *)
          (*-----------------------------------------------------------------*)

          IF LENGTH(err_string) = 1 THEN
            BEGIN;

              exec_char := err_string[1];
              action_code := POS(exec_char, action_letters);

              IF action_code = 0 THEN
                action_code := POS(exec_char, '1234567');

              IF (active_tcb^.uid_data.user_class < user_c_rsu)
                                                     AND (action_code > 4) THEN
                action_code := 0;

            END;

          (*-----------------------------------------------------------------*)
          (* Give bad action message if needed and loop                      *)
          (*-----------------------------------------------------------------*)

          IF action_code = 0 THEN
            send_message(message_bad_review_actn);

        UNTIL action_code <> 0;

      END;

    (*=======================================================================*)
    (* Main line of actual read subroutine                                   *)
    (*=======================================================================*)

    BEGIN;

      IF review_cmd THEN
        active_tcb^.curr_msg := msg_ptr^;

      (*---------------------------------------------------------------------*)
      (* See if we have to do a read or review.  If review then we may       *)
      (* use the line by line mode rather than full screen                   *)
      (*---------------------------------------------------------------------*)

      IF (NOT review_cmd)
                          OR (active_tcb^.tcb_type <> th_operator)
                          OR (opt_block.opt_no_full_review) THEN
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* Line by line mode read/review                                   *)
          (*-----------------------------------------------------------------*)

          (*-----------------------------------------------------------------*)
          (* Send the message data                                           *)
          (*-----------------------------------------------------------------*)

          read_a_msg(msg_ptr, '', 1, send_header, send_x400);

          (*-----------------------------------------------------------------*)
          (* Prompt user for what he wants done with message.  Stay in       *)
          (* loop until he gets it right                                     *)
          (*-----------------------------------------------------------------*)

          IF review_cmd THEN
            line_prompt;

        END (*----- End line by line processor ------------------------------*)
      ELSE
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* Full screen review mode                                         *)
          (*-----------------------------------------------------------------*)

          edit_exitcode  := 4; (* Default action is QUIT! *)

          edit_operation := 1;
          oper_emptr(msg_ptr);
          edit_operation := 0;

          action_code := edit_exitcode - 1;

        END;

      (*---------------------------------------------------------------------*)
      (* If necessary, mark the message as read                              *)
      (*---------------------------------------------------------------------*)

      read_sw := (action_code = 3) OR (action_code = 4) OR (action_code = 7);

      IF (active_tcb^.uid_data.user_id = msg_ptr^.msg_i_mb.msg_to)
                                         AND NOT (review_cmd AND read_sw) THEN
        BEGIN;
          msg_ptr^.msg_i_mb.msg_flag :=
                  msg_ptr^.msg_i_mb.msg_flag
                  OR mf_read
                  AND (NOT (mf_fwd_select OR mf_fwd_process));
          update_msg(msg_ptr);
          IF review_cmd THEN
            active_tcb^.curr_msg := msg_ptr^;
          read_sw := TRUE;
        END
      ELSE
        read_sw := FALSE;

      (*---------------------------------------------------------------------*)
      (* If this is not a review, we are done now                            *)
      (*---------------------------------------------------------------------*)

      IF NOT review_cmd THEN
        EXIT;

      (*---------------------------------------------------------------------*)
      (* Come here with a new action code for a message                      *)
      (*---------------------------------------------------------------------*)

msg_action_loop:

      (*---------------------------------------------------------------------*)
      (* Execute proper action on the message                                *)
      (*---------------------------------------------------------------------*)

      fl := msg_ptr^.msg_i_mb.msg_flag;

      CASE action_code OF

        (*-------------------------------------------------------------------*)
        (* Leave                                                             *)
        (*-------------------------------------------------------------------*)

        1: BEGIN;
             IF read_sw THEN
               send_message(message_flag_change)
             ELSE
               send_message(message_no_change);
             EXIT;
           END;

        (*-------------------------------------------------------------------*)
        (* Reply                                                             *)
        (*-------------------------------------------------------------------*)

        2: BEGIN;

             STR(msg_ptr^.msg_i_mb.msg_number, cmd_string);
             cmd_string := 'SR ' + cmd_string;

             active_tcb^.tcb_no_search_rel := TRUE;
             send_msg_cmd(cmd_string);
             active_tcb^.tcb_no_search_rel := FALSE;

             active_tcb^.curr_msg := msg_ptr^;
             line_prompt;
             GOTO msg_action_loop;

           END;

        (*-------------------------------------------------------------------*)
        (* Quit                                                              *)
        (*-------------------------------------------------------------------*)

        3: BEGIN;
             send_message(message_op_halted);
             active_tcb^.error_sw := TRUE;
             EXIT;
           END;

        (*-------------------------------------------------------------------*)
        (* Kill/Change                                                       *)
        (*-------------------------------------------------------------------*)

        4, 7: BEGIN;

             STR(msg_ptr^.msg_i_mb.msg_number, cmd_string);

             IF action_code = 4 THEN
               BEGIN;
                 cmd_string := 'K ' + cmd_string;
                 kill_msg_cmd(cmd_string);
               END
             ELSE
               BEGIN;
                 save_msg_no := msg_ptr^.msg_i_mb.msg_number;
                 cmd_string := 'EM ' + cmd_string;
                 oper_em(cmd_string);
                 msg_ptr := find_msg(save_msg_no);
                 test_pointer(msg_ptr);
               END;

             active_tcb^.curr_msg := msg_ptr^;
             IF (action_code = 4) AND (NOT active_tcb^.error_sw) THEN
               EXIT;

             active_tcb^.error_sw := FALSE;
             line_prompt;
             GOTO msg_action_loop;

           END;

        (*-------------------------------------------------------------------*)
        (* Release                                                           *)
        (*-------------------------------------------------------------------*)

        5: fl := fl AND NOT (mf_review OR mf_hold OR mf_kill);

        (*-------------------------------------------------------------------*)
        (* Hold                                                              *)
        (*-------------------------------------------------------------------*)

        6: BEGIN;
             fl := (fl AND (NOT mf_review)) OR mf_hold;
             msg_ptr^.msg_i_mb.msg_reason := message_reason_review_h;
           END;

      END;

      (*---------------------------------------------------------------------*)
      (* Log the action                                                      *)
      (*---------------------------------------------------------------------*)

      STR(action_code, err_string);
      log_data_is(msg_ptr^.msg_i_mb.msg_number, 'REVIEW ' + err_string);

      (*---------------------------------------------------------------------*)
      (* Update the message                                                  *)
      (*---------------------------------------------------------------------*)

      msg_ptr^.msg_i_mb.msg_flag := fl;
      update_msg(msg_ptr);

      (*---------------------------------------------------------------------*)
      (* Tell user                                                           *)
      (*---------------------------------------------------------------------*)

      active_tcb^.curr_msg := msg_ptr^;
      send_message(message_flag_change);

      (*---------------------------------------------------------------------*)
      (* Update routing pointer as needed                                    *)
      (*---------------------------------------------------------------------*)

      IF msg_route_num > msg_ptr^.msg_i_mb.msg_number THEN
        msg_route_num := msg_ptr^.msg_i_mb.msg_number;

    END; (*----- End read/review this message subroutine --------------------*)

  (*=========================================================================*)
  (* Main line of read message command                                       *)
  (*=========================================================================*)

  BEGIN;

    send_header := FALSE;
    send_x400   := TRUE;

    (*-----------------------------------------------------------------------*)
    (* Init                                                                  *)
    (*-----------------------------------------------------------------------*)

    log_data_s('L' + cmd_string);
    upcase_str_var(cmd_string);

    (*-----------------------------------------------------------------------*)
    (* Review command?                                                       *)
    (*-----------------------------------------------------------------------*)

    review_cmd := cmd_string[1] = 'V';

    (*-----------------------------------------------------------------------*)
    (* Execute right routine for this read                                   *)
    (*-----------------------------------------------------------------------*)

    word_count := WORDS(cmd_string);

    exec_char  := UPCASE(cmd_string[2]);

    IF (NOT review_cmd) AND ((exec_char = ' ') OR (exec_char = 'H')) THEN
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Read a single message (with or with/out headers)                  *)
        (*-------------------------------------------------------------------*)

        (*-------------------------------------------------------------------*)
        (* To header or not to header?  That is the question                 *)
        (*-------------------------------------------------------------------*)

        send_header := exec_char = 'H';

        (*-------------------------------------------------------------------*)
        (* Handle read without a number                                      *)
        (*-------------------------------------------------------------------*)

        IF word_count = 1 THEN
          BEGIN;

            (*---------------------------------------------------------------*)
            (* Ask for a number and get answer                               *)
            (*---------------------------------------------------------------*)

            send_message(message_rmc_num);

            cmd_string := read_tnc_data_str;

            strip_crlf(cmd_string);

            (*---------------------------------------------------------------*)
            (* If nothing send back, quit                                    *)
            (*---------------------------------------------------------------*)

            IF LENGTH(cmd_string) = 0 THEN
              BEGIN;
                send_message(message_nodata_can);
                EXIT;
              END;

            (*---------------------------------------------------------------*)
            (* If improper then quit                                         *)
            (*---------------------------------------------------------------*)

            IF LENGTH(cmd_string) > 253 THEN
              BEGIN;
                send_message(message_err_wrd);
                active_tcb^.error_sw := TRUE;
                EXIT;
              END;

            (*---------------------------------------------------------------*)
            (* Now we fake the read command                                  *)
            (*---------------------------------------------------------------*)

            cmd_string := 'R ' + cmd_string;
            word_count := WORDS(cmd_string);

          END;

        (*-------------------------------------------------------------------*)
        (* Validate number                                                   *)
        (*-------------------------------------------------------------------*)

        upcase_str_var(cmd_string);

        check_multiple_msg(@cmd_string, 2, word_count);
        IF active_tcb^.error_sw THEN
          EXIT;

        (*-------------------------------------------------------------------*)
        (* Loop thru all the messages                                        *)
        (*-------------------------------------------------------------------*)

        r_msg_no := get_next_multiple_msg;

        WHILE r_msg_no <> 0 DO
          BEGIN;

            (*---------------------------------------------------------------*)
            (* Fetch message pointer.  Give error if we can't find it        *)
            (*---------------------------------------------------------------*)

            msg_ptr := find_msg(r_msg_no);

            IF msg_ptr = NIL THEN
              BEGIN;

                (*-----------------------------------------------------------*)
                (* Message not found!                                        *)
                (*-----------------------------------------------------------*)

                active_tcb^.error_sw := TRUE;

                IF word_count = 1 THEN
                  BEGIN;
                    send_message(message_rmc_nf);
                    EXIT;
                  END
                ELSE
                  BEGIN;
                    STR(r_msg_no, err_string);
                    set_dollar1_parm(@err_string);
                    send_message(message_rmc_nf_wp);
                  END;

              END

            ELSE

              BEGIN;

                (*-----------------------------------------------------------*)
                (* Message found so read it                                  *)
                (*-----------------------------------------------------------*)

                read_this_one;
                IF active_tcb^.error_sw THEN EXIT;

              END;

            (*---------------------------------------------------------------*)
            (* Get next message                                              *)
            (*---------------------------------------------------------------*)

            r_msg_no := get_next_multiple_msg;

          END; (*----- End read multiple loop -------------------------------*)

        EXIT;

      END; (*----- End READ a certain message -------------------------------*)

    (*-----------------------------------------------------------------------*)
    (* Read everything else                                                  *)
    (*-----------------------------------------------------------------------*)

    (*-----------------------------------------------------------------------*)
    (* RM is a special case                                                  *)
    (*-----------------------------------------------------------------------*)

    IF exec_char = 'M' THEN
      BEGIN;

        IF word_count > 1 THEN
          BEGIN;
            send_message(message_err_wrd);
            active_tcb^.error_sw := TRUE;
            EXIT;
          END;

        cmd_string[2] := 'U';
        cmd_string := cmd_string + ' ' + active_tcb^.uid_data.user_id;

      END;

    (*-----------------------------------------------------------------------*)
    (* Build search block                                                    *)
    (*-----------------------------------------------------------------------*)

    cmd_string := COPY(cmd_string, 2, 255);
    set_search(cmd_string, @search_block);

    IF active_tcb^.error_sw THEN
      EXIT;

    (*-----------------------------------------------------------------------*)
    (* Start loop                                                            *)
    (*-----------------------------------------------------------------------*)

    search_msg(@search_block);

    IF search_block.search_last = NIL THEN
      BEGIN;
        send_message(message_lmc_nf);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Read the message we just found then loop searching/reading            *)
    (*-----------------------------------------------------------------------*)

    REPEAT

      msg_ptr := search_block.search_last;

      read_this_one;
      IF active_tcb^.error_sw THEN EXIT;

      search_msg(@search_block);

    UNTIL search_block.search_last = NIL;

    send_message(message_op_complete);

  END;

(*===========================================================================*)
(* Read a single message                                                     *)
(*    Destination can be another file!                                       *)
(*===========================================================================*)

PROCEDURE read_a_msg (msg_ptr : msg_index_ptr;
                      output_dest : file_name_str; write_mode : BYTE;
                      header_sw : BOOLEAN;         x400 : BOOLEAN);

  (* Write mode = 0  -- New file only      *)
  (*              1  -- Replace            *)
  (*              2  -- New file or append *)

  CONST
    x400_wide     = 10;
    out_max_print = 0;
    out_max_other = 10;

  VAR
    c           : CHAR;
    err         : BOOLEAN;
    file_id     : file_name_str;
    i           : INTEGER;
    in_head     : BOOLEAN;
    in_line     : STRING;
    out_counter : BYTE;
    out_max     : BYTE;
    out_to_file : BOOLEAN;
    r           : BYTE;
    t           : LONGINT;
    text_buffer : STRING[254];
    to_file     : TEXT;
    t1_str      : STRING;
    t2_str      : STRING;

  (*=========================================================================*)
  (* Subroutine to clean up resources so we can quit                         *)
  (*=========================================================================*)

  PROCEDURE clean_up;
    BEGIN;

      IF out_to_file THEN
        BEGIN;

          get_semaphore(semaphore_interrupts, sem_exclusive, FALSE);
          {$I-}
          CLOSE(to_file);
          {$I+}
          i := IORESULT;
          free_semaphore(semaphore_interrupts);

        END;

      t1_str := close_text_file;

    END;

  (*=========================================================================*)
  (* Read a line from the message                                            *)
  (*=========================================================================*)

  PROCEDURE read_a_line;

    BEGIN;

      WITH active_tcb^.io_fe^ DO
        BEGIN;

          READ(fe_text, text_buffer);

          IF EOLN(fe_text) THEN
            BEGIN;
              READLN(fe_text);
              in_line := text_buffer + cr;
            END
          ELSE
            in_line := text_buffer;

        END;

    END;

  (*=========================================================================*)
  (* Put a line to the destination                                           *)
  (*=========================================================================*)

  PROCEDURE put_a_line(this: STRING);

    CONST
      printer_codes = [159, 160];

    LABEL
      out_loop;

    BEGIN;

      (*---------------------------------------------------------------------*)
      (* Output to the user                                                  *)
      (*---------------------------------------------------------------------*)

      IF NOT out_to_file THEN
        BEGIN;
          send_tnc_data_str(this);
          err := more_check;
          EXIT;
        END;

      (*---------------------------------------------------------------------*)
      (* Output to a file!                                                   *)
      (*---------------------------------------------------------------------*)

out_loop:

      (*---------------------------------------------------------------------*)
      (* Output the line.  Add an LF if last character is a cr. Save the     *)
      (* error code for later                                                *)
      (*---------------------------------------------------------------------*)

      {$I-}
      IF this[LENGTH(this)] = cr THEN
        WRITE(to_file, this, lf)
      ELSE
        WRITE(to_file, this);
      {$I+}
      r := IORESULT;

      (*---------------------------------------------------------------------*)
      (* Add any needed delay                                                *)
      (*---------------------------------------------------------------------*)

      IF out_counter > out_max THEN
        BEGIN;
          out_counter := 0;
          task_switch;
        END
      ELSE
        INC(out_counter);

      (*---------------------------------------------------------------------*)
      (* If no errors, we are done                                           *)
      (*---------------------------------------------------------------------*)

      IF r = 0 THEN
        EXIT;

      (*---------------------------------------------------------------------*)
      (* Error occurred if we get here                                       *)
      (*---------------------------------------------------------------------*)

      (*---------------------------------------------------------------------*)
      (* If printer error and local user, let him fix it                     *)
      (*---------------------------------------------------------------------*)

      IF (out_max = out_max_print)
                       AND (r IN printer_codes)
                       AND ((active_tcb^.tcb_type = th_operator)
                               OR (active_tcb^.tcb_type = th_fwd_control)) THEN     (* Forward control thread      *)
        BEGIN;
          window_write_critical(active_tcb^.port_chan_s + 'R:',
                       'Printer error -- Alt-A to kill printer');
          task_wait(60, active_tcb^.tcb_type = th_operator);
          IF NOT kill_printer THEN
            GOTO out_loop;
        END;

      (*---------------------------------------------------------------------*)
      (* Printer error                                                       *)
      (*---------------------------------------------------------------------*)

      err := TRUE;
      clean_up;
      active_tcb^.error_sw := TRUE;
      send_tnc_data_str(dos_err_message(r) + cr);
      EXIT;

    END;

  (*=========================================================================*)
  (* Main line of READ A MESSAGE                                             *)
  (*=========================================================================*)

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Setup                                                                 *)
    (*-----------------------------------------------------------------------*)

    err := FALSE;

    out_max := out_max_other;

    active_tcb^.curr_msg := msg_ptr^;

    log_data_is(active_tcb^.curr_msg.msg_i_mb.msg_number, 'R');

    (*-----------------------------------------------------------------------*)
    (* Activate the output file if we have one                               *)
    (*-----------------------------------------------------------------------*)

    out_to_file := output_dest <> '';

    IF out_to_file THEN
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Slow down printer                                                 *)
        (*-------------------------------------------------------------------*)

        IF match_str(output_dest, printer_match) THEN
          out_max := out_max_print;

        (*-------------------------------------------------------------------*)
        (* Obtain the interrupt lock                                         *)
        (*-------------------------------------------------------------------*)

        get_semaphore(semaphore_interrupts, sem_exclusive, FALSE);

        (*-------------------------------------------------------------------*)
        (* Assign the file name                                              *)
        (*-------------------------------------------------------------------*)

        ASSIGN(to_file, output_dest);

        {$I-}
        RESET(to_file);
        {$I+}

        i := IORESULT;

        IF (write_mode = 0) AND (i = 0) THEN
          BEGIN;
            active_tcb^.error_sw := TRUE;
            free_semaphore(semaphore_interrupts);
            clean_up;
            send_message(message_file_exists);
            EXIT;
          END;

        {$I-}
        IF (write_mode = 2) AND (i = 0) THEN
          APPEND(to_file)
        ELSE
          REWRITE(to_file);
        {$I+}

        i := IORESULT;

        free_semaphore(semaphore_interrupts);

        IF i <> 0 THEN
          BEGIN;
            active_tcb^.error_sw := TRUE;
            clean_up;
            send_tnc_data_str(cr + dos_err_message(i) + cr);
            EXIT;
          END;

        out_counter := 0;

      END;

    (*-----------------------------------------------------------------------*)
    (* Send the message header line                                          *)
    (*-----------------------------------------------------------------------*)

    IF NOT (out_to_file OR x400) THEN
      BEGIN;
        send_msg_header(1);
        put_a_line(header_msg_block(msg_ptr, 1) + cr);
        IF err THEN EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Send X400 header                                                      *)
    (*-----------------------------------------------------------------------*)

    IF x400 THEN
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Message from                                                      *)
        (*-------------------------------------------------------------------*)

        t1_str := get_message(message_msg_from)
                                     + ' ' + msg_ptr^.msg_i_mb.msg_from;

        IF msg_ptr^.msg_i_mb.msg_from_at <> '' THEN
          BEGIN;
            t1_str := t1_str + ' @ ' + msg_ptr^.msg_i_mb.msg_from_at;
            IF msg_ptr^.msg_i_mb.msg_from_h <> '' THEN
              t1_str := t1_str + '.' + msg_ptr^.msg_i_mb.msg_from_h;
          END;

        put_a_line(t1_str + cr);
        IF err THEN EXIT;

        (*-------------------------------------------------------------------*)
        (* Message to                                                        *)
        (*-------------------------------------------------------------------*)

        t1_str := get_message(message_msg_to) + ' '
                                                  + msg_ptr^.msg_i_mb.msg_to;
        IF msg_ptr^.msg_i_mb.msg_to_at <> '' THEN
          t1_str := t1_str + ' @ ' + msg_ptr^.msg_i_mb.msg_to_at;

        IF msg_ptr^.msg_i_mb.msg_to_h <> '' THEN
          t1_str := t1_str + '.' + msg_ptr^.msg_i_mb.msg_to_h
        ELSE
          IF msg_ptr^.msg_i_mb.msg_to_at = opt_block.this_bb_sign THEN
            t1_str := t1_str + '.' + opt_block.this_bb_h;

        put_a_line(t1_str + cr);
        IF err THEN EXIT;

        (*-------------------------------------------------------------------*)
        (* Origination date time stamp                                       *)
        (*-------------------------------------------------------------------*)

        put_a_line(get_message(message_msg_date) + ' '
                         + time_str(msg_ptr^.msg_i_mb.msg_dt_orig, TRUE) + cr);
        IF err THEN EXIT;

        (*-------------------------------------------------------------------*)
        (* Message id                                                        *)
        (*-------------------------------------------------------------------*)

        t1_str := get_message(message_msg_msgid) + ' ';

        IF msg_ptr^.msg_i_mb.msg_type <> ' ' THEN
          t1_str := t1_str + msg_ptr^.msg_i_mb.msg_type;

        STR(msg_ptr^.msg_i_mb.msg_no_orig, t2_str);
        t1_str := t1_str + msg_flag_char(msg_ptr^.msg_i_mb.msg_flag) + ' '
                              + t2_str + '@' + msg_ptr^.msg_i_mb.msg_from_at;

        IF msg_ptr^.msg_i_mb.msg_from_at <> opt_block.this_bb_addr THEN
          BEGIN;
            STR(msg_ptr^.msg_i_mb.msg_number, t2_str);
            t1_str := t1_str + ', ' + t2_str + '@' + opt_block.this_bb_addr;
          END;

        IF msg_ptr^.msg_i_mb.msg_bid <> '' THEN
          t1_str := t1_str + ' $' + msg_ptr^.msg_i_mb.msg_bid;

        put_a_line(t1_str + cr);
        IF err THEN EXIT;

        (*-------------------------------------------------------------------*)
        (* Message subject                                                   *)
        (*-------------------------------------------------------------------*)

        put_a_line(get_message(message_msg_subj)
                                    + ' ' + msg_ptr^.msg_i_mb.msg_subj + cr);
        IF err THEN EXIT;

      END;

    (*-----------------------------------------------------------------------*)
    (* Find the file and open it                                             *)
    (*-----------------------------------------------------------------------*)

    STR(msg_ptr^.msg_i_mb.msg_number, file_id);
    file_id := opt_block.msg_file_dir + 'BB' + file_id + '.MSG';

    t2_str := open_text_file(file_id, TRUE);

    IF t2_str <> '' THEN
      BEGIN;

        active_tcb^.error_sw := TRUE;
        clean_up;
        put_a_line(t2_str + cr);

        EXIT;

      END;

    WITH active_tcb^.io_fe^ DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* If we are not showing headers, compute the path                   *)
        (*-------------------------------------------------------------------*)

        IF NOT header_sw THEN
          BEGIN;

            in_head := TRUE;
            in_line := '';
            t2_str  := '';

            (*---------------------------------------------------------------*)
            (* Loop reading header                                           *)
            (*---------------------------------------------------------------*)

            WHILE (NOT EOF(fe_text)) AND in_head DO
              BEGIN;

                (*-----------------------------------------------------------*)
                (* Read a line                                               *)
                (*-----------------------------------------------------------*)

                read_a_line;

                (*-----------------------------------------------------------*)
                (* Validity check                                            *)
                (*   1.  Must be long enough                                 *)
                (*   2.  Must have an @                                      *)
                (*   3.  @ must have room for a call                         *)
                (*   4.  Must start with R:                                  *)
                (*-----------------------------------------------------------*)

                i := POS('@', in_line);
                IF (LENGTH(in_line) < 10) OR (i = 0)
                                                 OR (i > (LENGTH(in_line) - 4))
                            OR (in_line[1] <> 'R') OR (in_line[2] <> ':') THEN
                  in_head := FALSE
                ELSE
                  BEGIN;

                    (*-------------------------------------------------------*)
                    (* Move pointer after call.  Compensate for @:           *)
                    (*-------------------------------------------------------*)

                    INC(i);
                    IF in_line[i] = ':' THEN
                      INC(i);

                    (*-------------------------------------------------------*)
                    (* Reduce down one word                                  *)
                    (*-------------------------------------------------------*)

                    t1_str := COPY(in_line, i, 255);
                    t1_str := subword(@t1_str, 1, 1);

                    (*-------------------------------------------------------*)
                    (* Scan forward for a non-alphameric that ends the call  *)
                    (*-------------------------------------------------------*)

                    i := 0;
                    REPEAT
                      INC(i);
                      IF i <= LENGTH(t1_str) THEN
                        c := t1_str[i]
                      ELSE
                        c := ' ';
                    UNTIL ((c <  'A') OR (c > 'Z'))
                                                 AND ((c <  '0') OR (c > '9'));

                    (*-------------------------------------------------------*)
                    (* If the size is ok, append the call sign to the path   *)
                    (*-------------------------------------------------------*)

                    IF i > 1 THEN
                      t2_str := t2_str + + '!' + COPY(t1_str, 1, i-1)
                    ELSE
                      in_head := FALSE

                  END;

              END; (*----- End loop for headers -----------------------------*)

            (*---------------------------------------------------------------*)
            (* If a path has been found, display it                          *)
            (*---------------------------------------------------------------*)

            IF t2_str <> '' THEN
              BEGIN;
                t2_str[1] := ' ';
                put_a_line(get_message(message_msg_path) + t2_str + cr);
                IF err THEN EXIT;
              END;

            (*---------------------------------------------------------------*)
            (* Display a blank line then the first non-header line           *)
            (*---------------------------------------------------------------*)

            put_a_line(cr);
            IF err THEN EXIT;
            put_a_line(in_line);
            IF err THEN EXIT;

          END;

        (*-------------------------------------------------------------------*)
        (* Loop thru file, sending what we have                              *)
        (*-------------------------------------------------------------------*)

        WHILE NOT EOF(fe_text) DO
          BEGIN;
            read_a_line;
            put_a_line(in_line);
            IF err THEN EXIT;
          END;

        (*-------------------------------------------------------------------*)
        (* If outputting to a file, correct day time stamp else put trail    *)
        (* message                                                           *)
        (*-------------------------------------------------------------------*)

        IF out_to_file THEN
          BEGIN;
            GETFTIME(fe_text, t);
            SETFTIME(to_file, t);
          END
        ELSE
          send_message(message_trail);

        clean_up;

      END;

  END;

END.
