(*===========================================================================*)
(* Semaphore control -- Part II                                              *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991 by H. Roy Engehausen.  All rights      *)
(*   reserved.                                                               *)
(*                                                                           *)
(*===========================================================================*)

{$DEFINE POINT_CHK}
{$UNDEF DEBUG_2} (* Find hang in get semaphore *)

UNIT BBSEMA2;

INTERFACE

  USES
    bbdummy;

CONST
  semaphore_fwd_route_use   = 1;
  semaphore_fwd_route_build = 2;
  semaphore_interrupts      = 3;
  semaphore_msg_list        = 4;
  semaphore_action_list     = 5;
  semaphore_port_start      = 6;

  semaphore_max   = 16;

  time_out_long   = 720;
  time_out_short  =  60;

TYPE
  semaphore_type  = (sem_exclusive, sem_shared, sem_clear);

  sem_process_ptr  = ^sem_process_type;

  sem_process_type = RECORD
                       next_process  : sem_process_ptr;
                       process_tcb   : tcb_ptr;
                       process_who   : POINTER;
                     END;

  sem_info_type    = RECORD
                       process_list  : sem_process_ptr;
                       excl_count    : BYTE;
                       sem_status    : semaphore_type;
                       time_out      : WORD;
                       last_unlock   : POINTER;
                     END;

VAR
  sem_info_array : ARRAY[1..semaphore_max] OF sem_info_type;
  high_semaphore : BYTE;

PROCEDURE process_free_semaphore(this_tcb : tcb_ptr);
PROCEDURE free_sema_tcb (sem_num : BYTE; this_tcb : tcb_ptr; nf_err : BOOLEAN);
PROCEDURE get_semaphore (sem_num          : BYTE;
                         type_get         : semaphore_type;
                         long_time_out_sw : BOOLEAN);

PROCEDURE free_semaphore (sem_num : BYTE);
FUNCTION  count_semaphore(sem_num : BYTE) : BYTE;
FUNCTION  test_semaphore (sem_num : BYTE) : semaphore_type;
FUNCTION  test_semaphore_tcb(sem_num : BYTE; type_test : semaphore_type)
                                                                  : BOOLEAN;

IMPLEMENTATION

USES
  CRT,
  bbbug,
  bbdump,
  bbmisc3,
  bbtask,
  bbtime;

(*===========================================================================*)
(* Free all semaphores for a task                                            *)
(*===========================================================================*)

PROCEDURE process_free_semaphore(this_tcb : tcb_ptr);
  VAR
    i : BYTE;

  BEGIN;
    FOR i := 1 TO semaphore_max DO
      free_sema_tcb(i, this_tcb, FALSE);
  END;

(*===========================================================================*)
(* Free a semaphores for a task                                              *)
(*===========================================================================*)

PROCEDURE free_sema_tcb(sem_num : BYTE; this_tcb : tcb_ptr; nf_err : BOOLEAN);

  VAR
    old_proc_l  : sem_process_ptr;
    prev_proc_l : sem_process_ptr;
    this_proc_l : sem_process_ptr;
    w_proc_l    : sem_process_ptr;

  BEGIN;

    WITH sem_info_array[sem_num] DO
      BEGIN;

        w_proc_l    := process_list;
        old_proc_l  := NIL;
        this_proc_l := NIL;

        (*-------------------------------------------------------------------*)
        (* Find this process's entry in the semaphore list                   *)
        (*-------------------------------------------------------------------*)

        WHILE (w_proc_l <> NIL) AND (this_proc_l = NIL) DO
          BEGIN;
            IF w_proc_l^.process_tcb = this_tcb THEN
              BEGIN;
                this_proc_l := w_proc_l;
                prev_proc_l := old_proc_l;
              END;
            old_proc_l := w_proc_l;
            w_proc_l   := w_proc_l^.next_process;
          END;

        (*-------------------------------------------------------------------*)
        (* No entry!                                                         *)
        (*-------------------------------------------------------------------*)

        IF this_proc_l = NIL THEN
          IF nf_err THEN
            BEGIN;
              WRITELN('Clearing a semaphore not held by task');
              semaphore_bug(sem_num, TRUE);
            END
          ELSE
            EXIT;

        (*-------------------------------------------------------------------*)
        (* Debug                                                             *)
        (*-------------------------------------------------------------------*)

        {$IFDEF POINT_CHK}
          test_pointer(this_proc_l);
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Unchain from list                                                 *)
        (*-------------------------------------------------------------------*)

        IF prev_proc_l = NIL THEN
          process_list := this_proc_l^.next_process
        ELSE
          BEGIN;
            {$IFDEF POINT_CHK}
              test_pointer(prev_proc_l);
            {$ENDIF}
            prev_proc_l^.next_process := this_proc_l^.next_process;
          END;

        (*-------------------------------------------------------------------*)
        (* If no one else on list then clear                                 *)
        (*-------------------------------------------------------------------*)

        IF process_list = NIL THEN
          sem_status   := sem_clear;

        (*-------------------------------------------------------------------*)
        (* Debug                                                             *)
        (*-------------------------------------------------------------------*)

        {$IFDEF POINT_CHK}
          IF process_list <> NIL THEN
            test_pointer(process_list);
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Free it                                                           *)
        (*-------------------------------------------------------------------*)

        DISPOSE(this_proc_l);

        (*-------------------------------------------------------------------*)
        (* Track it                                                          *)
        (*-------------------------------------------------------------------*)

        last_unlock := called_by(2);

      END;

  END; (*----- End free_sema_tcb --------------------------------------------*)

(*===========================================================================*)
(* Get a semaphores for a task                                               *)
(*===========================================================================*)

PROCEDURE get_semaphore (sem_num          : BYTE;
                         type_get         : semaphore_type;
                         long_time_out_sw : BOOLEAN);

  VAR
    excl_wait    : BYTE;
    i            : INTEGER;
    time_delta   : LONGINT;
    times_up     : LONGINT;
    w_tcb        : tcb_ptr;
    w_proc_l     : sem_process_ptr;

  BEGIN;

    {$IFDEF DEBUG_1}
      WRITELN('Semaphore startup');
      DELAY(300);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Make sure we ain't setting a clear!                                   *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG_1}
      WRITELN('Semaphore clear test');
      DELAY(300);
    {$ENDIF}

    IF type_get = sem_clear THEN
      BEGIN;
        WRITELN('Attempt to get a clear semaphore');
        semaphore_bug(sem_num, TRUE);
      END;

    (*-----------------------------------------------------------------------*)
    (* Check the number!                                                    *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG_1}
      WRITELN('Semaphore num test');
      DELAY(300);
    {$ENDIF}

    IF (sem_num < 1) OR (sem_num > semaphore_max) THEN
      BEGIN;
        WRITELN('Invalid semaphore number');
        semaphore_bug(sem_num, TRUE);
      END;


    WITH sem_info_array[sem_num] DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Set the time out as needed                                        *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_1}
          WRITELN('Semaphore timeout set');
          DELAY(300);
        {$ENDIF}

        IF sem_status = sem_clear THEN
          BEGIN;

            IF long_time_out_sw THEN
              time_out := time_out_long
            ELSE
              time_out := time_out_short;

          END;

        (*-------------------------------------------------------------------*)
        (* Make sure we are not getting a semaphore that we have             *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_1}
          WRITELN('task test');
          DELAY(300);
        {$ENDIF}

        IF sem_status <> sem_clear THEN
          BEGIN;
            w_proc_l := process_list;
            WHILE w_proc_l <> NIL DO
              BEGIN;

                {$IFDEF DEBUG_1}
                  WRITELN('task loop');
                  DELAY(300);
                {$ENDIF}

                IF w_proc_l^.process_tcb = active_tcb THEN
                  BEGIN;
                    WRITELN('Attempt to obtain a semaphore we already have');
                    semaphore_bug(sem_num, TRUE)
                  END;
                w_proc_l   := w_proc_l^.next_process
              END;
          END;

        (*-------------------------------------------------------------------*)
        (* If we want a shared semaphore and an exclusive is waiting, delay  *)
        (* is needed to prevent the exclusive from never obtaining the       *)
        (* semaphore                                                         *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_1}
          WRITELN('Excl delay');
          DELAY(300);
        {$ENDIF}

        IF (excl_count > 0) AND (sem_status = sem_shared) AND
                                                   (type_get = sem_shared) THEN
          BEGIN;

            excl_wait := excl_count;
            WHILE (excl_wait > 0) AND (sem_status = sem_shared) DO
              BEGIN;
                DEC(excl_wait);
                task_switch;
              END;

            task_switch;                (* This is to make sure the process  *)
            task_switch;                (* waiting for exclusive get's it    *)

          END;

        (*-------------------------------------------------------------------*)
        (* If the semaphore is exclusive, wait for it change status          *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_1}
          WRITELN('Excl wait');
          DELAY(300);
        {$ENDIF}


        IF sem_status = sem_exclusive THEN
          BEGIN;

            (*---------------------------------------------------------------*)
            (* Compute the time out time                                     *)
            (*---------------------------------------------------------------*)

            times_up := up_time_from_now(time_out);

            (*---------------------------------------------------------------*)
            (* Loop until we can gain control                                *)
            (*---------------------------------------------------------------*)

            REPEAT;

              {$IFDEF DEBUG_1}
                WRITELN('Excl time loop');
                DELAY(300);
              {$ENDIF}

              (*-------------------------------------------------------------*)
              (* Remember what time we started this check                    *)
              (*-------------------------------------------------------------*)

              time_delta := up_time;

              (*-------------------------------------------------------------*)
              (* Switch away                                                 *)
              (*-------------------------------------------------------------*)

              task_switch;

              (*-------------------------------------------------------------*)
              (* Compute how long we were away                               *)
              (*-------------------------------------------------------------*)

              time_delta := up_time - time_delta;

              (*-------------------------------------------------------------*)
              (* If we were away a long time then compensate for that        *)
              (*-------------------------------------------------------------*)

              IF up_time > 4 * up_ticks_per_sec THEN
                times_up := times_up + time_delta;

              (*-------------------------------------------------------------*)
              (* Check to see if time has expired                            *)
              (*-------------------------------------------------------------*)

              IF times_up < up_time THEN
                BEGIN;
                  WRITELN('Loop on exclusive semaphore status');
                  semaphore_bug(sem_num, TRUE);
                END;

            UNTIL sem_status <> sem_exclusive;

          END;

        (*-------------------------------------------------------------------*)
        (* If we want exclusive, we must wait for it to clear.               *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_1}
          WRITELN('Excl clear wait');
          DELAY(300);
        {$ENDIF}

        IF (type_get = sem_exclusive) THEN
          BEGIN;

            excl_count := 60;   (*----- Number of switches others will wait -*)

            IF sem_status <> sem_clear THEN
              BEGIN;

                times_up := up_time_from_now(time_out);

                REPEAT;
                  task_switch;

                  IF times_up < up_time THEN
                    BEGIN;
                      WRITELN('Loop on exclusive semaphore request');
                      semaphore_bug(sem_num, TRUE);
                    END;

                UNTIL sem_status = sem_clear;

              END;

            IF process_list <> NIL THEN
              BEGIN;
                WRITELN('Exclusive drop thru with something on list');
                semaphore_bug(sem_num, TRUE);
              END;

            excl_count := 0;    (*----- Let everyone thru from now on -------*)

          END;

        {$IFDEF DEBUG_1}
          WRITELN('Semaphore allocate');
          DELAY(300);
          dump_string('Semaphore allocate');
          dump_all;
          times_up := 0;
          REPEAT
            INC(times_up);
          UNTIL times_up > $FFFFFF;
        {$ENDIF}

        sem_status := type_get;

        NEW(w_proc_l);

        w_proc_l^.next_process := process_list;
        w_proc_l^.process_tcb  := active_tcb;
        w_proc_l^.process_who  := called_by(1);
        process_list           := w_proc_l;

        {$IFDEF DEBUG_1}
          WRITELN('Semaphore done allocate');
          DELAY(300);
        {$ENDIF}

      END;

  END;

(*===========================================================================*)
(* Free a semaphore                                                          *)
(*===========================================================================*)

PROCEDURE free_semaphore(sem_num : BYTE);
  BEGIN;

    free_sema_tcb(sem_num, active_tcb, TRUE);
    task_switch;

  END;

(*===========================================================================*)
(* Count processes holding a seamphore                                       *)
(*===========================================================================*)

FUNCTION count_semaphore(sem_num : BYTE) : BYTE;
  VAR
    countr   : BYTE;
    w_proc_l : sem_process_ptr;

  BEGIN;

    CASE sem_info_array[sem_num].sem_status OF
      sem_clear     : count_semaphore := 0;
      sem_exclusive : count_semaphore := 1;
      sem_shared    : WITH sem_info_array[sem_num] DO
                        BEGIN;

                          countr := 0;
                          w_proc_l   := process_list;

                          WHILE w_proc_l <> NIL DO
                            BEGIN;
                                INC(countr);
                                w_proc_l := w_proc_l^.next_process
                            END;

                          count_semaphore := countr;

                        END;

    END;

  END;

(*===========================================================================*)
(* See if a semaphore is set                                                 *)
(*===========================================================================*)

FUNCTION test_semaphore(sem_num : BYTE) : semaphore_type;
  BEGIN;

    test_semaphore := sem_info_array[sem_num].sem_status;

  END;

(*===========================================================================*)
(* See if a semaphore is held in a certain state by this task                *)
(*===========================================================================*)

FUNCTION test_semaphore_tcb(sem_num : BYTE; type_test : semaphore_type)
                                                                  : BOOLEAN;

  VAR
    b        : BOOLEAN;
    w_proc_l : sem_process_ptr;

  BEGIN;

    WITH sem_info_array[sem_num] DO
      BEGIN;

        b := type_test = sem_status;

        IF b THEN
          BEGIN;
            w_proc_l   := process_list;

            WHILE w_proc_l <> NIL DO
              BEGIN;
                IF w_proc_l^.process_tcb = active_tcb THEN
                  BEGIN;
                    test_semaphore_tcb := TRUE;
                    EXIT;
                  END;
                w_proc_l   := w_proc_l^.next_process
              END;
          END;

      END;

    test_semaphore_tcb := FALSE;

  END;

END.
