(*===========================================================================*)
(* Check setup routines                                                      *)
(*                                                                           *)
(*   Copyright 1988, 1989 by H. Roy Engehausen.  All rights reserved.        *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Check to see if this port is all set                                      *)
(*===========================================================================*)

PROCEDURE check_port;

  VAR
    b1       : BOOLEAN;
    b2       : BOOLEAN;
    look_ptr : port_block_ptr;

  BEGIN;

    err_save := err_sw;
    err_sw   := FALSE;

    WITH this_port^ DO
      BEGIN;
        IF port_type = port_unknown THEN
          err_line('Port type not set');
        IF com_number = 0 THEN
          err_line('COM number not set');
        IF (port_type <> port_pcpa) AND (data_rate = 0) THEN
          err_line('Speed not set');
        IF max_conn = 255 THEN
          err_line('Maximum connects not set');
        IF max_chan = 255 THEN
          err_line('Maximum channels not set');
        IF max_conn >= max_chan THEN
          err_line('Maximum channels not greater than maximum connects');
        IF time_out = 0 THEN
          err_line('Time out not set');
        IF fwd_min = 65535 THEN
          err_line('Forward time not set');
        IF port_pend = 0 THEN
          err_line('Max pending frames not set');
        IF max_chan = 0 THEN
          err_line('Maximum channels not set');
        IF port_sema <> 0 THEN
          BEGIN;
            look_ptr := port_chain;
            WHILE (look_ptr <> NIL) AND (look_ptr^.com_number <> port_sema) DO
              look_ptr := look_ptr^.next_port;
            IF look_ptr = NIL  THEN
              err_line('Cannot find port with interlock com number');
          END;

        IF (dflt_lang <> #0)
                         AND (POS(dflt_lang, opt_block.language_list) = 0) THEN
          err_line('This port has an invalid default language');

        b1 := attr_did_this IN port_parm_array[port_parm_ring].parm_attr;

        b2 := attr_did_this IN port_parm_array[port_parm_call].parm_attr;

        b2 := (attr_did_this IN port_parm_array[port_parm_bcst].parm_attr)
                       OR b2;

        IF b1 AND b2 THEN
    err_line('Mutually exclusive modem and non-modem parms set for this port');

      END;

    err_sw := err_sw OR err_save;

  END;

(*===========================================================================*)
(* Check to see if this file is all set                                      *)
(*===========================================================================*)

PROCEDURE check_fsb;
  BEGIN;

    err_save := err_sw;
    err_sw   := FALSE;

    WITH this_fsb^ DO
      BEGIN;
        IF fsb_name = '' THEN
          err_line('File Directory Name not set');
        IF fsb_path = '' THEN
          err_line('File Directory Path not set');
        IF fsb_path = 'ANY' THEN
          fsb_path := '';
      END;

    err_sw := err_sw OR err_save;

  END;
