(*===========================================================================*)
(* Setup process lines                                                       *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Process the port line.                                                    *)
(*===========================================================================*)

PROCEDURE process_port_line;
  VAR
    cm        : BYTE;
    look_port : port_block_ptr;

  BEGIN;

    IF (p_name = 'ENDPORT') AND (w_count = 1) THEN
      BEGIN;

        this_port^ := work_port;
        check_port;

        this_port^.dflt_order := this_port^.dflt_order + '   ';

        IF this_port^.port_type = port_modem THEN
          BEGIN;
            this_port^.port_monitor := FALSE;
            this_port^.port_bcst    := FALSE;
            this_port^.cr_timeout   := 0;
            this_port^.max_conn     := 1;
            this_port^.max_chan     := 1;
          END;

        IF this_port^.dflt_pac = 0 THEN
          this_port^.dflt_pac := this_port^.max_pac;

        port_flag := FALSE;

        IF (this_port^.port_type <> port_pcpa)
                               AND (this_port^.port_type <> port_bpqhost) THEN
          EXIT;

        WITH this_port^ DO
          BEGIN;
            look_port := next_port;
            cm        := com_number;
          END;

        WHILE look_port <> NIL DO
          WITH look_port^ DO
            BEGIN;

              IF (port_type = port_pcpa) AND (com_number = cm) THEN
                BEGIN;
                  IF (port_num = this_port^.port_num) THEN
                    BEGIN;
                      err_line('Two PC*PA ports defined the same');
                      EXIT;
                    END;
                  rel_port := this_port;
                  this_port^.rel_port := look_port;
                  EXIT;
                END;

              IF (port_type = port_bpqhost) AND (com_number = cm) THEN
                BEGIN;
                  IF (port_num = this_port^.port_num) THEN
                    BEGIN;
                      err_line('Two BPQ Host ports defined the same');
                      EXIT;
                    END;
                  rel_port := this_port;
                  this_port^.rel_port := look_port;
                  EXIT;
                END;

              look_port := next_port;

            END;

        EXIT;

      END; (*----- End process END PORT -------------------------------------*)

    get_the_parm(@port_parm_array);

  END;

(*===========================================================================*)
(* Process the file line                                                     *)
(*===========================================================================*)

PROCEDURE process_file_line;
  BEGIN;

    IF (p_name = 'ENDFILE') AND (w_count = 1) THEN
      BEGIN;

        this_fsb^ := work_fsb;
        check_fsb;
        fsb_flag := FALSE;
        EXIT;

      END;

    get_the_parm(@file_parm_array);

  END;

(*===========================================================================*)
(* Process the line                                                          *)
(*===========================================================================*)

PROCEDURE process_line;
  VAR
    i : BYTE;

  BEGIN;

    IF LENGTH(in_line) = 0 THEN EXIT;

    w_count := WORDS(in_line);

    IF w_count < 1 THEN
      BEGIN;
        err_line('Word count wrong');
        EXIT;
      END;

    p_name  := subword(@in_line, 1, 1);
    p_value := subword(@in_line, 3, 0);

    upcase_str_var(p_name);

    IF port_flag THEN
      BEGIN;
        process_port_line;
        EXIT;
      END;

    IF fsb_flag THEN
      BEGIN;
        process_file_line;
        EXIT;
      END;

    IF (p_name = 'PORT') AND (w_count = 1) THEN
      BEGIN;

        port_flag    := TRUE;

        FOR i := 1 TO parm_max DO
          WITH port_parm_array[i] DO
            parm_attr := parm_attr - [attr_did_this];

        FILLCHAR(work_port, SIZEOF(work_port), CHR(0));

        WITH work_port DO
          BEGIN;
            next_port  := NIL;
            rel_port   := NIL;
            com_number := 0;
            port_char  := CHR(port_cnt + ORD('A'));
            port_type  := port_unknown;
            port_sema  := 0;
            max_pac    := 128;
            data_rate  := 0;
            max_conn   := 255;
            max_chan   := 255;
            call_list  := NIL;
            port_color := default_data_color;
            fwd_min    := 65535;
            disc_delay := 10;
            dflt_order := '';
          END;

        NEW(this_port);

        IF port_chain = NIL THEN
          port_chain := this_port
        ELSE
          port_last_chn^.next_port := this_port;

        port_last_chn := this_port;

        port_cnt := port_cnt + 1;

        EXIT;
      END;

    IF (p_name = 'FILE_DIRECTORY') AND (w_count = 1) THEN
      BEGIN;

        fsb_flag     := TRUE;

        FOR i := 1 TO parm_max DO
          WITH file_parm_array[i] DO
            parm_attr := parm_attr - [attr_did_this];

        FILLCHAR(work_fsb, SIZEOF(work_fsb), CHR(0));

        WITH work_fsb DO
          next_fsb := fsb_base;

        NEW(this_fsb);

        fsb_base := this_fsb;

        fd_cnt   := fd_cnt + 1;

        EXIT;

      END;

    get_the_parm(@main_parm_array);

  END;
