(*===========================================================================*)
(* Procedure to send/receive things a G8BPQ Node in BPQ Host Mode            *)
(*                                                                           *)
(*   Copyright 1991, 1992 by H. Roy Engehausen.  All rights reserved         *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

{$DEFINE POINT_CHK}
{$UNDEF  DEBUG_1} (* Debug I/O calls *)
{$UNDEF  DEBUG_2} (* Debug send data *)

UNIT BBSRTBPQ;

INTERFACE

  PROCEDURE send_recv_tnc_BPQ(tnc_cmd_data : BYTE);

  (*-------------------------------------------------------------------------*)
  (* Monitor filter settings                                                 *)
  (*    I = .......1   -- I frames                                           *)
  (*    U = ......1.   -- UI frames                                          *)
  (*    S = .....1..   -- S frames                                           *)
  (*    F = ....1...   -- PID F0 frames                                      *)
  (*    C = ...1....   -- PID CF frames                                      *)
  (*    A = ..1.....   -- PID frames other than CF or F0                     *)
  (*-------------------------------------------------------------------------*)

  CONST
    bpq_monitor_i     = $01;
    bpq_monitor_ui    = $02;
    bpq_monitor_s     = $04;
    bpq_monitor_F0    = $08;
    bpq_monitor_CF    = $10;
    bpq_monitor_other = $20;

IMPLEMENTATION

USES
  DOS,
  bbax25,
  bbbug,
  bbdummy,
  bbdump,
  bbmisc3,
  bbmisc6,
  bbrunerr,
  bbstr,
  bbtask,
  bbtrace;

(*===========================================================================*)
(* Procedure to send/receive things to the tnc                               *)
(*    Parm = 0 -- Data                                                       *)
(*           1 -- Command                                                    *)
(*           2 -- General Poll                                               *)
(*           3 -- Data Poll                                                  *)
(*           4 -- Link Status Poll                                           *)
(*           5 -- Issue JHOST1                                               *)
(*           6 -- Force JHOST1                                               *)
(*                                                                           *)
(*    This parm is put in TNC_CMD_DATA.  The unit internally modifies        *)
(*    this to other states                                                   *)
(*           7 -- Ack status change                                          *)
(*           8 -- Connect.  Get call sign                                    *)
(*           9 -- Get status counds from link                                *)
(*          10 -- Monitor data                                               *)
(*          11 -- 'L' command.  Goes to state 9                              *)
(*          12 -- Set UI packet                                              *)
(*===========================================================================*)

PROCEDURE send_recv_tnc_BPQ(tnc_cmd_data : BYTE);

  (*-------------------------------------------------------------------------*)
  (* Local variables                                                         *)
  (*-------------------------------------------------------------------------*)

  VAR
    call_buffer    : STRING[10];
    cmd_char       : CHAR;
    i              : BYTE;
    l_link_pending : CHAR;
    l_link_state   : CHAR;
    tnc_registers  : REGISTERS;

    data_place     : RECORD
                       CASE byte OF
                         0 : (data_ptr : str_mixed_ptr);
                         1 : (data_off : WORD;
                              data_seg : WORD);
                     END;

    r_bsize        : WORD;
    ui_len         : WORD;


  LABEL
    tnc_loop;

  (*=========================================================================*)
  (* Command interpreter subroutine                                          *)
  (*=========================================================================*)

  {$I BBSRTBPC.PAS} (* Monitor *)

  (*=========================================================================*)
  (* Monitor subroutine                                                      *)
  (*=========================================================================*)

  {$I BBSRTBPM.PAS} (* Monitor *)

  (*=========================================================================*)
  (* Send UI frame subroutine                                                *)
  (*=========================================================================*)

  PROCEDURE prep_ui_frame;

    CONST
      b_size = SIZEOF(address_over) + 10;

    TYPE
      t = ARRAY[0..512] OF BYTE;

    VAR
      i         : BYTE;
      p         : ^t;

    BEGIN;
      p := POINTER(active_tcb^.tnc_htt);

      {$IFDEF POINT_CHK}
        test_pointer(p);
      {$ENDIF}

      build_ax25_header(active_port^.bcst_path, @p^[0], i);

      MOVE(data_place.data_ptr^.long_data,
           p^[i],
           data_place.data_ptr^.long_length);

      ui_len := i + data_place.data_ptr^.long_length;

    END;

  (*=========================================================================*)
  (* Set null response subroutine                                            *)
  (*=========================================================================*)

  PROCEDURE set_null;
    BEGIN;
      WITH active_tcb^ DO
        BEGIN;
          tnc_null             := TRUE;
          tnc_type             := t_to_h_ok;
          tnc_data.long_length := 0;
          tnc_data.str_data    := '';
        END;
    END;

  (*=========================================================================*)
  (* Main line                                                               *)
  (*=========================================================================*)

  BEGIN;

    {$IFDEF DEBUG_1}
      WRITELN('BPQ IO');
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Initialize some variables                                             *)
    (*-----------------------------------------------------------------------*)

    data_place.data_ptr := @active_tcb^.tnc_data;

    (*-----------------------------------------------------------------------*)
    (* This is where we loop to on a TNC busy or to redrive I/O              *)
    (*-----------------------------------------------------------------------*)

tnc_loop:

    (*-----------------------------------------------------------------------*)
    (* States                                                                *)
    (*    0 = Send data to switch                                            *)
    (*    1 = Send command to switch                                         *)
    (*        "L" command -- goes to state 11                                *)
    (*    2 = General poll -- Monitor channel goes to 10                     *)
    (*    3 = Status poll                                                    *)
    (*    4 = Status poll                                                    *)
    (*    5 = Status poll                                                    *)
    (*    6 = Status poll                                                    *)
    (*    7 = Ack status change                                              *)
    (*    8 = Get connecting station call sign                               *)
    (*    9 = Get link count                                                 *)
    (*   10 = Monitor recv data                                              *)
    (*   11 = L command                                                      *)
    (*   12 = Bad command                                                    *)
    (*-----------------------------------------------------------------------*)

    (*-----------------------------------------------------------------------*)
    (* Prevent execution if requested.  Always switch at least once          *)
    (*-----------------------------------------------------------------------*)

    REPEAT
      task_switch;
    UNTIL NOT active_tcb^.tcb_stop_tnc_io;

    (*-----------------------------------------------------------------------*)
    (* Get channel number                                                    *)
    (*-----------------------------------------------------------------------*)

    tnc_registers.AL := active_tcb^.channel + 1;

    {$IFDEF DEBUG_1}
      WRITELN('BPQ channel - ', tnc_registers.AL , ' - ', tnc_cmd_data);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Execute state information                                             *)
    (*-----------------------------------------------------------------------*)

    CASE tnc_cmd_data OF

      (*---------------------------------------------------------------------*)
      (* Send data to switch                                                 *)
      (*---------------------------------------------------------------------*)

      0 : BEGIN;

            IF (active_tcb^.channel <> 0)
                                         OR (active_port^.port_num = '0') THEN
              BEGIN;
                tnc_registers.AH := bpq_send_frame;
                tnc_registers.ES := data_place.data_seg;
                tnc_registers.SI := OFS(data_place.data_ptr^.long_data);
                tnc_registers.CX := data_place.data_ptr^.long_length;
              END
            ELSE
              BEGIN;
                prep_ui_frame;
                tnc_registers.AH := bpq_send_kiss;
                tnc_registers.AL := ORD(active_port^.port_num) - ORD('0');
                tnc_registers.ES := SEG(active_tcb^.tnc_htt^);
                tnc_registers.SI := OFS(active_tcb^.tnc_htt^);
                tnc_registers.CX := ui_len;
              END;


            {$IFDEF DEBUG_2}
              WRITELN('BPQ SDATA - AL = ', tnc_registers.AL,
                                 ' AH = ', tnc_registers.AH,
                                 ' CX = ', tnc_registers.CX,
                                 ' -- ', LENGTH(data_place.data_ptr^.str_data),
                                 ' -- ', data_place.data_ptr^.str_data);
            {$ENDIF}

          END;

      (*---------------------------------------------------------------------*)
      (* Command                                                             *)
      (*---------------------------------------------------------------------*)

      1 : BEGIN;

            cmd_char := UPCASE(data_place.data_ptr^.str_data[1]);

            CASE cmd_char OF

              (*-------------------------------------------------------------*)
              (* Link status                                                 *)
              (*-------------------------------------------------------------*)

              'L' : BEGIN;
                      tnc_cmd_data     := 11;
                      tnc_registers.AH := bpq_status;
                    END;

              (*-------------------------------------------------------------*)
              (* Disconnect                                                  *)
              (*-------------------------------------------------------------*)

              'D' : BEGIN;
                      tnc_registers.AH := bpq_cmd;
                      tnc_registers.CX := bpq_cmd_disconnect;
                    END;

              (*-------------------------------------------------------------*)
              (* Connect                                                     *)
              (*-------------------------------------------------------------*)

              'C' : BEGIN;
                      tnc_registers.AH := bpq_cmd;
                      tnc_registers.CX := bpq_cmd_connect;
                    END;

              (*-------------------------------------------------------------*)
              (* Monitor ON/OFF                                              *)
              (*-------------------------------------------------------------*)

              'M' : BEGIN;

                      i := bpq_monitor_command;
                      IF i = 2 THEN
                        tnc_cmd_data := 12
                      ELSE
                        BEGIN;
                          tnc_registers.AH := bpq_open;
                          tnc_registers.AL := 1;
                          tnc_registers.CL := bpq_appl_flag_sendcon_user
                                                  + bpq_appl_flag_sendcon_appl;

                          IF i > 0 THEN
                            tnc_registers.CL := tnc_registers.CL
                                                      OR bpq_appl_mask_monitor;
                          tnc_registers.DL := 0;
                        END;
                    END;

              (*-------------------------------------------------------------*)
              (* Anything else                                               *)
              (*-------------------------------------------------------------*)

              ELSE
                tnc_cmd_data := 12;

            END; (*----- End CASE for command type --------------------------*)

          END; (*----- End handling commands to tnc -------------------------*)

      (*---------------------------------------------------------------------*)
      (* Poll general/status -- This does a data poll now.  If this is a     *)
      (* general poll, we will change to status later                        *)
      (*---------------------------------------------------------------------*)

      2,
      3 : BEGIN;

            (*---------------------------------------------------------------*)
            (* Set up proper poll                                            *)
            (*---------------------------------------------------------------*)

            IF active_tcb^.channel <> 0 THEN
              BEGIN;

                (*-----------------------------------------------------------*)
                (* Data poll for regular channel                             *)
                (*-----------------------------------------------------------*)

                tnc_registers.AH := bpq_recv_frame;
                tnc_registers.ES := data_place.data_seg;
                tnc_registers.DI := OFS(data_place.data_ptr^.long_data);
                tnc_registers.CX := SIZEOF(data_place.data_ptr^.long_data);
              END
            ELSE
              BEGIN;

                (*-----------------------------------------------------------*)
                (* Data poll for monitor channel                             *)
                (*-----------------------------------------------------------*)

                IF active_port^.mon_filter =  0 THEN
                  BEGIN;
                    set_null;
                    EXIT;
                  END;

                tnc_cmd_data     := 10;
                tnc_registers.AH := bpq_monitor;
                tnc_registers.ES := SEG(active_tcb^.tnc_tth^);
                tnc_registers.DI := OFS(active_tcb^.tnc_tth^);
                tnc_registers.CX := active_tcb^.tnc_b_size;

              END;

            (*---------------------------------------------------------------*)
            (* Save size of receive buffer                                   *)
            (*---------------------------------------------------------------*)

            r_bsize := tnc_registers.CX;

          END;

      (*---------------------------------------------------------------------*)
      (* Poll status                                                         *)
      (*---------------------------------------------------------------------*)

      4 : tnc_registers.AH := bpq_status;

      (*---------------------------------------------------------------------*)
      (* Ack status change (Internal to this UNIT)                           *)
      (*---------------------------------------------------------------------*)

      7 : tnc_registers.AH := bpq_ack_status;

      (*---------------------------------------------------------------------*)
      (* Get call sign for connect (Internal to this UNIT)                   *)
      (*---------------------------------------------------------------------*)

      8 : BEGIN;
            tnc_registers.AH := bpq_get_call;
            tnc_registers.ES := SEG(call_buffer[1]);
            tnc_registers.DI := OFS(call_buffer[1]);
          END;

      (*---------------------------------------------------------------------*)
      (* Get the link counts (Internal to this UNIT)                         *)
      (*---------------------------------------------------------------------*)

      9 : tnc_registers.AH := bpq_get_counts;

      (*---------------------------------------------------------------------*)
      (* HOST switch stuff is a NOP                                          *)
      (*---------------------------------------------------------------------*)

      5,
      6 : WITH active_tcb^ DO
            BEGIN;
              set_null;
              EXIT;
            END;

    END; (*----- End of CASE on type of transaction -------------------------*)

    (*-----------------------------------------------------------------------*)
    (* Send the request                                                      *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG_1}
      WRITELN('BPQ IOREQ - AL = ', tnc_registers.AL,
                         ' AH = ', tnc_registers.AH,
                         ' CX = ', tnc_registers.CX,
                         ' DX = ', tnc_registers.DX);
    {$ENDIF}

    signal_place^ := $0200 + LO(signal_place^);

    IF tnc_cmd_data <> 12 THEN
      INTR(active_port^.com_number, tnc_registers);

    signal_place^ := $F800 + LO(signal_place^);

    {$IFDEF DEBUG_1}
      WRITELN('BPQ IODONE - AL = ', tnc_registers.AL,
                         ' AH = ', tnc_registers.AH,
                         ' CX = ', tnc_registers.CX,
                         ' DX = ', tnc_registers.DX);
      WRITELN('BPQ new state = ', tnc_cmd_data);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Interpret response based on command type                              *)
    (*-----------------------------------------------------------------------*)

    CASE tnc_cmd_data OF

      (*---------------------------------------------------------------------*)
      (* Link status counts returned.  Build result from "L" command         *)
      (*      Word 1 -- Number of link status messages waiting               *)
      (*      Word 2 -- Number of data packets waiting                       *)
      (*      Word 3 -- Number of packets awaiting 1st transmission          *)
      (*      Word 4 -- Number of frames send at least once                  *)
      (*      Word 5 -- Number of retries                                    *)
      (*      Word 6 -- Link status #                                        *)
      (*---------------------------------------------------------------------*)

      9 : BEGIN;

            active_tcb^.tnc_type := t_to_h_okmsg;

            (*---------------------------------------------------------------*)
            (* Convert pending receive buffers to characters                 *)
            (*---------------------------------------------------------------*)

            STR(tnc_registers.BX, call_buffer);

            (*---------------------------------------------------------------*)
            (* Start to build the output string                              *)
            (*---------------------------------------------------------------*)

            data_place.data_ptr^.str_data := l_link_pending + ' '
                                             + call_buffer + ' ';

            (*---------------------------------------------------------------*)
            (* Check number of buffers left in switch.  If low then force    *)
            (* a large pending count to ensure no busy.  Add to output       *)
            (* string.                                                       *)
            (*---------------------------------------------------------------*)

            IF tnc_registers.DX < 25 THEN
              call_buffer := '99'
            ELSE
              STR(tnc_registers.CX, call_buffer);

            (*---------------------------------------------------------------*)
            (* Finish output string.  Number of frames send once and number  *)
            (* retries are both always zero on this type port                *)
            (*---------------------------------------------------------------*)

            data_place.data_ptr^.str_data := data_place.data_ptr^.str_data
                                             + call_buffer + ' 0 0 '
                                             + l_link_state;

            data_place.data_ptr^.long_length :=
                                         LENGTH(data_place.data_ptr^.str_data);

            IF tnc_registers.CX <> 0 THEN
              trace_data('BPQPend', tnc_registers.CX, NIL,
                                         data_place.data_ptr^.str_data);

          END;

      (*---------------------------------------------------------------------*)
      (* L  Command                                                          *)
      (*---------------------------------------------------------------------*)

      11: BEGIN;

            (*---------------------------------------------------------------*)
            (* Save state and pending condition and get counts               *)
            (*---------------------------------------------------------------*)

            l_link_pending := CHR(tnc_registers.DX + ORD('0'));
            l_link_state   := CHR(tnc_registers.CX + ORD('0'));
            tnc_cmd_data := 9;
            GOTO tnc_loop;

          END; (*----- End handling 'L' command -----------------------------*)

      (*---------------------------------------------------------------------*)
      (* Send data, commands (but not 'L')                                   *)
      (*---------------------------------------------------------------------*)

      0,
      1 : set_null;

      (*---------------------------------------------------------------------*)
      (* Poll general, data                                                  *)
      (*---------------------------------------------------------------------*)

      2,
      3 : BEGIN;

            (*---------------------------------------------------------------*)
            (* If no data and a general poll, then we do a status poll.  This*)
            (* is because the actual poll for data doesn't return status     *)
            (* If no data and a true data poll, then return with null        *)
            (*---------------------------------------------------------------*)

            IF (tnc_registers.CX = 0) THEN
              BEGIN;
                IF (active_tcb^.channel <> 0) AND (tnc_cmd_data = 2) THEN
                  BEGIN;
                    tnc_cmd_data := 4;
                    GOTO tnc_loop;
                  END;
                set_null;
                EXIT;
              END;

            (*---------------------------------------------------------------*)
            (* Check the size                                                *)
            (*---------------------------------------------------------------*)

            IF r_bsize < tnc_registers.CX THEN
              BEGIN;
                WRITELN;
                WRITELN('***** BPQ Switch returned a packet that is ');
                WRITELN('***** larger than the buffer size.         ');
                WRITELN('***** Increase BPQ_BUFF in PARMS.BB        ');
                WRITELN;
                WRITELN('***** Current setting is ', r_bsize);
                WRITELN('***** Size received was ', tnc_registers.CX);
                WRITELN;
                RUNERROR(bpq_overrun);
              END;

            (*---------------------------------------------------------------*)
            (* Set the right answer code                                     *)
            (*---------------------------------------------------------------*)

            IF active_tcb^.channel <> 0 THEN
              active_tcb^.tnc_type := t_to_h_conn
            ELSE
              active_tcb^.tnc_type := t_to_h_mif;

            (*---------------------------------------------------------------*)
            (* Set the data length                                           *)
            (*---------------------------------------------------------------*)

            data_place.data_ptr^.long_length := tnc_registers.CX;
            IF tnc_registers.CX < 255 THEN
              data_place.data_ptr^.str_data[0] := CHR(tnc_registers.CX)
            ELSE
              data_place.data_ptr^.str_data[0] := CHR(255);

          END;

      (*---------------------------------------------------------------------*)
      (* Poll status                                                         *)
      (*---------------------------------------------------------------------*)

      4 : BEGIN;

            (*---------------------------------------------------------------*)
            (* If no change then answer null                                 *)
            (*---------------------------------------------------------------*)

            IF tnc_registers.DX = 0 THEN
              BEGIN;
                set_null;
                EXIT
              END;

            (*---------------------------------------------------------------*)
            (* If connect then we need call                                  *)
            (*---------------------------------------------------------------*)

            IF tnc_registers.CX = bpq_status_connect THEN
              BEGIN;
                tnc_cmd_data := 8;
                GOTO tnc_loop;
              END;

            (*---------------------------------------------------------------*)
            (* Only thing left is disconnect so set up buffers               *)
            (*---------------------------------------------------------------*)

            active_tcb^.tnc_type := t_to_h_links;
            data_place.data_ptr^.str_data := t_to_h_disc + 'SWITCH';
            data_place.data_ptr^.long_length :=
                                         LENGTH(data_place.data_ptr^.str_data);

            (*---------------------------------------------------------------*)
            (* Now go ack the change                                         *)
            (*---------------------------------------------------------------*)

            tnc_cmd_data := 7;
            GOTO tnc_loop;

          END;

      (*---------------------------------------------------------------------*)
      (* Monitor data                                                        *)
      (*---------------------------------------------------------------------*)

      10: BEGIN;

            (*---------------------------------------------------------------*)
            (* If no data then answer null                                   *)
            (*---------------------------------------------------------------*)

            IF tnc_registers.CX = 0 THEN
              BEGIN;
                set_null;
                EXIT
              END;

            (*---------------------------------------------------------------*)
            (* Set the result                                                *)
            (*---------------------------------------------------------------*)

            bpq_monitor_convert;

            active_tcb^.tnc_type := t_to_h_mh_noi;

          END;

      (*---------------------------------------------------------------------*)
      (* Ack status change (Internal to this UNIT) -- Don't touch anything   *)
      (* Its all set up by someone else                                      *)
      (*---------------------------------------------------------------------*)

      7 : ;

      (*---------------------------------------------------------------------*)
      (* Get call sign for connect                                           *)
      (*---------------------------------------------------------------------*)

      8 : BEGIN;

            (*---------------------------------------------------------------*)
            (* Remove trailing blanks from call sign which is left justified *)
            (* in the buffer                                                 *)
            (*---------------------------------------------------------------*)

            i := SIZEOF(call_buffer) - 1;
            WHILE (i > 0) AND (call_buffer[i] = ' ') DO
              DEC(i);
            call_buffer[0] := CHR(i);

            (*---------------------------------------------------------------*)
            (* If no call sign, make it switch                               *)
            (*---------------------------------------------------------------*)

            IF call_buffer = '' THEN
              call_buffer := 'SWITCH';

            (*---------------------------------------------------------------*)
            (* Set buffers                                                   *)
            (*---------------------------------------------------------------*)

            active_tcb^.tnc_type := t_to_h_links;
            data_place.data_ptr^.str_data := t_to_h_ct + call_buffer;
            data_place.data_ptr^.long_length :=
                                         LENGTH(data_place.data_ptr^.str_data);


            (*---------------------------------------------------------------*)
            (* Now go ack the change                                         *)
            (*---------------------------------------------------------------*)

            tnc_cmd_data := 7;
            GOTO tnc_loop;

          END; (*----- End handling of call sign ----------------------------*)

      (*---------------------------------------------------------------------*)
      (* Bad command                                                         *)
      (*---------------------------------------------------------------------*)

      12 : WITH active_tcb^ DO
             BEGIN;
               tnc_null             := FALSE;
               tnc_type             := t_to_h_badmsg;
               tnc_data.str_data    := t_to_h_badcmd;
               tnc_data.long_length := LENGTH(t_to_h_badcmd);
             END;

      ELSE
        BEGIN;
          WRITELN('Bad command code in BPQ HOST -- ', tnc_cmd_data);
          HALT;
        END;

    END; (*----- End of CASE on type of transaction -------------------------*)

    (*-----------------------------------------------------------------------*)
    (* Set null response if appropriate                                      *)
    (*-----------------------------------------------------------------------*)

    active_tcb^.tnc_null := active_tcb^.tnc_data.long_length = 0;

    {$IFDEF DEBUG_1}
      WRITELN('BPQ exit - ', active_tcb^.tnc_null,
                      ' - ', active_tcb^.tnc_type,
                      ' - ', active_tcb^.tnc_data.long_length,
                      ' - ', active_tcb^.tnc_data.str_data);

    {$ENDIF}

  END; (*----- End main send/receive TNC ------------------------------------*)

END.
