(*===========================================================================*)
(* Procedure to send/receive things to the PC*PA card                        *)
(*                                                                           *)
(*   Copyright 1988, 1989 by H. Roy Engehausen.  All rights reserved.        *)
(*   This software may be freely distributed and used, but it may not        *)
(*   under any circumstances be sold by anyone other than the author.        *)
(*   It may be distributed by a commercial company as long as it is          *)
(*   for no cost.                                                            *)
(*                                                                           *)
(*===========================================================================*)

PROCEDURE send_pcpa_mode;

  {$UNDEF DEBUG_ME}         (* Display PC*PA debugging info                  *)

  CONST

    rcvr_lockup    = 30; (* Number of receives done per time slice           *)
    xmit_pause     = 8;  (* A delay factor to prevent overrunning the        *)
                         (* TNCTSR code.  Requested by DRSI                  *)

  TYPE
    over_array = ARRAY[1..512] OF BYTE;

  VAR
    byte_array     : ^over_array;
    byte_count     : WORD;
    byte_pos       : WORD;
    int_no         : BYTE;
    rcvr_cnt       : BYTE;
    times_up       : LONGINT;

  BEGIN;

{$IFDEF DEBUG_ME}
WINDOW(1,1,80,25);
{$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Introduce a delay as per DRSI                                         *)
    (*-----------------------------------------------------------------------*)

    FOR int_no := 0 TO xmit_pause DO
      task_switch;

    (*-----------------------------------------------------------------------*)
    (* Set the interrupt number                                              *)
    (*-----------------------------------------------------------------------*)

    int_no := active_port^.com_number;

    (*-----------------------------------------------------------------------*)
    (* Ok.  Get addressing for transmit                                      *)
    (*-----------------------------------------------------------------------*)

    WITH active_tcb^.tnc_htt^, tnc_registers DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Get ready to transmit                                             *)
        (*-------------------------------------------------------------------*)

        byte_count := data_count + 4;
        byte_array := @channel;
        byte_pos   := 1;

        (*-------------------------------------------------------------------*)
        (* Loop for transmit                                                 *)
        (*-------------------------------------------------------------------*)

        WHILE byte_count > 0 DO
          BEGIN;

            AH := 1;
            AL := byte_array^[byte_pos];

{$IFDEF DEBUG_ME}
WRITELN('Send--', AL);
{$ENDIF}

            signal_place^ := $0100 + LO(signal_place^);

            INTR(int_no, tnc_registers);

            signal_place^ := $7800 + LO(signal_place^);

{$IFDEF DEBUG_ME}
WRITELN('Send<<', AX);
{$ENDIF}

            IF AH <> 0 THEN
              window_write_critical_i('Non-zero AH from PC*PA on send -- '
                                                    + active_tcb^.port_chan_s
                                                    + ' -- ' + CHR(AL)
                                                    + '--', AL);

            DEC(byte_count);
            INC(byte_pos);

          END; (*----- End transmit loop ------------------------------------*)

      END;

    (*-----------------------------------------------------------------------*)
    (* If this is master task then time check                                *)
    (*-----------------------------------------------------------------------*)

    IF master_thread THEN
      time_check;

    (*-----------------------------------------------------------------------*)
    (* If QRES then leave now.                                               *)
    (*-----------------------------------------------------------------------*)

    IF qres_resp THEN
      EXIT;

    (*-----------------------------------------------------------------------*)
    (* Ok.  Get addressing for receive                                       *)
    (*-----------------------------------------------------------------------*)

    WITH active_tcb^.tnc_tth^, tnc_registers DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Get ready to receive the first two characters.  They are constant *)
        (*-------------------------------------------------------------------*)

        byte_array := @channel;

        (*-------------------------------------------------------------------*)
        (* Loop reading the data                                             *)
        (*-------------------------------------------------------------------*)

        times_up := up_time + time_out_value;

        rcvr_cnt := 0;

        WHILE TRUE DO
          BEGIN;

            (*---------------------------------------------------------------*)
            (* Loop until buffer has something in it                         *)
            (*---------------------------------------------------------------*)

            REPEAT

              IF rcvr_cnt = 0 THEN
                BEGIN;
                  task_switch;
                  rcvr_cnt := rcvr_lockup;
                END
              ELSE
                DEC(rcvr_cnt);

              AX := 0;
              INTR(int_no, tnc_registers);

              signal_place^ := $F800 + LO(signal_place^);

{$IFDEF DEBUG_ME}
WRITELN('RECV--', AH, '.', AL);
{$ENDIF}

              IF AH = 0 THEN
                BEGIN;
                  task_switch;
                  rcvr_cnt := rcvr_lockup;

                  IF master_thread THEN
                    time_check;

                  IF (times_up < up_time) AND (AH = 0) THEN
                    BEGIN;
                      window_write_critical_i('Timeout PCPA loop -- '
                                                    + active_tcb^.port_chan_s
                                                    + ' -- Count = ',
                                               byte_count);
                      IF byte_count <= 5 THEN
                        WHILE byte_count <= 10 DO
                          BEGIN;
                            INC(byte_count);
                            byte_array^[byte_count] := $0;
                          END;
                      EXIT;
                    END;
                END;

            UNTIL AH <> 0; (*----- End loop for buffer ready ----------------*)

            INC(byte_count);
            byte_array^[byte_count] := AL;

            IF byte_count >= 2 THEN
              BEGIN;

                IF tnc_code = 0 THEN EXIT;

                (*-----------------------------------------------------------*)
                (* If this is a null terminated response and it is a null,   *)
                (* then we must be done                                      *)
                (*-----------------------------------------------------------*)

                IF (byte_array^[byte_count] = 0) AND (tnc_code < 6) THEN
                  EXIT;

                (*-----------------------------------------------------------*)
                (* Watch for weirdos                                         *)
                (*-----------------------------------------------------------*)

                IF byte_count > 260 THEN
                  BEGIN;
                    WRITELN;
                    WRITELN('Ser I/O Crash');
                    WRITELN('Byte cnt = ', byte_count);
                    WRITELN('AL = ', AL);
                    WRITELN('Type = ', tnc_code);
                    HALT;
                  END;

                (*-----------------------------------------------------------*)
                (* If this is a data count response and the count is correct *)
                (* then we must be done                                      *)
                (*-----------------------------------------------------------*)

                IF (byte_count >= (data67_count + 4)) AND (tnc_code >= 6)
                                                       AND (byte_count > 2) THEN
                  EXIT;

              END;

          END; (*----- End loop for receiving the response ------------------*)

      END;

  END;
