(*===========================================================================*)
(* *** LINKED and emergency switch processor                                 *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1991 by H. Roy Engehausen.  All rights reserved.  *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

{$UNDEF DEBUG1}

UNIT BBSTARL;

INTERFACE

  USES
    bbdummy;

  FUNCTION star_link(cmd_string : STRING) : BOOLEAN;

  PROCEDURE call_switch(VAR s: STRING);

IMPLEMENTATION

  USES
    CRT,
    bbauxm,
    bblog,
    bbmdata,
    bbmess,
    bbrdata,
    bbstr,
    bbuf;

(*===========================================================================*)
(* Process *** LINKED                                                        *)
(*===========================================================================*)

FUNCTION star_link(cmd_string : STRING) : BOOLEAN;

  VAR
    call_len      : BYTE;
    connect_call  : call_sign_str;
    connect_ssid  : ssid_sign_str;
    dash_pos      : BYTE;
    new_user      : BOOLEAN;
    subcall       : STRING[9];
    uid_i_current : user_index_ptr;
    wrd_cnt       : BYTE;

  PROCEDURE error_exit;
    BEGIN;
      send_message(message_not_on_port);
      star_link := FALSE;
    END;

  BEGIN;

    (*---------------------------------------------------------------------*)
    (* This handles either                                                 *)
    (*    *** LINKED TO AA4RE-1            (Standard format)               *)
    (*    *** LINKED TO AA4RE 1 NODE etc   (TEXNET 1 format)               *)
    (*    *** LINKED TO WB6ZVW1 NODE etc   (TEXNET 2 format)               *)
    (*    *** LINKED TO AA4RE  1 NODE etc  (TEXNET 3 format)               *)
    (*    *** LINKED TO WB6ZVW 1 NODE etc  (TEXNET 4 format)               *)
    (* For TEXNETs, note the SSID is in a fixed position                   *)
    (*---------------------------------------------------------------------*)

    (*---------------------------------------------------------------------*)
    (* Get ready to parse                                                  *)
    (*---------------------------------------------------------------------*)

    strip_crlf(cmd_string);

    upcase_str_var(cmd_string);

    subcall      := subword(@cmd_string, 4, 1);
    connect_ssid := subword(@cmd_string, 5, 1);
    call_len     := LENGTH(subcall);
    wrd_cnt      := words(cmd_string);
    dash_pos     := POS('-', subcall);

    {$IFDEF DEBUG1}
      WRITELN('SUBCALL=', LENGTH(subcall),      '=', subcall);
      WRITELN('SSID   =', LENGTH(connect_ssid), '=', connect_ssid);
      WRITELN('CALLLEN=', call_len);
      WRITELN('WRDCNT =', wrd_cnt);
      WRITELN('DASHPOS=', dash_pos);
      WRITELN('TASK   =', ORD(active_tcb^.tcb_type));
      WRITELN('T1     =', substr_compare(cmd_string, 1, '*** LINKED TO '));
      WRITELN('T1     =', active_tcb^.tcb_type <> th_user);
      DELAY(1000);
    {$ENDIF}

    (*---------------------------------------------------------------------*)
    (* Sanity checks                                                       *)
    (*   1) Less than 4 words                                              *)
    (*   2) *** LINKED TO                                                  *)
    (*   3) 4th word (callsign) must be 9 chars or less                    *)
    (*   4) 4th word (callsign) must be 3 chars or more                    *)
    (*   5) If 4 words then this is NOT TEXNET.  The 4 word is standard    *)
    (*      call-ssid format so anything longer than 6 characters must     *)
    (*      have a '-' in it.                                              *)
    (*   6) If more than 4 words then this is TEXNET.  The 4 word is       *)
    (*      may be the CALLwith SSID butted against it.  Make sure that    *)
    (*      no one has snuck in a '-'.  The 4 word can only be 7 chars max *)
    (*      and cannot be 6.  If 4th word is less than 7 then the SSID is  *)
    (*      in the 5th word and can only be 1 character                    *)
    (*   7) If we have a '-', it cannot be the first or last character     *)
    (*   8) Only "user" tasks can have *** LINKs                           *)
    (*---------------------------------------------------------------------*)

    IF (wrd_cnt < 4)
               OR (NOT substr_compare(cmd_string, 1, '*** LINKED TO '))
               OR (call_len > 9)
               OR (call_len < 3)
               OR ((wrd_cnt = 4) AND (call_len > 6) AND (dash_pos = 0))
               OR ((wrd_cnt > 4) AND ((dash_pos > 0)
                          OR (call_len > 7)

(* In an attempt to support TEXNET formats 3 & 4, the following test
   is removed
                          OR (call_len = 6)
*)

                          OR ((call_len < 7) AND (LENGTH(connect_ssid) <> 1))))
               OR (dash_pos = 1) OR (dash_pos = call_len)
               OR (active_tcb^.tcb_type <> th_user) THEN
      BEGIN;
        error_exit;
        EXIT;
      END;

    {$IFDEF DEBUG1}
      WRITELN('PARSE CALL');
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Now figure out the call                                               *)
    (*-----------------------------------------------------------------------*)

    IF dash_pos = 0 THEN
      BEGIN;

        {$IFDEF DEBUG1}
          WRITELN('DASHPOS0');
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* No '-' is present.  In the NON-TEXNET case then this is the       *)
        (* call and there is no SSID.  If this is TEXNET, then the SSID      *)
        (* is either the seventh character or the next word.                 *)
        (*                                                                   *)
        (* The connect_call assignment statement will truncate off the       *)
        (* SSID if necessary                                                 *)
        (*                                                                   *)
        (* TEXNET SSIDS are in HEX so convert to decimal                     *)
        (*-------------------------------------------------------------------*)

        connect_call := subcall;

        IF (wrd_cnt = 4) AND (call_len < 7) THEN
          connect_ssid := ''
        ELSE
          BEGIN;

            (*---------------------------------------------------------------*)
            (* TEXNET.  If WB6ZVW1 format, get the ssid out                  *)
            (*---------------------------------------------------------------*)

            IF call_len = 7 THEN
              connect_ssid := subcall[7];

            (*---------------------------------------------------------------*)
            (* Convert TEXNET SSID (hex) to decimal format                   *)
            (*---------------------------------------------------------------*)

            IF (connect_ssid >= 'A') AND (connect_ssid <= 'F') THEN
              connect_ssid :=
                   '1' + CHR((ORD(connect_ssid[1]) - ORD('A')) + ORD('0'));

            (*---------------------------------------------------------------*)
            (* Remove zero SSID                                              *)
            (*---------------------------------------------------------------*)

            IF connect_ssid = '0' THEN
              BEGIN;
                subcall      := connect_call;
                connect_ssid := '';
              END
            ELSE
              subcall := connect_call + '-' + connect_ssid;

          END;

      END
    ELSE
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* AA4RE-1 format                                                    *)
        (*-------------------------------------------------------------------*)

        connect_call := COPY(subcall, 1, dash_pos - 1);
        connect_ssid := COPY(subcall, dash_pos + 1, 0);

      END;

    (*-----------------------------------------------------------------------*)
    (* Create task name                                                      *)
    (*-----------------------------------------------------------------------*)

    IF LENGTH(subcall) > (SIZEOF(active_tcb^.tcb_name) - 2) THEN
      active_tcb^.tcb_name := subcall + 'L'
    ELSE
      active_tcb^.tcb_name := subcall + '/L';

    (*-----------------------------------------------------------------------*)
    (* Log it as neededed                                                    *)
    (*-----------------------------------------------------------------------*)

    log_data_s('L' + active_tcb^.port_chan_s + subcall
                                                 + ' ' + active_tcb^.tcb_name);

    (*-----------------------------------------------------------------------*)
    (* Get user's data                                                       *)
    (*-----------------------------------------------------------------------*)

    uid_i_current := find_uid(connect_call);

    new_user := uid_i_current = NIL;

    IF new_user THEN
      BEGIN;

        uid_init(@active_tcb^.uid_data);

        active_tcb^.uid_data.user_id  := connect_call;
        active_tcb^.uid_data.user_ssid := connect_ssid;

        add_uid(@active_tcb^.uid_data);

      END
    ELSE
      active_tcb^.uid_data := get_uid(uid_i_current)^;

    (*-----------------------------------------------------------------------*)
    (* Put in activity report                                                *)
    (*-----------------------------------------------------------------------*)

    add_mon_call(@dummy_port,  active_port^.port_char,
                                                 active_tcb^.uid_data.user_id);

    (*-----------------------------------------------------------------------*)
    (* Bump logon count                                                      *)
    (*-----------------------------------------------------------------------*)

    INC(active_tcb^.uid_data.user_l_cnt);

    (*-----------------------------------------------------------------------*)
    (* Set up the user class                                                 *)
    (*-----------------------------------------------------------------------*)

    IF NOT new_user THEN
      BEGIN;

        IF active_tcb^.uid_data.user_class = user_c_nu THEN
          active_tcb^.uid_data.user_class := user_c_uu;

        IF active_tcb^.uid_data.user_class > user_c_eu THEN
          active_tcb^.uid_data.user_class := user_c_eu;

        IF (active_tcb^.uid_data.user_flag AND user_f_sysop) <> 0 THEN
          active_tcb^.uid_data.user_class := user_c_eu;

        IF (active_tcb^.uid_data.user_flag AND user_f_bbs) <> 0 THEN
          active_tcb^.uid_data.user_class := user_c_bu;

      END;

    (*-----------------------------------------------------------------------*)
    (* Set up MAXPAC                                                         *)
    (*-----------------------------------------------------------------------*)

    active_tcb^.max_pac := active_port^.max_pac;
    IF (active_tcb^.max_pac > active_tcb^.uid_data.max_pac)
                                  AND (active_tcb^.uid_data.max_pac <> 0) THEN
      active_tcb^.max_pac := active_tcb^.uid_data.max_pac;
    IF active_tcb^.max_pac = 0 THEN
      BEGIN;
        WRITELN('MAXPAC = 0 ON *** LINKED');
        active_tcb^.max_pac := 128;
      END;

    (*-----------------------------------------------------------------------*)
    (* Finish TCB setup                                                      *)
    (*-----------------------------------------------------------------------*)

    active_tcb^.last_l_time := active_tcb^.uid_data.user_l_time;

    active_tcb^.tcb_sysop_pw_ok := FALSE;
    active_tcb^.tcb_access_ok   := FALSE;

    star_link := TRUE;

    status_window_change := TRUE;

  END;

(*===========================================================================*)
(* Carry out emergency mode call switch                                      *)
(*===========================================================================*)

PROCEDURE call_switch(VAR s: STRING);

  VAR
    old_call      : call_sign_str;
    uid_i_current : user_index_ptr;

  LABEL
    bad_call_prompt,
    good_call;

  BEGIN;

    old_call := active_tcb^.uid_data.user_id;

    REPEAT

      send_message(message_e_mode_switch);

      s := read_tnc_data_str;

      strip_crlf(s);

      upcase_str_var(s);

      IF words(s) <> 1 THEN
        GOTO bad_call_prompt;

      (*-----------------------------------------------------------------------*)
      (* Get user's data                                                       *)
      (*-----------------------------------------------------------------------*)

      uid_i_current := find_uid(s);

      IF uid_i_current = NIL THEN
        GOTO bad_call_prompt;

      active_tcb^.uid_data := get_uid(uid_i_current)^;

      IF (active_tcb^.uid_data.user_flag AND user_f_emerg) = 0 THEN
        GOTO bad_call_prompt;

      GOTO good_call;

      EXIT;

bad_call_prompt:

      send_message(message_bad_call);

    UNTIL FALSE;

good_call:

    (*-----------------------------------------------------------------------*)
    (* Log it as neededed                                                    *)
    (*-----------------------------------------------------------------------*)

    log_data_s('L' + active_tcb^.port_chan_s + s + ' ' + active_tcb^.tcb_name);

    IF LENGTH(s) > (SIZEOF(active_tcb^.tcb_name) - 2) THEN
      active_tcb^.tcb_name := s + 'L'
    ELSE
      active_tcb^.tcb_name := s + '/L';

    (*-----------------------------------------------------------------------*)
    (* Put in activity report                                                *)
    (*-----------------------------------------------------------------------*)

    add_mon_call(@dummy_port,  active_port^.port_char,
                                                 active_tcb^.uid_data.user_id);

    (*-----------------------------------------------------------------------*)
    (* Bump logon count                                                      *)
    (*-----------------------------------------------------------------------*)

    INC(active_tcb^.uid_data.user_l_cnt);

    (*-----------------------------------------------------------------------*)
    (* Set up the user class                                                 *)
    (*-----------------------------------------------------------------------*)

    IF active_tcb^.uid_data.user_class = user_c_nu THEN
      active_tcb^.uid_data.user_class := user_c_uu;

    IF active_tcb^.uid_data.user_class > user_c_eu THEN
      active_tcb^.uid_data.user_class := user_c_eu;

    IF (active_tcb^.uid_data.user_flag AND user_f_sysop) <> 0 THEN
      active_tcb^.uid_data.user_class := user_c_eu;

    IF (active_tcb^.uid_data.user_flag AND user_f_bbs) <> 0 THEN
      active_tcb^.uid_data.user_class := user_c_bu;

    (*-----------------------------------------------------------------------*)
    (* Set up MAXPAC                                                         *)
    (*-----------------------------------------------------------------------*)

    active_tcb^.max_pac := active_port^.max_pac;
    IF (active_tcb^.max_pac > active_tcb^.uid_data.max_pac)
                                  AND (active_tcb^.uid_data.max_pac <> 0) THEN
      active_tcb^.max_pac := active_tcb^.uid_data.max_pac;
    IF active_tcb^.max_pac = 0 THEN
      BEGIN;
        WRITELN('MAXPAC = 0 ON *** LINKED');
        active_tcb^.max_pac := 128;
      END;

    (*-----------------------------------------------------------------------*)
    (* Finish TCB setup                                                      *)
    (*-----------------------------------------------------------------------*)

    active_tcb^.last_l_time := active_tcb^.uid_data.user_l_time;

    active_tcb^.tcb_sysop_pw_ok := FALSE;
    active_tcb^.tcb_access_ok   := FALSE;

    active_tcb^.uid_data.user_name := old_call;

    status_window_change := TRUE;

  END;

END.
