(*===========================================================================*)
(* "Send msg to" processing                                                  *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991 by H. Roy Engehausen.  All rights      *)
(*   reserved.                                                               *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

{$UNDEF DEBUG_H} (* Debug home BBS *)

UNIT BBSTO;

INTERFACE

PROCEDURE send_msg_to_process(address_string : STRING);

IMPLEMENTATION

  USES
    bbdummy,
    bbmdata,
    bbmess,
    bbmisc,
    bbstr,
    bbuf,
    match;

(*===========================================================================*)
(* Send msg to processor                                                     *)
(*      Processes the 'TO' addressee data                                    *)
(*===========================================================================*)

PROCEDURE send_msg_to_process(address_string : STRING);

  VAR
    i           : BYTE;
    j           : BYTE;
    k           : BYTE;
    save_char   : CHAR;
    word_cnt    : BYTE;
    word_pos    : BYTE;
    word_string : STRING;

  LABEL no_more;

  {$I BBSTO2.PAS} (* Processes rest of send command *)

  (*=========================================================================*)
  (* Look up user bbs                                                        *)
  (*=========================================================================*)

  PROCEDURE user_bbs_lookup;

    VAR

      u_b_ptr  : user_record_ptr;
      u_ptr    : user_index_ptr;

    BEGIN;

      {$IFDEF DEBUG_H}
        WRITELN('BBS LOOKUP');
      {$ENDIF}

      u_ptr := find_uid(active_tcb^.curr_msg.msg_i_mb.msg_to);

      {$IFDEF DEBUG_H}
        WRITELN('UPTR = ', u_ptr <> NIL);
      {$ENDIF}

      IF u_ptr = NIL THEN EXIT;

      u_b_ptr := get_uid(u_ptr);

      {$IFDEF DEBUG_H}
        WRITELN('UBPTR = ', u_b_ptr <> NIL);
      {$ENDIF}

      active_tcb^.curr_msg.msg_i_mb.msg_to_at := u_b_ptr^.user_bbs;

      IF (u_b_ptr^.user_bbs <> opt_block.this_bb_sign)
                         AND (u_b_ptr^.user_bbs <> opt_block.this_bb_addr) THEN
        send_message(message_at_bbs_set);

    END;

  (*=========================================================================*)
  (* Main line                                                               *)
  (*=========================================================================*)

  BEGIN;

    WITH active_tcb^.curr_msg.msg_i_mb DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Initial load message header                                       *)
        (*-------------------------------------------------------------------*)

        msg_dt_in   := current_day_time;
        msg_dt_orig := current_day_time;

        msg_from    := active_tcb^.uid_data.user_id;
        msg_from_at := opt_block.this_bb_addr;

        (*-------------------------------------------------------------------*)
        (* Initial processing                                                *)
        (*-------------------------------------------------------------------*)

        strip_var(address_string, 'B');

        (*-------------------------------------------------------------------*)
        (* Change the magic word "at" to a @                                 *)
        (*-------------------------------------------------------------------*)

        word_string := upcase_str(address_string);
        i := POS(' AT ', word_string);

        IF i <> 0 THEN
          BEGIN;
            address_string[i+1] := '@';
            address_string[i+2] := ' ';
          END;

        (*-------------------------------------------------------------------*)
        (* If we have X@Y, change it to X @ Y                                *)
        (*-------------------------------------------------------------------*)

        i := POS('@', address_string);
        j := POS('.', address_string);

        IF (i > 1)
                   AND ((j = 0) OR (i < j))
                   AND (LENGTH(address_string) < 253)
                   AND ((address_string[i-1] <> ' ') OR
                                            (address_string[i+1] <> ' ')) THEN
          BEGIN;
            address_string := COPY(address_string, 1, i-1) + ' @ '
                                              + COPY(address_string, i+1, 255);
          END;

        (*-------------------------------------------------------------------*)
        (* Start parse of address string                                     *)
        (*-------------------------------------------------------------------*)

        word_cnt := words(address_string);

        word_string := upcase_str(subword(@address_string, 1, 1));

        (*-------------------------------------------------------------------*)
        (* If to address too big, error message please. Also bad format      *)
        (*-------------------------------------------------------------------*)

        IF (LENGTH(word_string) >= SIZEOF(msg_to))
                       OR ((NOT match_str(word_string, '+*'))
                                             AND (word_string[1] <> '=')) THEN
          BEGIN;
            send_message(message_addr_fmt);
            active_tcb^.error_sw := TRUE;
            EXIT;
          END;

        (*-------------------------------------------------------------------*)
        (* Set TO                                                            *)
        (*-------------------------------------------------------------------*)

        IF active_port^.port_suppress_ssid THEN
          word_string := strip_ssid(word_string);

        msg_to := word_string;

        IF (active_tcb^.tcb_abbs) OR (opt_block.opt_blank_to_bbs) THEN
          BEGIN;
            msg_to_at := '';
            msg_to_h  := '';
          END
        ELSE
          BEGIN;
            msg_to_at := opt_block.this_bb_addr;
            msg_to_h  := opt_block.this_bb_h;
          END;

        (*-------------------------------------------------------------------*)
        (* If we only had a TO, we can leave                                 *)
        (*-------------------------------------------------------------------*)

        IF word_cnt = 1 THEN
          GOTO no_more;

        (*-------------------------------------------------------------------*)
        (* More to go... This is somewhat variable                           *)
        (*-------------------------------------------------------------------*)

        address_string := subword(@address_string, 2, 0);

        more_address;

        (*-------------------------------------------------------------------*)
        (* Come here when ready to address resolve                           *)
        (*-------------------------------------------------------------------*)

no_more:

        (*-------------------------------------------------------------------*)
        (* If the TO_BBS is not blank or this is a BBS or no address         *)
        (* resolution wanted, then we are done                               *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_H}
          WRITELN('T1 = ', msg_to_at <> '');
          WRITELN('T2 = ', active_tcb^.uid_data.user_class = user_c_bu);
          WRITELN('T3 = ', NOT opt_block.opt_autoset_to_bbs);
        {$ENDIF}

        IF (msg_to_at <> '')
                         OR (active_tcb^.uid_data.user_class = user_c_bu)
                         OR (NOT opt_block.opt_autoset_to_bbs) THEN
          EXIT;

        (*-------------------------------------------------------------------*)
        (* See if we can resolve the address                                 *)
        (*-------------------------------------------------------------------*)

        user_bbs_lookup;

      END;

  END;

END.
