(*===========================================================================*)
(* Talk commands                                                             *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991 by H. Roy Engehausen.  All rights      *)
(*   reserved.                                                               *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

UNIT BBTCMD;

INTERFACE

PROCEDURE oper_t(cmd_in : STRING);
PROCEDURE talk_cmd(cmd_string : STRING);

IMPLEMENTATION

  USES
    CRT,
    bbconv,
    bbdummy,
    bbmdata,
    bbmess,
    bbmisc,
    bbmisc4,
    bbrdata,
    bbsdata,
    bbstr,
    bbtask,
    bbtime,
    bbuf,
    bbwin;

(*===========================================================================*)
(* Operator talk command                                                     *)
(*===========================================================================*)

PROCEDURE oper_t(cmd_in : STRING);
  VAR
    i            : WORD;
    find_this    : STRING[2];
    last_scb     : str_chain_ptr;
    save_port    : port_block_ptr;
    use_port     : port_block_ptr;
    talk_to_tcb  : tcb_ptr;
    t_chan       : BYTE;
    work_scb     : str_chain_ptr;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Process the command                                                   *)
    (*-----------------------------------------------------------------------*)

    IF cmd_in[2] <> ' ' THEN
      BEGIN;
        send_message(message_err_wrd);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    i := WORDS(cmd_in);

    IF i > 2 THEN
      BEGIN;
        send_message(message_err_wrd);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    IF (i < 2) OR (LENGTH(subword(@cmd_in, 2, 1)) <> 2) THEN
      BEGIN;
        send_message(message_2nd_bad);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* See if we can find the port                                           *)
    (*-----------------------------------------------------------------------*)

    find_this := upcase_str(subword(@cmd_in, 2, 1));

    save_port := active_port;

    IF NOT find_port(find_this[1]) THEN
      BEGIN;
        active_port := save_port;
        send_tnc_data_str('Cannot find the port');
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    t_chan := 0;
    WHILE (t_chan < SIZEOF(byte_to_char))
               AND (find_this[2] <> byte_to_char[t_chan]) DO
      INC(t_chan);

    IF (t_chan >= SIZEOF(byte_to_char)) OR (t_chan > active_port^.max_chan) THEN
      BEGIN;
        active_port := save_port;
        send_tnc_data_str('Invalid channel');
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* See if the task exists                                                *)
    (*-----------------------------------------------------------------------*)

    talk_to_tcb := find_tcb(find_this);

    IF talk_to_tcb <> NIL THEN
      WITH talk_to_tcb^ DO
        BEGIN;

          (*----------------------------------------------------------------*)
          (* Task exists -- Restore port pointer                            *)
          (*----------------------------------------------------------------*)

          active_port := save_port;

          (*----------------------------------------------------------------*)
          (* Task exists -- See if something special about it               *)
          (*----------------------------------------------------------------*)

          IF tcb_type <> th_user THEN
            BEGIN;
              send_tnc_data_str('Special task cannot be talked to');
              active_tcb^.error_sw := TRUE;
              EXIT;
            END;

          (*----------------------------------------------------------------*)
          (* OK we can talk to him so do it                                 *)
          (*----------------------------------------------------------------*)

          tcb_opr_talk := TRUE;

      END
    ELSE
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Task doesn't exist -- Set the channel                             *)
        (*-------------------------------------------------------------------*)

        WITH active_port^, active_tcb^ DO
          BEGIN;
            channel     := t_chan;
            port_chan_s := find_this;
          END;

        (*-------------------------------------------------------------------*)
        (* Start the new task                                                *)
        (*-------------------------------------------------------------------*)

        talk_to_tcb := task_create(@operator_terminal_task,
                                                          terminal_stack_size);

        (*-------------------------------------------------------------------*)
        (* Set the connected info                                            *)
        (*-------------------------------------------------------------------*)

        IF talk_to_tcb <> NIL THEN
          active_port^.connected^[talk_to_tcb^.channel] := talk_to_tcb;

        (*-------------------------------------------------------------------*)
        (* Restore port infomation                                           *)
        (*-------------------------------------------------------------------*)

        active_port             := save_port;
        active_tcb^.tcb_port    := save_port;
        active_tcb^.port_chan_s := 'LO';

        (*-------------------------------------------------------------------*)
        (* Check for thread start failure                                    *)
        (*-------------------------------------------------------------------*)

        IF talk_to_tcb = NIL THEN
          BEGIN;
            window_write('LO>:', 'Out of tasks for terminal');
            active_tcb^.error_sw := TRUE;
            EXIT;
          END;

        (*-------------------------------------------------------------------*)
        (* Initialize things                                                 *)
        (*-------------------------------------------------------------------*)

        WITH talk_to_tcb^ DO
          BEGIN;

            tcb_name          := 'TERMINAL';
            tcb_type          := th_opr_terminal;
            window            := window_operator;
            tcb_connect       := TRUE;
            tcb_ignore_lc     := TRUE;
            tcb_opr_talk      := TRUE;
            tcb_no_show_sdata := TRUE;
            max_pac           := 128;

            uid_init(@uid_data);

            WITH uid_data DO
              BEGIN;
                user_id      := tcb_name;
                user_ssid    := '';
                user_class   := user_c_lsu;
                user_flag    := user_f_sysop;
              END;

          END;

      END;

    (*-----------------------------------------------------------------------*)
    (* Set up to talk to the guy                                             *)
    (*-----------------------------------------------------------------------*)

    WITH talk_to_tcb^ DO
      BEGIN;
        tcb_opr_talk          := TRUE;
        conv_tcb              := active_tcb;
        active_tcb^.conv_tcb  := talk_to_tcb;
      END;

  END;

(*===========================================================================*)
(* User talk command                                                         *)
(*===========================================================================*)

PROCEDURE talk_cmd(cmd_string : STRING);
  VAR
    i          : WORD;
    j          : BYTE;
    time_to_go : BYTE;
    wait_for   : LONGINT;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Check command                                                         *)
    (*-----------------------------------------------------------------------*)

    IF WORDS(cmd_string) > 1 THEN
      BEGIN;
        send_message(message_err_wrd);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    IF (LENGTH(cmd_string) > 1) AND (cmd_string[2] <> ' ') THEN
      BEGIN;
        send_message(message_err_2nd);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* See if operator is in                                                 *)
    (*-----------------------------------------------------------------------*)

    IF NOT opt_block.opt_sysop_is_in THEN
      BEGIN;
        send_message(message_sysop_not_in);
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Prevent operator from talking to self!                                *)
    (*-----------------------------------------------------------------------*)

    IF active_tcb^.tcb_type = th_operator THEN
      BEGIN;
        send_tnc_data_str('You cannot talk to yourself!');
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Tell user we are waiting                                              *)
    (*-----------------------------------------------------------------------*)

    send_message(message_waiting_for_op);
    send_flush;

    (*-----------------------------------------------------------------------*)
    (* Wait for operator loop                                                *)
    (*-----------------------------------------------------------------------*)

    wait_for := up_time_from_now(60);

    WHILE up_time < wait_for DO
      WITH active_tcb^ DO
        BEGIN;

          (*----------------------------------------------------------------*)
          (* Give up some time                                              *)
          (*----------------------------------------------------------------*)

          FOR i := 1 TO 5 DO
            task_switch;

          (*----------------------------------------------------------------*)
          (* See if operator answered.  If so we fall out and get caught    *)
          (* in the main command loop at "Operator wants to talk to you"    *)
          (*----------------------------------------------------------------*)

          IF conv_tcb <> NIL THEN
            BEGIN;
              NOSOUND;
              EXIT;
            END;

          (*----------------------------------------------------------------*)
          (* See if anything is waiting from user.  If so cancel loop       *)
          (*----------------------------------------------------------------*)

          i := send_pending(FALSE);

          IF read_tnc_data_pending THEN
            BEGIN;
              send_message(message_wait_cancel);
              cmd_string := read_tnc_data_str;
              NOSOUND;
              EXIT;
            END;

          (*----------------------------------------------------------------*)
          (* Ring the bell                                                  *)
          (*----------------------------------------------------------------*)

          IF opt_block.opt_sysop_bell THEN
            BEGIN;
              time_to_go := current_day_time AND $03;
              IF time_to_go = 2 THEN
                BEGIN;
                  FOR i := 1 TO opt_block.bell_length2 DO
                    BEGIN;
                      SOUND(660);
                      FOR j := 1 TO opt_block.bell_length1 DO
                        task_switch;
                      SOUND(440);
                      FOR j := 1 TO opt_block.bell_length1 DO
                        task_switch;
                    END;
                  NOSOUND;
                END;
            END;

        END;

    (*-----------------------------------------------------------------------*)
    (* Send message that no SYSOP answered and leave                         *)
    (*-----------------------------------------------------------------------*)

    send_message(message_wait_no_sysop);

  END;

END.
