(*===========================================================================*)
(* Initialize TNC                                                            *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

{$DEFINE DEBUG_TINITT}
{$UNDEF  DEBUG_TINITT}

(*===========================================================================*)
(* Initialize tnc                                                            *)
(*===========================================================================*)

PROCEDURE init_this_tnc;

  VAR
    b          : BOOLEAN;
    i          : BYTE;
    r          : REGISTERS;
    s          : STRING;
    save_speed : WORD;

  BEGIN;

    WITH active_port^, active_tcb^ DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Clear buffer                                                      *)
        (*-------------------------------------------------------------------*)

        FILLCHAR(tnc_data, SIZEOF(tnc_data), 0);
        FILLCHAR(tnc_htt^, tnc_b_size, 0);

        (*-------------------------------------------------------------------*)
        (* Put in a phoney aux thread for now                                *)
        (*-------------------------------------------------------------------*)

        active_port^.aux_thread := active_tcb;

        (*-------------------------------------------------------------------*)
        (* Set port speed.  For modems we must force the freeze switch off.  *)
        (*-------------------------------------------------------------------*)

        save_speed := active_port^.cur_rate;

        IF active_port^.port_type = port_modem THEN
          BEGIN;
            b := active_port^.modem_freez;
            active_port^.modem_freez := FALSE;
          END;

        set_port_speed(active_port^.data_rate);

        IF active_port^.port_type = port_modem THEN
          active_port^.modem_freez := b
        ELSE
          active_port^.cur_rate := save_speed;

        (*-------------------------------------------------------------------*)
        (* Set the broadcast path to proper format                           *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_TINITT}
          WRITELN('BCST SET -- COM = ', com_number);
        {$ENDIF}

        IF (bcst_path <> '') AND (active_port ^.port_type <> port_bpqhost) THEN
         BEGIN;
            s := 'C ' + bcst_path;
            connect_format(@s, active_port);
            bcst_path := subword(@s, 2, 0);
          END;

        (*-------------------------------------------------------------------*)
        (* Skip port if subport                                              *)
        (*-------------------------------------------------------------------*)

        IF port_sub_port THEN
          EXIT;

        (*-------------------------------------------------------------------*)
        (* Set up for this port                                              *)
        (*-------------------------------------------------------------------*)

        tcb_port := active_port;
        w_color  := port_color;

        (*-------------------------------------------------------------------*)
        (* Set for the unproto channel                                       *)
        (*-------------------------------------------------------------------*)

        channel := 0;
        port_chan_s := port_char + '0';

        (*-------------------------------------------------------------------*)
        (* Collect garbage as needed                                         *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_TINITT}
          WRITELN('Garbage collect');
        {$ENDIF}

        {$IFDEF DEBUG_BPQHOST1}
          WRITELN('Garbage collect');
        {$ENDIF}

        gc;

        (*-------------------------------------------------------------------*)
        (* Set max connects                                                  *)
        (*-------------------------------------------------------------------*)

        STR(max_conn, t_str);

        (*-------------------------------------------------------------------*)
        (* Different initialization for different ports                      *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_TINITT}
           WRITELN('Port setup - ', ORD(port_type));
           DELAY(1000);
        {$ENDIF}

        CASE port_type OF

          (*-----------------------------------------------------------------*)
          (* BPQ host mode..  Set monitor                                    *)
          (*-----------------------------------------------------------------*)

          port_bpqhost :
            BEGIN;

              IF port_monitor THEN
                cmd_it('MIUSA')
              ELSE
                cmd_it('MN');

            END;

          (*-----------------------------------------------------------------*)
          (* PK-232 and BPQ 232 mode ports                                   *)
          (*-----------------------------------------------------------------*)

          port_aeapk232, port_g8bpq :
            BEGIN;

              IF port_monitor THEN
                BEGIN;

 {$IFDEF REMOVE} (* This code was for G8BPQ version 3.57 or earlier *)
                  IF port_type = port_aeapk232 THEN
                    BEGIN;
 {$ENDIF}
                      cmd_it('MN6');
                      cmd_it('MC6');
 {$IFDEF REMOVE} (* This code was for G8BPQ version 3.57 or earlier *)
                    END
                  ELSE
                    cmd_it('MN2');
 {$ENDIF}

                END
              ELSE
                BEGIN;
                  cmd_it('MN0');
                  cmd_it('MC0');
                END;

              IF port_type = port_aeapk232 THEN
                BEGIN;
                  cmd_it('MN');

                  IF call_set <> '' THEN
                    cmd_it('ML' + call_set);

                  IF bcst_path <> '' THEN
                    cmd_it('UN' + bcst_path);

                  cmd_it('CFALL');

                  cmd_it('PL0');
                  cmd_it('8BY');
                  cmd_it('AA0');
                  cmd_it('AIN');
                  cmd_it('CET');
                  cmd_it('CH00');
                  cmd_it('DLN');
                  cmd_it('ECN');
                  cmd_it('FLN');
                  cmd_it('MI0');
                  cmd_it('NEN');
                  cmd_it('RLN');
                  cmd_it('SO00');
                  cmd_it('ST00');
                  cmd_it('TFN');
                  cmd_it('XWN');
                  cmd_it('XF00');
                  cmd_it('XN00');

                END;

              cmd_it('UR' + t_str);

            END;

          (*-----------------------------------------------------------------*)
          (* Modem with AT command set                                       *)
          (*-----------------------------------------------------------------*)

          port_modem :
            BEGIN;

              {$IFDEF DEBUG_TINITT}
                WRITELN('Modem setup');
              {$ENDIF}

              modem_cmd('T E0 V1 X1 Q0 S7=60 S0=0');

            END;

          (*-----------------------------------------------------------------*)
          (* All others                                                      *)
          (*-----------------------------------------------------------------*)

          ELSE
            BEGIN;

              {$IFDEF DEBUG_TINITT}
                WRITELN('MIUSC');
              {$ENDIF}

              IF port_monitor THEN
                IF port_type = port_dedhost1 THEN
                  cmd_it('MIUSCRT')
                ELSE
                  cmd_it('MIUSC')
              ELSE
                cmd_it('MN');

              cmd_it('M');

              IF call_set <> '' THEN
                cmd_it('I' + call_set);

              IF bcst_path <> '' THEN
                cmd_it('C' + bcst_path);

              cmd_it('Y' + t_str);

            END;

          END; (*----- End Port-type case -----------------------------------*)

        (*-------------------------------------------------------------------*)
        (* Disconnect all the channels                                       *)
        (*-------------------------------------------------------------------*)

        FOR i := 0 TO max_chan DO
          BEGIN;

            {$IFDEF DEBUG_TINITT}
              WRITELN('Channel disconnect - ', i);
              DELAY(1000);
            {$ENDIF}

            channel := i;

            port_chan_s := port_char + byte_to_char[i];

            IF port_type = port_bpqhost THEN
              BEGIN;

                {$IFDEF DEBUG_BPQHOST1}
                  WRITELN('Issue open - ', i);
                {$ENDIF}

                r.AH := bpq_open;
                r.AL := i + 1;

                IF (i > 0) AND (i <= max_conn) THEN
                  BEGIN;
                    r.CL := bpq_appl_flag_sendcon_user
                            + bpq_appl_flag_sendcon_appl;
                    r.DL := opt_block.bpq_appl_num;
                  END
                ELSE
                  BEGIN;
                    r.CL := bpq_appl_flag_sendcon_user
                            + bpq_appl_flag_sendcon_appl;
                    IF (i = 0) AND port_monitor THEN
                      r.CL := r.CL OR bpq_appl_mask_monitor;
                    r.DL := 0;
                  END;

                INTR(com_number, r);

                {$IFDEF DEBUG_BPQHOST1}
                  WRITELN('Open complete');
                {$ENDIF}

              END;

            IF i <> 0 THEN
              BEGIN;

                {$IFDEF DEBUG_BPQHOST1}
                  WRITELN('Disconnect');
                {$ENDIF}

                cmd_it('D');

              END;

            {$IFDEF DEBUG_TINITT}
              WRITELN('Poll - ', i);
              DELAY(1000);
            {$ENDIF}

            REPEAT
              send_recv_tnc(2);
              IF NOT tnc_null THEN
                window_write('', tnc_data.str_data);
            UNTIL tnc_null;

          END;

        (*-------------------------------------------------------------------*)
        (* Read in first file                                                *)
        (*-------------------------------------------------------------------*)

        {$IFDEF DEBUG_TINITT}
          WRITELN('File init');
        {$ENDIF}

        IF first_load <> '' THEN
          t_file_init;

      END;

  END;
