(*===========================================================================*)
(* Update/find a userid                                                      *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1991 by H. Roy Engehausen.  All rights reserved.  *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Update a user_id                                                          *)
(*===========================================================================*)

PROCEDURE update_uid (uid_to_upd : user_record_ptr);

  VAR
    uid_index_current : user_index_ptr;

  {$UNDEF DEBUG_U}
  {$UNDEF TRACE_U}

  BEGIN;

    {$IFDEF POINT_CHK}
      test_pointer(uid_to_upd);
    {$ENDIF}

    {$IFDEF DEBUG_U}
      WRITELN('Update uid -- ', active_tcb^.tcb_number, '-- T = ',
              uid_to_upd^.user_l_time, ' -- u = ',
              uid_to_upd^.user_id);
      DELAY(2000);
    {$ENDIF}

    {$IFDEF TRACE_U}
      trace_data('UU', $7FFF AND uid_to_upd^.user_flag,
                                                     NIL, uid_to_upd^.user_id);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Obtain the interrupt lock                                             *)
    (*-----------------------------------------------------------------------*)

    get_semaphore(semaphore_interrupts, sem_exclusive, FALSE);

    (*-----------------------------------------------------------------------*)
    (* Get pointer to record                                                 *)
    (*-----------------------------------------------------------------------*)

    uid_index_current := uid_to_upd^.user_i_ptr;

    uid_index_current^.user_id := uid_to_upd^.user_id;

    (*-----------------------------------------------------------------------*)
    (* Sanity check                                                          *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF POINT_CHK}
      test_pointer(uid_index_current);
    {$ENDIF}

    IF uid_index_current^.user_recno > uid_total THEN
      BEGIN;
        WRITELN('Add fail! URec =', uid_index_current^.user_recno);
        WRITELN('Utot =', uid_total);
        uid_index_current^.user_recno := uid_total;
      END;

    (*-----------------------------------------------------------------------*)
    (* Write record                                                          *)
    (*-----------------------------------------------------------------------*)

    SEEK (uid_file, uid_index_current^.user_recno);
    WRITE(uid_file, uid_to_upd^);

    (*-----------------------------------------------------------------------*)
    (* Force the file out                                                    *)
    (*-----------------------------------------------------------------------*)

    flushdosfile(@uid_file);

    (*-----------------------------------------------------------------------*)
    (* Release the interrupt lock                                            *)
    (*-----------------------------------------------------------------------*)

    free_semaphore(semaphore_interrupts);

  END;

(*===========================================================================*)
(* Find a user in the index                                                  *)
(*===========================================================================*)

FUNCTION find_uid(uid_to_get : call_sign_str) : user_index_ptr;

  VAR
    uid_i_current : user_index_ptr;

  BEGIN;

    uid_i_current := uid_chain;

    WHILE uid_i_current <> NIL DO
      BEGIN;

        {$IFDEF POINT_CHK}
          test_pointer(uid_i_current);
        {$ENDIF}

        IF uid_to_get = uid_i_current^.user_id THEN
          BEGIN;
            find_uid := uid_i_current;
            EXIT;
          END;

        IF uid_to_get > uid_i_current^.user_id THEN
          uid_i_current := uid_i_current^.user_next
        ELSE
          uid_i_current := NIL;

      END;

    find_uid := NIL;

  END;
