(*===========================================================================*)
(* Get a user                                                                *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991 by H. Roy Engehausen.  All rights      *)
(*   reserved.                                                               *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Get the data about a user                                                 *)
(*===========================================================================*)

FUNCTION get_uid(uid_to_get : user_index_ptr) : user_record_ptr;

  VAR
    uid_buffer : user_record_ptr;

  BEGIN;

    {$IFDEF POINT_CHK}
      test_pointer(uid_to_get);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Get the uid buffer                                                    *)
    (*-----------------------------------------------------------------------*)

    uid_buffer := get_uid_buffer;

    (*-----------------------------------------------------------------------*)
    (* Obtain the interrupt lock                                             *)
    (*-----------------------------------------------------------------------*)

    get_semaphore(semaphore_interrupts, sem_exclusive, FALSE);

    (*-----------------------------------------------------------------------*)
    (* Read the record                                                       *)
    (*-----------------------------------------------------------------------*)

    SEEK(uid_file, uid_to_get^.user_recno);
    READ(uid_file, uid_buffer^);

    (*-----------------------------------------------------------------------*)
    (* Validate                                                              *)
    (*-----------------------------------------------------------------------*)

    IF uid_buffer^.user_id <> uid_to_get^.user_id THEN
      BEGIN;
        WRITELN('Get Usr fail -- Want =', uid_to_get^.user_id);
        WRITELN('Got =', uid_buffer^.user_id,
                                      ' at rec =', uid_to_get^.user_recno);
      END;

    (*-----------------------------------------------------------------------*)
    (* Fix up times                                                          *)
    (*-----------------------------------------------------------------------*)

    IF uid_buffer^.user_last   > current_day_time THEN
      uid_buffer^.user_last   := current_day_time;
    IF uid_buffer^.user_l_time > current_day_time THEN
      uid_buffer^.user_l_time := current_day_time;

    (*-----------------------------------------------------------------------*)
    (* Fix up "N" time.   Only needed in version 2 user files                *)
    (*-----------------------------------------------------------------------*)

    IF (uid_buffer^.user_n_time = 0)
                          OR (uid_buffer^.user_n_time > current_day_time) THEN
      uid_buffer^.user_n_time := uid_buffer^.user_last;

    (*-----------------------------------------------------------------------*)
    (* Update pointer                                                        *)
    (*-----------------------------------------------------------------------*)

    uid_buffer^.user_i_ptr := uid_to_get;

    (*-----------------------------------------------------------------------*)
    (* Set response                                                          *)
    (*-----------------------------------------------------------------------*)

    get_uid := uid_buffer;

    (*-----------------------------------------------------------------------*)
    (* Release the interrupt lock                                            *)
    (*-----------------------------------------------------------------------*)

    free_semaphore(semaphore_interrupts);

  END;
