PROGRAM bbufix;

(*===========================================================================*)
(* Fix user file                                                             *)
(*                                                                           *)
(*   Copyright 1989, 1990, 1991 by H. Roy Engehausen.  All rights reserved.  *)
(*   This software may be freely distributed and used, but it may not        *)
(*   under any circumstances be sold by anyone other than the author.        *)
(*   It may be distributed by a commercial company as long as it is          *)
(*   for no cost.                                                            *)
(*                                                                           *)
(*===========================================================================*)

{$R+}    {Range checking on}
{$B-}    {Boolean complete evaluation off}
{$S+}    {Stack checking on}
{$I+}    {I/O checking on}
{$V+}    {String var checks}
{$N-}    {No numeric coprocessor}

USES
  CRT,
  bbtime,
  bbdummy;

(*===========================================================================*)
(* Main program begins here                                                  *)
(*===========================================================================*)

VAR
  buffer          : STRING;
  code            : INTEGER;
  delete_new      : LONGINT;
  delete_unreg    : LONGINT;
  delete_user     : LONGINT;
  e_file          : FILE OF user_record_type;
  first_time      : BOOLEAN;
  fix_file        : FILE OF user_record_type;
  i               : WORD;
  i_name          : STRING;
  k               : WORD;
  li              : LONGINT;
  n_ok            : BOOLEAN;
  no_blk          : WORD;
  no_blk_max      : WORD;
  no_blk_tell     : WORD;
  o_name          : STRING;
  o_cnt           : WORD;
  rename_it       : BOOLEAN;
  reset_logon     : WORD;
  t_time          : LONGINT;
  user_buffer     : user_record_type;
  user_file       : FILE OF user_record_type;
  x_switch        : BOOLEAN;

LABEL
  loop,
  no_delete;

FUNCTION yn : BOOLEAN;
  VAR
    c : CHAR;
    x : STRING;

  BEGIN;

    IF x_switch THEN
      BEGIN;
        yn := TRUE;
        EXIT;
      END;

    WHILE TRUE DO
      BEGIN;
        WRITELN('Enter Y or N or X (Yes and dont ask anymore)');
        READLN(x);
        IF LENGTH(x) <> 1 THEN
          WRITELN('Wrong length')
        ELSE
          BEGIN;
            c := UPCASE(x[1]);
            IF c = 'X' THEN
              BEGIN;
                x_switch := TRUE;
                yn := TRUE;
                EXIT;
              END;
            IF (c = 'Y') OR (c = 'N') THEN
              BEGIN;
                yn := c = 'Y';
                EXIT;
              END;
            WRITELN('Answer not Y or N');
          END;

      END;
  END;

FUNCTION deleteme : BOOLEAN;
  BEGIN;
    WRITELN('Delete this record?');
    deleteme := yn;
  END;

(*===========================================================================*)
(* Main program begins here                                                  *)
(*===========================================================================*)

  BEGIN;

    DIRECTVIDEO := FALSE;

    x_switch := FALSE;

    o_cnt := 0;

    i_name  := PARAMSTR(1);
    o_name  := PARAMSTR(2);

    IF i_name = '' THEN
      i_name := 'USER.BB';

    IF o_name = '' THEN
      BEGIN;
        o_name := 'USER.OK';
        rename_it := TRUE;
      END
    ELSE
      rename_it := FALSE;

    (*-----------------------------------------------------------------------*)
    (* Open msg file for read                                                *)
    (*-----------------------------------------------------------------------*)

    ASSIGN(user_file, i_name);
    {$I-}
    RESET(user_file);
    {$I+}
    i := IORESULT;

    (*-----------------------------------------------------------------------*)
    (* Check for open OK                                                     *)
    (*-----------------------------------------------------------------------*)

    IF (i <> 0) AND (i <> 2) THEN
      BEGIN;
        WRITELN('Unable to open ', opt_block.user_file_name);
        WRITELN('DOS Error # ', i);
      END;

    (*-----------------------------------------------------------------------*)
    (* File doesn't exist.                                                   *)
    (*-----------------------------------------------------------------------*)

    IF i <> 0 THEN
      BEGIN;

        WRITELN('File does not exist');
        HALT;

      END;

    (*-----------------------------------------------------------------------*)
    (* Open msg file for write                                               *)
    (*-----------------------------------------------------------------------*)

    ASSIGN(fix_file, o_name);
    REWRITE(fix_file);

    (*-----------------------------------------------------------------------*)
    (* Check the version of the file                                         *)
    (*-----------------------------------------------------------------------*)

    READ(user_file, user_buffer);
    i := OFS(user_buffer.user_i_ptr^);
    IF i <> sys_version THEN
      BEGIN;
        WRITELN('Wanted version ', sys_version, ' user file');
        WRITELN('Got version ', i, ' user file');
        HALT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Write the version header                                              *)
    (*-----------------------------------------------------------------------*)

    WRITE(fix_file, user_buffer);

    (*-----------------------------------------------------------------------*)
    (* Get clock                                                             *)
    (*-----------------------------------------------------------------------*)

    FILLCHAR(today_time, SIZEOF(today_time), #0);

    new_date;

    WRITELN;

    buffer := time_str(current_day_time, TRUE);
    WRITELN('Today is ', COPY(buffer, 1, 6));

    (*-----------------------------------------------------------------------*)
    (* Get the number of days to delete new                                  *)
    (*-----------------------------------------------------------------------*)

    WRITELN;

    n_ok := FALSE;

    REPEAT;

      WRITELN('How many days before a new user gets deleted?');
      READLN(buffer);
      VAL(buffer, li, code);
      IF (code <> 0) OR (li < 0) OR (li > 999) THEN
        WRITELN('Not a valid number between 1 and 999')
      ELSE
        n_ok := TRUE;

    UNTIL n_ok;

    delete_new := current_day_time - li * ticks_per_day;

    (*-----------------------------------------------------------------------*)
    (* Get the number of days to delete unreg                                *)
    (*-----------------------------------------------------------------------*)

    WRITELN;

    n_ok := FALSE;

    REPEAT;

      WRITELN('How many days before an unregistered user gets deleted?');
      READLN(buffer);
      VAL(buffer, li, code);
      IF (code <> 0) OR (li < 0) OR (li > 999) THEN
        WRITELN('Not a valid number between 1 and 999')
      ELSE
        n_ok := TRUE;

    UNTIL n_ok;

    delete_unreg := current_day_time - li * ticks_per_day;

    (*-----------------------------------------------------------------------*)
    (* Get the number of days to delete users                                *)
    (*-----------------------------------------------------------------------*)

    WRITELN;

    n_ok := FALSE;

    REPEAT;

      WRITELN('How many days before a registered user gets deleted?');
      READLN(buffer);
      VAL(buffer, li, code);
      IF (code <> 0) OR (li < 0) OR (li > 999) THEN
        WRITELN('Not a valid number between 1 and 999')
      ELSE
        n_ok := TRUE;

    UNTIL n_ok;

    delete_user := current_day_time - li * ticks_per_day;

    (*-----------------------------------------------------------------------*)
    (* Reset logon count?                                                    *)
    (*-----------------------------------------------------------------------*)

    WRITELN;

    n_ok := FALSE;

    REPEAT;

      WRITELN('Reset logon counter for users with more than this');
      WRITELN('many logons.  0 means don''t reset anyone.');
      READLN(buffer);
      VAL(buffer, li, code);
      IF (code <> 0) OR (li < 0) OR (li > 999) THEN
        WRITELN('Not a valid number between 0 and 999')
      ELSE
        n_ok := TRUE;

    UNTIL n_ok;

    reset_logon := li;

    (*-----------------------------------------------------------------------*)
    (* Adjust current day time to prevent problems                           *)
    (*-----------------------------------------------------------------------*)

    current_day_time := current_day_time + ticks_per_day;

    (*-----------------------------------------------------------------------*)
    (* Read in the user file                                                 *)
    (*-----------------------------------------------------------------------*)

    no_blk      := 0;
    no_blk_tell := 0;
    no_blk_max  := FILESIZE(user_file) - 1;

loop:
    WHILE no_blk < no_blk_max DO
      BEGIN;

        li := SIZEOF(user_buffer);
        li := li * no_blk;

        i := no_blk DIV 10;
        IF i <> no_blk_tell THEN
          BEGIN;
            no_blk_tell := i;
            WRITELN('Processing record # ', no_blk);
          END;

        (*-------------------------------------------------------------------*)
        (* Read in the next record                                           *)
        (*-------------------------------------------------------------------*)

        INC(no_blk);

        READ(user_file, user_buffer);

        (*-------------------------------------------------------------------*)
        (* Remove deletes                                                    *)
        (*-------------------------------------------------------------------*)

        WITH user_buffer DO
          BEGIN;
            IF (user_flag AND user_f_delete) <> 0 THEN
              BEGIN;
                WRITELN('Delete -- ', user_id );
                GOTO loop;
              END;
          END;

        (*-------------------------------------------------------------------*)
        (* Don't remove excluded                                             *)
        (*-------------------------------------------------------------------*)

        IF (user_buffer.user_flag AND user_f_exclude) <> 0 THEN
          GOTO no_delete;

        (*-------------------------------------------------------------------*)
        (* Remove one timers                                                 *)
        (*-------------------------------------------------------------------*)

        WITH user_buffer DO
          BEGIN;

            IF (user_class = user_c_nu) AND (user_last < delete_new) THEN
              BEGIN;
                IF user_pw <> '' THEN
                  WRITELN('New user kept because of modem password -- ',
                          user_id, ' -- ', time_str(user_last, TRUE))
                ELSE
                  BEGIN;
                    WRITELN('Delete new user -- ', user_id,
                        ' -- ', time_str(user_last, TRUE));
                    GOTO loop;
                  END;
              END;

            IF (user_class = user_c_uu) AND (user_last < delete_unreg) THEN
              BEGIN;
                IF user_pw <> '' THEN
                  WRITELN('Unregistered user kept because of modem password',
                          ' -- ', user_id, ' -- ', time_str(user_last, TRUE))
                ELSE
                  BEGIN;
                    WRITELN('Delete unregistered user -- ', user_id,
                            ' -- ', time_str(user_last, TRUE));
                    GOTO loop;
                  END;
              END;

            IF ((user_class = user_c_ou) OR (user_class = user_c_eu))
                                 AND (user_last < delete_user) THEN
              BEGIN;
                IF user_pw <> '' THEN
                  WRITELN('Registered user kept because of modem password',
                          ' -- ', user_id, ' -- ', time_str(user_last, TRUE))
                ELSE
                  BEGIN;
                    WRITELN('Delete registered user -- ', user_id,
                            ' -- ', time_str(user_last, TRUE));
                    GOTO loop;
                  END;
              END;

          END;

        (*-------------------------------------------------------------------*)
        (* Validate date time                                                *)
        (*-------------------------------------------------------------------*)

no_delete:

        WITH user_buffer DO
          IF user_l_time > current_day_time THEN
            BEGIN;
              WRITELN('Bad L-time -- ', user_id , ' -- ',
                                            time_str(user_l_time, TRUE));
              user_l_time := current_day_time;
            END;

        (*-------------------------------------------------------------------*)
        (* Reset log                                                         *)
        (*-------------------------------------------------------------------*)

        WITH user_buffer DO
          IF (reset_logon > 0) AND (user_l_cnt > reset_logon) THEN
            BEGIN;
              WRITELN('User ', user_id, ' logon count of ', user_l_cnt,
                      ' has been reset to 1');
              user_l_cnt := 1;
            END;

        (*-------------------------------------------------------------------*)
        (* Write fix file                                                    *)
        (*-------------------------------------------------------------------*)

        INC(o_cnt);

        WRITE(fix_file, user_buffer);

      END; (*----- End of loop reading records ------------------------------*)

    WRITELN('Processed ', no_blk, ' records.');

    CLOSE(user_file);

    CLOSE(fix_file);

    WRITELN;
    WRITELN('Wrote ', o_cnt, ' records');
    WRITELN;

    IF rename_it THEN
      BEGIN;

        ASSIGN(e_file, 'USER.BAD');
        {$I-}
        ERASE(e_file);
        i := IORESULT;
        {$I+}

        RENAME(user_file, 'USER.BAD');
        RENAME(fix_file, 'USER.BB');

        WRITELN('The good file is still USER.BB');
        WRITELN('The bad file has been renamed USER.BAD');

      END;

  END.
