(*===========================================================================*)
(* Process wakeup line                                                       *)
(*                                                                           *)
(*   Copyright 1988, 1989 by H. Roy Engehausen.  All rights reserved.        *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Wakeup line processor                                                     *)
(*===========================================================================*)

PROCEDURE process_line;

  VAR
    action_time     : LONGINT;
    action_time_str : STRING[5];
    action_date     : LONGINT;
    action_date_str : STRING[8];
    code            : INTEGER;
    colon_pos       : BYTE;
    convert_error   : BOOLEAN;
    eq_sw           : BOOLEAN;
    hr              : BYTE;
    i               : INTEGER;
    last_time_str   : STRING[11];
    temp_time       : LONGINT;
    time_ok         : BOOLEAN;
    time_increment  : BOOLEAN;
    word_cnt        : BYTE;

  {$I BBWAKEDC}     (* Date check *)
  {$I BBWAKETC}     (* Time check *)

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Initialize a few things                                               *)
    (*-----------------------------------------------------------------------*)

    last_time     := current_day_time - ticks_per_day;
    last_time_str := '';

    did_today     := FALSE;

    event_time   := max_time;

    (*-----------------------------------------------------------------------*)
    (* Check to see if we have an old time                                   *)
    (*-----------------------------------------------------------------------*)

    wakeup_temp := subword(@line_buffer, 1, 1);

    IF LENGTH(wakeup_temp) = LENGTH(todays_date_time) THEN
      BEGIN;

        last_time_str := wakeup_temp;
        temp_time     := time_unstr(last_time_str, convert_error);
        IF NOT convert_error THEN
          last_time := temp_time
        ELSE
          BEGIN;
            window_write('WKE:', 'WAKEUP time conversion error -- Line #'
                                                               + line_no_str);
            EXIT;
          END;

        did_today :=
                 substr(todays_date_time, 1, 6) = substr(last_time_str, 1, 6);

        line_buffer := subword(@line_buffer, 2, 0);

      END;

    (*-----------------------------------------------------------------------*)
    (* Get the word count and validate                                       *)
    (*-----------------------------------------------------------------------*)

    word_cnt := words(line_buffer);

    IF word_cnt < 3 THEN
      BEGIN;
        window_write('WKE:',
                            'Not enough words on WAKEUP line #' + line_no_str);
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Get the action date                                                   *)
    (*-----------------------------------------------------------------------*)

    wakeup_temp := subword(@line_buffer, 1, 1);

    IF LENGTH(wakeup_temp) >= SIZEOF(action_date_str) THEN
      BEGIN;
        window_write('WKE:', 'Invalid WAKEUP date on line #' + line_no_str);
        EXIT;
      END;

    action_date_str := wakeup_temp;

    (*-----------------------------------------------------------------------*)
    (* Get the action time                                                   *)
    (*-----------------------------------------------------------------------*)

    wakeup_temp := subword(@line_buffer, 2, 1);

    IF wakeup_temp[1] = '+' THEN
      BEGIN;
        time_increment := TRUE;
        wakeup_temp := substr(wakeup_temp, 2, 0);
      END
    ELSE
      time_increment := FALSE;

    IF LENGTH(wakeup_temp) >= SIZEOF(action_time_str) THEN
      BEGIN;
        window_write('WKE:', 'Invalid WAKEUP date on line #' + line_no_str);
        EXIT;
      END;

    action_time_str := wakeup_temp;

    (*-----------------------------------------------------------------------*)
    (* Now check the dates                                                   *)
    (*-----------------------------------------------------------------------*)

    date_ok := FALSE;
    date_check;

    IF NOT date_ok THEN
      EXIT;

    (*-----------------------------------------------------------------------*)
    (* Convert action time to ticks since midnight                           *)
    (*-----------------------------------------------------------------------*)

    time_check;

    (*-----------------------------------------------------------------------*)
    (* If the time is not an increment and we ran today, then we are done    *)
    (*-----------------------------------------------------------------------*)

    IF (NOT time_ok)
                  OR (did_today AND (NOT time_increment) AND (NOT eq_sw)) THEN
      EXIT;

    (*-----------------------------------------------------------------------*)
    (* Time increment is simple                                              *)
    (*-----------------------------------------------------------------------*)

    IF time_increment THEN
      BEGIN;
        event_time := last_time + action_time;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Calculate event time.  If '==' in time area, then use action time     *)
    (*-----------------------------------------------------------------------*)

    IF NOT eq_sw THEN
      event_time := action_time + last_midnight
    ELSE
      event_time := action_time;

  END;
