(*===========================================================================*)
(* Scroll                                                                    *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1991 by H. Roy Engehausen.  All rights reserved.  *)
(*                                                                           *)
(*===========================================================================*)

PROCEDURE scr_window(up_or_down : BOOLEAN; number_of_lines : WORD);
  VAR
    s : STRING;
    i : INTEGER;
    w : window_data_ptr;
    l : window_data_ptr;

  BEGIN;

    WITH window_array[current_window] DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* If this window doesn't scroll, forget it                          *)
        (*-------------------------------------------------------------------*)

        IF NOT window_scrollable THEN
          EXIT;

        (*-------------------------------------------------------------------*)
        (* Loop for the right number of lines                                *)
        (*-------------------------------------------------------------------*)

        WHILE number_of_lines > 0 DO
          BEGIN;

            DEC(number_of_lines);

            (*---------------------------------------------------------------*)
            (* Handle direction                                              *)
            (*---------------------------------------------------------------*)

            IF up_or_down THEN
              BEGIN;

                (*-----------------------------------------------------------*)
                (* Direction is up!  If at the top already, we can't do much *)
                (*-----------------------------------------------------------*)

                IF window_at_top THEN
                  EXIT;

                (*-----------------------------------------------------------*)
                (* Find line before this one.  It had better not be NIL since*)
                (* we ain't at the top but check anyway                      *)
                (*-----------------------------------------------------------*)

                w := window_line^.last_line;

                IF w = NIL THEN
                  EXIT;

                {$IFDEF LOOP_POINT_CHK}
                  IF w = window_line THEN
                    BEGIN;
                      WRITELN('Window loop at scroll ');
                      dump_string('Wind loop at scroll ' + p2x(w));
                      dump_window(current_window);
                      HALT;
                    END;
                {$ENDIF}

                (*-----------------------------------------------------------*)
                (* Set previous line as current (bottom of window) and       *)
                (* then find the real top!  Again check for NIL to prevent   *)
                (* stupid errors                                             *)
                (*-----------------------------------------------------------*)

                window_line := w;

                w := find_window_top_line;

                IF w = NIL THEN
                  EXIT;                          (* Stupidity check          *)

                (*-----------------------------------------------------------*)
                (* Set the top switch as necessary                           *)
                (*-----------------------------------------------------------*)

                IF w^.last_line = NIL THEN
                  window_at_top := TRUE;
(*Was an exit here*)

                (*-----------------------------------------------------------*)
                (* Go write the line out                                     *)
                (*-----------------------------------------------------------*)

                GOTOXY(1,1);
                INSLINE;
                WITH w^ DO
                  BEGIN;
                    color(line_color);
                    CLREOL;                     (* Spreads color across line *)
                    WRITE(this_line);
                  END;

              END (*----- End scroll up -------------------------------------*)

            ELSE

              BEGIN;

                (*-----------------------------------------------------------*)
                (* Scroll down.  If at the bottom, then we must be done!     *)
                (*-----------------------------------------------------------*)

                IF window_line = window_last THEN
                  EXIT;

                (*-----------------------------------------------------------*)
                (* See if we are at the top.  If so, we might have a buffer  *)
                (* to free if we are over our limit.                         *)
                (*-----------------------------------------------------------*)

                IF window_at_top THEN
                  BEGIN;

                    IF (window_count > window_max_data) THEN
                      BEGIN;

                        (*---------------------------------------------------*)
                        (* Over the limit.  Find the last guy in the list    *)
                        (*---------------------------------------------------*)

                        w := window_line;
                        l := window_line^.last_line;
                        WHILE l <> NIL DO
                          BEGIN;
                            w := l;
                            l := w^.last_line;

                            {$IFDEF LOOP_POINT_CHK}
                              IF w = l THEN
                                BEGIN;
                                  WRITELN('Window loop at scroll ');
                                  dump_string('Wind loop at scroll ' + p2x(w));
                                  dump_window(current_window);
                                  HALT;
                                END;
                            {$ENDIF}

                          END;

                        (*---------------------------------------------------*)
                        (* Unchain it and throw it away                      *)
                        (*---------------------------------------------------*)

                        l := w;
                        w := w^.next_line;
                        w^.last_line := NIL;

                        {$IFDEF LOOP_POINT_CHK}
                          IF w = l THEN
                            BEGIN;
                              WRITELN('Window loop at scroll ');
                              dump_string('Wind loop at scroll ' + p2x(w));
                              dump_window(current_window);
                              HALT;
                            END;
                        {$ENDIF}

                        DISPOSE(l);

                        window_count := window_count - 1;

                      END (*----- End of destroying oldest line -------------*)

                    ELSE

                      (*-----------------------------------------------------*)
                      (* Remove top of window indicator since we ain't gonna *)
                      (* be when we scroll up                                *)
                      (*-----------------------------------------------------*)

                      window_at_top := FALSE;

                  END;

                (*-----------------------------------------------------------*)
                (* Scroll the window up                                      *)
                (*-----------------------------------------------------------*)

                window_line := window_line^.next_line;

                GOTOXY(1, window_location[window_loc].window_wl);

                color(window_line^.line_color);
                WRITELN;

                (*-----------------------------------------------------------*)
                (* Put the bottom line out                                   *)
                (*-----------------------------------------------------------*)

                put_bot_line(window_line, number_of_lines > 0);

              END; (*----- End of scroll down -------------------------------*)

          END; (*----- End of scroll count loop -----------------------------*)

      END; (*----- End with statement ---------------------------------------*)

  END; (*----- End scroll statement -----------------------------------------*)
