(*===========================================================================*)
(* Window write                                                              *)
(*   Put a line to the display for the active TCB                            *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1991 by H. Roy Engehausen.  All rights reserved.  *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Put a line in the window buffer                                           *)
(*    ecnt = number of linefeeds                                             *)
(*    s1 = first string  (usually the prefix)                                *)
(*    s2 = second string                                                     *)
(*===========================================================================*)

PROCEDURE put_wind_line(e_cnt : BYTE; VAR s1 : STRING; VAR s2 : STRING);

  VAR
    b            : BOOLEAN;
    color_to_use : BYTE;
    i            : WORD;
    s            : window_data_area;
    w            : window_data_ptr;

  BEGIN;

    {$IFDEF DEBUG}
      WRITELN('PWL -- ', e_cnt);
      DELAY(1000);
    {$ENDIF}

    color_to_use := active_tcb^.w_color;

    WITH window_array[current_window] DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Add one to line feed count for this line                          *)
        (*-------------------------------------------------------------------*)

        e_cnt := e_cnt + 1;

        (*-------------------------------------------------------------------*)
        (* If we are adding this line at the bottom of a window then         *)
        (* execute any linefeeds                                             *)
        (*-------------------------------------------------------------------*)

        IF (window_line^.next_line = NIL) AND window_act THEN
          BEGIN;
            REPEAT
                e_cnt := e_cnt - 1;
                color(color_to_use);
                WRITELN;
                CLREOL;
            UNTIL e_cnt < 1;
          END;

        (*-------------------------------------------------------------------*)
        (* Strip Blanks and cr                                               *)
        (*-------------------------------------------------------------------*)

        WHILE (LENGTH(s2) > 0) AND (s2[LENGTH(s2)] = cr) DO
          s2[0] := CHR(LENGTH(s2) - 1);

        strip_var(s2, 'T');

        (*-------------------------------------------------------------------*)
        (* Make sure the second string cannot be empty                       *)
        (*-------------------------------------------------------------------*)

        IF s2 = '' THEN
          s2 := ' ';

        (*-------------------------------------------------------------------*)
        (* Now start to build the output strings                             *)
        (*-------------------------------------------------------------------*)

        s := s1;

        WHILE s2 <> '' DO
          BEGIN;

            (*---------------------------------------------------------------*)
            (* Calculate how long the line can be                            *)
            (*---------------------------------------------------------------*)

            i  := 80 - LENGTH(s);

            (*---------------------------------------------------------------*)
            (* Snatch that much data from the remaining info to be put out   *)
            (*---------------------------------------------------------------*)

            s  := s + SUBSTR(s2, 1, i);

            (*---------------------------------------------------------------*)
            (* If the next thing after the data line we just got is a CR     *)
            (* then throw away that CR                                       *)
            (*---------------------------------------------------------------*)

            IF (LENGTH(s2) > i) AND (s2[i] <> cr) AND (s2[i+1] = cr) THEN
              i := i + 1;

            (*---------------------------------------------------------------*)
            (* This is now the remainder of data to be sent                  *)
            (*---------------------------------------------------------------*)

            s2 := SUBSTR(s2, i+1, 0);

            (*---------------------------------------------------------------*)
            (* If this is a scrollable window, then better handle it         *)
            (*---------------------------------------------------------------*)

            IF window_scrollable THEN
              BEGIN;

                (*-----------------------------------------------------------*)
                (* Get a buffer for the new line and initialize it           *)
                (*-----------------------------------------------------------*)

                w := get_new_window_line;

                WITH w^ DO
                  BEGIN;

                    next_line  := NIL;
                    last_line  := window_last;

                    this_line  := s;
                    line_color := color_to_use;

                  END;

                (*-----------------------------------------------------------*)
                (* Is the window positioned at the bottom?                   *)
                (*-----------------------------------------------------------*)

                b := window_last = window_line;

                (*-----------------------------------------------------------*)
                (* In any case make the bottom window line the new line we   *)
                (* just added.                                               *)
                (*-----------------------------------------------------------*)

                {$IFDEF LOOP_POINT_CHK}
                  IF w = window_last THEN
                    BEGIN;
                      WRITELN('Window loop put wind ');
                      dump_string('Wind loop at put wind ' + p2x(w));
                      dump_window(current_window);
                      HALT;
                    END;
                {$ENDIF}

                window_last^.next_line := w;
                window_last            := w;

                (*-----------------------------------------------------------*)
                (* If the window was scrolled to the bottom then we can      *)
                (* immediately write it out                                  *)
                (*-----------------------------------------------------------*)

                IF b THEN
                  BEGIN;

                    (*-------------------------------------------------------*)
                    (* Set the new window current line                       *)
                    (*-------------------------------------------------------*)

                    window_line := w;

                    (*-------------------------------------------------------*)
                    (* If the number of data lines now exceeds the window    *)
                    (* size, we turn off the top flag                        *)
                    (*-------------------------------------------------------*)

                    IF window_at_top
                            AND (window_count >
                                   window_location[window_loc].window_wl) THEN
                      window_at_top := FALSE;

                    (*-------------------------------------------------------*)
                    (* Finally, put the bottom line                          *)
                    (*-------------------------------------------------------*)

                    put_bot_line(w, NOT (LENGTH(s2) = 0));

                  END; (*----- End putting the bottom line ------------------*)

              END (*----- End scrollable window -----------------------------*)
            ELSE
              BEGIN;

                (*-----------------------------------------------------------*)
                (* Not a scrollable window so just write it out              *)
                (*-----------------------------------------------------------*)

                IF window_act THEN
                  BEGIN;
                    color(color_to_use);
                    WRITE(s);
                  END;

              END;

            (*---------------------------------------------------------------*)
            (* Clear away any prefix data, etc                               *)
            (*---------------------------------------------------------------*)

            s := ''

          END; (*----- End loop doing window writes -------------------------*)

      END;
    END;

(*===========================================================================*)
(* Write a line to the current window                                        *)
(*===========================================================================*)

PROCEDURE window_write_current(VAR put_prefix: STRING; VAR put_string: STRING);

  VAR
    i : BYTE;
    t_string : STRING;

  BEGIN;

    {$IFDEF DEBUG}
      WRITELN('WWC start');
      DELAY(1000);
    {$ENDIF}


    (*-----------------------------------------------------------------------*)
    (* Remove bells!                                                         *)
    (*-----------------------------------------------------------------------*)

    REPEAT
      i := POS(bell, put_string);
      IF i <> 0 THEN
        put_string[i] := #0;
    UNTIL i = 0;

    (*-----------------------------------------------------------------------*)
    (* Loop putting each string                                              *)
    (*-----------------------------------------------------------------------*)

    WHILE LENGTH(put_string) > 0 DO
      BEGIN;

        {$IFDEF DEBUG}
          WRITELN('WWC loop -- ', LENGTH(put_string));
          DELAY(1000);
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* If possible, break on a CR                                        *)
        (*-------------------------------------------------------------------*)

        i := POS(cr, put_string);
        IF i <> 0 THEN
          BEGIN;
            t_string   := COPY(put_string,   1,   i);
            put_string := COPY(put_string, i+1, 255);
          END
        ELSE
          BEGIN;
            t_string   := put_string;
            put_string := '';
          END;

        (*-------------------------------------------------------------------*)
        (* Remove a linefeed that is in front of the CR                      *)
        (*-------------------------------------------------------------------*)

        i := LENGTH(t_string);
        IF i >= 2 THEN
          IF (t_string[i-1] = lf) AND (t_string[i] = cr) THEN
            BEGIN;
              t_string[i-1] := cr;
              t_string[0]   := CHR(i - 1);
            END;

        (*-------------------------------------------------------------------*)
        (* Remove any linefeeds at the front.                                *)
        (*-------------------------------------------------------------------*)

        i := 0;
        WHILE t_string[1] = lf DO
          BEGIN;
            INC(i);
            IF LENGTH(t_string) > 1 THEN
              t_string := COPY(t_string, 2, 255)
            ELSE
              t_string[1] := cr;
          END;

        (*-------------------------------------------------------------------*)
        (* Readjust the linefeeds                                            *)
        (*-------------------------------------------------------------------*)

        IF i > 0 THEN
          DEC(i);

        (*-------------------------------------------------------------------*)
        (* Put the line                                                      *)
        (*-------------------------------------------------------------------*)

        put_wind_line(i, put_prefix, t_string);

        (*-------------------------------------------------------------------*)
        (* Clear the prefix for the next time around                         *)
        (*-------------------------------------------------------------------*)

        put_prefix := '';

      END; (*----- End loop putting -----*)

  END;


(*===========================================================================*)
(* Write a line to this tasks window                                         *)
(*===========================================================================*)

PROCEDURE window_write(put_prefix, put_string : STRING);

  VAR
    i : BYTE;
    j : WORD;
    k : BYTE;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Put us in the right window                                            *)
    (*-----------------------------------------------------------------------*)

    i := active_tcb^.window;

    {$IFDEF DEBUG}
      WRITELN('Window write -- ', i);
      DELAY(1000);
    {$ENDIF}

    IF i = 0 THEN EXIT;

    (*-----------------------------------------------------------------------*)
    (* Suppress prefix as needed                                             *)
    (*-----------------------------------------------------------------------*)

    IF opt_block.opt_suppress_pfx AND (i = window_operator) THEN
      put_prefix := COPY(put_prefix, 5, 255);

    (*-----------------------------------------------------------------------*)
    (* Write the capture file if any                                         *)
    (*-----------------------------------------------------------------------*)

    IF i = capture_this_window THEN
      BEGIN;

        {$IFDEF DEBUG}
          WRITELN('Window capture -- ', i);
          DELAY(1000);
        {$ENDIF}

        (*-------------------------------------------------------------------*)
        (* Write the prefix                                                  *)
        (*-------------------------------------------------------------------*)

        WRITE(capture_file, put_prefix);

        (*-------------------------------------------------------------------*)
        (* Loop looking for CRs                                              *)
        (*-------------------------------------------------------------------*)

        j := 1;
        REPEAT
          k := POS(cr, COPY(put_string, j, 255));
          IF k = 0 THEN
            BEGIN;
              WRITELN(capture_file, COPY(put_string, j, 255));
              k := 255;
            END
          ELSE
            BEGIN;
              WRITELN(capture_file, COPY(put_string, j, k-1));
              INC(k, j-1);
            END;

          {$IFDEF DEBUG}
          WRITELN('K1=', k, ' J1=', j);
          DELAY(500);
          {$ENDIF}

          IF (k < LENGTH(put_string)) AND (put_string[k+1] = lf) THEN
            INC(k);

          j := k + 1;

          {$IFDEF DEBUG}
          WRITELN('K2=', k, ' J2=', j, ' L2=', LENGTH(put_string));
          DELAY(500);
          {$ENDIF}

        UNTIL j > WORD(LENGTH(put_string));

      END;

    (*-----------------------------------------------------------------------*)
    (* Select it                                                             *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG}
      WRITELN('Window select -- ', i);
      DELAY(1000);
    {$ENDIF}

    window_select(i);

    (*-----------------------------------------------------------------------*)
    (* Write to window                                                       *)
    (*-----------------------------------------------------------------------*)

    {$IFDEF DEBUG}
      WRITELN('Window write -- ', i);
      DELAY(1000);
    {$ENDIF}

    window_write_current(put_prefix, put_string);

  END;

(*===========================================================================*)
(* Select critical window                                                    *)
(*===========================================================================*)

PROCEDURE select_critical_window;

  VAR
    this_win : BYTE;

  BEGIN;

    IF edit_window_busy THEN
      this_win := who_is_in_window(window_top_screen)
    ELSE
      this_win := who_is_in_window(window_bottom_screen);

    IF NOT window_array[this_win].window_act THEN
      this_win := window_operator;

    (*-----------------------------------------------------------------------*)
    (* Select it                                                             *)
    (*-----------------------------------------------------------------------*)

    window_select(this_win);

  END;

(*===========================================================================*)
(* Write a line to a window.  Its critical!                                  *)
(*===========================================================================*)

PROCEDURE window_write_critical(put_prefix, put_string : STRING);

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Select critical window                                                *)
    (*-----------------------------------------------------------------------*)

    select_critical_window;

    (*-----------------------------------------------------------------------*)
    (* Write to window                                                       *)
    (*-----------------------------------------------------------------------*)

    put_prefix := 'CRIT*> ' + put_prefix;

    window_write_current(put_prefix, put_string);

  END;

(*===========================================================================*)
(* Write a line to a window.  Its critical!  This call has a number attached *)
(*===========================================================================*)

PROCEDURE window_write_critical_i(put_prefix : STRING; put_number : LONGINT);

  VAR
    s        : STRING;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Select critical window                                                *)
    (*-----------------------------------------------------------------------*)

    select_critical_window;

    (*-----------------------------------------------------------------------*)
    (* Convert number                                                        *)
    (*-----------------------------------------------------------------------*)

    STR(put_number, s);

    (*-----------------------------------------------------------------------*)
    (* Write to window                                                       *)
    (*-----------------------------------------------------------------------*)

    put_prefix := 'CRIT*> ' + put_prefix;

    window_write_current(put_prefix, s);

  END;

(*===========================================================================*)
(* Write the bottom line to the real screen                                  *)
(*===========================================================================*)

PROCEDURE put_bot_line(in_w : window_data_ptr; more_to_come : BOOLEAN);

  VAR
    w: window_data_ptr;

  BEGIN;

    {$IFDEF DEPTH}
    stack_depth;
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* If its not active, we can leave now                                   *)
    (*-----------------------------------------------------------------------*)

    IF NOT window_array[current_window].window_act THEN
      EXIT;

    (*-----------------------------------------------------------------------*)
    (* Set addressability                                                    *)
    (*-----------------------------------------------------------------------*)

    WITH window_location[window_array[current_window].window_loc], in_w^ DO
      BEGIN;

        (*-------------------------------------------------------------------*)
        (* Write the line out                                                *)
        (*-------------------------------------------------------------------*)

        color(line_color);
        CLREOL;
        WRITE(this_line);

        (*-------------------------------------------------------------------*)
        (* We are done if any of this is true                                *)
        (*     1) The line wasn't 80 bytes long                              *)
        (*     2) There is more to come                                      *)
        (*     3) The last character was a CR                                *)
        (*-------------------------------------------------------------------*)

        IF (LENGTH(this_line) <> 80) OR more_to_come
                                                  OR (this_line[80] = cr) THEN
          EXIT;

        (*-------------------------------------------------------------------*)
        (* This code handles the case of a full line and the screen          *)
        (* upspacing when we didn't want it to.  We push the creen down and  *)
        (* rewrite the top line                                              *)
        (*-------------------------------------------------------------------*)

        (*-------------------------------------------------------------------*)
        (* Locate top line of window                                         *)
        (*-------------------------------------------------------------------*)

        w := find_window_top_line;

        (*-------------------------------------------------------------------*)
        (* Push the window down 1 line                                       *)
        (*-------------------------------------------------------------------*)

        GOTOXY(1,1);
        INSLINE;

        (*-------------------------------------------------------------------*)
        (* Write the line out                                                *)
        (*-------------------------------------------------------------------*)

        IF w <> NIL THEN
          BEGIN;
            color(w^.line_color);
            WRITE(w^.this_line);
          END;

        GOTOXY(80, window_wl);

        (*-------------------------------------------------------------------*)
        (* Position cursor back where it belongs                             *)
        (*-------------------------------------------------------------------*)

      END;

  END;
