(*===========================================================================*)
(* X/Ymodem Protocol processor -- Common                                     *)
(*                                                                           *)
(*   Copyright 1989, 1990 by H. Roy Engehausen.  All rights reserved.        *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* Show size so far                                                          *)
(*===========================================================================*)

PROCEDURE show_size;

  VAR
    s : STRING[12];

  BEGIN;

    STR(curr_size, s);
    window_write(p, s + ' bytes transferred');

  END;

(*===========================================================================*)
(* Build checksum                                                            *)
(*===========================================================================*)

PROCEDURE build_sum;

  VAR
    i : WORD;
    j : WORD;
    s : BYTE;

  BEGIN;

    s := 0;

    j := b_size AND $FFF0;

    FOR i := 1 TO j DO
      s := $FF AND (s + ORD(big_buff^.y_b[i]));

    check_num := s;

  END;

(*===========================================================================*)
(* Build CRC                                                                 *)
(*===========================================================================*)

PROCEDURE build_crc;

  VAR

    crc    : WORD;
    i      : WORD;
    j      : BYTE;
    k      : WORD;

  BEGIN;

    crc := 0;

    k := b_size AND $FFF0;

    FOR i := 1 TO k DO
      BEGIN;

        crc := crc XOR (ORD(big_buff^.y_b[i]) SHL 8);

        FOR j := 1 TO 8 DO
          IF (crc AND $8000) <> 0 THEN
            crc := (($7FFF AND crc) SHL 1) XOR $1021
          ELSE
            crc := ($7FFF AND crc) SHL 1;

      END;

    check_num := crc;

  END;

(*===========================================================================*)
(* Receive a block                                                           *)
(*===========================================================================*)

PROCEDURE get_a_block;

  CONST
    debug = FALSE;

  TYPE
    buff_array = ARRAY[0..2048] OF CHAR;

  VAR
    b          : BOOLEAN;
    buff       : buff_ptr;
    buff_a_ptr : ^buff_array;
    buff_cntr  : WORD;
    buff_size  : WORD;
    cnt        : WORD;
    i          : BYTE;
    time_out   : LONGINT;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Empty buffer                                                          *)
    (*-----------------------------------------------------------------------*)

    send_flush;

    (*-----------------------------------------------------------------------*)
    (* Initialize counter                                                    *)
    (*-----------------------------------------------------------------------*)

    buff_cntr := 0;

    to_sw     := FALSE;

    (*-----------------------------------------------------------------------*)
    (* Select buffer                                                         *)
    (*-----------------------------------------------------------------------*)

    IF up_it THEN
      BEGIN;
        buff      := big_buff;
        buff_size := SIZEOF(buff_type);
      END
    ELSE
      BEGIN;
        buff      := small_buff;
        buff_size := small_buff_size;
      END;

    IF NOT crc_sw THEN
      DEC(buff_size);

    (*-----------------------------------------------------------------------*)
    (* Clear buffer                                                          *)
    (*-----------------------------------------------------------------------*)

    FILLCHAR(buff^, buff_size, 0);

    (*-----------------------------------------------------------------------*)
    (* Loop for whole block                                                  *)
    (*-----------------------------------------------------------------------*)

    buff_a_ptr := POINTER(buff);

IF debug THEN
WRITELN('Get = ', buff_size);

    REPEAT

      (*---------------------------------------------------------------------*)
      (* Set time out                                                        *)
      (*---------------------------------------------------------------------*)

      time_out := up_time_from_now(time_to_wait);

      (*---------------------------------------------------------------------*)
      (* Loop until data appears or we time out                              *)
      (*---------------------------------------------------------------------*)

      REPEAT

        task_switch;

        i := send_pending(FALSE);

      UNTIL (active_tcb^.i_data.long_length > 0) OR (time_out < up_time);

      (*---------------------------------------------------------------------*)
      (* If not data then we must have timed out                             *)
      (*---------------------------------------------------------------------*)

      IF active_tcb^.i_data.long_length = 0 THEN
        BEGIN;
          to_sw := TRUE;
          EXIT;
        END;

      (*---------------------------------------------------------------------*)
      (* Show binary data as needed                                          *)
      (*---------------------------------------------------------------------*)

      IF opt_block.opt_show_binary THEN
        window_write(active_tcb^.port_chan_s + '<:',
                                                 active_tcb^.i_data.str_data);

      (*---------------------------------------------------------------------*)
      (* Move data from tnc buffer into XY buffer                            *)
      (*---------------------------------------------------------------------*)

      cnt := active_tcb^.i_data.long_length;

IF debug THEN
WRITELN('C1 = ', cnt);

      IF cnt > (buff_size - buff_cntr) THEN
        cnt := buff_size - buff_cntr;

IF debug THEN
WRITELN('C2 = ', cnt);

      MOVE(active_tcb^.i_data.long_data, buff_a_ptr^[buff_cntr], cnt);

      INC(buff_cntr, cnt);

IF debug THEN
WRITELN('BC = ', buff_cntr);

      (*---------------------------------------------------------------------*)
      (* Clear TNC buffer pointer                                            *)
      (*---------------------------------------------------------------------*)

      active_tcb^.i_data.long_length := 0;

      (*---------------------------------------------------------------------*)
      (* Get the block type                                                  *)
      (*---------------------------------------------------------------------*)

      block_type := buff^.xy_type;

IF debug THEN
WRITELN('BT = ', block_type);

      (*---------------------------------------------------------------------*)
      (* If not a data buffer or if buffer is full then we are done          *)
      (*---------------------------------------------------------------------*)

      b := ((block_type <> soh) AND (block_type <> stx))
                                                    OR (buff_cntr = buff_size);

IF debug THEN
WRITELN('B1 = ', b);

      (*---------------------------------------------------------------------*)
      (* If a 128 byte data block then we could also be done                 *)
      (*---------------------------------------------------------------------*)

      IF (NOT b) AND (block_type = soh) THEN
        BEGIN;
          IF (NOT crc_sw) AND (buff_cntr >= (3 + 128)) THEN
            b := TRUE
          ELSE
            IF buff_cntr >= (4 + 128) THEN
              b := TRUE;
        END;

IF debug THEN
WRITELN('B2 = ', b);

      (*---------------------------------------------------------------------*)
      (* Loop until done                                                     *)
      (*---------------------------------------------------------------------*)

    UNTIL b

  END;

(*===========================================================================*)
(* Test a block against the different checks                                 *)
(*===========================================================================*)

FUNCTION check_block(buff : buff_ptr) : BOOLEAN;

  VAR
    b : BOOLEAN;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Assume good                                                           *)
    (*-----------------------------------------------------------------------*)

    check_block := TRUE;

    (*-----------------------------------------------------------------------*)
    (* Get block type                                                        *)
    (*-----------------------------------------------------------------------*)

    block_type := buff^.xy_type;

    (*-----------------------------------------------------------------------*)
    (* Handle non-data blocks                                                *)
    (*-----------------------------------------------------------------------*)

    IF (block_type <> soh) AND (block_type <> stx) THEN
      BEGIN

        (*-------------------------------------------------------------------*)
        (* If not a cancel then we should have only one char followed by null*)
        (*-------------------------------------------------------------------*)

        IF (block_type <> can)
                AND (buff^.xy_b_no <> 0)
                AND (buff^.xy_b_no_invert <> 0) THEN
          BEGIN;
            do_mess_local(message_bin_proto_err);
            check_block := FALSE;
          END;

        (*-------------------------------------------------------------------*)
        (* All others are AOK                                                *)
        (*-------------------------------------------------------------------*)

        EXIT;

      END;

    (*-----------------------------------------------------------------------*)
    (* Check data block number                                               *)
    (*-----------------------------------------------------------------------*)

    IF buff^.xy_b_no <> (NOT buff^.xy_b_no_invert) THEN
      BEGIN;
        do_mess_local(message_bin_bad_blkno);
        check_block := FALSE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Check tail area                                                       *)
    (*-----------------------------------------------------------------------*)

    IF block_type = soh THEN
      BEGIN;
        b_size   := 128;
        tail_ptr := @buff^.x_tail;
      END
    ELSE
      BEGIN;
        b_size   := 1024;
        tail_ptr := @buff^.y_tail;
      END;

    IF crc_sw THEN
      BEGIN;
        build_crc;
        b := (tail_ptr^.xy_crc_hi = HI(check_num))
                                     AND (tail_ptr^.xy_crc_lo = LO(check_num));
      END
    ELSE
      BEGIN;
        build_sum;
        b := tail_ptr^.xy_chksum = check_num;
      END;

    (*-----------------------------------------------------------------------*)
    (* If a goof, tell everybody                                             *)
    (*-----------------------------------------------------------------------*)

    IF NOT b THEN
      BEGIN;
        do_mess_local(message_bin_bad_block);
        check_block := FALSE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* If a header, reset the header switch                                  *)
    (*-----------------------------------------------------------------------*)

    IF first_sw AND (buff^.xy_b_no = 0) THEN
      BEGIN;
        EXIT;
      END;

    first_sw := FALSE;

    (*-----------------------------------------------------------------------*)
    (* Check for valid block number                                          *)
    (*-----------------------------------------------------------------------*)

    IF buff^.xy_b_no <> block_no THEN
      BEGIN;
        do_mess_local(message_bin_bad_blkno);
        check_block := FALSE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* If we reach here then everything is AOK                               *)
    (*-----------------------------------------------------------------------*)

  END;

(*===========================================================================*)
(* Show header                                                               *)
(*===========================================================================*)

PROCEDURE show_header;

  VAR
    s : STRING;

  BEGIN;

    set_dollar1_parm(@bin_mode);
    s := get_message(message_bin_hdr) + ' ' + header_str;
    window_write(p, s);

  END;
