(*===========================================================================*)
(* X/Ymodem Protocol processor -- Upload                                     *)
(*                                                                           *)
(*   Copyright 1989 by H. Roy Engehausen.  All rights reserved.              *)
(*                                                                           *)
(*===========================================================================*)

PROCEDURE xy_upload;

  CONST debug = FALSE;

  VAR
    b         : BOOLEAN;
    i         : BYTE;
    ok_proto  : BOOLEAN;
    s         : STRING[10];
    size_left : LONGINT;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Open the block for write                                              *)
    (*-----------------------------------------------------------------------*)

    REWRITE(data_file^, 1);
    free_semaphore(semaphore_interrupts);

    (*-----------------------------------------------------------------------*)
    (* Initialize switches                                                   *)
    (*-----------------------------------------------------------------------*)

    first_sw  := TRUE;

    block_no  := 1;

    curr_size := 0;
    file_size := 999999999;

    (*-----------------------------------------------------------------------*)
    (* Wait before the first C                                               *)
    (*-----------------------------------------------------------------------*)

    task_wait (15, FALSE);

    (*-----------------------------------------------------------------------*)
    (* Send the CRC C if necessary and wait for an answer                    *)
    (*-----------------------------------------------------------------------*)

    IF crc_sw THEN
      BEGIN;

        error_cnt := 0;

        REPEAT

          send_tnc_data_str('C');
          send_drain;

          get_a_block;

          INC(error_cnt);

IF debug THEN
WRITELN('BLK1 IN = ', ORD(big_buff^.xy_type));

        UNTIL (NOT to_sw) OR (error_cnt = 10);

      END;

    (*-----------------------------------------------------------------------*)
    (* If timeout or this is not CRC mode then send the beginning NAK        *)
    (*-----------------------------------------------------------------------*)

    IF to_sw OR NOT crc_sw THEN
      BEGIN;

        error_cnt := 0;

        REPEAT

          send_tnc_data_str(nak);
          send_drain;
IF debug THEN
  WRITELN('Sent NAK1');

          get_a_block;

          INC(error_cnt);

IF debug THEN
WRITELN('BLK2 IN = ', ORD(big_buff^.xy_type));

        UNTIL (NOT to_sw) OR (error_cnt = 10);

      END;

    (*-----------------------------------------------------------------------*)
    (* No response to beginning NAK.  Dump connection                        *)
    (*-----------------------------------------------------------------------*)

    IF to_sw THEN
      BEGIN;
        do_mess(message_bin_time_out);
        abort_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Receive loop                                                          *)
    (*-----------------------------------------------------------------------*)

    error_cnt := 0;

    REPEAT

IF debug THEN
WRITELN('BLK3 IN = ', ORD(big_buff^.xy_type));

      ok_proto := FALSE;

      b := check_block(big_buff);

      IF (block_type = stx) OR (block_type = soh) THEN
        BEGIN;

          ok_proto := TRUE;

          IF NOT b THEN
            BEGIN;
              send_tnc_data_str(nak);
IF debug THEN
  WRITELN('Sent NAK2');
            END
          ELSE
            BEGIN;

              send_tnc_data_str(ack);
              send_flush;
IF debug THEN
  WRITELN('Sent ACK3');

              size_left := file_size - curr_size;

              IF big_buff^.xy_b_no = 0 THEN
                BEGIN;

                  IF batch_sw THEN
                    BEGIN;
                      send_tnc_data_str('C');
                      send_flush;
IF debug THEN
  WRITELN('Sent C4');
                    END;

                  i := SIZEOF(header_str) - 1;
                  MOVE(big_buff^.x_b, header_str[1], i);

                  WHILE (i > 0)
                          AND ((header_str[i] = '') OR (header_str[i] = #0)) DO
                    DEC(i);

                  header_str[0] := CHR(i);

                  WHILE i > 0 DO
                    BEGIN;
                      IF header_str[i] = #0 THEN
                        header_str[i] := ' ';
                      DEC(i);
                    END;

                  show_header;

                  s := subword(@header_str, 2, 1);
                  VAL(s, file_size, check_num);

                  IF check_num <> 0 THEN
                    file_size := 999999999;

                  block_no := 0;

                END
              ELSE
                BEGIN;
                  IF block_type = soh THEN
                    BEGIN;
                      IF size_left > 128 THEN
                        size_left := 128;
                    END
                  ELSE
                    IF size_left > 1024 THEN
                      size_left := 1024;
                  BLOCKWRITE(data_file^, big_buff^.x_b, size_left);
                  INC(curr_size, size_left);

                END;

              show_size;

              IF block_no = 255 THEN
                block_no := 0
              ELSE
                INC(block_no);

              send_drain;

            END;
        END
      ELSE
        IF b THEN
          BEGIN

            IF block_type = eot THEN
              BEGIN;

                s[1] := eof;
                WHILE (curr_size AND $FF) <> 0 DO
                  BEGIN;
                    BLOCKWRITE(data_file^, s[1], 1);
                    INC(curr_size);
                  END;

                send_tnc_data_str(ack);
                send_flush;
IF debug THEN
  WRITELN('Sent ACK5');

                EXIT;

              END;

            IF block_type = can THEN
              BEGIN;

                do_mess_local(message_bin_abort_in);
                abort_sw := TRUE;
                EXIT;

              END;

          END;

      IF NOT ok_proto THEN
        BEGIN;
          send_tnc_data_str(nak);
IF debug THEN
  WRITELN('Sent NACK6');
          do_mess_local(message_bin_proto_err);
        END;

      get_a_block;

    UNTIL FALSE; (*----- Never ending receive loop --------------------------*)

  END;
