(*===========================================================================*)
(* YAPP Protocol processor -- AA4RE routines                                 *)
(*                                                                           *)
(*   Copyright 1989, 1991, 1992 by H. Roy Engehausen.  All rights reserved.  *)
(*                                                                           *)
(*===========================================================================*)

(*===========================================================================*)
(* These routines are all compatibility for YAPP mainline                    *)
(*===========================================================================*)

(*===========================================================================*)
(* Check for yapp timeout                                                    *)
(*===========================================================================*)

FUNCTION timeout  : BOOLEAN;
  BEGIN;
    timeout := yapp_timer <= up_time;
  END;

(*===========================================================================*)
(* Set yapp timer                                                            *)
(*===========================================================================*)

PROCEDURE set_timer (time_to_go : BYTE);
  BEGIN;
    yapp_timer := up_time_from_now(time_to_go* opt_block.yapp_time_exp);
  END;

(*===========================================================================*)
(* Query if anything is in input buffer                                      *)
(*===========================================================================*)

FUNCTION inready  : BOOLEAN;

  VAR
    i : BYTE;

  BEGIN;

    IF active_tcb^.i_data.long_length = 0 THEN
      BEGIN;

        i := send_pending(FALSE);
        IF opt_block.opt_show_binary
                              AND (active_tcb^.i_data.long_length > 0) THEN
          window_write(active_tcb^.port_chan_s + '<:',
                                              active_tcb^.i_data.str_data);

      END;

    inready := active_tcb^.i_data.long_length > 0

  END;

(*===========================================================================*)
(* Receive one character                                                     *)
(*===========================================================================*)

FUNCTION recvchar : CHAR;

  VAR
    i : INTEGER;

  BEGIN;

    i := active_tcb^.i_data.long_length - 1;
    IF i < 0 THEN
      BEGIN;
        WRITELN('YAPP recvchar with empty buffer');
        RUNERROR(yapp_error);
      END;

    recvchar := active_tcb^.i_data.long_data[1];

    IF i > 0 THEN
      MOVE(active_tcb^.i_data.long_data[2], active_tcb^.i_data.long_data[1], i);

    active_tcb^.i_data.long_length := i;

    active_tcb^.i_data.str_data[0] := CHR(min_w(255, i));

  END;

(*===========================================================================*)
(* Show a YAPP message                                                       *)
(*===========================================================================*)

PROCEDURE showmsg (mess_no : BYTE);

  VAR s : STRING;

  BEGIN;

    CASE mess_no OF
      1      : mess_no := message_bin_started;
      8      : mess_no := message_bin_done;
      10     : mess_no := message_bin_abort_out;
      11     : mess_no := message_bin_abort_in;
      12     : mess_no := message_bin_time_out;
      13     : mess_no := message_bin_proto_err;
      ELSE
        BEGIN;
          STR(mess_no, s);
          window_write(p + 'YAPP message # ', s);
          EXIT;
        END;
    END;

    do_mess_local(mess_no);

  END;

(*===========================================================================*)
(* Show a YAPP message -- Escaping                                           *)
(*===========================================================================*)

PROCEDURE escmsg (mess_no : BYTE);
  BEGIN;

    showmsg(mess_no);

  END;

(*===========================================================================*)
(* Transmit a string                                                         *)
(*===========================================================================*)

PROCEDURE xmitstr(out_str : STRING);
  BEGIN;

    send_tnc_data_ub(@out_str[1], LENGTH(out_str));
    send_flush;
    send_switch := TRUE;

  END;

(*===========================================================================*)
(* Transmit a line.  Used by YAPP to send commands to TNC.  Ignore these     *)
(*===========================================================================*)

PROCEDURE xmitline( b : STRING);
  BEGIN;
    (* NOOP *)
  END;

(*===========================================================================*)
(* Put TNC in command mode.  Ignore this                                     *)
(*===========================================================================*)

PROCEDURE cmdmode; INLINE ($90);

(*===========================================================================*)
(* Display a packet because of error                                         *)
(*===========================================================================*)

PROCEDURE disppacket;
  BEGIN;
    WRITELN('DISPPACKET');
  END;

(*===========================================================================*)
(* Show file header                                                          *)
(*===========================================================================*)

PROCEDURE showheader;
  VAR
    s : STRING;

  BEGIN;

    set_dollar1_parm(@bin_mode);
    s := get_message(message_bin_hdr);

    window_write(p, s + ' ' + xferhdr);

  END;

(*===========================================================================*)
(* Show bytes transferred                                                    *)
(*===========================================================================*)

PROCEDURE showbytes;
  VAR
    s : STRING[10];

  BEGIN;
    STR(xfercnt, s);
    window_write(p, s + ' bytes transferred');
  END;

(*===========================================================================*)
(* Show state of YAPP transfer                                               *)
(*===========================================================================*)

PROCEDURE showstate;

  CONST
    yapp_state_str =  ('SendInit' +
                      ' Initretry' +
                      ' SendHeader' +
                      ' SendData' +
                      ' SendEof' +
                      ' SendEOT' +
                      ' RcvWait' +
                      ' RcvHeader' +
                      ' RcvData' +
                      ' SndABORT' +
                      ' WaitAbtAck' +
                      ' RcdABORT' +
                      ' Start');

  VAR
    s : STRING;

  BEGIN;

    IF opt_block.opt_yapp_state THEN
      BEGIN;
        s := yapp_state_str;
        window_write(p, 'YAPP state is ' + subword(@s, 1 + ORD(state), 1));
      END;

  END;

(*===========================================================================*)
(* Delay some number of milliseconds                                         *)
(*===========================================================================*)

PROCEDURE yapp_delay(delay_time : WORD);
  BEGIN;
    delay_time := (delay_time + 500) div 1000;
    IF delay_time > 0 THEN
      task_wait(delay_time, TRUE)
    ELSE
      task_switch;
  END;

(*===========================================================================*)
(* Flush the transmit buffer                                                 *)
(*===========================================================================*)

PROCEDURE flush;
  BEGIN;
    send_flush;
  END;
