/*
   bmh.c : Copyright Paul Healy, EI9GL, 1992.

   Derived from bm source. See the file copyright.bm for conditions of use.

   920709 : Created.
   920825 : Mucked around with haveread1 logic a bit.
   930413 : Redid to use Mailfile struct.
*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "config.h"
#include "misc.h"
#include "help.h"
#include "pc.h"
#include "rc.h"
#include "spawn.h"
#include "mailer.h"
#include "buf.h"
#include "mailfile.h"
#include "bmh.h"
#include "biff.h"
#include "extract.h"

#ifdef BMH
#define main bmh_main
#endif

static int bmhhelp(int argc, char *argv[]);
static int shell(int argc, char *argv[]);
static int dummy(int argc, char *argv[]);
static int readmsg(Mailfile *m, int argc, char *argv[]);
static int notesfile(Mailfile *m, int argc, char *argv[]);
static int groups(Mailfile *m, int argc, char *argv[]);
static int xit(int argc, char *argv[]);
static int savemsg(Mailfile *m, int argc, char *argv[]);
static int writemsg(Mailfile *m, int argc, char *argv[]);
static int printmsg(Mailfile *m, int argc, char *argv[]);
static int delete(Mailfile *m, int argc, char *argv[]);
static int move(Mailfile *m, int argc, char *argv[]);
static int undelete(Mailfile *m, int argc, char *argv[]);
static int sync(int argc, char *argv[]);

static struct cmdstruct {
   char cmd;
   MAINFN main; /* main(argc, argv) style function */
   MAILFN mail; /* x_main(m, argc, argv) style function */
} cmds[] = {
  '?',   bmhhelp,    NULL,
  '.',   NULL,       show,
  '!',   shell,      NULL,
  '#',   dummy,      NULL,
  '\0',  NULL,       readmsg,
  '\n',  NULL,       readmsg,
  'a',   ali_main,   NULL,
  'b',   NULL,       dist,
  'd',   NULL,       delete,
  'f',   NULL,       forw,
  'g',   NULL,       groups,
  'h',   NULL,       scan,
  'i',   showissue,  NULL,
  'k',   smtp_main,  NULL,    
  'l',   smtp_main,  NULL,
  'm',   comp_main,  NULL,
  'n',   NULL,       notesfile,
  'p',   NULL,       printmsg,
  'r',   NULL,       repl,
  's',   NULL,       savemsg,
  'u',   NULL,       undelete,
  'v',   pager_main, NULL,
  'w',   NULL,       writemsg,
  'x',   xit,        NULL,
  '$',   sync,       NULL,
  'M',   NULL,       move,
  'N',   folders_main, NULL,
  'P',   NULL,       post,
  'S',   NULL,       pick
};
/* c e j o t y z - = */
const int ncmds = sizeof(cmds) / sizeof(struct cmdstruct);

static int
callcmd(Mailfile *m, char *line)
{
   char *arg[MAXARGS];
   int nargs = parse(rip(line), arg, MAXARGS);
   int ch = nargs > 0 ? arg[0][0] : 0;
   int i;

   if (line == NULL)
      return -1;

   if (ch == 'q')
      return 'q';
   else if (isdigit(ch)) {
      arg[nargs] = arg[0];
      return show(m, nargs+1, arg);
      }
   else 
      for (i=0; i<ncmds; i++)
         if (ch == cmds[i].cmd)
            if (cmds[i].main != NULL)
               return cmds[i].main(nargs, arg);
            else if (cmds[i].mail != NULL)
               return cmds[i].mail(m, nargs, arg);
            else 
               return -1;

   return -1;
}

int
main(int argc, char *argv[])
{
   Mailfile m;
   char line[LINELEN];
   int ch;

   if (ishelp(argc, argv)) {
      printoptions();
      return 0;
      }

   mail_zero(&m);
   
   if (scan(&m, argc, argv) == -1) {
      fprintf(stderr, "bmh: scan failed\n");
      return -1;
      }

   do
      {
      null_check("main loopb 1");
      mail_resetpos(&m);      /* move to start of mailfile for each command */
      null_check("main loop 2");
      if ((ch = callcmd(&m, biff(&m, line, sizeof(line)))) == -1)
         printf("Invalid command - ? for help\n");
      }
   while ( ch != 'q' );

   return mail_savestate(&m);
}

/* ------------------------------------------------------------------- */

static int
bmhhelp(int argc, char *argv[])
{
   (void) argc; (void) argv;
   printf("bmh ");
   version(stdout);
   printf(" d [msglist]         delete a message\n");
   printf(" m                   mail a message\n");
   printf(" s [msglist] [file]  save message in file (default mbox)\n");
   printf(" w [msglist] [file]  save message in file no headers (default mbox)\n");
   printf(" f [msg]             forward message\n");
   printf(" b [msg]             bounce message (remail)\n");
   printf(" r [msg]             reply to a message\n");
   printf(" p [msglst]          print message on printer\n");
   printf(" .                   display current message\n");
   printf(" h                   list message headers in notefile\n");
   printf(" l                   list unsent messages\n");
   printf(" k                   kill unsent messages\n");
   printf(" n [file]            display or change notefile\n");
   printf(" N                   display notefile (verbose listing)\n");
   printf(" g                   display newsgroups\n");
   printf(" #                   read message, where # is the number of the message\n");
   printf(" q                   quit\n");
   printf(" ! cmd               run shell command\n");
   printf(" a [alias]           display mail alias(es)\n");
   printf(" i                   show configuration/version/copyright information\n");
   printf(" M [msglist] file    move mail to another notefile\n");
   printf(" S [search param]    search message file\n");
   printf(" ?                   print this help screen\n");

   return 0;
}

static int
shell(int argc, char *argv[])
{
   argv[argc] = NULL;
#ifdef SWAP
   if ( (argc > 1) && (run_vp(P_WAIT, argv[1], &argv[1]) == -1) )
#else
   if ( (argc > 1) && (spawnvp(P_WAIT, argv[1], &argv[1]) == -1) )
#endif
      perror("bmh");

   return 0;
}

static int
dummy(int argc, char *argv[])
{
   (void) argc; (void) argv;
   printf("'#' is not a bmh command: enter a number, e.g. \n 1\n");

   return 0;
}

static int
readmsg(Mailfile *m, int argc, char *argv[])
{
   return next(m, argc, argv);
}

#ifdef notdef
   if (mail_isread(m, mail_curmsg(m))) {
      next(m, 1, argv);
      mail_setcur(m, mail_curmsg(m)+1);
      }
   else 
      show(m, 1, argv);

   mail_markasread(m, mail_curmsg(m));
#endif

static int
notesfile(Mailfile *m, int argc, char *argv[])
{
   if (argc == 1) {
      argv[1] = "-short";
      folders_main(2, argv);
      putchar('\n');
      }
   else {
      mail_fin(m);
      if (mail_init(m, argv[1]) == -1)
         fprintf(stderr, "bmh: can't change to %s\n", argv[1]);
       else
         scan(m, 1, argv);
      }

   return 0;
}

static int
groups(Mailfile *m, int argc, char *argv[])
{
   (void)m; (void)argc;
   argv[1] = "-news";
   folders_main(2, argv);

   return 0;
}

static int
xit(int argc, char *argv[])
{
   (void) argc; (void) argv;
   printf("bmh: warning mail file may already have been updated\n");
   return 0;
}

static int
savemsg(Mailfile *m, int argc, char *argv[])
{
   return extract(m, argc, argv, EXT_MSG);
}

static int
writemsg(Mailfile *m, int argc, char *argv[])
{
   return extract(m, argc, argv, EXT_BODY);
}

static int
printmsg(Mailfile *m, int argc, char *argv[])
{
#ifdef MSDOS
   argv[argc++] = "prn";
#else
#error "you need to specify a file name that is an alias for the printer"
#endif
   argv[argc] = NULL;

   return extract(m, argc, argv, 1);
}

static int
delete(Mailfile *m, int argc, char *argv[])
{
   rmm(m, argc, argv);

   return mail_reset(m);
}

static int
move(Mailfile *m, int argc, char *argv[])
{
   if (argc > 2) {
      char *tmp = argv[1];
      argv[1] = argv[argc-1];
      argv[argc-1] = tmp;
      }

   refile(m, argc, argv);

   return mail_reset(m);
}

static int
undelete(Mailfile *m, int argc, char *argv[])
{
   (void) m; (void) argc; (void) argv;
   fprintf(stderr, "bmh: undelete is not supported\n");
   return 0;
}

static int
sync(int argc, char *argv[])
{
   (void) argc; (void) argv;
   fprintf(stderr, "bmh: there is no need to sync the mailfile\n");
   return 0;
}


