#ifndef _BOOTPD_H
#define _BOOTPD_H
/*************************************************/
/* Center for Information Technology Integration */
/*           The University of Michigan          */
/*                    Ann Arbor                  */
/*                                               */
/* Dedicated to the public domain.               */
/* Send questions to info@citi.umich.edu         */
/*                                               */
/* BOOTP is documented in RFC 951 and RFC 1048   */
/*************************************************/

/*
 *	Jan 92	Bill Simpson	rationalized header files.
 */

#ifndef _SOCKET_H
#include "socket.h"
#endif

#ifndef _ARP_H
#include "arp.h"
#endif

#define MHOSTS  12     /* max number of 'hosts' structs */
#define BP_MAXDNS 5

struct bootp {
	char	op;			/* packet opcode type */
	char	htype;			/* hardware addr type */
	char	hlen;			/* hardware addr length */
	char	hops;			/* gateway hops */
	int32	xid;			/* transaction ID */
	int16	secs;			/* seconds since boot began */
	int16	unused;
	struct in_addr	ciaddr;		/* client IP address */
	struct in_addr	yiaddr;		/* 'your' IP address */
	struct in_addr	siaddr;		/* server IP address */
	struct in_addr	giaddr;		/* gateway IP address */
	char	chaddr[16];		/* client hardware address */
	char	sname[64];		/* server host name */
	char	file[128];		/* boot file name */
	char	vend[64];		/* vendor-specific area */
};

/*
 * UDP port numbers, server and client.
 */
#define	IPPORT_BOOTPS		67
#define	IPPORT_BOOTPC		68

#define BOOTREQUEST		1
#define BOOTREPLY		2

#define BOOTP_PAD		0
#define BOOTP_SUBNET		1
#define BOOTP_GATEWAY		3
#define BOOTP_DNS		6
#define BOOTP_HOSTNAME		12
#define BOOTP_END		0xff

/*
 * "vendor" data permitted for Stanford boot clients.
 */
struct vend {
        unsigned char  v_magic[4];     /* magic number */
        unsigned long  v_flags;        /* flags/opcodes, etc. */
        unsigned char  v_unused[56];   /* currently unused */
};

#define VM_STANFORD     "STAN"  /* v_magic for Stanford */

/* v_flags values */
#define VF_PCBOOT       1       /* an IBMPC or Mac wants environment info */
#define VF_HELP         2       /* help me, I'm not registered */

struct bootpd_stat {
	int	rcvd;
	int	bad_size;
	int	bad_op;

};

struct host {
	struct host *next;	/* next in chain */
        char name[31];		/* host name (and suffix) */
        unsigned char  htype;	/* hardware type */
        char   haddr[MAXHWALEN];/* hardware address */
        struct in_addr iaddr;	/* internet address */
	char bootfile[32];	/* default boot file name */
};

#define NULLHOST (struct host *) 0

extern struct host *bp_hosts;

extern struct udp_cb *Bootpd_cb;

extern char    homedir[];	 /* bootfile homedirectory */
extern char    defaultboot[];	 /* default file to boot */

extern int32   bp_DefaultDomainNS[]; /* default domain name server */

extern char *ArpNames[];

void bootpd __ARGS((struct iface *iface, struct udp_cb *sock, int cnt));
void bootp_print_packet __ARGS((struct bootp *bp));


/* in bootpdip.c */
extern char bp_ascii[];

int readtab __ARGS((void));
void readtab_shut __ARGS((void));
void bp_log __ARGS((char *fmt,...));
void da_status __ARGS((struct iface *iface));
int da_assign __ARGS((struct iface *iface,char *hwaddr,int32 *ipaddr));
int da_init __ARGS((void));
void da_shut __ARGS((void));
int da_done_net __ARGS((struct iface *iface));
int da_serve_net __ARGS((struct iface *iface,int32 rstart,int32 rend));

#endif
