/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */
/* This file has the generic call setup routines */

/* Call Request assumes to have a set of linked peers and wants to
	be passed the one with out a parent, and we will try to
	find a link to adopt it (sorry couldn't resist) */

#define USERPORT 0
extern unsigned char node_addr[16];
extern struct ax25_addr CONFIG;

call_request(vc)
register struct VCS *vc;
{
	static unsigned int i, cause;
	static unsigned char *altlist;
	static struct ax25_addr l2_x121d, l2_x121, ax_src, ax_dest;
	static struct VCS *vcx;
	int l2_reset(), l2_clear(), l2_send(), l2_recv(), NULLFCN();
	int clr_config(), con_config(), send_config(), recv_config();

	cause=Congestion+120; /* Default to no memory */

	puthex2("Call",vc,"For peer @",vc->peer->parent);
	get_nsap(vc,&ax_dest);
	if ((check_x121(node_addr,vc->addr) != -1)) /* This node, Local call */
		if (callcmp(&ax_dest,&CONFIG)) /* Remote Configuration adr */ {
			vc->RESET=NULLFCN;
			vc->CLEAR=clr_config;
			vc->CONNECTED=con_config;
			vc->RECV=recv_config;
			vc->SEND=send_config;
			vc->P=P2;
			set_p(vc,P4,0);
			return;
		} else
		if (i=find_call(&ax_dest)<0) /* Not a Level 3 user */ {
			vc->RESET=l2_reset;
			vc->CLEAR=l2_clear;
			vc->CONNECTED=NULLFCN;
			vc->SEND=l2_send;
			vc->RECV=l2_recv;

			vcx=vc->peer;
			vc->lcn=0; /* Level 2 User */
			vc->P=P2;
			l2_adr(vcx->addr,&l2_x121d,&l2_x121);
			get_nsap(vcx,&ax_src);
			if (! (vc->parent=L2_CONN(
				USERPORT,	/* Default level 2 user port */
				&ax_src,	/* Source user */
				&ax_dest,	/* Destination */
				&l2_x121d,	/* Network DNIC */
				&l2_x121,	/* Network address */
				&L3CALL[0],	/* Digi via me */
				&vc->digi,	/* Optional digi */
				TRUE) )		/* Level 2 Call */
				) goto error;
			insert_lcn(vc);	 /* Queue call, no alternatives */
			return;
		}
		else /* Local Level 3 Call */ {
			if (i<16) /* A network node, illegal */ {
				cause=0x8000+Access_Barred+0xf5;
				goto error;
			}
			vc->rand=0;	/* Leaving the network */
			vc->addr[0]=vc->peer->addr[0]=0;
			if (! linked_to(&ax_dest,&vc->parent)) /* New Link */ {
				if (! (vc->parent=L2_CONN(
					L3PORT[i],	/* Port */
					&L3CALL[0],	/* Switch is source */
					&L3CALL[i],	/* User is dest. */
					&L3DIGI[i],	/* Optional digi */
					NULL,
					NULL,
					NULL,
					FALSE) )	/* Level 3 Link */
					) goto error;
			}
			if (vc->parent->R == R1) set_p(vc,P2,0);
			else /* Link is not ready, queue call in llcn */ {
				vc->lcn=255; /* Level 3 call pending */
				insert_lcn(vc); /* append call */
			}
			return;
		}
	else /* Remote Level 3 call do routing */ {
		i = find_route(&altlist, vc->addr);

		if (i == -1) /* Modified address, re-route */ {
			vc->alt = 255;	/* Clear, Address changed */
			queue(NULL, call_request, 0, vc);
			return;
		} else
		if (i == 0) /* No Path */ {
			cause=Not_Obtainable;
			goto error;
		}
		i = (altlist[0]>>4) & 0x0f;
		if (!i) /* Internal ERROR */ {
			cause=Not_Obtainable+119;
			goto error;
		}
		vc->lcn = 255;
		vc->tx_queue = (struct datastr *)altlist;
		if ((find_path(vc) == TRUE)) return;
		if (vc->alt == 255) cause=Out_of_Order;	/* No good routes */
		/* else No memory */
	}
error: /* Fall Through, error */
	vc->peer->peer = NULL;
	vc->parent = NULL;
	set_p(vc->peer,P6, cause);
	vcx=vc;
	free_vc(vcx);
}
