/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */

#include <string.h>
#undef NULL
#include <conio.h>
#include <stdio.h>
#define TRUE 1
#define FALSE 0

struct rt_tab {
	char cmd;
	char dta;
	char next[10];
};
#define MaxRT 100
struct rt_tab rtable[MaxRT]={0};
int numrt=0;

struct dnic {
	unsigned char adr[3];
	struct rt_tab *rt;
};
struct dnic rtdn[8];
int ndnics;

unsigned char calls[16][7],
	      digi[16][7],
	      port[16],
	      maxvc[16],
	      l3wi[16];
unsigned int  delay[16];
int rtsize, maxalt;
#define rtmax 256
unsigned char usaalt[15][16],
	      usa[1000],
	      nodeadr[16],
	      alt[8],
/*	      rtable[256], */
              l2txt[2208];	/* Don't use it all, Please! */
int l2txtl;

FILE *fnam;
char zarr[100];

get_alt(alt)
unsigned char alt[];
{
	int i,j,pos, low;
	unsigned char a[16];

	printf("Enter the LINK NUMBER, best choice to worst.\n");
	printf("Exit loop with 0\n");
	for (i=1;i<16;i++) {
		a[i]=0;
		printf("      Alternative # %d ?",i);
		getline(zarr, 80);
		sscanf(zarr," %d ",&j);
		if (j==0) break;
		a[i]=j;
	}
	alt[0]=i-1;	/* Length */
	if (i > 1) {
		low=0;
		pos=1;
		for (j=1;j<i;j++) {
			if (low) {
				alt[pos] = alt[pos] + (a[j] & 0x0f);
				pos++;
			}
			else alt[pos] = (a[j] & 0x0f) << 4;
			low = !low;
		}
	}
	alt[0] = alt[0]<<4;	/* Shift into right pos */
}

int
getnum(s,min,max,deflt)
char *s;
int min, max, deflt;
{
	int x,f;
	f=0;
	do {
		if (f) printf("*** That Number is out of range ***");
		printf("Min= %d, Max= %d, Suggest= %d\n",min,max,deflt);
		printf("%s",s);
		getline(zarr,80);
		sscanf(zarr,"%d",&x);
		f=1;
	} while (x<min || x>max);
	return x;
}

get_call(stn)
char stn[];
{
	int i,j,max;
	char adr[80];

	if (getlin(adr,80) <= 0) return -1;
	max=strlen(adr);
	j = 0; /* Source pos */
	while ((adr[j]==0x20) && (j<max)) j++;
	if (j>=max) return -1;
	for (i=0;i<6;i++) {
		adr[j]=toupper(adr[j]);
		if ((j==max) || (adr[j] == 0x2d))
			stn[i]=0x40;
		else /* Normal ch */
			stn[i]=(adr[j++]<<1);
	}
	stn[i]=0;
	if (adr[j++] == 0x2d && j <= max && adr[j] >= 0x30 && adr[j] <= 0x39) {
		stn[i]=(adr[j++]-0x30)<<1;
		if (j <= max && adr[j] >= 0x30 && adr[j] <= 0x39) {
			stn[i] = stn[i]*10 + (adr[j] - 0x30)<<1;
		}
		if (stn[i] > 30) stn[i]=0;
	}
	return 0;
}

int
getlin(str,max)
char *str;
int max;
{
	int ch, pos;

	ch=pos=0;
	while (ch != 0x0a && pos<max) {
		ch=getchar();
		if (ch == -1) return -1;
		str[pos++]=ch;
		str[pos]=0;
	}
	if (ch== 0x0a) str[pos-1]=0;
	return (strlen(str));
}

int
getline(str,max)
char *str;
int max;
{
	int l;

	while ((l=getlin(str,max)) == 0) ;
	return l;
}

int check_x121(ptrx,adr)
register unsigned char adr[];
unsigned char ptrx[];
{
	static unsigned char i, j, alen, *ptr;

	ptr=ptrx;
	alen=ptr[0] & 0x0f;
	if ( (adr[0] & 0x0f < alen) || (alen == 0) ) return -1;
	j=alen>>1;

	for (i=1; i<=j; i++)
		if (adr[i] != ptr[i]) return -1;

	if (alen & 1 && /* Length is odd, check last nibble */
		 ( ptr[i] &0xf0 != (adr[i] & 0x0f0) )) return -1;
	if (alen == adr[0]&0x0f) return 0;
	return 1;
}

l3_adr(str,x121)
char str[];
unsigned char x121[16];
{
	int i,len,ch,hi;
	char x[6];

	x121[0]=len=0;
	for (len=0;len<7;len++) /* Insure valid numbers */ {
		ch=str[len];
		if (ch == 0) {
			break; /* End of "callsign" */
		}
		if ((ch < 0x30) || (ch > 0x39)) return; /* Non-digit */
	}
	x121[0]=4;
	x121[1]=0x31;
	x121[2]=0x00;
	x[0]=len;
	ch=0;
	hi=0;
	for (i=0;i<len;i++) {
		hi = !hi;
		if (hi) ch = (str[i] - 0x30) << 4;
		else {
			ch = ch | (str[i] - 0x30);
			x[ (i>>1) + 1 ]= ch;
		}
	}
	if (hi) x[ (i>>1) + 1 ]= ch;
	adr_copy(0,x,1,x121);
	x121[0]=((x121[0] & 0x0f) + ((x121[0] & 0x0f0) >> 4));
}

adr_copy(src,istr,des,ostr)
int src;
unsigned char istr[16];
int des;
unsigned char ostr[16];

{
	int ipos, opos, ilow, olow, len, i, acc;

	if(src) /* Copy High order call (calling address) */ {
		ilow=istr[0]&1;
		ipos= ( (istr[0] & 0x0f) + ilow) >> 1;
		len=(istr[0] & 0x0f0)>>4;
		if (!ilow) ipos++;
	}
	else /* Copy low order call (called address) */ {
		ipos=1;
		ilow=0;
		len=istr[0] & 0x0f;
	}
	if (des) /* Copy to high order call (calling) */ {
		olow=ostr[0] & 1;
		opos=( (ostr[0] & 0x0f) + olow) >> 1;
		ostr[0]=ostr[0] & 0x0f | len<<4;
		if (!olow) opos++;
	}
	else /* Copy to low order call (called) */ {
		olow=0;
		opos=1;
		ostr[0]=ostr[0] & 0xf0 | len;
	}
	acc=0;
	if (olow)
		acc=ostr[opos] & 0xf0; /* Init acc */
	for (i=1;i<=len;i++) {
		if (ilow) acc += istr[ipos++] & 0x0f;
		else acc += (istr[ipos] & 0x0f0) >> 4;
		ilow = !ilow;
		if (olow) /* Output a byte */ {
			ostr[opos++]=acc;
			acc=0;
		}
		else acc=acc<<4;
		olow = !olow;
	}
	if (olow) ostr[opos]=acc;
}

next_alt(ptr,index)
unsigned char ptr[], index;
{	/* ptr points to a routing table entry and index  is 1..15 */
	int i, j;

	i=ptr[0];
	j=i & 0x0f;
	i=i>>4;
	if (index > i) return EOF;
	j += j&1;
	j= (j + --index) >> 1;
	i= ptr[++j];
	if (index&1) i=i & 0x0f;
	else i=(i & 0x0ff) >> 4;
	return i;
}

int new_rt()
{
	int j;

	if (numrt+1 == MaxRT) /* Table overflow */ {
		printf("MaxRT not big enough, is %d !! ***\n",MaxRT);
		exit(1);
	}
	rtable[numrt].cmd = 0;	/* Default to Invalid Terminator */
	rtable[numrt].dta = 0;
	for (j=0;j<10;j++) rtable[numrt].next[j]=0;
	return numrt++;
}

add_route(adr,alt)	/* Can have a specific station listed more than once */
unsigned char adr[];
unsigned char alt;
{
	unsigned char j, l, jl, kl;
	int ptr;
	int i, k;

	if (numrt == 0) numrt++;
	ptr=0;
	i=5;	/* Point PAST dnic */
	adr[0] = adr[0]<<4;

	do {
		k=next_alt(adr,i++);
		if (k == EOF) /* All done! */ {
			rtable[ptr].cmd = 1;	/* It's Valid! */
			rtable[ptr].dta = alt;
			if (usaalt[alt][0] & 0x0f) rtable[ptr].cmd = 2;
			return;
		}
		if (rtable[ptr].next[k] == 0) rtable[ptr].next[k] = new_rt();
		ptr = rtable[ptr].next[k];
	} while (ptr != 0);
}

new_ref(old,new)
int old,new;
{
	int i,j;

	for (i=0;i<numrt;i++)
		for (j=0;j<10;j++)
			if( rtable[i].next[j] == old ) rtable[i].next[j] = new;
}

unique_rt(src,low,high)
int src, low, high;
{
	int i, j, l;

	for (i=low;i<high;i++)
	   if ( (rtable[i].cmd != 99) && (rtable[src].dta == rtable[i].dta) ) {
		l = TRUE;
		for (j=0;j<10;j++)
			if (rtable[src].next[j] != rtable[i].next[j]) {
				l = FALSE;
				break;
			}
		if (l) return i;
	}
	return -1;
}

dump_rt()
{
	int i, j;

	printf("rec cmd dta   Next...\n");
	for (i=0;i<numrt;i++) {
		printf("%2.2d  %2.2d   %2.2d   ",i,rtable[i].cmd,rtable[i].dta);
		for (j=0;j<10;j++) printf("%d ",rtable[i].next[j]);
		printf("\n");
	}
}

compress_rt()	/* Only supports cmd=0 or cmd=1 */
{
	int i, j, k, l;

	dump_rt();
	j=numrt;
	for (i=0;i<j;i++) /* Scan the original list only */ {
		if (rtable[i].cmd != 0) /* May be a global terminator */ {
			l=0;
			for (k=0;k<10;k++) l = l || rtable[i].next[k];
			if (l == 0) /* Is a global terminator */ {
				if ( (l=unique_rt(i,j,numrt)) == -1) {
					l=new_rt();
					rtable[l].cmd=rtable[i].cmd;
					rtable[l].dta=rtable[i].dta;
					for (k=0;k<10;k++)
					   rtable[l].next[k]=rtable[i].next[k];
				} else {
					if (rtable[l].cmd == 0)
						rtable[l].cmd = rtable[i].cmd;
				}
				rtable[i].cmd = 99;	/* void this record */
				new_ref(i,l);	/* Change ref of I to L */
			}
		}
	}
	dump_rt();
	/* now remove the voided records */
	for (i=0;i<numrt;i++) {
		if (rtable[i].cmd == 99) /* Delete this record */ {
			for (j=i;j<numrt-1;j++) /* Copy Record up one */ {
				rtable[j].cmd=rtable[j+1].cmd;
				rtable[j].dta=rtable[j+1].dta;
				for (k=0;k<10;k++)
					rtable[j].next[k]=rtable[j+1].next[k];
				rtable[j+1].cmd = 99;
				new_ref(j+1,j);
			}
			numrt--;
			i--;
		}
	}
	dump_rt();
}

int
get_yes()
{
	register char ch;

	do {
		ch=getch();
		ch=toupper(ch);
	} while ( (ch!=0x59) && (ch!=0x4e));
	return (ch == 0x59); /* True is Yes */
}

putxch(val, f)
FILE f;
int val;
{
	char v;

	v= (val >> 4) &0x0f;
	if (v > 9) v += 7;
	fputc(v+0x30, f);
	v=(val & 0x0f);
	if (v > 9) v += 7;
	fputc(v+0x30, f);
}

int
wrap(col, f)
int col;
FILE f;
{
	if (col > 70) {
		fputc(0x0a, f);
		return 1;
	}
	return 0;
}
#define WRITE 01
#define READ 02
#include "config.h"

preamble(cmd, obj, len)
int cmd, obj, len;
{
	fputc(0x3a,fnam);
	putxch(cmd & 0xff, fnam);
	putxch(obj & 0xff, fnam);
	putxch(len & 0xff, fnam);
	putxch((len >> 8) & 0xff, fnam);
}

main()
{
	int i,j,h,m,l,pos,ch, checksum;
	char fname[20], adr[20], str[20], *chp;
	int digilen, callen;

	printf("ROSE Configuration Generator\n\n");
	do {
		printf("Configuration file to create? ");
		getline(fname, 80);
		fnam=fopen(fname,"wt");
	} while (fnam == NULL);

	printf("Each switch must have a Switch Address, in the ROSE we are\n");
	printf("using the Area Code and Exchange of the switch's QTH for this\n");
	printf("This address is used to route calls to the switch from others\n");
	printf("All switches in a given Area Code need to know about the others\n");
	printf("You will be asked to enter the routing entries to the other\n");
	printf("switches a little later.\n\n");
	do {
		printf("This switch's address (ex 201779)? ");
		getline(str,80);
		l3_adr(str,nodeadr);
	} while (nodeadr[0] == 0);

	printf("Define all links to the switches THIS switch knows about\n");
	printf("Each time through the loop you will be told the LINK NUMBER\n");
	printf("you will need to reference the links with this number later\n");
	printf("so please keep track of this for the moment.  [i will fix this\n");
	printf("later so you can just use the callsign, but i'm being lazy now!]\n");

	callen= -1;
	digilen=0;
	do {
		callen++;
		printf("\nHit RETURN without entering a call to exit\n");
		if (callen == 0) printf("Callsign of THIS switch? ");
		else printf("Callsign of switch for Link #%d? ",callen);
		if (get_call(calls[callen]) == -1) break;

		if (callen==0) printf("Digipeat Call of THIS switch? ");
		else {
	printf("The path between two switches can have up to ONE digipeater\n");
	printf("For the LINKS that don't need a digipeater just\n");
	printf("hit RETURN to skip it\n");
	printf("\nDigipeater needed to access the switch for LINK #%d? ",callen);
		}
		if (get_call(digi[callen]) != -1) digilen=callen+1;
		if (callen != 0) digi[callen][6] |= 0x80;

	if (callen > 0) {
	printf("For the port number must be 0 for TNC 2's\n");
	printf("The DR-200 can support 2 Ports (0 and 1)\n");
		printf("\nPort to use for LINK #%d? ",callen);
		scanf("%d", &i);    
		port[callen]=i;

	printf("\nOn an X.25 link (what ROSE uses) you can limit the number\n");
	printf("of Virtual Circuits (VC's) or Calls or Users that can use\n");
	printf("a given link at a time.\n");

		maxvc[callen]=getnum("\nMax VC's? ",5,254,20);

	printf("For each VC there is a window size (like MAXFRAME for AX.25)\n");

		l3wi[callen]=getnum("\nPacket Window Size? ",2,7,4);

	printf("When a link is not working this switch will mark it as\n");
	printf("bad for a given time period, this is done so we don't\n");
	printf("keep trying to route calls over a dead link one right after\n");
	printf("the other.  I think a good delay might be 15 mins (900 sec).\n");

		delay[callen]=getnum("\nBad link timer value (seconds)? ",
					0,28800,900);
	} else {
		port[0]=0;
		maxvc[0]=0;
		l3wi[0]=0;
		delay[0]=0;
	}

	printf("END OF LINK # %d\n", callen);

	} while (callen != 15);

	printf("[now for the fun part]\n\n");
	printf("The heart of any real network is the routing tables,\n");
	printf("which tell each switch where to forward calls to for other\n");
	printf("switches.\n");
	printf("For a given address there can be up to 15 alternative LINKS\n");
	printf("that the call to be routed to.  This is so that if a LINK to\n");
	printf("the next switch is down or can't accept anymore VC's we will\n");
	printf("be able to attempt to complete the call if another RF path exists.\n");
	printf("\nPlease consider the RF Paths between the switches and the various\n");
	printf("routing tables you could have.  Don't worry about causing a\n");
	printf("routing loop, as the switch will detect it and try the next\n");
	printf("alternative in the list, until it runs out.\n");

	printf("Since by and large the routes to get around in an area are\n");
	printf("very common you will now be asked for Routing sets.\n");
	printf("A routing set is a list of LINK numbers defining the best\n");
	printf("to worst path to a specific area.\n");

	for (i=1;i<16;i++) {
		printf("\nEnter Routing Set # %d\n",i);
		get_alt(usaalt[i]);
		if (usaalt[i][0] == 0) break;
	}
	maxalt = i;

	printf("Now enter a partial address, such as 201, or a node address\n");
	printf("followed by the routing set to use for that address\n");

	rtdn[0].adr[0]=4;	/* DNIC Len */
	rtdn[0].adr[1]=0x31;	/* 3100 */
	rtdn[0].adr[2]=0x00;
	rtsize=0;
	while (rtsize == 0) {
		printf("\nAddress? ");
		getline(str,80);
		if (strlen(str) == 1 && str[0]==0x30) break;
		l3_adr(str,adr);
		printf("Routing Set to use? ");
		getline(zarr, 80);
		sscanf(zarr," %d",&i);
		if (i<0 || i>15 || usaalt[i][0]==0) {
			printf("Invalid routing set\n");
			continue;
		}
		add_route(adr,i); /* Route "adr" to alt[i] */
	}

	printf("Now enter an address followed by the switch address\n");
	printf("of the node that covers that address.\n");
	printf("This is used for routing within a LAN.\n");

	while (rtsize == 0) {
		printf("\nAddress? ");
		getline(str,80);
		if (strlen(str) == 1 && str[0]==0x30) break;
		l3_adr(str,adr);
		printf("\nSwitch address to route calls to? ");
		getline(str,80);
		if (strlen(str) == 1 && str[0]==0x30) break;
		l3_adr(str,zarr);
		for (i=0;i<maxalt;i++) /* Find address in table */ {
			if (check_x121(usaalt[i],zarr) != -1) break;
		}
		if (i == maxalt) /* New entry */ {
			adr_copy(0,zarr,0,usaalt[i]);
			maxalt++;
		}
		add_route(adr,i);
	}

/*	rtdn[ndnics].rt=rtable; */
	ndnics++;
	compress_rt();

	printf("Now Please enter the text message that goes to any\n");
	printf("Level 2 user who connects to the switch Directly.\n");
	printf("It should give them some general info and directions\n");
	printf("of how to make a call via the switch.\n");
	printf("I usually have something like this:\n");
	printf("W2VY-3 ROSE Switch, address 201779\n\n");
	printf("While disconnected issue a command like:\n\n");
	printf("C dest VIA W2VY-3 201779\n");
	printf("\nFor more information download ROSEINFO.DOC from KD6TH-4\n");
	printf("or call me at (201) 779-W2VY (9289)\n\n");
	printf("Now please enter lines of text you want sent.\n");
	printf("End with Control-Z\n\n");
	pos=0;
	while (pos<2048) {
		getlin(zarr,160);
		if (feof(stdin)) break;
		chp=zarr;
		while (*chp) l2txt[pos++] = *(chp++);
		l2txt[pos++] = 0x0d;
	}
	l2txtl=pos;

	printf("Please Wait, saving data to disk...\n");

	/* Now write out the Callsigns */

	preamble(WRITE,CALL,callen*7);
	checksum=0;
	pos=9;
	for (i=0;i<callen;i++) {
		for (j=0;j<7;j++) {
			ch=calls[i][j];
			checksum += ch;
			if (wrap(pos, fnam)) pos=0;
			pos += 2;
			putxch(ch,fnam);
		}
	}
	putxch((-checksum) &0xff, fnam);
	fputc(0x0a, fnam);


	/* Now write out the digipeaters */

	preamble(WRITE,DIGI,digilen*7);
	checksum=0;
	pos=9;
	for (i=0;i<digilen;i++) {
		for (j=0;j<7;j++) {
			ch=digi[i][j];
			checksum += ch;
			if (wrap(pos, fnam)) pos=0;
			pos += 2;
			putxch(ch,fnam);
		}
	}
	putxch((-checksum) & 0xff, fnam);
	fputc(0x0a, fnam);


	/* Now write out the ports */

	preamble(WRITE,PORT,callen);
	checksum=0;
	pos=9;
	for (i=0;i<callen;i++) {
		ch=port[i];
		checksum += ch;
		if (wrap(pos, fnam)) pos=0;
		pos += 2;
		putxch(ch,fnam);
	}
	putxch((-checksum) & 0xff, fnam);
	fputc(0x0a, fnam);

	/* Now write out the Max Vc's */
	preamble(WRITE,MAXVC,callen);
	checksum=0;
	pos=9;
	for (i=0;i<callen;i++) {
		ch=maxvc[i];
		checksum += ch;
		if (wrap(pos, fnam)) pos=0;
		pos += 2;
		putxch(ch,fnam);
	}
	putxch((-checksum) & 0xff, fnam);
	fputc(0x0a, fnam);


	/* Next is the Level 3 Packet Window */

	preamble(WRITE,L3WI,callen);
	checksum=0;
	pos=9;
	for (i=0;i<callen;i++) {
		ch=l3wi[i];
		checksum += ch;
		if (wrap(pos, fnam)) pos=0;
		pos += 2;
		putxch(ch,fnam);
	}
	putxch((-checksum) & 0xff, fnam);
	fputc(0x0a, fnam);


	/* Now the link retry delay */

	preamble(WRITE,DELAY,callen*2);
	checksum=0;
	pos=9;
	for (i=0;i<callen;i++) {
		j=delay[i];
		ch = j&0xff;
		checksum += ch;
		if (wrap(pos, fnam)) pos=0;
		pos += 2;
		putxch(ch,fnam);
		ch= (j>>8) &0xff;
		checksum += ch;
		if (wrap(pos, fnam)) pos=0;
		pos += 2;
		putxch(ch,fnam);
	}
	putxch((-checksum) & 0xff, fnam);
	fputc(0x0a, fnam);


	/* DNICS and routing table for the dnic */

	preamble(WRITE,DNICS,3+(numrt*12));
	checksum=0;

	pos=9;
	for (i=0;i<3;i++) {
		ch=rtdn[0].adr[i];
		checksum += ch;
		if (wrap(pos, fnam)) pos=0;
		pos += 2;
		putxch(ch,fnam);
	}
	for (i=0;i<numrt;i++) {
		ch=rtable[i].cmd;
		checksum += ch;
		if (wrap(pos, fnam)) pos=0;
		pos += 2;
		putxch(ch,fnam);

		ch=rtable[i].dta;
		checksum += ch;
		if (wrap(pos, fnam)) pos=0;
		pos += 2;
		putxch(ch,fnam);

		for (j=0;j<10;j++) {
			ch=rtable[i].next[j];
			if ((ch == i) && (ch != 0))
				printf("*** Warning BR +0, ch=%d\n",ch);
			if (ch != 0) ch = (ch-i)*12;
			if (ch < 0) printf("*** Warning Disp = %d ***\n",ch);
			checksum += ch;
			if (wrap(pos, fnam)) pos=0;
			pos += 2;
			putxch(ch,fnam);
		}
	}
	putxch((-checksum) & 0xff, fnam);
	fputc(0x0a, fnam);


	/* Switch Address */

	preamble(WRITE,NODENUM,7);
	checksum=0;
	pos=9;
	for (i=0;i<7;i++) {
		ch=nodeadr[i];
		checksum += ch;
		if (wrap(pos, fnam)) pos=0;
		pos += 2;
		putxch(ch,fnam);
	}
	putxch((-checksum) & 0xff, fnam);
	fputc(0x0a, fnam);


	/* USA Routes (the routing alternatives) */

	preamble(WRITE,USAROUTE,120);
	checksum=0;
	pos=0;
	for (i=0;i<15;i++) {
		for (j=0;j<8;j++) {
			ch=usaalt[i][j];
			checksum += ch;
			if (wrap(pos, fnam)) pos=0;
			pos += 2;
			putxch(ch,fnam);
		}
	}
	putxch((-checksum) & 0xff, fnam);
	fputc(0x0a, fnam);


	/* L2USER Text Message */

	preamble(WRITE,L2USER,l2txtl);
	checksum=0;
	pos=9;
	for (i=0;i<l2txtl;i++) {
		ch=l2txt[i];
		checksum += ch;
		if (wrap(pos, fnam)) pos=0;
		pos += 2;
		putxch(ch,fnam);
	}
	putxch((-checksum) & 0xff, fnam);
	fputc(0x0a, fnam);

	/* All Done, close file */
	fclose(fnam);
	printf("All Done!\n\n");
}
