/* convers server - based on conversd written by DK5SG
 * ported to WNOS by DB3FL - 9109xx/9110xx
 * ported to NOS by PE1NMB - 920120
 * Mods by PA0GRI
 * Cleanup, and additional mods by WG7J
 */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#ifdef	UNIX
#include <sys/types.h>
#include <sys/stat.h>
#endif
#include <io.h>
#include "global.h"
#include "config.h"
#ifdef MAILBOX
#include "mailbox.h"
#include "netuser.h"
#endif
#include "timer.h"
#include "cmdparse.h"
#include "socket.h"
#include "session.h"
#include "files.h"

#ifdef CONVERS

#define	LINK	1
#define space


#ifdef LINK
static int nolinks;
static void connect_permlinks __ARGS((int a,void *b,void *c));
#endif
void conv_incom __ARGS((int s,void *t,void *p));

int Sconv = -1;
static char *myhostname = NULLCHAR;
#ifdef space
static char cnumber[] = "*** Channel numbers must be in the range 0..%d.\n";
#else
static char cnumber[] = "* range 0..%d.\n";
#endif

#define MAXCHANNEL 	32767
#define LINELEN 	256
#define MAX_WAITTIME	(60*60*3)
#define NAMELEN 10

struct convection {
	int  type;			/* Connection type */
#define CT_UNKNOWN      0
#define CT_USER         1
#define CT_HOST         2
#define CT_CLOSED       3
    char  name[NAMELEN+1];         /* Name of user or host */
    char  host[NAMELEN+1];         /* Name of host where user is logged on */
	struct convection *via;		/* Pointer to neighbor host */
    char *data;                 /* room for some personal data */
    int channel;            /* Channel number */
	int32 time;			/* Connect time */
	int locked;			/* Set if mesg already sent */
	int fd;				/* Socket descriptor */
	int fmask;			/* Socket mask */
	char ibuf[LINELEN];		/* Input buffer */
	int  received;			/* Number of bytes received */
	int  xmitted;			/* Number of bytes transmitted */
	struct convection *next;	/* Linked list pointer */
};

#define CM_UNKNOWN	(1 << CT_UNKNOWN)
#define CM_USER		(1 << CT_USER)
#define CM_HOST		(1 << CT_HOST)
#define CM_CLOSED	(1 << CT_CLOSED)

#define NULLCONNECTION	((struct convection *) 0)

static struct convection *convections;

struct permlink {
    char name[NAMELEN+1];   /* Name of link */
    int32 addr;             /* address to link to */
	struct convection *convection;	/* Pointer to associated connection */
    int32 statetime;        /* Time of last (dis)connect */
	int  tries;			/* Number of connect tries */
	int32 waittime;			/* Time between connect tries */
	int32 retrytime;		/* Time of next connect try */
	int fd;				/* socket descriptor */
	struct permlink *next;		/* Linked list pointer */
};
#define NULLPERMLINK ((struct permlink *) 0)

static struct permlink *permlinks;
static int fTerminating;
/* Stop convers server */
int
conv0(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	fTerminating = TRUE;
	close_s(Sconv);
	Sconv = -1;
	return 0;
}

/* Start up convers server */
int
conv1(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct sockaddr_in lsocket;
	int s;
#ifdef LINK
    struct proc *child;
#endif

	if(Sconv != -1){
		return 0;
	}
	psignal(Curproc,0); 		/* Don't keep the parser waiting */
	chname(Curproc,"Convers listener");
	fTerminating = FALSE;
	lsocket.sin_family = AF_INET;
	lsocket.sin_addr.s_addr = INADDR_ANY;
	lsocket.sin_port = (argc < 2) ? IPPORT_CONVERS : atoi(argv[1]);

	Sconv = socket(AF_INET,SOCK_STREAM,0);
	bind(Sconv,(char *)&lsocket,sizeof(lsocket));
	listen(Sconv,1);
    /* needed for stop/restart of server - WG7J */
    if(myhostname) {
        free(myhostname);
        myhostname = NULLCHAR;
    }
#ifdef LINK
    child = newproc("cperm",1024,connect_permlinks,0,0,NULL,0);
#else
	myhostname = strdup(Hostname);
#endif
	while(!fTerminating){
		if((s = accept(Sconv,NULLCHAR,(int *)NULL)) == -1)
			break;			/* Service is shutting down */
        /* Low mem check now done in tcpin.c - WG7J */
        /* Spawn a server */
        newproc("conversd",2048,conv_incom,s,0,NULL,0);
	}
#ifdef LINK
    if(!nolinks)
        killproc(child);
#endif
	return 0;
}

static void
free_connection(cp)
register struct convection *cp;
{
	register struct permlink *p;

	for(p = permlinks; p; p = p->next)
        if(p->convection == cp)
            p->convection = NULLCONNECTION;
    if(cp->fmask)
        close_s(cp->fd);
	free((char *) cp);
}

static void
free_closed_connections()
{
	register struct convection *cp,*p;
	time_t currtime;

	currtime = time(&currtime);

	for(p = NULLCONNECTION,cp = convections; cp; )
		if(cp->type == CT_CLOSED ||
			cp->type == CT_UNKNOWN && cp->time + 300 < currtime) {
		if(p) {
			p->next = cp->next;
			free_connection(cp);
			cp = p->next;
		} else {
			convections = cp->next;
			free_connection(cp);
			cp = convections;
		}
	} else {
		p = cp;
		cp = cp->next;
	}
}

static void
update_permlinks(name,cp)
char  *name;
struct convection *cp;
{
	register struct permlink *p;
	time_t currtime;

	for(p = permlinks; p; p = p->next)
		if(!strcmp(p->name,name)) {
			currtime = time(&currtime);
			p->convection = cp;
			p->statetime = currtime;
			p->tries = 0;
			p->waittime = 60;
			p->retrytime = currtime + p->waittime;
		}
}

static struct convection *
alloc_connection(fd)
int  fd;
{
	register struct convection *cp;
	time_t currtime;

    currtime = time(NULL);

	cp = (struct convection *)callocw(1,sizeof(struct convection ));
    cp->fd = fd;
    cp->fmask = 1;          /* close on exit, by default */
	cp->time = currtime;
	cp->next = convections;
	convections = cp;
	return cp;
}

#ifdef LINK
void
connect_permlinks(a,b,c)
int a;
void *b;
void *c;
{
	int s,len = SOCKSIZE;
	register struct permlink *p;
	struct sockaddr_in cport;
    int32 addr;
    FILE * fp;
    char host_name[NAMELEN+1],link_to[LINELEN];
    char line[LINELEN];
	time_t currtime;

	if((fp = fopen(Convers,"r")) == NULLFILE) 
		{
		if(myhostname == NULLCHAR)
			myhostname = strdup(Hostname);
		}
	else
		{
		while(fgets(line,LINELEN,fp))
			{
			if(*line == '#')
				continue;
			host_name[0] = '\0';
			link_to[0] = '\0';
			sscanf(line,"%10s %*s %s",host_name,link_to);
			if(host_name[0] == '\0')
				continue;
			if(myhostname == NULLCHAR)
				{
				myhostname = strdup(host_name);
				continue;
				}
			if(link_to[0] != '\0')
				{
				/* check to see if we already have a link to such animal,
				 * this happens when we stop and restart the server - WG7J
				*/
				if((addr = resolve(link_to)) == 0)
				/* next try the 'name' field */
				if((addr = resolve(host_name)) == 0)
					continue;   /* Don't know this one ! */
				for(p=permlinks;p;p=p->next)
					if(p->addr == addr)
						continue;       /* already have this one ! */

				/* Seems like a new link ! Go add it */
				p = (struct permlink *)callocw(1,sizeof(struct permlink ));
				strcpy(p->name,host_name);
				p->addr = addr;
				p->next = permlinks;
				permlinks = p;
				update_permlinks(host_name,NULLCONNECTION);
				}
			}
		fclose(fp);
		}

    /* If there are no links,
     * there is no reason the stay alive - WG7J
     */
	if(permlinks == NULLPERMLINK)
		{
		nolinks = 1;      /* tell our parent */
		return;
		}

	while(!fTerminating)
		{
		pause(1000L);
		for(p = permlinks; p; p = p->next)
			{
#ifdef NMB
			printf("Hostname      = %s\n",p->name);
			printf("Link addr     = %s\n",inet_ntoa(p->addr));
			printf("convection    = %ld\n",p->convection );
			printf("retrytime     = %ld\n",p->retrytime);
#endif
			currtime = time(&currtime);
			if(p->convection || p->retrytime > currtime)
				continue;
			p->tries++;
			p->waittime <<= 1;
			if(p->waittime > MAX_WAITTIME)
				p->waittime = MAX_WAITTIME;
			p->retrytime = p->waittime + currtime;
			cport.sin_family = AF_INET;
			cport.sin_port = IPPORT_CONVERS;
			cport.sin_addr.s_addr = p->addr; /* we've resolved this earlier */
			if((s = socket(AF_INET,SOCK_STREAM,0)) == -1)
				continue;
			if(connect(s,(char *)&cport,len) == -1)
				{
				shutdown(s,2);  /* to make sure it doesn't linger around */
				close_s(s);     /* WG7J - 9207228 */
				continue;
				}
			p->fd = s;
			newproc("permlink",2048,conv_incom,s,0,NULL,0);
			}
		}
}
#endif


static void
clear_locks()
{
	register struct convection *p;

    for(p = convections;p;p = p->next)
        p->locked = 0;
}

static void
send_user_change_msg(name,host,oldchannel,newchannel)
char  *name,*host;
int  oldchannel,newchannel;
{
	register struct convection *p;

	for(p = convections; p; p = p->next) {
		if(p->type == CT_USER && !p->via && !p->locked) {
			if(p->channel == oldchannel) {
				if(newchannel >= 0)
					p->xmitted += usprintf(p->fd,
					    "*** %s switched to channel %d.\n",
					     name,newchannel);
				else
					p->xmitted += usprintf(p->fd,
						"*** %s signed off.\n",name);
				p->locked = 1;
			}
			if(p->channel == newchannel) {
				p->xmitted += usprintf(p->fd,
                        "*** %s signed on.\n",name);
				p->locked = 1;
			}
		}
		if(p->type == CT_HOST && !p->locked) {
			p->xmitted += usprintf(p->fd,
				"/\377\200USER %s %s %d %d %d\n",
				name,host,0,oldchannel,newchannel);
			p->locked = 1;
		}
	}
	return;
}

static char *
formatline(prefix,text)
char  *prefix,*text;
{

#define PREFIXLEN 10
#define CONVLINELEN   79

	register char  *f,*t,*x;
	register int  l,lw;

    static char buf[2*LINELEN];

	for(f = prefix,t = buf; *f; *t++ = *f++) ;
	l = (int)(t - buf);
	f = text;

	while(!fTerminating) {
		while(isspace(uchar(*f)))
			f++;
		if(!*f) {
			*t++ = '\n';
			*t = '\0';
			return buf;
		}
		for(x = f; *x && !isspace(uchar(*x)); x++) ;
		lw = (int)(x - f);
		if(l > PREFIXLEN && l + 1 + lw > CONVLINELEN) {
			*t++ = '\n';
			l = 0;
		}
		do {
			*t++ = ' ';
			l++;
		} while(l < PREFIXLEN);
		while(lw--) {
			*t++ = *f++;
			l++;
		}
	}
}

static void
send_msg_to_user(fromname,toname,text)
char  *fromname,*toname,*text;
{
	register struct convection *p;
	char buffer[2*LINELEN];

	for(p = convections; p; p = p->next) {
		if(p->type == CT_USER && !strcmp(p->name,toname))
			if(p->via) {
				if(!p->via->locked) {
					p->via->xmitted += usprintf(p->via->fd,
                    "/\377\200UMSG %s %s %s\n",fromname,toname,text);
					p->via->locked = 1;
				}
			} else {
				if(!p->locked) {
					if(strcmp(fromname,"conversd")) {
						sprintf(buffer,"<*%s*>:",fromname);
						p->xmitted += usprintf(p->fd,"%s",formatline(buffer,text));
				} else {
					p->xmitted += usprintf(p->fd,"%s\n",text);
				}
				p->locked = 1;
			}
		}
	}
	return;
}

static void
send_msg_to_channel(fromname,channel,text)
char  *fromname;
int  channel;
char  *text;
{
	char  buffer[3*LINELEN];
	register struct convection *p;

	for(p = convections; p; p = p->next) {
		if(p->type == CT_USER && p->channel == channel)
			if(p->via) {
				if(!p->via->locked) {
					p->via->xmitted += usprintf(p->via->fd,
					"/\377\200CMSG %s %d %s\n",fromname,channel,text);
					p->via->locked = 1;
				}
			} else {
				if(!p->locked) {
					sprintf(buffer,"<%s>:",fromname);
					p->xmitted += usprintf(p->fd,"%s",formatline(&buffer[0],text));
					p->locked = 1;
				}
			}
	}
}

extern char *Months[];		/* in smtpserv.c */

static char  *
timestring(gmt)
long  gmt;
{
	static char  buffer[10];
	struct tm *tm;
	time_t currtime;

	currtime = time(&currtime);


	tm = localtime(&gmt);
	if(gmt + 24 * 60 * 60 > currtime)
		sprintf(buffer," %2d:%02d",tm->tm_hour,tm->tm_min);
	else
		sprintf(buffer,"%-3.3s %2d",Months[tm->tm_mon - 1],tm->tm_mday);
	return buffer;
}

#ifdef space
    char invitetext[] = "\n*** Message from %s at%s ...\nPlease join convers channel %d.\n\n";
    char mbinvitetext[] = "\n*** Message from %s at%s ...\nPlease type 'C' to join convers, then goto channel %d.\n\n";
    char responsetext[] = "*** Invitation sent to %s @ %s";
#else
    char invitetext[] = "\n*** Msg frm %s at%s ...\nPse join ch. %d.\n\n";
    char mbinvitetext[] = "\n*** Msg frm %s at%s ...\nPse hit 'C' for convers; goto ch. %d.\n\n";
    char responsetext[] = "*** sent to %s @ %s";
#endif
    char cnvd[] = "conversd";

static void
send_invite_msg(fromname,toname,channel)
char  *fromname,*toname;
int  channel;
{

	char buffer[LINELEN];
	struct convection *p;
#ifdef MAILBOX
	int i;
	struct mbx *m;
#endif
	time_t currtime;

	currtime = time(&currtime);

#ifdef MAILBOX
    /* Check users in the mailbox that aren't active */
    for(i = 0; i < NUMMBX; i++){
        if((m = Mbox[i]) != NULLMBX){
            if(m->state == MBX_CMD && !strcmp(m->name,toname)) {
//                p->xmitted += usprintf(m->user,mbinvitetext,fromname,timestring(currtime),channel);
                usflush(m->user);
                clear_locks();
                sprintf(buffer,responsetext,toname,"BBS@");
                strcat(buffer,Hostname);
                send_msg_to_user(cnvd,fromname,buffer);
                return;
            }
        }
    }
#endif

    /* check the current convers users */
    for(p = convections; p; p = p->next) {
		if(p->type == CT_USER && !strcmp(p->name,toname)) {
			if(p->channel == channel) {
                clear_locks();
                sprintf(buffer,"*** User %s is already on this channel.",toname);
                send_msg_to_user(cnvd,fromname,buffer);
                return;
            }
			if(!p->via && !p->locked) {
                p->xmitted += usprintf(p->fd,invitetext,fromname, \
                                        timestring(currtime),channel);
                clear_locks();
                sprintf(buffer,responsetext,toname,myhostname);
                send_msg_to_user(cnvd,fromname,buffer);
                return;
            }
			if(p->via && !p->via->locked) {
                p->via->xmitted += usprintf(p->via->fd,
                        "/\377\200INVI %s %s %d\n",fromname,toname,channel);
                return;
			}
        }
	}
    /* Nothing found locally, invite user on all links */
	for(p = convections; p; p = p->next) {
		if(p->type == CT_HOST && !p->locked) {
            p->xmitted += usprintf(p->fd,
                        "/\377\200INVI %s %s %d\n",fromname,toname,channel);
		}
	}
	return;
}

static void
bye_command(cp)
struct convection *cp;
{
	register struct convection *p;

	switch(cp->type) {
	case CT_UNKNOWN:
		cp->type = CT_CLOSED;
		break;
	case CT_USER:
		cp->type = CT_CLOSED;
		clear_locks();
		send_user_change_msg(cp->name,cp->host,cp->channel,-1);
		break;
	case CT_HOST:
		cp->type = CT_CLOSED;
		update_permlinks(cp->name,NULLCONNECTION);
        for(p = convections; p; p = p->next)
			if(p->via == cp) {
                p->type = CT_CLOSED;
                clear_locks();
                send_user_change_msg(p->name,p->host,p->channel,-1);
			}
		break;
    case CT_CLOSED:
		break;
    }
}

static void
channel_command(cp)
struct convection *cp;
{
	char  s[7];
	int  newchannel;

	s[0] = '\0';
	sscanf(cp->ibuf,"%*s %6s",s);
	if(s[0] == '\0') {
#ifdef space
	cp->xmitted += usprintf(cp->fd,"*** You are on channel %d.\n",cp->channel);
#else
	cp->xmitted += usprintf(cp->fd,"* On channel %d.\n",cp->channel);
#endif
		return;
	}
	newchannel = atoi(s);
	if(newchannel < 0 || newchannel > MAXCHANNEL) {
        cp->xmitted += usprintf(cp->fd,cnumber,MAXCHANNEL);
        return;
	}
	if(newchannel == cp->channel) {
        cp->xmitted += usprintf(cp->fd,
                            "*** Already on channel %d.\n",cp->channel);
		return;
	}
	send_user_change_msg(cp->name,cp->host,cp->channel,newchannel);
	cp->channel = newchannel;
	cp->xmitted += usprintf(cp->fd,"*** Now on channel %d.\n",cp->channel);
	return;
}

static void
help_command(cp)
struct convection *cp;
{
#ifdef space
	cp->xmitted += usprintf(cp->fd,"Commands may be abbreviated. Commands are:\n");
	cp->xmitted += usprintf(cp->fd,"/? or /HELP\t\tPrint help information\n");
	cp->xmitted += usprintf(cp->fd,"/BYE\t\t\tTerminate the convers session\n");
	cp->xmitted += usprintf(cp->fd,"/CHANNEL n\t\tSwitch to channel n\n");
	cp->xmitted += usprintf(cp->fd,"/EXIT\t\t\tTerminate the convers session\n");
	cp->xmitted += usprintf(cp->fd,"/INVITE user\t\tInvite user to join your channel\n");
	cp->xmitted += usprintf(cp->fd,"/LINKS [LONG]\t\tList all connections to other hosts\n");
	cp->xmitted += usprintf(cp->fd,"/MSG user text...\tSend a private message to user\n");
	cp->xmitted += usprintf(cp->fd,"/PERSONAL text\t\tSet or show some personal text\n");
	cp->xmitted += usprintf(cp->fd,"/QUIT\t\t\tTerminate the convers session\n");
	cp->xmitted += usprintf(cp->fd,"/WHO [QUICK]\t\tList all users and their channel numbers\n");
	cp->xmitted += usprintf(cp->fd,"/WRITE user text...\tSend a private message to user\n***\n");
#else
	cp->xmitted += usprintf(cp->fd,"no help\n");
#endif
	return;
}

static void
invite_command(cp)
struct convection *cp;
{
	char toname[10];

	toname[0] = '\0';
    sscanf(cp->ibuf,"%*s %10s",toname);
	if(toname[0] != '\0')
	send_invite_msg(cp->name,toname,cp->channel);
	return;
}

static void
links_command(cp)
struct convection *cp;
{
	char  tmp[20];
	char full[3];
	struct convection *pc;
	struct permlink *pp;

	full[0] = '\0';
	sscanf(cp->ibuf,"%*s %2s",full);
    cp->xmitted += usprintf(cp->fd,"Host       State         Since%s\n",
	(full[0] != 'l') ? "" : " NextTry Tries Queue Receivd Xmitted");
	for(pc = convections; pc; pc = pc->next)
		if(pc->type == CT_HOST) {
			cp->xmitted += usprintf(cp->fd,
				(*full == 'l') ?
                "%-10s %-12s %s     %5d %7d %7d\n" :
                "%-10s %-12s %s\n",
				pc->name,
				"Connected",
				timestring(pc->time),
				0,	/* always 0, cuz NOS handles the tx-queue */
				pc->received,
				pc->xmitted);
		}
	for(pp = permlinks; pp; pp = pp->next)
		if(!pp->convection || pp->convection->type != CT_HOST) {
			strcpy(tmp,timestring(pp->retrytime)),
				cp->xmitted += usprintf(cp->fd,
				(*full == 'l') ?
                "%-10s %-12s %s  %s %5d\n" :
                "%-10s %-12s %s\n",
				pp->name,
				pp->convection ? "Connecting" : "Disconnected",
				timestring(pp->statetime),
				tmp,
				pp->tries);
	}
	cp->xmitted += usprintf(cp->fd,"***\n");
	return;
}

static void
msg_command(cp)
struct convection *cp;
{

    char dummy[LINELEN],toname[NAMELEN+1],*text;
	register struct convection *p;

	toname[0] = '\0';
    sscanf(cp->ibuf,"%s %10s",dummy,toname);
	text = &cp->ibuf[0];
	text += strlen(dummy) + strlen(toname) + 2;

	if(!*text)
		return;
	for(p = convections; p; p = p->next)
		if(p->type == CT_USER && !strcmp(p->name,toname))
			break;
	if(!p)
		cp->xmitted += usprintf(cp->fd,"*** No such user: %s.\n",toname);
	else
		send_msg_to_user(cp->name,toname,text);
	return;
}

static void
name_command(cp)
struct convection *cp;
{
	int  newchannel;
	char dummy[7];

	cp->name[0] = '\0';
	dummy[0] = '\0';
    sscanf(cp->ibuf,"%*s %10s %6s",cp->name,dummy);
	newchannel = atoi(dummy);
	if(cp->name[0] == '\0')
		return;
	strlwr(cp->name);
    strcpy(cp->host,myhostname);
	cp->type = CT_USER;
	cp->xmitted += usprintf(cp->fd,
        "Conference @ %s  Type /HELP for help.\n",myhostname);
	if(newchannel < 0 || newchannel > MAXCHANNEL) {
		cp->xmitted += usprintf(cp->fd,cnumber,MAXCHANNEL);
	} else
		cp->channel = newchannel;
	send_user_change_msg(cp->name,cp->host,-1,cp->channel);
	return;
}

/* Set some personal data, like name and qth - WG7J */
static void
personal_command(cp)
struct convection *cp;
{
    struct convection *p;

    if(strlen(cp->ibuf) > 3) {
        if(cp->data)
            free(cp->data);
        rip(cp->ibuf);      /* get rid of ending '\n' */
        cp->data = strdup(&cp->ibuf[3]);
        /* update all links too ! - WG7J */
        for(p=convections;p;p=p->next)
            if(p->type == CT_HOST)
                p->xmitted += usprintf(p->fd,"/\377\200UDAT %s %s %s\n",
                                    cp->name,cp->host,cp->data);
    } else {
        cp->xmitted += usprintf(cp->fd,"*** data set to: %s\n", \
            cp->data ? cp->data : "" );
    }
    return;
}

static void
who_command(cp)
struct convection *cp;
{

	char buffer[LINELEN];
	int channel,full = 0,quick = 0;
	struct convection *p;
#ifdef MAILBOX
	int i;
	struct mbx *m;
#endif

	buffer[0] = '\0';
	sscanf(cp->ibuf,"%*s %2s",buffer);
	switch(tolower(buffer[0])) {
	case 'q':
		quick = 1;
		break;
	}

	if(quick) {
		cp->xmitted += usprintf(cp->fd,"Channel Users\n");
		clear_locks();
		do {
			channel = -1;
			for(p = convections; p; p = p->next) {
				if(p->type == CT_USER &&
				  !p->locked &&
				  (channel < 0 || channel == p->channel)) {
					if(channel < 0) {
						channel = p->channel;
						sprintf(buffer,"%7d",channel);
					}
					strcat(buffer," ");
					strcat(buffer,p->name);
					p->locked = 1;
				}
			}
			if(channel >= 0) {
				cp->xmitted += usprintf(cp->fd,"%s\n",buffer);
			}
		} while(channel >= 0);
	} else {
		cp->xmitted += usprintf(cp->fd,
            "User       Host       Via        Channel   Time Personal\n");
		for(p = convections; p; p = p->next) {
			if(p->type == CT_USER) {
				cp->xmitted += usprintf(cp->fd,
                "%-10s %-10s %-10s %7d %s %s\n",
					p->name,
					p->host,
					p->via ? p->via->name : "",
					p->channel,
					timestring(p->time),
                    p->data ? p->data : "" );
			}
		}
	}
#ifdef MAILBOX
	for(i = 0; i < NUMMBX; i++) {
	if((m = Mbox[i]) != NULLMBX) {
		if(m->state == MBX_CMD) {
		if(quick)
            cp->xmitted += usprintf(cp->fd," BBS %s\n",m->name);
		else
            cp->xmitted += usprintf(cp->fd,"%-10s BBS@%s\n",m->name,Hostname);
		}
	}
	}
#endif
	cp->xmitted += usprintf(cp->fd,"***\n");
	return;
}

/* this is called from the finger-daemon when 'conf' is fingered - WG7J */
void ShowConfUsers(s)
int s;
{
    struct convection *p;

    usputs(s,"User       Host       Via        Channel   Time Personal\n");
    for(p=convections;p;p=p->next)
        if(p->type == CT_USER)
            usprintf(s,"%-10s %-10s %-10s %7d %s %s\n",
                    p->name,p->host,p->via ? p->via->name : "",
                    p->channel,timestring(p->time),p->data);

}

static void
h_cmsg_command(cp)
struct convection *cp;
{
    char *text;
    int  channel;
    char name[LINELEN],dummy[40];

    sscanf(cp->ibuf,"%s %10s %d",dummy,name,&channel);
	text = &cp->ibuf[0];
	text += strlen(dummy) + strlen(name) + 2;
	while(isspace(*text) == 0)
		text++;
	text++;
	if(isprint(*text) != 0)
		send_msg_to_channel(name,channel,text);
	return;
}

static void
h_host_command(cp)
struct convection *cp;
{

    char name[NAMELEN+1];
	register struct convection *p;
	register struct permlink *pp;

	name[0] = '\0';
    sscanf(cp->ibuf,"%*s %10s",name);
	if(name[0] == '\0')
		return;
	for(p = convections; p; p = p->next)
		if(!strcmp(p->name,name))
			bye_command(p);
	for(pp = permlinks; pp; pp = pp->next)
		if(!strcmp(pp->name,name) && pp->convection && pp->convection != cp)
			bye_command((strcmp(myhostname,name) < 0) ? pp->convection : cp);
	if(cp->type != CT_UNKNOWN)
		return;
	cp->type = CT_HOST;
	strcpy(cp->name,name);		/* already allocated */
	update_permlinks(name,cp);
	cp->xmitted += usprintf(cp->fd,"/\377\200HOST %s\n",myhostname);
	for(p = convections; p; p = p->next)
		if(p->type == CT_USER) {
			cp->xmitted += usprintf(cp->fd,
				"/\377\200USER %s %s %d %d %d\n",
				p->name,p->host,0,-1,p->channel);
            if(p->data)
                cp->xmitted += usprintf(cp->fd,
                        "/\377\200UDAT %s %s %s\n",
                        p->name,p->host,p->data);
        }
	return;
}

static void
h_invi_command(cp)
struct convection *cp;
{
    char dummy[10];
	char fromname[10],toname[10];
	int  channel;

    sscanf(cp->ibuf,"%*s %10s %10s %d",fromname,toname,&channel);
	send_invite_msg(fromname,toname,channel);
	return;
}

static void
h_loop_command(cp)
struct convection *cp;
{
  char host[LINELEN];

  sscanf(cp->ibuf,"%*s %s",host);
  log(cp->fd, "conversd rx: LOOP %s",host);
  bye_command(cp);
}

/* Command to take user's personal data across a link - WG7J */
void
h_udat_command(cp)
struct convection *cp;
{
    char *name,*host,*data;
    struct convection *p;

    /* do a validity check first */
    if((name = strchr(cp->ibuf,' ')) == NULLCHAR)
        return;
    name++;
    if((host = strchr(name,' ')) == NULLCHAR)
        return;
    *host++ = '\0';
    if((data = strchr(host,' ')) == NULLCHAR)
        return;
    *data++ = '\0';
    if(strlen(data) == 0)
        return;
    /* everything seems fine, now find user ! */
    for(p=convections;p;p=p->next)
        if(!strcmp(p->name,name) && !strcmp(p->host,host)) {
            if(p->data)
                free(p->data);
            p->data = strdup(data);
        }
}

static void
h_umsg_command(cp)
struct convection *cp;
{
    char dummy[LINELEN],fromname[NAMELEN+1],toname[NAMELEN+1],*text;

    sscanf(cp->ibuf,"%s %10s %10s",dummy,fromname,toname);
	text = &cp->ibuf[0];
	text += strlen(dummy) + strlen(fromname) + strlen(toname) + 3;
	if(*text)
	send_msg_to_user(fromname,toname,text);
	return;
}

static void
h_user_command(cp)
struct convection *cp;
{
    char host[NAMELEN+1],name[NAMELEN+1];
	int  newchannel,oldchannel;
	register struct convection *p;
	time_t currtime;

	currtime = time(&currtime);

    sscanf(cp->ibuf,"%*s %10s %10s %*s %d %d",name,host,&oldchannel,&newchannel);

	for(p = convections; p; p = p->next)
        if(p->type == CT_USER) {
            /* new 920705 dl9sau */
            /* If Neighbour2 registers a user on HostX, while someone has already
             * been registered for HostX via Neighbour1, then we definitely have
             * a loop !  We send a loop detect message and then close the link:
             * /..LOOP <myhostname> <myneighbour> <host>
             *
             * The LOOP PREVENTION CODE detects ONLY a loop if it starts at this
             * host. That's, why I suggest this code to be implemented in every
             * conversd implementation.
             */
            if (oldchannel < 0 && p->via != cp && !strcmp(p->host, host)) {
                usprintf(cp->fd,"/\377\200LOOP %s %s %s\n", \
                    myhostname, host,p->via ? p->via->name : myhostname);
                log(cp->fd, "conversd sent: LOOP %s",host);
                bye_command(cp);
                return;
            }
            if(p->channel == oldchannel && p->via == cp && 
                !strcmp(p->name,name) && !strcmp(p->host,host))
			break;
        }
	if(!p) {
		p = (struct convection *)callocw(1,sizeof(struct convection ));
		p->type = CT_USER;
        strcpy(p->name,name);
        strcpy(p->host,host);
		p->via = cp;
		p->channel = oldchannel;
		p->time = currtime;
		p->next = convections;
		convections = p;
	}
	if((p->channel = newchannel) < 0)
		p->type = CT_CLOSED;
	send_user_change_msg(name,host,oldchannel,newchannel);
	return;
}

static void
process_commands(cp,mbxuser)
struct convection *cp;
int mbxuser;
{
    char arg[LINELEN];
    int arglen,size;
    char *ccp,*cccp;

    struct cmdtable {
        char  *name;
        void (*fnc)();
        int  states;
    } cmdtable[] = {
        "?",        help_command,       CM_USER,
        "bye",      bye_command,        CM_USER,
        "channel",  channel_command,    CM_USER,
        "exit",     bye_command,        CM_USER,
        "help",     help_command,       CM_USER,
        "invite",   invite_command,     CM_USER,
        "links",    links_command,      CM_USER,
        "msg",      msg_command,        CM_USER,
        "name",     name_command,       CM_UNKNOWN,
        "personal", personal_command,   CM_USER,
        "quit",     bye_command,        CM_USER,
        "who",      who_command,        CM_USER,
        "write",    msg_command,        CM_USER,

        "\377\200cmsg", h_cmsg_command,     CM_HOST,
        "\377\200host", h_host_command,     CM_UNKNOWN,
        "\377\200invi", h_invi_command,     CM_HOST,
        "\377\200loop", h_loop_command,     CM_HOST,
        "\377\200udat", h_udat_command,     CM_HOST,
        "\377\200umsg", h_umsg_command,     CM_HOST,
        "\377\200user", h_user_command,     CM_HOST,
        "\377\200CMSG", h_cmsg_command,     CM_HOST,
        "\377\200HOST", h_host_command,     CM_UNKNOWN,
        "\377\200INVI", h_invi_command,     CM_HOST,
        "\377\200LOOP", h_loop_command,     CM_HOST,
        "\377\200UDAT", h_udat_command,     CM_HOST,
        "\377\200UMSG", h_umsg_command,     CM_HOST,
        "\377\200USER", h_user_command,     CM_HOST,
        0,      0,          0,
    };
    struct cmdtable *cmdp;

    for(;;) {
loop:
        if(mbxuser)
            usflush(cp->fd);    /* might not be neccesary */
        if(cp->type == CT_CLOSED)
            break;
        memset(cp->ibuf,0,sizeof(cp->ibuf));
        if((size = recvline(cp->fd,cp->ibuf,sizeof(cp->ibuf))) <= 0)
            break;
        cp->received += size;
        clear_locks();
        cp->locked = 1;
        if(*cp->ibuf == '/') {
            ccp = &cp->ibuf[1];
            arg[0] = '\0';
            sscanf(ccp,"%s",arg);
            arglen = strlen(arg);
            for(cmdp = cmdtable; cmdp->name; cmdp++) {
#if !defined(OS2)
                if(!strncmpi(cmdp->name,arg,arglen)) {
#else
                if(!strncmp(cmdp->name,arg,arglen)) {
#endif
                    if(cmdp->states & (1 << cp->type))
                        (*cmdp->fnc)(cp);
                    goto loop;
                }
            }
            if(cp->type == CT_USER)
                cp->xmitted += usprintf(cp->fd,
                    "*** Unknown command '/%s'. Type /HELP for help.\n",arg);
            goto loop;
        }
        if((ccp = strpbrk(cp->ibuf,"\r\n")) != NULLCHAR)
            *ccp = '\0';
        if(isprint(cp->ibuf[0]) != 0 && cp->type == CT_USER)
            send_msg_to_channel(cp->name,cp->channel,cp->ibuf);
    }
    sockblock(cp->fd,SOCK_BLOCK);
    if(mbxuser)
        setflush(cp->fd,-1);
    bye_command(cp);
    free_closed_connections();
}

/* Incoming convers session */
void
conv_incom(s,t,p)
int s;
void *t;
void *p;
{
	struct convection *cp;
	struct permlink *pl;

	sockowner(s,Curproc);	/* We own it now */
	sockmode(s,SOCK_BINARY);
        sockblock(s,SOCK_NOTXBLOCK);   /* prevent backlogs ! */
	cp = alloc_connection(s);

	for(pl = permlinks; pl; pl = pl->next)
		if(pl->fd == s) {
			pl->convection = cp;
			cp->xmitted += usprintf(s,"/\377\200HOST %s\n",myhostname);
		}

	if(pl == NULLPERMLINK) {
			usprintf(cp->fd,"\npse login with '/n <call>'\n\n");
	}
    process_commands(cp,0);
}

#ifdef MAILBOX
/* this is for Mailbox users */
void
mbox_converse(m)
struct mbx *m;
{
	struct convection *cp;

    sockblock(m->user,SOCK_NOTXBLOCK);   /* prevent backlogs ! */
    cp = alloc_connection(m->user);
    strcpy(cp->name,m->name);
    if (!myhostname)
	myhostname = strdup(Hostname);
    strcpy(cp->host,myhostname);
    cp->type = CT_USER;
    cp->fmask = 0;          /* do not close socket on exit */
    setflush(m->user,'\n');       /* automatic flushing each line */
    cp->xmitted += usprintf(m->user,
            "Conference @ %s  Type /HELP for help.\n",myhostname);
    clear_locks();
    cp->locked = 1; /* send to everyone but ourself */
    send_user_change_msg(cp->name,cp->host,-1,0);

    process_commands(cp,1);
}
#endif /* MAILBOX */

#endif /* CONVERS */
