/*
   dist.c : Copyright Paul Healy, EI9GL, 1992.

   920802 : Created.
   920830 : Added Resent headers
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "config.h"
#include "buf.h"
#include "mailfile.h"
#include "bmh.h"
#include "rc.h"
#include "misc.h"
#include "help.h"
#include "header.h"
#include "mailer.h"
#include "extract.h"
#include "send.h"

#ifdef BMH
#define main dist_main
#endif

/*
 * forward a message in its orginal form
 */
static int
bouncemsg(FILE *fp, Header_t *h)
{
   struct addr *list;
   char *argv[MAXARGS], *to_copy = NULL;
   int argc, ret = 0;
   char *to = HEADERP(h, TO);

   if (to == NULL) {
      fprintf(stderr, "dist: no recipients\n");
      ret = -1;
      }
   else if ((to_copy = strdup(to)) == NULL) {
      fprintf(stderr, "dist: no room\n");
      ret = -1;
      }
   else if ( ((argc = parse(to, argv, MAXARGS))< 1) ||
             ((list = make_tolist(argc, argv)) == NULL)
           ) {
      fprintf(stderr, "dist: no recpients, send aborted\n");
      ret = -1;
      }
   else {
      char filename[MAXNAMLEN];
      FILE *tmpf = tempfile("dist", filename, "w+");

      if (tmpf == NULL) {
         fprintf(stderr, "dist: can't open temporary file\n");
         ret = -1;
         }
      else {
         fprintf(tmpf, "Resent-Date: %s", ptime(time(NULL)));

         fprintf(tmpf, "Resent-From: %s@%s", getrc(username), getrc(hostname));
         if ( getrc(fullname) != NULL )
            fprintf(tmpf, " (%s)", getrc(fullname));
         fprintf(tmpf, "\n");

         if (getrc(replyto) != NULL)
            fprintf(tmpf, "Resent-Reply-To: %s\n", getrc(replyto));
         if (to_copy != NULL)
            fprintf(tmpf, "Resent-To: %s\n", to_copy);

         if (dupfile(fp, tmpf) == -1) {
            fprintf(stderr, "dist: error writing to %s\n", filename);
            ret = -1;
            }
         else 
            queuejob(tmpf, list, 0L);
         }
      del_addrlist(list);
      fclose(tmpf);
      unlink(filename);
      }

   if (to_copy != NULL)
      free(to_copy);

   return ret;
}

int
dist(Mailfile *m, int argc, char *argv[])
{
   char filename[MAXNAMLEN];
   FILE *fp;

   if (mail_getfile(m, argc, argv, "dist") == -1)
      return -1;

   if ( (fp = tempfile("dist", filename, "w")) == NULL)
      return -1;
   else
      fclose(fp);

   if ( (addstr("dist", filename, "To: \n--------\n") == -1) ||
        (append("dist", m, filename, mail_thismsg(m, argc, argv)) == -1) ||
        (call_ed("dist", filename) == -1)
      ) {
      unlink(filename);
      return -1;
      }

   return dowhatnow("dist", filename, bouncemsg);
}

int
main(int argc, char *argv[])
{
   Mailfile m;

   if (dohelp(argc, argv, "dist [msg] (distribute/bounce a message)"))
      return 0;

   mail_zero(&m);

   dist(&m, argc, argv);

   return mail_savestate(&m);
}
