/* Interface driver for the DRSI PCPA board for the IBM PC connected
 * to a WA4DSY 56kbps modem
*/
#include <stdio.h>
#include "global.h"
#include "mbuf.h"
#include "iface.h"
#include "dr.h"
#include "8530.h"
#include "ax25.h"
#include "trace.h"

struct dr dr[NDR];
int dr0vec();
int (*drhandle[])() = { dr0vec };
struct hdlc hdlc[2*NDR];
int16 ndr;

/* Branch table for interrupt handler */
int htxint(), hexint(), hrxint(), hspint();
static int (*svec[])() = {
	htxint, hexint, hrxint, hspint
};

/* Master interrupt handler for the PC-100 card. All interrupts come
 * here first, then are switched out to the appropriate routine.
 */
drint(dev)
int16 dev;
{
	register char iv;
	register int16 drbase;
	struct hdlc *hp;
	
	dr[dev].ints++;
	drbase = dr[dev].addr;

	/* Read interrupt vector, including status, from channel B */
	iv = read_sio(drbase+CHANB,R2);

	hp = &hdlc[2 * dev + ((iv & 0x80)? 0 : 1)];

	/* Now switch to appropriate routine */
	(*svec[(iv>>1) & 0x3])(hp);

	/* Reset interrupt pending state (register A only) */
	write_sio(drbase+CHANA,R0,RET_INT);

	/* Wang the 8530 hardware interrupt acknowledge line - Bdale */
	inportb(drbase+INTACK);
}
/* HDLC Special Receive Condition interrupt
 * The most common event that triggers this interrupt is the
 * end of a frame; it can also be caused by a receiver overflow.
 */
static
hspint(hp)
register struct hdlc *hp;
{
	register char c;

	hp->spints++;
	c = read_sio(hp->base,R1);	/* Fetch latched bits */

	if((c & (END_FR|CRC_ERR)) == END_FR && hp->rcvbuf != NULLBUF
		&& hp->rcvbuf->cnt > 1){
		/* End of valid frame */
		hp->rcvbuf->cnt--;	/* Toss 1st crc byte */
		enqueue(&hp->rcvq,hp->rcvbuf);
		hp->rcvbuf = NULLBUF;
		hp->rcvcnt++;
	} else {
		/* An overflow or CRC error occurred; restart receiver */
		hp->crcerr++;
		if(hp->rcvbuf != NULLBUF){
			hp->rcp = hp->rcvbuf->data;
			hp->rcvbuf->cnt = 0;
		}
	}
	write_sio(hp->base,R0,ERR_RES);
}
/* HDLC SIO External/Status interrupts
 * The only one of direct interest is a receiver abort; the other
 * usual cause is a change in the modem control leads, so kick the
 * transmit interrupt routine.
 */
static
hexint(hp)
register struct hdlc *hp;
{
	hp->exints++;
	hp->status = read_sio(hp->base,R0);	/* Fetch status */
	if((hp->status & BRK_ABRT) && hp->rcvbuf != NULLBUF){
		hp->aborts++;
		/* Restart receiver */
		hp->rcp = hp->rcvbuf->data;
		hp->rcvbuf->cnt = 0;
	}
	write_sio(hp->base,R0,RES_EXT_INT);
	write_sio(hp->base,R0,RET_INT);
	/* Kick the transmit interrupt routine for a possible modem change */
	htxint(hp);
}
/* HDLC receiver interrupt handler. Allocates buffers off the freelist,
 * fills them with receive data, and puts them on the receive queue.
 */
static
hrxint(hp)
register struct hdlc *hp;
{
	register struct mbuf *bp;
	register int16 base;

	hp->rxints++;
	base = hp->base;
	/* Allocate a receive buffer if not already present */
	if((bp = hp->rcvbuf) == NULLBUF){
		bp = hp->rcvbuf = alloc_mbuf(hp->bufsiz);
		if(bp == NULLBUF){
			/* No memory, abort receiver */
			hp->nomem++;
			write_sio(base,R3,ENT_HM|RxENABLE|RxCRC_ENAB|Rx8);
			(void) inportb(base+DATA);
			return;
		}
		hp->rcp = hp->rcvbuf->data;
	}
	while(read_sio(base,R0) & Rx_CH_AV){
		if(bp->cnt++ >= hp->bufsiz){
			/* Too large; abort the receiver, toss buffer */
			hp->toobig++;
			write_sio(base,R3,ENT_HM|RxENABLE|RxCRC_ENAB|Rx8);
			(void) inportb(base+DATA);
			free_p(bp);
			hp->rcvbuf = NULLBUF;
			break;
		}
		/* Normal save */
		*hp->rcp++ = inportb(base+DATA);
	}
}
int ctswait;
/* HDLC transmit interrupt service routine
 *
 * The state variable tstate, along with some static pointers,
 * represents the state of the transmit "process".
 */
static
htxint(hp)
register struct hdlc *hp;
{
	register int16 base;
	char i_state,c;

	i_state = disable();
	hp->txints++;
	base = hp->base;
	while(read_sio(base,R0) & Tx_BUF_EMP){
		switch(hp->tstate){
		/* First here for efficiency */
		case ACTIVE:		/* Sending frame */
			if(pullup(&hp->sndbuf,&c,1) == 1){
				outportb(base+DATA,c);
			} else {
				/* Do this after sending the last byte */
				write_sio(base,R0,RES_Tx_P);
				if((hp->sndbuf = dequeue(&hp->sndq)) == NULLBUF){
					switch(hp->mode){
					case CSMA:
						/* Begin transmitter shutdown */
						hp->tstate = FLUSH;
						break;
					case FULLDUP:
						hp->tstate = IDLE;
						break;
					}
				}
			}
			continue;
		case IDLE:
			/* Transmitter idle. Find a frame for transmission */
			if((hp->sndbuf = dequeue(&hp->sndq)) == NULLBUF)
				goto ret;

		case DEFER:	/* note fall-thru */
			if(hp->mode == CSMA && (hp->status & DCD)){
				hp->tstate = DEFER;
				goto ret;
			}
			rts(base,ON);	/* Transmitter on */
		case KEYUP:	/* note fall-thru */
			if((hp->status & CTS) == 0){
				ctswait++;
				hp->tstate = KEYUP;
				goto ret;
			}
			write_sio(base,R0,RES_Tx_CRC);
			pullup(&hp->sndbuf,&c,1);
			outportb(hp->base+DATA,c);
			hp->tstate = ACTIVE;
			write_sio(base,R0,RES_EOM_L);
			continue;
		case FLUSH:	/* Sending flush character */
			outportb(hp->base+DATA,(char)0);
			hp->tstate = FIN2;
			continue;
		case FIN2:
			write_sio(base,R0,SEND_ABORT);
			hp->tstate = IDLE;
			rts(base,OFF);
			write_sio(base,R0,RES_Tx_P);
			continue;
		}
	}
ret:	restore(i_state);
}

/* Set request-to-send on modem */
static
rts(base,x)
int16 base;
int x;
{
	int16 cmd;

	if(x)
		cmd = TxCRC_ENAB | RTS | TxENAB | Tx8 | DTR;
	else
		cmd = TxCRC_ENAB | TxENAB | Tx8 | DTR;
	write_sio(base,R5,cmd);
}
/* (re)Initialize HDLC controller parameters */
static
hdlcparam(hp)
register struct hdlc *hp;
{
	int16 tc;
	char i_state;
	register int16 base;

	/* Initialize 8530 channel for SDLC operation */
	base = hp->base;
	i_state = disable();

	switch(base & 2){
	case 0:
		write_sio(base,R9,CHRA);	/* Reset channel A */
		break;
	case 2:
		write_sio(base,R9,CHRB);	/* Reset channel B */
		break;
	}
	/* Wait/DMA disable, Int on all Rx chars + spec condition,
	 * parity NOT spec condition, TxINT enable, Ext Int enable
	 */
	write_sio(base,R1,INT_ALL_Rx | TxINT_ENAB | EXT_INT_ENAB);

	/* Dummy interrupt vector, will be modified by interrupt type
	 * (This probably isn't necessary)
	 */
	write_sio(base,R2,0);

	/* 8 bit RX chars, auto enables off, no hunt mode, RxCRC enable,
	 * no address search, no inhibit sync chars, enable RX
	 */
	write_sio(base,R3,Rx8|RxCRC_ENAB|RxENABLE);

	/* X1 clock, SDLC mode, Sync modes enable, parity disable
	 * (Note: the DPLL does a by-32 clock division, so it's not necessary
	 * to divide here).
	 */
	write_sio(base,R4,X1CLK | SDLC | SYNC_ENAB);

	/* DTR On, 8 bit TX chars, no break, TX enable, SDLC CRC,
	 * RTS off, TxCRC enable
	 */
	write_sio(base,R5,DTR|Tx8|TxENAB|TxCRC_ENAB);

	/* SDLC flag */
	write_sio(base,R7,FLAG);

	/* No reset, status low, master int enable, enable lower chain,
	 * no vector, vector includes status
	 */
	write_sio(base,R9,MIE|NV|VIS);
	/* CRC preset 1, NRZI encoding, no active on poll, flag idle,
	 * flag on underrun, no loop mode, 8 bit sync
	 */
	write_sio(base,R10,CRCPS|NRZI);

	/* Board no longer channel-specific for clk.  The board should be set
	 * up to run from the 4.9152Mhz onboard crystal connected to PCLK.
	 * Both channels get receive clock at 32x from PCLK via the DPLL,
	 * with TRxC as an output, via a 4040 div by 32 counter to RTxC set
	 * us as an input to provide the transmit clock.
	 */

	/*            TRxC = BR Generator Output, TRxC O/I,
	 *	      transmit clock = RTxC pin, 
	 *	      receive clock = DPLL output
	 */
	write_sio(base,R11,TRxCBR|TRxCOI|TCRTxCP|RCDPLL);

	/* Compute and load baud rate generator time constant
	 * DPLL needs x32 clock
	 * XTAL is defined in dr.h to be the crystal clock / (2 * 32)
	 */
	tc = XTAL/(hp->speed) - 2;
	write_sio(base,R12,tc & 0xff);
	write_sio(base,R13,(tc >> 8) & 0xff);

	write_sio(base,R14,SNRZI);	/* Set NRZI mode */
	write_sio(base,R14,SSBR);	/* Set DPLL source = BR generator */
	write_sio(base,R14,SEARCH);	/* Enter search mode */
	/* Set baud rate gen source = PCLK, enable baud rate gen */
	write_sio(base,R14,BRENABL|BRSRC);

	/* Break/abort IE, TX EOM IE, CTS IE, no SYNC/HUNT IE, DCD IE,
	 * no Zero Count IE
	 */
	write_sio(base,R15,BRKIE|TxUIE|CTSIE|DCDIE);

	restore(i_state);
	if(hp->mode == FULLDUP){
		rts(base,ON);
	} else if(hp->tstate == IDLE){
		rts(base,OFF);
	}
	return 0;
}
/* Attach a dr-100 interface to the system
 * argv[0]: hardware type, must be "dr"
 * argv[1]: I/O address, e.g., "0x380"
 * argv[2]: vector, e.g., "2"
 * argv[3]: mode, must be:
 *	    "ax25" (AX.25 UI frame format)
 * argv[4]: interface label, e.g., "dr0"
 * argv[5]: receiver packet buffer size in bytes
 * argv[6]: maximum transmission unit, bytes
 * argv[7]: interface speed, e.g, "9600"
 */
dr_attach(argc,argv)
int argc;
char *argv[];
{
	register struct interface *if_dra,*if_drb;
	extern struct interface *ifaces;
	struct hdlc *hp;
	int dev;
	int dr_init();
	int dodr();
	int dr_stop();
	int ax_send();
	int ax_output();
	int dr_raw();
	int (*getirq())();	/* Getirq is a function returning a pointer to
				 * a function returning int */

	if(ndr >= Ndr){
		printf("Too many dr controllers\n");
		return -1;
	}
	dev = ndr++;

	/* Initialize hardware-level control structure */
	dr[dev].addr = htoi(argv[1]);
	dr[dev].save.vec = htoi(argv[2]);

	/* Save original interrupt vector */
	dr[dev].save.vec = getirq(dr[dev].vec);
	/* Set new interrupt vector */
	if(setirq(dr[dev].vec,pchandle[dev]) == -1){
		printf("IRQ %u out of range\n",dr[dev].vec);
		ndr--;
		return -1;
	}
	/* Create interface structures and fill in details */
	if_pca = (struct interface *)calloc(1,sizeof(struct interface));
	if_pcb = (struct interface *)calloc(1,sizeof(struct interface));

	if_pca->name = malloc((unsigned)strlen(argv[4])+1);
	strcpy(if_pca->name,argv[4]);
	if_pcb->name = malloc((unsigned)strlen(argv[4])+1);
	strcpy(if_pcb->name,argv[4]);
	if_pcb->name[strlen(argv[4]) - 1]++;	/* kludge */
	if_pcb->mtu = if_pca->mtu = atoi(argv[6]);
	if_pca->dev = 2*dev;
	if_pcb->dev = 2*dev + 1;
	if_pcb->recv = if_pca->recv = dopc;
	if_pcb->stop = if_pca->stop = pc_stop;
	if_pcb->output = if_pca->output = ax_output;
	if_pcb->raw = pc_raw;

	if(strcmp(argv[3],"ax25") == 0){

		axarp();
		if(mycall.call[0] == '\0'){
			printf("set mycall first\n");
			free((char *)if_pca);
			free((char *)if_pcb);
			return -1;
		}		
		if_pcb->send = if_pca->send = ax_send;
		if(if_pcb->hwaddr == NULLCHAR)
			if_pcb->hwaddr = malloc(sizeof(mycall));
		memcpy(if_pcb->hwaddr,(char *)&mycall,sizeof(mycall));
	} else {
		printf("Mode %s unknown for interface %s\n",
			argv[3],argv[4]);
		free((char *)if_pca);
		free((char *)if_pcb);
		return -1;
	}
	if_pca->next = if_pcb;
	if_pcb->next = ifaces;
	ifaces = if_pca;

	hp = &hdlc[2*dev+1];
	hp->speed = (int16)atoi(argv[7]);
	hp->base = dr[dev].addr + CHANB;
	hp->bufsiz = atoi(argv[5]);
	hdlcparam(hp);

	hp = &hdlc[2*dev];
	hp->speed = (int16)atoi(argv[7]);
	hp->base = dr[dev].addr + CHANA;
	hp->bufsiz = atoi(argv[5]);
	hdlcparam(hp);

	/* Clear mask (enable interrupt) in 8259 interrupt controller */
	maskon(dr[dev].vec);

	return 0;
}
dopc(interface)
struct interface *interface;
{
	struct hdlc *hp;
	struct mbuf *bp;

	hp = &hdlc[interface->dev];
	while((bp = dequeue(&hp->rcvq)) != NULLBUF){
		hp->rcvcnt--;
		dump(interface,IF_TRACE_IN,TRACE_AX25,bp);
		ax_recv(interface,bp);
	}	
}
pc_stop(iface)
struct interface *iface;
{
	int16 dev;

	dev = iface->dev;
	if(dev & 1)
		return;
	dev >>= 1;	/* Convert back into dr number */
	/* Turn off interrupts */
	maskoff(dr[dev].vec);

	/* Restore original interrupt vector */
	setirq(dr[dev].vec,dr[dev].save.vec);

	/* Force hardware reset */
	write_sio(dr[dev].addr + CHANA,R9,FHWRES);
}
	
/* Send raw packet on PC-100 */
pc_raw(interface,bp)
struct interface *interface;
struct mbuf *bp;
{
	char kickflag;
	struct hdlc *hp;

	dump(interface,IF_TRACE_OUT,TRACE_AX25,bp);
	hp = &hdlc[interface->dev];
	kickflag = (hp->sndq == NULL);
	enqueue(&hp->sndq,bp);
	if(kickflag)
		htxint(&hdlc[interface->dev]);
}
