/* Driver for 3COM Ethernet card */

#define        TIMER        20000        /* Timeout on transmissions */

#include <stdio.h>
#include "machdep.h"
#include "ether.h"
#include "ec.h"
#include "mbuf.h"
#include "iface.h"
#include "timer.h"
#include "ip.h"
#include "pc.h"
#include "arp.h"

unsigned char inportb();

int ec0vec();
int (*ecvec[])() = {ec0vec};
struct ec ec[EC_MAX];                /* Per-controller info */
unsigned nec;

#ifdef        notdef
static struct timer audit;
#endif        notdef
#ifdef        TRACE
#include "trace.h"
int ether_dump();
#endif        TRACE
        
/* Initialize interface */
ec_init(interface,bufsize)
struct interface *interface;
unsigned bufsize;        /* Maximum size of receive queue in PACKETS */
{
        register struct ec *ecp;
        register unsigned base;
        void eaudit();
        unsigned short getcs();
        int16 dev;
        char *malloc();

        dev = interface->dev;
        ecp = &ec[dev];
        base = ecp->base;
        ecp->rcvmax = bufsize;

        /* Pulse IE_RESET */
        outportb(IE_CSR(base),IE_RESET);

        /* Set interrupt vector */
        setvec(8+ecp->vec,getcs(),(unsigned)ecvec[dev]);

        /* Clear mask (enable interrupt) in 8259 interrupt controller */
        clrbit(INTMASK,(char)(1<<ecp->vec));

#ifdef        notdef
        if(inportb(IE_CSR(base)) != 0x80)
                printf("%s: can't reset IE_CSR\r\l",interface->name);
        if((inportb(EDLC_XMT(base)) & 0x0f) != 0)
                printf("%s: can't clear EDLC_XMT\r\l",interface->name);
        if((inportb(EDLC_RCV(base))&0x9f) != EDLC_STALE)
                printf("%s: can't reset EDLC_RCV\r\l",interface->name);
#endif        notdef

        if(interface->hwaddr == NULLCHAR)
                interface->hwaddr = malloc(EADDR_LEN);
        getaddr(base,interface->hwaddr);
        setaddr(base,interface->hwaddr);

        /* Enable DMA/interrupt request, gain control of buffer */
        outportb(IE_CSR(base),IE_RIDE|IE_SYSBFR);

        /* Turn off all transmit interrupts, since they aren't used */
        outportb(EDLC_XMT(base),0);

        /* Set up the receiver interrupts and flush status */
        outportb(EDLC_RCV(base),EDLC_MULTI|EDLC_GOOD|EDLC_ANY|EDLC_SHORT
         |EDLC_DRIBBLE|EDLC_FCS|EDLC_OVER);
        inportb(EDLC_RCV(base));

        /* Start receiver */
        outportw(IE_RP(base),0);        /* Reset read pointer */
        outportb(IE_CSR(base),IE_RIDE|IE_RCVEDLC);
#ifdef        notdef
        audit.start = 1;
        audit.arg = dev;
        audit.func = eaudit;
        start_timer(&audit);
#endif        notdef
}
#ifdef        notdef
/* Kick interrupt routine periodically to avoid lockups */
void
eaudit(dev)
int dev;
{
        ecint(dev);
        start_timer(&audit);
}
#endif        notdef

/* Send an IP datagram on Ethernet */
ec_send(bp,interface,gateway,precedence,delay,throughput,reliability)
struct mbuf *bp;                /* Buffer to send */
struct interface *interface;        /* Pointer to interface control block */
int32 gateway;                        /* IP address of next hop */
char precedence;
char delay;
char throughput;
char reliability;
{
        char *egate,*res_arp();

        egate = res_arp(interface,ARP_ETHER,gateway,bp);
        if(egate != NULLCHAR)
                (*interface->output)(interface,egate,interface->hwaddr,IP_TYPE,bp);
}

/* Send a packet with Ethernet header */
ec_output(interface,dest,source,type,bp)
struct interface *interface;                /* Pointer to interface control block */
char dest[];                /* Destination Ethernet address */
char source[];                /* Source Ethernet address */
int16 type;                /* Type field */
struct mbuf *bp;        /* Data field */
{
        register struct ether *ep;
        struct mbuf *hdr;

        /* Create the Ethernet packet header and prepend it to data */
        if((hdr = alloc_mbuf(HDR)) == NULLBUF){
                /* No space for header, drop packet and return */
                free_p(bp);
                return;
        }
        ep = (struct ether *)hdr->data;
        hdr->cnt = HDR;
        bcopy(dest,ep->dest,EADDR_LEN);
        bcopy(source,ep->source,EADDR_LEN);
        ep->type = htons(type);
        hdr->next = bp;
        bp = hdr;
        (*interface->raw)(interface,bp);
}

/* Send raw packet (caller provides header) */
ec_raw(interface,bp)
struct interface *interface;                /* Pointer to interface control block */
struct mbuf *bp;        /* Data field */
{
        register struct ec *ecp;
        register unsigned base;
        register int i;
        char i_state;
        int try,xstat;
        short size;

        ecp = &ec[interface->dev];
        base = ecp->base;

#ifdef        TRACE
        if(trace & TRACE_EC){
                if(trace & TRACE_HDR)
                        printf("%s sent:\r\l",interface->name);
                dotrace(ether_dump,bp);
        }
#endif        TRACE
        size = len_mbuf(bp);
        /* Pad the size out to the minimum, if necessary,
         * with junk from the last packet (nice security hole here)
         */
        if(size < RUNT)
                size = RUNT;
        size = (size+1) & ~1;        /* round size up to next even number */

#ifdef        notdef
        /* If board appears to have stuck in transmit state, reset it */
        if((inportb(IE_CSR(base)) & IE_BUFCTL) == IE_XMTEDLC){
                outportb(IE_CSR(base),IE_RESET|IE_RIDE|IE_SYSBFR);
                outportb(IE_CSR(base),IE_RIDE|IE_SYSBFR);
        }
#endif        notdef
        i_state = disable();        /* No distractions from the receiver */

        /* Get control of the board buffer and disable receiver */
        outportb(IE_CSR(base),IE_RIDE|IE_SYSBFR);
        /* Point GP at beginning of packet */
        outportw(IE_GP(base),BFRSIZ-size);
        /* Actually load each piece with a fast assembler routine */
        while(bp != NULLBUF){
                outbuf(IE_BFR(base),bp->data,bp->cnt);
                bp = free_mbuf(bp);
        }
        for(try=0;try < MAXTRIES;try++){
                /* Crank counter back to beginning and start transmit */
                outportw(IE_GP(base),BFRSIZ-size);
                outportb(IE_CSR(base),IE_RIDE|IE_XMTEDLC);
                outportw(IE_RP(base),0);
                /* Busy wait for the transmitter to complete (it's fast) */
                for(i=TIMER;(inportb(IE_CSR(base)) & IE_XMTBSY) && i != 0;i--)
                        ;
                if(i == 0)
                        break;        /* Timeout */
                xstat = inportb(EDLC_XMT(base));
                if(xstat & EDLC_JAM){
                        ecp->estats.jams++;
#ifdef        notdef
                        printf("%s: jam\r\l",interface->name);
                        fflush(stdout);
#endif        notdef
                        continue;        /* Give it another shot */
                } else {
                        break;
                }
        }
        restore(i_state);
        if(i == 0)
                ecp->estats.timeout++;
#ifdef        notdef
        /* Restart receiver if necessary */
        if(!(inportb(IE_CSR(base)) & IE_RCVBSY)){
                outportb(IE_CSR(base),IE_RESET|IE_RIDE|IE_SYSBFR);
                outportb(IE_CSR(base),IE_RIDE|IE_RCVEDLC);
        }
#endif        notdef
}

/* Ethernet interrupt handler */
int
ecint(dev)
int dev;
{
        register struct ec *ecp;
        register unsigned base;
        struct mbuf *bp;
        int16 size;
        char rxstat,once;

        ecp = &ec[dev];
        base = ec[dev].base;
        once = 0;
        ecp->estats.ints++;

        for(;;){
                rxstat = inportb(EDLC_RCV(base));
                if(rxstat & EDLC_STALE){
                        if(!once)
                                ecp->estats.stale++;
                        break;
                }
                once++;
                if(rxstat & EDLC_SHORT){
                        ecp->estats.shorts++;
                        rcv_fixup(base);
                        continue;
                }
                if(rxstat & EDLC_OVER){
                        ecp->estats.over++;
                        rcv_fixup(base);
                        continue;
                }
                if(rxstat & EDLC_DRIBBLE){
                        ecp->estats.dribble++;
                        continue;
                }
                if(rxstat & EDLC_FCS){
                        ecp->estats.fcs++;
                        continue;
                }
                if(rxstat & EDLC_ANY){
                        ecp->estats.any++;
                        /* Get control of the buffer */
                        outportw(IE_GP(base),0);
                        outportb(IE_CSR(base),IE_RIDE|IE_SYSBFR);
                
                        /* Allocate mbuf and copy the packet into it */
                        size = inportw(IE_RP(base));
                        if(size < RUNT || size > GIANT)
                                ecp->estats.badsize++;
                        else if(ecp->rcvcnt >= ecp->rcvmax)
                                ecp->estats.drop++;
                        else if((bp = alloc_mbuf(size)) == NULLBUF)
                                ecp->estats.nomem++;
                        else {
                                inbuf(IE_BFR(base),bp->data,size);
                                bp->cnt = size;
                                ecp->rcvcnt++;
                                enqueue(&ecp->rcvq,bp);/* Put it on the receive queue */
                        }
                        outportb(IE_CSR(base),IE_RIDE|IE_RCVEDLC);
                        outportb(IE_RP(base),0);
                }
        }
}
static
rcv_fixup(base)
register unsigned base;
{
        outportb(IE_CSR(base),IE_RIDE|IE_SYSBFR);
        outportb(IE_CSR(base),IE_RIDE|IE_RCVEDLC);
        outportb(IE_RP(base),0);
}
/* Process any incoming Ethernet packets on the receive queue */
int
ec_recv(interface)
struct interface *interface;
{
        void arp_input(),ip_route();
        struct ec *ecp;
        struct mbuf *bp;
        struct ether hdr;
        char multicast;

        ecp = &ec[interface->dev];
        if((bp = dequeue(&ecp->rcvq)) != NULLBUF){
                ecp->rcvcnt--;
#ifdef        TRACE
        if(!(bp->data[0] & 1) || trace & TRACE_BDCST){
                if(trace & TRACE_HDR)
                        printf("%s recv:\r\l",interface->name);
                dotrace(ether_dump,bp);
        }
#endif        TRACE
                /* Remove Ethernet header and kick packet upstairs */
                pullup(&bp,(char *)&hdr,HDR);
                if(hdr.dest[0] & 1)
                        multicast = 1;
                else
                        multicast = 0;
                switch(ntohs(hdr.type)){
                case ARP_TYPE:
                        arp_input(interface,bp);
                        break;
                case IP_TYPE:
                        ip_route(bp,multicast);
                        break;
                default:
                        ecp->estats.badtype++;
                        free_p(bp);
                        break;
                }
        }
}
/* Read Ethernet address from controller PROM */
static
getaddr(base,cp)
register unsigned base;
register char *cp;
{
        register int i;

        for(i=0;i<EADDR_LEN;i++){
                outportw(IE_GP(base),i);
                *cp++ = inportb(IE_SAPROM(base));
        }
}
/* Set Ethernet address on controller */
static
setaddr(base,cp)
register unsigned base;
register char *cp;
{
        register int i;

        for(i=0;i<EADDR_LEN;i++)
                outportb(EDLC_ADDR(base)+i,*cp++);
}
/* Shut down the Ethernet controller */
ec_stop(interface)
struct interface *interface;
{
        register unsigned base;
        int16 dev;

        dev = interface->dev;
        base = ec[dev].base;
        /* Set mask (disable interrupt) in 8259 interrupt controller */
        setbit(INTMASK,(char)(1<<ec[dev].vec));

        /* Pulse IE_RESET */
        outportb(IE_CSR(base),IE_RESET);
        outportb(IE_CSR(base),0);
}
/* Attach a 3-Com model 3C500 Ethernet controller to the system
 * argv[0]: hardware type, must be "3c500"
 * argv[1]: I/O address, e.g., "0x300"
 * argv[2]: vector, e.g., "3"
 * argv[3]: mode, must be "arpa"
 * argv[4]: interface label, e.g., "ec0"
 * argv[5]: maximum number of packets allowed on receive queue, e.g., "5"
 * argv[6]: maximum transmission unit, bytes, e.g., "1500"
 */
ec_attach(argc,argv)
int argc;
char *argv[];
{
        register struct interface *if_ec;
        extern struct interface *ifaces;
        unsigned dev;
        char *calloc(),*malloc();
        int ec_init();
        int ec_send();
        int ec_recv();
        int ec_stop();

        if(nec >= EC_MAX){
                printf("Too many Ethernet controllers\r\l");
                return -1;
        }
        dev = nec++;
        if_ec = (struct interface *)calloc(1,sizeof(struct interface));

        if_ec->name = malloc((unsigned)strlen(argv[4])+1);
        strcpy(if_ec->name,argv[4]);
        if_ec->mtu = atoi(argv[6]);
        if_ec->send = ec_send;
        if_ec->output = ec_output;
        if_ec->raw = ec_raw;
        if_ec->recv = ec_recv;
        if_ec->stop = ec_stop;
        if_ec->dev = dev;
        if_ec->flags = IF_BROADCAST;

        ec[dev].base = htoi(argv[1]);
        ec[dev].vec = htoi(argv[2]);

        if(strcmp(argv[3],"arpa") != 0){
                printf("Mode %s unknown for interface %s\r\l",
                        argv[3],argv[4]);
                free((char *)if_ec);
                return -1;
        }
        if_ec->next = ifaces;
        ifaces = if_ec;
        /* Initialize device */
        ec_init(if_ec,(unsigned)atoi(argv[5]));
        return 0;
}
#ifdef        TRACE

ether_dump(bpp)
struct mbuf **bpp;
{
        struct ether ehdr;
        char s[20],d[20];

        pullup(bpp,(char *)&ehdr,sizeof(struct ether));
        pether(s,ehdr.source);
        pether(d,ehdr.dest);
        printf("Ether: %s->%s type 0x%x len %u\r\l",s,d,ntohs(ehdr.type),
                sizeof(struct ether) + len_mbuf(*bpp));

        if((trace & TRACE_HDR) > 2){
                switch(ntohs(ehdr.type)){
                        case IP_TYPE:
                                ip_dump(bpp);
                                break;
                        case ARP_TYPE:
                                arp_dump(bpp);
                                break;
                }
        }
}
int
doetherstat(argc,argv)
int argc;
char *argv[];
{
        struct ec *ecp;

        for(ecp = ec;ecp < &ec[nec]; ecp++){
                printf("Controller %u:\r\l",ecp-ec);

                printf("intrpt    spur      shorts    dribble   fcs       overf     any       jams\r\l");
                printf("%-10ld%-10u%-10u%-10u%-10u%-10u%-10ld%-u\r\l",
                 ecp->estats.ints,ecp->estats.spur,ecp->estats.shorts,
                 ecp->estats.dribble,ecp->estats.fcs,ecp->estats.over,ecp->estats.any,
                 ecp->estats.jams);

                printf("stale     badsize   output    timeout   badtype   drop      nomem\r\l");
                printf("%-10u%-10u%-10ld%-10u%-10u%-10u%-10u\r\l",
                 ecp->estats.stale,ecp->estats.badsize,ecp->estats.out,
                 ecp->estats.timeout,ecp->estats.badtype,ecp->estats.drop,
                 ecp->estats.nomem);
        }
        return 0;
}
#endif        TRACE

