/* Driver for Western Digital WD8003E Ethernet card                       */
/* By Bob Clements, K1BC, May 1988 for the KA9Q TCP/IP package            */
/* Symbol prefix "EW" is for Ethernet, Western-digital card               */

#define TIMER	20000	/* Timeout on transmissions */

#include <stdio.h>
#include <memory.h>
#include "global.h"
#include "mbuf.h"
#include "enet.h"
#include "iface.h"
#include "ec.h"
#include "ew.h"
#include "timer.h"
#include "arp.h"
#include "trace.h"
#include "pc.h"


extern char *abstoptr();

int ew0vec();
int ew1vec();
int ew2vec();
int (*ewvec[])() = {ew0vec,ew1vec,ew2vec};
extern struct ec ec[EC_MAX];		/* Per-controller info */
extern unsigned nec;
extern int ec_output();
extern int ec_recv();

/* Initialize interface */
int
ew_init(interface,bufsize)
struct interface *interface;
unsigned bufsize;	/* Maximum size of receive queue in PACKETS */
{
	register struct ec *ecp;
	register unsigned base;
	void eaudit();
	unsigned short getcs();
	int16 dev;
	int16 membase;

	dev = interface->dev;
	ecp = &ec[dev];
	base = ecp->base;
	ecp->rcvmax = bufsize;
	ecp->iface = interface;

	/* Reset the board */
 	outportb(EW_CMD(base),EW_RESET);
 	outportb(EW_CMD(base),0);
	outportb(EW_CCMD(base),EWC_NODMA+EWC_PAGE0);
	outportb(EW0_ISR(base), EWISR_ALL);	/* Clear all interrupt flags */

	outportb(EW0_DCFG(base),EWDCFG_BM8);	/* Configure fifo organization */
	outportb(EW0_RCNTLO(base),0);	/* Clear byte count regs */
	outportb(EW0_RCNTHI(base),0);
	outportb(EW0_RXCR(base),EWRXCR_MON);	/* Set Rcvr to monitor mode */
	outportb(EW0_TXCR(base),0);				/* Set Xmtr to normal */

	/* Turn on the shared memory block */
	outportb(EW_CMD(base),EW_MEMEN + (ecp->mem>>9 & EW_MEM_MASK));
	/* Set up control of shared memory, buffer ring, etc. */
	outportb(EW0_STARTPG(base),SM_RSTART_PG);
	outportb(EW0_STOPPG(base),SM_RSTOP_PG);
	outportb(EW0_BOUNDARY(base),SM_RSTART_PG);
	ecp->txb_use = 0;			/* Transmit buffer pointers */

	outportb(EW0_ISR(base),0xff);			/* Clear all pending interrupts */

	if(setirq(ecp->vec,ewvec[dev]) == -1){
		printf("IRQ %u out of range\n",ecp->vec);
		return -1;
	}
	maskon(ecp->vec);		/* Enable interrupt */

	/* Set interrupt vector */
	outportb(EW0_IMR(base),EWISR_ALL);			/* Interrupts we will take */
	outportb(EW_CCMD(base),EWC_NODMA+EWC_PAGE1);
	outportb(EW1_CURPAG(base),SM_RSTART_PG+1);	/* Current shared page */
	outportb(EW_CCMD(base),EWC_NODMA+EWC_PAGE0);
 
	if(interface->hwaddr == NULLCHAR)
		interface->hwaddr = malloc(EADDR_LEN);
	getewaddr(base,interface->hwaddr);
	setewaddr(base,interface->hwaddr);
	if(memcmp(interface->hwaddr,ether_bdcst,EADDR_LEN) == 0){
		printf("wd8003e address PROM contains broadcast address!!\n");
		return -1;
	}

	outportb(EW_CCMD(base),EWC_START+EWC_NODMA); /* Start DS8390 */
	outportb(EW0_RXCR(base),EWRXCR_BCST);	/* Receive broadcasts */

	return 0;
}

/*
 * Here is the real transmit-frame routine.  The device allows
 * multiple frames to be loaded into the shared memory for transmission.
 * Since each maximum-size frame takes up one sixth of the memory,
 * I have chosen to only queue one buffer at a time into the shared
 * memory, leaving most of it for received frames.  This means that
 * we can't transmit two frames with minimum interframe spacing.
 * That doesn't bother me very much.  -- K1BC
 */

/* Send raw packet (caller provides header) */
ew_raw(interface,bp)
struct interface *interface;	/* Pointer to interface control block */
struct mbuf *bp;		/* Data field */
{
	register struct ec *ecp;
	register unsigned base;
	register int i;
	char i_state;
	short size;
	unsigned txbase;
	unsigned txoffset;

	ecp = &ec[interface->dev];
	base = ecp->base;

	ecp->estats.xmit++;
	dump(interface,IF_TRACE_OUT,TRACE_ETHER,bp);

	size = len_mbuf(bp);
	/* Pad the size out to the minimum, if necessary,
	 * with junk from the last packet (nice security hole here)
	 */
	if(size < RUNT)
		size = RUNT;

	/* Wait for transmitter to be idle, if necessary. */

	for(i=TIMER;(ecp->txb_use != 0) && i != 0;i--)
		;
	if(i == 0){
		ecp->estats.timeout++;
		free_p(bp);
		return;
	}						/* Now TX buffer is free to fill */

	ecp->txb_use++;			/* Make buffer busy */
	ecp->size = size;

	txbase   = ecp->mem + (SM_TSTART_PG << 4);
	txoffset = 0;

	while(bp != NULLBUF){	/* Move the data into shared memory */
		ew_outbuf(txbase,txoffset,bp->data,bp->cnt);
		txoffset += bp->cnt;
		bp = free_mbuf(bp);
	}

	/* Tell transmitter the byte count of the frame */
	outportb(EW0_TCNTLO(base), size & 0xff);
	outportb(EW0_TCNTHI(base), (size >> 8) & 0xff);
	outportb(EW0_TPSR(base), SM_TSTART_PG);		/* Starting page number */

	/* Start transmitter */
	outportb(EW_CCMD(base),EWC_TRANS+EWC_NODMA);
}

/* Ethernet interrupt handler */
int
ewint(dev)
int dev;
{
	register struct ec *ecp;
	register unsigned base;
	struct mbuf *bp;
	int i;
	int16 size;
	int16 pg;
	int16 curr;
	char stat;
	char pstat;
	unsigned rxpara;
	unsigned rxoffset;

	ecp = &ec[dev];
	base = ec[dev].base;
	ecp->estats.intrpt++;

check_isr:
	/* Was there an interrupt from this card? */
	if ((EWISR_ALL & (stat = inportb(EW0_ISR(base)))) == 0)
		goto interrupt_done;

/* First, a messy procedure for handling the case where the rcvr
 * over-runs its ring buffer.  This is spec'ed by National for the chip. */

	if (stat & EWISR_OVER){
		outportb(EW_CCMD(base), EWC_STOP+EWC_NODMA);	/* Stop the card */
		/* Remove one frame from the ring */
		pg = inportb(EW0_BOUNDARY(base));
        if (++pg == SM_RSTOP_PG)
			pg = SM_RSTART_PG;

		rxpara = ecp->mem + (pg << 4);
		pstat = ew_gbyte(rxpara, EW_RBUF_STAT);

		if (pstat & EWPS_RXOK)
			rcv_frm(ecp, pg);		/* Accept this frame */
		pg = ew_gbyte(rxpara, EW_RBUF_NXT_PG);
		if (--pg < SM_RSTART_PG)
			pg = SM_RSTOP_PG - 1;	/* Wrap-around if needed */
		outportb(EW0_BOUNDARY(base), pg);	/* Set new boundary */
		outportb(EW0_RCNTLO(base), 0);	/* Clear Rcv byte count */
		outportb(EW0_RCNTHI(base), 0);
		for (i = TIMER; i > 0; i--)
			if (inportb(EW0_ISR(base)) & EWISR_RESET)
				break;				/* Wait for it to reset */
		outportb(EW0_TXCR(base), EWTXCR_LOOP);	/* Set loopback mode */
		outportb(EW_CCMD(base), EWC_START + EWC_NODMA);	/* Start it */
		outportb(EW0_TXCR(base), 0);		/* Turn off loopback */
		outportb(EW0_ISR(base), EWISR_OVER);	/* Clear overrun int req */
		ecp->estats.over++;					/* Count rcv overruns */
	goto check_isr;
	}
 
/* Handle receive flags, normal and with error (but not overrun). */

	if (stat & (EWISR_RX | EWISR_RX_ERR)){
		for (;;){	/* Loop over all buffered RX frames */
			outportb(EW_CCMD(base),EWC_NODMA+EWC_PAGE1);
			curr = inportb(EW1_CURPAG(base));
			outportb(EW_CCMD(base),EWC_NODMA+EWC_PAGE0);
			pg = inportb(EW0_BOUNDARY(base));
			if (++pg == SM_RSTOP_PG)
				pg = SM_RSTART_PG;
			if (pg == curr)
				break;					/* Done all frames in shared memory */

			rxpara = ecp->mem + (pg << 4);

			pstat = ew_gbyte(rxpara, EW_RBUF_STAT);	/* Status for this frame */
			if (pstat & EWPS_RXOK)
				rcv_frm(ecp, pg);		/* Accept this frame */
			pg = ew_gbyte(rxpara, EW_RBUF_NXT_PG);	/* Starting page of next frame */
			if (--pg < SM_RSTART_PG)
				pg = SM_RSTOP_PG - 1;	/* Wrap-around if needed */
			outportb(EW0_BOUNDARY(base), pg);	/* Set new boundary */
			}
		outportb(EW0_ISR(base), EWISR_RX | EWISR_RX_ERR | EWISR_OVER);
		goto check_isr ;
	}
  
/* Handle transmit flags. */

	if (stat & (EWISR_TX | EWISR_TX_ERR)){
		pstat = inportb(EW0_TSR(base));
		if (stat & EWISR_TX){					/* If not an error */
			if (pstat & EWTSR_COLL)
				ecp->estats.jam++;				/* Collided at least once */
			if (pstat & EWTSR_COLL16)
				ecp->estats.jam16++;			/* Coll 16 times, dropped */
			outportb(EW0_ISR(base), EWISR_TX);	/* Clear TX done flag */
		} else if (stat & EWISR_TX_ERR) {		/* If TX error */
			if (pstat & EWTSR_FU)
				ecp->estats.jam16++;			/* Need a counter for underrun */
		/* Count TX underruns */
			outportb(EW0_ISR(base), EWISR_TX_ERR);	/* Clear TX error flag */
		}

	ecp->txb_use = 0;			/* Count one fewer outstanding TX buf */
								/* Should be --, not just clear, if multi */
								/* buffers in use. */
	
	/* If TX queue and/or TX shared memory ring buffer were being
	 * used, logic to step through them would go here.  However,
	 * in this version, we just clear the flags and the txb_use
	 * count for background to notice.
	 */

	goto check_isr ;			/* See if any other interrupts on */
	}

/* Now check to see if any counters are getting full */

	if (stat & EWISR_COUNTERS){
		i = inportb(EW0_COUNTER0(base));	/* Get alignment err counter */
			ecp->estats.bad += i;			/* Could be individual cntrs */
		i = inportb(EW0_COUNTER1(base));	/* Get CRC err counter */
			ecp->estats.bad += i;	
		i = inportb(EW0_COUNTER2(base));	/* Get missed frame err counter */
			ecp->estats.bad += i;	
		outportb(EW0_ISR(base), EWISR_COUNTERS); /* Clear counters flag */
	goto check_isr ;			/* Anything else to do? */
	}

interrupt_done: ;
}


/* Copy the received frame from shared memory to buffer */
/* Arg is page number of frame in shared memory */

static
rcv_frm(ecp, pg)
struct ec *ecp;			/* Ethernet control info structure */
int16 pg;				/* Page number in shared memory */
{
	struct mbuf *bp;
	unsigned     size;
	int16        n;
	char         pstat;
	char		*cp;
	unsigned	 rxpara;

	rxpara = ecp->mem + (pg << 4);

	size = (ew_gbyte(rxpara, EW_RBUF_SIZE_HI) << 8) +
            ew_gbyte(rxpara, EW_RBUF_SIZE_LO) - EW_RBUF_NHDR;

	if (size < RUNT || size > GIANT)
		ecp->estats.bad++;
	else if (ecp->rcvcnt >= ecp->rcvmax)		/* Queue getting too big? */
		ecp->estats.drop++;
	else if ((bp = alloc_mbuf(size)) == NULLBUF)
		ecp->estats.nomem++;
	else {
		cp = bp->data;						/* Where to put them */
		ecp->estats.recv++;

		if ( (((size + 255 + EW_RBUF_NHDR) >> 8) + pg) > SM_RSTOP_PG){
			/* Copy in two pieces due to buffer wraparound. */

			n = ((SM_RSTOP_PG - pg) << 8) - EW_RBUF_NHDR;	/* To top of mem */
			ew_inbuf(rxpara, EW_RBUF_NHDR, cp, n);	/* Copy that part */

			cp += n;
			n = size - n;
			rxpara = ecp->mem + (SM_RSTART_PG << 4);
			ew_inbuf(rxpara, 0, cp, n);

		} else {
			ew_inbuf(rxpara, EW_RBUF_NHDR, cp, size);
		}
		bp->cnt = size;
		ecp->rcvcnt++;
		enqueue(&ecp->rcvq, bp);	/* Put it in the RX queue */
	}
}

/* Read Ethernet address from controller PROM */
static
getewaddr(base,cp)
register unsigned base;
register char *cp;
{
	register int i;

	for(i=0; i<EADDR_LEN; i++){
		*cp++ = inportb(EW_SAPROM(base + i));
	}
}
/* Set Ethernet address on controller */
static
setewaddr(base,cp)
register unsigned base;
register char *cp;
{
	register int i;

	outportb(EW_CCMD(base),EWC_NODMA+EWC_PAGE1);
	for(i=0; i<EADDR_LEN; i++)				/* Set board's address filter */
		outportb(EW1_PHYS(base)+i,*cp++);

	for(i=0;i<8;i++)						/* Clear multicast addr */
		outportb(EW1_MULT(base)+i,0);

	outportb(EW_CCMD(base),EWC_NODMA+EWC_PAGE0);
}

/* Shut down the Ethernet controller */
ew_stop(interface)
struct interface *interface;
{
	register unsigned base;
	int16 dev;

	dev = interface->dev;
	base = ec[dev].base;

	/* Disable interrupt */
	maskoff(ec[dev].vec);

	/* Pulse IE_RESET */
	outportb(EW_CCMD(base),EWC_STOP+EWC_NODMA); /* Stop DS8390 */
	outportb(EW0_ISR(base),0xff);			/* Clear all pending interrupts */
	outportb(EW0_IMR(base),0x00);			/* Turn off all enables */
}

/* Attach a Western Digital model 8003E Ethernet controller to the system
 * argv[0]: hardware type, must be "wd8003e"
 * argv[1]: I/O address, e.g., "0x280"
 * argv[2]: vector, e.g., "2"
 * argv[3]: mode, must be "arpa"
 * argv[4]: interface label, e.g., "ec0"
 * argv[5]: maximum number of packets allowed on receive queue, e.g., "5"
 * argv[6]: maximum transmission unit, bytes, e.g., "1500"
 * argv[7]: Paragraph # of base address for shared memory area, e.g., "0xc400"
 */
ew_attach(argc,argv)
int argc;
char *argv[];
{
	register struct interface *if_ec;
	extern struct interface *ifaces;
	unsigned dev;
	int ew_init();
	int ec_send();
	int ec_recv();
	int ew_stop();
	int pether(),gaether();

	if(strcmp(argv[0],"wd8003e") != 0){
		printf("Device type %s unknown for interface %s\n",
			argv[0],argv[4]);
		free(if_ec->name);
		free((char *)if_ec);
		return -1;
	}

	arp_init(ARP_ETHER,EADDR_LEN,IP_TYPE,ARP_TYPE,ether_bdcst,pether,gaether);
	if(nec >= EC_MAX){
		printf("Too many Ethernet controllers\n");
		return -1;
	}
	dev = nec++;
	if((if_ec = (struct interface *)calloc(1,sizeof(struct interface))) == NULLIF
	 ||(if_ec->name = malloc((unsigned)strlen(argv[4])+1)) == NULLCHAR){
		printf("ew_attach: no memory!\n");
		return -1;
	}
	strcpy(if_ec->name,argv[4]);
	if_ec->mtu = atoi(argv[6]);
	if_ec->send = ec_send;
	if_ec->output = ec_output;
	if_ec->raw = ew_raw;
	if_ec->recv = ec_recv;
	if_ec->stop = ew_stop;
	if_ec->dev = dev;

	ec[dev].base = htoi(argv[1]);
	ec[dev].vec = htoi(argv[2]);
	ec[dev].mem = argc > 7 ? (htoi(argv[7]) & 0xfe00) : SM_BASE;
								 /* 8K boundary, default 0xc400 */

	if(strcmp(argv[3],"arpa") != 0){
		printf("Mode %s unknown for interface %s\n",
			argv[3],argv[4]);
		free(if_ec->name);
		free((char *)if_ec);
		return -1;
	}
	/* Initialize device */
	if(ew_init(if_ec,(unsigned)atoi(argv[5])) != 0){
		free(if_ec->name);
		free((char *)if_ec);
		return -1;
	}
	if_ec->next = ifaces;
	ifaces = if_ec;

	return 0;
}
