/* Stuff specific to the Western Digital WD003E Ethernet controller board */
/* By Bob Clements, K1BC, May 1988 for the KA9Q TCP/IP package            */
/* Symbol prefix "EW" is for Ethernet, Western-digital card               */

/* Find base address of specified device */
#define	EW(dev)	(ec_dev[dev].addr)

/* The EW registers - First, the board registers */
#define	EW_CMD(num)	((num)+0x0)		/* Board's command register */
#define	EW_SAPROM(num)	((num)+0x8)	/* window on station addr prom */

/* The EW registers - Next, the DS8390 chip registers */
/* There are two (really 3) pages of registers in the chip. You select *
 * which page you want, then address them at offsets 10-1F from base.  *
 * The chip command register (EW_CCMD) appears in both pages.          */
#define EW_CCMD(num)	((num)+0x10)	/* Chip's command register     */
/* Page 0 */
#define EW0_STARTPG(num) ((num)+0x11)	/* Starting page of ring bfr */
#define EW0_STOPPG(num) ((num)+0x12)	/* Ending page +1 of ring bfr */
#define EW0_BOUNDARY(num) ((num)+0x13)	/* Boundary page of ring bfr */
#define EW0_TSR(num)	((num)+0x14)	/* Transmit status reg */
#define EW0_TPSR(num)	((num)+0x14)	/* Transmit starting page */
#define EW0_TCNTLO(num) ((num)+0x15)	/* Low  byte of tx byte count */
#define EW0_TCNTHI(num) ((num)+0x16)	/* High byte of tx byte count */
#define EW0_ISR(num)	((num)+0x17)	/* Interrupt status reg */
#define EW0_RCNTLO(num)	((num)+0x1a)	/* Remote byte count reg */
#define EW0_RCNTHI(num)	((num)+0x1b)	/* Remote byte count reg */
#define EW0_RXCR(num)	((num)+0x1c)	/* RX control reg */
#define EW0_TXCR(num)	((num)+0x1d)	/* TX control reg */
#define EW0_COUNTER0(num) ((num)+0x1d)	/* Rcv alignment error counter */
#define EW0_DCFG(num)	((num)+0x1e)	/* Data configuration reg */
#define EW0_COUNTER1(num) ((num)+0x1e)	/* Rcv CRC error counter */
#define EW0_IMR(num)	((num)+0x1f)	/* Interrupt mask reg */
#define EW0_COUNTER2(num) ((num)+0x1f)	/* Rcv missed frame error counter */
/* Page 1 */
#define EW1_PHYS(num)	((num)+0x11)	/* This board's physical enet addr */
#define EW1_CURPAG(num)	((num)+0x17)	/* Current memory page */
#define EW1_MULT(num)	((num)+0x18)	/* Desired multicast addr */


/* Board commands in EW_CMD */
#define EW_RESET 0x80		/* Reset the board */
#define EW_MEMEN 0x40		/* Enable the shared memory */
#define EW_MEM_MASK	0x3f	/* B18-B13 of address of the shared memory */

/* Chip commands in EW_CCMD */
#define EWC_STOP 0x01		/* Stop the chip */
#define EWC_START 0x02		/* Start the chip */
#define EWC_TRANS 0x04		/* Transmit a frame */
#define EWC_NODMA 0x20		/* No remote DMA used on this card */
#define EWC_PAGE0 0x00		/* Select page 0 of chip registers */
#define EWC_PAGE1 0x40		/* Select page 1 of chip registers */

/* Commands for RX control reg */
#define EWRXCR_MON 0x20		/* Monitor mode */
#define EWRXCR_BCST 0x04	/* Accept broadcasts */

/* Commands for TX control reg */
#define EWTXCR_LOOP 0x02	/* Set loopback mode */

/* Bits in EW0_DCFG - Data config register */
#define EWDCFG_BM8 0x48		/* Set burst mode, 8 deep FIFO */

/* Bits in EW0_ISR - Interrupt status register */
#define EWISR_RX 0x01		/* Receiver, no error */
#define EWISR_TX 0x02		/* Transmitter, no error */
#define EWISR_RX_ERR 0x04	/* Receiver, with error */
#define EWISR_TX_ERR 0x08	/* Transmitter, with error */
#define EWISR_OVER 0x10		/* Receiver overwrote the ring */
#define EWISR_COUNTERS 0x20	/* Counters need emptying */
#define EWISR_RESET 0x80	/* Reset completed */
#define EWISR_ALL 0x3f		/* Interrupts we will enable */

/* Bits in received packet status byte and EW0_RSR */
#define EWPS_RXOK 0x01		/* Received a good packet */

/* Bits in TX status reg */

#define EWTSR_COLL 0x04		/* Collided at least once */
#define EWTSR_COLL16 0x08	/* Collided 16 times and was dropped */
#define EWTSR_FU 0x20		/* TX FIFO Underrun */

/* Shared memory management parameters */

#define SM_TSTART_PG 0		/* First page of TX buffer */
#define SM_RSTART_PG 6		/* Starting page of ring */
#define SM_RSTOP_PG 32		/* Last page +1 of ring */
#define SM_BASE 0xC400		/* Default para where shared memory starts */
							/* Real value set at attach time. */

#define EW_RBUF_STAT 0			/* Received frame status */
#define EW_RBUF_NXT_PG 1		/* Page after this frame */
#define EW_RBUF_SIZE_LO 2		/* Length of this frame  */
#define EW_RBUF_SIZE_HI 3		/* Length of this frame  */
#define EW_RBUF_NHDR 4			/* Length of above header area */

/* End of EW.H */
