; EWVEC.ASM - Routines to move blocks of memory from shared
;             memory of WD8003E Ethernet controller.  Needed because
;             this data area is outside the "small data" model range.
; This is a slightly modified version of Phil's routine for I/O port moves,
; by Bob Clements, K1BC, May 1988 for the KA9Q TCP/IP package.

; Note: There should not be ew0vec...ew2vec. They should be integrated
; 	into ec0vec...ec2vec, which should call ewint_ or ecint_ as
;	appropriate for the device.


	include lmacros.h

; Conditional ES save/restore macros not in lmacros.h
pushes	macro
	ifdef LONGPTR
	push	es
	endif
	endm

popes	macro
	ifdef LONGPTR
	pop	es
	endif
	endm

	assume	ds:dataseg
	assume	cs:codeseg
	public	sssave,spsave,intstk

	ifdef	FARPROC
	extrn	doret:far,ewint_:far,getintds:far
	else
	extrn	doret:near,ewint_:near,getintds:near
	endif


; ew0vec - Ethernet interrupt handler
	public	ew0vec_

ew0vec_ proc	far
	push	ds		; save on user stack
	call	getintds	; establish interrupt data segment

	mov	ds:sssave,ss	; stash user stack context
	mov	ds:spsave,sp

	push	ds
	pop	ss
	lea	sp,intstk+512

	push	ax		; save user regs on interrupt stack
	push	bx
	push	cx
	push	dx
	push	bp
	push	si
	push	di
	push	es
	push	ds
	pop	es

	mov	ax,0		; arg for service routine
	push	ax
	call	ewint_
	pop	ax
	jmp	doret
ew0vec_	endp

; ew1vec - Ethernet interrupt handler
	public	ew1vec_

ew1vec_ proc	far
	push	ds		; save on user stack
	call	getintds	; establish interrupt data segment

	mov	ds:sssave,ss	; stash user stack context
	mov	ds:spsave,sp

	push	ds
	pop	ss
	lea	sp,intstk+512

	push	ax		; save user regs on interrupt stack
	push	bx
	push	cx
	push	dx
	push	bp
	push	si
	push	di
	push	es
	push	ds
	pop	es

	mov	ax,1		; arg for service routine
	push	ax
	call	ewint_
	pop	ax
	jmp	doret
ew1vec_	endp

; ew2vec - Ethernet interrupt handler
	public	ew2vec_

ew2vec_ proc	far
	push	ds		; save on user stack
	call	getintds	; establish interrupt data segment

	mov	ds:sssave,ss	; stash user stack context
	mov	ds:spsave,sp

	push	ds
	pop	ss
	lea	sp,intstk+512

	push	ax		; save user regs on interrupt stack
	push	bx
	push	cx
	push	dx
	push	bp
	push	si
	push	di
	push	es
	push	ds
	pop	es

	mov	ax,2		; arg for service routine
	push	ax
	call	ewint_
	pop	ax
	jmp	doret
ew2vec_	endp


; ew_outbuf - put a buffer to an area of memory
	procdef ew_outbuf,<<opara,word>,<ooffs,word>,<obuf,ptr>,<ocnt,word>>
	pushf
	push	si
	pushds
	push	es
	push	di
	mov	dx,opara
	mov	es,dx
	mov	dx,ooffs
	mov	di,dx		; es:di = shared mem buffer
	mov	cx,ocnt
	ldptr	si,obuf,ds	; ds:si = obuf
	cld

; If buffer doesn't begin on a word boundary, send the first byte
	test	si,1	; (buf & 1) ?
	jz	obufeven ; no
	movsb		; *di++ = *si++;
	dec	cx	; cx--;
	mov	ocnt,cx	; save for later test
obufeven:
	shr	cx,1	; cx = cnt >> 1; (convert to word count)
; Do the bulk of the buffer, a word at a time
	jcxz	onobuf	; if(cx != 0){
	rep	movsw	; do { *di++ = *si++; (si is word pointer)
			; } while(--cx != 0); }
; now check for odd trailing byte
onobuf:	mov	cx,ocnt
	test	cx,1
	jz	ocnteven
	movsb		; *di++ = *si++;
ocnteven:
	pop	di
	pop	es
	popds
	pop	si
	popf
	pret
	pend	ew_outbuf

; ew_inbuf - get a buffer from an area of memory
	procdef ew_inbuf,<<ipara,word>,<ioffs,word>,<ibuf,ptr>,<icnt,word>>
	pushf
	push	di
	pushes
	push	si
	push	ds
	mov	dx,ioffs
	mov	si,dx
	mov	cx,icnt
	ldptr	di,ibuf,es	; es:di = ibuf (es already set in small model)
	mov	dx,ipara
	mov	ds,dx		; ds:si = shared mem pointer
	cld

; If buffer doesn't begin on a word boundary, get the first byte
	test	di,1	; if(buf & 1){
	jz	ibufeven ;
	movsb		; *di++ = *si++;
	dec	cx	; cx--;
	mov	icnt,cx	; icnt = cx; } save for later test
ibufeven:
	shr	cx,1	; cx = cnt >> 1; (convert to word count)
; Do the bulk of the buffer, a word at a time
	jcxz	inobuf	; if(cx != 0){
	rep	movsw	; *di++ = *si++; (di is word pointer)
			; } while(--cx != 0);
; now check for odd trailing byte
inobuf:	mov	cx,icnt
	test	cx,1
	jz	icnteven
	movsb		; *di++ = *si++;
icnteven:
	pop	ds
	pop	si
	popes
	pop	di
	popf
	pret
	pend	ew_inbuf


; ew_gbyte - get a byte from rx frame. 
; For status and next-frame (bytes 0 and 1) and size (bytes 3-4)
	procdef ew_gbyte,<<spara,word>,<soffs,word>>
	pushf
	push	di
	push	es
	mov	dx,soffs
	mov	di,dx
	mov	dx,spara
	mov	es,dx		; es:di = shared mem pointer
	mov	al,es:[di]
	xor	ah,ah
	pop	es
	pop	di
	popf
	pret
	pend	ew_gbyte

	end
