/*
   expire.c : Copyright Paul Healy, EI9GL, 1992.

   920913 : Inspired by expire in wg7j v1.03 (!?)
   930316 : To do -
             . add a high water mark over which the expiry starts
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "config.h"
#include "buf.h"
#include "mailfile.h"
#include "header.h"
#include "rc.h"
#include "lock.h"
#include "mailer.h"
#include "misc.h"
#include "help.h"
#include "expirlib.h"
#include "pc.h"
         
#ifdef BMH
#define main expire_main
#endif

static time_t
parsetime(char *line)
{
   char *p = strdup(line), *arg[MAXARGS];
   int nargs;
   time_t t;
   int year, month;

   if (p == NULL) {
      fprintf(stderr, "expire: no room\n");
      return -1L;
      }

   if ( (nargs = parse(p, arg, MAXARGS)) < 3) {
      fprintf(stderr, "expire: error parsing from line\n");
      free(p);
      return -1L;
      }

   t = atol(arg[nargs-1]);

   if (t > 3000L) {
      free(p);
      return t;                        /* nntp client format */
      }

   if (t < 100L)
      year = (int) (t + 1900L);
   else
      year = (int) t;

   if ( (month = asc2month(arg[nargs-4])) == 0) {
      fprintf(stderr, "expire: bad month symbol\n");
      free(p);
      return -1L;
      }

             /* Hours               Days  */
   t = maket(atoi(arg[nargs-2]), atoi(arg[nargs-3]), month, year);

   free(p);

   return t;
}

static int
doexpire(FILE *oldfp, FILE *newfp, int days)
{
   char line[LINELEN];
   time_t datum = time(NULL) - (days <= 0 ? DefaultAge : days) * 24 * 60 * 60;

   while (fgets(line, sizeof(line), oldfp) != NULL) {
      time_t msgt;

      if (!ismsg(line))
         continue;

      if ( (msgt = parsetime(line)) == -1L) {
         fprintf(stderr, "expire: warning can't parse\n   %s\n  - exiting\n", line);
         return -1;
         }
      else if (msgt > datum) {
         if (fputs(line, newfp) == EOF) {
            fprintf(stderr, "expire: error writing new message file\n");
            return -1;
            }
         return dupfile(oldfp, newfp);
         }

      }

   return 0;
}

int
expire(Mailfile *m, int argc, char *argv[])
{
   if (mail_getfile(m, argc, argv, "expire") == -1)
      return -1;

   return updatef("expire", getrc(maildir), mail_folder(m), EXT, doexpire, argc, argv);
}

int
main(int argc, char *argv[])
{
   Mailfile m;

   if (dohelp(argc, argv, "expire [+folder] [<days>] (expire mail)"))
      return 0;

   mail_zero(&m);

   expire(&m, argc, argv);

   return mail_savestate(&m);
}
