/*
   expirlib.c : Copyright Paul Healy, EI9GL, 1992.

   930110 : Created.
*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <io.h>
#include <time.h>
#include <dos.h>
#include "config.h"
#include "expirlib.h"
#include "lock.h"
#include "misc.h"

int
updatef(char *prog, char *dir, char *file, char *ext, UFN f, int argc, char *argv[])
{
   char cur[MAXNAMLEN];
   char new[MAXNAMLEN];
   char old[MAXNAMLEN];
   FILE *curfp, *newfp;
   int days;

   (void) ext;

   sprintf(cur, "%s/%s", dir, file);

   if (bm_lock(cur) == -1) {
      fprintf(stderr, "%s: can't lock %s\n", prog, cur);
      return -1;
      }

   if ( (curfp = fopen(cur, "r")) == NULL) {
      fprintf(stderr, "%s: can't open %s\n", prog, cur);
      bm_unlock(cur);
      return -1;
      }

   sprintf(new, "%s/%s.new", dir, file);

   if ( (newfp = fopen(new, "w")) == NULL) {
      fprintf(stderr, "%s: can't open %s\n", prog, new);
      fclose(curfp);
      bm_unlock(cur);
      return -1;
      }

   if ( (argc>1) && isdigit(argv[1][0]) )
      days = atoi(argv[1]);
   else
      days = DefaultAge;

   if (f(curfp, newfp, days) == -1) {
      fclose(curfp);
      fclose(newfp);
      bm_unlock(cur);
      return -1;
      }

   fclose(curfp);
   fclose(newfp);

   sprintf(old, "%s/.old", dir, file);

   if (access(old, 0) == 0)
      if (unlink(old) == -1) {
         fprintf(stderr, "%s: error removing %s\n", prog, old);
         bm_unlock(cur);
         return -1;
         }

   if (rename(cur, old) == -1) {
      perror(prog);
      bm_unlock(cur);
      return -1;
      }

   if (rename(new, cur) == -1) {
      perror(prog);
      bm_unlock(cur);
      return -1;
      }

   return bm_unlock(cur);
}

time_t
maket(int hour, int day, int month, int year)
{
   struct date d;
   struct time t;

   if (year < 2000)
      year += 1900;
   else if (year < 2100)
      year += 2000;

   if ( (year<1980) || (year>2099) ||
        (month<1)   || (month>12)  ||
        (day<1)     || (day>31)    ||
        (hour<0)    || (hour>23)  
      ) {
      fprintf(stderr, "expire: error making time %d:%d:%d:%d\n",
         year, month, day, hour);
      return (time_t) -1L;
      }

   d.da_year = year;
   d.da_day = (char) day;
   d.da_mon = (char) month;

   t.ti_min = 0;
   t.ti_hour = (unsigned char) hour;
   t.ti_hund = 0;
   t.ti_sec = 0;

   return (time_t) dostounix(&d, &t);
}
