/*
   expirnn.c : Copyright Paul Healy, EI9GL, 1992.

   Todo: added some respect for the Expires: header

   920913 : Created.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <limits.h>
#include "config.h"
#include "buf.h"
#include "mailfile.h"
#include "header.h"
#include "rc.h"
#include "misc.h"
#include "help.h"
#include "lock.h"
#include "mailer.h"
#include "expirlib.h"
#include "pc.h"
#include "nntp.h"

#ifdef BMH
#define main expirnn_main
#endif

static int Verbose = 0;

static int
asc2num(char *s)
{
   return ((s[0]-'0') * 10) + (s[1]-'0');
}

static time_t
nnt2unix(char *date, char *time)
{
   int hour = asc2num(time);
   int day = asc2num(&date[4]);
   int month = asc2num(&date[2]);
   int year = asc2num(date);

   return maket(hour, day, month, year);
}

/*
 * The article parameter is something like:
 *   control/3
 */
static int
expire(char *article)
{
   char *p = strchr(article, '/');
   char directory[MAXNAMLEN], filename[MAXNAMLEN], oldname[MAXNAMLEN];

   if (p == NULL)
      return -1;
   else
      *p = 0;

   if (getnnpath(article, directory, MAXNAMLEN) == NULL)
      return -1;

   sprintf(filename, "%s/%s", directory, p+1);
   sprintf(oldname, "%s/%s.old", directory, p+1);

   fprintf(stderr, "expirnn: expiring %s", filename);

   if (rename(filename, oldname) == -1) {
      fprintf(stderr, " - Failed!\n");
      return -1;
      }
   else {
      fprintf(stderr, "\n");
      return 0;
      }
}

/*
 * Return values are:
 *  0 - article was successfully expired
 *  1 - have reached a point where there should be no further expires
 * -1 - an error occurred.
 */
static int
checkarticle(char *line, time_t datum)
{
   char *arg[MAXARGS];
   int nargs;
   time_t t;

   if ( (nargs = parse(line, arg, MAXARGS)) < 4) {
      fprintf(stderr, "expirenn: error parsing history for %s\n", line);
      return -1;
      }
   /*
    * <11537@ei9gl.ampr.org> 930113 234727 control/3
    */
   if ((t = nnt2unix(arg[1], arg[2])) == -1L) {
      fprintf(stderr, "expirenn: bad date/time for %s %s %s\n",
         arg[0], arg[1], arg[2]);
      return -1;
      }

   if ( t < datum ) {
      /*
       * Want to expire each copy of the article
       */
      int i;

      for (i=3; i<nargs; i++)
         if (expire(arg[i]) != 0)
            fprintf(stderr, "expirnn: warning expire failed for %s %s %s %s\n",
               arg[0], arg[1], arg[2], arg[i]);

      return 0;
      }
   else 
      return 1;   /* do a file dup here */
}

static int
expirenn(FILE *oldfp, FILE *newfp, int days)
{
   char line[LINELEN], copy[LINELEN];
   time_t datum = time(NULL) - (days <= 0 ? DefaultAge : days) * 24 * 60 * 60;

   while (fgets(line, sizeof(line), oldfp) != NULL) {

      if (Verbose)
         fprintf(stderr, "expirnn: checking %s", line);

      switch (checkarticle(strcpy(copy, line), datum) ) {
         case 0:  /* article was expired */
            break;
         case 1:
            fputs(line, newfp);
            return dupfile(oldfp, newfp);
         case -1:
            /*
             * abort on error?
             * return'ing -1 from here will mean that the history file will
             * not be updated.
             */
            break;
         }
      }

   return 0;
}

static int
update_active(FILE *oldfp, FILE *newfp, int day)
{
   char line[LINELEN];
   char group[MAXNAMLEN];

   (void) day;

   while (fgets(line, sizeof(line), oldfp) != NULL) {
      long high, low;
      char posting, moderated = ' ';

      if (Verbose)
         fprintf(stderr, "expirnn: checking newsgroup %s", line);

      /*
       * newsgroup_name HHHHH LLLLL Y M
       */
      sscanf(line, "%s %ld %ld %c %c", group, &high, &low, &posting, &moderated);

      if (getnnrange(group, &high, &low) == -1) 
         fprintf(stderr, "expirnn: can't update active for %s\n", group);

      fprintf(newfp,"%s %5.5ld %5.5ld %c", group, high, low, posting);
      if (moderated != ' ')
         fprintf(newfp," %c\n", moderated);
      else 
         fprintf(newfp,"\n", moderated);
      }
   return 0;
}

int
main(int argc, char *argv[])
{
   if (dohelp(argc, argv, "expirnn [<days>] [-verbose] (expire news articles)"))
      return 0;

   Verbose = isflag("-verbose", argc, argv);

   if ( updatef("expirnn",getrc(newsdir),"history", "", expirenn, argc, argv)==0)
      return updatef("expirnn", getrc(newsdir), "active", "",
         update_active, argc, argv);
   else
      return -1;
}
