/*
   export.c : Copyright Paul Healy, EI9GL, 1992.

   921213 : Created.
            To do: split up the to part of the address.
                   Optionally add R: line
                   Add address override option
                   switch to add (some) rfc822 headers.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "rc.h"
#include "misc.h"
#include "help.h"
#include "pager.h"
#include "buf.h"
#include "mailfile.h"
#include "header.h"
#include "mailer.h"
#include "message.h"

#ifdef BMH
#define main export_main
#endif

/*
 * follows the convention used in wg7j v1.07 (mailbox.c:makecl)
 */
static char *
getbid(char *msgid, char *bid)
{
   char *at = strchr(msgid, '@');

   if (at != NULL) {
      if(stricmp(&msgid[strlen(msgid) - 5], ".bbs>") == 0)
         *at = '\0';       /* retain the bid given by user */
      else {
         char *dot = strchr(at, '.');

         *at = '_';
         if (dot != NULL)
            *dot = '\0';
         }  
      }
          
   bid[0] = '$';
   strncpy(bid+1, msgid+1, 13);
   bid[13] = '\0';            /* BIDs should be no longer than 13 bytes */

   return bid;
}

static char *
getfrom(Header_t *h, char *id)
{
   if (HEADERP(h, FROM) == NULL)
      return strupr(strcpy(id, getrc(username)));
   else {
      char *from = HEADERP(h, FROM);
      char *at = strchr(from, '@');
      char *percent = strchr(from, '%');

      if (at != NULL)
         *at = '\0';
      if (percent != NULL)
         *percent = '\0';
          
      strncpy(id, from, 7);
      id[6] = '\0';

      return strupr(id);
      }
}

static char *
getto(char *to, char *id)
{
   char *space = strchr(to, ' ');

   if (space != NULL)
      *space = '\0';
          
   return strupr(strcpy(id, to));
}

#define hdrchk(h, hdr, s, msg) \
   if ( HEADER(h, hdr) == NULL) { \
      fprintf(stderr, "export: missing %s header in message %d\n", s, msg); \
      return MSG_QUIT; \
      } 

static MsgResult
doexport(Mailfile *m, char *line)
{
   Header_t h;
   char toaddr[20], from[20], bid[50];    /* s[p|b] toaddr < from $bid */
   static char Hold[] = "HOLD";
   int msgnum = mail_curmsg(m);

   null_check(line);

   if (gethdrb(m, &h) == -1)
      return MSG_QUIT;

   hdrchk(h, TO, Hdr[TO], msgnum);
   hdrchk(h, FROM, Hdr[FROM], msgnum);
   hdrchk(h, MESSAGEID, Hdr[MESSAGEID], msgnum);

   printf("S%c %s < %s %s\n%s\n",
      HEADER(h, XBBSMSGTYPE) == NULL ? 'P' : toupper(HEADER(h, XBBSMSGTYPE)[0]),
      HEADER(h, TO) == NULL ? Hold : getto(HEADER(h, TO), toaddr),
      getfrom(&h, from),
      strupr( HEADER(h, MESSAGEID) == NULL ? "" : getbid(HEADER(h, MESSAGEID), bid) ),
      HEADER(h, SUBJECT) == NULL ? "" : HEADER(h, SUBJECT)
      );

   null_check("export: header done");

   while (mail_getstring(m, line, LINELEN) != NULL)
      if (ismsg(line)) {
         puts("/EX");
         fflush(stdout);
         return MSG_NEXT;               /* have hit next message */
         }
      else if (puts(line) == EOF) {
         fprintf(stderr, "export: can't write to standard output\n");
         return MSG_QUIT;
         }
         
   puts("/EX");
   fflush(stdout);

   null_check("end of export");
                     
   return MSG_EOF;
}

int
export(Mailfile *m, int argc, char *argv[])
{
   if (mail_getfile(m, argc, argv, "export") == -1)
      return -1;

   if (argc == 1)                       /* any arguments ? */
      walk2msg(m, doexport);
   else {
      int msg = mail_lastmsg(m, argc, argv);

      walkselect(m, doexport, argc, argv);
      mail_setcur(m, msg);
      }


   return 0;
}

int
main(int argc, char *argv[])
{
   Mailfile m;

   if (dohelp(argc, argv, "export [+folder] [msglist] (mail to ax25 pbbs converter)"))
      return 0;

   mail_zero(&m);

   export(&m, argc, argv);

   return mail_savestate(&m);
}
