/*
   extract.c: Copyright 1993, Paul Healy.

   930412 : Created
*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "buf.h"
#include "mailfile.h"
#include "message.h"
#include "header.h"
#include "extract.h"
#include "mailer.h"
#include "misc.h"

#ifdef BMH
#define main extract_main
#endif

static FILE *fp = NULL;

/*
 * save message including any headers.
 */
static MsgResult
dumpmsg(Mailfile *m, char *line)
{
   return mail_msg2file(m, fp, mail_getstring(m, line, LINELEN));
}

/*
 * save message, with no headers.
 */
static MsgResult
savemsg(Mailfile *m, char *line)
{
   Header_t h;

   if (gethdrb(m, &h) == -1) {
      fprintf(stderr, "extract: bad header on message %d\n", mail_curmsg(m));
      return MSG_QUIT;
      }
   else 
      return mail_msg2file(m, fp, mail_getstring(m, line, LINELEN));
}

int
msg2file(Mailfile *m, FILE *f, int msg, int wantheaders)
{
   fp = f;
   mail_setcur(m, msg);

   return walk2msg(m, wantheaders ? dumpmsg : savemsg);
}

int
append(char *prog, Mailfile *m, char *filename, int msg)
{
   int ret =0;
#ifdef MSDOS
   FILE *fp = fopen(filename, "ab");
#else
   FILE *fp = fopen(filename, "a");
#endif

   if (fp == NULL)
      ret = -1;
   else if(msg2file(m, fp, msg, 1) == -1)
      ret = -1;
   
   if (fp != NULL)
      fclose(fp);

   if (ret != 0)
      fprintf(stderr, "%s: unable to append message %d to %s\n",
         prog, mail_curmsg(m), filename);

   return ret;
}

int
xtract(Mailfile *m, char *filename, int argc, char *argv[], int wanthdr)
{
   if ( (fp = fopen(filename, "ab")) == NULL) {
      fprintf(stderr, "extract: can't open %s\n", filename);
      return -1;
      }

   if (argc == 1)            
      walk2msg(m, wanthdr ? dumpmsg : savemsg);
   else 
      walkselect(m, wanthdr ? dumpmsg : savemsg, argc, argv);

   return fclose(fp);
}

int
extract(Mailfile *m, int argc, char *argv[], int wanthdr)
{
   char *filename;

   if (mail_getfile(m, argc, argv, "extract") == -1)
      return -1;

   if ((argc > 1) && !isdigit(argv[argc-1][0]))
      filename = argv[--argc];
   else 
      filename = "mbox";

   if ( (fp = fopen(filename, "ab")) == NULL) {
      fprintf(stderr, "extract: can't open %s\n", filename);
      return -1;
      }

   if (argc == 1)            
      walk2msg(m, wanthdr ? dumpmsg : savemsg);
   else 
      walkselect(m, wanthdr ? dumpmsg : savemsg, argc, argv);

   printf("bmh: appended msg(s) to %s\n", filename);

   return fclose(fp);
}

#ifdef notdef

int
main(int argc, char *argv[])
{
   Mailfile m;

   if (dohelp(argc, argv, "extract [+folder] [-noheaders] [msglist] [file]"))
      return 0;

   mail_zero(&m);

   extract(&m, argc, argv);

   return mail_savestate(&m);
}

#endif
