/*
   folders.c : Copyright Paul Healy, EI9GL, 1992.

   Derived from bm.

   Copyright 1986 Bdale Garbee, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.
   Copyright 1987 1988 Dave Trulli NN2Z, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.

   911222 : Added this header.
   920601 : Added test program to display info on folders - define FOLDER_PROG
   920604 : folders.exe, release v0.1
   920610 : release v0.2, stars public areas
   920616 : release v0.3, simpler date, flag locked folders
   920708 : v0.5, cleanup for release
   920715 : Cleanup bmh v0.1 release, fixed month display
   920719 : Added recursive search of mail directory for news files
   920828 : To do - List folders by date
   930309 : Made area handling more robust (for news files mainly)
            Added support for the dayfirst rc flag
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <ctype.h>
#include <io.h>
#include <time.h>
#include "config.h"
#include "pc.h"
#include "rc.h"
#include "lock.h"
#include "misc.h"
#include "help.h"
#include "pager.h"
#include "getopts.h"

#ifdef BMH
#define main folders_main
#endif

#undef DEBUG

static const char PublicMark = '*';

static int Verbose = 0;
static int Zaplocks = 0;
static int News = 0;

static int FolderCount = 0;
static long ByteCount = 0L;

static int listfolders_long(char *maildir, char *prefix);

/* -------------------------------------------------------------------- */

#ifdef AX25BBS

static int
area_count(char *name)
{
   static int count = -1;   /* simple cache of area count */
   FILE *fp;

   if (count != -1)
      return count;
   else if ( (name == NULL) || ((fp = fopen(name, "r")) == NULL) )
      return count = 0;
   else {
      char line[LINELEN];

      count = 0;
      while (fgets(line, sizeof(line), fp) != NULL)
         if ( isspace(*line) || (*line=='\0') )
            continue;
         else if (isalnum(*line))
            count++;
      fclose(fp);

      return count;
      }
}

/*
static void
area_free(char *area[], int maxarea)
{
   for (int i=0; i<maxarea; i++)
      free(area[i]);
   free(area);
}
*/

static int
area_cmp(const void *p1, const void *p2)
{
   char **s1 = (char **) p1;
   char **s2 = (char **) p2;

   return strcmp(*s1, *s2);
}

static void
area_sort(char *area[], int maxarea)
{
#ifdef DEBUG
   int i;
#endif

   if (maxarea > 0)
      qsort(area, maxarea, sizeof(char *), area_cmp);
#ifdef DEBUG
   for (i=0; i<maxarea; i++)
      fprintf(stderr, "sortarea: %d = <%s>\n", i, area[i]);
#endif
}

static int
area_get(char *name, char *area[], int maxarea)
{
   FILE *fp;

   if ( (maxarea<1) || (name == NULL) || ((fp = fopen(name, "r")) == NULL) )
      return 0;
   else {
      char line[LINELEN];
      int i = 0;

      while (fgets(line, sizeof(line), fp) != NULL)
         if ( isspace(*line) || (*line=='\0') )
            continue;
         else if (isalnum(*line)) {
            char *cp = line;

            while (*cp)
               if (isspace(*cp))
                  *cp = '\0';
               else
                  cp++;
            if ((area[i] = strdup(line)) == NULL) {
               fprintf(stderr, "folders: no room\n");
               fclose(fp);
               return -1;
               }
#ifdef DEBUG
            else
               fprintf(stderr, "getarea: storing <%s>\n", area[i]);
#endif
            i++;
            }
      fclose(fp);
      area_sort(area, maxarea);
      return i;
      }
}

static int
area_ispublic(char *s, char *area[], int maxarea)
{
   int i;

   for (i=0; i<maxarea; i++)
      if (strcmp(s, area[i]) == 0)
         return 1;

   return 0;
/*
   fprintf(stderr, "ispublic: checking <%s> %d (%d, %d, %d)\n", s, res, maxarea,
      bsearch(s, area, maxarea, sizeof(char *), area_cmp),
      lfind(s, area, maxarea, sizeof(char *), area_cmp)
      );

   if (maxarea > 0)
      return bsearch(s, area, maxarea, sizeof(char *), area_cmp) != 0;
   else
      return 0;
*/
}

static char **
area_init(int *maxarea)
{
   char filename[MAXNAMLEN];
   static char **p = NULL;
   static int count = -1;

   *maxarea = count;

   if (p != NULL)
      return p;

   sprintf(filename, "%s/areas", getrc(spooldir));

   if ( (count = area_count(filename)) < 1)
      return NULL;
   else if ( (p = (char **) malloc(sizeof(char *)*count)) == NULL ) {
      fprintf(stderr, "folders: no room\n");
      exit(-1);
      }
   else if (area_get(filename, p, count) == -1)
      exit(-1);

   *maxarea = count;

   return p;
}
#endif /* AX25BBS */

/* -------------------------------------------------------------------- */

static void
listfolders_short(char *maildir)
{
   char pattern[MAXNAMLEN], *p;
   Directory d;

   sprintf(pattern, "%s/*%s", maildir, EXT);

   if (dir_init(&d, pattern) < 1)
      return ;

#ifdef DEBUG
   null_check("after dir init");
#endif

   while( ( p = dir_read(&d) ) != NULL ) {
      char *dot = strchr(p, '.');
      if (dot != NULL)
         *dot = '\0';
      printf("%10s", p);
      }
   dir_fin(&d);
#ifdef DEBUG
   null_check("listfolders_short");
#endif
}

/* -------------------------------------------------------------------- */

static char *
bmtime(const time_t *t, char *s)
{
   struct tm *tmp = localtime(t);
   char *p = getrc(dayfirst);

   if (*p == 'y')
      sprintf(s, "%2d/%02d/%2d", tmp->tm_mday, tmp->tm_mon+1, tmp->tm_year);
   else 
      sprintf(s, "%2d/%02d/%2d", tmp->tm_mon+1, tmp->tm_mday, tmp->tm_year);

   return s;
}

/* -------------------------------------------------------------------- */
#ifdef NEWS
static int
listnews(char *maildir, char *prefix, char *buf)
{
   Directory d;
   int ret = 0;
   char *p;
#ifdef DEBUG
   fprintf(stderr, "listnews: <%s> <%s> %u\n", maildir, prefix, coreleft());
#endif
   if (News)
      News = 0;
   else if ( listfolders_long(maildir, prefix) == -1)
      return -1;

   sprintf(buf, "%s/*.  ", maildir);
   if (dir_init(&d, buf) < 1)
      return 0;

   while( ( p = dir_read(&d) ) != NULL ) {
      struct stat sbuf;

      sprintf(buf, "%s/%s", maildir, p);
      if ( stat(buf, &sbuf) == -1) {
         fprintf(stderr, "\nfolders: warning can't stat %s\n", buf);
         ret = -1;
         break;
         }

      if (!(sbuf.st_mode & S_IFDIR) || (*p == '.'))
         continue;
      else {
         int dirlen = strlen(maildir);
         int prelen = strlen(prefix);

         sprintf(&maildir[dirlen], "/%s", p);
         sprintf(&prefix[prelen], "%s.", p);
         
         if ( listnews(maildir, prefix, buf) == -1) {
            ret = -1;
            break;
            }
         maildir[dirlen] = 0;
         prefix[prelen] = 0;
         }
      }

   dir_fin(&d);
   return ret;
}

static int
listall(void)
{
   char dir[MAXNAMLEN], prefix[MAXNAMLEN], buf[MAXNAMLEN];
   int ret = 0;

   prefix[0] = 0;
   FolderCount = 0;
   ByteCount = 0L;

   ret = listnews(strcpy(dir, getrc(maildir)), prefix, buf);

   if (ret != -1)
      return page_printf("\rTotal of %d folders, %ld bytes used",
         FolderCount, ByteCount);
   else 
      return ret;
}
#endif /* NEWS */

/* -------------------------------------------------------------------- */

static char **Area = NULL;
static int NArea = -1;

static int
mail_info(char *dir, char *name, char *prefix, long *size, long *time, int *isl, int *ispub)
{
   char filename[MAXNAMLEN];
   struct stat sbuf;

   sprintf(filename, "%s/%s", dir, name);

   if (stat(filename, &sbuf) == -1) {
      fprintf(stderr, "\nfolders: warning can't stat %s\n", filename);
      return -1;
      }
   else {
      char *dot = strstr(filename, EXT);

      *size = sbuf.st_size;
      *time = sbuf.st_atime;

      if (dot != NULL)
         *dot = '\0';
      *isl = islocked(filename);

      sprintf(filename, "%s%s", prefix, name);
#ifdef AX25BBS
      *ispub = area_ispublic(filename, Area, NArea);
#else
      *ispub = 0;
#endif
      return 0;
      }
}

static char *
zaplock(int islocked, char *dir, char *name)
{
   if (islocked && Zaplocks) {
      char file[MAXNAMLEN];
      char *dot;

      sprintf(file, "%s/%s", dir, name);
      dot = strstr(file, EXT);
      if (dot != NULL)
         *dot = '\0';
      if (bm_unlock(file) == 0)
         return "[Lock removed]";
      else
         return "[ERROR removing lock]";
      }
   else
      return "";
}

static int
listfolders_long(char *maildir, char *prefix)
{
   char line[MAXNAMLEN], *p;
   Directory d;
   int ret = 0;

#ifdef DEBUG
   fprintf(stderr, "listfolders: <%s> <%s> %u\n", maildir, prefix, coreleft());
#endif

   sprintf(line, "%s/*%s", maildir, EXT);

   if (dir_init(&d, line) < 1)
      return 0;

#ifdef DEBUG
   fprintf(stderr, "listfolders: dir init ok %u\n", coreleft());
#endif

   while( ( p = dir_read(&d) ) != NULL ) {
      char date[10], *dot;
      long size, time;
      int islocked, ispub;

      if (mail_info(maildir, p, prefix, &size, &time, &islocked, &ispub) == -1) {
         ret = -1;
         break;
         }
#ifdef DEBUG
      fprintf(stderr, "listfolders: mail info ok %s\n", p);
#endif
      dot = strstr(p, EXT);
      if (dot != NULL)
         *dot = '\0';
      if ( page_printf(" %c %s%-*s%8ld   %s  %s %s",
               ispub ? PublicMark : ' ',
               prefix,
               20 - strlen(prefix), p,
               size,
               bmtime(&time, date),
               islocked ? "[locked]" : "",
               zaplock(islocked, maildir, p)) == -1
         ) {
         ret = -1;
         break;
         }
      FolderCount++;
      ByteCount+=size;
      }
   dir_fin(&d);
   return ret;
}

static const char *optv[] = {
   "?|help",
   "s|short",
   "m|mail",
   "n|news",
   "z|zaplocks",
   "v|verbose",
   NULL
};

int
main(int argc, char *argv[])
{
   int optchar;
   char *optarg;
   options_t *opts;
   int errors = 0;

   int Short = 0;
   int Mail = 0;

#ifdef AX25BBS
   static unsigned runcount = 0;
#endif

   if (dohelp(argc, argv, "folders [-short|-mail|-news|-zaplocks] (list folders)"))
      return 0;

#ifdef AX25BBS
   if ( (runcount == 0) && ((Area = area_init(&NArea)) == NULL))
      printf("folders: no areas file found\n");

   runcount++;
#endif

   opts = optinit(*argv, optv);
   optctrl(opts, opt_LONG_ONLY);

   --argc, ++argv;
   while( (optchar = getopts(opts, argc, argv, &optarg)) != 0 )
      switch(optchar) {
         case 's' : Short = 1; break;
         case 'm' : Mail = 1; break;
         case 'n' : News = 1; break;
         case 'z' : Zaplocks = 1; break;
         case 'v' : Verbose = 1; break;
         default : ++errors; break;
         }

   if (errors) {
      optusage(opts, "", stderr);
      optfree(opts), opts = NULL;
      return -1;
      }
   else 
      optfree(opts), opts = NULL;

   page_setup("folders");

   if (Short)
      listfolders_short(getrc(maildir));
   else if (Mail)
      listfolders_long(getrc(maildir), "");
#ifdef NEWS
   else 
      listall();
#endif

   page_fin("folders");

   return 0;
}
