/* Stuff common to both the FTP server and client */
#include <stdio.h>
#include "machdep.h"
#include "mbuf.h"
#include "netuser.h"
#include "timer.h"
#include "tcp.h"
#include "ftp.h"
#include "session.h"

/* FTP data channel receive upcall handler */
void
r_ftpd(tcb,cnt)
struct tcb *tcb;
int16 cnt;
{
        register struct ftp *ftp;
        struct mbuf *bp;
#ifdef        UNIX
        char c;
#endif
#ifdef        OSK
        char c;
#endif


        ftp = (struct ftp *)tcb->user;
        if(ftp->state != RECEIVING_STATE){
                close_tcp(tcb);
                return;
        }
        /* This will likely also generate an ACK with window rotation */
        recv_tcp(tcb,&bp,cnt);

#ifdef        UNIX
        while(pullup(&bp,&c,1) == 1){
                if(ftp->type == IMAGE_TYPE || c != '\r')
                        putc(c,ftp->fp);
        }
#else
#ifndef OSK
        while(bp != NULLBUF){
                if(bp->cnt != 0)
                        fwrite(bp->data,1,(unsigned)bp->cnt,ftp->fp);
                bp = free_mbuf(bp);
        }
#else
        while(pullup(&bp,&c,1) == 1){
                if(ftp->type == IMAGE_TYPE || c != '\l')
                        putc(c,ftp->fp);
        }
#endif
#endif
}
/* FTP data channel transmit upcall handler */
void
t_ftpd(tcb,cnt)
struct tcb *tcb;
int16 cnt;
{
        struct ftp *ftp;
        struct mbuf *bp;
        char *cp;
        int c;

        ftp = (struct ftp *)tcb->user;
        if(ftp->state != SENDING_STATE){
                close_tcp(tcb);
                return;
        }
        if((bp = alloc_mbuf(cnt)) == NULLBUF){
                /* Hard to know what to do here */
                return;
        }
        cp = bp->data;
        while(cnt != 0 && (c = getc(ftp->fp)) != EOF){
#ifdef        CPM
                if(ftp->type == ASCII_TYPE && c == CTLZ)
                        break;        /* CTLZ is CP/M's text EOF marker */
#endif
#ifdef        MSDOS
                /* CTLZ is sometimes used by MS-DOS editors too */
                if(ftp->type == ASCII_TYPE && c == CTLZ)
                        break;        
#endif
#ifdef        UNIX
                if(ftp->type == ASCII_TYPE && c == '\n'){
                        *cp++ = '\r';
                        bp->cnt++;
                        cnt--;
                }
#endif
#ifdef        OSK
                if(ftp->type == ASCII_TYPE && c == '\r'){
                        *cp++ = '\r';
                        bp->cnt++;
                        c = '\l';
                        cnt--;
                }
#endif
                *cp++ = c;
                bp->cnt++;
                cnt--;
        }
        if(bp->cnt != 0)
                send_tcp(tcb,bp);
        else
                free_p(bp);

        if(cnt != 0){        /* EOF seen */
                fclose(ftp->fp);
                ftp->fp = NULLFILE;
                close_tcp(tcb);
        }
}
/* Allocate an FTP control block */
struct ftp *
ftp_create(bufsize)
unsigned bufsize;
{
        void ftp_delete();
        char *calloc(),*malloc();
        register struct ftp *ftp;

        if((ftp = (struct ftp *)calloc(1,sizeof (struct ftp))) == NULLFTP)
                return NULLFTP;
        if(bufsize != 0 && (ftp->buf = malloc(bufsize)) == NULLCHAR){
                ftp_delete(ftp);
                return NULLFTP;
        }
        ftp->state = COMMAND_STATE;
        ftp->type = ASCII_TYPE;        /* Default transfer type */
        return ftp;
}
/* Free resources, delete control block */
void
ftp_delete(ftp)
register struct ftp *ftp;
{
        if(ftp->fp != NULLFILE)
                fclose(ftp->fp);
        if(ftp->data != NULLTCB)
                del_tcp(ftp->data);
        if(ftp->username != NULLCHAR)
                free(ftp->username);
        if(ftp->path != NULLCHAR)
                free(ftp->path);
        if(ftp->buf != NULLCHAR)
                free(ftp->buf);
        if(ftp->cd != NULLCHAR)
                free(ftp->cd);
        if(ftp->session != NULLSESSION)
                freesession(ftp->session);
        free((char *)ftp);
}


