/* FTP Server state machine - see RFC 959 */

#define        LINELEN                128        /* Length of command buffer */

#include <stdio.h>

#ifdef DATE
#include <time.h>
#endif

#ifdef OSK
#include <modes.h>
#endif

#include "machdep.h"
#include "mbuf.h"
#include "netuser.h"
#include "timer.h"
#include "tcp.h"
#include "ftp.h"

/* Command table */
static char *commands[] = {
        "user",
#define        USER_CMD        0
        "acct",
#define        ACCT_CMD        1
        "pass",
#define        PASS_CMD        2
        "type",
#define        TYPE_CMD        3
        "list",
#define        LIST_CMD        4
        "cwd",
#define        CWD_CMD                5
        "dele",
#define        DELE_CMD        6
        "name",
#define        NAME_CMD        7
        "quit",
#define        QUIT_CMD        8
        "retr",
#define        RETR_CMD        9
        "stor",
#define        STOR_CMD        10
        "port",
#define        PORT_CMD        11
        "nlst",
#define        NLST_CMD        12
        "pwd",
#define        PWD_CMD                13
        "xpwd",                        /* For compatibility with 4.2BSD */
#define        XPWD_CMD        14
        NULLCHAR
};

/* Response messages */
static char banner[] = "220 %s FTP Ready\r\l";
static char badcmd[] = "500 Unknown command\r\l";
static char givepass[] = "331 Enter password\r\l";
static char logged[] = "230 Logged in\r\l";
static char typeok[] = "200 Type OK\r\l";
static char deleok[] = "250 File deleted\r\l";
static char delefail[] = "550 Delete failed\r\l";
static char pwdmsg[] = "257 \"%s\" is current directory\r\l";
static char badtype[] = "501 Unknown type \"%s\"\r\l";
static char badport[] = "501 Bad port syntax\r\l";
static char unimp[] = "502 Command not yet implemented\r\l";
static char bye[] = "221 Goodbye!\r\l";
static char nodir[] = "553 Can't read directory \"%s\"\r\l";
static char cantopen[] = "550 Can't read file \"%s\"\r\l";
static char sending[] = "150 Opening data connection for %s %s\r\l";
static char cantmake[] = "553 Can't create file \"%s\"\r\l";
static char portok[] = "200 Port command okay\r\l";
static char rxok[] = "226 File received OK\r\l";
static char txok[] = "226 File sent OK\r\l";
static char noperm[] = "550 Permission denied\r\l";
static char noconn[] = "425 Data connection refused\r\l";

static struct tcb *ftp_tcb;

/* Start up FTP service */
ftp_start(argc,argv)
int argc;
char *argv[];
{
        struct socket lsocket;
        void r_ftp(),s_ftp();

        lsocket.address = ip_addr;
        if(argc < 2)
                lsocket.port = FTP_PORT;
        else
                lsocket.port = atoi(argv[1]);

        ftp_tcb = open_tcp(&lsocket,NULLSOCK,TCP_SERVER,0,r_ftp,NULLVFP,s_ftp,0,(int *)NULL);
}
ftp_stop()
{
        if(ftp_tcb != NULLTCB)
                close_tcp(ftp_tcb);
}
/* FTP server control channel connection state change upcall handler */
static
void
s_ftp(tcb,old,new)
struct tcb *tcb;
char old,new;
{
        extern char hostname[];
        struct ftp *ftp,*ftp_create();
        void ftp_delete();
        char *inet_ntoa();
#ifdef DATE
        long t;
        char msg2[100];
#endif

        switch(new){
#ifdef        QUICKSTART
        case SYN_RECEIVED:
#else
        case ESTABLISHED:
#endif
                if((ftp = ftp_create(LINELEN)) == NULLFTP){
                        /* No space, kill connection */
                        close_tcp(tcb);
                        return;
                }
                ftp->control = tcb;        /* Downward link */
                tcb->user = (int *)ftp;        /* Upward link */

                /* Set default data port */
                ftp->port.address = tcb->conn.remote.address;
                ftp->port.port = FTPD_PORT;

                /* Note current directory */
                log(tcb,"open FTP");
#ifndef DATE
                tprintf(ftp->control,banner,hostname);
#else
                time(&t);
                sprintf(msg2,"%s (%s",hostname,ctime(&t));
                msg2[strlen(msg2)-1] = ')';
                tprintf(ftp->control,banner,msg2);
#endif
                break;                
        case CLOSE_WAIT:
                close_tcp(tcb);
                break;
        case CLOSED:
                log(tcb,"close FTP");
                if((ftp = (struct ftp *)tcb->user) != NULLFTP)
                        ftp_delete(ftp);
                /* Check if server is being shut down */
                if(tcb == ftp_tcb)
                        ftp_tcb = NULLTCB;
                del_tcp(tcb);
                break;
        }
}

/* FTP control channel receiver upcall handler */
static
void
r_ftp(tcb,cnt)
struct tcb *tcb;
int16 cnt;
{
        register struct ftp *ftp;
        char *index(),c;
        struct mbuf *bp;
        void docommand();

        if((ftp = (struct ftp *)tcb->user) == NULLFTP){
                /* Unknown connection, just kill it */
                close_tcp(tcb);
                return;
        }
        switch(ftp->state){
        case COMMAND_STATE:
                /* Assemble an input line in the session buffer. Return if incomplete */
                recv_tcp(tcb,&bp,0);
                while(pullup(&bp,&c,1) == 1){
                        switch(c){
                        case '\r':        /* Strip cr's */
                                continue;
#ifdef OSK
                        case '\l':        /* Complete line; process it */
#else
                        case '\n':        /* Complete line; process it */
#endif
                                ftp->buf[ftp->cnt] = '\0';
                                docommand(ftp);
                                ftp->cnt = 0;
                                break;
                        default:        /* Assemble line */
                                if(ftp->cnt != LINELEN-1)
                                        ftp->buf[ftp->cnt++] = c;
                                break;
                        }
                }
                /* else no linefeed present yet to terminate command */
                break;
        case SENDING_STATE:
        case RECEIVING_STATE:
                /* Leave commands pending on receive queue until
                 * present command is done
                 */
                break;
        }
}

/* FTP server data channel connection state change upcall handler */
void
s_ftpd(tcb,old,new)
struct tcb *tcb;
char old,new;
{
        register struct ftp *ftp;

        if((ftp = (struct ftp *)tcb->user) == NULLFTP){
                /* Unknown connection. Kill it */
                del_tcp(tcb);
        } else if((old == FINWAIT1 || old == CLOSING) && ftp->state == SENDING_STATE){
                /* We've received an ack of our FIN while sending; we're done */
                ftp->state = COMMAND_STATE;
                tprintf(ftp->control,txok);
                /* Kick command parser if something is waiting */
                if(ftp->control->rcvcnt != 0)
                        r_ftp(ftp->control,ftp->control->rcvcnt);
        } else if(ftp->state == RECEIVING_STATE && new == CLOSE_WAIT){
                /* FIN received on incoming file */
#ifdef        CPM
                if(ftp->type == ASCII_TYPE)
                        putc(CTLZ,ftp->fp);
#endif
                close_tcp(tcb);
                fclose(ftp->fp);
                ftp->fp = NULLFILE;
                ftp->state = COMMAND_STATE;
                tprintf(ftp->control,rxok);
                /* Kick command parser if something is waiting */
                if(ftp->control->rcvcnt != 0)
                        r_ftp(ftp->control,ftp->control->rcvcnt);
        } else if(new == CLOSED){
                if(old == SYN_SENT){
                        /* Data connection was refused, complain about it */
                        tprintf(ftp->control,noconn);
                        /* And clean up */
                        if(ftp->fp != NULLFILE){
                                fclose(ftp->fp);
                                ftp->fp = NULLFILE;
                        }
                        ftp->state = COMMAND_STATE;
                        /* Kick command parser if something is waiting */
                        if(ftp->control->rcvcnt != 0)
                                r_ftp(ftp->control,ftp->control->rcvcnt);
                }
                /* Clear only if another transfer hasn't already started */
                if(ftp->data == tcb)
                        ftp->data = NULLTCB;
                del_tcp(tcb);
        }
}

/* Parse and execute ftp commands */
static
void
docommand(ftp)
register struct ftp *ftp;
{
        void r_ftpd(),t_ftpd(),s_ftpd();
        char *cmd,*arg,*cp,**cmdp,*file;
        char *index(),*malloc(),*strcpy(),*pathname();
        struct socket dport;
        
#ifndef        CPM
        FILE *dir();
#endif

        cmd = ftp->buf;
        if(ftp->cnt == 0){
                /* Can't be a legal FTP command */
                tprintf(ftp->control,badcmd);
                return;
        }        
        cmd = ftp->buf;

        /* Translate entire buffer to lower case */
        for(cp = cmd;*cp != '\0';cp++)
                *cp = tolower(*cp);

        /* Find command in table; if not present, return syntax error */
        for(cmdp = commands;*cmdp != NULLCHAR;cmdp++)
                if(strncmp(*cmdp,cmd,strlen(*cmdp)) == 0)
                        break;
        if(*cmdp == NULLCHAR){
                tprintf(ftp->control,badcmd);
                return;
        }
        arg = &cmd[strlen(*cmdp)];
        while(*arg == ' ')
                arg++;

        /* Execute specific command */
        switch(cmdp-commands){
        case USER_CMD:
                if((ftp->username = malloc((unsigned)strlen(arg)+1)) == NULLCHAR){
                        close_tcp(ftp->control);
                        break;
                }
                strcpy(ftp->username,arg);
                tprintf(ftp->control,givepass);
                break;
        case TYPE_CMD:
                switch(arg[0]){
                case 'a':        /* Ascii */
                        ftp->type = ASCII_TYPE;
                        tprintf(ftp->control,typeok);
                        break;
                case 'b':        /* Binary */
                case 'i':        /* Image */
                        ftp->type = IMAGE_TYPE;
                        tprintf(ftp->control,typeok);
                        break;
                default:        /* Invalid */
                        tprintf(ftp->control,badtype,arg);
                        break;
                }
                break;
        case QUIT_CMD:
                tprintf(ftp->control,bye);
                close_tcp(ftp->control);
                break;
        case RETR_CMD:
                /* Disk operation; return ACK now */
                tcp_output(ftp->control);
                file = pathname(ftp->cd,arg);
                if(!permcheck(ftp,RETR_CMD,file)){
                         tprintf(ftp->control,noperm);
                } else if((ftp->fp = fopen(file,"r")) == NULLFILE){
                        tprintf(ftp->control,cantopen,file);
                } else {
                        log(ftp->control,"RETR %s/%s",ftp->cd,arg);
                        dport.address = ip_addr;
                        dport.port = FTPD_PORT;
                        ftp->state = SENDING_STATE;
                        tprintf(ftp->control,sending,"RETR",arg);
                        ftp->data = open_tcp(&dport,&ftp->port,TCP_ACTIVE,
                         0,NULLVFP,t_ftpd,s_ftpd,ftp->control->tos,(int *)ftp);
                }
                free(file);
                break;
        case STOR_CMD:
                /* Disk operation; return ACK now */
                tcp_output(ftp->control);
                file = pathname(ftp->cd,arg);
                if(!permcheck(ftp,STOR_CMD,file)){
                         tprintf(ftp->control,noperm);
                        free(file);
                         break;
                } else if((ftp->fp = fopen(file,"w")) == NULLFILE){
                        tprintf(ftp->control,cantmake,file);
                } else {
                        log(ftp->control,"STOR %s/%s",ftp->cd,arg);
                        dport.address = ip_addr;
                        dport.port = FTPD_PORT;
                        ftp->state = RECEIVING_STATE;
                        tprintf(ftp->control,sending,"STOR",arg);
                        ftp->data = open_tcp(&dport,&ftp->port,TCP_ACTIVE,
                         0,r_ftpd,NULLVFP,s_ftpd,ftp->control->tos,(int *)ftp);
                }
                free(file);
                break;
        case PORT_CMD:
                if(pport(&ftp->port,arg) == -1){
                        tprintf(ftp->control,badport);
                } else {
                        tprintf(ftp->control,portok);
                }
                break;
#ifndef CPM
        case LIST_CMD:
                /* Disk operation; return ACK now */
                tcp_output(ftp->control);

                file = pathname(ftp->cd,arg);
                if(!permcheck(ftp,RETR_CMD,file)){
                         tprintf(ftp->control,noperm);
                } else if((ftp->fp = dir(file,1)) == NULLFILE){
                        tprintf(ftp->control,nodir,file);
                } else {
                        dport.address = ip_addr;
                        dport.port = FTPD_PORT;
                        ftp->state = SENDING_STATE;
                        tprintf(ftp->control,sending,"LIST",file);
                        ftp->data = open_tcp(&dport,&ftp->port,TCP_ACTIVE,
                         0,NULLVFP,t_ftpd,s_ftpd,ftp->control->tos,(int *)ftp);
                }
                free(file);
                break;
        case NLST_CMD:
                /* Disk operation; return ACK now */
                tcp_output(ftp->control);

                file = pathname(ftp->cd,arg);
                if(!permcheck(ftp,RETR_CMD,file)){
                         tprintf(ftp->control,noperm);
                } else if((ftp->fp = dir(file,0)) == NULLFILE){
                        tprintf(ftp->control,nodir,file);
                } else {
                        dport.address = ip_addr;
                        dport.port = FTPD_PORT;
                        ftp->state = SENDING_STATE;
                        tprintf(ftp->control,sending,"NLST",file);
                        ftp->data = open_tcp(&dport,&ftp->port,TCP_ACTIVE,
                         0,NULLVFP,t_ftpd,s_ftpd,ftp->control->tos,(int *)ftp);
                }
                free(file);
                break;
        case CWD_CMD:
                tcp_output(ftp->control);        /* Disk operation; return ACK now */

                file = pathname(ftp->cd,arg);
                if(!permcheck(ftp,RETR_CMD,file)){
                         tprintf(ftp->control,noperm);
                        free(file);
#ifdef        MSDOS
                /* Don'tcha just LOVE %%$#@!! MS-DOS? */
                } else if(strcmp(file,"\\") == 0 || access(file) == 0){
#else
#ifdef OSK
                } else if(access(file,S_IREAD+S_IFDIR) == 0){  /* See if it exists */
#else
                } else if(access(file) == 0){        /* See if it exists */
#endif
#endif
                        /* Succeeded, record in control block */
                        free(ftp->cd);
                        ftp->cd = file;
                        tprintf(ftp->control,pwdmsg,file);
                } else {
                        /* Failed, don't change anything */
                        tprintf(ftp->control,nodir,file);
                        free(file);
                }
                break;
        case XPWD_CMD:
        case PWD_CMD:
                tprintf(ftp->control,pwdmsg,ftp->cd);
                break;
#else
        case LIST_CMD:
        case NLST_CMD:
        case CWD_CMD:
        case XPWD_CMD:
        case PWD_CMD:
#endif
        case ACCT_CMD:                
                tprintf(ftp->control,unimp);
                break;
        case DELE_CMD:
                file = pathname(ftp->cd,arg);
                if(!permcheck(ftp,DELE_CMD,file)){
                         tprintf(ftp->control,noperm);
                } else if(unlink(file) == 0){
                        tprintf(ftp->control,deleok);
                } else {
                        tprintf(ftp->control,delefail);
                }
                free(file);
                break;
        case PASS_CMD:
                tcp_output(ftp->control);        /* Send the ack now */
                ftplogin(ftp,arg);                        
                break;
        }
}
static
int
pport(sock,arg)
struct socket *sock;
char *arg;
{
        int32 n;
        int atoi(),i;

        n = 0;
        for(i=0;i<4;i++){
                n = atoi(arg) + (n << 8);
                if((arg = index(arg,',')) == NULLCHAR)
                        return -1;
                arg++;
        }
        sock->address = n;
        n = atoi(arg);
        if((arg = index(arg,',')) == NULLCHAR)
                return -1;
        arg++;
        n = atoi(arg) + (n << 8);
        sock->port = n;
        return 0;
}
/* Attempt to log in the user whose name is in ftp->username and password
 * in pass
 */
static
ftplogin(ftp,pass)
struct ftp *ftp;
char *pass;
{
        char buf[80],*cp,*cp1,*malloc();
        FILE *fp;

        if((fp = fopen(USERFILE,"r")) == NULLFILE){
                /* Userfile doesn't exist */
                tprintf(ftp->control,noperm);
                return;
        }
        while(fgets(buf,sizeof(buf),fp),!feof(fp)){
                if(buf[0] == '#')
                        continue;        /* Comment */
                if((cp = index(buf,' ')) == NULLCHAR)
                        /* Bogus entry */
                        continue;
                *cp++ = '\0';                /* Now points to password */
                if(strcmp(ftp->username,buf) == 0)
                        break;                /* Found user name */
        }
        if(feof(fp)){
                /* User name not found in file */
                fclose(fp);
                tprintf(ftp->control,noperm);
                return;
        }
        fclose(fp);
        /* Look for space after password field in file */
        if((cp1 = index(cp,' ')) == NULLCHAR){
                /* Invalid file entry */
                tprintf(ftp->control,noperm);
                return;
        }
        *cp1++ = '\0';        /* Now points to path field */
        if(strcmp(cp,"*") != 0 && strcmp(cp,pass) != 0){
                /* Password required, but wrong one given */
                tprintf(ftp->control,noperm);
                return;
        }
        if((cp = index(cp1,' ')) == NULLCHAR){
                /* Permission field missing */
                tprintf(ftp->control,noperm);
                return;
        }
        *cp++ = '\0';        /* now points to permission field */

        /* Set up current directory and path prefix */
        ftp->cd = malloc(strlen(cp1)+1);
        ftp->path = malloc(strlen(cp1)+1);
        strcpy(ftp->cd,cp1);
        strcpy(ftp->path,cp1);
        
        /* And finally set the permission bits */
        ftp->perms = atoi(cp);

        tprintf(ftp->control,logged);
        log(ftp->control,"%s logged in",ftp->username);
}                

#ifdef        MSDOS
/* Illegal characters in a DOS filename */
char badchars[] = "\"[]:|<>+=;,";
#endif

/* Return 1 if the file operation is allowed, 0 otherwise */
permcheck(ftp,op,file)
struct ftp *ftp;
int op;
char *file;
{
#ifdef        MSDOS
        char *cp,*index();
        
        /* Check for characters illegal in MS-DOS file names */
        for(cp = badchars;*cp != '\0';cp++){
                if(index(file,*cp) != NULLCHAR)
                        return 0;        
        }
#endif
        /* The target file must be under the user's allowed search path */
        if(strncmp(file,ftp->path,strlen(ftp->path)) != 0)
                return 0;

        switch(op){
        case RETR_CMD:
                /* User must have permission to read files */
                if(ftp->perms & FTP_READ)
                        return 1;
                return 0;
        case DELE_CMD:
                /* User must have permission to (over)write files */
                if(ftp->perms & FTP_WRITE)
                        return 1;
                return 0;
        case STOR_CMD:
                /* User must have permission to (over)write files, or permission
                 * to create them if the file doesn't already exist
                 */
                if(ftp->perms & FTP_WRITE)
                        return 1;
#ifdef OSK
                if(access(file,0) == -1 && (ftp->perms & FTP_CREATE))
#else
                if(access(file) == -1 && (ftp->perms & FTP_CREATE))
#endif
                        return 1;
                return 0;
        }
        return 0;        /* "can't happen" -- keep lint happy */
}

