/*
   header.c: Copyright Paul Healy, EI9GL, 1992.

   Derived from bm.

   Copyright 1986 Bdale Garbee, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.
   Copyright 1987 1988 Dave Trulli NN2Z, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.

   920712: Added this header.
   920717: Todo - add line continuation support
   920823: Cleaned up considerably.
   921227: Todo - add a maximum limit to the size of individual headers
*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <string.h>
#include "config.h"
#include "buf.h"
#include "rc.h"
#include "mailfile.h"
#include "send.h"
#include "header.h"
#include "misc.h"

#undef DEBUG
#undef STRICT

#define Wanted(hdr) ( 1 /* (hdr!=STATUS) && (hdr!=RECEIVED) */)
#define MaxHdrLength 16

char *Hdr[] = {
   "Cc: ",        
   "Date: ",      
   "From: ",      
   "Message-ID: ",
   "Newsgroups: ",
   "Received: ",  
   "Reply-To: ",  
   "Sender: ",    
   "Status: ",    
   "Subject: ",   
   "To: "
#ifdef AX25BBS       /* Note the extra comma for bbs-msg-type */
   , "X-BBS-Msg-Type: "
#endif
};

static struct header {
   char str[MaxHdrLength];
   Header hdr;
} hd[] = {
   /*
    * Currently only interested in parsing these headers. Note alphabetical
    * order is required for bsearch'ing.
    */
   "cc",             CC,
   "date",           DATE,
   "from",           FROM,
   "message-id",     MESSAGEID,
   "newsgroups",     NEWSGROUPS,
   "received",       RECEIVED,
   "reply-to",       REPLYTO,
   "sender",         SENDER,
   "status",         STATUS,
   "subject",        SUBJECT,
   "to",             TO
#ifdef AX25BBS                       /* Note the extra comma for bbs-msg-type */
   , "x-bbs-msg-type", XBBSMSGTYPE
#endif
};

static const int NumHeaders = sizeof(hd) / sizeof(struct header);

#ifdef DEBUG
static void
dump(Header_t *h)
{
   int i;

   fflush(stdout);
   fputc('\n', stderr);

   for (i=0; i<MAXHDR; i++)
      if (h->header[i] != NULL)
         fprintf(stderr, "%s: %s\n", hd[i].str, h->header[i]);

   putuhdr(stderr, h);
   fprintf(stderr, "dump: space used=%d, remaining=%d\n",
      h->hwm-h->buf, HDRBUFLEN - (h->hwm-h->buf));
}
#endif

void
hdr_init(Header_t *h)
{
   int i;

   for (i=0; i<MAXHDR; i++)
      h->header[i] = NULL;

   h->hwm = h->buf;
   h->lasth = NOHEADER;
}

static int
copyin(Header_t *h, char *s)
{
   char *cp = h->hwm, *end = &h->buf[HDRBUFLEN]-1;

   while (isspace(*s))
      s++;

   while (*s && (*s!='\n'))
      if (cp < end)
         *cp++ = *s++;
      else {
         fflush(stdout);
         fprintf(stderr, "parseheader: no buffer space\n");
         *(cp-1) = 0;
         return -1;
         }

   *cp = 0;
   h->hwm = cp + 1;

   return 0;
}

int
hdr_add(Header_t *h, Header hdr, char *s)
{
/*
   printf("hdradd: %s: %s\n", hd[hdr].str, s);
*/
   h->lasth = hdr;

/*
   if ( !Wanted(hdr) )
      return 0;
   else
*/
   if (h->header[hdr] != NULL) {
#ifdef STRICT
      if ( hdr == SUBJECT )           /* NOS smtp -> nntp gateway duplicates */
         return 0;                    /* the subject line */
      fflush(stdout);
      fprintf(stderr, "parseheader: duplicate %s line: - %s\n", hd[hdr].str, s);
      return -1;
#else
      return 0;
#endif
      }
   else {
      h->header[hdr] = h->hwm;
      copyin(h, s);

      return 0;
      }
}

static int
hdr_cont(Header_t *h, char *s)
{
/*
   printf("hdrcont: %s\n", s);
   if ( !Wanted(h->lasth) )
      return 0;
   else
*/
   if (h->hwm == 0) {
      fflush(stdout);
      fprintf(stderr, "parseheader: bad continuation - %.25s ...\n", s);
      return -1;
      }
   else {
/*
      if (h->lasth != UNKNOWN)
*/
         *(h->hwm-1) = ' ';
      copyin(h, s);

      return 0;
      }
}

static Header
loosehtype(char *s, char **data)
{
   char *p = s;
   struct header *result;

   /*
    * check to see if there is a ':' 
    */
   while ((*p != ':') && *p && !isspace(*p))
      p++;

   if ( *p != ':' )
      return NOHEADER;

   *p = '\0';
   result = (struct header *) bsearch(s, hd, NumHeaders,
         sizeof(struct header), (CMPFN) stricmp);

   *p = ':';  /* overwrite nul with ':' (which was originally there) */

   if (result != NULL) {
      /*
       * find start of header data
       */
      char *k = p + 1;

      while (isspace(*k))
         k++;
      *data = k;

      return result->hdr;
      }
   else 
      return UNKNOWN;
}

#ifdef notdef
static Header
htype(char *s)
{
   char *dummy;
   
   return loosehtype(s, &dummy);
}
#endif

static int
eoh(char line[])        /* end of header */
{
   return (line[0] == '\0') || (line[0] == '\n') || (line[0] == '-');
}

static int
checkheader(char *line, Header_t *ht)
{
   Header h;
   char *data;

   if (*line == '>')                           /* Escaped '>From' */
      return 0;

   if ((*line == ' ') || (*line == '\t'))      /* line continuation */
      if (ht->lasth < NOHEADER)
         return hdr_cont(ht, line);
      else
         return 0;

   h = loosehtype(line, &data);
/*
   printf("checkheader: <%s>, %d\n", line, h);
*/
   if ( h == NOHEADER ) {
      fflush(stdout);
      fprintf(stderr, "parseheader: bad header - %.45s ...\n", line);
      return -1;
      }
   else if (h == UNKNOWN) {
      ht->lasth = UNKNOWN;
      return copyin(ht, line);
      }
   else 
      return hdr_add(ht, h, data);
}

/*
 *  Read from buffer parsing headers. Fill in details in the argv array.
*/
int
gethdrb(Mailfile *m, Header_t *h)
{
   int ret = 0;
   char line[LINELEN];

   hdr_init(h);

#ifdef DEBUG
   fflush(stdout);
   fprintf(stderr, "header: starting new message\n", line);
#endif

   while (mail_getstring(m, line, sizeof(line)) != NULL)
      if (eoh(line))
         break;
      else if (checkheader(line, h) == -1) {
         ret = -1;
         break;
         }
#ifdef DEBUG
      else 
         fprintf(stderr, "header: checking <%s>\n", line);

   fprintf(stderr, "header: header finishing on line <%s>\n", line);
   dump(h);
#endif

   return ret;
}

/*
 *  Read from a file parsing headers. Fill in details in the hdr struct.
 */
int
gethdrf(FILE *fp, Header_t *h)
{
   int ret = 0;
   char line[LINELEN];

   hdr_init(h);

   while (fgets(line, sizeof(line), fp) != NULL)
      if (eoh(line))
         break;
      else if (checkheader(line, h) == -1) {
         ret = -1;
         break;
         }

#ifdef DEBUG
   fprintf(stderr, "header: finishing on line %s\n", line);
   dump(h);
#endif

   return ret;
}

static int
isknown(Header_t *h, char *p)
{
   int i;

   for (i=0; i<MAXHDR; i++)
      if (p == h->header[i])
         return 1;

   return 0;
}

/*
 * dump any unknown headers found (while hdr parsing) to a file
 */
int
putuhdr(FILE *fp, Header_t *h)
{
   char *p = h->buf;

   while (p < h->hwm)
      if (!isknown(h, p)) {
         while (*p)
            if (*p == '\n') {
               fputs("\n\t", fp);
               p++;
               }
            else if (fputc(*p++, fp) == EOF)
               return -1;
         fputc('\n', fp);
         p++;
         }
      else 
         while (*p++)
            ;

   return 0;
}

FILE *
gethdrfn(char *filename, Header_t *h)
{
   FILE *fp = fopen(filename, "r");

   if (fp == NULL)
      return NULL;
   else if (gethdrf(fp, h) == -1) {
      fclose(fp);
      return NULL;
      }
   else 
      return fp;
}

static int
writes(FILE *fp, struct addr *tp)
{
   int len = 0;

   if (tp->next != NULL) {
      len = writes(fp, tp->next);  /* recurse */
      if (len > 50) {
         fputs(",\n\t", fp);
         len = 0;
         }
      else
         fputs(", ", fp);
      }
      
   fputs(tp->user, fp);
   len += strlen(tp->user);
   if ( (tp->host != NULL) || (*tp->host != '\0') ) {
      fprintf(fp, "@%s", tp->host);
      len += strlen(tp->host) + 1;
      }

#ifdef DEBUG
   fprintf(stderr, "send: mailing %s@%s\n", tp->user,
      tp->host == NULL ? getrc(hostname) : tp->host);
#endif
   return len;
}

static int
write2s(FILE *fp, char *header, struct addr *tolist)
{
   fputs(header, fp);
   writes(fp, tolist);

   return fputc('\n', fp) == EOF ? -1 : 0;
}

/*
 * Returns:
 *  -1L on error
 *   0L if a passed in message-id was used
 *   otherwise the local msgid for the message/article.
 */
long
putrfc822(FILE *fp, Header_t *h, struct addr *tolist, struct addr *cclist)
{
   long id = 0L;

   /*
    * Message-ID:       (optional, for mail - but not news)
    */
   if (HEADERP(h, MESSAGEID) == NULL) {
      if ((id = get_msgid(getrc(mqueue))) == -1L)
         return -1L;
      else
         fprintf(fp, "%s<%ld@%s>\n", Hdr[MESSAGEID], id, getrc(hostname));
      }
   else if (*(HEADERP(h, MESSAGEID)) == '<')
      fprintf(fp, "%s%s\n", Hdr[MESSAGEID], HEADERP(h, MESSAGEID));
   else 
      fprintf(fp, "%s<%s>\n", Hdr[MESSAGEID], HEADERP(h, MESSAGEID));

   /*
    * Date:
    */
   if (fprintf(fp, "%s%s", Hdr[DATE], ptime(time(NULL))) == EOF)
      return -1L;

   /*
    * From:
    */

   if (fputs(Hdr[FROM], fp) == EOF)
      return -1L;
   else if (HEADERP(h, FROM) == NULL) {
      if (getrc(fullname) != NULL)
         fprintf(fp, "%s@%s (%s)\n", getrc(username), getrc(hostname), getrc(fullname));
      else 
         fprintf(fp, "%s@%s\n", getrc(username), getrc(hostname));
      }
   else if (fprintf(fp, "%s\n", HEADERP(h, FROM)) == EOF)
      return -1L;

   /*
    * Subject:   (optional)
    */
   if (fprintf(fp, "%s%s\n", Hdr[SUBJECT],
      HEADERP(h, SUBJECT) == NULL ? "" : HEADERP(h, SUBJECT)) == EOF)
      return -1L;

   /*
    * Sender:
    */
   if (HEADERP(h, FROM) != NULL) {
      if (fputs(Hdr[SENDER], fp) == EOF)
         return -1L;
      if (getrc(fullname) != NULL)
         fprintf(fp, "%s@%s (%s)\n", getrc(username), getrc(hostname), getrc(fullname));
      else 
         fprintf(fp, "%s@%s\n", getrc(username), getrc(hostname));
      }

   /*
    * Reply-To:
    */
   if (HEADERP(h, REPLYTO) != NULL)
      fprintf(fp, "%s%s\n", Hdr[REPLYTO], HEADERP(h, REPLYTO));
   else if (getrc(replyto) != NULL)
      fprintf(fp, "%s%s\n", Hdr[REPLYTO], getrc(replyto));

   if (tolist != NULL)
      write2s(fp, Hdr[TO], tolist);
   if (cclist != NULL)
      write2s(fp, Hdr[CC], cclist);

#ifdef AX25BBS
   if (HEADERP(h, XBBSMSGTYPE) != NULL)
      if (fprintf(fp, "%s%s\n", Hdr[XBBSMSGTYPE], HEADERP(h, XBBSMSGTYPE)) == EOF)
         return -1L;
#endif

   if ( (putuhdr(fp, h) == -1) ||
        (putc('\n', fp) == EOF)
      )
      return -1L;
   else
      return id;
}

#ifdef notdef
static int
write_recip(FILE *fp, char *header, int argc, char *argv[])
{
   int i;
   int len = 0;

   fputs(header, fp);

   for (i=0; i<(argc-1); i++) {
      len += strlen(argv[i]);

      if (len > 50) {
         fprintf(fp, "%s,\n\t", argv[i]);
         len = 0;
         }
      else
         fprintf(fp, "%s, ", argv[i]);
      }

   return fprintf(fp, "%s\n", argv[argc]) == EOF ? -1 : 0;
}
#endif

