/*
   help.c : Copyright Paul Healy, EI9GL, 1993.

   920708 : Created.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "config.h"
#include "help.h"
#include "mailer.h"
#include "rc.h"
#include "pc.h"
#include "misc.h"

#ifdef __TURBOC__
#ifndef __BORLANDC__
extern char *tzname[2];
#endif
#endif

static char Version[] = "Version 0.3 (" __DATE__ ")\n";

static int help_option = 0;

int
version(FILE *fp)
{
   return fputs(Version, fp) == EOF ? -1 : 0;
}

static void
copyr(FILE *fp)
{
   char *perm = "Copyright, %s permission granted for non-commercial use%c";

   version(fp);
   fprintf(fp, perm, "1992-1993 Paul Healy EI9GL", '\n');
   fprintf(fp, perm, "1987 Bdale Garbee", '\n');
   fprintf(fp, perm, "1988 Dave Trulli NN2Z", ' ');
}

int
issue(FILE *fp)
{
   char buf[MAXNAMLEN];
   char *p;

   fputs("bmh: a Mail Handler for NOS based on bm     ", fp);
   copyr(fp);

   fprintf(fp, "\n\noptions:\n [LINELEN=%d] [MAXNAMLEN=%d] [HDRBUFLEN=%d]",
      LINELEN, MAXNAMLEN, HDRBUFLEN);
#ifdef MAIL
   fprintf(fp, " [MAIL]");
#endif
#ifdef NEWS
   fprintf(fp, " [NEWS]");
#endif
#ifdef EXPIRY
   fprintf(fp, " [EXPIRY]");
#endif
#ifdef AX25BBS
   fprintf(fp, " [AX25BBS]");
#endif
#ifdef INDEX
   fprintf(fp, " [INDEX]");
#endif

#ifdef SWAP
   fprintf(fp, " [SWAP]");
#endif

#ifdef __TURBOC__
   fprintf(fp, " [STACKSIZE=%d] [TURBOC=0x%x]", _stklen, __TURBOC__);
#endif

   fprintf(fp, "\n");

   p = getruncom(buf);
   printf("configuation:\n");
   printf(" [rcfile=%s]", p == NULL ? "(" RUNCOM " not found)" : p);
   printf(" [user=%s@%s]\n", getrc(username), getrc(hostname));
   printf(" [timezone=%s] [dayfirst=%s]", gettz(buf), getrc(dayfirst));
#ifdef SWAP
   printf(" [swapping=%s]", getrc(swap));
#endif
   printf(" [bufsize=%u]", (size_t) atol(getrc(bufsize)));
   p = getrc(index);
   printf(" [indexing=%s]\n", *p=='y' ? "on" : "off" );

   printf(" [spool=%s]\n [mail=%s]\n [queue=%s]\n [tmp=%s]\n", getrc(spooldir),
      getrc(maildir), getrc(mqueue), gettmpdir());

   printf("environment:\n");
#ifdef __TURBOC__
   printf(" [dos=v%d.%d]", _osmajor, _osminor);
   if (dpmipresent() != -1)
      printf(" [cpu=80%d86] [dpmi=v%d.%d]", _cputype, _dpmimajor, _dpmiminor);
   if (dvpresent() != -1)
      printf(" [desqview=v%d.%d]", _dvmajor, _dvminor);
   if (cedpresent() != -1)
      printf(" [ced]");
   putc('\n', stdout);
   printf(" [tz=%s] [dst=%s]",
      tzname[0]==NULL ? "not set" : tzname[0],
      tzname[1]==NULL ? "not set" : tzname[1]);
   printf(" [timezone=%ld] [daylight=%d]\n", timezone, daylight);
   printf(" [directvideo=%d]", directvideo);
   printf(" [program=%s]", strlwr(_argv[0]));
   printf(" [cwd=%s]\n", strlwr(getcwd(buf, sizeof(buf))));
   printf(" [coreleft=%u] [farcoreleft=%lu]\n", coreleft(), farcoreleft());
#endif
   return 0;
}

int
showissue(int argc, char *argv[])
{
   if (dohelp(argc, argv, "issue (show configuration/version/copyright information)"))
      return 0;

   return issue(stdout);
}

int
sethelp(int i)
{
   return help_option = i;
}

int
helplevel(void)
{
   return help_option;
}

static int
help(char *usage)
{
   if (help_option == 0) {
      fprintf(stdout, "usage: %s\n\n", usage);
      issue(stdout);
      exit(0);
      }
   else 
      fputs(usage, stdout);

   return 1;
}

int
ishelp(int argc, char *argv[])
{  
   if (argc < 2)
      return 0;
   else 
      return isflag("-help", argc, argv)  ||
             isflag("-h", argc, argv)  ||
             isflag("-?", argc, argv);
}

/*
 * returns non zero result if help was requested
 */
int
dohelp(int argc, char *argv[], char *usage)
{
   if (ishelp(argc, argv))
      return help(usage);
   else 
      return 0;
}
