#ifndef	_IFACE_H
#define	_IFACE_H

#ifndef	_GLOBAL_H
#include "global.h"
#endif

#ifndef	_MBUF_H
#include "mbuf.h"
#endif

#ifndef _PROC_H
#include "proc.h"
#endif

/* Packet driver interface classes */
#define	CL_NONE		0
#define	CL_ETHERNET	1
#define	CL_PRONET_10	2
#define	CL_IEEE8025	3
#define	CL_OMNINET	4
#define	CL_APPLETALK	5
#define	CL_SERIAL_LINE	6
#define	CL_STARLAN	7
#define	CL_ARCNET	8
#define	CL_AX25		9
#define	CL_KISS		10
#define CL_IEEE8023	11
#define CL_FDDI 	12
#define CL_INTERNET_X25 13
#define CL_LANSTAR	14
#define CL_SLFP 	15
#define	CL_NETROM	16
#define CL_PPP		17
#define NCLASS		18

/* Interface encapsulation mode table entry. An array of these structures
 * are initialized in config.c with all of the information necessary
 * to attach a device.
 */
struct iftype {
	char *name;		/* Name of encapsulation technique */
	int type;		/* Type field for network process */

	int (*init)	__ARGS((struct iface *));
	char *(*format) __ARGS((char *,char *));
				/* Function that formats addresses */
	int (*scan)	__ARGS((char *,char *));
				/* Reverse of format */
	int hwalen;		/* Length of hardware address, if any */
};
#define	NULLIFT	(struct iftype *)0
extern struct iftype Iftypes[];


/* Interface control structure */
struct iface {
	struct iface *next;	/* Linked list pointer */
	char *name;		/* Ascii string with interface name */

	int32 addr;		/* IP address */
	int32 broadcast;	/* Broadcast address */
	int32 netmask;		/* Network mask */

	int16 mru;		/* Maximum receive unit size */
	int16 mtu;		/* Maximum transmission unit size */
#define MXU_DEFAULT	1500		/* default for both MRU and MTU */
#define MXU_MINIMUM	 576		/* minimum for both MRU and MTU */
#define MTU_SLIP	1006		/* SLIP default for MTU */

	int16 flags;		/* Configuration flags */
#define DATAGRAM_MODE	0		/* Send datagrams in raw link frames */
#define CONNECT_MODE	1		/* Send datagrams in connected mode */

	int16 trace;		/* Trace flags */
#define IF_TRACE_OUT	0x0001		/* Output packets */
#define IF_TRACE_IN	0x0010		/* Packets to me except broadcast */
#define IF_TRACE_ASCII	0x0100		/* Dump packets in ascii */
#define IF_TRACE_HEX	0x0200		/* Dump packets in hex/ascii */
#define IF_TRACE_NOBC	0x1000		/* Suppress broadcasts */
#define IF_TRACE_RAW	0x2000		/* Raw dump, if supported */
	char *trfile;		/* Trace file name, if any */
	FILE *trfp;		/* Stream to trace to */

	struct iface *forw;	/* Forwarding interface for output, if rx only */

	struct proc *rxproc;	/* Receiver process, if any */
	struct proc *txproc;	/* Transmitter process, if any */
	struct proc *supv;	/* Supervisory process, if any */

/* Device dependent */
	void *ddv;		/* Device dependent values, if any */
	int16 d_index;		/* Subdevice number */
	int16 d_type;		/* Device type */
#define IFD_NONE	0
#define IFD_PACKET	1
#define IFD_COM_PORT	2
#define IFD_3C501	3
#define IFD_DRSI	4
#define IFD_EAGLE	5
#define IFD_HAPN	6
#define IFD_HS		7
#define IFD_PC100	8
#define IFD_PI		9
#define IFD_SCC 	10

	char   *hwaddr; 	/* Device hardware address, if any */

			/* send mbufs to device */
	int (*iosend)		__ARGS((int,struct mbuf *));
			/* fetch input chars from device */
	int (*ioget)		__ARGS((int));
			/* To device -- control */
	int32 (*ioctl)	__ARGS((struct iface *,int cmd,int set,int32 val));
			/* From device -- when status changes */
	int (*iostatus) 	__ARGS((struct iface *,int cmd,int32 val));
			/* Call before detaching */
	int (*iostop)		__ARGS((struct iface *));
			/* Display status */
	void (*ioshow)		__ARGS((struct iface *));

/* Encapsulation dependent */
	void *edv;		/* Encapsulation dependent values, if any */
	int16 e_type;		/* Encapsulation type */

	struct iftype *iftype;	/* Pointer to appropriate iftype entry */

			/* Encapsulate an IP datagram */
	int (*send) __ARGS((struct mbuf *,struct iface *,int32,int,int,int,int));
			/* Encapsulate any link packet */
	int (*put) __ARGS((struct iface *,char *,char *,int16,struct mbuf *));
			/* Send raw packet */
	int (*raw)		__ARGS((struct iface *,struct mbuf *));
			/* Display status */
	void (*show)		__ARGS((struct iface *));
			/* Call to detach */
	void (*detach)		__ARGS((struct iface *));

	int (*discard)		__ARGS((struct iface *,struct mbuf *));
	int (*echo)		__ARGS((struct iface *,struct mbuf *));

	/* Counters */
	int32 ipsndcnt; 	/* IP datagrams sent */
	int32 iprecvcnt;	/* IP datagrams received */

	int32 lastsent; 	/* Clock time of last send */
	int32 lastrecv;		/* Clock time of last receive */

	int32 rawrecvcnt;	/* Raw packets received */
	int32 InOctets; 	/* # octets received */
	int32 InUnicast;	/* # unicast packets */
	int32 InMulticast;	/* # multicast packets */
	int32 InBroadcast;	/* # broadcast packets */
	int16 InDiscards;	/* # discards without error */
	int16 InErrors; 	/* # error packets */

	int32 rawsndcnt;	/* Raw packets sent */
	int32 OutOctets;	/* # octets sent */
	int32 OutUnicast;	/* # unicast packets */
	int32 OutMulticast;	/* # multicast packets */
	int32 OutBroadcast;	/* # broadcast packets */
	int16 OutDiscards;	/* # discards without error */
	int16 OutErrors;	/* # errors on send */

	struct mbuf *OutQ;
	int16 OutQLength;	/* # packets on Output Queue */

	int16 Speed;		/* current bandwidth, bits per second */
};
#define	NULLIF	(struct iface *)0
extern struct iface *Ifaces;	/* Head of interface list */
extern struct iface *Loopback;	/* Optional loopback interface */
extern struct iface *IP_tunnel; /* IP-in-IP pseudo interface */

extern char Noipaddr[];

/* In iface.c: */
struct iface *ismyaddr __ARGS((int32 addr));
struct iface *if_lookup __ARGS((char *name));
struct iface *if_attach __ARGS((char *name));
int if_detach __ARGS((struct iface *ifp));
char *if_name __ARGS((struct iface *ifp,char *comment));
int setencap __ARGS((struct iface *ifp,char *mode));
int bitbucket __ARGS((struct iface *ifp,struct mbuf *bp));

/* In config.c: */
int net_route __ARGS((struct iface *ifp,int ptype,struct mbuf *bp));

#endif	/* _IFACE_H */
