/*
   import.c : Copyright 1992-1993 Paul Healy, EI9GL. 

   Takes one or more messages in ax25 pbbs export format (optionally from
   stdin) and queues them for delivery.

   911012 : Created.
   911124 : Changed /EX comparision to catch a line with space(s) after the
            X. (strcmp -> strncmp).
   920501 : Cater for "@@" on some ANS bulletins
   920718 : bmh version
   921108 : Program will not now try to add a bid, if the lookup for the bid
            failed.
   930206 : Fixed problem with import silently doing nothing when there is
            no history file to start off with.
   930718 : Added R: line handling using code from N5KNX & WG7J as a guide.
            Also added support for reading directly from mail folders, with
            optional removal of the folder if the import goes successfully.
            The only thing missing from the N5KNX mods is rewrite handling &
            setting the rhs of the msgid field to relate to the original
            bbs - this may be important for nntp handling...
*/
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <time.h>
#include <ctype.h>
#include <stdlib.h>
#include "config.h"
#include "misc.h"
#include "rc.h"
#include "pc.h"
#include "lock.h"
#include "help.h"
#include "mailer.h"
#include "getopts.h"
#include "buf.h"
#include "mailfile.h"
#include "header.h"
#include "send.h"
#include "mkpath.h"

#ifdef BMH
#define main import_main
#endif

static char Verbose = 0;
static int Errors = 0;

static char *
grab(char *to, char *ptr)
{
   char *start = to;

   if (ptr==NULL)
      return NULL;
   while (*ptr && (*ptr==' ') )
      ptr++;
   while (*ptr && (*ptr!=' ') && (*ptr!='@')
               && (*ptr!='$') && (*ptr!='<') && (*ptr!='\n'))
      *to++ = *ptr++;
   *to = '\0';

   return start;
}

static int
addbid(char *bid)
{
   FILE *fp = fopenlocked(getrc(spooldir), "history", "", "a");
   char *tsb = getrc(tsbids);
   
   if (fp == NULL) {
      Errors++;
      return -1;
      }

   if ( (tsb != NULL) && (tolower(*tsb) == 'y') )
      fprintf(fp, "%s %ld\n", bid, time(NULL));   /* wg7j format */
   else 
      fprintf(fp, "%s\n", bid);

   if (ferror(fp)) {
      Errors++;
      fprintf(stderr, "addbid: could not add %s to the history file\n", bid);
      }

   return fcloselocked(fp, getrc(spooldir), "history");
}

static int
checkhistory(char *bid)
{
   FILE *fp = fopenlocked(getrc(spooldir), "history", "", "r");
   char line[LINELEN];
   int len = strlen(bid);
   
   if (fp == NULL) {
      fprintf(stderr, "import: warning can't open %s/history\n", getrc(spooldir));
      /*
       * If the file doesn't exist then the fopen (above) will have
       * failed. So try again. The proper way to do this would be to create
       * the history file at the start if it doesn't exist.
       */
      if ((fp = fopenlocked(getrc(spooldir), "history", "", "a")) == NULL) {
         fprintf(stderr, "import: unable to open %s/history\n", getrc(spooldir));
         Errors++;
         return -1;
         }
      else {
         fprintf(stderr, "import: created %s/history\n", getrc(spooldir));
         fcloselocked(fp, getrc(spooldir), "history");
         return addbid(bid);
         }
      }

   while ( fgets(line, sizeof(line), fp) != NULL )
      if (strncmp(line, bid, len) == 0) {
         fprintf(stderr, "import: %s found in history file. Skipping...\n", bid);
         fcloselocked(fp, getrc(spooldir), "history");
         return -1;
         }

   fcloselocked(fp, getrc(spooldir), "history");

   return addbid(bid);
}

static int
digestline(char *line, char *bbstype, char *to, char *from, char *msgid, int nohist)
{
   char temp[256], temp2[256], bid[64], *cp;

   *bbstype = line[1];

   /*
    * Digest TO fields
    */
   if ( (cp = strchr(line, '@')) != NULL)
      if (*(cp+1) == '@')                 /* double @ on some ANS bulletins */
         cp++;

   if ( cp != NULL)
      sprintf(to, "%s@%s", grab(temp, line+3), grab(temp2, cp+1));
   else
      sprintf(to, "%s@%s", grab(temp, line+3), getrc(hostname));

   /*
    * Digest FROM field
    */
   if ( (cp = strchr(line, '<')) != NULL) 
      sprintf(from, "%s@", grab(temp, cp+1));
   else
      from[0] ='\0';

   /*
    * Digest BID field
    */
   if ( (cp = strchr(line, '$')) != NULL) {
      sprintf(msgid, "%s@%s.bbs", grab(bid, cp+1), getrc(username));
      return nohist ? 0 : checkhistory(bid);
      }
   else {
      long id = get_msgid(getrc(mqueue));

      if (id == -1L) {
         fprintf(stderr, "import: unable to get message-id\n");
         abort();
         }
      else 
         sprintf(msgid, "%ld@%s", id, getrc(hostname));

      return 0;
      }
}

/*
 * returns:
 *   0 if an R: line wasn't found
 *   1 of an R: line was found
 *  -1 for an error
 *
 * Find '@[:]CALL.STATE.COUNTRY' or '?[:]CALL.STATE.COUNTRY' string
 * The : is optional.
 */
static int
rline(char *line, char *sending_bbs, int maxlen)
{
   if (strncmp(line, "R:", 2) != 0)
      return 0;
   else {
      char *call = strchr(line, '@');  /* I haven't seen '?', but ... */

      if ( (call == NULL) && ((call = strchr(line, '?')) == NULL) ) {
         fprintf(stderr, "rline: bad R: line %s...\n", line);
         return -1;
         }
      else 
         call += *(call+1) == ':' ? 2 : 1;

      maxlen--; /* leave room for the terminating nul */
      while (!isspace(*call) && maxlen)
         *sending_bbs++ = *call++, maxlen--;

      *sending_bbs = '\0';

      if (maxlen == 0) {
         fprintf(stderr, "rline: warning bbs address too long %s...\n", sending_bbs);
         return -1;
         }
      else 
         return 1;
      }
}

static int
import(FILE *fp, int nohistory, char *toaddr)
{
   char line[LINELEN];

   while (fgets(line, sizeof(line), fp) != NULL) {
      char bbstype[2], msgid[LINELEN], to[LINELEN], 
           subj[LINELEN], from[LINELEN], frombbs[LINELEN],
           msgfn[MAXNAMLEN];
      FILE *msgfp;
      Header_t h;

      if ( (strlen(line) < 3) || (line[0] != 'S') )
         continue;
      if ( !( (line[1] == 'B') || (line[1] == 'P') ) ) /* T ? */
         continue;

      fprintf(stderr,"import: %s\n", rip(line));

      if (digestline(strlwr(line), bbstype, to, from, msgid, nohistory) == -1) {
         while (fgets(line, sizeof(line), fp) != NULL)
            if (strncmpi(line, "/EX", 3) == 0)
               break;
         continue;
         }

       strcpy(frombbs, getrc(hostname)); /* default from bbs */

      /*
       * Get SUBJECT
       */
      if (fgets(subj, sizeof(subj), fp) == NULL)
         subj[0] = '\0';
      else 
         rip(subj);

      if ( (msgfp = tempfile("import", msgfn, "w+")) == NULL)
         return -1;

      while (!feof(fp) && (fgets(line, sizeof(line), fp) != NULL) )
         if (strncmpi(line, "/EX", 3) == 0)
            break;
         else if (fputs(line, msgfp) == EOF) {
            fprintf(stderr, "import: unable to write to %s\n", msgfn);
            fclose(msgfp);
            unlink(msgfn);
            return -1;
            }
         else if (rline(line, frombbs, LINELEN) == -1) {
            fclose(msgfp);
            unlink(msgfn);
            return -1;
            }

      fseek(msgfp, 0L, SEEK_SET);

      hdr_init(&h);
      hdr_add(&h, TO, toaddr != NULL ? toaddr : to);
      hdr_add(&h, SUBJECT, subj);

      /*
       *  Let the header print routine add the from line if there wasn't
       *  a from field, a Sender: header would be added otherwise
       */
      if (from[0] != '\0')   
         hdr_add(&h, FROM, strcat(from, strlwr(frombbs)));
      else
         sprintf(from, "%s@%s", getrc(username), strlwr(frombbs));

      hdr_add(&h, MESSAGEID, msgid);
      bbstype[1] = 0;
      hdr_add(&h, XBBSMSGTYPE, strupr(bbstype));

      if (Verbose) {
         fprintf(stderr, "        MAIL FROM:<%s>\n", from);
         fprintf(stderr, "        RCPT TO:<%s>\n", HEADER(h, TO));
         fprintf(stderr, "        ARTICLE:<%s>\n", msgid);
         }

      if (dosend(msgfp, &h) == -1) {
         fprintf(stderr, "import: unable to queue message to %s\n", HEADER(h, TO));
         fclose(msgfp);
         unlink(msgfn);
         return -1;
         }
      fflush(stdout);
      fclose(msgfp);
      unlink(msgfn);
      }

   return 0;
}

/*
 * input can come from:
 *  1. stdin
 *  2. a named file (& optional removal if successful)
 *  3. a mailbox which will be locked during use (& optionally removed
 *    when finished). If no `+folder' appears on the command line, then
 *    the normal defaults come into play - contents of current.bmh, then
 *    the user's file from bm.rc (ie the unlink option is a bit dangerous).
 */
static FILE *
iopen(Mailfile *m, char *filename, char *name)
{
   null_check("iopen");

   if (!isatty(fileno(stdin))) {
      strcpy(name, "stdin");
      return stdin;
      }
   else if ( (filename != NULL) && (*filename != '+') )
      return fopen(strcpy(name, filename), "r");
   else if (mail_lock(m) != -1)
      return fopen(strcpy(name, mail_filename(m)), "r");
   else {
      fprintf(stderr, "import: can't lock %s\n", strcpy(name, mail_filename(m)));
      return NULL;
      }
}

static int
iclose(FILE *fp, Mailfile *m, char *filename, int unlinkit)
{
   null_check("iclose");
   if (fp == stdin)
      return 0;
   else if ( (filename != NULL) && (*filename != '+') ) {
      fclose(fp);
      if (unlinkit)
         return unlink(filename);
      else
         return 0;
      }
   else {
      fclose(fp);
      if (unlinkit)
         unlink(mail_filename(m));
      return mail_unlock(m);
      }
}

static char *optv[] = {
   "n|nohistory",
   "v|trace",
   "t:to",
   "u|unlink",
   NULL
};

int
main(int argc, char *argv[])
{
   int optchar;
   char *optarg;
   options_t *opts;
   int errors = 0;

   Mailfile m;
   FILE *fp;
   int nohist = 0, unlinkit = 0;
   char *to = NULL;
   int filename;
   char name[MAXNAMLEN];

   if (dohelp(argc, argv, "import [file|+folder|-nohist|-to addr|-unlink|-trace] (ax25 pbbs mailer)"))
      return 0;

   if (makepath(getrc(spooldir)) == -1)         /* once bitten... */
      return -1;

   mail_zero(&m);

   if ((opts = mail_opts(&m, &argc, &argv, optv, "import")) == NULL)
      return -1;

   while( (optchar = getopts(opts, argc, argv, &optarg)) != 0 )
      switch(optchar) {
         case 'n' : nohist = 1; break;
         case 't':  to = optarg; break;
         case 'u' : unlinkit = 1; break;
         case 'v' : Verbose = 1; break;
         default : ++errors; break;
         }

   if ((filename = mail_endopts(errors, opts)) == -1) {
      fprintf(stderr, "import: error parsing command line\n");
      return -1;
      }

   if ((fp = iopen(&m, argv[filename], name)) == NULL) {
      fprintf(stderr, "import: can't open %s\n", name);
      return -1;
      }
   else if (Verbose)
      fprintf(stderr, "import: reading from %s\n", name);

   Errors = 0;

   if (import(fp, nohist, to) != 0)
      Errors++;

   if (Errors)
      fprintf(stderr, "import: error count =%d\n", Errors);

   return iclose(fp, &m, argv[filename], (Errors == 0) && unlinkit);
}   

