/*
   inc.c : Copyright Paul Healy, EI9GL, 1993.

   inc ei.news             -- dump spool/news/ei/news to spool/mail/ei/news.txt
   inc +ei.news ei.news    -- same as above
   inc +news ei.news       -- dump spool/news/ei/news to spool/mail/news.txt
   inc                     -- dump all active groups

   930124 : Created.
*/
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "config.h"
#include "misc.h"
#include "lock.h"
#include "rc.h"
#include "buf.h"
#include "mailfile.h"
#include "header.h"
#include "nntp.h"
#include "help.h"
#include "misc.h"
#include "mailer.h"

#ifdef BMH
#define main inc_main
#endif

/*
 * Each nntp (server-style) newsgroup has an extra file recording
 * the last message incorporated in the normal mail system.
 */
static long
currentinc(char *newsgroup)
{
   FILE *fp;
   char filename[MAXNAMLEN];
   long num;

   sprintf(filename, "%s/%s/current.inc", getrc(newsdir), newsgroup);
   if ( (fp = fopen(filename, "r")) == NULL)
      return 1L;
   if (fscanf(fp, "%ld", &num) != 1) {
      fclose(fp);
      return -1L;
      }
   fclose(fp);

   return num;
}

static long
setcurrentinc(char *newsgroup, long num)
{
   FILE *fp;
   char filename[MAXNAMLEN];

   sprintf(filename, "%s/%s/current.inc", getrc(newsdir), newsgroup);
   if ( (fp = fopen(filename, "w")) == NULL)
      return -1L;
   if (fprintf(fp, "%ld", num) == EOF) {
      fclose(fp);
      return -1L;
      }
   fclose(fp);

   return num;
}

/* --------------------------------------------------------------------- */

static int
addfrom(FILE *in, FILE *out)
{
   Header_t h;

   gethdrf(in, &h);

   if (HEADER(h, FROM) == NULL)
      fprintf(out, "From nntp@%s %ld\n", getrc(hostname), time(NULL));
   else 
      fprintf(out, "From %s %ld\n", HEADER(h, FROM), time(NULL));

   if (fseek(in, 0L, SEEK_SET) == 0)
      return 0;
   else
      return -1;
}

static int      
add2mail(FILE *fp, char *mailfile)
{
   FILE *mailfp;
   int ret;

   if ( (fp == NULL) || (mailfile == NULL) ||
        ((mailfp = fopenlocked(getrc(maildir), mailfile, ".txt", "a")) == NULL)
      )
      return -1;

   if (addfrom(fp, mailfp) == -1)
      ret = -1;
   else 
      ret = dupfile(fp, mailfp);

   fcloselocked(mailfp, getrc(maildir), mailfile);

   return ret;
}

typedef long (*wpfn)(char *newsgroup, char *dir, char *mailfile);

static long
dump(char *newsgroup, char *newsdir, char *mailfile)
{
   long i;
   long start = currentinc(newsdir);
   long high, low;

   if ( (start == -1L) || (newsgroup == NULL) || (newsdir == NULL) )
      return -1L;
   if (mailfile == NULL)
      mailfile = newsgroup;
   news2dir(mailfile);

   if (getnnrange(newsgroup, &high, &low) == -1) {
      fprintf(stderr, "inc: error checking active for %s\n", newsgroup);
      return -1L;
      }

   for (i=start; i<high; i++) {
      char filename[MAXNAMLEN];
      FILE *fp;

      sprintf(filename, "%ld", i);
      if ( (fp = fopenlocked(newsdir, filename, "", "r")) ==NULL) {
         fprintf(stderr, "inc: warning can't open %s\n", filename);
         continue;
         }
      if (add2mail(fp, mailfile) == -1) {
         fprintf(stderr, "inc: warning append %s to %s failed\n", filename, mailfile);
         fcloselocked(fp, newsdir, filename);
         return -1L;
         }
      fcloselocked(fp, newsdir, filename);
      }

   return setcurrentinc(newsgroup, high);
}

static long
walkpointer(wpfn f)
{
   FILE *fp = fopenlocked(getrc(newsdir), "pointer", "", "r");
   char line[LINELEN];

   if (fp == NULL) {
      fprintf(stderr, "inc: can't open news pointer file (%s/pointer)\n", getrc(newsdir));
      return -1L;
      }

   while (fgets(line, LINELEN, fp) != NULL) {
      char *p = strchr(line, ' ');

      if (p == NULL) {
         fprintf(stderr, "inc: bad line in pointer file: %s\n", line);
         continue;
         }
      *p = 0;
      f(line, news2dir(p+1), NULL);
      }
   return (long) fcloselocked(fp, getrc(newsdir), "pointer");
}

static long
dumpall(void)
{
   return walkpointer(dump);
}

int
main(int argc, char *argv[])
{
   char path[MAXNAMLEN];

   if (dohelp(argc, argv, "inc [+folder] [newsgroup] (incorporate news into the mail system)"))
      return 0;

   if (argc == 1)
      return (int) dumpall();
   else if (argc == 2) {
      if (argv[1][0] == '+') {
         fprintf(stderr, "inc: missing newsgroup for mail folder %s\n", argv[1]);
         return -1;
         }
      return (int) dump(argv[1], getnnpath(argv[1], path, MAXNAMLEN), NULL);
      }
   else if (argc == 3)
      return (int) dump(&argv[1][1], getnnpath(&argv[1][1], path, MAXNAMLEN), NULL);

   return -1;
}
