/* IP header tracing routines
 * Copyright 1991 Phil Karn, KA9Q
 */
 /* Mods by PA0GRI */
#include <stdio.h>
#include "global.h"
#include "config.h"
#include "mbuf.h"
#include "internet.h"
#include "iface.h"
#include "ip.h"
#include "trace.h"
#include "netuser.h"

void
ip_dump(ifp,bpp,check)
struct iface *ifp;
struct mbuf **bpp;
int check;
{
	struct ip ip;
	int16 ip_len;
	int16 length;
	int16 csum;

	if(bpp == NULLBUFP || *bpp == NULLBUF)
		return;	

	prtTrace(ifp,"IP:");
	/* Sneak peek at IP header and find length */
	ip_len = ((*bpp)->data[0] & 0xf) << 2;
	if(ip_len < IPLEN){
		prtTrace(ifp," bad header\n");
		return;
	}
	if(check)
		csum = cksum(NULLHEADER,*bpp,ip_len);
	else
		csum = 0;

	ntohip(&ip,bpp);	/* Can't fail, we've already checked ihl */

	/* Trim data segment if necessary. */
	length = ip.length - ip_len;	/* Length of data portion */
	trim_mbuf(bpp,length);	
	prtTrace(ifp," len %u",ip.length);
	prtTrace(ifp," %s",inet_ntoa(ip.source));
	prtTrace(ifp,"->%s ihl %u ttl %u",
		inet_ntoa(ip.dest),ip_len,uchar(ip.ttl));
	if(ip.tos != 0)
		prtTrace(ifp," tos %u",uchar(ip.tos));
	if(ip.offset != 0 || ip.flags.mf)
		prtTrace(ifp," id %u offs %u",ip.id,ip.offset);
	if(ip.flags.congest)
		prtTrace(ifp," CE");
	if(ip.flags.df)
		prtTrace(ifp," DF");
	if(ip.flags.mf){
		prtTrace(ifp," MF");
		check = 0;	/* Bypass host-level checksum verify */
	}
	if(csum != 0)
		prtTrace(ifp," CHECKSUM ERROR (%u)",csum);

	if(ip.offset != 0){
		prtTrace(ifp, "\n");
		return;
	}
	switch(uchar(ip.protocol)){
	case TCP_PTCL:
		prtTrace(ifp," prot TCP\n");
		tcp_dump(ifp,bpp,ip.source,ip.dest,check);
		break;
	case UDP_PTCL:
		prtTrace(ifp," prot UDP\n");
		udp_dump(ifp,bpp,ip.source,ip.dest,check);
		break;
	case ICMP_PTCL:
		prtTrace(ifp," prot ICMP\n");
		icmp_dump(ifp,bpp,ip.source,ip.dest,check);
		break;
	case IP_PTCL:
		prtTrace(ifp," prot IP\n");
		ip_dump(ifp,bpp,check);
		break;
	case AX25_PTCL:
		prtTrace(ifp," prot AX25\n");
		ax25_dump(ifp,bpp,check);
		break;
#ifdef	RSPF
	case RSPF_PTCL:
		prtTrace(ifp," prot RSPF\n");
		rspf_dump(ifp,bpp,ip.source,ip.dest,check);
		break;
#endif
	default:
		prtTrace(ifp," prot %u\n",uchar(ip.protocol));
		break;
	}
}
