/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */
/* This file has the level 2 routines that are needed for the Level 3 test
   environment and will be replaced with real level 2 counterparts when
   the level 3 code is intergerated with the level 2 (from KE3Z) */

extern struct axcb *links;
extern struct interface *ifaces[8];
extern int L3TRIED[];
extern int USER_MAXVC;	/* Configurable limit on # of vc's a user can have */
extern int USER_L3_W;	/* User Level 3 Window */

extern struct ax25_parms *l2parms, *l3parms;
extern struct ax25_parms dl2parms, dl3parms;

int find_call();
 
struct axcb *
L2_CONN(port,sourcex,destinx,digi1x,digi2x,digi3x,digi4x,l2user)
int port;
struct ax25_addr *sourcex,*destinx,*digi1x,*digi2x,*digi3x,*digi4x;
int l2user;
{
	static struct axcb *lnk;
	register struct ax25_addr *ch;
	static struct ax25_addr *source,*destin[5];
	static int alen, i;

	source=sourcex;
	destin[0]=destinx;
	destin[1]=digi1x;
	destin[2]=digi2x;
	destin[3]=digi3x;
	destin[4]=digi4x;

	if (source == NULL || destin[0] == NULL) return NULL;

	ch= (struct ax25_addr *)str_work;

	alen=0;
	for (i=0;i<5;i++) {
		if ((destin[i]) && (destin[i]->call[0])) {
			memcpy(ch,destin[i],AXALEN);
			ch++;
			alen++;
		}
	}
	ch--;
	ch->ssid |= E; /* Set eof bit in address */

	lnk=open_ax25(source,str_work,ifaces[port],alen);
	if (!lnk) return lnk;
	if (!l2user) lnk_l3_pars(lnk);
	lnk->port = port;
	lnk->R = 0;
	axconnect(lnk);
	return lnk;
}

struct VCS *
new_vc()
{
	register struct VCS *vc;
	int l3_reset(), l3_clear(), l3_send(), l3_recv(), l3_connected();

	vc=(struct VCS *)calloc(1,sizeof(struct VCS));
	if (vc != NULL) {
		vc->WI=USER_L3_W;	/* Assume it's a user */
		vc->RESET=l3_reset;
		vc->CLEAR=l3_clear;
		vc->CONNECTED=l3_connected;
		vc->SEND=l3_send;
		vc->RECV=l3_recv;
	}
	return (vc);
}

struct axcb *
new_link()
{
	register struct axcb *lnk;

	lnk=(struct axcb *)calloc(1,sizeof(struct axcb));
	if (lnk != NULL) {
		lnk->maxvc=USER_MAXVC;	/* Assume a user link */
	}
	return (lnk);
}
