/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */
/* These are the routines that manage the L3_ tables
   They are:
	L3_CALLS - The known Level 3 nodes/Users
	L3_DIGI  - An optional Digipeater to use to get to them
	L3_PORT  - The Port they were last using
	L3_WI	 - Packet Window to use
	L3_MAXVC - Max # of VC's for this Level 3 User

   There are also 2 tables associated with nodes only
	L3_RETRY_DELAY - The time to wait before attempting to use the
			 link after it failed to come up during a Call Setup
			 (Can be 0 to always try the link)
	L3_TRIED       - The current value of the above count down timer

	These tables are indexed with the "Level 3 Callsign number"
	obtained from find_call(callsign) (which searches L3_CALLS)

	Nodes are index 1-15
	Users are index 16-L3_CALL_SIZE
	The callsign of THIS node is L3_CALLS[0]
*/

extern
int	maxl3dev;		/* The max size of the L3CALL tables w/users */
extern
struct ax25_addr L3CALL[],
		 L3DIGI[];
extern
unsigned char	 L3PORT[],
		 L3_WI[],
		 L3_MAXVC[];
extern int	 L3DELAY[16],
		 L3CSIZ,
		 L3NSIZ;

int		 L3TRIED[16];

int
callcmp(stn1, stn2x)
register struct ax25_addr *stn1;
struct ax25_addr *stn2x;
{
	static struct ax25_addr *stn2;
	static int j;

	if ( !stn1 || !(stn2=stn2x)) return FALSE;

	if ((stn1->ssid & SSID) != (stn2->ssid & SSID)) return FALSE;

	for (j=0; j<ALEN; j++)
		if (stn1->call[j] != stn2->call[j]) return FALSE;
	return TRUE;
}
/*		if ((stn1->call[j] & ~E) != (stn2->call[j] & ~E)) {
			return FALSE;
		}
*/

int find_call(stn)
struct ax25_addr *stn;
{
	static int i;

	for (i=0; i<=maxl3dev; i++)
		if (L3CALL[i].call[0] && callcmp(&L3CALL[i], stn)) return i;
	return -1;
}

int
linked_to(station, p)
struct ax25_addr *station;
struct axcb **p;
{
	register struct axcb *p1;

	p1=links;
	while (p1 != NULL) /* Find correct for station */ {
		if (callcmp(&p1->path[0], station)) break; /* Found it! */
		p1 = p1->next;
	}
	*p=p1;
	return (p1 != NULL);
}

#if 0
ndef z80
int adcsgn(stn, digi, port, siz, max)
struct ax25_addr *stn, *digi;
int port, *siz, max;
{
	static unsigned char i;
	register int size;

	size= *siz;
	if (size>=max) return FALSE;
	if (stn->call[0] == 0) return FALSE;
	size++;
	for (i=0; i<ALEN; i++) {
		L3CALL[size].call[i] = stn->call[i];
		L3DIGI[size].call[i] = digi->call[i];
	}
	L3CALL[size].ssid = stn->ssid & SSID;
	L3DIGI[size].ssid = digi->ssid & SSID;
	L3PORT[size]=port;
	*siz=size;
	return TRUE;
}

int
adcl( stn, digi, port)
struct ax25_addr *stn, *digi;
int port;
{
	return adcsgn(stn, digi, port, &L3CSIZ, maxl3dev);
}

int
addnod( stn, digi, port, time)
struct ax25_addr *stn, *digi;
int port, time;
{
	if (adcsgn(stn, digi, port, &L3NSIZ, 15)) {
		L3DELAY[L3NSIZ]=time;
		return TRUE;
	}
	return FALSE;
}
#endif

void
link_count()
{
/* These should queue a timer when link fails and kill it when link comes ready */
	int i;

	/* Now update the L3TRIED lists for the routing manager */
	queue(NULL,link_count,1,0);	/* Dec link counters every second */
	for (i=1;i<16;i++) /* Decrement each tried counter for Nodes only */
		if (L3TRIED[i]>0) L3TRIED[i]--;
}
