/*
   letter.h: Copyright 1993, Paul Healy.

   It would be nice to stick this sort of data up high...

   930616 : Created
*/
#ifndef _letter_h_
#define _letter_h_

/*
 * Turbo C doesn't like bitfields longer than 16 bits, so...
 */
typedef enum { 
   Let_status   = 0x80000000, /* set if the message has been read */
   Let_replied  = 0x40000000, /* set if the message has been replied to */
   Let_overflow = 0x20000000, /* set if the message is longer than 2^28 - 1 bytes */
   Let_deleted  = 0x10000000, /* temporarily set while deleting messages */
   Let_length   = 0x0fffffff, /* mask giving the length in bytes of the message */
   Let_flags    = 0xf0000000
};

typedef unsigned long letter;

#define LET_CURMSG 0   /* slot 0 holds the current ('+') message */
#define LET_MAXMSG 1   /* slot 1 holds the number of messages */
#define LET_MSG0   2   /* slot 2 holds details on message 0 */

typedef letter *folder;

int let_setup(folder f, int maxmsg);   /* alloc memory, init vars */
int let_setrange(folder f, int lo, int hi, long flag);
int let_unsetrange(folder f, int lo, int hi, long flag);

#define let_fin(folder f) { free(f); f = NULL; }
#define let_isread(f, m) (f[m+LET_MSG0] & Let_status)
#define let_isreplied(f, m) (f[m+LET_MSG0] & Let_replied)
#define let_isdeleted(f, m) (f[m+LET_MSG0] & Let_deleted)
#define let_maxmsg(f) (f[LET_MAXMSG])

#define let_markasread(f, lo, hi) let_setrange(f, lo, hi, Let_status)
#define let_markreplied(f, m) let_setrange(f, m, m, Let_replied)
#define let_markdeleted(f, lo, hi) let_setrange(f, lo, hi, Let_deleted)
#define let_markundeleted(f, lo, hi) let_unsetrange(f, lo, hi, ~Let_deleted)

#ifdef notdef
long let_setstart(folder f, int msg, long start);
long let_getstart(folder f, int msg);
int let_save(folder f, char *name, int curmsg);
int let_load(folder f, char *name);

#define let_length(f, m) (f[m+LET_MSG0] & Let_length)
#endif

#endif /* _letter_h_ */

