/* Internet LPD Server filter definitions
 *   written by David Johnson (dave@cs.olemiss.edu)
 *
 * This code is in the public domain.
 *
 * Revision History:
 *
 * Revision 1.1  91/09/14  dave
 * Added device field to filter_stream to allow access to printer status
 *
 * Revision 1.0  91/09/04  dave
 * Initial Release
 *
 */

/*
 * Supported filters (found in lpdfilt.c)
 *
 * These (other than pr_filter and print_copy) should _not_ be called
 * directly.  Only through the filter_list below.
 */
extern void text_filter __ARGS((int unused, void *vparms, void *unused2));
extern void ps_switch_filter __ARGS((int unused, void *vparms, void *unused2));
extern void text_ps_filter __ARGS((int unused, void *vparms, void *unused2));
extern void pr_filter __ARGS((int unused, void *vparms, void *unused2));
extern int print_copy __ARGS((int fd, int device_fd));


struct LPDfilter {
	char *name;		/* filter name, used in printcap */
	void (*process) __ARGS((int unused,void *parms,void *unused2));
};

/* in lpdfilt.c */
extern struct LPDfilter filter_list[];

/*
 * Input and Output "streams" for filters
 */
struct filter_stream {
	char type;
#define IO_SOCKET	0
#define IO_FILE		1
#define IO_LOCAL	2
	union {
		int socket;
		FILE *fp;
		struct iface *ifp;
	} stream;
	struct LPDdevice *device;
};

/*
 * Generic parameters for filters
 *
 * When the job does not specify one of these parameters, the default
 * value in parentheses () is used.
 */
struct filter_parms {
	char *printer_name;	/* printer name */
	int  width;		/* page width in characters (D_WIDTH) */
	int  length;		/* page length in lines (D_LENGTH) */
	int  xwidth;		/* page width in pixels (D_PWIDTH) */
	int  ylength;		/* page length in pixels (D_PLENGTH) */
	int  literal;		/* binary? (0) */
	int  indent;		/* amount to indent in characters (0) */
	char *class;		/* job class (NULL) */
	char *jobname;		/* job name (NULL) */
	char *account;		/* accounting information (NULL) */
	char *login_name;	/* user's login name (NULL) */
	char *host;		/* originating host (NULL) */
	char format;		/* format of job */
	char *af_file;		/* accounting file (NULL) */
	int  default_ps;	/* default to postscript mode? (0) */
	char *ps_start;		/* postscript start init string (NULL) */
	char *ps_end;		/* postscript end string (NULL) */
	struct filter_stream input;	/* Input stream */
	struct filter_stream output;	/* Output stream */
	int  pages;		/* number of pages previously printed, */
                                /* goes to af */
	int  *sync;		/* syncronization address */
};


/*
 * pr(1) settings
 */
#define PR_TOP_MARGIN      5
#define PR_BOTTOM_MARGIN   5

/*
 * Parameters for pr(1) filter
 */
struct pr_parms {
	FILE *file_fp;		/* input file */
	int  out_s;		/* output socket */
	char *title;		/* title to print in header */
	int  length;		/* length of page (D_LENGTH) */
	int  width;		/* width of page (D_WIDTH) */
};

