/*
   mail.c : Copyright Paul Healy, EI9GL, 1993.

   Todo: add the mail/bm style escape sequences some time

   920103 : Created.
*/
#include <stdio.h>
#include <string.h>
#include "config.h"
#include "help.h"
#include "rc.h"
#include "getopts.h"
#include "buf.h"
#include "mailfile.h"
#include "header.h"
#include "send.h"

#ifdef BMH
#define main mail_main
#endif

static int
prompt(char to[], Header_t *h)
{
   printf("To: ");
   if (fgets(to, HDRMAXLEN, stdin) == NULL) {
      fprintf(stderr, "mail: no recipient, send aborted\n");
      return -1;
      }
   if (HEADERP(h, CC) == NULL) {
      char cc[100];
      printf("Cc: ");
      fgets(cc, sizeof(cc), stdin);
      hdr_add(h, CC, cc);
      }
   if (HEADERP(h, SUBJECT) == NULL) {
      char subj[100];
      printf("Subject: ");
      fgets(subj, sizeof(subj), stdin);
      hdr_add(h, SUBJECT, subj);
      }
   printf("-------\n");

   return 0;
}

static const char *optv[] = {
   "?|help",
   "s:subject",
   "f:from",
   "m:msgid",
   NULL
   };

static int
getmsg(int argc, char *argv[], Header_t *h)
{
   char to[HDRMAXLEN];
   int optchar;
   char *optarg;
   options_t *opts;
   int errors = 0;

   hdr_init(h);

   opts = optinit(*argv, optv);
   optctrl(opts, opt_LONG_ONLY);

   --argc, ++argv;
   while( (optchar = getopts(opts, argc, argv, &optarg)) != 0 )
      switch(optchar) {
         case '?' :
         case 'H' :
            ++errors;
            break;
         case 's' : hdr_add(h, SUBJECT, optarg);  break;
         case 'f' : hdr_add(h, FROM, optarg);     break;
         case 'm' : hdr_add(h, MESSAGEID, optarg); break;
         default : ++errors; break;
         }

   if (errors) {
      optusage(opts, "error, type 'mail -help' for help\n", stderr);
      optfree(opts), opts = NULL;
      return -1;
      }

   to[0] = 0;

   if (getoptind(opts) >= argc) {
      if (prompt(to, h) == -1) {
         optfree(opts), opts = NULL;
         return -1;
         }
      }
   else {
      int i;
      int curtolen = 1;

      for (i=getoptind(opts); i<argc; i++) {
         curtolen += strlen(argv[i]) + 1; 
         if (curtolen > HDRMAXLEN) {
            fprintf(stderr, "mail: recipient list too long, no mail sent\n");
            optfree(opts), opts = NULL;
            return -1;
            }
         else {
            strcat(to, argv[i]);
            strcat(to, " ");
            }
         }
      }
   hdr_add(h, TO, to);

   optfree(opts), opts = NULL;

   return 0;
}

int
main(int argc, char *argv[])
{
   Header_t h;

   if (dohelp(argc, argv, "mail [-s subj] [-f from] [-m msgid] [recip1 ... recipN]"))
      return 0;

   if (getmsg(argc, argv, &h) == -1)
      return -1;
   else 
      return dosend(stdin, &h);
}
