/* There are only two functions in this mailbox code that depend on the
 * underlying protocol, namely mbx_getname() and dombchat(). All the other
 * functions can hopefully be used without modification on other stream
 * oriented protocols than AX.25 or NET/ROM.
 *
 * SM0RGV 890506, most work done previously by W9NK
 *
 *** Changed 900114 by KA9Q to use newline mapping features in stream socket
 *	interface code; everything here uses C eol convention (\n)
 *
 *	Numerous new commands and other changes by SM0RGV, 900120
 *
 * Gateway function now support outgoing connects with the user's call
 * with inverted ssid. Users can connect to system alias as well...
 * See also several mods in socket.c,ax25.c and others
 * 11/15/91, WG7J/PA3DIS
 * Additional privileges, such as send, read,is_bbs added 26/1/92 WG7J
 *
 * Mods by PA0GRI
 */
#include <stdio.h>
#include <time.h>
#include <ctype.h>
#ifdef	UNIX
#include <sys/types.h>
#include <sys/stat.h>
#endif
#include "global.h"
#include "config.h"
#include "timer.h"
#include "proc.h"
#include "socket.h"
#include "usock.h"
#include "session.h"
#include "smtp.h"
#include "dirutil.h"
#include "telnet.h"
#include "ftp.h"
#include "ftpserv.h"
#include "commands.h"
#include "netuser.h"
#include "files.h"
#include "bm.h"
#include "pktdrvr.h"
#include "ax25.h"
#include "mailbox.h"
#include "ax25mail.h"
#include "nr4mail.h"
#include "cmdparse.h"

/*
#define MBDEBUG
*/

struct mbx *Mbox[NUMMBX];
static char *MMotd = NULLCHAR;
static int MAttended = TRUE;	/* default to attended mode */
unsigned Maxlet = BM_NLET;
#define	MAXPWDLEN 30
static char *Mbpasswd = NULLCHAR;
#ifdef	MBFWD
char *Mbzip = NULLCHAR;
char *Mbqth = NULLCHAR;
char *Mbhaddress = NULLCHAR;
char *Mbfwdinfo = NULLCHAR;
int Mbsmtptoo = 0;
int Mtrace = 0;
int UtcOffset = 0;
#endif
int MBSecure = 0;
char Mbnrid[20];		/* 6 + ':' + 10 + "} " + 1 */

char Noperm[] = "Permission denied.\n";
char Nosock[] = "Can't create socket\n";
extern int Nokeys;		/* set when shelled out */
extern char TooManyDigis[];

static char Mbbanner[] = "Welcome %s,\nto the %s TCP/IP Mailbox (%s)\n%s%s\n";

static char Mbwarning[] = "Third Party mail not permitted.\n";

static char Mbmenu[] = "Current msg# %d :\n ?,A,B,C,D,E,F,H,I,J,K,L,M,N,O,P,R,S,T,U,V,W,X,Z >\n";

static char Longmenu[] = "(?)help    (A)rea  (B)ye    (C)onnect (D)ownload (E)scape  (F)inger\n"
			 "(H)elp     (I)nfo  (J)heard (K)ill    (L)ist     (M)busers (N)odes\n"
			 "(O)perator (P)orts (R)ead   (S)end    (T)elnet   (U)pload  (V)erbose\n"
			 "(W)hat     (X)pert (Z)ap\n";
static char Loginbanner[] = "\nKA9Q NOS (%s)\n\n";
static char Howtoend[] = "Terminate with /EX or ^Z in first column (^A aborts):\n";
static int doattend __ARGS((int argc,char *argv[],void *p));
static int domaxmsg __ARGS((int argc,char *argv[],void *p));
static int domotd __ARGS((int argc,char *argv[],void *p));
static int dotimeout __ARGS((int argc,char *argv[],void *p));
static int dombusers __ARGS((int argc,char *argv[],void *p));
static int dombsecure __ARGS((int argc,char *argv[],void *p));
static int doarea __ARGS((int argc,char *argv[],void *p));
static int dosetexpert __ARGS((int argc,char *argv[],void *p));
static int dojumpstart __ARGS((int argc,char *argv[],void *p));
static int dombnrid __ARGS((int argc,char *argv[],void *p));
static int dombpassword __ARGS((int argc,char *argv[],void *p));
#ifdef	MBFWD
static int dombtrace __ARGS((int argc,char *argv[],void *p));
static int dombzipcode __ARGS((int argc,char *argv[],void *p));
static int dombfwdinfo __ARGS((int argc,char *argv[],void *p));
static int dombqth __ARGS((int argc,char *argv[],void *p));
static int dombhaddress __ARGS((int argc,char *argv[],void *p));
static int dombsmtptoo __ARGS((int argc,char *argv[],void *p));
static int dombutc __ARGS((int argc,char *argv[],void *p));
#endif

int Mbexpert = 1;
int Mbjumpstart = 0;
int Usenrid = 0;

/* mbox subcommand table */
static struct cmds Mbtab[] = {
	"attend",	doattend,	0, 0, NULLCHAR,
	"expert",	dosetexpert,	0, 0, NULLCHAR,
#ifdef	MBFWD
	"fwdinfo",	dombfwdinfo,	0, 0, NULLCHAR,
	"haddress",	dombhaddress,	0, 0, NULLCHAR,
#endif
	"jumpstart",	dojumpstart,	0, 0, NULLCHAR,
#ifdef	AX25
#ifdef	MBFWD
	"kick",		dombkick,	0, 0, NULLCHAR,
#endif
#endif
	"maxmsg",	domaxmsg,	0, 0, NULLCHAR,
	"motd",		domotd,		0, 0, NULLCHAR,
	"nrid",		dombnrid,	0, 0, NULLCHAR,
	"password",	dombpassword,	0, 0, NULLCHAR,
#ifdef	MBFWD
	"qth",		dombqth,	0, 0, NULLCHAR,
#endif
	"secure",	dombsecure,	0, 0, NULLCHAR,
#ifdef	MBFWD
	"smtptoo",	dombsmtptoo,	0, 0, NULLCHAR,
#endif
	"status",	dombusers,	0, 0, NULLCHAR,
#ifdef	AX25
#ifdef	MBFWD
	"timer",	dombtimer,	0, 0, NULLCHAR,
#endif
#endif
 	"tiptimeout",	dotimeout,	0, 0, NULLCHAR,
#ifdef	MBFWD
	"trace",	dombtrace,	0, 0, NULLCHAR,
	"utc",		dombutc,	0, 0, NULLCHAR,
	"zipcode",	dombzipcode,	0, 0, NULLCHAR,
#endif
	NULLCHAR,
};

int
dombox(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	if(argc == 1)
		return dombusers(argc,argv,p);
	return subcmd(Mbtab,argc,argv,p);
}

static int
dosetexpert(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return setbool(&Mbexpert,"Expert user as default",argc,argv);
}

static int
dojumpstart(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return setbool(&Mbjumpstart,"Jump start on connect",argc,argv);
}

static int
dombpassword(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	int len;

	if(Command->input != Current->input)
		return 0;
	if(argc != 2){
		tprintf("Usage: mbox password \"<password>\"\n");
		return 0;
	}

	if((len = strlen(argv[1])) == 0){
		tprintf("pw = %s\n",Mbpasswd);
		return 0;		/* zero length, don't do anything */
	}

	if(len > MAXPWDLEN){
		tputs("Too long\n");
		return 0;
	}
	if(Mbpasswd != NULLCHAR)
		free(Mbpasswd);
	Mbpasswd = mallocw(len+1);
	strcpy(Mbpasswd,argv[1]);
	Mbpasswd[len] = '\0';
	return 0;
}

#ifdef	MBFWD
static int
dombsmtptoo(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return setbool(&Mbsmtptoo,"SMTP headers incl",argc,argv);
}

static int
dombzipcode(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	int len;

	if(argc > 2){
		tprintf("Usage: mbox zipcode <zip>\n");
		return 0;
	}
	if(argc == 1){
		if(Mbzip != NULLCHAR)
			tprintf("%s\n",Mbzip);
	} else {
#define	ZIPSIZE 7
		if((len = strlen(argv[1])) > ZIPSIZE){
			tprintf("zip > %d\n",ZIPSIZE);
			return 0;
		}
		if(Mbzip != NULLCHAR){
			free(Mbzip);
			Mbzip = NULLCHAR;
		}
		if(len){
			Mbzip = mallocw(len);
			strcpy(Mbzip,argv[1]);
		}
	}
	return 0;
}

static int
dombqth(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	int len;

	if(argc > 2){
		tprintf("Usage: mbox qth <qth>\n");
		return 0;
	}
	if(argc == 1){
		if(Mbqth != NULLCHAR)
			tprintf("%s\n",Mbqth);
	} else {
		if(Mbqth != NULLCHAR){
			free(Mbqth);
			Mbqth = NULLCHAR;
		}
		len = strlen(argv[1]);
		if(len){
			Mbqth = mallocw(len);
			strcpy(Mbqth,argv[1]);
		}
	}
	return 0;
}

static int
dombhaddress(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	int len;

	if(argc > 2){
		tprintf("Usage: mbox haddress <address>\n");
		return 0;
	}
	if(argc == 1){
		if(Mbhaddress != NULLCHAR)
			tprintf("%s\n",Mbhaddress);
	} else {
		if(Mbhaddress != NULLCHAR){
			free(Mbhaddress);
			Mbhaddress = NULLCHAR;
		}
		len = strlen(argv[1]);
		if(len){
			Mbhaddress = mallocw(len);
			strcpy(Mbhaddress,argv[1]);
			strupr(Mbhaddress);
		}
	}
	return 0;
}

static int
dombfwdinfo(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	int len;

	if(argc > 2){
		tprintf("Usage: mbox fwdinfo <fwdinfo>\n");
		return 0;
	}
	if(argc == 1){
		if(Mbfwdinfo != NULLCHAR)
			tprintf("%s\n",Mbfwdinfo);
	} else {
		if(Mbfwdinfo != NULLCHAR){
			free(Mbfwdinfo);
			Mbfwdinfo = NULLCHAR;
		}
		len = strlen(argv[1]);
		if(len){
			Mbfwdinfo = mallocw(len);
			strcpy(Mbfwdinfo,argv[1]);
		}
	}
	return 0;
}

static int
dombutc(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	if(argc < 2){
		tprintf("Local Time is UTC");
		if(UtcOffset == 0)
			tputc('\n');
		else {
			if(UtcOffset > 0)
				tputc('+');
			tprintf("%d\n",UtcOffset);
		}
		return 0;
	}
	UtcOffset = atoi(argv[1]);
	return 0;
}
#endif

static int
dombnrid(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return setbool(&Usenrid,"Netrom id",argc,argv);
}

/* if unattended mode is set, ax25, telnet and maybe other sessions will
 * be restricted.
 */
static int
doattend(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return setbool(&MAttended,"Mailbox Attended flag",argc,argv);
}

static int
domaxmsg(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return setuns(&Maxlet,"Maximum messages per area",argc,argv);
}

static int
domotd(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	if(argc > 2) {
		tprintf("Usage: mbox motd <message>\n");
		return 0;
	}

	if(argc < 2) {
		if(MMotd != NULLCHAR)
			tputs(MMotd);
	}
	else {
		if(MMotd != NULLCHAR){
			free(MMotd);
			MMotd = NULLCHAR;	/* reset the pointer */
		}
		if(!strlen(argv[1]))
			return 0;		/* clearing the buffer */
		MMotd = mallocw(strlen(argv[1])+5);/* allow for the EOL char */
		strcpy(MMotd, argv[1]);
		strcat(MMotd, "\n");		/* add the EOL char */
	}
	return 0;
}

int
dombusers(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	int i, len;
	char fsocket[MAXSOCKSIZE];
	char *cp, *cp1;
	char upl[40], down[40], tmp[AXBUF];
	struct usock *up, *up1;
	int s;

#ifdef	NETROM
	struct nrroute_tab *np;
	char temp[AXBUF], *cp2, *cp3;
	char incircuit[] = "Circuit  (%s%s%s %s)";
	char outcircuit[] = "Circuit  (%s%s%s)";
#endif

	m = (struct mbx *)p;
	tputs("Users:\n");
	for (i = 0; i < NUMMBX; i++){
		if((m = Mbox[i]) != NULLMBX){
			len = MAXSOCKSIZE;
			getpeername(m->user,fsocket,&len);
			cp = strdup(psocket(fsocket));

			upl[0] = '\0';
			down[0] = '\0';

			switch(m->family) {		/* UPLINK */
#ifdef AX25
			case AF_AX25:
				sprintf(upl,"Uplink  (%s)",cp);
				break;
#endif
#ifdef NETROM
			case AF_NETROM:
				if((cp1 = strchr(cp,' ')) != NULLCHAR)
					*cp1 = '\0';
				cp1 += 3;
				setcall(temp,cp1);
				np = find_nrroute(temp);
				cp2 = strdup(np->alias);
				if((cp3 = strchr(cp2,' ')) != NULLCHAR)
					*cp3 = '\0';
			/*
				if(*cp2 == '#' || *cp2 == '\0')
					sprintf(upl,incircuit,"","",cp1,cp);
				else
			 */
				sprintf(upl,incircuit,cp2,":",cp1,cp);
				free(cp2);
				break;
#endif
			case AF_INET:
				if((cp1 = strchr(cp,':')) != NULLCHAR)
					*cp1 = '\0';
				sprintf(upl,"Telnet  (%s @ %s)",m->name,cp);
				break;
			case AF_LOCAL:
				sprintf(upl,"Local   (%s)",Hostname);
				break;
			default:
				strcpy(upl,"Connect");
				break;
			}
			free(cp);
			tprintf("%-36s",upl);

			for(s = SOCKBASE; s < Nusock + SOCKBASE; s++) {
				if((up = itop(s)) == NULLUSOCK || s == m->user)
					continue;
				up1 = itop(m->user);
				if(up->owner == up1->owner) {
					getpeername(s,fsocket,&len);
					cp = strdup(psocket(fsocket));
					switch(up->type) {
						case TYPE_TCP:
							if((cp1 = strchr(cp,':')) != NULLCHAR)
								*cp1 = '\0';
							sprintf(down,"Telnet   (%s)",cp);
							break;
#ifdef AX25
						case TYPE_AX25I:
						 /*
							if((cp1 = strchr(cp,' ')) != NULLCHAR)
								*cp1 = '\0';
							/
							sprintf(down,"Downlink (%s)",cp);
							break;
#endif
#ifdef NETROM
						case TYPE_NETROML4:
							/*get rid of usercall*/
							if((cp1 = strchr(cp,' ')) != NULLCHAR)
								*cp1 = '\0';
							cp1 += 3;   /*get rid of ' @ '*/
							setcall(temp,cp1);	/*get node call*/
							np = find_nrroute(temp);/*find alias, if any*/
							cp2 = strdup(np->alias);
							if((cp3 = strchr(cp2,' ')) != NULLCHAR)
								*cp3 = '\0';
							/*
							memcpy(cp,m->call,AXALEN);
							cp[ALEN] ^= 0x1e;
							*/
						  /*
							if(*cp2 == '#' || *cp2 == '\0')
								sprintf(down,outcircuit,"","",cp1);
							else
						   */
							sprintf(down,outcircuit,cp2,":",cp1);
							free(cp2);
							break;
#endif
						case TYPE_LOCAL_STREAM:
						case TYPE_LOCAL_DGRAM:
							sprintf(down,"Local    (%s)",Hostname);
							break;
						default:
							strcpy(down,"Connect");
							break;
					}
					free(cp);
					tprintf("<-->  %s",down);
					break;
				}
			}
		tputc('\n');
		}
	}
	tputc('\n');
	return 0;
}

/*Enlighten them a bit!
 */
static void
dombconnecthelp() {
	tputs("Syntax: 'C <node>'        for NET/ROM connects\n"
	      "        'C <port> <call>' for AX.25 connects\n\n");
}

int
dombports(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct iface *ifp;

	tputs("Available ports:\n");
	for(ifp=Ifaces;ifp!=NULLIF;ifp=ifp->next)
		if(ifp->type == CL_AX25) {
			tprintf("%-7s",ifp->name);
			if(ifp->descr != NULLCHAR)
				tprintf(":  %s",ifp->descr);
			else
				tputc('\n');
		}
	tputc('\n');
	return 0;
}

static int
dotimeout(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return setuns(&Tiptimeout,"Tip connection timeout",argc,argv);
}

#ifdef	MBFWD
static int
dombtrace(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return setbool(&Mtrace,"Mailbox trace flag",argc,argv);
}
#endif

static int
dombsecure(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return setbool(&MBSecure,"Mailbox gateway secure flag",argc,argv);
}

/**********************************************************************/

void
listusers(s)
int s;
{
	int outsave;

	usprintf(s,"\nCurrent remote users:\n");
	outsave = Curproc->output;
	Curproc->output = s;
	dombusers(0,NULLCHARP,NULL);
	Curproc->output = outsave;
}
struct mbx *
newmbx()
{
	int i;
	struct mbx *m;

	for(i = 0; i < NUMMBX; i++){
		if(Mbox[i] == NULLMBX){
			m = Mbox[i] = (struct mbx *)callocw(1,sizeof(struct mbx));
			m->mbnum = i;
			return m;
		}
	}
	/* If we get here, there are no free mailbox sessions */
	return NULLMBX;
}

static int
mbx_getname(m)
struct mbx *m;
{
	char *cp;
	union sp sp;
	char tmp[MAXSOCKSIZE];
	char buf[MBXLINE];
	int len = MAXSOCKSIZE;
	int anony = 0;
	int oldmode;
	int founddigit = 0;

	sp.p = tmp;
	sp.sa->sa_family = AF_LOCAL;	/* default to AF_LOCAL */
	getpeername(m->user,tmp,&len);
	m->family = sp.sa->sa_family;
	m->path = mallocw(MBXLINE);
	/* This is one of the two parts of the mbox code that depends on the
	 * underlying protocol. We have to figure out the name of the
	 * calling station. This is only practical when AX.25 or NET/ROM is
	 * used. Telnet users have to identify themselves by a login procedure.
	 */
	switch(sp.sa->sa_family){
#ifdef	AX25
	case AF_NETROM:
	case AF_AX25:
		/* NETROM and AX25 socket address structures are "compatible" */
		/* Save users call, in case she/he wants to use a gateway */
		memcpy(m->call,sp.ax->ax25_addr,AXALEN);
		m->call[ALEN] &= 0xfc;	/* Make sure E-bit isn't set! */
		pax25(m->name,sp.ax->ax25_addr);
		cp = strchr(m->name,'-');
		if(cp != NULLCHAR)			/* get rid of SSID */
			*cp = '\0';
		/* SMTP wants the name to be in lower case */
		cp = m->name;
		while(*cp){
			if(isupper(*cp))
				*cp = tolower(*cp);
			++cp;
		}
		anony = 1;
		/* Try to find the privileges of this user from the userfile */
#ifdef notdef
		if((m->privs = userlogin(m->name,buf,&m->path,MBXLINE,&anony,0L)) == -1)
			if((m->privs = userlogin("bbs",buf,&m->path,MBXLINE,&anony,0L)) == -1)
				if((m->privs = userlogin("anonymous",buf,&m->path,MBXLINE,
					 &anony,0L)) == -1){
						m->privs = 0;
						free(m->path);
						m->path = NULLCHAR;
				}
#else
		if((m->privs = userlogin(m->name,buf,&m->path,MBXLINE,&anony,0L)) == -1){
			m->privs = 0;
			free(m->path);
			m->path = NULLCHAR;
		}
#endif
		if(m->privs & EXCLUDED_CMD)
			return -1;
		if((cp = strchr(m->path,';')) != 0)
			*cp = '\0';	/* use only first directory */
		return 0;
#endif
	case AF_LOCAL:
	case AF_INET:
		m->state = MBX_LOGIN;
		tprintf(Loginbanner,Hostname);
		for(;;){
			oldmode = sockmode(m->user,SOCK_ASCII);
			tputs("Login: ");
			usflush(m->user);
			if(mbxrecvline(m->user,m->name,sizeof(m->name),-1) == EOF)
				return -1;
			if(*m->name == '\0')
				continue;
			if(sp_user(m->name)) { /* Special user, encrypted password */
				time(&m->ttim);
				tprintf("Challenge : %016.16lx --> ",m->ttim);
				usflush(m->user);
				if(mbxrecvline(m->user,buf,MBXLINE,-1) == EOF)
					return -1;
				anony = 0;
			} else { /* Normal User, so ask him for password */
				tprintf("Password: %c%c%c",IAC,WILL,TN_ECHO);
				usflush(m->user);
				sockmode(m->user,SOCK_BINARY);
				if(mbxrecvline(m->user,buf,MBXLINE,-1) == EOF)
					return -1;
				tprintf("%c%c%c",IAC,WONT,TN_ECHO);
				sockmode(m->user,oldmode);
				tputc('\n');
				usflush(m->user);
			}
			/* This is needed if the password was send before the
			 * telnet no-echo options were receied. We neeed to
			 * flush the eold sequence from the input buffers, sigh
			 */
			if(socklen(m->user,0))/* discard any remaining input */
				recv_mbuf(m->user,NULL,0,NULLCHAR,0);
			if(sp_user(m->name)){
				m->privs = userlogin(m->name,buf,&m->path,MBXLINE,&anony,m->ttim);
			} else {
				m->privs = userlogin(m->name,buf,&m->path,MBXLINE,&anony,0L);
			}
			if(m->privs != -1){
				if(anony)
					log(m->user,"MBOX Login: %s Password: %s",m->name,buf);
				else
					log(m->user,"MBOX Login: %s",m->name);
				if(m->privs & EXCLUDED_CMD)
					return -1;
/* try to set the name as the user-call. This is a very crude test.!
 * Login id must have at least 1 digit (0-9) in it,
 * and it must be possible to convert it to a callsign.
 * Be careful. If this doesn't work, disallow the gateway command.
 * Someone with the name johan and permissions set to allow gateway/netrom
 * will go out as johan-15 or johan-0 !!!!  WG7J/PA3DIS
 */
				for(cp = m->name;*cp != '\0';cp++)
					if(isdigit((int)*cp))
						break;
				if(*cp != '\0')
					founddigit = 1;
				if((setcall(m->call,m->name) == -1) ||
				   (!founddigit)){
					m->privs &= ~AX25_CMD;
					m->privs &= ~NETROM_CMD;
				}
				if((cp = strchr(m->path,';')) != 0)
					*cp = '\0';
				return 0;
			}
			tprintf("Login incorrect\n");
			*m->name = '\0';	/* wipe any garbage */
		}
	}
	return 0;
}

/* Give the prompt */
void
putprompt(m)
struct mbx *m;
{
	if(m->sid &MBX_SID)
		tputs(">\n");
	else {
		if(Usenrid){
			tputs(Mbnrid);
			tprintf((m->sid & MBX_EXPERT) ? "" : Mbmenu, m->current);
		} else {
			tprintf((m->sid & MBX_EXPERT) ? "> " : Mbmenu, m->current);
		}
	}
}

/* Incoming mailbox session */
void
mbx_incom(s,t,p)
int s;
void *t;
void *p;
{
	struct mbx *m;
	struct usock *up;
	char *buf[3];
	int rval;

	sockmode(s,SOCK_ASCII);
	sockowner(s,Curproc);	/* We own it now */
	/* Secede from the parent's sockets, and use the network socket that
	 * was passed to us for both input and output. The reference
	 * count on this socket will still be 1; this allows the domboxbye()
	 * command to work by closing that socket with a single call.
	 * If we return, the socket will be closed automatically.
	 */
	close_s(Curproc->output);
	close_s(Curproc->input);
	Curproc->output = Curproc->input = s;

	/* We'll do our own flushing right before we read input */
	setflush(s,-1);

	log(s,"open MBOX");
	if((m = newmbx()) == NULLMBX){
		tputs(TooManySessions);
		return;
	}
	m->user = s;
	m->escape = 24;		/* default escape character is Ctrl-X */
	m->type = (int) t;
	/* get the name of the remote station */

	if(mbx_getname(m) == -1) {
		exitbbs(m);
		return;
	}

	if(Mbexpert)		/* Expert status is default */
		m->sid |= MBX_EXPERT;
	m->state = MBX_CMD;	/* start in command state */

	/* Flush any remaining input from AX.25 users */
	if((m->family == AF_AX25) && (socklen(m->user,0) != 0))
		recv_mbuf(m->user,NULL,0,NULLCHAR,0);

	/* Now say hi */
	tputs("[NOS-H$]\n");
	if(m->privs & IS_BBS)
		m->sid ^= IS_BBS;	
	else {
		
/*		if(m->family == AF_INET) */
			tprintf(Mbbanner,m->name,Hostname,Version,
				ThirdParty ? "" : Mbwarning,
				(MMotd != NULLCHAR) ? MMotd : "");
/*		else
			tprintf("(%s)\n\n%s\n",Version,
				(MMotd != NULLCHAR) ? MMotd :""); */
	}

	/* Enable our local message area */
	buf[1] = m->name;
	doarea(2,buf,m);

	/* send the prompt */
	putprompt(m);

	while(mbxrecvline(s,m->line,MBXLINE,-1) != EOF){
		if((rval = mbx_parse(m)) == -2)
			break;
		if(rval == 1)
			tprintf("Bad syntax.\n");
		if(!(m->sid & MBX_SID) && isnewprivmail(m) > 0L)
			tprintf("You have new mail.\n");
		scanmail(m);
		putprompt(m);
		m->state = MBX_CMD;
	}
	exitbbs(m);
	/* nasty hack! we may have screwed up reference count */
	/* by invoking newproc("smtp_send",....); Fudge it!   */
	if((up = itop(Curproc->output)) != NULLUSOCK)
		up->refcnt = 1;
	close_s(Curproc->output);
}

void
exitbbs(m)
struct mbx *m;
{
	closenotes(m);
	free(m->to);
	free(m->tofrom);
	free(m->origto);
	free(m->tomsgid);
	free(m->path);
	if(m->stdinbuf != NULLCHAR)
		free(m->stdinbuf);
	if(m->stdoutbuf != NULLCHAR)
		free(m->stdoutbuf);
	free((char *)m->mbox);
	Mbox[m->mbnum] = NULLMBX;
	free((char *)m);
}

/**********************************************************************/

static int dombchat __ARGS((int argc,char *argv[],void *p));
static int dombconnect __ARGS((int argc,char *argv[],void *p));
static int dodownload __ARGS((int argc,char *argv[],void *p));
static int dombupload __ARGS((int argc,char *argv[],void *p));
static int dowhat __ARGS((int argc,char *argv[],void *p));
static int dozap __ARGS((int argc,char *argv[],void *p));
static int dosend __ARGS((int argc,char *argv[],void *p));
static int dosid __ARGS((int argc,char *argv[],void *p));
static int dosysop __ARGS((int argc,char *argv[],void *p));
static int dostars __ARGS((int argc,char *argv[],void *p));
static int dombnrnodes __ARGS((int argc,char *argv[],void *p));
static int dombhelp __ARGS((int argc,char *argv[],void *p));
static int dombtelnet __ARGS((int argc,char *argv[],void *p));
static int dombfinger __ARGS((int argc,char *argv[],void *p));
static int dombexpert __ARGS((int argc,char *argv[],void *p));
static void gw_alarm __ARGS((void *p));
static void gw_input __ARGS((int s,void *notused,void *p));
static void gw_superv __ARGS((int null,void *proc,void *p));
static int mbx_to __ARGS((int argc,char *argv[],void *p));
static int mbx_data __ARGS((struct mbx *m,struct list *cclist,char *extra));
static int msgidcheck __ARGS((char *string));
static int uuencode __ARGS((FILE *infile,int s,char *infilename));
static int thirdparty __ARGS((struct mbx *m));

static struct cmds Mbcmds[] = {
	"",		doreadnext,	0, 0, NULLCHAR,
	"?",		dombhelp,	0, 0, NULLCHAR,
	"area",		doarea,		0, 0, NULLCHAR,
	"bye",		domboxbye,	0, 0, NULLCHAR,
	"connect",	dombconnect,	0, 0, NULLCHAR,
	"download",	dodownload,	0, 2, "D[U] filename",
	"escape",	dombescape,	0, 0, NULLCHAR,
	"finger",	dombfinger,	0, 0, NULLCHAR,
	"help",		dombhelp,	0, 0, NULLCHAR,
	"info",		dombhelp,	0, 0, NULLCHAR,
#ifdef	AX25
	"jheard",	doaxheard,	0, 0, NULLCHAR,
#endif
	"kill",		dodelmsg,	0, 2, "K numbers",
	"list",		dolistnotes,	0, 0, NULLCHAR,
	"musers",	dombusers,	0, 0, NULLCHAR,
	"nodes",	dombnrnodes,	0, 0, NULLCHAR,
	"operator",	dombchat,	0, 0, NULLCHAR,
	"ports",	dombports,	0, 0, NULLCHAR,
	"read",		doreadmsg,	0, 2, "R numbers",
	"send",		dosend,		0, 0, NULLCHAR,
	"telnet",	dombtelnet,	0, 2, "T hostname",
	"upload",	dombupload,	0, 0, NULLCHAR,
	"verbose",	doreadmsg,	0, 2, "V numbers",
	"what",		dowhat,		0, 0, NULLCHAR,
	"xpert",	dombexpert,	0, 0, NULLCHAR,
	"zap",		dozap,		0, 2, "Z filename",
	"[",		dosid,		0, 0, NULLCHAR,
#ifdef	AX25
#ifdef	MBFWD
	"f>",		dorevfwd,	0, 0, NULLCHAR,
#endif
#endif
	"@",		dosysop,	0, 0, NULLCHAR,
	"***",		dostars,	0, 0, NULLCHAR,
	NULLCHAR,	NULLFP,		0, 0, "Huh?",
};

/* "twocmds" defines the MBL/RLI two-letter commands, eg. "SB", "SP" and so on.
 * They have to be treated specially since cmdparse() wants a space between
 * the actual command and its arguments.
 * "SP FOO" is converted to "s  foo" and the second command letter is saved
 * in m->stype. Longer commands like "SEND" are unaffected, except for
 * commands starting with "[", i.e. the SID, since we don't know what it will
 * look like.
 */
static char twocmds[] = "slrd[";	/* S,L,R,D are two-letter commands */
int
mbx_parse(m)
struct mbx *m;
{
	char *cp;
	int i;
	char *newargv[2];
	/* Translate entire buffer to lower case */
	for (cp = m->line; *cp != '\0'; ++cp)
		if(isupper(*cp))
			*cp = tolower(*cp);
	/* Skip any spaces at the begining */
	for(cp = m->line;isspace(*cp);++cp)
		;
	m->stype = ' ';
	if(*cp != '\0' && *(cp+1) != '\0')
	for(i=0; i<strlen(twocmds); ++i){
		if(*cp == twocmds[i] && (isspace(*(cp+2)) || *(cp+2) == '\0'
		 || *cp == '[')){
			if(islower(*(++cp)))
				m->stype = toupper(*cp); /* Save the second character */
			else
				m->stype = *cp;
			*cp = ' ';
			break;
		}
	}
	/* See if the input line consists solely of digits */
	cp = m->line;
	for(cp = m->line;isspace(*cp);++cp)
		;
	newargv[1] = cp;
	for(;*cp != '\0' && isdigit(*cp);++cp)
		;
	if(*cp == '\0' && strlen(newargv[1]) > 0) {
		newargv[0] = "read";
		return doreadmsg(2,newargv,(void *)m);
	}
	else
		return cmdparse(Mbcmds,m->line,(void *)m);
}

/* This works like recvline(), but telnet options are answered and the
 * terminating newline character is not put into the buffer. If the
 * incoming character equals the value of escape, any queued input is
 * flushed and -2 returned.
 */
int
mbxrecvline(s,buf,len,escape)
int s;
char *buf;
int len;
int escape;
{
	int c, cnt = 0, opt;
	if(buf == NULLCHAR)
		return 0;
	usflush(Curproc->output);
	while((c = recvchar(s)) != EOF){
		if(c == IAC){		/* Telnet command escape */
			if((c = recvchar(s)) == EOF)
				break;
			if(c > 250 && c < 255 && (opt = recvchar(s)) != EOF){
#ifdef	foo
				switch(c){
				case WILL:
					tprintf("%c%c%c",IAC,DONT,opt);
					break;
				case WONT:
					tprintf("%c%c%c",IAC,DONT,opt);
					break;
				case DO:
					tprintf("%c%c%c",IAC,WONT,opt);
					break;
				case DONT:
					tprintf("%c%c%c",IAC,WONT,opt);
				}
#endif
/* to be fixed 			usflush(Curproc->output);*/
				continue;
			}
			if(c != IAC && (c = recvchar(s)) == EOF)
				break;
		}
		/* ordinary character */
		if(c == '\r' || c == '\n')
			break;
		if(uchar(c) == escape){
			if(socklen(s,0)) /* discard any remaining input */
				recv_mbuf(s,NULL,0,NULLCHAR,0);
			cnt = -2;
			break;
		}
		*buf++ = c;
		++cnt;
		if(cnt == len - 1)
			break;
	}
	if(c == EOF && cnt == 0)
		return -1;
	*buf = '\0';
	return cnt;
}

int
domboxbye(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;

	m = (struct mbx *)p;
	/* Now say goodbye */
	tprintf("Thank you %s ,for calling the %s NOS Tcp/Ip Mailbox.\n",m->name,
		Hostname);
	if(m->type == TIP)
		tprintf("Please hang up now.\n");

	return -2;	/* signal that exitbbs() should be called */
}

static int
dombhelp(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	char buf[255];
	int i;
	FILE *fp;
	if(*argv[0] == '?') {
		tputs(Longmenu);
		return 0;
	}
	buf[0] = '\0';
	if(argc > 1)
		for(i=0; Mbcmds[i].name != NULLCHAR; ++i)
			if(!strncmp(Mbcmds[i].name,argv[1],strlen(argv[1]))) {
				sprintf(buf,"%s/%s.hlp",Helpdir,Mbcmds[i].name);
				break;
			}
	if(buf[0] == '\0')
		if(*argv[0] == 'i'){			/* INFO command */
			tprintf("KA9Q NOS ver. %s\n",Version);
			sprintf(buf,"%s/info.hlp",Helpdir);
		} else
			sprintf(buf,"%s/help.hlp",Helpdir);
	if((fp = fopen(buf,READ_TEXT)) != NULLFILE) {
		sendfile(fp,Curproc->output,ASCII_TYPE,0);
		fclose(fp);
	} else
		tprintf("No help available. (%s not found)\n",buf);
	return 0;
}

static int
dombexpert(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;

	m = (struct mbx *)p;
	m->sid ^= MBX_EXPERT;
	return 0;
}

static int
dombchat(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	char buf[8], *newargv[3];

	/* test for shelled out */
	if(MAttended && !Nokeys){
		newargv[0] = "telnet";
		newargv[1] = Hostname;
		sprintf(buf,"%d",IPPORT_TTYLINK);
		newargv[2] = buf;
		return dombtelnet(3,newargv,p);
	}
	else {
		tprintf("Sorry - the system is unattended.\007\n");
	}
	/* It returns only after a disconnect or refusal */
	return 0;
}

static int
dosend(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	int cccnt = 0, fail = 0;
	char *host, *cp, fullfrom[MBXLINE], sigwork[LINELEN], *rhdr = NULLCHAR;
	struct list *ap, *cclist = NULLLIST;
	struct mbx *m;
	FILE *fp;

	m = (struct mbx *)p;
	if(MBSecure)
		if((m->family != AF_AX25) && (m->family != AF_NETROM)) {
			tputs("Not allowed.\n");
			return 0;
		}

	if((m->stype != 'R' || (m->sid & MBX_SID)) && mbx_to(argc,argv,m)
	   == -1){
		if(m->sid & MBX_SID)
			tprintf("NO - syntax error\n");
		else {
			tprintf("S command syntax error - format is:\n");
			tprintf("  S[F] name [@ host] [< from_addr] [$bulletin_id]\n");
			tprintf("  SR [number]\n");
		}
		return 0;
	}
	/* check for permission - WG7J */
	if(m->privs & NO_SENDCMD) {
 		if(stricmp(m->to,"sysop")){
			tputs("Sorry, to 'sysop' only\n");
			return 0;
		}
	}
	if(m->stype != 'R' && msgidcheck(m->tomsgid)) {
		if(m->sid & MBX_SID)
			tputs("NO - ");
		tprintf("Already have %s\n",m->tomsgid);
		return 0;
	}
	if(m->stype == 'R' && !(m->sid & MBX_SID) &&
	   mbx_reply(argc,argv,m,&cclist,&rhdr) == -1)
		return 0;
	if((cp = rewrite_address(m->to)) != NULLCHAR)
		 if(strcmp(m->to,cp) != 0){
			m->origto = m->to;
			m->to = cp;
		 } else
			free(cp);

	if(!ThirdParty && !(m->privs & SYSOP_CMD) || (m->privs & NO_3PARTY))
		if(thirdparty(m) == 0){
			tputs(Mbwarning);
			return 0;
		}

	if((m->origto != NULLCHAR || m->stype == 'R') && !(m->sid & MBX_SID))
		tprintf("To: %s\n", m->to);
	if(validate_address(m->to) == 0){
		if(m->sid & MBX_SID)
			tprintf("NO - bad address\n");
		else
			tprintf("Bad user or host name\n");
		free(rhdr);
		del_list(cclist);
		/* We don't free any more buffers here. They are freed upon
		 * the next call to mbx_to() or to domboxbye()
		 */
		return 0;
	}
	/* Display the Cc: line (during SR command) */
	for(ap = cclist; ap != NULLLIST; ap = ap->next) {
		if(cccnt == 0){
			tprintf("%s",Hdrs[CC]);
			cccnt = 4;
		}
		else {
			tputs(", ");
			cccnt += 2;
		}
		if(cccnt + strlen(ap->val) > 80 - 3) {
			tputs("\n    ");
			cccnt = 4;
		}
		tputs(ap->val);
		cccnt += strlen(ap->val);
	}
	if(cccnt)
		tputc('\n');
	m->state = MBX_SUBJ;
	if(m->stype != 'R' || (m->sid & MBX_SID) != 0) {
		tprintf((m->sid & MBX_SID) ? "OK\n" : "Subject: ");
		if(mbxrecvline(m->user,m->line,MBXLINE,-1) == -1)
			return 0;
	}
	else				/* Replying to a message */
		tprintf("Subject: %s\n",m->line);
	if(mbx_data(m,cclist,rhdr) == -1){
		free(rhdr);
		del_list(cclist);
		tputs("Can't create temp file for mail\n");
		return 0;
	}
	free(rhdr);
	m->state = MBX_DATA;
	if((m->sid & MBX_SID) == 0 && m->stype != 'F')
		tprintf("Enter message.  %s",Howtoend);

	if(m->stype != 'F' || (m->sid & MBX_SID) != 0)
		while(mbxrecvline(m->user,m->line,MBXLINE,-1) != -1){
			if(m->line[0] == 0x01){  /* CTRL-A */
				fclose(m->tfile);
				tputs("Aborted.\n");
				del_list(cclist);
				return 0;
			}
			if(m->line[0] != CTLZ && stricmp(m->line, "/ex"))
				fprintf(m->tfile,"%s\n",m->line);
			else
				break;	/* all done */
		}
	else {
		fprintf(m->tfile,"----- Forwarded message -----\n\n");
		msgtofile(m,m->current,m->tfile,0);
		fprintf(m->tfile,"----- End of forwarded message -----\n");
	}

	/* Insert customised signature if one is found */
	if(!(m->sid & MBX_SID)) {	/* not a forwarding BBS */
	     sprintf(sigwork,"%s/%s.sig",Signature,
		     m->tofrom ? m->tofrom : m->name);
	     if((fp = fopen(sigwork,READ_TEXT)) != NULLFILE){
		  while(fgets(sigwork,LINELEN,fp) != NULLCHAR)
			fputs(sigwork,m->tfile);
		  fclose(fp);
	     }
	}

	if((host = strrchr(m->to,'@')) == NULLCHAR) {
		host = Hostname;	/* use our hostname */
		if(m->origto != NULLCHAR) {
			/* rewrite_address() will be called again by our
			 * SMTP server, so revert to the original address.
			 */
		 	free(m->to);
			m->to = m->origto;
			m->origto = NULLCHAR;
		}
	}
	else
		host++;	/* use the host part of address */

	/* make up full from name for work file */
	if(m->tofrom != NULLCHAR)
		sprintf(fullfrom,"%s%%%s@%s",m->tofrom, m->name, Hostname);
	else
		sprintf(fullfrom,"%s@%s",m->name,Hostname);
	if(cclist != NULLLIST && stricmp(host,Hostname) != 0) {
		fseek(m->tfile,0L,0);	/* reset to beginning */
		fail = queuejob(m->tfile,Hostname,cclist,fullfrom);
		del_list(cclist);
		cclist = NULLLIST;
	}
	addlist(&cclist,m->to,0);
	fseek(m->tfile,0L,0);
	fail += queuejob(m->tfile,host,cclist,fullfrom);
	del_list(cclist);
	fclose(m->tfile);
	if(fail)
	     tputs("Couldn't queue message for delivery\n");
	else
	     if(m->tomsgid != NULLCHAR &&
		(fp = fopen(Historyfile,APPEND_TEXT)) != NULLFILE) {
		  fprintf(fp,"%s\n",m->tomsgid); /* Save BID in history file */
		  fclose(fp);
	     }
	smtptick(NULL);		/* wake SMTP to send that mail */
	return 0;
}

static int
dombnrnodes(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	if(argc < 2)
		return doroutedump();
	return dorouteinfo(argc,argv,p);
}

static int
dosid(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	char *cp;

	m = (struct mbx *)p;
	if(argc == 1)
		return 1;
	if(argv[1][strlen(argv[1]) - 1] != ']') /* must be an SID */
		return 1;
	m->sid = MBX_SID;
	/* Now check to see if this is an RLI board.
	 * As usual, Hank does it a bit differently from
	 * the rest of the world.
	 */
	if(m->stype == 'R' && strncmp(argv[1],"li",2) == 0)/* [RLI] at a minimum */
		m->sid |= MBX_RLI_SID;
	/* Check to see if the BBS supports a kludge called "hierarchical
	 * routing designators."
	 *
	 * No need to check for ']' -- it must be there or this is not
	 * a valid mbox id -- it is checked earlier (fix de OH3LKU)
	 */
	if((cp = strchr(argv[1],'-')) != NULLCHAR
	 && (cp=strchr(cp+1,'h')) != NULLCHAR
	 && strchr(cp+1,'$'))
		m->sid |= MBX_HIER_SID;
	return 0;
}

int
dombescape(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;

	m = (struct mbx *)p;
	if(argc < 2){
		tputs("The escape character is: ");
		if(m->escape < 32)
			tprintf("CTRL-%c\n",m->escape+'A'-1);
		else
			tprintf("'%c'\n",m->escape);
		return 0;
	}
	if(strlen(argv[1]) > 1)
		if(isdigit(*argv[1]))
			m->escape = (char) atoi(argv[1]);
		else
			return 1;
	else
		m->escape = *argv[1];
	return 0;
}

static int
dodownload(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	FILE *fp;
	char *file;

	m = (struct mbx *)p;
	file = pathname(m->path,argv[1]);
	if(!permcheck(m->path,m->privs,RETR_CMD,file)){
		tputs(Noperm);
		return 0;
	}
	if((fp = fopen(file,READ_TEXT)) == NULLFILE)
		tprintf("Can't open \"%s\": %s\n",file,sys_errlist[errno]);
	else
		if(m->stype == 'U'){			/* uuencode ? */
			fclose(fp);
			fp = fopen(file,READ_BINARY);	/* assume non-ascii */
			uuencode(fp,m->user,file);
		} else
			sendfile(fp,m->user,ASCII_TYPE,0);
	free(file);
	fclose(fp);
	return 0;
}

static int
dombupload(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	FILE *fp;
	char *file, buf[LINELEN];

	m = (struct mbx *)p;
	file = pathname(m->path,argv[1]);
	if(!permcheck(m->path,m->privs,STOR_CMD,file)){
		tputs(Noperm);
		return 0;
	}
	if((fp = fopen(file,WRITE_TEXT)) == NULLFILE){
		tprintf("Can't create \"%s\": %s\n",file,sys_errlist[errno]);
		free(file);
		return 0;
	}
	log(m->user,"MBOX upload: %s",file);
	tprintf("Send file,  %s",Howtoend);
	for(;;){
		if(mbxrecvline(m->user,buf,LINELEN,-1) == -1){
			unlink(file);
			break;
		}
		if(buf[0] == 0x01){  /* CTRL-A */
			unlink(file);
			tputs("Aborted.\n");
			break;
		}
		if(buf[0] == CTLZ || !stricmp("/ex",buf))
			break;
		fputs(buf,fp);
#if !defined(UNIX) && !defined(__TURBOC__) && !defined(AMIGA)
		/* Needed only if the OS uses a CR/LF
		 * convention and putc doesn't do
		 * an automatic translation
		 */
		if(putc('\r',fp) == EOF)
			break;
#endif
		if(putc('\n',fp) == EOF)
			break;
	}
	free(file);
	fclose(fp);
	return 0;
}

static int
dowhat(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	FILE *fp;
	char *file;

	m = (struct mbx *)p;
	if(argc < 2)
		file = strdup(m->path);
	else
		file = pathname(m->path,argv[1]);
	if(!permcheck(m->path,m->privs,RETR_CMD,file)){
		tputs(Noperm);
		return 0;
	}
	if((fp = dir(file,1)) == NULLFILE)
		tprintf("Can't read directory: \"%s\": %s\n",file,sys_errlist[errno]);
	else
		sendfile(fp,m->user,ASCII_TYPE,0);
	free(file);
	fclose(fp);
	return 0;
}

static int
dozap(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	char *file;

	m = (struct mbx *)p;
	file = pathname(m->path,argv[1]);
	if(!permcheck(m->path,m->privs,DELE_CMD,file)){
		tputs(Noperm);
		return 0;
	}
	if(unlink(file))
		tprintf("Zap failed: %s\n",sys_errlist[errno]);
	log(m->user,"MBOX Zap: %s",file);
	free(file);
	return 0;
}

/* password protection added - 920118, WG7J */
static int
dosysop(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	int c;
	int len,i,pwdc[5],valid = 0;
	char *cp;
	extern struct cmds Cmds[];

	m = (struct mbx *) p;
	if(!(m->privs & SYSOP_CMD)){
		tputs(Noperm);
		return 0;
	}

	tputs("\nType CTRL-X to return\n\nBe VERY careful!!\n");

	if(Mbpasswd != NULLCHAR){
		len = strlen(Mbpasswd);
		for(i = 0;i < 5;i++)
			tprintf("%d ",(pwdc[i] = random(len))); /* get random */
		tputc('\n');	
		while(1){
			c = mbxrecvline(Curproc->input,m->line,MBXLINE,m->escape);
			if(c == EOF || c == -2)
				return 0;
			if(*m->line == '\0')
				break;
			cp = m->line;
			for(i = 0; i < 5;i++)
				if(*cp++ != Mbpasswd[pwdc[i]])
					break;
			if(i == 5)
				valid = 1;
		}
		if(!valid)
			return 0;
	}

	for(;;){
		tputs("Net> ");
		usflush(Curproc->output);
		c = mbxrecvline(Curproc->input,m->line,MBXLINE,m->escape);
		if(c == EOF || c == -2)
			break;
		log(m->user,"MBOX sysop: %s",m->line);
		if(cmdparse(Cmds,m->line,NULL) == -2)
			break;
	}
	return 0;
}

/* Handle the "*** Done" command when reverse forwarding ends or the
 * "*** LINKED to" command.
 */
static int
dostars(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	int anony = 1;
	char *c;

	m = (struct mbx *)p;
	/* The "*** LINKED to" command is only allowed to stations with
	 * SYSOP privileges to prevent others from obtaining the same.
	 */
	if((m->privs & SYSOP_CMD) && argc == 4 && !strcmp(argv[1],"linked")) {
		strcpy(m->name,argv[3]);
		/* Try to find the privileges of this user from the userfile */
#ifdef notdef
		if((m->privs = userlogin(m->name,NULLCHAR,&m->path,MBXLINE,
					 &anony,0L)) == -1)
		     if((m->privs = userlogin("bbs",NULLCHAR,&m->path,
				      MBXLINE,&anony,0L)) == -1)
			  if((m->privs = userlogin("anonymous",NULLCHAR,
					   &m->path,MBXLINE,&anony,0L)) == -1){
						m->privs = 0;
						free(m->path);
						m->path = NULLCHAR;
			  }
#else
		if((m->privs = userlogin(m->name,NULLCHAR,&m->path,MBXLINE,
					 &anony,0L)) == -1){
				m->privs = 0;
				free(m->path);
				m->path = NULLCHAR;
		}
#endif
		if((c = strchr(m->path,';')) != 0)
			*c = '\0';	/* use only first directory */
		tprintf("Oh, hello %s.\n",m->name);
		if(m->privs & EXCLUDED_CMD)
			return domboxbye(0,NULLCHARP,p);
		changearea(m,m->name);
		return 0;
	}
	if(argc > 1 && (m->sid & MBX_SID))	/* "*** Done" or similar */
		return 2;
	return -1;
}

static int
doarea(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	char *cp, *area;
	FILE *fp;

	m = (struct mbx *) p;
	if(argc < 2){
		area = strdup(m->area);
		while((cp = strchr(area,'/')) != NULLCHAR)
			*cp = '.';
		tprintf("Current message area is: %s\n",area);
		free(area);

		tprintf("Available areas are:\n%-15s  Your private mail area\n",
		  m->name);
		if((fp = fopen(Arealist,READ_TEXT)) == NULLFILE)
			return 0;
		sendfile(fp,m->user,ASCII_TYPE,0);
		fclose(fp);
		return 0;
	}
	if((m->privs & SYSOP_CMD) || strcmp(m->name,argv[1]) == 0){
		changearea(m,argv[1]);
		if(m->nmsgs){
			if(!strcmp(m->name,m->area))
				tprintf("You have ");
			else {
				area = strdup(m->area);
				while((cp = strchr(area,'/')) != NULLCHAR)
					*cp = '.';
				tprintf("%s: ",area);
				free(area);
			}
			tprintf("%d message%s -  %d new.\n", m->nmsgs,
			  m->nmsgs == 1 ? " " : "s ", m->newmsgs);
		}
		return 0;
	}
	if(isarea(argv[1])) {
		changearea(m,argv[1]);
		area = strdup(m->area);
		while((cp = strchr(area,'/')) != NULLCHAR)
			*cp = '.';
		tprintf("%s: %d message%s.\n", area, m->nmsgs,
		  m->nmsgs == 1 ? "" : "s");
		free(area);
	}
	else
		tprintf("No such message area: %s\n",argv[1]);
	return 0;
}

/* subroutine to do the actual switch from one area to another */
void
changearea(m,area)
struct mbx *m;
char *area;
{
	char *cp;

	closenotes(m);
	m->nmsgs = m->newmsgs = m->current = 0;
	strcpy(m->area,area);
	while((cp = strchr(m->area,'.')) != NULLCHAR)
		*cp = '/';
	while((cp = strchr(m->area,'\\')) != NULLCHAR)
		*cp = '/';
	scanmail(m);
}

static int
dombtelnet(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	int s, len, i;
	char dsocket[MAXSOCKSIZE];
	struct sockaddr_in fsocket;

	m = (struct mbx *) p;
	fsocket.sin_family = AF_INET;
	if(argc < 3)
		fsocket.sin_port = IPPORT_TELNET;
	else
		fsocket.sin_port = atoi(argv[2]);

	if((fsocket.sin_addr.s_addr = resolve(argv[1])) == 0){
		tprintf(Badhost,argv[1]);
		return 0;
	}
	/* Only local telnets are are allowed to the unprivileged user */
	if(!(m->privs & TELNET_CMD) && !ismyaddr(fsocket.sin_addr.s_addr)){
		tprintf(Noperm);
		return 0;
	}
	if((s = socket(AF_INET,SOCK_STREAM,0)) == -1){
		tprintf(Nosock);
		return 0;
	}
	if(fsocket.sin_port == IPPORT_TTYLINK) {
		m->startmsg = mallocw(80);
		len = MAXSOCKSIZE;
		i = getpeername(m->user,dsocket,&len);
		sprintf(m->startmsg,"*** Incoming call from %s@%s ***\n",
			m->name,i != -1 ? psocket(dsocket): Hostname);
	}
	return gw_connect(m,s,(struct sockaddr *)&fsocket,SOCKSIZE);
}

static int
dombfinger(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	char *host, *user = NULLCHAR, buf[8], *newargv[3];

	if(argc > 2){
		tprintf("Usage: F user@host  or  F @host  or  F user.\n");
		return 0;
	}
	host = Hostname;
	if(argc == 2){
		if((host = strchr(argv[1], '@')) != NULLCHAR){
			*host = '\0';
			host++;
		} else
			host = Hostname;
		user = argv[1];
	}
	m = (struct mbx *) p;
	m->startmsg = mallocw(80);
	if(user != NULLCHAR)
		sprintf(m->startmsg,"%s\n",user);
	else
		strcpy(m->startmsg,"\n");
	newargv[0] = "telnet";
	newargv[1] = host;
	sprintf(buf,"%d",IPPORT_FINGER);
	newargv[2] = buf;
	return dombtelnet(3,newargv,p);
}

/* Generic mbox gateway code. It sends and frees the contents of m->startmsg
 * when the connection has been established unless it a null pointer.
 */
static int
gw_connect(m,s,fsocket,len)
struct mbx *m;
int s;
struct sockaddr *fsocket;
int len;
{
	int c;
	char *cp, *cp1;
	struct proc *child;
	struct gwalarm *gwa;
	char *node, *tocall, whereto[18], buf[80];
	char temp[AXBUF];
	struct nrroute_tab *rp;

	sockmode(s,SOCK_ASCII);
	child = newproc("gw supervisor",256,gw_superv,0,Curproc,m,0);
	tputs("Trying ...  The escape character is: ");
	if(m->escape < 32)
		tprintf("CTRL-%c\n",m->escape+'A'-1);
	else
		tprintf("'%c'\n",m->escape);
	usflush(Curproc->output);

	/*find out where we're going to*/
	tocall = strdup(psocket(fsocket));
	if((cp1 = strchr(tocall,' ')) != NULLCHAR)
		*cp1 = '\0';
	if(fsocket->sa_family == AF_NETROM) {
		/*find the node alias*/
		setcall(temp,tocall);
		rp = find_nrroute(temp);
		node = strdup(rp->alias);
		if((cp1 = strchr(node,' ')) != NULLCHAR)
			*cp1 = '\0';
		sprintf(whereto,"%s:%s",node,tocall);
		free(node);
	} else
		strcpy(whereto,tocall);
	free(tocall);

	if(connect(s,(char *)fsocket,len) == -1){
		if((cp = sockerr(s)) != NULLCHAR) {
			switch(cp[0]) {
			case 'R':
				sprintf(buf,"%susy from",
					(m->family == AF_NETROM)?"B":"*** b");
				break;
			case 'T':
				if(m->family != AF_NETROM) {
					sprintf(buf,"*** timeout with");
					break;
				}
			default:
				sprintf(buf,"%sailure with",
					(m->family == AF_NETROM)?"F":"*** f");
				break;
			}
			tprintf("%s%s %s\n\n",
				(m->family == AF_NETROM) ? Mbnrid : "",buf,whereto);
		}
		shutdown(s,2);  /* HB9RWM suggestion */
		close_s(s);
		killproc(child);
		return 0;
	}
	/* The user did not type the escape character */
	killproc(child);
	tprintf("%s%sonnected to %s.\n",
		(m->family == AF_NETROM) ? Mbnrid : "",
		(m->family == AF_NETROM) ? "C" : "***c",whereto);
	
	if(m->startmsg != NULLCHAR){
		usputs(s,m->startmsg);
		free(m->startmsg);
		m->startmsg = NULLCHAR;
	}
	/* Since NOS does not flush the output socket after a certain
	 * period of time, we have to arrange that ourselves.
	 */
	gwa = (struct gwalarm *) mallocw(sizeof(struct gwalarm));
	gwa->s1 = Curproc->output;
	gwa->s2 = s;
	set_timer(&gwa->t,2*1000L);
	gwa->t.func = gw_alarm;
	gwa->t.arg = (void *) gwa;
	start_timer(&gwa->t);
	/* Fork off the receive process */
	child = newproc("gw in",1024,gw_input,s,m,Curproc,0);
	
	for(;;){
		if((c = recvchar(Curproc->input)) == EOF)
			break;
		if(c == m->escape){
			if(socklen(Curproc->input,0))
				recv_mbuf(Curproc->input,NULL,0,NULLCHAR,0);
			break;
		}
		if(usputc(s,c) == EOF)
			break;
	}
	stop_timer(&gwa->t);
	free((char *)gwa);
	close_s(s);
	killproc(child); /* get rid of the receive process */
	if(m->family == AF_INET)
		tprintf("%c%c%c\n",IAC,WONT,TN_ECHO);
	return 0;
}

static void
gw_input(s,mbx,p)
int s;
void *mbx;
void *p;
{
	int c;
	char *cp, *cp1;
	struct proc *parent;
	struct mbx *m;

	parent = (struct proc *) p;
	m = (struct mbx *)mbx;
	cp1 = strdup(Mbnrid);
	if((cp = strchr(cp1,'}')) != NULLCHAR)
		*cp = '\0';
	strupr(cp1);
	while((c = recvchar(s)) != EOF)
		tputc(c);
	if((cp = sockerr(s)) != NULLCHAR && m->family != AF_NETROM) {
		switch(cp[0]) {
		case 'T':
			usprintf(m->user,"\n*** %s: Link failure",cp1);
			break;
		case 'R':
			usprintf(m->user,"*** DM received");
			break;
		}
	}
	usprintf(m->user,"\n%s%seconnected to %s\n\n",
		(m->family == AF_NETROM) ? Mbnrid : "",
		(m->family == AF_NETROM) ? "R" : "*** r",
		cp1);

	free(cp1);
	cp1 = NULLCHAR;

	/* Tell the parent that we are no longer connected */
	alert(parent,ENOTCONN);
	pwait(Curproc); /* Now wait to be killed */
}

/* Check if the escape character is typed while the parent process is busy
 * doing other things. 
 */
static void
gw_superv(null,proc,p)
int null;
void *proc;
void *p;
{
	struct proc *parent;
	struct mbx *m;
	int c;
	parent = (struct proc *) proc;
	m = (struct mbx *) p;
	while((c = recvchar(Curproc->input)) != EOF)
		if(c == m->escape){
			/* flush anything in the input queue */
			if(socklen(Curproc->input,0))
				recv_mbuf(Curproc->input,NULL,0,NULLCHAR,0);
			break;
		}
	alert(parent,EINTR);	 /* Tell the parent to quit */
	pwait(Curproc);		 /* Please kill me */
}

static void
gw_alarm(p)
void *p;
{
	struct gwalarm *gwa = (struct gwalarm *)p;
	char oldbl;
	struct usock *up;

	/* Flush sockets s1 and s2, but first make sure that the socket
	 * is set to non-blocking mode, to prevent the flush from blocking
	 * if the high water mark has been reached.
	 */
	if((up = itop(gwa->s1)) != NULLUSOCK) {
		oldbl = up->noblock;
		up->noblock = 1;
		usflush(gwa->s1);
		up->noblock = oldbl;
	}
	if((up = itop(gwa->s2)) != NULLUSOCK) {
		oldbl = up->noblock;
		up->noblock = 1;
		usflush(gwa->s2);
		up->noblock = oldbl;
	}
	start_timer(&gwa->t);
}

/* States for send line parser state machine */
#define		LOOK_FOR_USER		2
#define		IN_USER			3
#define		AFTER_USER		4
#define		LOOK_FOR_HOST		5
#define		IN_HOST			6
#define		AFTER_HOST		7
#define		LOOK_FOR_FROM		8
#define		IN_FROM			9
#define		AFTER_FROM		10
#define		LOOK_FOR_MSGID		11
#define		IN_MSGID		12
#define		FINAL_STATE		13
#define		ERROR_STATE		14

/* Prepare the addressee.  If the address is bad, return -1, otherwise
 * return 0
 */
static int
mbx_to(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	register char *cp;
	int state, i;
	char *user, *host, *from, *msgid;
	int userlen = 0, hostlen = 0, fromlen = 0, msgidlen = 0;
	struct mbx *m;

	m = (struct mbx *)p;
	/* Free anything that might be allocated
	 * since the last call to mbx_to() or mbx_reply()
	 */
	free(m->to);
	m->to = NULLCHAR;
	free(m->tofrom);
	m->tofrom = NULLCHAR;
	free(m->tomsgid);
	m->tomsgid = NULLCHAR;
	free(m->origto);
	m->origto = NULLCHAR;

	if(argc == 1)
		return -1;
	i = 1;
	cp = argv[i];
	state = LOOK_FOR_USER;
	while(state < FINAL_STATE){
#ifdef MBDEBUG
		tprintf("State is %d, char is %c\n", state, *cp);
#endif
		switch(state){
		case LOOK_FOR_USER:
			if(*cp == '@' || *cp == '<' || *cp == '$'){
				state = ERROR_STATE;		/* no user */
			} else {
				user = cp;			/* point at start */
				userlen++;			/* start counting */
				state = IN_USER;
			}
			break;
		case IN_USER:
			switch(*cp){
			case '\0':
				state = AFTER_USER;		/* done with username */
				break;
			case '@':
				state = LOOK_FOR_HOST;		/* hostname should follow */
				break;
			case '<':
				state = LOOK_FOR_FROM;		/* from name should follow */
				break;
			case '$':
				state = LOOK_FOR_MSGID;	/* message id should follow */
				break;
			default:
				userlen++;			/* part of username */
			}
			break;
		case AFTER_USER:
			switch(*cp){
			case '@':
				state = LOOK_FOR_HOST;		/* hostname follows */
				break;
			case '<':
				state = LOOK_FOR_FROM;		/* fromname follows */
				break;
			case '$':
			state = LOOK_FOR_MSGID;	/* message id follows */
				break;
			default:
				state = ERROR_STATE;
			}
			break;
		case LOOK_FOR_HOST:
			if(*cp == '@' || *cp == '<' || *cp == '$'){
				state = ERROR_STATE;
				break;
			}
			if(*cp == '\0')
				break;
			host = cp;
			hostlen++;
			state = IN_HOST;
			break;
		case IN_HOST:
			switch(*cp){
			case '\0':
				state = AFTER_HOST;		/* found user@host */
				break;
			case '@':
				state = ERROR_STATE;		/* user@host@? */
				break;
			case '<':
				state = LOOK_FOR_FROM;		/* fromname follows */
				break;
			case '$':
				state = LOOK_FOR_MSGID;	/* message id follows */
				break;
			default:
				hostlen++;
			}
			break;
		case AFTER_HOST:
			switch(*cp){
			case '@':
				state = ERROR_STATE;		/* user@host @ */
				break;
			case '<':
				state = LOOK_FOR_FROM;		/* user@host < */
				break;
			case '$':
				state = LOOK_FOR_MSGID;	/* user@host $ */
				break;
			default:
				state = ERROR_STATE;		/* user@host foo */
			}
			break;
		case LOOK_FOR_FROM:
			if(*cp == '@' || *cp == '<' || *cp == '$'){
				state = ERROR_STATE;
				break;
			}
			if(*cp == '\0')
				break;
			from = cp;
			fromlen++;
			state = IN_FROM;
			break;
		case IN_FROM:
			switch(*cp){
			case '\0':
				state = AFTER_FROM;		/* user@host <foo */
				break;
			case '<':
				state = ERROR_STATE;		/* user@host <foo< */
				break;
			case '$':
				state = LOOK_FOR_MSGID;	/* message id follows */
				break;
			default:
				fromlen++;
			}
			break;
		case AFTER_FROM:
			switch(*cp){
			case '@':				/* user@host <foo @ */
			case '<':				/* user@host <foo < */
				state = ERROR_STATE;
				break;
			case '$':
				state = LOOK_FOR_MSGID;	/* user@host <foo $ */
				break;
			default:
				state = ERROR_STATE;		/* user@host foo */
			}
			break;
		case LOOK_FOR_MSGID:
			if(*cp == '\0')
				break;
			msgid = cp;
			msgidlen++;
			state = IN_MSGID;
			break;
		case IN_MSGID:
			if(*cp == '\0')
				state = FINAL_STATE;
			else
				msgidlen++;
			break;
		default:
			/* what are we doing in this state? */
			state = ERROR_STATE;
		}
		if(*(cp) == '\0'){
			++i;
			if(i < argc)
			cp = argv[i];
			else break;
		} else
			++cp;
	}
	if(state == ERROR_STATE || state == LOOK_FOR_HOST
	 || state == LOOK_FOR_FROM || state == LOOK_FOR_MSGID)
		return -1;		/* syntax error */

	m->to = mallocw(userlen + hostlen + 2);

	strncpy(m->to, user, userlen);
	m->to[userlen] = '\0';

	if(hostlen){
		m->to[userlen] = '@';
		strncpy(m->to + userlen + 1, host, hostlen);
		m->to[userlen + hostlen + 1] = '\0';
	}
	if(fromlen){
		m->tofrom = mallocw(fromlen + 1);
		strncpy(m->tofrom, from, fromlen);
		m->tofrom[fromlen] = '\0';
	}
	if(msgidlen){
		m->tomsgid = mallocw(msgidlen + 1);
		strncpy(m->tomsgid, msgid, msgidlen);
		m->tomsgid[msgidlen] = '\0';
	}
	return 0;
}

/* This opens the data file and writes the mail header into it.
 * Returns 0 if OK, and -1 if not.
 */
static int
mbx_data(m,cclist,extra)
struct mbx *m;
struct list *cclist;	/* list of carbon copy recipients */
char *extra;		/* optional extra header lines */
{
	time_t t;
	struct list *ap;
	int cccnt = 0;
	
	if((m->tfile = tmpfile()) == NULLFILE)
		return -1;
	time(&t);
#ifdef	notdef
	/* these 2 lines get added again by smtp-server */
	fprintf(m->tfile,Hdrs[RECEIVED]);
	if(m->tofrom != NULLCHAR)
		fprintf(m->tfile,"from %s ",m->name);
	fprintf(m->tfile,"by %s (%s)\n\tid AA%ld ; %s",
		Hostname, Version, get_msgid(), ptime(&t));
#endif
	fprintf(m->tfile,"%s%s",Hdrs[DATE],ptime(&t));
	fprintf(m->tfile,Hdrs[MSGID]);
	if(m->tomsgid)
		fprintf(m->tfile,"<%s@%s>\n", m->tomsgid, m->name);
	else
		fprintf(m->tfile,"<%ld@%s>\n",get_msgid(), Hostname);
	fprintf(m->tfile,Hdrs[FROM]);
	if(m->tofrom)
		fprintf(m->tfile,"%s%%%s@%s\n",
			m->tofrom, m->name, Hostname);
	else
		fprintf(m->tfile,"%s@%s\n", m->name, Hostname);
	fprintf(m->tfile,"%s%s\n",Hdrs[TO],m->origto != NULLCHAR ? m->origto : m->to);
	/* Write Cc: line */
	for(ap = cclist; ap != NULLLIST; ap = ap->next) {
		if(cccnt == 0){
			fprintf(m->tfile,"%s",Hdrs[CC]);
			cccnt = 4;
		}
		else {
		       fprintf(m->tfile,", ");
		       cccnt += 2;
		}
		if(cccnt + strlen(ap->val) > 80 - 3) {
		       fprintf(m->tfile,"\n    ");
		       cccnt = 4;
		}
		fputs(ap->val,m->tfile);
		cccnt += strlen(ap->val);
	}
	if(cccnt)
		fputc('\n',m->tfile);
	fprintf(m->tfile,"%s%s\n",Hdrs[SUBJECT],m->line);
	if(!isspace(m->stype) && ((m->stype != 'R' && m->stype != 'F') ||
	  (m->sid & MBX_SID) !=0))
		  fprintf(m->tfile,"%s%c\n", Hdrs[BBSTYPE],m->stype);
	if(extra != NULLCHAR)
		fprintf(m->tfile,extra);
	fprintf(m->tfile,"\n");

	return 0;
}

/* Returns true if string is in history file or if string appears to be a
 * message id generated by our system.
 */
static int
msgidcheck(string)
char *string;
{
	FILE *fp;
	char buf[LINELEN], *cp;
	if(string == NULLCHAR)
	return 0;
	/* BID's that we have generated ourselves are not kept in the history
	 * file. Such BID's are in the nnnn_hhhh form, where hhhh is a part of
	 * our hostname, truncated so that the BID is no longer than 11
	 * characters.
	 */
	if((cp = strchr(string,'_')) != NULLCHAR && *(cp+1) != '\0' && 
	strnicmp(cp+1,Hostname,strlen(cp+1)) == 0)
		return 1;

	if((fp = fopen(Historyfile,READ_TEXT)) == NULLFILE)
		return 0;
	while(fgets(buf,LINELEN,fp) != NULLCHAR) {
		rip(buf);
		if(stricmp(string,buf) == 0) {	/* found */
			fclose(fp);
			return 1;
		}
	}
	fclose(fp);
	return 0;
}

/* Read the rewrite file for lines where the first word is a regular
 * expression and the second word are rewriting rules. The special
 * character '$' followed by a digit denotes the string that matched
 * a '*' character. The '*' characters are numbered from 1 to 9.
 * Example: the line "*@*.* $2@$1.ampr.org" would rewrite the address
 * "foo@bar.xxx" to "bar@foo.ampr.org".
 * $H is replaced by our hostname, and $$ is an escaped $ character.
 * If the third word on the line has an 'r' character in it, the function
 * will recurse with the new address.
 */
char *
rewrite_address(addr)
char *addr;
{
	char *argv[10], buf[MBXLINE], *cp, *cp2, *retstr;
	int cnt;
	FILE *fp;
	if ((fp = fopen(Rewritefile,READ_TEXT)) == NULLFILE)
		return NULLCHAR;
	memset((char *)argv,0,10*sizeof(char *));
	while(fgets(buf,MBXLINE,fp) != NULLCHAR) {
		if(*buf == '#')		/* skip commented lines */
			continue;
		if((cp = strchr(buf,' ')) == NULLCHAR) /* get the first word */
			if((cp = strchr(buf,'\t')) == NULLCHAR)
				continue;
		*cp = '\0';
		if((cp2 = strchr(buf,'\t')) != NULLCHAR){
			*cp = ' ';
			cp = cp2;
			*cp = '\0';
		}
		if(!wildmat(addr,buf,argv))
			continue;		/* no match */
		rip(++cp);
		cp2 = retstr = (char *) callocw(1,MBXLINE);
		while(*cp != '\0' && *cp != ' ' && *cp != '\t')
			if(*cp == '$') {
				if(isdigit(*(++cp)))
					if(argv[*cp - '0'-1] != '\0')
						strcat(cp2,argv[*cp - '0'-1]);
				if(*cp == 'h' || *cp == 'H') /* Our hostname */
					strcat(cp2,Hostname);
				if(*cp == '$')	/* Escaped $ character */
					strcat(cp2,"$");
				cp2 = retstr + strlen(retstr);
				cp++;
			}
			else
				*cp2++ = *cp++;
		for(cnt=0; argv[cnt] != NULLCHAR; ++cnt)
			free(argv[cnt]);
		fclose(fp);
		/* If there remains an 'r' character on the line, repeat
		 * everything by recursing.
		 */
		if(strchr(cp,'r') != NULLCHAR || strchr(cp,'R') != NULLCHAR) {
			if((cp2 = rewrite_address(retstr)) != NULLCHAR) {
				free(retstr);
				return cp2;
			}
		}
		return retstr;
	}
	fclose(fp);
	return NULLCHAR;
}

/* uuencode a file -- translated from C++; both versions copyright 1990
   by David R. Evans, G4AMJ/NQ0I
*/

static int
uuencode(infile,s,infilename)
FILE *infile;
int s;			/* output socket */
char *infilename;
{
	int n_read_so_far = 0, n_written_so_far = 0, in_chars, n, mode = 0755;
	unsigned long cnt = 0;
	unsigned char in[3], out[4], line[100];
#ifdef UNIX
	struct stat stb;
  
	if(stat(infilename,&stb) != -1)
		 mode = stb.st_mode & 0777;	/* get real file protection mode */
#endif
	usprintf(s, "begin %03o %s\n", mode, infilename);

	/* do the encode */
	for(;;){	
		in_chars = fread(in, 1, 3, infile);
		out[0] = in[0] >> 2;
		out[1] = in[0] << 6;
		out[1] = out[1] >> 2;
		out[1] = out[1] | (in[1] >> 4);
		out[2] = in[1] << 4;
		out[2] = out[2] >> 2;
		out[2] = out[2] | (in[2] >> 6);
		out[3] = in[2] << 2;
		out[3] = out[3] >> 2;
		for (n = 0; n < 4; n++)
			out[n] += ' ';
		n_read_so_far += in_chars;
		for (n = 0; n < 4; n++)
			line[n_written_so_far++] = out[n];
		if (((in_chars != 3) || (n_written_so_far == 60)) && n_read_so_far > 0) {
			line[(n_read_so_far + 2) / 3 * 4] = '\0';

			usprintf(s,"%c%s\n",n_read_so_far + ' ', line);
			cnt += n_read_so_far;
			n_read_so_far = 0;
			n_written_so_far = 0;
		}
		if (in_chars == 0)
			break;
	}
	if (usprintf(s," \nend\nsize %lu\n", cnt) == EOF)
		return 1;
	return 0;
}

/* Attempt to determine if this is third-pary mail. */
static int 
thirdparty(m)
struct mbx *m;
{
	char buf[MBXLINE], *cp, *rp;
	FILE *fp;

	if(strchr(m->to,'@') != NULLCHAR || strchr(m->to,'%') != NULLCHAR
			|| strchr(m->to,'!') != NULLCHAR)
		return 0;

	rp = strdup(Hostname);

	if((cp = strchr(rp, '.')) != NULLCHAR)
		*cp = '\0';

	if(stricmp(m->to,rp) == 0){
		free(rp);
		return -1;
	}
	free(rp);

	if(stricmp(m->to,"sysop") == 0)
		return -1;

	if((fp = fopen(Arealist,READ_TEXT)) == NULLFILE)
		return 0;

	while(fgets(buf,MBXLINE,fp) != NULLCHAR){
		/* The first word on each line is all that matters */
		if((cp = strchr(buf, ' ')) != NULLCHAR)
			*cp = '\0';
		if((cp = strchr(buf,'\t')) != NULLCHAR)
			*cp = '\0';
		if(stricmp(m->to,buf) == 0){
			fclose(fp);
			return -1;
		}
	}
	fclose(fp);
	return 0;
}

static int
dombconnect(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	char buf[7];
	struct nrroute_tab *np;
	int ndigis,i,s;
	struct sockaddr_nr lsocket,fsocket;
	struct sockaddr_ax alsocket;    /*the local socket*/
	struct sockaddr_ax afsocket;    /*the remote socket*/
	struct iface *ifp;
	char alias[AXBUF];
	char local_call[AXALEN];
	char digis[MAXDIGIS][AXALEN];
	char target[AXALEN];

	m = (struct mbx *) p;

	if(MBSecure)
		if((m->family != AF_AX25) && (m->family != AF_NETROM)) {
			tputs("Not allowed.\n");
			return 0;
		}

	if (argc == 1){ /* this was chat with sysop, now he doesn't know!*/
		dombconnecthelp();
		return 0;
	}

	if (argc == 2) {
	/*NETROM connection wanted*/
		if(!(m->privs & NETROM_CMD)){
			tputs(Noperm);
			return 0;
		}
		if(Nrifaces[0].iface == NULLIF){
			tputs("NET/ROM not activated.\n\n");
			return 0;
		}
		/* See if the requested destination is a known alias or call,
		 * use it if it is.  Otherwize give an error message.
		 */
		putalias(alias,argv[1],0);
		strupr(argv[1]);    /*make sure it's upper case*/
		if((np = find_nrboth(alias,argv[1])) == NULLNRRTAB){
			/*no such call or node alias*/
			tputs("no such node\n\n");
			dombconnecthelp();
			dombports(0,NULL,NULL);
			return 0;
		}

		if((s = socket(AF_NETROM,SOCK_SEQPACKET,0)) == -1){
			tputs(Nosock);
			return 0;
		}
		lsocket.nr_family = AF_NETROM;

		/* Set up our local username, bind would use Mycall instead */
		memcpy(lsocket.nr_addr.user,m->call,AXALEN);
		/* Set up our source address */
		memcpy(lsocket.nr_addr.node,Nr_iface->hwaddr,AXALEN);

		bind(s,(char *)&lsocket,sizeof(struct sockaddr_nr));

		memcpy(fsocket.nr_addr.user,np->call,AXALEN);
		memcpy(fsocket.nr_addr.node,np->call,AXALEN);
		fsocket.nr_family = AF_NETROM;
		return gw_connect(m,s,(struct sockaddr *)&fsocket, sizeof(struct sockaddr_nr));
	}

	if(argc > 2) {
	/*AX25 gateway connection wanted*/
		if(!(m->privs & AX25_CMD)){
			tputs(Noperm);
			return 0;
		}
		if((ifp = if_lookup(argv[1])) == NULLIF) {
			tprintf("Unknown port %s\n",argv[1]);
			dombports(0,NULL,NULL);
			return 0;
		}
		if(ifp->type != CL_AX25){
			tprintf("Port %s not usable for AX.25 connects\n",argv[1]);
			dombports(0,NULL,NULL);
			return 0;
		}
		if(setcall(target,argv[2]) == -1){
			tprintf("Bad call %s\n",argv[2]);
			return 0;
		}
		/* If digipeaters are given, put them in the routing table */
		if(argc > 3){
			ndigis = argc - 3;
			if(ndigis > MAXDIGIS){
				tputs(TooManyDigis);
				return 0;
			}
			for(i=0;i<ndigis;i++){
				if(setcall(digis[i],argv[i+3]) == -1){
					tprintf("Bad digipeater %s\n",argv[i+3]);
					return 0;
				}
			}
			if(ax_add(target,AX_LOCAL,digis,ndigis) == NULLAXR){
				tputs("AX25 route add failed\n");
				return 0;
			}
		}
		if((s = socket(AF_AX25,SOCK_STREAM,0)) == -1){
			tputs(Nosock);
			return 0;
		}

		/*fill in the known stuff*/
		alsocket.sax_family = afsocket.sax_family = AF_AX25;

		/*the remote call to connect to*/
		setcall(afsocket.ax25_addr,argv[2]);

		/*the outgoing interface*/
		strncpy(afsocket.iface,argv[1],ILEN);

		/*now set local user call, invert ssid*/
		memcpy(local_call,m->call,AXALEN);
		local_call[AXALEN-1] ^= 0x1e;
		memcpy(alsocket.ax25_addr,local_call,AXALEN);
		/*and bind it (otherwize Mycall will be used!)*/
		bind(s,(char *)&alsocket,sizeof(struct sockaddr_ax));

		return gw_connect(m,s,(struct sockaddr *)&afsocket, sizeof(struct sockaddr_ax));
	}
	return 0;
}
