/*
   mailbuf.c: Copyright 1993, Paul Healy.

   930518 : Created
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "buf.h"
#include "mailfile.h"
#include "mailer.h"
#include "rc.h"
#include "pc.h"
#include "lock.h"
#include "buf.h"
#include "message.h"

#undef DEBUG

int
mail_refill(Mailfile *m)
{
   if (buf_eof(&m->buffer)) {
#ifdef DEBUG
      fprintf(stderr, "refill: eof set\n");
#endif
      return -1;
      }
   else if (mail_open(m) == -1) {
#ifdef DEBUG
      fprintf(stderr, "bmh: mail refill open failed\n");
#endif
      return -1;
      }
   else {
      int ret = buf_fill(m->fd, &m->buffer);
      if ( (mail_close(m) == -1) || (ret==-1) ) {
#ifdef DEBUG
         fprintf(stderr, "refill: failure (%d)\n", ret);
#endif
         return -1;
         }
      else
         return 0;
      }
}

int
mail_getmsg(Mailfile *m, char *line, int max)
{
   buf *bp = &m->buffer;

   line[0] = 0;

#ifdef DEBUG
   fprintf(stderr, "getmsg: called\n");
#endif

   if (buf_eof(bp))
      return -1;

   do
      {
      buf_findF(bp);
      if (buf_eob(bp)) {
#ifdef DEBUG
         fprintf(stderr, "getmsg: refill called\n");
#endif
         if (mail_refill(m) == 0)
            continue;
         else 
            return -1;
         }
      m->offset = buf_offset(&m->buffer, buf_curp(bp));
      }
   while ((mail_getstring(m, line, max)!=NULL) && !ismsg(line));

   return 0;
}

char *
mail_getstring(Mailfile *m, char *line, int max)
{
   char *s = line;
   buf *bp = &m->buffer;
   uchar *p;

   if (buf_eof(bp))
      return NULL;

   do {
      uchar *e = buf_endp(bp);
      p = buf_curp(bp);

      while ( (p < e) && ((*s++ = *p++) != '\n') && max--)
         ;

      if (*(s-1) == '\n')
         break;

      if (p>=e) {
         if (mail_refill(m) == 0)
            continue;
         else {
            buf_seteof(bp);
            *(s-1) = '\0';
            return line;
            }
         }
      }
   while ( (*(s-1) != '\n') && max);

#ifdef MSDOS
   if (*(s-2) == '\r') /* dispose of the extra msdos line marker */
      *(s-2) = '\0';
   else 
#endif
      *(s-1) = '\0';

   if (max == 0)
      fprintf(stderr, "bmh: warning line too long (%d max): \"%.35s ...\"\n", LINELEN, line);

   buf_setptr(bp, p);

   return line;
}

#ifdef DEBUG
   static int REFILL = 0;

   if (REFILL)
      fprintf(stderr, "gs: start p=%p, e=%p, max=%d\n", buf_curp(bp),  buf_endp(bp), max);
#endif

/*
      *s=0;
      fprintf(stderr, "gs: p=%p, e=%p, max=%d, s=%s\n", p, e, max, line);
*/
#ifdef DEBUG
         REFILL = 1;
         fflush(stdout);
         *s=0;
         fprintf(stderr, "gs: refill p=%p, e=%p, max=%d, s=%s\n", p, e, max, line);
#endif
#ifdef DEBUG
            fprintf(stderr, "gs: buf = <%.65s>\n", buf_curp(bp));
#endif
#ifdef DEBUG
   if (REFILL)
      fprintf(stderr, "gs: end p=%p, max=%d, s=%s\n", p, max, line);
#endif


int
mail_skipline(Mailfile *m)
{
   buf *bp = &m->buffer;
   uchar *p, *e;
   int ret;

   do {
      ret = 0;
      p = buf_curp(bp);
      e = buf_endp(bp);

      while ( (p < e) && (*p != '\n'))
         p++;   
      }
   while ( (p == e) && ((ret=mail_refill(m)) == 0) );

   buf_setptr(bp, p);

   return ret;
}

int
mail_curc(Mailfile *m)
{
   buf *bp = &m->buffer;

   if (!buf_eob(bp) || (mail_refill(m) == 0))
      return *(buf_curp(bp));
   else
      return EOF;
}

int
mail_lastc(Mailfile *m)
{
   return *(buf_curp(&m->buffer) - 1);
}

int
mail_getc(Mailfile *m)
{
   int ch = mail_curc(m);

   if (ch != EOF)
      buf_inc(&m->buffer);

   return ch;
}

/*
 * 0 => not from
 * 1 => is from
 * -1=> error
 */
static int
isfrom(Mailfile *m, FILE *fp, char *line)
{
   if (mail_lastc(m) != '\n') {
      if (putc(mail_getc(m), fp) != EOF)
         return 0;
      else {
         fprintf(stderr, "rmm: can't write to temporary file\n");
         return -1;
         }
      }
   else {   /* \n followed by F => read the line */
      if (mail_getstring(m, line, LINELEN) == NULL)
         return -1;
      else {
         if (ismsg(line))
            return 1;
         else 
            return putline(fp, line);
         }
      }
}

int
mail_msg2file(Mailfile *m, FILE *fp, char *line)
{
   buf *bp = &m->buffer;

   if (putline(fp, line) == -1)
      return MSG_QUIT;

   while (1) {
      char *start = (char *)buf_curp(bp);
      size_t num = (size_t)(buf_findF(bp) - start);

      if ( (num > 0) && (fwrite(start, sizeof(char), num, fp) != num)) {
         fprintf(stderr, "rmm: can't write to %s\n", mail_newname(m));
         return MSG_QUIT;
         }

      if (buf_eob(bp) && (mail_refill(m) == -1) )
         return MSG_EOF;
      else {
         int result = isfrom(m, fp, line);
         if (result != 0)
            return result;
         }
      }
}
