/*
   mailer.c : Copyright Paul Healy, EI9GL, 1992.

   920802 : Created.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "config.h"
#include "rc.h"
#include "pc.h"
#include "buf.h"
#include "mailfile.h"
#include "header.h"
#include "send.h"
#include "misc.h"
#include "mailer.h"
#include "pager.h"
#include "nntp.h"
#include "spawn.h"

char *
gettmpdir(void)
{
   char *tmp = getrc(tmpdir);

   if (tmp == NULL)
      if ( (tmp = getenv("BMHTMP")) == NULL)
         tmp = getenv("TMP");

   return tmp == NULL ? "./" : tmp;
}

/*
 * eg, set TMP=g:\   where g: is a ram disk
 */
FILE *
tempfile(char *prog, char *filename, char *mode)
{  
   FILE *fp = NULL;

   sprintf(filename, "%s%.6sXXXXXX", gettmpdir(), prog);

   if ((mktemp(filename) == NULL) || ( (fp = fopen(filename, mode)) == NULL) )
      fprintf(stderr, "%s: can't make temporary file %s\n", prog, filename);
#ifdef DEBUG
   printf("tempfile: made %s\n", filename);
#endif
   return fp;
}

FILE *
fopenbuf(char *filename, char *mode, int bufsize)
{
   FILE *fp = fopen(filename, mode);

   if (fp == NULL) {
      fprintf(stderr, "open: can't open %s\n", filename);
      return NULL;
      }
   else if (setvbuf(fp, NULL, _IOFBF, bufsize) == 0)
      return fp;
   else {
      fprintf(stderr, "bmh: can't buffer %s (tried %d)\n", filename, bufsize);
      fclose(fp);
      return NULL;
      }
}

int
putline(FILE *fp, char *s)
{
   if (s==NULL)
      return -1;
   else if (*s != '\0')
      if ( fputs(s, fp) == EOF )
         return -1;
   if (
#ifdef MSDOS
      (putc('\r', fp) != EOF) &&
#endif
      (putc('\n', fp) != EOF) )
      return 0;
   else
      return -1;
}

static int
call_edit(char *prog, char *ed, char *filename)
{
   if ( ed == NULL) {
      fprintf(stderr, "%s: no editor defined\n", prog);
      return -1;
      }
#ifdef SWAP
   if (run_lp(P_WAIT, ed, filename) != 0) {
#else
   if (spawnlp(P_WAIT, ed, ed, filename, NULL) != 0) {
#endif
      fprintf(stderr, "%s: error calling editor '%s %s'\n", prog, ed, filename);
      perror(prog);
      return -1;
      }

   return 0;
}

int
call_ed(char *prog, char *filename)
{
   return call_edit(prog, getrc(editor), filename);
}

static int
what_now(char *filename)
{
   char s[32];

   while (1) {
      char *arg[MAXARGS];
      int nargs;

      printf("What now? ");
      if (fgets(s, sizeof(s), stdin) == NULL)
         return -1;
         
      nargs = parse(rip(s), arg, MAXARGS);

      switch (nargs > 0 ? arg[0][0] : 0) {
         case 'e':
            call_edit("whatnow", nargs > 1 ? arg[1] : getrc(editor), filename);
            break;
         case 'l':
            page_file("bmh", filename);
            putchar('\r');
            break;
         case 'p':
         case 's':
         case '.':   /* bm/mail ness */
            return 0;
         case 'q':
            return -1;
         case 0:
         case '?':
            printf(" edit [<editor>]\n");
            printf(" list\n");
            printf(" post\n");
            printf(" quit\n");
            printf(" ? - print this message\n");
            break;
         default:
            printf("Type ? for help\n");
            break;
         }
      }
}

int
dowhatnow(char *prog, char *filename, Agent agent)
{
   if (what_now(filename) == 0) {
      Header_t h;
      FILE *fp = gethdrfn(filename, &h);

      if (fp == NULL) {
         fprintf(stderr, "%s: error reading %s\n", prog, filename);
         unlink(filename);
         return -1;
         }
      agent(fp, &h);
      fclose(fp);
      }

   return unlink(filename);
}

int
dosignature(FILE *fp)
{
   int ret = 0;
   FILE *sigfp;

   if (getrc(signature) == NULL)
      return 0;

   if ( (sigfp = fopen(getrc(signature), "r")) == NULL) {
      fprintf(stderr, "bmh: can't open signature file %s\n", getrc(signature));
      return -1;
      }
/*
   fprintf(fp, "--\n");
*/
   if (dupfile(sigfp, fp) != 0) {
      fprintf(stderr, "bmh: error adding signature file to message\n");
      ret = -1;
      }

   fclose(sigfp);

   return ret;
}

int
dupfile(FILE *in, FILE *out)
{
   static char *buf = NULL;
   int r;

   if ( (buf==NULL) && (buf = (char *) malloc(LINELEN*4)) == NULL) {
      fprintf(stderr, "dupfile: no room\n");
      return -1;
      }

   while ( (r = fread(buf, sizeof(char), sizeof(buf), in)) > 0 ) {
      if ( fwrite(buf, sizeof(char), r, out) != r) {
         fprintf(stderr, "bmh: error writing file\n");
         return -1;
         }
      if (r != sizeof(buf))
         break;
      }

   return 0;
}

int
isflag(char *flag, int argc, char *argv[])
{
   int i;

   for (i=1; i<argc; i++)
      if (strcmp(flag, argv[i]) == 0)
         return 1;

   return 0;
}

int
asc2month(char *s)
{
   char *result;
   static char m[][4] = {
         "Apr", "Aug", "Dec", "Feb", "Jan", "Jul",
         "Jun", "Mar", "May", "Nov", "Oct", "Sep" };
   static char m2num[] = {4, 8, 12, 2, 1, 7, 6, 3, 5, 11, 10, 9 };

   result = (char *) bsearch(s, m, 12, 4, (CMPFN) strcmp);
   return result == NULL ? 0 : (int) m2num[( (result - (char *) m) / 4 )];
}

/*
 * This function looks for a pattern :dd:dd (d=decimal number), eg
 * Fri Jan 22 00:45:02 1988
 */
static int
isdate(char *s)
{
   char *p;

   /*
    * Assume that if a '@' appears anywhere on the line, then we have
    * found the start of a mail message. (needed for messages delivered
    * by the nntp client).
    */
   if (strchr(s, '@') != NULL)  /* could also look for '!' here */
      return 1;

   p = strrchr(s, ':');

   if (p == NULL)
      return 0;

   p = p - 3;              /* move to the first ':' */

   if (p[0] != ':')
      return 0;

   if (isdigit(p[1]) && isdigit(p[2]) && isdigit(p[4]) && isdigit(p[5]))
      return 1;
   else
      return 0;
}

int
ismsg(char *s)
{
   if (*s != 'F')
      return 0;
   /*
    * This is a bit of a hack to enable the tcp group archives from ucsd
    * to be read correctly. It depends upon the local smtp server always
    * adding a 'From user ddd mmm d hh:mm:ss yyyy' header - the : is the
    * important part. ('@' won't work because some servers won't add an
    * at part for local mail, and misses uucp bang paths as well). But of
    * course this won't work for (see above for an attempt):
    * From @RELAY.CS.NET:jpd%usl-pc@USL.CSNET  Thu Feb 18 00:14:19 1988
    */
   else if ( (strncmp(s, "From ", 5) == 0) && isdate(s+5))
      return 1;
   else
      return 0;
}

/* ------------------------------------------------------------------------ */

char *
news2dir(char *s)
{
   char *p = s;

   if (s == NULL)
      return NULL;

   while (*s) {
      if (*s == '.')
         *s = '/';
      s++;
      }
   return p;
}

char *
dir2news(char *s)
{
   char *p = s;

   if (s == NULL)
      return NULL;

   while (*s) {
      if (*s == '/')
         *s = '.';
      s++;
      }
   return p;
}

int
mktemplate(char *prog, char *filename, char *to, char *subj, char *message, int sig)
{
   int ret = 0;
   FILE *fp = tempfile(prog, filename, "w");

   if ( (fp == NULL) ||
        (fprintf(fp, "To: %s\nCc: \nSubject: %s\n--------\n%s", to, subj, message) == EOF) ||
        (sig && (dosignature(fp) == -1))
      )
      ret = -1;

   if (fp != NULL)
      fclose(fp);

   if (ret == -1)
      fprintf(stderr, "%s: unable to write to %s\n", prog, filename);

   return ret;
}

int
addstr(char *prog, char *filename, char *s)
{
   int ret = 0;
   FILE *fp = fopen(filename, "a");

   if (fp == NULL)
      ret = -1;
   else {
      if (fputs(s, fp) == EOF)
         ret = -1;
      fclose(fp);
      }

   if (ret == -1)
      fprintf(stderr, "%s: unable to write to %s\n", prog, filename);

   return ret;
}

int
addquotedmsg(char *prog, char *filename, FILE *fpin, char *quotestr)
{
   FILE *fpout = fopen(filename, "a");
   int ret = 0;

   if ( (fpin == NULL) || (fpout == NULL))
      ret = -1;
   else {
      char buf[LINELEN];

      while (fgets(buf, sizeof(buf), fpin) != NULL)
         if (*buf == '\n')
            fputc('\n', fpout);
         else if (fprintf(fpout, "%s%s", quotestr, buf) == EOF) {
            ret = -1;
            break;
            }

      fclose(fpout);
      }

   if (ret == -1)
      fprintf(stderr, "%s: unable to write to %s\n", prog, filename);

   return ret;
}

/*
 * remove rfc822 style comments from a string, and remove leading
 * and trailing blank characters.
 * returns:
 *  NULL, if there is a malformed strings, ie () which don't match
 *  a pointer to the decommented string otherwise.
 */
char *
decomment(char *s)
{
   char *p;

   while ((p=strchr(s, '(')) != NULL) {
      char *q = strchr(p, ')');

      if (q==NULL)
         return NULL;
      else
         strcpy(p, q+1);
      }

   p=s;
   while (*p && isspace(*p))
      p++;
   if (p!=s)
      strcpy(s, p);          /* trim leading spaces */

   p = strchr(s, '\0');
   while ( (p>s) && isspace(*(p-1)) )
      p--;
   *p='\0';                  /* trim trailing spaces */

   return s;
}
