/*
   mailfile.c: Copyright 1993, Paul Healy.

   930201 : Created
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include "config.h"
#include "buf.h"
#include "mailfile.h"
#include "mailer.h"
#include "rc.h"
#include "pc.h"
#include "lock.h"
#include "buf.h"

#undef DEBUG

#ifdef DEBUG

int
mail_debug(Mailfile *m)
{
   printf("Folder: %s\n", mail_folder(m));
   printf("Location: %s\n", mail_stub(m));
   printf("Filename: %s\n", mail_filename(m));
   printf("Newsgroup: %s\n", mail_newsgroup(m));
   printf("New name: %s\n", mail_newname(m));
   printf("Old name: %s\n", mail_oldname(m));
   printf("Current msg: %d\n", mail_curmsg(m));
   printf("Last msg: %d\n", mail_lastmsg(m));
   fflush(stdout);

   return 0;
}

#endif

/* ---------------------------------------------------------------------- */

int
mail_lastmsg(Mailfile *m, int argc, char *argv[])
{
   int last = 0, i, this;

   for (i=1; i<argc; i++)
      if (isdigit(argv[i][0])) {
         char *p = strchr(argv[i], '-');

         if ( p != NULL)
            this=atoi(p+1);
         else
            this=atoi(argv[i]);

         if (this>last)
            last = this;
         }

   if (last > 0)
      return last;
   else if (m == NULL)
      return 1;
   else 
      return mail_curmsg(m);
}

int
mail_thismsg(Mailfile *m, int argc, char *argv[])
{
   int i;

   for (i=1; i<argc; i++)
      if (isdigit(argv[i][0]))
         return atoi(argv[i]);

   return mail_curmsg(m) > 0 ? mail_curmsg(m) : 1;
}

int
mail_prevmsg(Mailfile *m, int argc, char *argv[])
{
   int first = INT_MAX, i, this;

   for (i=1; i<argc; i++)
      if (isdigit(argv[i][0])) {
         this=atoi(argv[i]);
         if (this<first)
            first = this;
         }

   if (first != INT_MAX)
      return first;
   else if (m == NULL)
      return 1;
   else 
      return (mail_curmsg(m)-1) > 0 ? (mail_curmsg(m)-1) : 1;
}

/* ---------------------------------------------------------------------- */

int
mail_open(Mailfile *m)
{
#ifdef DEBUG
   fprintf(stderr, "mailopen: %s fd=%d lock=%d\n", mail_filename(m), m->fd, m->keeplock);
#endif

   if ( (m->fd != -1) && (m->keeplock) )
      return 0;
   else if (mail_lock(m) == -1) {
      fprintf(stderr, "bmh: can't lock %s\n", mail_filename(m));
      return -1;
      }
   else {
      m->fd = open(mail_filename(m), O_RDONLY | O_BINARY);

      if (m->fd == -1) {
/*
         fprintf(stderr, "bmh: error opening %s\n", mail_filename(m));
*/
         if (mail_unlock(m) == -1)
            fprintf(stderr, "bmh: error unlocking %s\n", mail_filename(m));
         return -1;
         }
      else
         return m->fd;
      }
}

int
mail_close(Mailfile *m)
{
#ifdef DEBUG
   fprintf(stderr, "mailclose: %s, fd=%d, lock=%d\n", mail_filename(m), m->fd, m->keeplock);
#endif

   if (m->keeplock || (m->fd == -1))
      return 0;
   else if (close(m->fd) == -1) {
      fprintf(stderr, "bmh: error closing %s (%d)\n", mail_filename(m), m->fd);
      m->fd = -1;
      mail_unlock(m);
      return -1;
      }
   else if (mail_unlock(m) == -1) {
      fprintf(stderr, "bmh: error unlocking %s\n", mail_filename(m));
      m->fd = -1;
      return -1;
      }
   else {
      m->fd = -1;
      return 0;
      }
}

long
mail_size(Mailfile *m)
{
   long size;

   if (mail_open(m) == -1)
      return -1L;
   else if ( (size=filelength(m->fd))==0L) {
      if ( (close(m->fd)==-1) ||
           (unlink(mail_filename(m)) == -1) ||
           (mail_unlock(m) == -1)
         )
         return -1L;
      else
         return 0L;
      }
   else if (mail_close(m)==-1)
      return -1L;
   else
      return size;
}

#define mail_check(condition, msg, file) if (condition) { \
 fprintf(stderr, msg, prog, file); \
 unlink(mail_newname(m)); \
 return -1; \
}

int
mail_update(Mailfile *m, char *prog)
{
   mail_check((access(mail_oldname(m), 0)==0) && (unlink(mail_oldname(m)) == -1), "%s: can't unlink %s\n", mail_oldname(m));
   mail_check(rename(mail_filename(m), mail_oldname(m)) == -1, "%s: can't rename %s to make backup\n", mail_filename(m));
   mail_check(rename(mail_newname(m), mail_filename(m)) == -1, "%s: can't rename %s to make new mailfile\n", mail_newname(m));

   return 0;
}
#undef mail_check
