/*
   mailfile.h: Copyright 1993, Paul Healy, EI9GL.

   930201 : Created
*/
#ifndef _mailfile_h_
#define _mailfile_h_

typedef struct {
   char *filename;            /* fully qualified file name */
   char *basename;
   char *newsname;            /* newsgroup form of the name */
   char *newname;             /* fully qualified name, .new extension */
   char *oldname;             /* fully qualified name, .old extension */
   char *foldname;            /* fully qualified name, no extension */
   char *lockname;            /* fully qualified name, .lck extension */
   char *indexname;           /* fully qualified name, .idx extension */

   int curmsg;
   long offset;

   int keeplock;              /* between reads */
   int fd;
   buf buffer;
} Mailfile;

/* in mailinit.c */
void mail_zero(Mailfile *m);
int mail_init(Mailfile *m, char *basename); /* init data struct */
int mail_fin(Mailfile *m);                  /* tidy up when finished */
int mail_loadstate(Mailfile *m);            /* get details from disk */
int mail_savestate(Mailfile *m);            /* put details to disk */
int mail_getfile(Mailfile *m, int argc, char *argv[], char *prog);
#ifdef _GETOPTS_H
options_t *mail_opts(Mailfile *m, int *argc, char ***argv, char *optv[], char *prog);
int mail_endopts(int errors, options_t *opts);
#endif
int mail_reset(Mailfile *m);

/* in mailfile.c */
int mail_debug(Mailfile *m);
int mail_lastmsg(Mailfile *m, int argc, char *argv[]);
int mail_thismsg(Mailfile *m, int argc, char *argv[]);
int mail_prevmsg(Mailfile *m, int argc, char *argv[]);
int mail_open(Mailfile *m);
int mail_close(Mailfile *m);
long mail_size(Mailfile *m);
int mail_update(Mailfile *m, char *prog);

/* in mailbuf.c */
int mail_refill(Mailfile *m);
char *mail_getstring(Mailfile *m, char *line, int max);
int mail_skipline(Mailfile *m);
int mail_getmsg(Mailfile *m, char *line, int max);
int mail_curc(Mailfile *m);
int mail_lastc(Mailfile *m);
int mail_getc(Mailfile *m);
int mail_msg2file(Mailfile *m, FILE *fp, char *line);

#define mail_folder(m) ((m)->basename)    /* returns the current folder */
#define mail_stub(m) ((m)->foldname)      /* full path, no extension */
#define mail_filename(m) ((m)->filename)  /* full file name */
#define mail_newsgroup(m) ((m)->newsname) /* returns the current newsgroup */
#define mail_newname(m) ((m)->newname)    /* new file name */
#define mail_oldname(m) ((m)->oldname)    /* old file name */
#define mail_idxname(m) ((m)->indexname)  

#define mail_curmsg(m) ((m)->curmsg)      /* returns the current msg num */
#define mail_setcur(m, n) ((m)->curmsg = n > 0 ? n : 1) /* sets the current msg */

#define mail_keeplock(m) ((m)->keeplock = 1)
#define mail_releaselock(m) ((m)->keeplock = 0)

#define mail_lock(m) mklock((m)->lockname)
#define mail_unlock(m) rmlock((m)->lockname)
#define mail_resetpos(m) buf_reset((&(m)->buffer))
#define mail_offset(m) ((m)->offset)

#endif /* _mailfile_h_ */
