/*
   mailinit.c: Copyright 1993, Paul Healy.

   930518 : Created
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "buf.h"
#include "mailfile.h"
#include "mailer.h"
#include "rc.h"
#include "pc.h"
#include "lock.h"
#include "buf.h"
#include "getopts.h"
#include "misc.h"

#undef DEBUG

void
mail_zero(Mailfile *m)
{
#ifdef DEBUG
   fprintf(stderr, "mail zero called\n");
#endif

   m->filename = NULL;     
   m->basename = NULL;
   m->newsname = NULL;     
   m->newname = NULL;      
   m->oldname = NULL;      
   m->foldname = NULL;      
   m->lockname = NULL;      
   m->indexname = NULL;      

   m->curmsg = 1;
   m->keeplock = 0;

   m->fd = -1;
   buf_init(&m->buffer);
}

int           
mail_init(Mailfile *m, char *basename)
{
   char buf[MAXNAMLEN];
   char *path = getrc(maildir);

#ifdef DEBUG
   fprintf(stderr, "mail init called <%s>\n", basename);
#endif

   if (m == NULL)
      return -1;
   else if (basename == NULL)
      basename = getrc(username);

   if (basename == NULL)
      return -1;

   news2dir(basename);

   sprintf(buf, "%s/%s%s", path, basename, EXT);
   m->filename = strdup(buf);
   m->basename = strdup(basename);
   m->newsname = strdup(basename);      dir2news(m->newsname);

   sprintf(buf, "%s/%s.new", path, basename);
   m->newname = strdup(buf);

   sprintf(buf, "%s/%s.old", path, basename);
   m->oldname = strdup(buf);
   sprintf(buf, "%s/%s", path, basename);
   m->foldname = strdup(buf);
   sprintf(buf, "%s/%s.lck", path, basename);
   m->lockname = strdup(buf);
   sprintf(buf, "%s/%s.idx", path, basename);
   m->indexname = strdup(buf);

   mail_setcur(m, 1);

   m->keeplock = 0;
   m->fd = -1;

   if ( (m->filename == NULL) ||
        (m->basename == NULL) ||
        (m->newsname == NULL) ||
        (m->newname == NULL)  ||
        (m->oldname == NULL)  ||
        (m->foldname == NULL) ||
        (m->lockname == NULL) ||
        (m->indexname == NULL)
      ) {
      return -1;
      }
   else
      return 0;
}

/*
 * free any malloced memory mainly
 */
int
mail_fin(Mailfile *m)
{
#ifdef DEBUG
   fprintf(stderr, "mail fin called <%s>\n", m->filename);
#endif

   if (m == NULL)
      return -1;

   free(m->filename);
   free(m->basename);
   free(m->newsname);
   free(m->newname); 
   free(m->oldname);
   free(m->foldname);
   free(m->lockname);

   buf_free(&m->buffer, 1);

   mail_zero(m);

   return 0;
}

/* --------------------------------------------------------------------- */

int
mail_loadstate(Mailfile *m)
{
   char *filename = getrc(current);
   char basename[MAXNAMLEN];
   FILE *fp;
   int msg;

   if ( (m == NULL) || (filename == NULL) || (fp=fopen(filename, "r")) == NULL)
      return -1;
   else if (fscanf(fp, "%s %d", basename, &msg) != 2) {
      fclose(fp);
      return -1;
      }
   else {
      fclose(fp);

      if ( mail_init(m, basename) == -1)
         return -1;
      else {
         mail_setcur(m, msg < 1 ? 1 : msg);

#ifdef DEBUG
         fprintf(stderr, "mailload: %s %d\n", basename, msg);
#endif
         return 0;
         }
      }
}

int
mail_savestate(Mailfile *m)
{
   char *filename = getrc(current);
   FILE *fp;

#ifdef DEBUG
   fprintf(stderr, "mail save called %s %d\n", mail_folder(m), mail_curmsg(m));
#endif

   if ( (m == NULL) || (filename == NULL) || (fp=fopen(filename, "w")) == NULL)
      return -1;
   else {
      fprintf(fp, "%s %d", mail_folder(m), mail_curmsg(m) < 1 ? 1 : mail_curmsg(m));
      return fclose(fp);
      }
}

/* ---------------------------------------------------------------------- */
/*
 * Get the current state information from -
 *  1) the command line, 'show +inbox 1'
 *  2) from current state
 *  3) from current.bmh
 */
int
mail_getfile(Mailfile *m, int argc, char *argv[], char *prog)
{
   if (m==NULL)
      return -1;
   else if ((argc > 1) && (argv[1][0] == '+')) {
      if (mail_folder(m) != NULL)
         mail_fin(m);
      if (mail_init(m, &argv[1][1]) == -1) {
         fprintf(stderr, "%s: unable to initialiase for %s\n", prog, &argv[1][1]);
         return -1;
         }
      else {
         mail_setcur(m, 1);
         return 0;
         }
      }
   else if (mail_folder(m) != NULL)
      return 0;                             /* use current folder/msg info */
   else if (mail_loadstate(m) == -1)        /* possibly no current.bmh     */
      return mail_init(m, getrc(username)); /* so try the default username */
   else
      return 0;
}

options_t *
mail_opts(Mailfile *m, int *argc, char ***argv, char *optv[], char *prog)
{
   options_t *opts;

   opts = optinit(prog, optv);
   optctrl(opts, opt_LONG_ONLY);
   null_check("opts");

   if (mail_getfile(m, *argc, *argv, prog) == 0) {
      null_check("opts getfile");
      if ((*argc>1) && ((*argv)[1][0] == '+')) {
         *argc = *argc - 1;
         *argv = *argv + 1;
         null_check("opts +");
         }
      *argc = *argc - 1;
      *argv = *argv + 1;
      null_check("opts +2");

      return opts;
      }
   else {
      optfree(opts);
      return NULL;
      }
}

int
mail_endopts(int errors, options_t *opts)
{
   if (errors) {
      optusage(opts, "", stderr);
      optfree(opts);
      return -1;
      }
   else {
      int optind = getoptind(opts);
      optfree(opts);
      return optind;
      }
}

int
mail_reset(Mailfile *m)
{
   char *name;
   int msg;

   name = strdup(mail_newsgroup(m));
   msg = mail_curmsg(m);

   mail_fin(m);

   if (mail_init(m, name) == -1) {
      fprintf(stderr, "bmh: can't change to %s\n", name);
      free(name);
      return -1;
      }
   else {
      mail_setcur(m, msg);
      free(name);
      return 0;
      }
}
