/* bm.c
 *	Simple mail user interface for KA9Q IP/TCP package.
 *	A.D. Barksdale Garbee II, aka Bdale, N3EUA
 *	Copyright 1986 Bdale Garbee, All Rights Reserved.
 *	Permission granted for non-commercial copying and use, provided
 *	this notice is retained.
 *
 * revision history:
 *
 *  v3.0 871225 D. Trulli nn2z
 *	Heavily restructured program to use a index array of message.
 *	More compatible with UNIX type mail commands.
 *  v2.7 871214 D. Trulli nn2z
 *	Cleaned up header and message parsing to prevent lock up.
 *	Revised command structure.
 *  v2.6 870826 Bdale
 *	integrate PA0GRI's interface/gateway to the WA7MBL PBBS
 *  v2.5 870713 Bdale
 *      integrate additional patches from PA0GRI, minor cleanup
 *  v2.4 870614 - P. Karn
 *      "smart gateway" function moved to smtp client code in net.exe.
 *	User interface for sending mail reworked to resemble Berkeley Mail
 *  v2.3 870524
 *      Extensive addition/revision by Gerard PA0GRI.  Now supports a healthy
 *      set of real commands.  
 *  v2.1,2.2
 *      exact change history lost.
 *  v2.0 c.870115
 *      First version with command parser, only send and quit work.
 *  v1.0
 *      First attempt.  Send messages only.
 */

#include <stdio.h>
#include "bm.h"
#include "stdlib.h"		/* added in the Atari Version (system()) */

extern char version[];

#ifdef ATARI
extern char implementation[];
#include "stdlib.h"
extern long _32K;
long _32K = 0x8000;
extern long _MNEED;
long _MNEED = 100000;
int bmmode = 0;			/* 0=SMTP, 1=UUCP */
#endif

static char copyright[] = 
"Copyright 1987 Bdale Garbee, permission granted for non-commercial use.";

FILE	*mfile = NULLFILE;
char 	hostname[SLINELEN],	/* name of this host from rc file */
	username[SLINELEN],	/* name of this user from rc file */
	fullname[SLINELEN],	/* full name of this user from rc file */
	replyto[SLINELEN],	/* address for reply-to header */
/*	inreplyto[SLINELEN],*/	/* original message-id */
	timezone[4],		/* users time zone */
	maildir[LINELEN],	/* defined mail directory */
	editor[SLINELEN],	/* user's favorite text editor */
	notename[9],		/* name of current notesfile */
	notefile[LINELEN];	/* full path name of the mail text file */
int	current;		/* the current message number */
int	nmsgs;			/* the number of messages in the notesfile */
int	change;			/* indicates that the mail file has been
				 * changed in this session */
unsigned maxlet = NLET+1;	/* max number of messages in mailbox */
int	tty;			/* tells if stdin is a tty */
char	*malloc();
struct let *mbox;

/* command parser, handle at least read, send, headers, delete */
main(argc,argv) 
int argc;
char *argv[];
{
	FILE *tfile, *tmpfile();
	char command[80];	/* command line */
	char	line[80];
	char	*notel, *tch, *index();
	int	msgnum;
	int	len;
	long	tmp;

#ifndef ATARI
	if (isatty(fileno(stdin))) {
#endif
		/* announce ourselves */
		printf("\n%s\n",version);
#ifdef ATARI
		(void)xbios(21,2);		/* make the cursor blink */
		printf("%s\n\n",implementation);
		tty = 1;
#ifndef ATARI
	} else
		tty = 0;
#endif

	/* initialize globals from config file */
	*editor = '\0';  /* default no editor */
	*maildir = '\0';
	*replyto = '\0';
	current = 1;
	nmsgs = 0;
	strcpy(timezone,"GMT");
	read_rc();

	if(*maildir == '\0')
		strcpy(maildir,"\\spool\\inbound\\");	/* was: /spool/mail/ */
	else if (maildir[strlen(maildir) -1] != '\\')	/* was: '/' */
#ifdef ATARI
			strcat(maildir,"\\");
#else
			strcat(maildir,"/");
#endif

	strncpy(notename,username,8);
	notename[8] = '\0';
	if(argc > 1){
			dosmtpsend(NULLFILE,argv[1],NULLCHAR,NULLCHAR);
			exit(0);
	}

	/*
	* Since we are in the dos small model make sure that we
	* don't overflow a unsigned short on the number bytes
	* need for mallloc. If don;t check still malloc will
	* succeed and we will be trashing ourself in now time.
	*/
	tmp = (long)maxlet * (long)sizeof(struct let);
	if ((tmp & 0xffff0000) ||
	(mbox=(struct let *)malloc((unsigned short)tmp)) == (struct let *)NULL){
		printf("Not enough more to allocate table for %d messages\n",
			maxlet);
		exit(1);
	}

	if (bmmode == 0) {
		sprintf(notefile,"%s%s.txt",maildir,notename);
	} else {
		sprintf(notefile,"%s%s.mai",maildir,notename);
	}
	initnotes();
	listnotes();
	printf("\nType ? for help.\n");
 
	/* command parsing loop */
	while(1) {
		if (bmmode == 0) {
			printf("[smtp] %s> ",notename);
		} else {
			printf("[uucp] %s> ",notename);
		}
	  gets(command);
	  if (*command != '\0') {
		  notel = &command[1];
		  while (*notel != '\0')
			if (*notel == ' ' || *notel == '\t')
				notel++;
			else
				break;
	  }

	  switch (*command) {
	    case 'm':		/* send msg */
		if (*notel != '\0')
			dosmtpsend(NULLFILE,notel,NULLCHAR,NULLCHAR);
		else
			dosmtpsend(NULLFILE,NULLCHAR,NULLCHAR,NULLCHAR);
		break;

	    case 's':		/* save current msg to file */
		if(*notel == '\0') 
			notel = "mbox";
		if((tfile = fopen(notel,"a")) == NULLFILE) {
			printf("Can't open '%s'\n",notel);
			break;
		} else {
			savemsg(current,tfile,0);
			fclose(tfile);
			printf("%s appended\n",notel);
		}
		break;

	    case 'w':		/* save current msg to file */
		if(*notel == '\0') 
			notel = "mbox";
		if((tfile = fopen(notel,"a")) == NULLFILE) {
			printf("Can't open '%s'\n",notel);
			break;
		} else {
			savemsg(current,tfile,1);
			fclose(tfile);
			printf("%s appended\n",notel);
		}
		break;

	    case 't':		/* transmit msg in file*/
		if (*notel == '\0') {
			printf("Enter filename containing message: ");
			gets(command);
			tch = command;
		} else
			tch = notel;
		if((tfile = fopen(tch,"r")) == NULLFILE)
			printf("\n Cannot open file %s\n",tch);
		else {
			dosmtpsend(tfile,NULLCHAR,NULLCHAR,NULLCHAR);
			fclose(tfile);
		}
		break;

	    case 'q':		/* quit */
		closenotes();
		unlink("\\xbm0.tmp");	/* used in tmpfile() */
		unlink("\\xbm1.tmp");	/* used during reply */
		unlink("\\xbm2.tmp");	/* temp file used during forward */
	    	exit(0);
		break;

	    case 'x':		/* abort */
		fclose(mfile);
	    	exit(0);
		break;

	    case 'p':		/* printer on or off */
		savemsg(current,stdout,0);
		break;

	    case 'r':			/* reply */
		if (*notel == '\0')
			reply(current);
		else {
			msgnum = atoi(notel);
			if( msgnum >= 1 && msgnum <= nmsgs)
				reply(atoi(notel));
			else
				printf("Invalid mesage number %d\n",msgnum);
		}
		break;

	    case 'u':
		if (*notel == '\0')
			msgnum = current;
		else
			msgnum = atoi(notel);
		if( msgnum >= 1 && msgnum <= nmsgs)
			mbox[msgnum].status &= ~DELETE;
		else
			printf("Invalid mesage number %d\n",msgnum);
		break;

	    case 'f':		/* forward a message */
		if(*notel != '\0') {
			msgnum = atoi(notel);
/*			if((tfile = tmpfile()) == NULLFILE)	*/
			if((tfile = fopen("\\xbm2.tmp","w+")) == NULLFILE)
				printf("\nCannot open temp file\n");
			else {
				savemsg(msgnum,tfile,0);
				fseek(tfile,0L,0);
				dosmtpsend(tfile,NULLCHAR,NULLCHAR,NULLCHAR);
				fclose(tfile);
			}
		}
	    case '!':
		system(notel);
		break;

	    case 'l':		/* display unsent messages */
	        printf("List of unsent messages.\n");
		filedir(mailqueue,0,line);
			while(line[0] != '\0') {
				printf("message in -> %s \n",line);
				filedir(mailqueue,1,line);
			}
		break;

	    case 'n':		/* display or change notefiles */
		if(*notel != '\0') {
			closenotes();
			strncpy(notename,notel,8);
			notename[8] = '\0';
			if (bmmode == 0) {
				sprintf(notefile,"%s%s.txt",maildir,notename);
			} else {
				sprintf(notefile,"%s%s.mai",maildir,notename);
			}
			initnotes();
			listnotes();
		} else {  /* (s)he wants to see what notefiles there are */
			if (bmmode == 0) {
				sprintf(command,"%s*.txt",maildir,notename);
			} else {
				sprintf(command,"%s*.mai",maildir,notename);
			}
			filedir(command,0,line);
			while(line[0] != '\0') {
				tch = index(line,'.');
				*tch = '\0';
				printf("notesfile -> %s\n",line);
				filedir(command,1,line);
			}
		}
		break;

	    case 'd':		/* delete a message */
		if (*notel != '\0')
			msgnum = atoi(notel);
		else
			msgnum = current;
		delmsg(msgnum);
		break;

	    case 'h':		/* list message headers in notesfile */
		listnotes();
		break;

	    case '-':		/* previous */
		if (current > 1)
			current--;
		else
			printf("No more messages\n");
		break;
	    case '+':		/* next */
		if (current < nmsgs)
			current++;
		else
			printf("Last message\n");
		break;
	
	    case '.':
		bmmode = ~bmmode;
		if(*notel != '\0') {
			closenotes();
			strncpy(notename,notel,8);
			notename[8] = '\0';
			if (bmmode == 0) {
				sprintf(notefile,"%s%s.txt",maildir,notename);
			} else {
				sprintf(notefile,"%s%s.mai",maildir,notename);
			}
			initnotes();
			listnotes();
		} else {
			printf("You must specify a mailbox name!\n");
			bmmode = ~bmmode;
		}
/*		closenotes();
		bmmode = ~bmmode;
		initnotes();
		listnotes();	*/
		break;

	    case '?':		/* help */
		printf(" d [msg]     delete a message\n");
		printf(" m user      mail a message\n");
		printf(" s [file]    save message in file (default mbox)\n");
		printf(" w [file]    save message in file no header (default mbox)\n");
		printf(" t file      send a message from a file\n");
		printf(" r [msg]     reply to a message\n");
		printf(" f [msg]     forward a message \n");
		printf(" u [msg]     undelete a message\n");
		printf(" +           goto next message\n");
		printf(" -           goto previous message\n");
		printf(" p [msg]     print message on terminal\n");
		printf(" h           display message headers in notefile\n");
		printf(" l           list unsent messages\n");
		printf(" n [file]    display or change notesfile\n");
		printf(" #           number of message to read in notesfile\n");
		printf(" x           quit without changing mail file\n");
		printf(" q           quit\n");
		printf(" .           toggle between SMTP and UUCP mode\n");
		printf(" ?           print this help screen\n");
		break;

	    default:
		if((msgnum = atoi(command)) > 0) {
			if (msgnum < 0 || msgnum > nmsgs)
				printf("Message number out of range\n");
			else {
				current = msgnum;
				savemsg(msgnum,stdout,0);
			}
		} else
			printf(" commands <d,m,s,w,,+,-,t,r,f,u,p,h,l,n,#,?,q,x>\n");
	    	break;
	  }
	} 
}

read_rc()
{
	FILE	*rcfp;	/* handle for the configuration file */
	char	rcline[LINELEN]; /* buffer for config file reading */
	if((rcfp = fopen(BMRC,"r")) == NULLFILE) {	/* open config file */
		printf("Cannot open \BM.RC, check your installation\n");
		exit(-1);
	} else
	    while (!feof(rcfp)) {
		fgets(rcline,LINELEN,rcfp);
		rip(rcline);
		if (rcline[0] == ';') continue;
		if (strlen(rcline) < 5) continue;
		switch (rc_line_type(rcline)) {
			case HOST:
				strncat(hostname,&rcline[5],sizeof(hostname));
				break;
			case USER:
				strncat(username,&rcline[5],sizeof(username));
				break;
			case REPLY:
				strncat(replyto,&rcline[6],sizeof(replyto));
				break;
			case EDIT:
				strncat(editor,&rcline[5],sizeof(editor));
				break;
			case ZONE:
				strncpy(timezone,&rcline[5],3);
				timezone[3] = '\0';
				break;
			case SMTP: 
				*maildir = '\0';
				strncat(maildir,&rcline[5],sizeof(maildir));
				break;
			case NAME:
				*fullname = '\0';
				strncat(fullname,&rcline[9],sizeof(fullname));
				break;

			case MAXLET:
				maxlet = atoi(&rcline[7])+1;
				break;
			default:
				break;
		}	
	    }
	fclose(rcfp);
}

/* return the line_type from a line of the configuration file */
rc_line_type(s)
char *s;
{
	if (strncmp("fido",s,4) == 0) return (FIDO);
	if (strncmp("smtp",s,4) == 0) return (SMTP);
	if (strncmp("host",s,4) == 0) return (HOST);
	if (strncmp("user",s,4) == 0) return (USER);
	if (strncmp("edit",s,4) == 0) return (EDIT);
	if (strncmp("zone",s,4) == 0) return (ZONE);
	if (strncmp("fullname",s,8) == 0) return (NAME);
	if (strncmp("reply",s,5) == 0) return (REPLY);
	if (strncmp("maxlet",s,6) == 0) return (MAXLET);
	return (NONE);
}

/* replace terminating end of line marker(s) with null */
rip(s)
register char *s;
{
	register char *cp;
	char *index();

	if((cp = index(s,'\r')) != NULLCHAR)
		*cp = '\0';
	if((cp = index(s,'\n')) != NULLCHAR)
		*cp = '\0';
}
