/*
   main.c: Copyright 1993, Paul Healy.

   To do: add global options here -
      . debug
      . verbose
      . editor
      . moreproc

   930418 : Created
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "buf.h"
#include "mailfile.h"
#include "bmh.h"
#include "help.h"
#include "misc.h"
#include "mailer.h"

int Debug = 0;

static struct jumptable {
   char jump[9];
   MAINFN prog;
} table[] = {
#ifdef MAIL
    "ali",     ali_main,
    "bmh",     bmh_main,
    "comp",    comp_main,
    "dist",    dist_main,
#endif
#ifdef AX25BBS
    "expirbid",expirbid_main,
    "expire",  expire_main,
#endif
#ifdef NEWS
    "expirnn", expirnn_main,
#endif
#ifdef AX25BBS
    "export",  export_main,
#endif
#ifdef MAIL
    "folders", folders_main,
    "forw",    forw_main,
#endif
#ifdef AX25BBS
    "import",  import_main,
#endif
#ifdef NEWS
    "inc",     inc_main,
#endif
    "issue",   showissue,
#ifdef MAIL
    "mail",    mail_main,
#endif
#ifdef INDEX
    "map",     map_main,
#endif
#ifdef MAIL
    "next",    next_main,
    "pick",    pick_main,
#endif
#ifdef NEWS
    "post",    post_main,
#endif
#ifdef MAIL
    "refile",  refile_main,
    "repl",    repl_main,
    "rmm",     rmm_main,
    "scan",    scan_main,
    "show",    show_main,
    "smtp",    smtp_main
#endif
};

int
printoptions(void)
{
   const int num = sizeof(table) / sizeof(struct jumptable);
   int i;
   char *argv[2];

   argv[1] = "-help";
   sethelp(1);

   fprintf(stdout, "usage: bmh [-help|-?|many options] (read/send mail) ");
   version(stdout);

   for (i=0; i<num; i++)
      if (strcmp(table[i].jump, "bmh") != 0) {
/*
         fprintf(stdout, "       bmh -");
*/
         fprintf(stdout, " bmh -");
         argv[0] = table[i].jump;
         (void) table[i].prog(2, argv);
         if (i<num-1)
            fputc('\n', stdout);
         }

   fflush(stdout);

   sethelp(0);

   return 0;
}

int
callprog(int argc, char *argv[])
{
   struct jumptable *result;
   char *cp, *lookup;
   char *start = argv[0] + strlen(argv[0]);

   while ( (start>=argv[0]) && (*start != '/') && (*start != '\\')) {
      *start = tolower(*start);
      start--;
      }
   start++;
   if ((cp = strchr(start, '.')) != NULL)
      *cp = 0;

   if ( (strcmp(start, "bmh") == 0) && (argc > 1) && (argv[1][0] == '-') )
      lookup = &argv[1][1];
   else 
      lookup = start;

   if (Debug)
      printf("bmh: command is <%s>\n", lookup);

   result = (struct jumptable *)
      bsearch(lookup,
         table,
         sizeof(table) / sizeof(struct jumptable),
         sizeof(struct jumptable),
         (CMPFN) strcmp);

   if (result == NULL)
      return bmh_main(argc, argv);
   else {
      if (lookup == start)
         return (*result->prog) (argc, argv);
      else {
         argv[1]++; /* skip '-' */
         return (*result->prog) (argc-1, &argv[1]);
         }
      }
}

int
main(int argc, char *argv[])
{
   null_init();

   if (isflag("-debug", argc, argv))
      Debug = 1;

   if (setupbm()==-1)
      return -1;

   return callprog(argc, argv);
}
