/* Main network program - provides both client and server functions */
#define        NSESSIONS        10        /* Maximum interactive client sessions */
#define HOSTNAMELEN 32                /* changed from 16 by Bdale 860812 */

#ifndef        STARTUP
#define	       STARTUP (char*) makefname("TCPIP_STARTUP","/net.startup")
#endif

#include <stdio.h>
#include <time.h>
#include "machdep.h"
#include "mbuf.h"
#include "netuser.h"
#include "timer.h"
#include "icmp.h"
#include "iface.h"
#include "ip.h"
#include "tcp.h"
#include "ftp.h"
#include "telnet.h"
#include "session.h"
#include "cmdparse.h"
#include "pc.h"

extern struct interface *ifaces;
struct session sessions[NSESSIONS];
struct session *current;
extern char major_rev[], minor_rev[];
extern struct mbuf *loopq;

int mode;
FILE *logfp;
char badhost[] = "Unknown host %s\r\l";

#ifdef        TRACE
#include "trace.h"
int32 trace;
#endif

char hostname[HOSTNAMELEN];        
int32 resolve();
int16 lport = 1001;
char prompt[] = "net> ";
char nospace[] = "No space!!\r\l";        /* Generic malloc fail message */
static char notval[] = "Not a valid TCB\r\l";

#ifndef        MSDOS                        /* PC uses F-10 key always */
static char escape = 0x1d;        /* default escape character is ^] */
#endif

/* Command lookup and branch table */
int go(),cmdmode(),dodigipeat(),
        doipaddr(),dotelnet(),doexit(),doclose(),dohostname(),
        doreset(),dotcpstat(),dotrace(),doescape(),dohelp(),
        dowindow(),doroute(),doecho(),
        dolog(),doipstat(),doicmpstat(),doetherstat(),memstat(),doarp(),
        dosession(),doftp(),domss(),dostart(),dostop(),doattach(),
        domycall(),dosmtptick(),doudpstat(),dottl(),doparam(),dotcpkick(),
        doeol(),dogateway(),dobeacon();

#ifdef HAPN
int        dohapnstat();
#endif HAPN

#ifdef        EAGLE
int doegstat();
#endif

static struct cmds cmds[] = {
        /* The "go" command must be first */
        "",                go,                0, NULLCHAR,        NULLCHAR,
        "arp",                doarp,                0, NULLCHAR,        NULLCHAR,
        "attach",        doattach,        2,
                "attach <hardware> <hw specific options>", NULLCHAR,
        "close",        doclose,        0, NULLCHAR,        NULLCHAR,
#ifdef        AX25
        "digipeat",        dodigipeat,        0, NULLCHAR,        NULLCHAR,
#endif
#ifdef        EAGLE
        "eaglestat",        doegstat,        0, NULLCHAR,        NULLCHAR,
#endif
        "echo",                doecho,                0, NULLCHAR,        "echo [refuse|accept]",
        "eol",                doeol,                0, NULLCHAR,
                "eol options: unix, standard",
#ifndef        MSDOS
        "escape",        doescape,        0, NULLCHAR,        NULLCHAR,   
#endif
#ifdef        PC_EC
        "etherstat",        doetherstat,        0, NULLCHAR,        NULLCHAR,
#endif  PC_EC
        "exit",                doexit,                0, NULLCHAR,        NULLCHAR,
        "ftp",                doftp,                2, "ftp <address>",        NULLCHAR,
        "gateway",        dogateway,        0, NULLCHAR,        NULLCHAR,
#ifdef HAPN
        "hapnstat",        dohapnstat,        0, NULLCHAR,        NULLCHAR,
#endif HAPN
        "help",                dohelp,                0, NULLCHAR,        NULLCHAR,
        "hostname",        dohostname,        0, NULLCHAR,        NULLCHAR,
        "log",                dolog,                0, NULLCHAR,        NULLCHAR,
        "icmpstat",        doicmpstat,        0, NULLCHAR,        NULLCHAR,
        "ipaddr",        doipaddr,        0, NULLCHAR,        NULLCHAR,
        "ipstat",        doipstat,        0, NULLCHAR,        NULLCHAR,
        "smtp",                dosmtptick,        0, NULLCHAR,        NULLCHAR,
        "memstat",        memstat,        0, NULLCHAR,        NULLCHAR,
        "mss",                domss,                0, NULLCHAR,        NULLCHAR,
#ifdef        AX25
        "mycall",        domycall,        0, NULLCHAR,        NULLCHAR,
#endif
        "param",        doparam,        2, "param <interface>", NULLCHAR,
        "beacon",       dobeacon,       3, "beacon <interface> <timer>", NULLCHAR,
        "reset",        doreset,        0, NULLCHAR,        NULLCHAR,
        "route",        doroute,        0, NULLCHAR,        NULLCHAR,
        "session",        dosession,        0, NULLCHAR,        NULLCHAR,
#ifdef        SERVERS
        "start",        dostart,        2, "start <servername>",NULLCHAR,
        "stop",                dostop,                2, "stop <servername>",        NULLCHAR,
#endif
        "tcpstat",        dotcpstat,        0, NULLCHAR,        NULLCHAR,
        "tcpkick",        dotcpkick,        0, NULLCHAR,        NULLCHAR,
        "telnet",        dotelnet,        2, "telnet <address>",        NULLCHAR,
        "trace",        dotrace,        0, NULLCHAR,        NULLCHAR,
        "ttl",                dottl,                0, NULLCHAR,        NULLCHAR,
        "udpstat",        doudpstat,        0, NULLCHAR,        NULLCHAR,
        "window",        dowindow,        0, NULLCHAR,        NULLCHAR,
        "?",                dohelp,                0, NULLCHAR,        NULLCHAR,
        NULLCHAR,        NULLFP,                0,
                "Unknown command; type \"?\" for list",   NULLCHAR, 
};

/* "route" subcommands */
int doadd(),dodrop();
static struct cmds rtcmds[] = {
        "add", doadd, 3,
        "route add <dest addr>[/<bits>] <if name> [gateway] [metric]",
        "Add failed",

        "drop", dodrop, 2,
        "route drop <dest addr>[/<bits>]",
        "Not in table",

        NULLCHAR, NULLFP, 0,
        "route subcommands: add, drop",
        NULLCHAR, 
};

#ifdef        SERVERS
/* "start" and "stop" subcommands */
int ftp_start(),smtp_start(),discard_start(),echo_start(),telnet_start();
static struct cmds startcmds[] = {
        "discard",        discard_start,        0, NULLCHAR, NULLCHAR,
        "echo",                echo_start,        0, NULLCHAR, NULLCHAR,
        "ftp",                ftp_start,        0, NULLCHAR, NULLCHAR,
        "smtp",                smtp_start,        0, NULLCHAR, NULLCHAR,
        "telnet",        telnet_start,        0, NULLCHAR, NULLCHAR,
        NULLCHAR,        NULLFP,                0,
                "start options: discard, echo, ftp, smtp, telnet", NULLCHAR,
};
int ftp_stop(),smtp_stop(),echo_stop(),discard_stop(),telnet_stop();
static struct cmds stopcmds[] = {
        "discard",        discard_stop,        0, NULLCHAR, NULLCHAR,
        "echo",                echo_stop,        0, NULLCHAR, NULLCHAR,
        "ftp",                ftp_stop,        0, NULLCHAR, NULLCHAR,
        "smtp",                smtp_stop,        0, NULLCHAR, NULLCHAR,
        "telnet",        telnet_stop,        0, NULLCHAR, NULLCHAR,
        NULLCHAR,        NULLFP,                0,
                "stop options: discard, echo, ftp, smtp, telnet",  NULLCHAR,
};
#endif

main(argc,argv)
int argc;
char *argv[];
{
        static char inbuf[BUFSIZ];        /* keep it off the stack */
        int c;
        char *ttybuf,*fgets();
        int16 cnt;
        int ttydriv();
        int cmdparse();
        void check_time();
        FILE *fp;
        struct interface *ifp;
        struct mbuf *bp;

        ioinit();
        printf("KA9Q Internet Protocol Package, v%s.%s\n\l",major_rev,minor_rev);
        if((fp = fopen(STARTUP,"r")) != NULLFILE){
                while(fgets(inbuf,BUFSIZ,fp) != NULLCHAR){
                        cmdparse(cmds,inbuf);
                }
                fclose(fp);
        }                
        cmdmode();
        smtpclinit();                        /* arm SMTP background client timer */

        /* Main commutator loop */
        for(;;){
                /* Process any keyboard input */
                while((c = kbread()) != -1){
#ifdef        MSDOS
                        /* c == -2 means the command escape key (F10) */
                        if(c == -2){
                                if(mode != CMD_MODE){
                                        printf("\r\l");
                                        cmdmode();
                                }
                                continue;
                        }
#endif
                        if((cnt = ttydriv(c,&ttybuf)) == 0)
                                continue;
                        switch(mode){
                        case CMD_MODE:
                                (void)cmdparse(cmds,ttybuf);
                                fflush(stdout);
                                break;
                        case CONV_MODE:
#ifndef        MSDOS
                                if(ttybuf[0] == escape && escape != 0){
                                        printf("\r\l");
                                        cmdmode();
                                } else
#endif MSDOS
                                        if(current->parse != NULLFP)
                                                (*current->parse)(ttybuf,cnt);

                                break;
                        }
                        if(mode == CMD_MODE){
                                printf(prompt);
                                fflush(stdout);
                        }
                }
                /* Service the loopback queue */
                while((bp = dequeue(&loopq)) != NULLBUF)
                        ip_recv(bp,0);

                /* Service the interfaces */
                for(ifp = ifaces; ifp != NULLIF; ifp = ifp->next){
                        if(ifp->recv != NULLFP)
                                (*ifp->recv)(ifp);
                }

                /* Service the clock if it has ticked */
                check_time();

#ifdef        MSDOS
                /* Tell DoubleDos to let the other task run for awhile.
                 * If DoubleDos isn't active, this is a no-op
                 */
                giveup();
#else
                /* Wait until interrupt, then do it all over again */
                eihalt();
#endif
        }
}

/* Enter command mode */
int
cmdmode()
{
        if(mode != CMD_MODE){
                mode = CMD_MODE;
                cooked();
                printf(prompt);
                fflush(stdout);
        }
        return 0;
}
/* Select and display sessions */
static
dosession(argc,argv)
int argc;
char *argv[];
{
        unsigned i;
        struct session *s;
        extern char *tcpstates[];
        char *psocket();

        if(argc < 2){
                printf(" #  TCB  Type    State        Remote socket\r\l");
                for(s=sessions;s < & sessions[NSESSIONS];s++){
                        switch(s->type){
                        case TELNET:
                                printf("%c%-3d%4x Telnet  %-13s%-s:%d\r\l",
                                 (current == s)? '*':' ',
                                 s - sessions,
                                 (int)s->cb.telnet->tcb,
                                 tcpstates[s->cb.telnet->tcb->state],
                                 s->name,s->cb.telnet->tcb->conn.remote.port);
                                break;
                        case FTP:
                                printf("%c%-3d%4x FTP     %-13s%-s:%d\r\l",
                                 (current == s)? '*':' ',
                                 s - sessions,
                                 (int)s->cb.ftp->control,
                                 tcpstates[s->cb.ftp->control->state],
                                 s->name,s->cb.ftp->control->conn.remote.port);
                                break;
                        }
                }
                return 0;
        }
        i = atoi(argv[1]);
        if(i > NSESSIONS){
                printf("Invalid session: %d\r\l",i);
                return 1;
        }
        if(sessions[i].type == FREE){
                printf("Inactive session: %d\r\l",i);
                return 1;
        }
        current = &sessions[i];
        go();
        return 0;
}
/* Enter conversational mode with current session */
int
go()
{
        void rcv_char(),r_ctl();

        if(current == NULLSESSION || current->type == FREE)
                return 0;
        mode = CONV_MODE;
        switch(current->type){
        case TELNET:
                if(current->cb.telnet->remote[TN_ECHO])
                        raw();        /* Re-establish raw mode if it was set */
                rcv_char(current->cb.telnet->tcb,0); /* Get any pending input */
                break;
        case FTP:
                r_ctl(current->cb.ftp->control,0);
                break;
        }
        return 0;
}
#ifdef        AX25
extern int digipeat;
dodigipeat(argc,argv)
int argc;
char *argv[];
{
        if(argc == 1) {
                printf("digipeat %s\r\l",digipeat ? "on" : "off");
        } else {
                if(strcmp(argv[1],"on") == 0)
                        digipeat = 1;
                else
                        digipeat = 0;
        }
}
#endif
static
doipaddr(argc,argv)
int argc;
char *argv[];
{
        char *inet_ntoa();
        int32 n;

        if(argc < 2) {
                printf("%s\r\l",inet_ntoa(ip_addr));
        } else if((n = resolve(argv[1])) == 0){
                printf(badhost,argv[1]);
                return 1;
        } else
                ip_addr = n;
        return 0;
}
static
dogateway(argc,argv)
int argc;
char *argv[];
{
        char *inet_ntoa();
        extern int32 gateway;
        int32 n;

        if(argc < 2){
                printf("%s\r\l",inet_ntoa(gateway));
        } else if((n = resolve(argv[1])) == 0){
                printf(badhost,argv[1]);
                return 1;
        } else
                gateway = n;
        return 0;
}
static
doexit(argc,argv)
int argc;
char *argv[];
{
        iostop();
        exit(0);
}
static
doclose(argc,argv)
int argc;
char *argv[];
{
        struct session *s;
        unsigned int i;

        if(argc < 2){
                /* Default is current session */
                if(current == NULLSESSION || current->type == FREE){
                        printf("No current session\r\l");
                        return 1;
                }
                s = current;
        } else {
                i = atoi(argv[1]);
                if(i > NSESSIONS){
                        printf("Invalid session: %d\r\l",i);
                        return 1;
                }
                s = &sessions[i];
                if(s->type == FREE){
                        printf("Inactive session: %d\r\l",i);
                        return 1;
                }
        }
        switch(s->type){
        case TELNET:
                close_tcp(s->cb.telnet->tcb);
                break;
        case FTP:
                close_tcp(s->cb.ftp->control);
                break;
        }
        return 0;
}
static
doreset(argc,argv)
int argc;
char *argv[];
{
        long htol();
        struct tcb *tcb;

        if(argc < 2){
                if(current == NULLSESSION){
                        printf("No current session\r\l");
                        return 1;
                }
                switch(current->type){
                case TELNET:
                        tcb = current->cb.telnet->tcb;
                        break;
                case FTP:
                        tcb = current->cb.ftp->control;
                        break;
                }
        } else
                tcb = (struct tcb *)htol(argv[1]);
        if(!tcpval(tcb)){
                printf(notval);
                return 1;
        }
        close_self(tcb,RESET);
        return 0;
}
static
int
dotcpstat(argc,argv)
int argc;
char *argv[];
{
        long htol();
        register struct tcb *tcb;

        if(argc < 2){
                tcpstat();
        } else {
                tcb = (struct tcb *)htol(argv[1]);
                if(tcpval(tcb))
                        state_tcp(tcb);
                else
                        printf(notval);
        }
        return 0;
}
static
int
dotcpkick(argc,argv)
int argc;
char *argv[];
{
        long htol();
        register struct tcb *tcb;
        void tcp_timeout();

        if(argc < 2){
                if(current == NULLSESSION){
                        printf("No current session\r\l");
                        return 1;
                }
                switch(current->type){
                case TELNET:
                        tcb = current->cb.telnet->tcb;
                        break;
                case FTP:
                        tcb = current->cb.ftp->control;
                        break;
                }
        } else
                tcb = (struct tcb *)htol(argv[1]);

        if(!tcpval(tcb)){
                printf(notval);
                return 1;
        }
        if(argc > 2){
                /* Optional value for SRTT */
                tcb->srtt = atoi(argv[2]);
        }
        /* Don't actually do anything unless something is pending */
        if(tcb->sndcnt != 0){
                tcp_timeout((int *)tcb);
        }
        return 0;
}
static
int
dotrace(argc,argv)
int argc;
char *argv[];
{
        long htol();
#ifdef TRACE
        if(argc < 2)
                printf("trace level 0x%08lx\r\l",trace);
        else
                trace = htol(argv[1]);
        return 0;
#else
        printf("Trace not enabled - recompile with TRACE defined\r\l");
#endif
}
static
dohostname(argc,argv)
int argc;
char *argv[];
{
        char *strncpy();

        if(argc < 2)
                printf("%s\r\l",hostname);
        else 
                strncpy(hostname,argv[1],HOSTNAMELEN);
        return 0;
}
static
int
dolog(argc,argv)
int argc;
char *argv[];
{
        char *strncpy();

        static char logname[15];
        if(argc < 2){
                if(logfp)
                        printf("Logging to %s\r\l",logname);
                else
                        printf("Logging off\r\l");
                return 0;
        }
        if(logfp){
                fclose(logfp);
                logfp = NULLFILE;
        }
        if(strcmp(argv[1],"stop") != 0){
                strncpy(logname,argv[1],15);
                logfp = fopen(logname,"a+");
        }
        return 0;
}
#ifndef        MSDOS
static
int
doescape(argc,argv)
int argc;
char *argv[];
{
        if(argc < 2)
                printf("0x%x\r\l",escape);
        else 
                escape = atoi(argv[1]);
        return 0;
}
#endif        MSDOS
static
int
dohelp(argc,argv)
int argc;
char *argv[];
{
        register struct cmds *cmdp;
        int i,j;

        printf("Main commands:\r\l");
        for(i=0,cmdp = cmds;cmdp->name != NULL;cmdp++,i++){
                printf("%s",cmdp->name);
                if((i % 4) == 3)
                        printf("\r\l");
                else {
                        for(j=strlen(cmdp->name);j < 16; j++)
                                putchar(' ');
                }
        }
        if((i % 4) != 0)
                printf("\r\l");
        return 0;
}
static
int
domss(argc,argv)
int argc;
char *argv[];
{
        if(argc < 2)
                printf("%d\r\l",tcp_mss);
        else
                tcp_mss = atoi(argv[1]);
        return 0;
}
static
int
dowindow(argc,argv)
int argc;
char *argv[];
{
        if(argc < 2)
                printf("%d\r\l",tcp_window);
        else
                tcp_window = atoi(argv[1]);
        return 0;
}
static
int
dottl(argc,argv)
char *argv[];
{
        if(argc < 2)
                printf("%u\r\l",ip_ttl & 0xff);
        else
                ip_ttl = atoi(argv[1]);
        return 0;
}
struct session *
newsession()
{
        register int i;

        for(i=0;i<NSESSIONS;i++)
                if(sessions[i].type == FREE)
                        return &sessions[i];
        return NULLSESSION;
}
freesession(s)
struct session *s;
{
        if(s == NULLSESSION)
                return;
        s->type = FREE;
        if(s->name != NULLCHAR){
                free(s->name);
                s->name = NULLCHAR;
        }
                
}
/* Display and/or manipulate routing table */
int
doroute(argc,argv)
int argc;
char *argv[];
{
        if(argc < 2){
                dumproute();
                return 0;
        }
        return subcmd(rtcmds,argc,argv);
}
/* Add an entry to the routing table
 * E.g., "add 1.2.3.4 ax0 5.6.7.8 3"
 */
int
doadd(argc,argv)
int argc;
char *argv[];
{
        struct interface *ifp;
        int32 dest,gateway;
        unsigned bits;
        char *bitp,*index();
        int metric;

        if(strcmp(argv[1],"default") == 0){
                dest = 0;
                bits = 0;
        } else {
                if((dest = resolve(argv[1])) == 0){
                        printf(badhost,argv[1]);
                        return 1;
                }

                /* If IP address is followed by an optional slash and
                 * a length field, (e.g., 128.96/16) get it;
                 * otherwise assume a full 32-bit address
                 */
                if((bitp = index(argv[1],'/')) != NULLCHAR){
                        bitp++;
                        bits = atoi(bitp);
                } else
                        bits = 32;
        }
        for(ifp=ifaces;ifp != NULLIF;ifp = ifp->next){
                if(strcmp(argv[2],ifp->name) == 0)
                        break;
        }
        if(ifp == NULL){
                printf("Interface \"%s\" unknown\r\l",argv[2]);
                return 1;
        }
        if(argc > 3){
                if((gateway = resolve(argv[3])) == 0){
                        printf(badhost,argv[3]);
                        return 1;
                }
        } else {
                gateway = 0;
        }
        if(argc > 4)
                metric = atoi(argv[4]);
        else
                metric = 0;

        rt_add(dest,bits,gateway,metric,ifp);
        return 0;
}
/* Drop an entry from the routing table
 * E.g., "drop 128.96/16
 */
int
dodrop(argc,argv)
int argc;
char *argv[];
{
        char *bitp,*index();
        unsigned bits;
        int32 n;

        /* If IP address is followed by an optional slash and length field,
         * (e.g., 128.96/16) get it; otherwise assume a full 32-bit address
         */
        if((bitp = index(argv[1],'/')) != NULLCHAR){
                bitp++;
                bits = atoi(bitp);
        } else
                bits = 32;

        if((n = resolve(argv[1])) == 0){
                printf(badhost,argv[1]);
                return 1;
        }
        return rt_drop(n,bits);
}

#ifdef SERVERS
dostart(argc,argv)
int argc;
char *argv[];
{
        return subcmd(startcmds,argc,argv);
}
dostop(argc,argv)
int argc;
char *argv[];
{
        return subcmd(stopcmds,argc,argv);
}
#endif SERVERS

doecho(argc,argv)
int argc;
char *argv[];
{
        extern int refuse_echo;

        if(argc < 2){
                if(refuse_echo)
                        printf("Refuse\r\l");
                else
                        printf("Accept\r\l");
        } else {
                if(argv[1][0] == 'r')
                        refuse_echo = 1;
                else if(argv[1][0] == 'a')
                        refuse_echo = 0;
                else
                        return -1;
        }
        return 0;
}
/* set for unix end of line for remote echo mode telnet */
doeol(argc,argv)
int argc;
char *argv[];
{
        extern int unix_line_mode;

        if(argc < 2){
                if(unix_line_mode)
                        printf("Unix\r\l");
                else
                        printf("Standard\r\l");
        } else {
                if(strcmp(argv[1],"unix") == 0)
                        unix_line_mode = 1;
                else if(strcmp(argv[1],"standard") == 0)
                        unix_line_mode = 0;
                else {
                        return -1;
                }
        }
        return 0;
}
/* List of supported hardware devices */
#ifdef        PC_EC
int ec_attach();
#endif

int asy_attach();

#ifdef        PC100
int pc_attach();
#endif

#ifdef        EAGLE
int eg_attach();
#endif

#ifdef HAPN
int hapn_attach();
#endif

struct cmds attab[] = {
#ifdef        PC_EC
        /* 3-Com Ethernet interface */
        "3c500", ec_attach, 7, 
        "attach 3c500 <address> <vector> arpa <label> <buffers> <mtu>",
        "Could not attach 3c500",
#endif
        /* Ordinary PC asynchronous adaptor */
        "asy", asy_attach, 8, 
        "attach asy <address> <vector> slip|ax25 <label> <buffers> <mtu> <speed>",
        "Could not attach asy",
#ifdef        PC100
        /* PACCOMM PC-100 8530 HDLC adaptor */
        "pc100", pc_attach, 8, 
        "attach pc100 <address> <vector> ax25 <label> <buffers> <mtu> <speed>",
        "Could not attach pc100",
#endif
#ifdef        EAGLE
        /* EAGLE RS-232C 8530 HDLC adaptor */
        "eagle", eg_attach, 8,
        "attach eagle <address> <vector> ax25 <label> <buffers> <mtu> <speed>",
        "Could not attach eagle",
#endif
#ifdef        HAPN
        /* Hamilton Area Packet Radio (HAPN) 8273 HDLC adaptor */
        "hapn", hapn_attach, 8,
        "attach hapn <address> <vector> ax25 <label> <rx bufsize> <mtu> csma|full",
        "Could not attach hapn",
#endif
        NULLCHAR, NULLFP, 0,
        "Unknown device",
        NULLCHAR,
};

/* Attach an interface
 * Syntax: attach <hw type> <I/O address> <vector> <mode> <label> <bufsize> [<speed>]
 */
doattach(argc,argv)
int argc;
char *argv[];
{
        return subcmd(attab,argc,argv);
}
/* Manipulate I/O device parameters */
doparam(argc,argv)
int argc;
char *argv[];
{
        register struct interface *ifp;

        for(ifp=ifaces;ifp != NULLIF;ifp = ifp->next){
                if(strcmp(argv[1],ifp->name) == 0)
                        break;
        }
        if(ifp == NULLIF){
                printf("Interface \"%s\" unknown\r\l",argv[1]);
                return 1;
        }
        if(ifp->ioctl == NULLFP){
                printf("Not supported\r\l");
                return 1;
        }
        /* Pass rest of args to device-specific code */
        return (*ifp->ioctl)(ifp,argc-2,argv+2);
}
/* Log messages of the form
 * Tue Jan 31 00:00:00 1987 44.64.0.7:1003 open FTP
 */
/*VARARGS2*/
log(tcb,fmt,arg1,arg2,arg3,arg4)
struct tcb *tcb;
char *fmt;
int arg1,arg2,arg3,arg4;
{
        char *cp,*index();
        long t;

        if(logfp == NULLFILE)
                return;
        time(&t);
        cp = (char*) ctime(&t);
        rip(cp);
        fprintf(logfp,"%s %s - ",cp,psocket(&tcb->conn.remote));
        fprintf(logfp,fmt,arg1,arg2,arg3,arg4);
        fprintf(logfp,"\n");
        fflush(logfp);
}
/* Define null definitions for unused interfaces to avoid pulling in code
 * from the library
 */
#ifndef        ETHER
int ec_output() {}
int pether() {}
int gether() {}
char *ether_bdcst = NULLCHAR;
#endif ETHER

#ifndef AX25
int setcall() {}
int psax25() {}
struct ax25_addr ax25_bdcst,mycall;
#endif AX25


dobeacon(argc,argv)
int argc;
char *argv[];
{
        register struct interface *ifp;

        for(ifp=ifaces;ifp != NULLIF;ifp = ifp->next){
                if(strcmp(argv[1],ifp->name) == 0)
                        break;
        }
        if(ifp == NULLIF){
                printf("Interface \"%s\" unknown\r\l",argv[1]);
                return 1;
        }

        return ax25_beacon(ifp,argc-2,argv+2);
}

