/*
 *	Simple mail user interface for KA9Q IP/TCP package.
 *	A.D. Barksdale Garbee II, aka Bdale, N3EUA
 *	Copyright 1986 Bdale Garbee, All Rights Reserved.
 *	Permission granted for non-commercial copying and use, provided
 *	this notice is retained.
 *
 * revision history:
 *
 *  v3.1 870117 D. Trulli nn2z
 *	Add multiple arguments mail commands.
 *	Send to multiple users.
 *	Add status commands;
 *  v3.0 871225 D. Trulli nn2z
 *	Heavily restructured program to use a index array of message.
 *	More compatible with UNIX type mail commands.
 *  v2.7 871214 D. Trulli nn2z
 *	Cleaned up header and message parsing to prevent lock up.
 *	Revised command structure.
 *  v2.6 870826 Bdale
 *	integrate PA0GRI's interface/gateway to the WA7MBL PBBS
 *  v2.5 870713 Bdale
 *      integrate additional patches from PA0GRI, minor cleanup
 *  v2.4 870614 - P. Karn
 *      "smart gateway" function moved to smtp client code in net.exe.
 *	User interface for sending mail reworked to resemble Berkeley Mail
 *  v2.3 870524
 *      Extensive addition/revision by Gerard PA0GRI.  Now supports a healthy
 *      set of real commands.  
 *  v2.1,2.2
 *      exact change history lost.
 *  v2.0 c.870115
 *      First version with command parser, only send and quit work.
 *  v1.0
 *      First attempt.  Send messages only.
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "bm.h"

extern char version[];
static char copyright[] = 
"Copyright 1987 Bdale Garbee, permission granted for non-commercial use.";

/* comands valid in bm.rc */

struct token rccmds[] = {
	"smtp", SMTP,
	"host", HOST,
	"user", USER,
	"edit", EDIT,
	"fullname", NAME,
	"reply", REPLY,
	"maxlet", MAXLET,
	"mbox", MBOX,
	"record", RECORD,
	"screen", SCREEN,
	NULLCHAR, 0
};

FILE	*mfile = NULLFILE;
char 	*hostname = NULLCHAR;		/* name of this host from rc file */
char	*username = NULLCHAR;		/* name of this user from rc file */
char	*fullname = NULLCHAR;		/* fullname of this user from rc file */
char	*replyto = NULLCHAR;		/* address for reply-to header */
char	*maildir = NULLCHAR;		/* defined mail directory */
char	*savebox = NULLCHAR;		/* name of the mbox text file */
char	*record = NULLCHAR;		/* file to save a copy of outbound mail */
char	*editor = NULLCHAR;	/* user's favorite text editor */
char	notename[9];		/* name of current notesfile */
char	notefile[SLINELEN];	/* full path name of the mail text file */
char	*mfilename;		/* used for -f option name of mail file */
int	current;		/* the current message number */
int	nmsgs;			/* the number of messages in the notesfile */
int	newmsgs;		/* Number of new unread message */
int	change;			/* indicates that the mail file has been
				 * changed in this session */
unsigned maxlet = NLET+1;	/* max number of messages in mailbox */
int	tty;			/* tells if stdin is a tty */
struct let *mbox;


char usage[] = "Usage: bm [-u user] [ -f file]\n  or bm addr .. addr\n";
char badmsg[] = "Invalid Message number %d\n";
char nomail[] = "No mail for %s\n";
char noaccess[] = "Unable to access %s\n";

main(argc,argv)
int argc;
char *argv[];
{
	extern int optind;
	extern char *optarg;
	long	tmp;
	int	c;

#if	defined(__TURBOC__) || defined(MICROSOFT)
	fclose(stdaux);
	fclose(stdprn);
#endif

	getconfig();

	if (isatty(fileno(stdin))) {
		/* announce ourselves */
		screen_clear();
		printf(" %s\n%s\n\n",version,copyright);
		tty = 1;
	} else
		tty = 0;

		/* initialize globals from config file */
	current = 1;
	nmsgs = 0;

	/* check for important directories */
	if(access(maildir,0)){
		printf(noaccess,maildir);
		exit(1);
	}
	if(access(mailqdir,0)){
		printf(noaccess,mailqdir);
		exit(1);
	}
	strncpy(notename,username,8);
	notename[8] = '\0';
	mfilename = NULLCHAR;

	while ((c = getopt(argc,argv,"u:f:")) != -1) {
		switch(c) {
		case 'f':
			mfilename = optarg;
			break;
		case 'u':
			strncpy(notename,optarg,8);
			notename[8] = '\0';
			break;
		case '?':
			printf(usage);
			exit(1);
		}
	}

	/* set any signal handlers to catch break */
	setsignals();

	if(optind < argc){
		dosmtpsend(NULLFILE,&argv[optind],argc-optind,NULLCHAR);
		(void) exit(0);
	}

	tmp = (long)maxlet * (long)sizeof(struct let);
#ifdef	MSDOS
	/*
	* Since we are in the dos small model make sure that we
	* don't overflow a unsigned short on the number bytes
	* need for mallloc. If not checked malloc will
	* succeed and we will be trashing ourself in no time.
	*/
	if ((tmp & 0xffff0000) ||
	    (mbox=(struct let *)malloc((unsigned short)tmp)) == (struct let *)NULL){
#else
	if ((mbox=(struct let *)malloc((unsigned)tmp)) == (struct let *)NULL){
#endif
		fprintf(stderr,"Not enough memory to allocate table for %d messages\n",
		    maxlet);
		(void) exit(1);
	}

	sprintf(notefile,"%s/%s.txt",maildir,notename);
	if (mfilename == NULLCHAR) {
		if (lockit())
			exit(1);
		initnotes(notefile);
		rmlock(maildir,notename);
		listnotes(notefile);
	} else {
		initnotes(mfilename);
		listnotes(mfilename);
	}
	
	getcommand();
}

getconfig()
{
	FILE	*rcfp;	/* handle for the configuration file */
	char	rcline[LINELEN]; /* buffer for config file reading */
	register char *p;
#ifdef UNIX
	static char runcom[LINELEN];
	char *getenv();
#endif /* UNIX */

#ifdef UNIX
	/* Try to open $HOME/RUNCOM */
	if ((p = getenv("HOME")) != NULLCHAR) {
		sprintf(runcom, "%s/%s", p, RUNCOM);
		rcfp = fopen(runcom, "r");
	}

	/* If $HOME not defined or file missing
	 * try current directory
	 */
	if (p == NULLCHAR || rcfp == NULLFILE) {
		if ((rcfp = fopen(RUNCOM, "r")) == NULLFILE) {
			fprintf(stderr,"Cannot open '%s', check your installation\n", RUNCOM);
			(void) exit(1);
		}
	}
#else
	if((rcfp = fopen(RUNCOM,"r")) == NULLFILE) {	/* open config file */
		printf("Cannot open '%s', check your installation\n", RUNCOM);
		(void) exit(1);
	}
#endif
	while (!feof(rcfp)) {
		if (fgets(rcline,LINELEN,rcfp) == NULLCHAR)
			break;
		rip(rcline);
		if (*rcline == '\0' || *rcline == ';'|| *rcline == '#')
			continue;
		/* find the argument to the command */
		p = rcline;
		/* skip the command */
		while(*p && *p != ' ' && *p != '\t')
				p++;
		/* skip the white space */
		while(*p == ' ' || *p == '\t')
				p++;
		switch (rc_line_type(rcline)) {
		case HOST:
			hostname = savestr(p);
			break;
		case USER:
			username = savestr(p);
			break;
		case REPLY:
			replyto = savestr(p);
			break;
		case EDIT:
			editor = savestr(p);
			break;
		case SMTP:
			maildir = savestr(p);
			break;
		case NAME:
			fullname = savestr(p);
			break;
		case MAXLET:
			maxlet = atoi(p)+1;
			break;
		case MBOX:
			savebox = savestr(p);
			break;
		case RECORD:
			record = savestr(p);
			break;
		case SCREEN:
#if	defined(__TURBOC__)
			setvideo(p);
#endif
			break;
		default:
			fprintf(stderr,
			"%s: Invalid command: %s\n",RUNCOM,rcline);
			break;
		}
	}
	fclose(rcfp);
	if(maildir == NULLCHAR)
		maildir = mailspool;
	if(savebox == NULLCHAR)
		savebox = "mbox";
	if(hostname == NULLCHAR) {
		fprintf(stderr,"%s: hostname not set\n",RUNCOM);
		exit(1);
	}
	if(username == NULLCHAR) {
		fprintf(stderr,"%s: username not set\n",RUNCOM);
		exit(1);
	}
}

/* return the line_type from a line of the configuration file */
rc_line_type(s)
register char *s;
{
	register struct token *tp;
	for (tp = rccmds; tp->str != NULLCHAR; tp++) {
		if (strncmp(tp->str,s,strlen(tp->str)) == 0)
			return tp->type;
	}
	return (NONE);
}

/* replace terminating end of line marker(s) with null */
rip(s)
register char *s;
{
	for (; *s; s++) {
		if (*s == '\r' || *s == '\n') {
			*s = '\0';
			break;
		}
	}
}

/* list jobs wating to be sent in the mqueue */
listqueue()
{
	char tstring[80];
	char line[20];
	char host[SLINELEN];
	char to[SLINELEN];
	char from[SLINELEN];
	char *p;
	char	status;
	FILE *fp;

	printf("Outbound mail queue\n");
	printf("S Job      Host                 From                      To\n");
	filedir(mailqueue,0,line);
	while(line[0] != '\0') {
		sprintf(tstring,"%s/%s",mailqdir,line);
		if ((fp = fopen(tstring,"r")) == NULLFILE) {
			printf("Can't open %s\n",tstring);
			continue;
		}
		if ((p = strrchr(line,'.')) != NULLCHAR)
			*p = '\0';
		sprintf(tstring,"%s/%s.lck",mailqdir,line);
		if (access(tstring,0))
			status = ' ';
		else
			status = 'L';
		fgets(host,sizeof(host),fp);
		rip(host);
		fgets(from,sizeof(from),fp);
		rip(from);
		printf("%c %-8s %-20s %-25s",status,line,host,from);
		while (fgets(to,sizeof(to),fp) != NULLCHAR) {
			rip(to);
			printf(" %s",to);
		}
		printf("\n");
		fclose(fp);
		filedir(mailqueue,1,line);
	}
}

/* kill a job in the mqueue */
killjob(j)
char *j;
{
	char s[SLINELEN];
	char tbuf[SLINELEN];
	char *p;
	sprintf(s,"%s/%s.lck",mailqdir,j);
	p = strrchr(s,'.');
	if (!access(s,0)) {
		printf("Warning Job %s is locked by SMTP. Remove (y/n)? ",j);
		gets(tbuf);
		if (*tbuf != 'y')
			return;
		(void) unlink(s);
	}
	strcpy(p,".wrk");
	if (unlink(s))
		printf("Job id %s not found\n",j);
	strcpy(p,".txt");
	unlink(s);
}

/* copy a string return a pointer to it */
char *
savestr(s)
char *s;
{
	char *p;
	p = malloc(strlen(s)+1);
	if (p == NULLCHAR) 
		printf("No more memory\n");
	else
		strcpy(p,s);
	return p;
}

dohelp()
{
	screen_clear();
	printf("\n\n");
	printf(" d [msglist]           delete a message\n");
	printf(" m userlist            mail a message\n");
	printf(" s [msglist] [file]    save message in file (default mbox)\n");
	printf(" w [msglist] file      save message in file no header\n");
	printf(" f [msg]               forward message\n");
	printf(" r [msg]               reply to a message\n");
	printf(" u [msglst]            undelete a message\n");
	printf(" p                     print message on terminal\n");
	printf(" h                     display message headers in notefile\n");
	printf(" l                     list unsent messages\n");
	printf(" k                     kill unsent messages\n");
	printf(" n [file]              display or change notesfile\n");
	printf(" #                     where # is the number of message to read\n");
	printf(" x                     quit without changing mail file\n");
	printf(" q                     quit\n");
	printf(" ! cmd                 run dos command\n");
	printf(" ?                     print this help screen\n");
}


/* save a message list in a file in mailbox format */
savemsg(argc,argv)
int argc;
char *argv[];
{
	char *cp;
	int msgnum;
	int	i;
	FILE *tfile;

	if (mfile == NULLFILE){
		printf(nomail,notename);
		return;
	}
	if (argc == 0 || isdigit(*argv[argc - 1])) {
		cp = savebox;
	} else {
		cp = argv[argc - 1];
		--argc;
	}
	if ((tfile = fopen(cp,"a")) == NULLFILE) {
		printf("Can't open '%s'\n",cp);
		return;
	}
	if (argc == 0)
		msgtofile(current, tfile, 0);
	else {
		for (i = 0; i < argc; i++) {
			msgnum = atoi(argv[i]);
			if (msgnum >= 1 && msgnum <= nmsgs)
				msgtofile(msgnum, tfile, 0);
			else
				printf(badmsg,msgnum);
		}
	}
		fclose(tfile);
		printf("%s appended\n",cp);
}

writemsg(argc,argv)
int argc;
char *argv[];
{
	char *cp;
	int msgnum;
	int	i;
	FILE *tfile;

	if (mfile == NULLFILE){
		printf(nomail,notename);
		return;
	}
	if (argc == 0 || isdigit(*argv[argc - 1])) {
		printf("no file specified\n");
		return;
	} else {
		cp = argv[argc - 1];
		--argc;
	}
	if ((tfile = fopen(cp,"a")) == NULLFILE) {
		printf("Can't open '%s'\n",cp);
		return;
	}
	if (argc == 0)
		msgtofile(current, tfile, 1);
	else {
		for (i = 0; i < argc; i++) {
			msgnum = atoi(argv[i]);
			if (msgnum >= 1 && msgnum <= nmsgs)
				msgtofile(msgnum, tfile, 1);
			else
				printf(badmsg,msgnum);
		}
	}
	fclose(tfile);
	printf("%s appended\n",cp);
}

bmexit(x)
int x;
{
	if (mfilename == NULLCHAR) {
		if(lockit())
			exit(1);
		closenotes(notefile);
		rmlock(maildir,notename);
	} else {
		closenotes(mfilename);
	}
	screen_clear();
	exit(x);
}

/* this is the main command processing loop */
getcommand()
{
	FILE *tfile, *tmpfile();
	char	command[LINELEN];	/* command line */
	char	*args[MAXARGS];
	int	nargs;
	char	*cp;
	register int	msgnum;
	register int	i;

	printf("\nType ? for help.\n");

	/* command parsing loop */
	while(1) {
		printf("\"%s\"> ",notename);
		gets(command);

		if (feof(stdin))		/* someone hit ^Z! */
			strcpy(command,"q");	/* treat it like 'q' */
			
		if(*command == '!') {
			system(&command[1]);
			continue;
		}
		if (*command) {
			cp = command;
			while (*cp && *cp != ' ')
				cp++;
			nargs = parse(cp,args,MAXARGS);
		}

		switch (*command) {
		case 'm':		/* send msg */
			if (nargs != 0) {
				dosmtpsend(NULLFILE,args,nargs,NULLCHAR);
			} else {
				printf("To: ");
				gets(command);
				nargs = parse(command,args,MAXARGS);
				dosmtpsend(NULLFILE,args,nargs,NULLCHAR);
			}
			break;

		case 's':		/* save current msg to file */
			savemsg(nargs,args);
			break;

		case 'w':		/* write current msg to file */
			writemsg(nargs,args);
			break;

		case 'q':		/* quit */
			screen_clear();
			bmexit(0);
			/* NOTREACHED */
			break;
		case 'x':		/* abort */
			fclose(mfile);
			(void) exit(0);
			/* NOTREACHED */
			break;

		case 'p':		/* print message */
			displaymsg(current);
			break;

		case 'r':			/* reply */
			if (nargs == 0)
				msgnum = current;
			else
				msgnum = atoi(args[0]);
			if (msgnum >= 1 && msgnum <= nmsgs)
				reply(msgnum);
			else
				printf(badmsg,msgnum);
			break;

		case 'f':		/* forward a message */
			if(nargs == 0)
				msgnum = current;
			else 
				msgnum = atoi(args[0]);
			if (msgnum < 1 || msgnum > nmsgs) {
				printf(badmsg,msgnum);
				break;
			}
			if((tfile = tmpfile()) == NULLFILE)
				printf("\nCannot open temp file\n");
			else {
				msgtofile(msgnum,tfile,0);
				fseek(tfile,0L,0);
				printf("To: ");
				gets(command);
				nargs = parse(command,args,MAXARGS);
				dosmtpsend(tfile,args,nargs,NULLCHAR);
				fclose(tfile);
			}
			break;

		case 'u':
			if (nargs == 0)
				mbox[current].status &= ~DELETE;
			else
				for (i = 0; i < nargs; i++) {
					msgnum = atoi(args[i]);
					if( msgnum < 1 && msgnum > nmsgs)
						printf(badmsg,msgnum);
					else
						mbox[msgnum].status &= ~DELETE;
				}
			break;

		case 'l':		/* display unsent messages */
			listqueue();
			break;
		case 'k':
			if (nargs == 0)
				printf("No job id specified\n");
			else
				for (i = 0; i < nargs; i++) 
					killjob(args[i]);
			break;

		case 'n': {		/* display or change notefiles */
			char	line[80];
			if(nargs != 0) {
				if (mfilename == NULLCHAR) {
					if(lockit())
						break;
					closenotes(notefile);
					rmlock(maildir,notename);
				} else {
					closenotes(mfilename);
				}
				mfilename = NULLCHAR;
				strncpy(notename,args[0],8);
				notename[8] = '\0';
				sprintf(notefile,"%s/%s.txt",maildir,notename);
				if (lockit()) {
					mfile = NULLFILE;
					printf("Mail file is busy\n");
					break;
				}
				initnotes(notefile);
				rmlock(maildir,notename);
				listnotes(notefile);
			} else {  /* he wants to see what notefiles there are */
				sprintf(command,"%s/*.txt",maildir,notename);
				filedir(command,0,line);
				while(line[0] != '\0') {
					cp = strchr(line,'.');
					*cp = '\0';
					printf("notesfile -> %s\n",line);
					filedir(command,1,line);
				}
			}
			break;
		}

		case 'd':		/* delete a message */
			if (nargs == 0)
				delmsg(current);
			else
				for ( i = 0; i < nargs; i++) {
					msgnum = atoi(args[i]);
					if( msgnum >= 1 && msgnum <= nmsgs)
						delmsg(msgnum);
					else
						printf(badmsg,msgnum);
				}
			break;

		case 'h':		/* list message headers in notesfile */

			if (mfilename == NULLCHAR)
				listnotes(notefile);
			else
				listnotes(mfilename);
			break;

		case '\0':	/* a blank line prints next message */
			printnext();
			break;

		case '?':		/* help */
			dohelp();
			break;

		default:
			if (!isdigit(*command))
				printf("Invalid command - ? for help\n");
			else {
				msgnum = atoi(command);
				if (msgnum < 0 || msgnum > nmsgs)
					printf(badmsg,msgnum);
				else {
					current = msgnum;
					displaymsg(current);
				}
			}
			break;
		}
	}
}
