/*
   map.c: Copyright 1993, Paul Healy.

   Based substantially (reverse engineered) on MAIL2IDX, which is:
   (C) 1993 Johan. K. Reinalda, WG7J
   Freely usable for non-commercial use only !

   930714 : Created
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "buf.h"
#include "mailfile.h"
#include "header.h"
#include "misc.h"
#include "help.h"
#include "pager.h"
#include "mailer.h"
#include "buf.h"
#include "mailfile.h"
#include "message.h"
#include "bmh.h"
#include "rc.h"
#include "lock.h"

#ifdef BMH
#define main map_main
#endif

static FILE *fp = NULL;

static long
messageid(Header_t *h)
{
   char *s = HEADERP(h, RECEIVED);
   char *cp;

   /*
    * Received: from ei9gl.ampr.org by levin.ei9gl.ampr.org (NOS) with SMTP
    *     id AA13724 ; Fri, 16 Jul 93 17:54:44 UTC
    */
   if (s == NULL)
      return 0L;
   else if ((cp=strstr(s, "AA")) != NULL)  /* very NOS-centric... */
      return atol(cp+2);
   else
      return 0L;
}

static char *
from(Header_t *h, char *buf)
{
   char *f = repladdr(buf, HEADERP(h, REPLYTO), HEADERP(h, FROM));
   char *p = strchr(f, '@');

   if (p!=NULL)
      *p='\0';

   return f;
}

static char *
date(char *d, char *result)
{
   char *arg[MAXARGS];
   int nargs = parse(rip(d), arg, MAXARGS);

   if (nargs < 3)
      return "";
   else if (isdigit(arg[0][0]))  /* nntp style */
      sprintf(result, "%2d %s", atoi(arg[0]), arg[1]);
   else 
      sprintf(result, "%2d %s", atoi(arg[1]), arg[2]);

   return result;
}

static int
buffer(char *newline, long newoffset)
{
   static long oldoffset = 0L;
   static char *oldline = NULL;

   if (oldline == NULL) {
      if ((oldline = (char *) malloc(sizeof(char)*LINELEN)) == NULL) {
         fprintf(stderr, "map: no room\n");
         return -1;
         }
      }
   else {
      char num[10];

      if (newline != NULL)
         sprintf(num, "%7ld", newoffset - oldoffset);
      else 
         sprintf(num, "%7ld", newoffset - oldoffset + 2L); /* newoffset is mail_size which is 2 less than jnos offset */

      strncpy(&oldline[39], num, 7);

      if (fputs(oldline, fp) == EOF) {
         fprintf(stderr, "map: write failed\n");
         return -1;
         }
      }

   if (newline != NULL) {
      oldoffset = newoffset;
      strcpy(oldline, newline);
      }
   else {
      free(oldline);
      oldline = NULL;
      }

   return 0;
}

static MsgResult
mapmsg(Mailfile *m, char *line)
{
   Header_t h;
   char tmp[LINELEN], tmp2[LINELEN];

   if (gethdrb(m, &h) == -1)
      return MSG_QUIT;
   else {
      sprintf(line,
            /* msgid  BP R  TO       FROM   DATE   LEN   SUBJECT */
              "%5.5ld %c %c %-13.13s %-8.8s %-6.6s %6.6s %-35.35s\n",
              messageid(&h),
              HEADER(h, XBBSMSGTYPE) == NULL ? ' ' : HEADER(h, XBBSMSGTYPE)[0],
              HEADER(h, STATUS) == NULL ? 'N' : HEADER(h, STATUS)[0],
              HEADER(h, TO) == NULL ? "" : HEADER(h, TO),
              from(&h, tmp2),
              date(HEADER(h, DATE), tmp),
              "length",
              HEADER(h, SUBJECT) == NULL ? "" : HEADER(h, SUBJECT)
              );
      if (buffer(line, mail_offset(m)) == -1)
         return MSG_QUIT;
      else {
         mail_getmsg(m, line, LINELEN);
         return MSG_NEXT;
         }
      }
}

int
mapit(Mailfile *m)
{
   char *idx = getrc(index);

   if (*idx == 'n')
      return 0;
/*
   mail_resetpos(m);
*/
   if (mail_reset(m) == -1)
      return -1;
/*
   if (mail_lock(m) == -1) {
      fprintf(stderr, "map: can't lock %s\n", mail_filename(m));
      return -1;
      }
*/
   if ( (fp = fopen(mail_idxname(m), "w")) == NULL) {
      fprintf(stderr, "map: can't open %s\n", mail_idxname(m));
/*
      mail_unlock(m);
*/
      return -1;
      }
   else {
      walkall(m, mapmsg);
      buffer(NULL, mail_size(m));  /* mail size is too small compared to offsets? */
      fclose(fp);

      return 0 /* mail_unlock(m)*/ ;
      }
}

int
map(Mailfile *m, int argc, char *argv[])
{
   if (mail_getfile(m, argc, argv, "map") == -1)
      return -1;
   else 
      return mapit(m);
}

int
main(int argc, char *argv[])
{
   Mailfile m;

   if (dohelp(argc, argv, "map [+folder] (create index)"))
      return 0;

   mail_zero(&m);

   map(&m, argc, argv);

   return mail_savestate(&m);
}

