/*
   message.c: Copyright 1993, Paul Healy.

   930408 : Created
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "buf.h"
#include "mailfile.h"
#include "message.h"
#include "mailer.h"

#undef DEBUG

typedef struct {
   int argc;
   char **argv;
   int argc_active;
   int default_message;
   int fin;
   } Want;

typedef int (*WFN)(Want *wp, int msg);
static int walk(Mailfile *m, MsgFn fn, WFN wantfn, Want *wp);

/* ------------------------------------------------------------------------ */
/*
 * For things like: show 1
 */
static int
want1msg(Want *wp, int msg)
{
   int result = (wp->default_message == msg);

   if (result)
      wp->fin = 1;

#ifdef DEBUG
   fprintf(stderr, "bmh: want1msg called - msg = %d, result = %d\n", msg, result);
#endif

   return result;
}

int
walk2msg(Mailfile *m, MsgFn fn)
{
   Want w;

#ifdef DEBUG
   fprintf(stderr, "bmh: walk2msg called\n");
#endif

   w.default_message = mail_curmsg(m);
   w.argc = 0;
   w.argc_active = 0;
   w.fin = 0;

   return walk(m, fn, want1msg, &w);
}

/* ------------------------------------------------------------------------ */
/*
 * For things like: scan
 */
static int
wantall(Want *w, int m)
{
   (void)w; (void)m;
   return 1;
}

int
walkall(Mailfile *m, MsgFn fn)
{
   Want w;

   w.default_message = mail_curmsg(m);
   w.argc = 0;
   w.argc_active = 0;
   w.fin = 0;

   return walk(m, fn, wantall, &w);
}
/* ------------------------------------------------------------------------ */
int
wanted(int msg, int argc, char *argv[])
{
   int i;
   static char Done[] = "";

   for (i=1; i<argc; i++)
      if (isdigit(argv[i][0])) {
         int m = atoi(argv[i]);
         char *p = strchr(argv[i], '-');

         if (p != NULL) {
            if (msg >= m)
               if (msg <= atoi(p+1))
                  return 1;
               else
                  argv[i] = Done;
             }
         else if (m == msg) {
            argv[i] = Done;
            return 1;
            }
         }
      else if (strcmp(argv[i], "all") == 0)
         return 1;

   return 0;
}

/*
 * For things like: show 1 4 7-9
 */
static int
wantsel(Want *wp, int msg)
{
   int i;
   static char Done[] = "";

   for (i=1; i<wp->argc; i++)
      if (isdigit(wp->argv[i][0])) {
         int m = atoi(wp->argv[i]);
         char *p = strchr(wp->argv[i], '-');

         if (p != NULL) {
            if (msg >= m)
               if (msg <= atoi(p+1))
                  return 1;
               else {
                  wp->argv[i] = Done;
                  wp->argc_active--;
                  }
             }
         else if (m == msg) {
            wp->argv[i] = Done;
            wp->argc_active--;
            return 1;
            }
         }

   return 0;
}

int
walkselect(Mailfile *m, MsgFn fn, int argc, char *argv[])
{
   Want w;

   if ( (argc == 2) && (strcmp(argv[1], "all") == 0))
      return walkall(m, fn);

   w.argc = argc;
   w.argv = argv;
   w.fin = 0;
   w.argc_active = argc;

   return walk(m, fn, wantsel, &w);
}
/* ------------------------------------------------------------------------ */
/*
 * For things like: rmm 
 */
static int
wantnot1msg(Want *wp, int msg)
{
   int ret = !want1msg(wp, msg);

   wp->fin = 0;

   return ret;
}

int
walknot2msg(Mailfile *m, MsgFn fn, int msg)
{
   Want w;

   w.default_message = msg;
   w.argc = 0;
   w.argc_active = 0;
   w.fin = 0;

   return walk(m, fn, wantnot1msg, &w);
}
/* ------------------------------------------------------------------------ */
/*
 * For things like: rmm 1 4 7-9
 */
static int
wantnotsel(Want *wp, int msg)
{
   return !wantsel(wp, msg);
}

int
walknotselect(Mailfile *m, MsgFn fn, int argc, char *argv[])
{
   Want w;

   if ( (argc == 2) && (strcmp(argv[1], "all") == 0))
      return 0;

   w.argc = argc;
   w.argc_active = 0;  /* need to see all messages */
   w.argv = argv;
   w.fin = 0;

   return walk(m, fn, wantnotsel, &w);
}
/* ------------------------------------------------------------------------ */

static int
chkfound(Want *wp)
{
   int i;
   int found = 1;

   if (wp == NULL)
      return 0;

   for (i=1; i<wp->argc; i++) {
#ifdef DEBUG
      fprintf(stderr, "chkfound: checking %d = <%s>\n", i, wp->argv[i]);
#endif
      if (wp->argv[i][0] && (strchr(wp->argv[i], '-') == NULL) ) {
         fprintf(stderr, "%s %s", found ? "bmh: message(s)" : "", wp->argv[i]);
         found = 0;
         }
      }

   if (found == 0)
      fprintf(stderr, " not found\n");

   return found;
}

static int
walkfin(Want *w)
{
   return w->fin || (w->argc_active == 1);
}

/*
 * General mail file walker.
 * fn returns one of three values = MsgResult
 */
static int
walk(Mailfile *m, MsgFn fn, WFN wantfn, Want *wp)
{
   int msg = mail_curmsg(m);
   static char *line = NULL;
   int num = 0, res;

   if ( (line == NULL) && ( (line = (char *) malloc(LINELEN)) == NULL) ) {
      fprintf(stderr, "bmh: no room in walk\n");
      return -1;
      }
   else if (mail_refill(m) == -1) {
      fprintf(stderr, "bmh: can't read %s\n", mail_filename(m));
      return -1;
      }

   while ( (!walkfin(wp)) && (mail_getmsg(m, line, LINELEN) != -1)) {
#ifdef DEBUG
      fflush(stdout); fprintf(stderr, "walk: msg %d\n", num);
#endif
      while ( (!walkfin(wp)) && ismsg(line) ) {
         num++;
         mail_setcur(m, num);
         if (wantfn(wp, num)) {
            res = fn(m, line);
            if ((res==MSG_QUIT) || (res==MSG_EOF))
               wp->fin = 1;
            }
         else
            break;
         }
      }

   if (res != MSG_QUIT)
      chkfound(wp);

   return mail_setcur(m, msg);
}
