/*
   misc.c : Copyright Paul Healy, EI9GL, 1992.

   Derived from bm.

   Copyright 1986 Bdale Garbee, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.
   Copyright 1987 1988 Dave Trulli NN2Z, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.

   920603 : Created.
   920802 : Cleaned up parse a little.
   930113 : Added ':' and '=' to list of possible delimiters - allows
            variables in bm.rc like:
            alias: /home/aliasfile -- mh style or
            alias=/home/aliasfile
            (NB parse is now called all over the place...)
   930124 : moved get_msgid here to avoid makefile awkwardness
   930410 : Dropped ':' as a delimiter - too many problems with msdos
            paths including a drive.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <time.h>
#include <io.h>
#include <alloc.h>
#include "config.h"
#include "misc.h"
#include "rc.h"
#include "lock.h"
#include "buf.h"
#include "mailfile.h"
#include "header.h"
#include "mailer.h"

static const char Seperators[] = " \t,\n="; /* used to have ':' */

/*
 * parse a line into argv array. Return argc, -1 on error.
 */
int
parse(char *line, char *argv[], int maxargs)
{
   int argc = 0;

   while ( argc < maxargs ) {
      int qflag = 0;

      while (isspace(*line) || (*line == ',') )  /* Skip leading white space */
         line++;
      if(*line == '\0')
         break;

      if(*line == '"') {  /* Check for quoted token */
         line++;          /* Suppress quote */
         qflag = 1;
         }
      argv[argc++] = line;   /* Beginning of token */

      /*
       * Find terminating delimiter
       */

      if (qflag) {
         if ( (line = strchr(line, '"')) == NULL )
            return -1;
         *line++ = '\0';
         }
      else {
         char *cp;

         /* Find space or tab. If not present,
          * then we've already found the last
          * token.
          * 920802 : noted that bm would wrongly leave trailing tab chars
          *          after a token
          */
         if( (cp = strpbrk(line, Seperators)) == NULL)
            break;
         *cp++ = '\0';
         line = cp;
      }
   }
   if (argc == maxargs)
      return -1;

   argv[argc] = NULL;

   return argc;
}

/*
 * replace terminating end of line marker with null
 */
char *
rip(char *s)
{
   char *p = s;

   while (*s)
      if (*s == '\n')
         *s = '\0';
      else
         s++;

   return p;
}

#ifdef notdef
int
catchit(void)     /* dummy do nothing */
{
   return 1;      /* returning 0 causes the program to abort */
}
#endif

int
setupbm(void)
{
   static int done = 0;

   if (done)             /* don't keep loading config in bmh */
      return 0;
   else
      done = 1;

   if (loadconfig()==-1)
      return -1;

   if ((setvbuf(stdout, NULL, _IOFBF, STDOUT_BUFSIZE) != 0)) {
      fprintf(stderr, "bmh: can't buffer stdout\n");
      return -1;
      }

   atexit( (atexit_t) flushall); /* don't know why this is needed */

#ifdef __TURBOC__
   /* ctrlbrk(catchit); */
   fclose(stdprn);
   fclose(stdaux);
#endif
   if (signal(SIGINT, SIG_IGN) == SIG_ERR)
      perror("signal");

   return 0;
}

/*
 * Turbo C doesn't appear to use the timezone variable, despite what
 * the documentation says about localtime, tzset etc
 */
char *
gettz(char *tz)
{
   char *p;

   if ((p = getrc(zone)) != NULL)
      strcpy(tz, p);
   else if ( (p = getenv("TZ")) == NULL)
      strcpy(tz, "+0000 (GMT)");
   else if (strcmp(p, "GMT") == 0)
      strcpy(tz, "+0000 (GMT)");
   else if (strcmp(p, "UT") == 0)
      strcpy(tz, "+0000 (UT)");
   else if (strcmp(p, "UTC") == 0)
      strcpy(tz, "+0000 (UTC)");
   else if ((p[3]=='-') || (p[3]=='+') || isdigit(p[3]))
      sprintf(tz, "%c%02d00 (%.3s)",
         p[3] == '-' ? '-' : '+', abs(atoi(&p[3])), p);
   else {
      strncpy(tz, p, 3);
      tz[3] = 0;
      }

   return tz;
}

/*
 * get the time and date field as
 * "DAY day MONTH year hh:mm:ss ZONE"
 */
char *
ptime(long t)
{
   struct tm *ltm = localtime(&t);  /* does this set/use the timezone variable? */
   static char str[60];
   static char *days[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
   static char *months[] = {
      "Jan", "Feb", "Mar", "Apr", "May", "Jun",
      "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
   char tz[15];

   /*
    * rfc 822 format - now rfc 1123 for year
    */

   sprintf(str,"%s, %d %s %04d %02d:%02d:%02d %s\n",
      days[ltm->tm_wday],
      ltm->tm_mday,
      months[ltm->tm_mon],
      ltm->tm_year + (ltm->tm_year < 91 ? 2000 : 1900),
      ltm->tm_hour,
      ltm->tm_min,
      ltm->tm_sec,
      gettz(tz));

   return str;
}

/*
 * get a message id from the sequence file
 */
long 
get_msgid(char *mqueue)
{
   FILE *sfile = fopenlocked(mqueue, "sequence", ".seq", "r+");
   long sequence = 0L;

   /*
    *  if sequence file exists, get the value, otherwise set it
    */
   if (sfile != NULL) {
      if ( fscanf(sfile, "%ld", &sequence) == EOF)
         sequence = -1L;
      else {
         /*
          *  Keep it in range, use 8 digit number for dos name prefix.
          */
         if ( (sequence < 0L) || (sequence > 99999999L) )
            sequence = 0L;
         if (fseek(sfile, 0L, SEEK_SET) != 0)
            sequence = -1L;
         }
      }
   else {
      if ( (sfile = fopenlocked(mqueue, "sequence", ".seq", "w")) == NULL) {
         fprintf(stderr, "smtp: can't make %s/sequence.seq\n", mqueue);
         return -1L;
         }
      else 
         fprintf(stderr, "smtp: %s/sequence.seq created\n", mqueue);
      }

   if (sequence != -1L)
      fprintf(sfile, "%ld", ++sequence);
   else
      fprintf(stderr, "smtp: error reading %s/sequence.seq\n", mqueue);

   fcloselocked(sfile, mqueue, "sequence");

   return sequence;
}

static char *
getname(char *cp)
{
   char *cp1;

   if((cp = strchr(cp,'<')) == NULL)
      return NULL;
   cp++;                               /* cp -> first char of name */
   if( (cp1 = strchr(cp,'>') ) == NULL )
      return NULL;
   *cp1 = '\0';

   return cp;
}

char *
repladdr(char *buf, char *replyto, char *from)
{
   char *g;

   if (replyto != NULL)
      strcpy(buf, replyto);
   else if (from != NULL)
      strcpy(buf, from);
   else
      return strcpy(buf, "");

   decomment(buf);

   if ( (g = getname(buf)) != NULL)
      return g;          /* found <x@y> => x@y */
   else                  /* not <x@y> form */
      return buf;
}

char *
name(char *from)
{
   char *p, *q;

   if ( (p=strchr(from, '(')) != NULL) {
      if ( (q=strchr(p, ')')) != NULL)
         *q='\0';
      else
         return from;
      return name(p+1); 
      }
   if ( (p=strchr(from, '"')) != NULL) {       /* remove quotes if present */
      if ( (q=strchr(p+1, '"')) != NULL)
         *q='\0';
      else
         return from;
      return p+1;
      }
   return from;
}

static int oldNull;

void
null_init(void)
{
   oldNull = *(unsigned short *)NULL;
#ifdef __BORLANDC__
   if (heapfillfree(0x56AA) == _HEAPCORRUPT)
      fprintf(stderr, "null_init: heap has been corrupted\n");
#endif
}

void
null_check(char *s)
{
   if (*(unsigned short *)NULL != oldNull) {
      fprintf(stderr, "\nnull: location smashed at %s\n", s);
      abort();
      }

#ifdef __BORLANDC__
   if (heapcheck() == _HEAPCORRUPT)
      fprintf(stderr, "null_check: heap has been corrupted at %s\n", s);
   if (heapcheckfree(0x56AA) == _HEAPCORRUPT)
      fprintf(stderr, "null_check: heap has been corrupted (checkfree) at %s\n", s);
#endif
}


