/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */
struct VCS *clear_vcs(vc)
register struct VCS *vc;
{
	static struct VCS *ret, *last, *ptr;
	static struct VCS *vcp;

	ret=NULL;
	last=NULL;
	while (vc != NULL) {
		if (vc->lcn != 255) {
			vcp=vc->peer;
			if (vcp != NULL) {
				vcp->peer=NULL;
				set_p(vcp, P6,0); /* need better diag */
				vc->peer=NULL;
			}
			ptr=vc->next;
			set_p(vc, P1, 0);
			vc=ptr;
		}
		else {
			if (last==NULL) /* first one */ {
				ret=vc;
				last=vc;
			}
			else /* Got another */ {
				last->next=vc;
				last=vc;
			}
			vc=last->next;
			last->next=NULL;
		}
	}
	return ret;
}

process_calls(lnkx)
struct axcb *lnkx;
{
	static struct VCS *vcn;
	register struct VCS *vcp;
	static struct axcb *lnk;

	while (lnk->llcn->lcn == 255) /* First on list is pending */ {
		vcn=lnk->llcn;
		lnk->llcn=vcp->next;
		vcn->next=NULL;
		set_p(vcn,P2,0); /* Issue call request if you can */
	} /* The above loop will loop until there is 1 successful call */
	vcp=lnk->llcn;
	while (vcp->next != NULL) /* There are more calls to do */ {
		if (vcp->next->lcn == 255) /* Got a call to process */ {
			vcn=vcp->next;
			vcp->next=vcn->next;
			vcn->next=NULL;
			set_p(vcn,P2,0);
		}
		vcp=vcp->next;
	}
}

adr_copy(src,istrx,des,ostrx)
int src;
unsigned char istrx[16];
int des;
unsigned char ostrx[16];

{
	static int ipos, opos, ilow, olow, len, i, acc;
	static unsigned char *istr, *ostr;

	istr=istrx;
	ostr=ostrx;
	if(src) /* Copy High order call (calling address) */ {
		ilow=istr[0]&1;
		ipos= ( (istr[0] & 0x0f) + ilow) >> 1;
		len=(istr[0] & 0x0f0)>>4;
		if (!ilow) ipos++;
	}
	else /* Copy low order call (called address) */ {
		ipos=1;
		ilow=0;
		len=istr[0] & 0x0f;
	}
	if (des) /* Copy to high order call (calling) */ {
		olow=ostr[0] & 1;
		opos=( (ostr[0] & 0x0f) + olow) >> 1;
		ostr[0]=ostr[0] & 0x0f | len<<4;
		if (!olow) opos++;
	}
	else /* Copy to low order call (called) */ {
		olow=0;
		opos=1;
		ostr[0]=ostr[0] & 0xf0 | len;
	}
	acc=0;
	if (olow)
		acc=ostr[opos] & 0xf0; /* Init acc */
	for (i=1;i<=len;i++) {
		if (ilow) acc += istr[ipos++] & 0x0f;
		else acc += (istr[ipos] & 0x0f0) >> 4;
		ilow = !ilow;
		if (olow) /* Output a byte */ {
			ostr[opos++]=acc;
			acc=0;
		}
		else acc=acc<<4;
		olow = !olow;
	}
	if (olow) ostr[opos]=acc;
}

unsigned char
glcnlow(lnk)	/* Get lcn, return 0 if all in use */
struct axcb *lnk;
{
	static unsigned char i;
	register struct VCS *p;

	i=1;
	p=lnk->llcn;
	while (p) /* There are more vc's */ {
		if (p->lcn != i) /* Found a Hole ! */ return i;
		i++;
		p=p->next;
	}
	if (i<=lnk->maxvc) return i;
	return 0;
}

unsigned char
glcnhigh(lnkx)	/* Get lcn, return 0 is all in use */
struct axcb *lnkx;
{
	static unsigned char i, j, max;
	register struct VCS *p;
	static struct axcb *lnk;

	lnk=lnkx;
	max=lnk->maxvc;
	p=lnk->llcn;
	if (!p) /* No vc's yet */ return max;
	if (p->lcn == 255) /* No active vc's yet */ return max;
	j=0;
	i=1;
	while (p) /* There are more to check */ {
		if (p->lcn != i) /* Found Hole */ {
			i=p->lcn;
			j=i-1; /* Save highest lcn # in hole */
		}
		i++;
		p=p->next;
		if (p) if (p->lcn == 255) break; /* Pending calls are at end */
	}
	if (i <= lnk->maxvc) return max;
	return j;
}

insert_lcn(vcx)	/* Insert a vc onto it's parent */
struct VCS *vcx;
{
	register struct VCS *p;
	static unsigned char i;
	static struct VCS *vc;
	static struct axcb *lnk;

	vc=vcx;
	lnk=vc->parent;
	i=vc->lcn;
	p=lnk->llcn;
	vc->next=NULL;	/* Just in case... */
	if (!p) /* No VC's yet, it's easy... */ {
		lnk->llcn=vc;
		return;
	}
	if (p->lcn > i) /* Insert at head */ {
		vc->next=p;	
		lnk->llcn=vc;
		return;
	}
	while (p->next) /* More to check */ {
		if (p->next->lcn > i) /* Insert before the next */ {
			vc->next=p->next;
			p->next=vc;
			return;
		}
		p=p->next;
	}
	p->next=vc;	/* Fell through, append this guy */
}

addlcn(vc)
register struct VCS *vc;

{
	static struct axcb *p;

	if (p=vc->parent) {
		if (p->am_dce) vc->lcn=glcnlow(p);
		else vc->lcn=glcnhigh(p);
		if (vc->lcn) /* Not zero, no error */ {
			insert_lcn(vc);
			return TRUE;
		}
	}
	return FALSE;
}

dellcn(vc)
register struct VCS *vc;
{
	static struct VCS **l;
	static struct VCS *vc1;
	extern struct VCS *pending_calls;

	if (!vc->parent) /* on pending_calls */ l= &pending_calls;
	else l= &vc->parent->llcn;
	if (*l) /* Not NULL */ {
		vc1= *l;
		if (vc1==vc) *l=vc->next;
		else {
			while (vc1) /* Not NULL */ {
			   if (vc1->next==vc) break;
			   vc1=vc1->next;
			}
		if (vc1) vc1->next=vc->next;
		}
	}
	vc->next=NULL;
	vc->parent=NULL;
}

