/*
   nntp.c : Paul Healy, EI9GL.

   920816 : Added this header, bmh version.
            Todo:
               . check that the date is ok for nntp
               . generally check file writes for success
   921108 : moved makepath routines to mkpath.c
   930114 : cleaned up header generation a bit.

 * NNTP Server/Client - See RFC977
 * Jeffrey R. Comstock. - NR0D - Bloomington, Minnesota USA
 * Copyright 1990 Jeffrey R. Comstock, All Rights Reserved.
 * Permission granted for non-commercial copying and use, provided
 * this notice is retained.
 *
 * DB3FL 9107xx: heavily rewritten and bug fixing in file-handling
 *
 * ported to NOS by PE1NMB - 920120
 *
 * Mods by PA0GRI and WG7J
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <string.h>
#include <limits.h>
#include "config.h"
#include "buf.h"
#include "mailfile.h"
#include "pc.h"
#include "mailer.h"
#include "send.h"
#include "header.h"
#include "nntp.h"
#include "lock.h"
#include "rc.h"
#include "misc.h"
#include "mkpath.h"

int
getnnrange(char *group, long *high, long *low)
{
   char directory[MAXNAMLEN];
   char filespec[MAXNAMLEN];
   char filename[MAXNAMLEN];
   
   if (getnnpath(group, directory, MAXNAMLEN) == NULL)
      return -1;

   *high = 0L;
   *low = LONG_MAX;

   sprintf(filespec, "%s/*.*", directory);
   filedir(filespec, 0, filename);

   while(filename[0] != '\0') {
      long num;
      
      if (isdigit(filename[0]))
         num = atol(filename);
      else
         continue;

      if (num > *high)
         *high = num;
      if (num < *low)
         *low = num;

      filedir(filespec, 1, filename);
      }

   /*
    * An empty newsgroup has 0 for the highest article and 1 for first
    * article:             HHHHH LLLLL P
    * empty-newsgroup.test 00000 00001 y
    */
   if ((*high == 0L) || (*low < 1L))
      *low = 1L;

   return 0;
}

/*
 * paths are for pseudo symbolic links
 */
static int
addpath(char *group)
{
   char path[MAXNAMLEN];
   FILE *fp = fopenlocked(getrc(newsdir), "pointer", "", "a");

   if (fp == NULL) {
      fprintf(stderr, "nntp: can't open %s/pointer\n", getrc(newsdir));
      return -1;
      }

   sprintf(path, "%s/%s", getrc(newsdir), group);

   fprintf(fp, "%s %s\n", group, news2dir(path));  /* check for errors here */
   fcloselocked(fp, getrc(newsdir), "pointer");

   return makepath(path);
}

char *
getnnpath(char *group, char *path, int maxpath)
{
   FILE *fp = fopenlocked(getrc(newsdir), "pointer", "", "r");
   int grplen = strlen(group);
   char *result = path;

   sprintf(path, "%s/junk", getrc(newsdir));

   if (fp == NULL)
      return result;

   while (fgets(path, maxpath, fp) != NULL)
      if (strncmp(group, path, grplen) == 0) {
         rip(path);
         result = &path[grplen+1];
         break;
         }

   fcloselocked(fp, getrc(newsdir), "pointer");

   return result;
}

static int
addhistory(long id, char *group, long num, char *msgid)
{
   long currtime = time(NULL);
   struct tm *stm = localtime(&currtime);
   FILE *fp = fopenlocked(getrc(newsdir), "history", "", "a");

   if (fp == NULL) {
      fprintf(stderr, "nntp: error opening %s/history\n", getrc(newsdir));
      return -1;
      }

   if (id == 0L)
      if (*msgid == '<')
         fputs(msgid, fp);
      else
         fprintf(fp, "<%s>", msgid);
   else 
      fprintf(fp, "<%ld@%s>", id, getrc(hostname));

   fprintf(fp, " %2.2d%2.2d%2.2d %2.2d%2.2d%2.2d %s/%ld\n",
      stm->tm_year, stm->tm_mon+1, stm->tm_mday,
      stm->tm_hour, stm->tm_min, stm->tm_sec,
      group, num);

   return fcloselocked(fp, getrc(newsdir), "history");
}

static long
updateactivelist(char *group)
{
   FILE *activefp = fopenlocked(getrc(newsdir), "active", "", "r+");
   FILE *tmpfp;
   long number = 0L;
   char buf[LINELEN], filename[MAXNAMLEN];
   int grplen = strlen(group);

   if ( (activefp == NULL) &&
        ( (activefp = fopenlocked(getrc(newsdir), "active", "", "w+")) == NULL)
      ) {
      fprintf(stderr, "post: can't open active file\n");
      return -1L;
      }

   if ((tmpfp = tempfile("post", filename, "w")) == NULL) {
      fcloselocked(activefp, getrc(newsdir), "active");
      fprintf(stderr, "post: can't open temporary file\n");
      return -1;
      }

   while (fgets(buf, sizeof(buf), activefp) != NULL)
      if (strncmp(group, buf, grplen) == 0) {
         long low;
         char posting, moderated = ' ';

         sscanf(&buf[grplen+1], "%ld %ld %c %c",
            &number, &low, &posting, &moderated);
         fprintf(tmpfp,"%s %5.5ld %5.5ld %c", group, ++number, low, posting);
         if (moderated != ' ')
            fprintf(tmpfp," %c\n", moderated);
         else 
            fprintf(tmpfp,"\n", moderated);
         }
      else
         fputs(buf, tmpfp);

   if (number == 0L) {
      addpath(group);
      number = 1L;
      fprintf(tmpfp, "%s 00001 00001 y\n", group);
      }

   rewind(tmpfp);
   rewind(activefp);
   dupfile(tmpfp, activefp);

   fcloselocked(activefp, getrc(newsdir), "active");
   fclose(tmpfp);
   unlink(filename);

   return number;
}

static long
xferart(FILE *fpin, char *group)
{
   long num;

   if ((num = updateactivelist(group)) > 0L) {
      char Num[20], pathname[MAXNAMLEN];
      char *path;
      FILE *fp;

      sprintf(Num, "%ld", num);
      path = getnnpath(group, pathname, sizeof(pathname));

      if ( (fp = fopenlocked(path, Num, "", "w")) == NULL) {
         fprintf(stderr, "post: can't open %s/%s\n", path, Num);
         return -1L;
         }
      dupfile(fpin, fp);
      fcloselocked(fp, path, Num);

      return num;
   }
   else {
      fprintf(stderr, "post: can't update active list\n");
      return -1L;
      }
}

int
postit(FILE *fpin, Header_t *h)
{
   char filename[MAXNAMLEN];
   long msgid;
   FILE *fp;
   long ret;

   if (makepath(getrc(newsdir)) == -1)
      return -1;

   if ((fp = tempfile("post", filename, "w")) == NULL)
      return -1;

   /*
    * Path:
    */
   fprintf(fp, "Path: %s!not-for-mail\n",
      getrc(newshost) != NULL ? getrc(newshost) : getrc(hostname));

   /*
    * Newsgroups:
    */
   fprintf(fp, "%s%s\n", Hdr[NEWSGROUPS],
      HEADERP(h, NEWSGROUPS) == NULL ? "ampr.test" : HEADERP(h, NEWSGROUPS));

   /*
    * Organization:
    */
   if (getrc(organ) != NULL)
      fprintf(fp, "Organization: %s\n", getrc(organ));

   if ((msgid=putrfc822(fp, h, NULL, NULL)) == -1L) {
      fprintf(stderr, "post: error writing mail headers\n");
      fclose(fp);
      unlink(filename);
      return -1;
      }

   dupfile(fpin, fp);
   rewind(fp);

   if ((ret=xferart(fp, HEADERP(h, NEWSGROUPS))) == -1L)
      fprintf(stderr, "post: failed to send article\n");
   else
      addhistory(msgid, HEADERP(h, NEWSGROUPS), ret, HEADERP(h, MESSAGEID));

   fclose(fp);
   unlink(filename);

   return 0;
}
