/* NET/ROM header tracing routines
 * Copyright 1991 Phil Karn, KA9Q
 */
#include <stdio.h>
#include "global.h"
#include "mbuf.h"
#include "netrom.h"
#include "nr4.h"
#include "trace.h"

/* Display NET/ROM network and transport headers */
void
netrom_dump(ip,bpp,check)
struct iface  *ip;
struct mbuf **bpp;
int check;
{
	char src[AXALEN],dest[AXALEN];
	char tmp[AXBUF];
	char thdr[NR4MINHDR];
	register i;

	if(bpp == NULLBUFP || *bpp == NULLBUF)
		return;
	/* See if it is a routing broadcast */
	if(uchar(*(*bpp)->data) == NR3NODESIG) {
		(void)PULLCHAR(bpp);		/* Signature */
		pullup(bpp,tmp,ALEN);
		tmp[ALEN] = '\0';
		prtTrace(ip,"NET/ROM Routing: %s\n",tmp);
		for(i = 0;i < NRDESTPERPACK;i++) {
			if (pullup(bpp,src,AXALEN) < AXALEN)
				break;
			prtTrace(ip,"        %12s",pax25(tmp,src));
			pullup(bpp,tmp,ALEN);
			tmp[ALEN] = '\0';
			prtTrace(ip,"%8s",tmp);
			pullup(bpp,src,AXALEN);
			prtTrace(ip,"    %12s",pax25(tmp,src));
			tmp[0] = PULLCHAR(bpp);
			prtTrace(ip,"    %3u\n",uchar(tmp[0]));
		}
		return;
	}
	/* Decode network layer */
	pullup(bpp,src,AXALEN);
	prtTrace(ip,"NET/ROM: %s",pax25(tmp,src));

	pullup(bpp,dest,AXALEN);
	prtTrace(ip,"->%s",pax25(tmp,dest));

	i = PULLCHAR(bpp);
	prtTrace(ip," ttl %d\n",i);

	/* Read first five bytes of "transport" header */
	pullup(bpp,thdr,NR4MINHDR);
	switch(thdr[4] & NR4OPCODE){
 	case NR4OPPID:	/* network PID extension */
		if (thdr[0] == NRPROTO_IP && thdr[1] == NRPROTO_IP) {
 			ip_dump(ip,bpp,check) ;
			return;
		}
 		else
 			prtTrace(ip,"         protocol family %x, proto %x",
			 uchar(thdr[0]), uchar(thdr[1])) ;
 		break ;
	case NR4OPCONRQ:	/* Connect request */
		prtTrace(ip,"         conn rqst: ckt %d/%d",uchar(thdr[0]),uchar(thdr[1]));
		i = PULLCHAR(bpp);
		prtTrace(ip," wnd %d",i);
		pullup(bpp,src,AXALEN);
		prtTrace(ip," %s",pax25(tmp,src));
		pullup(bpp,dest,AXALEN);
		prtTrace(ip,"@%s",pax25(tmp,dest));
		break;
	case NR4OPCONAK:	/* Connect acknowledgement */
		prtTrace(ip,"         conn ack: ur ckt %d/%d my ckt %d/%d",
		 uchar(thdr[0]), uchar(thdr[1]), uchar(thdr[2]),
		 uchar(thdr[3]));
		i = PULLCHAR(bpp);
		prtTrace(ip," wnd %d",i);
		break;
	case NR4OPDISRQ:	/* Disconnect request */
		prtTrace(ip,"         disc: ckt %d/%d",
		 uchar(thdr[0]),uchar(thdr[1]));
		break;
	case NR4OPDISAK:	/* Disconnect acknowledgement */
		prtTrace(ip,"         disc ack: ckt %d/%d",
		 uchar(thdr[0]),uchar(thdr[1]));
		break;
	case NR4OPINFO:	/* Information (data) */
		prtTrace(ip,"         info: ckt %d/%d",
		 uchar(thdr[0]),uchar(thdr[1]));
		prtTrace(ip," txseq %d rxseq %d",
		 uchar(thdr[2]), uchar(thdr[3]));
		break;
	case NR4OPACK:	/* Information acknowledgement */
		prtTrace(ip,"         info ack: ckt %d/%d",
		 uchar(thdr[0]),uchar(thdr[1]));
		prtTrace(ip," txseq %d rxseq %d",
		 uchar(thdr[2]), uchar(thdr[3]));
		break;
	default:
 		prtTrace(ip,"         unknown transport type %d",
		 thdr[4] & 0x0f) ;
		break;
	}
	if(thdr[4] & NR4CHOKE)
		prtTrace(ip," CHOKE");
	if(thdr[4] & NR4NAK)
		prtTrace(ip," NAK");
	if(thdr[4] & NR4MORE)
		prtTrace(ip," MORE");
	prtTrace(ip, "\n");
}

