/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */
/* This file has the operating environment routines needed for task scheduling

	a task can be scheduled to run asap or after a timer has expired
	timers (or a scheduled task) can be killed
		if the timer/task parameter block is known */

struct tmrblk *taskq, *timerq;
char *malloc();

#include "timer.h"
#include "data.h"
#include "buffer.h"
#include "ax25.h"
#include "ax25l2.h"
#include "l3struc.h"

#include "buffer.c"

#include "memory.c"

int
task()
{
	register struct tmrblk *t;

	if (!(t=taskq)) {
		space_used();	/* Have nothing to do, collect garbage */
		return FALSE;	/* Nothing waiting */
	}
	t=taskq;
	taskq = t->next;
	if (taskq) taskq->prev = t->prev;
	t->prev = NULL;
	t->next = NULL;
	if (t->back)
		if (*(t->back) == t) *(t->back)=NULL;
	t->back=NULL;
	(*t->task)(t->param);
	free(t);
	OS_QUE(task,0);
	return TRUE; /* This says we executed something */
}

void
queue(ptrx,rtn,delay,par)
struct tmrblk **ptrx;
char (*rtn)();
int delay;
int par;

{
	register struct tmrblk *p;
	static struct tmrblk *p1;
	static struct tmrblk **ptr;

	ptr=ptrx;

	p=(struct tmrblk *)malloc(sizeof(struct tmrblk));

	if (p==NULL) return; /* We are dead */
	if (delay < 0) /* Illegal request assume 0 */ delay=0;
	p->task=rtn;
	p->param=par;
	p->delay=delay;
	p->prev=NULL;
	p->next=NULL;
	p->back=NULL;

	if (ptr) {
		p->back=ptr;
		if (*ptr) kill(*ptr);	/* A timer was running, kill it */
		*ptr=p;
	}

	if (delay == 0) /* Do this when we get a chance */ {
		qtask(p);
		return;
	}
	/* It's a timer request, add it to the timer queue */

	if (timerq == NULL) /* Timer Chain is empty */ {
		p->prev=p;
		timerq=p;
		return;
	}
	/* There are things waiting */
	if (timerq->delay > p->delay) /* New Top */ {
		timerq->delay -= p->delay;
		p->prev=timerq->prev;
		timerq->prev=p;
		p->next=timerq;
		timerq=p;
		return;
	}
	/* Insert this guy in correct pos */
	p1=timerq;
loop:
	p->delay -= p1->delay;
	if (p1->next == NULL) {	/* New bottom */
		p->prev=p1;
		p1->next=p;
		timerq->prev=p;
		return;
	}
	else /* Before next? */ {
		p1=p1->next;
		if (p1->delay > p->delay) /* goes before p1 */ {
			p1->delay -= p->delay;
			p->prev=p1->prev;
			p->next=p1;
			p->prev->next=p;
			p1->prev=p;
			return;
		}
	}
	goto loop;
}

void
kill(pblk)
register struct tmrblk *pblk;
{
	if (pblk == NULL) return;

	if (pblk == timerq) /* The next timer to expire */ {
		timerq=pblk->next;
		if (timerq) /* and wasn't the only item */ {
			timerq->prev=pblk->prev;
			timerq->delay += pblk->delay;
		}
	}
	else {
		if (pblk == taskq) /* The next task, just made it */ {
			taskq=pblk->next;
			if (taskq) taskq->prev=pblk->prev;
		}
		else /* It's in the middle of a list */ {
			if (pblk->next) /* Not the last item */ {
				pblk->next->delay += pblk->delay;
				pblk->next->prev=pblk->prev;
				pblk->prev->next=pblk->next;
			}
			else /* The bottom of a queue */ {
				pblk->prev->next=NULL;
				if (taskq) if (taskq->prev == pblk)
							taskq->prev=pblk->prev;
				if (timerq) if (timerq->prev == pblk)
							timerq->prev=pblk->prev;
			}
		}
	}
	if (pblk->back) if (*(pblk->back) == pblk) *(pblk->back)=NULL;
	pblk->back=NULL;
	free(pblk);
	return;
}

/* If there is a need to have auxilary timers/tasks running can have a queue
   or a list of maintance routines that get run... */

service_timers()
{
	register struct tmrblk *p;
	static int i;

	if (timerq != NULL) /* Decrement the top item */ {
		timerq->delay--;
		if (timerq->delay == 0) /* We have tasks to schedule */ {
			while (timerq != NULL) {
				if (timerq->delay == 0) /* This is ready */ {
					p=timerq;
					timerq=timerq->next;
					if (timerq) timerq->prev=p->prev;
					qtask(p);
				}
				else break;
			}
		}
	}
}

qtask(t)
register struct tmrblk *t;
{
	if (taskq == NULL) /* Nothing is waiting */ {
		t->next=NULL;
		t->prev=t;	/* taskq->prev always points to last item */
		taskq=t;	/* even if it is also the first */
		OS_QUE(task,0);
	}
	else /* Queue up behind the last guy */ {
		t->prev=taskq->prev;
		t->prev->next=t;	/* Make last guy's next point to me */
		taskq->prev=t;
		t->next=NULL;
	}
}
