/* OS- and machine-dependent stuff for OSK machines */

#include <time.h>
#include <stdio.h>
#include <sgstat.h>
#include <modes.h>
#include <dir.h>


#include "machdep.h"
#include "mbuf.h"
#include "internet.h"
#include "iface.h"
#include "cmdparse.h"
#include "pc.h"

static char fnamebuffer[100];
char *makefname(env_root, name)
char *env_root, *name;
{
	strncpy(fnamebuffer, getenv(env_root), 100);
	strcat(fnamebuffer, name);
	return fnamebuffer;
}

movmem(from,to,amount)
register char *from, *to;
register int amount;
{
	while (--amount>=0)
		*to++ = *from++;
}


memstat()
{
}

#ifndef OSKV21

time(t)
long *t;
{
}

char *
ctime(t)
long *t;
{
	static char mytime[30],
	           *monthname[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
	                            "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
	                          };
    struct sgtbuf s;

    getime(&s);
    sprintf(mytime, "%02d-%s-%02d %02d:%02d:%02d\n"
           ,s.t_day
           ,monthname[s.t_month-1]
           ,s.t_year
           ,s.t_hour
           ,s.t_minute
           ,s.t_second
           );
	return mytime;
}
        
#endif OSKV21

restore()
{
}


disable()
{
}



kbread()
{
	unsigned char c;

    if (getstat(1,0) < 0)
      return -1;

	read(0, &c, 1);
	return c;
}


struct asy asy[ASY_MAX];

/* Interface list header */
struct interface *ifaces = NULL;


/* Called at startup time to set up console I/O, memory heap */
ioinit()
{
	char termname[40];
	struct sgbuf sbuf;

	_gs_devn(2, termname+1);
	termname[0] = '/';
	fclose(stdin);
	if (fopen(termname, "r+") != stdin)
	  exit(1000);
	  
	fclose(stdout);
	if (fopen(termname, "w+") != stdout)
	  exit(2000);
	
    if (fileno(stdin) != 0)
      exit(3000);
      
    if (fileno(stdout) != 1)
      exit(4000);

	_gs_opt(0,&sbuf);
    sbuf.sg_case = 0;
    sbuf.sg_backsp = 0;
    sbuf.sg_delete = 0;
    sbuf.sg_echo = 0;
    sbuf.sg_alf = 0;
    sbuf.sg_nulls = 0;
    sbuf.sg_pause = 0;
    sbuf.sg_page = 0;
    sbuf.sg_bspch = 0;
    sbuf.sg_dlnch = 0;
    sbuf.sg_eorch = 0;
    sbuf.sg_eofch = 0;
    sbuf.sg_rlnch = 0;
    sbuf.sg_dulnch = 0;
    sbuf.sg_psch = 0;
    sbuf.sg_bsech = 0;
    sbuf.sg_bellch = 0;
    sbuf.sg_xon = 0;
    sbuf.sg_xoff = 0;
    sbuf.sg_tabcr = 0;
    sbuf.sg_tabsiz = 0;
	_ss_opt(0,&sbuf);
	
	_gs_opt(1,&sbuf);
    sbuf.sg_case = 0;
    sbuf.sg_backsp = 0;
    sbuf.sg_delete = 0;
    sbuf.sg_echo = 0;
    sbuf.sg_alf = 0;
    sbuf.sg_nulls = 0;
    sbuf.sg_pause = 0;
    sbuf.sg_page = 0;
    sbuf.sg_bspch = 0;
    sbuf.sg_dlnch = 0;
    sbuf.sg_eorch = 0;
    sbuf.sg_eofch = 0;
    sbuf.sg_rlnch = 0;
    sbuf.sg_dulnch = 0;
    sbuf.sg_psch = 0;
    sbuf.sg_bsech = 0;
    sbuf.sg_bellch = 0;
    sbuf.sg_xon = 0;
    sbuf.sg_xoff = 0;
    sbuf.sg_tabcr = 0;
    sbuf.sg_tabsiz = 0;
	_ss_opt(1,&sbuf);
	
	setbuf(stdin,NULL);
}

/* Called just before exiting to restore console state */
iostop()
{
}


/* checks the time then ticks and updates ISS */
static int clkval = 0;
void
check_time()
{
        int32 iss();

        if(clkval != clksec()){
                clkval = clksec();
                tick();
                (void)iss();
        }
}



/* Create a directory listing in a temp file and return the resulting file
 * descriptor. If full == 1, give a full listing; else return just a list
 * of names.
 *
 */
FILE *
dir(path,full)
char *path;
int full;
{
        register FILE *fp;
        register char *name = "/dd/a", *cp;
        char cmd[256], disk[256];
        
        strcpy(disk, path);
        for (cp=disk; *cp && *++cp!='/'; )
          ;
        *cp = '\0';

        if(full){
                sprintf(cmd,"(dir -de %s; echo; free %s; echo)>-%s",path,disk,name);
                system(cmd);
        } else {
                sprintf(cmd,"(dir -d %s; echo; free %s; echo)>-%s",path,disk,name);
                system(cmd);
               }
        fp = fopen(name,"r");
        return fp;
}


FILE *
tmpfile()
{
	char template[50];

	strcpy(template, "/dd/smtp_tmpXXXXXX");
	return fopen(mktemp(template), "w+");
}


eihalt()
{
    tsleep(1);
}


#asm
clksec:   moveq   #1,d0
          os9     F$Time
          rts
#endasm



/* Initialize asynch port "dev" */
int
asy_init(dev,bufsize)
int16 dev;
unsigned bufsize;
{
        register unsigned base;
        register struct asy *ap;

        static char termname[] = "/t?";
        struct sgbuf sbuf;
        
        ap = &asy[dev];

        base = ap->addr;   /* 1 or 2 for "/t1" or "/t2" */

        termname[2] = base + '0';
        if ( (ap->vec=open(termname, S_IREAD+S_IWRITE)) < 0) {
        	printf("asy%d: Open failed.\r\l", dev);
        	fflush(stdout);
        	return -1;
        }
        
        _gs_opt(ap->vec, &sbuf);
        sbuf.sg_case = 0;
        sbuf.sg_backsp = 0;
        sbuf.sg_delete = 0;
        sbuf.sg_echo = 0;
        sbuf.sg_alf = 0;
        sbuf.sg_nulls = 0;
        sbuf.sg_pause = 0;
        sbuf.sg_page = 0;
        sbuf.sg_bspch = 0;
        sbuf.sg_dlnch = 0;
        sbuf.sg_eorch = 0;
        sbuf.sg_eofch = 0;
        sbuf.sg_rlnch = 0;
        sbuf.sg_dulnch = 0;
        sbuf.sg_psch = 0;
        sbuf.sg_kbich = 0;
        sbuf.sg_kbach = 0;
        sbuf.sg_bsech = 0;
        sbuf.sg_bellch = 0;
        sbuf.sg_xon = 0;
        sbuf.sg_xoff = 0;
        sbuf.sg_tabcr = 0;
        sbuf.sg_tabsiz = 0;
        _ss_opt(ap->vec, &sbuf);

        return 0;
}


int
asy_stop(iface)
struct interface *iface;
{
	return 0;
}


/* Asynchronous line I/O control */
asy_ioctl(interface,argc,argv)
struct interface *interface;
int argc;
char *argv[];
{
        if(argc < 1){
                printf("%d\r\l",asy[interface->dev].speed);
                return 0;
        }
        return asy_speed(interface->dev,atoi(argv[0]));
}


/* Set asynch line speed */
int
asy_speed(dev,speed)
int dev;
int speed;
{
	return 0;
}


/* Send a buffer to serial transmitter */
asy_output(dev,buf,cnt)
unsigned dev;
char *buf;
unsigned short cnt;
{
        if(dev >= nasy)
            return ;

        write(asy[dev].vec, buf, cnt);
}

/* Receive characters from asynch line
 * Returns count of characters read
 */
unsigned
asy_recv(dev,buf,cnt)
int dev;
char *buf;
unsigned cnt;
{
        register struct asy *ap;
        register int charpend;

        ap = &asy[dev];

        if ( (charpend=_gs_rdy(ap->vec)) < 0)
          return 0;

        charpend = read(ap->vec, buf, min(charpend, cnt));
        return charpend < 0 ? 0 : charpend;
}


int stxrdy(dev)
int dev;
{
	return 1;
}


char *
mytermname()
{
	static char termname[40];

	_gs_devn(2, termname+1);
	termname[0] = '/';

	return termname;
}


/* dirutil.c
 *        OSK directory reading stuff.
 *        Matthias Bohlen
 *        Copyright 1987 Matthias Bohlen, All Rights Reserved.
 *        Permission granted for non-commercial copying and use, provided
 *        this notice is retained.
 */

static DIR *d;
static struct direct *r;


/* wildcard filename lookup */
filedir (dirname, times, ret_str)
char *dirname;
int times;
char *ret_str;
{
	register int masklen;
	register char *cp;

	char dname[80], filemask[80];
	
	for (cp=dirname+strlen(dirname); cp>dirname && *--cp != '/'; )
	  ;
	strncpy(dname, dirname, cp-dirname);
	dname[cp-dirname] = '\0';
	strcpy(filemask, cp+1);
	
    masklen = strlen(filemask);
    
	if (times==0) {   /* first time */
		if ( (d=opendir(dname)) == NULL ) {
			*ret_str = '\0';
			return;
		}
	}
	
	while (1) {
		if ( (r=readdir(d)) == NULL ) {
			*ret_str = '\0';
			closedir(d);
			return;
		}
		if (_cmpnam(r->d_name, filemask, masklen) == 0)
			if (r->d_name[0] != '.') {
				strcpy(ret_str, r->d_name);
				return;
			}
	}
}

/* E O F */
