/*
   pager.c : Copyright Paul Healy, EI9GL, 1992.

   911218: Added this header
   921105: Told pager about tabs, still needs some intelligence about graphic
           characters.
   930124: pager is no longer a standalone program.
*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <dir.h>
#include <string.h>
#include "config.h"
#include "pager.h"
#include "rc.h"
#include "pc.h"
#include "misc.h"
#include "mailer.h"

#ifdef BMH
#define main pager_main
#endif

/* ---------------------------------------------------------------------- */

static int x = -1;
static int max_x = 23;
static int max_y = 79;
static int tty = 0;
static int fin = 0;

static int UseViewer = 0;
static FILE *Page_fp = NULL;
static char *PageFileName = NULL;

/* ---------------------------------------------------------------------- */

static int
ask(void)
{
   static char ClearMore[] = "\r     \r";

   printf("\rMore?");   /* \r will make up for any tab expansion */
   fflush(stdout);

   while (1)
      switch ( tolower(getch()) ) {
         case 'q':
         case 'n':
            fin = -1;
            return -1;
         case '?':
         case 'h':
            printf("\r\n");
            printf(" <space>   move forward one page\n");
            printf(" <return>  move down one line of text\n");
            printf(" q         finish viewing\n\n");
            fflush(stdout);
            break;
         case 'y':
         case ' ':
            printf(ClearMore);
            fflush(stdout);
            x = 0;
            return 0;
         case '\r':
         case '\n':
            printf(ClearMore);
            fflush(stdout);
            x--;
            return 0;
         default:
            putchar(7);     /* quiet variable */
            fflush(stdout);
            break;
         }
}

/*
 * Output a null terminated string to stdout.
 *  return values:
 *   0 => continue on displaying
 *  -1 => user has requested a halt   
 */
static int
page_putstring(char *sp)
{
   int y = 0;
   unsigned char *s = (unsigned char *) sp;

   if (fin)
      return -1;

   if (!tty) {
      puts(sp);
      return 0;
      }

   if ( (x >= max_x) && (ask() == -1))
      return -1;

   while (*s) {
      if (isprint(*s)) {  /* bs, del chars mess up the pager */
         putchar(*s);
         y++;
         }
      else if (*s == '\t') {
         putchar(' ');
         putchar(' ');
         putchar(' ');
         y+=3;
         if (y>max_y)
            putchar('\r');  /* resync */
         }
#ifdef MSDOS
      else if ( *s > 0x7f ) {      /* Allow IBM PC graphics chars thru */
         putchar(*s);
         y++;
         }
#endif
      if ( (y>max_y) || (*s == '\n') ) {
         y = 0;
         x++;
         if ( (x >= max_x) && (ask() == -1))
            return -1;
         }
      s++;
      }
   x++;

   return 0;
}

static int
view_file(char *prog, char *filename)
{
   char *lister = getrc(viewer);

   if (lister == NULL)
      return -1;

   if (spawnlp(P_WAIT, lister, lister, filename, NULL) != 0) {
      fprintf(stderr, "%s: error calling file viewer '%s %s'\n",
         prog, lister, filename);
      perror(prog);
      return -2;
      }
   else 
      return 0;
}

/* ---------------------------------------------------------------------- */

int
page_setup(char *prog)
{
   x = 0;
   max_x = atoi(getrc(maxrow)) - 1;
   max_y = atoi(getrc(maxcol)) - 1;

   tty = isatty(fileno(stdout));
   fin = 0;

   UseViewer = getrc(viewer) != NULL;
   if (UseViewer) {
      char filename[MAXNAMLEN];

      Page_fp = tempfile(prog, filename, "wb");
      if (Page_fp == NULL)
         UseViewer = 0;
      else
         PageFileName = strdup(filename);
      }

   return 0;
}

/*
   Write a string to the console with paged output. The string should not be
   terminated with a newline character (\n). Embedded \n's should be ok.
*/
int
page_puts(char *s)
{
   if (Page_fp != NULL)
      return putline(Page_fp, s);
   else {
      int ret = page_putstring(s);

      if ( tty && (ret == 0) )
         putchar('\n');

      return ret;
      }
}

/*
 * note this always (incorrectly) adds a \n
 */
int
page_printf(const char *format, ...)
{
   char line[LINELEN];
   int ret;
   va_list vp;

   va_start(vp, format);
   ret = vsprintf(line, format, vp);
   va_end(vp);

   if (ret == -1)
      return -1;
   else 
      return page_puts(line);
}

int
page_fin(char *prog)
{
   if (UseViewer) {
      if ((PageFileName == NULL) || (Page_fp == NULL))
         return -1;
      fclose(Page_fp);
      view_file(prog, PageFileName);
      unlink(PageFileName);
      free(PageFileName);
      }
   else 
      fflush(stdout);

   return 0;
}

int
page_next(void)
{
   static char ClearMore[] = "\r     \r";

   if (UseViewer || !tty)
      return 0;

   printf("\rNext?");   /* \r will make up for any tab expansion */
   fflush(stdout);

   while (1)
      switch ( tolower(getch()) ) {
         case 'q':
            return -1;
         case '?':
         case 'h':
            printf("\r\n");
            printf(" <space>   next message\n");
            printf(" q         finish viewing\n\n");
            fflush(stdout);
            break;
         case 'y':
         case ' ':
         case 'n':
         case 'N':
            printf(ClearMore);
            fflush(stdout);
            return 0;
         default:
            putchar(7);     /* quiet variable */
            fflush(stdout);
            break;
         }
}

int
page_file(char *prog, char *filename)
{
   FILE *fp;
   char s[LINELEN];

   if (view_file(prog, filename) != -1)
      return 0;

   if ( (fp=fopen(filename, "r")) == NULL) {
      fprintf(stderr, "%s: can't open %s\n", prog, filename);
      return -1;
      }

   page_setup(prog);

   while (fgets(s, sizeof(s), fp) != NULL) {
      char *p = s;

      for (; *p; p++)
         if (*p == '\n') {
            *p = '\0';
            break;
            }

      if ( page_puts(s) != 0 )
         break;
      }
   fclose(fp);

   page_fin(prog);

   return 0;
}

/*
#ifdef PAGER_PROG
*/
int
main(int argc, char *argv[])
{
   if (argc != 2) {
      fprintf(stderr, "Usage: pager <file>\n");
      return -1;
      }

   return page_file("pager", argv[1]);
}
/*
#endif
*/
