/*
   pc.c : Copyright Paul Healy, EI9GL, 1992.

   Derived from bm.

   Copyright 1986 Bdale Garbee, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.
   Copyright 1987 1988 Dave Trulli NN2Z, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.

   This file contains MSDOS specific functions

   911215 : Added this header
   920809 : Added option for command line editing with CED
*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <bios.h>
#include "config.h"
#include "misc.h"
#include "pc.h"
#include "rc.h"

#undef DEBUG

#ifdef __TURBOC__
extern unsigned _stklen = STACKSIZE;
#endif

void
setvideo(char *s)
{
#ifdef __TURBOC__
   if (strncmp("bios", s, 4) == 0)
      directvideo = 0;
   else
      directvideo = 1;
#endif
}

int
cedpresent(void)
{
   union	REGS regs;

   regs.h.ah = 0xff;
   regs.h.al = 0x02;
   int86(0x21, &regs, &regs);
#ifdef DEBUG
   fprintf(stderr, "\nced: cf=%d, ax=0x%x, ah=0x%x\n", regs.x.cflag, regs.x.ax, regs.h.ah);
#endif
   if (regs.x.cflag || (regs.h.ah == 0xff))
      return -1;
   else
      return 0;
}

#ifdef notdef
char *
congets(char *line, int len)
{
   union	REGS regs;
   struct SREGS seg;
   char far *p = (char far *) line;

   segread(&seg);

   regs.h.ah = 0x0a;           /* buffered input function */
   seg.ds = FP_SEG(p);
   regs.x.dx = FP_OFF(p);

   line[0] = len - 2;
   line[1] = 0;

   int86(0x21, &regs, &regs);

   return &line[2];
}
#endif

char *
getline(char *line, int size)
{
   char *opt = getrc(editline);

   if ( (opt == NULL) || (tolower(*opt) == 'n'))
      return fgets(line, size, stdin);
   else {
      char *p;

      size = size - 2;
      if (size > 255)
         size = 255;   /* only have one byte to store the length */

      line[0] = size;
      line[1] = 0;     /* INT 21h, AH=0Ah, DS:DX[1] = length of previous line */
                       /* that can be returned - just for safety */
      p = cgets(line); 
      putchar('\n');   /* emulate fgets */
      fflush(stdout);  /* for bc++ v3.1 */

      return p;
      }
}

int
kbready(void)
{
   return bioskey(1) != 0;
}

unsigned char _dvmajor = 0;
unsigned char _dvminor = 0;

int
dvpresent(void)
{
   char *mtask = getrc(multitask);
   union	REGS regs;

   if (*mtask == 'n')
      return -1;

   regs.x.ax = 0x2b01;        /* DOS + cmos set date; ask for DV version if present */
   regs.x.cx = 0x4445;        /* 'DE' (year) */
   regs.x.dx = 0x5351;        /* 'SQ' (month+day) */
   int86(0x21, &regs, &regs);

   if (regs.h.al == 0xff)      
      return -1;             /* invalid -  system date was unchanged */
   else {
      _dvmajor = regs.h.bh;
      _dvminor = regs.h.bl;

      return 0;
      }
}

unsigned char _cputype = 0;
unsigned char _dpmimajor = 0;
unsigned char _dpmiminor = 0;

int
dpmipresent(void)
{
   union	REGS regs;

   regs.x.ax = 0x1687;
   int86(0x2f, &regs, &regs);

   if (regs.x.ax != 0)      
      return -1;
   else {
      _cputype = regs.h.cl;
      _dpmimajor = regs.h.dh;
      _dpmiminor = regs.h.dl;
      return 0;
      }
}

int
deschedule(void)
{
   static int checked = 0;
   static int dv = 0;
   static int dpmi = 0;
   union	REGS regs;

   if (!checked) {
      dv = dvpresent();
      dpmi = dpmipresent();
      checked = 1;
      }

   /*
    * qdpmi doesn't give use the dpmi pause call to give up time to dv.
    */
   if (dv != -1) {
      regs.x.ax = 0x1000;          /* topview pause */
      int86(0x15, &regs, &regs);
      return 0;
      }
   else if (dpmi != -1) {
      regs.x.ax = 0x1680;          /* dpmi pause (windows, os2) */
      int86(0x2f, &regs, &regs);
      if (regs.h.al != 0)
         dpmi = -1;
      return 0;
      }
   else
      return -1;
}

/* ----------------------------------------------------------------------- */

static int
numfiles(char *name)
{
   struct ffblk sbuf;
   int n = 0;

   if (findfirst(name, &sbuf, FA_HIDDEN | FA_SYSTEM | FA_DIREC ) == 0) {
      n++;
      while (findnext(&sbuf) == 0)
         n++;
      }

   return n;
}

/* directory search utility for DOS */

void
filedir(char *name, int times, char *result)
{
   static struct ffblk sbuf;

   result[0] = 0;

   if (times == 0) {
      if (findfirst(name, &sbuf, FA_HIDDEN | FA_SYSTEM | FA_DIREC ))
         return ;
      }
   else
      if (findnext(&sbuf))
         return ;

   strcpy(result, sbuf.ff_name);
   strlwr(result);
}

static int
getdir(char *name, String buf[], int count)
{
   int i;

   for (i=0; i<count; i++) {
      filedir(name, i, buf[i]);
      if (buf[i][0] == '\0')
         return -1;
#ifdef DEBUG
      else
         null_check(buf[i]);
#endif
      }
   qsort(buf, count, sizeof(String), (CMPFN) strcmp);
#ifdef DEBUG
   null_check("after qsort");
#endif
   return count;
}

/* ----------------------------------------------------------------------- */

int
dir_init(Directory *d, char *pattern)
{
   int nfound = numfiles(pattern);

#ifdef DEBUG
   null_check("after counting files");
#endif

   d->buf = NULL;
   d->ptr = -1;
   d->num = nfound;

   if (nfound < 1)
      return 0;

   if ( (d->buf = (String *) malloc(sizeof(String) * nfound) ) == NULL) {
      fprintf(stderr, "initdir: no room for directory listing (%d entries)\n", nfound);
      return -1;
      }
#ifdef DEBUG
   null_check("after malloc");
#endif

   d->num = getdir(pattern, d->buf, nfound);

#ifdef DEBUG
   null_check("after getdir");
#endif

   if (d->num != nfound) {
      fprintf(stderr, "initdir: warning, directory changed during listing\n");
      free(d->buf);
      d->buf = NULL;
      return -1;
      }

   return nfound;
}

char *
dir_read(Directory *d)
{
   d->ptr = d->ptr + 1;

   if ( (d->ptr >= d->num) || (d->buf == NULL) )
      return NULL;
   else 
      return d->buf[d->ptr];
}

int
dir_fin(Directory *d)
{
   if (d->buf != NULL) {
      free(d->buf);
      d->buf = NULL;
      }
   return 0;
}

/*
int
sortbytime(Directory *d)
{
   int i;

   for (i=0; i<d->num; i++) {
      char s[MAXNAMLEN];
      struct statbuf sbuf;

      sprintf(s, "%s/%s", d->[i].name);
      stat(s, &sbuf);
      d->[i].time = sbuf.atime;
      }
   qsort();
}

*/

long
filelen(char *filename)
{
#ifdef STATOK
   struct stat statbuf;
   return stat(filename, &statbuf) == -1 ? -1L : statbuf.st_size;
#else
   int fd = open(filename, O_RDONLY, 0);
   long size;

   if (fd == -1)
      return -2L;

#ifdef MSDOS
   size = filelength(fd);
#else
   size = lseek(fd, 0L, SEEK_END); /* size is -1L on error */
#endif

   close(fd);

#ifdef DEBUG
   fprintf(stderr, "size: %s is %ld\n", filename, size);
#endif

   return size;
#endif
}

