MODEL        equ        0
        include lmacros.h
        assume        ds:dataseg
        public sssave,spsave,intstk,doret,intds

; common routine for interrupt return
doret:        mov        al,20h                ; 8259 end-of-interrupt command
        out        20h,al
        pop        es
        pop        di
        pop        si
        pop        bp
        pop        dx
        pop        cx
        pop        bx
        pop        ax
        mov        ss,sssave
        mov        sp,spsave        ; restore original stack context
        pop        ds
        iret

; setvec - set interrupt vector
; called from C as follows
; setvec(vec,csval,ipval)
; char vec;                /* Interrupt number */
; unsigned short csval;        /* CS register for vector */
; unsigned short ipval;        /* IP for vector */

        procdef        setvec,<<vec,byte>,<csval,word>,<ipval,word>>
        mov        cs:intds,ds        ; save data segment pointer
        mov        ah,25h
        mov        al,vec
        mov        dx,ipval
        push        ds                ; save
        mov        cx,csval
        push        cx
        pop        ds                ; ds == csval
        int        21h
        pop        ds                ; restore
        pret
        pend        setvec
intds        dw        0        ; save loc for ds (must be in code segment)


; getvec - return current interrupt vector
; called from C as
; long                /* Returns CS in high word, IP in low word */
; getvec(vecnum)
; char vecnum;        /* Interrupt number */

        procdef        getvec,<<vecnum,byte>>
        mov        ah,35h
        mov        al,vecnum
        push        es        ; save, since DOS uses it for a return value
        int        21h
        mov        dx,es        ; CS value into DX (C high word)
        mov        ax,bx        ; IP value into AX (C low word)
        pop        es        ; restore es
        pret
        pend        getvec

; clksec - return current second value
        procdef        clksec
        mov        ah,2ch
        int        21h
        mov        al,dh
        mov        ah,0
        pret
        pend        clksec

; kbraw - raw, nonblocking read from console
; If character is ready, return it; if not, return -1

        procdef        kbraw
        mov        ah,06h        ; Direct Console I/O
        mov        dl,0ffh        ; Read from keyboard
        int        21h        ; Call DOS
        jz        nochar        ; zero flag set -> no character ready
        mov        ah,0        ; valid char is 0-255
        pret
nochar:
        mov        ax,-1        ; no char, return -1
        pret
        pend        kbraw

        procdef        getcs
        mov        ax,cs
        pret
        pend        getcs

        procdef        getds
        mov        ax,ds
        pret
        pend        getds

; disable - disable interrupts and return previous state: 0 = disabled,
;           1 = enabled

        procdef        disable
        pushf                        ; save state on stack
        cli                        ; interrupts off
        pop        ax                ; original flags -> ax
        and        ax,200h                ; 1<<9 is IF bit
        jnz        ion                ; nonzero -> interrupts were on
        pret
ion:        mov        ax,1
        pret
        pend        disable

; restore - restore interrupt state: 0 = off, nonzero = on

        procdef        restore,<<istate,byte>>
        test        istate,0ffh
        jz        ioff
        sti
ioff:        pret
        pend        restore

; Halt until an interrupt occurs, then return
        procdef eihalt
        sti        ; make sure interrupts are enabled
        hlt
        pret
        pend        eihalt

; Relinquish processor so other doubledos task can run
        procdef        giveup
        mov        al,2                ; 110 ms
        mov        ah,0eeh
        int        21h
        pret
        pend        giveup

        finish
dataseg        segment        para public 'data'
        bss        sssave:word,2
        bss        spsave:word,2
        bss        intstk:byte,512
dataseg        ends
        end

