/*
   pick.c: Copyright 1993, Paul Healy.

   930418 : Created
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "misc.h"
#include "help.h"
#include "mailer.h"
#include "buf.h"
#include "getopts.h"
#include "mailfile.h"
#include "message.h"
#include "spawn.h"
#include "header.h"
#include "pc.h"

#ifdef BMH
#define main pick_main
#endif

static char *ExecProg = NULL;

static int
exec(int n)
{
   if (ExecProg == NULL)
      return printf("%d ", n);
   else {
      char *argv[3];
      char num[15];

      argv[0] = ExecProg;
      sprintf(num, "%d", n);
      argv[1] = num;
      argv[2] = NULL;
      if (run_vp(P_WAIT, ExecProg, argv) != 0) {
         fprintf(stderr, "pick: error calling %s %d\n", ExecProg, n);
         return -1;
         }
      else {
         /* need a more prompt here */
         return 0;
         }
      }
}

static int SearchCmd = -1;
static char *SearchStr = NULL;

static int
ispart(char *str, char *substr)
{
#ifdef DEBUG
   fprintf(stderr, "pick: checking <%s> for <%s>\n", str, substr);
#endif
/*
   return strstr(strlwr(str), substr) != NULL;
*/
   return strstr(str, substr) != NULL;
}

static MsgResult
match(Mailfile *m, char *line)
{
   Header_t h;

   if (gethdrb(m, &h) == -1) {
      fprintf(stderr, "pick: bad header on message %d\n", mail_curmsg(m));
      return MSG_QUIT;
      }

   if (SearchCmd != -1) {
      if (ispart(HEADER(h, SearchCmd), SearchStr))
         exec(mail_curmsg(m));
      }
   else {
      while (mail_getstring(m, line, LINELEN) != NULL)
         if (ismsg(line))
            return MSG_NEXT;
         else if (ispart(line, SearchStr)) {
            exec(mail_curmsg(m));
            break;
            }
      }

   mail_getmsg(m, line, LINELEN);
   return MSG_NEXT;
}

int
pick(Mailfile *m, int argc, char *argv[])
{
   int optchar;
   char *optarg;
   options_t *opts;
   int errors = 0;

   static const char *optv[] = {
      "c:cc",             
      "d:date",
      "e:exec",
      "f:from",           
      "m:message-id",     
      "n:newsgroups",     
      "r:reply-to",       
      "S:search",         
      "s:subject",        
      "t:to",             
#ifdef AX25BBS
      "x:x-bbs-msg-type",
#endif
      NULL
      };

   ExecProg = NULL;
   SearchCmd = -1;
   SearchStr = NULL;

   if (mail_getfile(m, argc, argv, "pick") == -1)
      return -1;

   opts = optinit(*argv, optv);
   optctrl(opts, opt_LONG_ONLY);

   --argc, ++argv;
   while( (optchar = getopts(opts, argc, argv, &optarg)) != 0 )
      switch(optchar) {
         case '?' :
         case 'H' :
            ++errors;
            break;
         case 'c' : SearchStr = optarg; SearchCmd = CC; break;
         case 'd' : SearchStr = optarg; SearchCmd = DATE; break;
         case 'e' : ExecProg = optarg; break;
         case 'f' : SearchStr = optarg; SearchCmd = FROM; break;
         case 'm' : SearchStr = optarg; SearchCmd = MESSAGEID; break;
         case 'n' : SearchStr = optarg; SearchCmd = NEWSGROUPS; break;
         case 'r' : SearchStr = optarg; SearchCmd = REPLYTO; break;
         case 'S' : SearchStr = optarg; SearchCmd = -1; break;
         case 's' : SearchStr = optarg; SearchCmd = SUBJECT; break;
         case 't' : SearchStr = optarg; SearchCmd = TO; break;
         case 'x' : SearchStr = optarg; SearchCmd = XBBSMSGTYPE; break;
         default : ++errors; break;
         }

   if (errors || (SearchStr == NULL)) {
      optusage(opts, "", stderr);
      optfree(opts), opts = NULL;
      return -1;
      }
   else 
      optfree(opts), opts = NULL;
/*
   strlwr(SearchStr);
*/
   walkall(m, match);
/*
   fprintf(stderr, "pick: argv[-1] = %s\n", argv[-1]);
*/
   if (argv[-1][0] == 'S')
      printf("\n");

   return 0;
}

int
main(int argc, char *argv[])
{
   Mailfile m;

   if (dohelp(argc, argv, "pick [+folder] [-exec prog] [conditions] (pick messages)"))
      return 0;

   mail_zero(&m);

   pick(&m, argc, argv);

   return mail_savestate(&m);
}

