;***************************************************************
;* MACROS.ASM -- General Purpose Macros 		       *
;*							       *
;* Copyright (C) 1990, 1991 by Alef Null. All rights reserved. *
;* Author(s): Jarkko Vuori, OH2LNS			       *
;* Modification(s):					       *
;***************************************************************

;********************
;*    DSP CARD 3    *
;*     equates	    *
;********************

; DCD light handling
dcdon	macro
	bset	#$2,x:m_pcd
	endm
dcdoff	macro
	bclr	#$2,x:m_pcd
	endm

; PTT handling
ptton	macro
	bset	#$4,x:m_pcd
	endm
pttoff	macro
	bclr	#$4,x:m_pcd
	endm


;********************
;*  Useful macros   *
;********************

; macro for immediate move
movi	macro	data,dest
	move		    #data,a1
	move		    a1,dest
	endm
movib	macro	data,dest
	move		    #data,b1
	move		    b1,dest
	endm

; macro for convolution
conv	macro	mode,n,xd,cf,acc
	clr	acc	    x:(xd)+,x0	y:(cf)+,y0
	rep	#n-1
	mac	x0,y0,acc   x:(xd)+,x0	y:(cf)+,y0
	if	mode
	macr	x0,y0,acc   (xd)-
	else
	macr	x0,y0,acc
	endif
	endm

; first order high-pass IIR filter
; input in a, output in a
; frq is -3dB point (fc/fs)
hpass	macro	acc,frq,store
c1	set	-1.0/@cos(6.28319*frq)*(1.0-@sin(6.28319*frq))
	move		    acc,x0
	move		    #(1.0-c1)/2.0,x1
	mpyr	x0,x1,acc   x:<store,x1
	move		    acc,x:<store
	sub	x1,acc	    x:<store+1,x0
	move		    #c1,x1
	macr	-x0,x1,acc
	move		    acc,x:<store+1
	endm

; first order low-pass IIR filter
; input in a, output in a
; frq is -3dB point (fc/fs)
lpass	macro	acc,frq,store
c1	set	-1.0/@cos(6.28319*frq)*(1.0-@sin(6.28319*frq))
	move		    acc,x0
	move		    #(1.0+c1)/2.0,x1
	mpyr	x0,x1,acc   x:<store,x1
	move		    acc,x:<store
	add	x1,acc	    x:<store+1,x0
	move		    #c1,x1
	macr	-x0,x1,acc
	move		    acc,x:<store+1
	endm


;********************
;*  Contex swapping *
;*     macros	    *
;********************

; macro for entering interrupt service routine
enter	macro	contex
	move		    x,l:<contex+0
	move		    y,l:<contex+1
	move		    a10,l:<contex+2
	move		    b10,l:<contex+3
	move		    a2,x:<contex+4
	move		    b2,y:<contex+4
	endm


; macro for leaving interrupt service routine
leave	macro	contex
	move		    l:<contex+0,x
	move		    l:<contex+1,y
	move		    l:<contex+2,a10
	move		    l:<contex+3,b10
	move		    x:<contex+4,a2
	move		    y:<contex+4,b2
	rti
	endm


; macro for  ALU contex space reserving
roomab	macro
	ds	5
	endm


; macro for address register contex saving
stor	macro	contex,reg
	move		    r\reg,y:<contex+0
	move		    m\reg,y:<contex+1
	move		    n\reg,y:<contex+2
	endm


; macro for a and x registers contex restoring
rclr	macro	contex,reg
	move		    y:<contex+0,r\reg
	move		    y:<contex+1,m\reg
	move		    y:<contex+2,n\reg
	endm


; macro for  a and x contex space reserving
roomr	macro
	ds	3
	endm


;********************
;*   Debug macros   *
;********************

; macro for debugging
bug	macro	value
	move	#value,a1
	swi
	endm

; macro for Nth sample debugging
bugN	macro	N
	move		    y:<counter,b
	move		    #>1,x0
	sub	x0,b
	jne	<_stok
	swi
	move		    #N,b1
_stok	move		    b1,y:<counter
	endm

; macro for post mortem dump
dump	macro
	ori	#$03,mr

	jclr	#m_htde,x:m_hsr,*
	movep	#>1313,x:m_htx

	move	#0,r0
	move	#$ffff,m0
	move	#0,n0

	move	#512+4064-1,a1
	do	a1,_dploop
	jclr	#m_htde,x:m_hsr,*
	movep	p:(r0)+,x:m_htx
_dploop

	move	#0,r0
	do	#256+2048-1_,dxloop
	jclr	#m_htde,x:m_hsr,*
	movep	x:(r0)+,x:m_htx
_dxloop

	move	#0,r0
	do	#256+2048+1,_dyloop
	jclr	#m_htde,x:m_hsr,*
	movep	y:(r0)+,x:m_htx
_dyloop

	stop
	endm


;********************
;*     TLC32044     *
;********************

; primary communication controls
noadj	equ	%00<<8
retard	equ	%01<<8
advance equ	%10<<8


; output to ssi
outssi	macro	data
	jclr	#m_tde,x:m_sr,*
	movep		    #data,x:m_tx
	endm


; program the Texas A/D & D/A converter chip
pgmtlc	macro	ctrl,tra,tra2,trb
	outssi	$000000 			    ; flush any garbage out
	outssi	$000300 			    ; begin secondary communication
	outssi	((ctrl<<2)|%11)<<8		    ; set control word
	outssi	$000300 			    ; begin secondary communication
	outssi	((tra<<9)|(tra<<2)|%00)<<8	    ; set TA and RA registers
	outssi	$000300 			    ; begin secondary communication
	outssi	((tra2<<9)|(tra2<<2)|%01)<<8	    ; set TA' and RA' registers
	outssi	$000300 			    ; begin secondary communication
	outssi	((trb<<9)|(trb<<2)|%10)<<8	    ; set TB and RB registers
	endm


;********************
;*   Modem status   *
;*    handling	    *
;********************

	xref	status

xmt	equ	0				    ; currently transmitting data
dcd	equ	1				    ; carrier detected
ptt	equ	2				    ; transmitter on
req	equ	3				    ; xmit request pending
tim	equ	4				    ; timer request
shut	equ	5				    ; xmitter shutdown request
