	page	132,63,1,1
	opt	rc
	title	'DSP CARD 3 initialization'

;***************************************************************
;* INIT.ASM -- DSP56001 interrupt vector table and DSP CARD 3  *
;*	       initialization code			       *
;*							       *
;* Here are all DSP56001 interrupt vectors and space for       *
;* interrupt contexts. Also included are initalization code    *
;* and some stuff for debugging.			       *
;*							       *
;* Copyright (C) 1990, 1991 by Alef Null. All rights reserved. *
;* Author(s): Jarkko Vuori, OH2LNS			       *
;* Modification(s):					       *
;***************************************************************

	nolist
	include 'ioequlc'
	list

	section Vector
	xref	sci_rec,sci_xmt,sci_tim
	xref	m_rx,m_sr,m_tx,m_ssr,m_srxl

	nolist
	include 'intequlc'
	list

; Reset vector
	org	p:i_reset
	jmp	start

; SSI receive interrupt
	org	p:i_ssird
	movep		    x:m_rx,p:(r7)+
	nop

; SSI receive interrupt with errors
	org	p:i_ssirde
	movep		    x:m_sr,p:(r7)	    ; clear ROE
	movep		    x:m_rx,p:(r7)

; SSI transmitter interrupt
	org	p:i_ssitd
	movep		    #0,x:m_tx

; SSI transmitter interrupt with errors
	org	p:i_ssitde
	movep		    x:m_sr,x:m_tx	    ; clear TUE

; SCI receive interrupt
	org	p:i_scird
	jsr	sci_rec

; SCI receive interrupt with errors
	org	p:i_scirde
	movep		    x:m_ssr,x:m_tx

; SCI transmitter interrupt
	org	p:i_scitd
	jsr	sci_xmt

	endsec


	section Init
	xdef	start
	xref	ssi_ini,sci_ini
	xref	fft_i
	xref	m_loop
	xref	m_bcr,m_pcc,m_pcddr,m_pcd,m_ipr

	nolist
	include 'macros'
	list

	org	p:


;****************************
;*   Start of the program   *
;****************************
; no wait states on external memory
start	movep		    #$0000,x:m_bcr	    ; no wait states

; initialize peripherals
	jsr	<ssi_ini
	jsr	<sci_ini

; initialize port C
	movep		    #$01eb,x:m_pcc	    ; TXD,RXD,SC0,SC2,SCK,SRD,STD
	movep		    #$0014,x:m_pcddr	    ; SCLK,SC1 as output
	movep		    #$0000,x:m_pcd

; initialize status
	movi	0,x:<status

; initialize modules (with external hardware)
	jsr	<fft_i

; start interrupts
	movep		    #$b400,x:m_ipr	    ; SSI=IPL2,SCI=IPL1,HOST=IPL0
	andi	#$fc,mr 			    ; unmask interrupts

; and jump to the main modem module
	jmp	<m_loop

	endsec



;****************************
;*   STATUS OF THE MODEM    *
;****************************

	section Status

	xdef	status

	org	x:

status	ds	1

	endsec


;****************************
;*  CONTEX STORE FOR INTS   *
;****************************

	section Contexts
	xdef	scix,scir,scit
	xdef	scixr,scirr,swir

	nolist
	include 'macros'
	list

	org	l:

scix	roomab
scir	roomab
scit	roomab

	org	y:

scixr	roomr
scirr	roomr
swir	roomr

	endsec

	end
