	page	132,63,1,1
	opt	rc
	title	'SCI port handling'

;***************************************************************
;* SCI.ASM -- SCI drivers				       *
;*							       *
;* Provides interrupt based SCI handling, and data buffering.  *
;*							       *
;* This module uses registers as follows:		       *
;*  r3 - xmit data pointer				       *
;*							       *
;* Copyright (C) 1990, 1991 by Alef Null. All rights reserved. *
;* Author(s): Jarkko Vuori, OH2LNS			       *
;* Modification(s):					       *
;***************************************************************

	nolist
	include 'ioequlc'
	list

	section Sci
	xdef	baud
	xdef	sci_rec,sci_xmt
	xdef	sci_ini,putblk
	xref	m_scr,m_sccr,m_srxl,m_ssr,m_tie,m_stxl,m_stxm,m_stxh

	org	p:

	nolist
	include 'macros'
	list

; Serial interface parameters
baud	equ	19200


;****************************
;*    SCI initialization    *
;****************************
sci_ini
; initialize SCI
	IF	!DEBUG
	movep		    #$0b02,x:m_scr	    ; async 8,n,1
	movep		    #20736000/(2*2*16*baud),x:m_sccr
	ELSE
	movep		    #$0b02,x:m_scr	    ; async 8,n,1
	movep		    #20736000/(10*2*2*16*baud),x:m_sccr
	ENDIF

	rts


;****************************
;*  Request to xmit block   *
;****************************
;starting address of the block in x0
;ending   address of the block in x1
;returns   C if transmitter busy
;	  NC if request accepted
putblk
; check if transmitter running
	jset	#m_tie,x:m_scr,putblke
; no, start transmitting header
	move		    x0,x:<stxblk
	move		    x1,x:<endblk
	move		    #heablk,r3
	bset	#header,x:<status
	movep		    p:(r3),x:m_stxl
	bset	#shift,x:<status
	bset	#m_tie,x:m_scr

putblke rts


;****************************
;*   SCI xmitter interrupt  *
;****************************
sci_xmt
; save registers to be used
	enter	scix

; send appropriate byte of the word
	jset	#shift,x:<status,scix1
	movep		    p:(r3),x:m_stxl
	bset	#shift,x:<status
	jmp	<scx_end
scix1	bchg	#shftcnt,x:<status
	jcs	<scix2
	movep		    p:(r3),x:m_stxm
	jmp	<scx_end
scix2	movep		    p:(r3)+,x:m_stxh
	bclr	#shift,x:<status

; whole word send
	jclr	#header,x:<status,scix3
; sending header, check if end of it
	move		    r3,a
	move		    #heablk+healen,x0
	cmp	x0,a
	jne	<scx_end
; yes, header sent, start sending datablock
	bclr	#header,x:<status
	move		    x:<stxblk,r3
	jmp	<scx_end
; sending datablock, check if the block has been sent
scix3	move		    r3,a
	move		    x:<endblk,x0
	cmp	x0,a
	jne	<scx_end
; yes, shut down transmitter
	bclr	#m_tie,x:m_scr

; restore registers
scx_end leave	scix



;****************************
;*   SCI receive interrupt  *
;****************************
sci_rec
; save registers to be used
	enter	scir

; read received character
	movep		    x:m_srxl,x0

; check what to do with it
	move		    #>'+',a
	cmp	x0,a	    #>'-',a
	jeq	<sci_r1
	cmp	x0,a
	jne	<scr_end

; decrement N
	move		    p:NN2,a
	asr	a
	move		    a,p:NN2
	jmp	<scr_end

; increment N
sci_r1	move		    p:NN2,a
	asl	a
	move		    a,p:NN2

; restore registers
scr_end leave	scir


;****************************
;*	 DATA - AREA	    *
;****************************

	xdef	NN2

healen	equ	4
heablk	dc	$010964
version dc	290791
samplef dc	8000				    ; sample frequency
NN2	dc	16				    ; accumulator count

	org	x:

; KISS-to-data queue
stxblk	ds	1
endblk	ds	1

	endsec

	end
