/*
   refile.c : Copyright 1992-1993, Paul Healy, EI9GL.

   open current folder in read mode  Lock
   open current.new in write mode
   open to folder in append mode     Lock

   write selected msgs to end of folder
   write unselected msgs to current.new

   Unlock folder
   rename current -> current.old
   rename current.new -> current      Unlock

   920727 : Created.
*/
#include <stdio.h>
#include <string.h>
#include "config.h"
#include "rc.h"
#include "help.h"
#include "lock.h"
#include "buf.h"
#include "mailfile.h"
#include "message.h"
#include "bmh.h"
#include "mailer.h"

#ifdef BMH
#define main refile_main
#endif

static int rcurmsg = 0;
static int rargc = 0;
static char **rargv = NULL;
static FILE *curr_fp = NULL;
static FILE *dest_fp = NULL;

static MsgResult
cpmsg(Mailfile *m, char *line)
{
   if ( wanted(mail_curmsg(m), rargc, rargv) ||
        ( (rargc == 2) && (mail_curmsg(m) == rcurmsg) )
      )
      return mail_msg2file(m, dest_fp, line);
   else 
      return mail_msg2file(m, curr_fp, line);
}

static int
xtract(Mailfile *m_curr, Mailfile *m_dest, int argc, char *argv[])
{
   rcurmsg = mail_curmsg(m_curr);
   rargc = argc;
   rargv = argv;

   if ( (curr_fp = fopenbuf(mail_newname(m_curr), WRITE, STDOUT_BUFSIZE)) == NULL) {
      fprintf(stderr, "refile: can't open %s\n", mail_newname(m_curr));
      return -1;
      }
   else if ( (dest_fp = fopenbuf(mail_filename(m_dest), APPEND, STDOUT_BUFSIZE)) == NULL) {
      fprintf(stderr, "refile: can't open %s\n", mail_newname(m_dest));
      fclose(curr_fp);
      return -1;
      }
   else {
      int ret = walkall(m_curr, cpmsg);

      fclose(curr_fp);
      fclose(dest_fp);

      return ret;
      }
}

int
refile(Mailfile *m_curr, int argc, char *argv[])
{
   int ret;
   char *dest = argv[1];
   Mailfile m_to;
   Mailfile *m_dest = &m_to;

   if (argc < 2) {
      fprintf(stderr, "refile: no folder?\n");
      return -1;
      }
   else if (*dest == '+')
      dest++;

   if (mail_getfile(m_curr, argc-1, ++argv, "refile") == -1)
      return -1;

   if (strcmp(dest, mail_newsgroup(m_curr)) == 0) {
      fprintf(stderr, "refile: source and destination folders are the same\n");
      return -1;
      }

   mail_zero(m_dest);

   if (mail_init(m_dest, dest) == -1) {
      fprintf(stderr, "refile: failed setting up to read %s\n", dest);
      return -1;
      }
   else if (mail_lock(m_dest) == -1) {
      fprintf(stderr, "refile: can't lock %s\n", mail_filename(m_dest));
      mail_fin(m_dest);
      return -1;
      }

   mail_keeplock(m_curr);

   ret = xtract(m_curr, m_dest, argc, argv);

   mail_releaselock(m_curr);
   mail_close(m_curr);

   if (ret == -1)
      unlink(mail_newname(m_curr));
   else 
      mail_update(m_curr, "refile");

   mail_unlock(m_curr);
   mail_unlock(m_dest);

#ifdef INDEX
   mapit(m_curr);
   mail_reset(m_curr);
   mapit(m_dest);
#endif

   mail_fin(m_dest);

   if (mail_size(m_curr)==-1) {
/*
      fprintf(stderr, "refile: warning check empty failed, %s may still be locked\n", mail_filename(m_curr));
*/
      return -1;
      }
   else 
      return ret;
}

int
main(int argc, char *argv[])
{
   Mailfile m;

   if (dohelp(argc, argv, "refile [+folder] [msglist] (move mail)"))
      return 0;

   mail_zero(&m);

   refile(&m, argc, argv);

   return mail_savestate(&m);
}
