	page	132,63,1,1
	opt	rc
	title	'DSP CARD 3 initialization'

;***************************************************************
;* INIT.ASM -- DSP56001 interrupt vector table and DSP CARD 3  *
;*	       initialization code			       *
;*							       *
;* Here are all DSP56001 interrupt vectors and space for       *
;* interrupt contexts. Also included are initalization code    *
;* and some stuff for debugging.			       *
;*							       *
;* Copyright (C) 1990, 1991 by Alef Null. All rights reserved. *
;* Author(s): Jarkko Vuori, OH2LNS			       *
;* Modification(s):					       *
;***************************************************************

	nolist
	include 'ioequlc'
	list

	section Vector
	xref	sci_rec,sci_err,sci_xmt,sci_tim
	xref	m_rx,m_sr,m_tx

	nolist
	include 'intequlc'
	list

; Reset vector
	org	p:i_reset
	jmp	start

; SWI interrupt
	org	p:i_swi
	jsr	cpu_swi

; SSI receive interrupt
	org	p:i_ssird
	movep		    x:m_rx,x:(r2)-
	nop

; SSI receive interrupt with errors
	org	p:i_ssirde
	movep		    x:m_sr,x:(r2)	    ; clear ROE
	movep		    x:m_rx,x:(r2)

; SSI transmitter interrupt
	org	p:i_ssitd
	movep		    x:(r3)+,x:m_tx
	nop

; SSI transmitter interrupt with errors
	org	p:i_ssitde
	movep		    x:m_sr,x:m_tx	    ; clear TUE

; SCI receive interrupt
	org	p:i_scird
	jsr	sci_rec

; SCI receive interrupt with errors
	org	p:i_scirde
	jsr	sci_err

; SCI transmitter interrupt
	org	p:i_scitd
	jsr	sci_xmt

; SCI timer
	org	p:i_scitm
	jsr	sci_tim

; Host xmitter interrupt
	org	p:i_hsttd
	jsr	host_x

	endsec


	section Init
	xdef	cpu_swi,host_x,start
	xdef	counter
	xref	ssi_ini,sci_ini
	xref	HDLC_i,modem_i
	xref	m_loop
	xref	m_htie,m_hcr,m_htx,m_bcr,m_pcc,m_pcddr,m_pcd,m_ipr

	nolist
	include 'macros'
	list

	org	p:

;****************************
;*    Software interrupt    *
;****************************
; mainly for debug purposes, a1 contains bug trap id
cpu_swi bset	#m_htie,x:m_hcr
	jcc	<cpu_sw2
	jset	#0,y:<hbstat,cpu_sw3
	stor	swir,0
	move		    y:<hbwptr,r0
	move		    #0,n0
	move		    #hbuflen-1,m0
	nop
	move		    a1,x:(r0)+
	move		    r0,y:<hbwptr

; check if buffer full
	move		    #>$7ff,a
	move		    r0,x0
	cmp	x0,a
	jne	<cpu_sw1
	bset	#0,y:<hbstat
cpu_sw1 rclr	swir,0
	rti

cpu_sw2 movep		    a1,x:m_htx
cpu_sw3 rti


;****************************
;*    Host xmit interrupt   *
;****************************
host_x
	enter	hostx
	stor	hostxr,0

	move		    y:<hbrptr,r0
	move		    #0,n0
	move		    #hbuflen-1,m0
;	 bclr	 #0,y:<hbstat

; check if buffer empty
	move		    y:<hbwptr,a
	move		    r0,x0
	cmp	x0,a
	jeq	<hostx1

; no, put data out
	movep	x:(r0)+,x:m_htx
	move		    r0,y:<hbrptr
	jmp	<hostx2

; yes, shut down xmitter
hostx1	bclr	#m_htie,x:m_hcr

hostx2	rclr	hostxr,0
	leave	hostx


;****************************
;*   Start of the program   *
;****************************
; no wait states on external memory
start	movep		    #$0000,x:m_bcr	    ; no wait states

; initialize peripherals
	jsr	<ssi_ini
	jsr	<sci_ini

; initialize port C
	movep		    #$01eb,x:m_pcc	    ; TXD,RXD,SC0,SC2,SCK,SRD,STD
	movep		    #$0014,x:m_pcddr	    ; SCLK,SC1 as output
	movep		    #$0000,x:m_pcd

; initialize debug system and status
	movi	0,y:<hbstat
	movi	hbuf,y:<hbrptr
	movi	hbuf,y:<hbwptr
	movi	1,y:<counter

	movi	0,x:<status

; initialize modules (with external hardware)
	jsr	<HDLC_i
	jsr	<modem_i

; start interrupts
	movep		    #$b400,x:m_ipr	    ; SSI=IPL2,SCI=IPL1,HOST=IPL0
	andi	#$fc,mr 			    ; unmask interrupts

; and jump to the main modem module
	jmp	<m_loop

	org	x:

hbuflen equ	1024
hbuf	ds	hbuflen


	org	y:
hbstat	ds	1
hbwptr	ds	1
hbrptr	ds	1
counter ds	1

	endsec



;****************************
;*   STATUS OF THE MODEM    *
;****************************

	section Status

	xdef	status

	org	x:

status	ds	1

	endsec


;****************************
;*  CONTEX STORE FOR INTS   *
;****************************

	section Contexts
	xdef	scix,scir,scit,hostx
	xdef	scixr,scirr,swir,hostxr

	nolist
	include 'macros'
	list

	org	l:

scix	roomab
scir	roomab
scit	roomab
hostx	roomab

	org	y:

scixr	roomr
scirr	roomr
swir	roomr
hostxr	roomr

	endsec

	end
