/*
 *  DSPLDRV.C -- Display drivers
 *
 *  Copyright (C) 1989, 1990 by Jarkko Vuori. All rights reserved.
 *  Author(s): J Vuori
 *  Modification(s):
 */

#include <graph.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "dspldrv.h"


struct {
    struct xycoord lowerleft,
		   highright;
    int 	   pixperchrx,
		   pixperchry;
    int 	   colors;
} dimensions;
static int *prev;


/*
 * Initializes display
 */
int InitDspl(int dx, int dy) {
    struct videoconfig config;
    int 	       x, y;
    char	       buf[80];

    if (!(prev = calloc(dx, sizeof(int))) ||
	 _setvideomode(_VRES16COLOR) ||
	 _setvideomode(_HERCMONO)) {
	_getvideoconfig(&config);

	dimensions.colors = config.numcolors;

	/* calculate screen dimensions */
	dimensions.lowerleft.xcoord = (config.numxpixels-dx)/2;
	dimensions.lowerleft.ycoord = (config.numypixels-dy)/2;

	dimensions.highright.xcoord = dimensions.lowerleft.xcoord+dx;
	dimensions.highright.ycoord = dimensions.lowerleft.ycoord+dy;

	dimensions.pixperchrx = config.numxpixels/80;
	dimensions.pixperchry = config.numypixels/25;

	/* plot surrounding rectangle */
	_setbkcolor(_BLUE); _setcolor(7);
	_rectangle(_GBORDER, dimensions.lowerleft.xcoord-1,
			     dimensions.lowerleft.ycoord,
			     dimensions.highright.xcoord,
			     dimensions.highright.ycoord+1);
	/* plot horizontal ticks */
	for (x = dimensions.lowerleft.xcoord; x <= dimensions.highright.xcoord; x += dx/16) {
	    _moveto(x, dimensions.highright.ycoord+1);
	    _lineto(x, dimensions.highright.ycoord+1+dimensions.pixperchry/2);
	}
	/* plot vertical ticks */
	for (y = dimensions.highright.ycoord; y >= dimensions.lowerleft.ycoord; y -= dy/8) {
	    _moveto(dimensions.lowerleft.xcoord-2, y);
	    _lineto(dimensions.lowerleft.xcoord-2-dimensions.pixperchrx, y);
	}


	_settextcolor(7);
	sprintf(buf, "DSP CARD 3 Data Grapher (%s)", __DATE__);
	_settextposition(5+config.numypixels/dimensions.pixperchry, 1+(80-strlen(buf))/2);
	_outtext(buf);

	_setcliprgn(dimensions.lowerleft.xcoord, dimensions.lowerleft.ycoord+1,
		    dimensions.highright.xcoord-1, dimensions.highright.ycoord);

	return(0);
    } else
	return (-1);
}


/*
 * Plot one data value
 */
void PlotLine(int x, int y) {
    if (x) {
	_setcolor((dimensions.colors > 2) ? 3 : 7);
	_moveto(dimensions.lowerleft.xcoord+x-1, dimensions.highright.ycoord-prev[x-1]);
	_lineto(dimensions.lowerleft.xcoord+x, dimensions.highright.ycoord-y);
    }

    prev[x] = y;
}


/*
 * Plot circle
 */
void PlotCircle(x, y) {
    _setcolor((dimensions.colors > 2) ? 3 : 7);
    _ellipse(_GBORDER, dimensions.lowerleft.xcoord+x-3, dimensions.highright.ycoord-(y-3),
		       dimensions.lowerleft.xcoord+x+3, dimensions.highright.ycoord-(y+3));
}

/*
 * Plot zero line
 */
void PlotZeroLine(int y) {
    _setcolor((dimensions.colors > 2) ? 3 : 7);
    _moveto(dimensions.lowerleft.xcoord, dimensions.highright.ycoord-y);
    _lineto(dimensions.lowerleft.xcoord+512, dimensions.highright.ycoord-y);
}

/*
 * Erase current plot area
 */
void ErasePlot(void) {
    _clearscreen(_GVIEWPORT);
}


/*
 * releases display
 */
void ReleaseDspl() {
    _setvideomode(_DEFAULTMODE);
    free(prev);
}
