/*
   repl.c : Copyright Paul Healy, EI9GL, 1992.

   Todo: pull the @ file into another buffer for editors that are capable

   Derived from bm.

   Copyright 1986 Bdale Garbee, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.
   Copyright 1987 1988 Dave Trulli NN2Z, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.

   920708 : Created
   921228 : redid using the buf library routines
   930811 : finished quoted reply. Still needs auto cc work.
*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>
#include "config.h"
#include "rc.h"
#include "buf.h"
#include "mailfile.h"
#include "header.h"
#include "send.h"
#include "misc.h"
#include "help.h"
#include "mailer.h"
#include "message.h"
#include "extract.h"
#include "pc.h"

#ifdef BMH
#define main repl_main
#endif

#define REPLNAME "@"

static int
quotemsg(FILE *fp, char *filename, Header_t *h)
{
   char *qchar = getrc(quote);
   char buf[LINELEN];

   if ( (qchar == NULL) || (qchar[0] == 'n') )
      return 0;

   sprintf(buf, "%s writes:\n", name(HEADERP(h, FROM)));
   if ( (addstr("repl", filename, buf) == -1) ||
        (addquotedmsg("repl", filename, fp, qchar) == -1)
      )
      return -1;
   else
      return 0;
}

static int
getreply(Header_t *hdr, char *to, char *subj)
{
   repladdr(to, HEADERP(hdr, REPLYTO), HEADERP(hdr, FROM));

   if (HEADERP(hdr, SUBJECT) != NULL) {
      if (strnicmp(HEADERP(hdr, SUBJECT), "Re:", 3))          /* No Re: yet? */
         sprintf(subj, "Re: %s", HEADERP(hdr, SUBJECT));
      else   /* there's an Re:, let's not add another */
         strcpy(subj, HEADERP(hdr, SUBJECT));
      }
   else
      strcpy(subj, "Re:");

   return 0;
}

static int
addsig(char *filename)
{
   FILE *fp = fopen(filename, "a");

   if (fp == NULL)
      return -1;
   else {
      int ret = dosignature(fp);
      fclose(fp);
      return ret;
      }
}


static int
newfile(void)
{
   int fd = creat(REPLNAME, S_IREAD | S_IWRITE);

   if (fd == -1)
      return -1;
   else
      return close(fd);
}

static int
setup(Mailfile *m, int argc, char *argv[], char *filename)
{
   Header_t h;
   char to[LINELEN], subj[LINELEN];
   FILE *fp;

   if ( (newfile() == -1) ||
        (append("repl", m, REPLNAME, mail_thismsg(m, argc, argv)) == -1) ||
        ((fp=gethdrfn(REPLNAME, &h)) == NULL) ||
        (getreply(&h, to, subj) == -1) ||
        (mktemplate("repl", filename, to, subj, "", 0) == -1) ||
        (quotemsg(fp, filename, &h) == -1) ||
        (addsig(filename) == -1)
      ) {

      if (fp != NULL)
         fclose(fp);

      unlink(filename);
      return -1;
      }
   else
      return fclose(fp);
}

int
repl(Mailfile *m, int argc, char *argv[])
{
   char filename[MAXNAMLEN];

   if (mail_getfile(m, argc, argv, "repl") == -1)
      return -1;
   else if (setup(m, argc, argv, filename) == -1) {
      unlink(REPLNAME);
      return -1;
      }
   else if (call_ed("repl", filename) == -1) {
      unlink(filename);
      unlink(REPLNAME);
      return -1;
      }
   else {
      int ret = dowhatnow("repl", filename, dosend);
      unlink(REPLNAME);
      return ret;
      }
}

/* ---------------------------------------------------------------------- */

int
main(int argc, char *argv[])
{
   Mailfile m;

   if (dohelp(argc, argv, "repl [+folder] [msg] (reply to a message)"))
      return 0;

   mail_zero(&m);

   repl(&m, argc, argv);

   return mail_savestate(&m);
}
