/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */

rsrt42(rec)
struct table_entry *rec;
{
	link->llcn=clear_vcs(link->llcn);
	kill(link->t20);	/* clear_timer if one was running */
	link->am_dce= (bgetch(&pbuff) == DTE_Orig);
	link->coll=0;		/* Reset Restart Collision Flag */
	set_r(link,R3,0);	/* Send RestartC */
	set_r(link,R1,0);	/* We are now ready */
}

void
rsrtco(lnk)
register struct axcb *lnk;
{
	if (lnk->R == R0) /* Try to bring up link */ {
/*		kill(lnk->t20); */
		set_r(lnk,R2,DTE_Orig);
		queue(&lnk->t20,rsrtco,rnd(COLLTIME),lnk);
	}
}

rsrt43(rec)
struct table_entry *rec;
{
	register struct VCS *vc;
	static struct VCS *vcx;
	void rx_up_ignore(),tx_up_flush();
	int st_up_wait();

	kill(link->t20);
	link->llcn=clear_vcs(link->llcn);
	if (bgetch(&pbuff) == (DTE_Orig>>8)) /* Checking the Cause field */ {
		link->coll += 1;	/* Inc N20 */
		if (link->coll > COLLMAX) /* N20 count exceeded */ {
			/* Clear calls queued and tear down link */
			while (link->llcn) /* != NULL */ {
				vc=link->llcn;
				link->llcn=vc->next;
				vcx=vc->peer;
				vcx->peer=NULL;
				set_p(vcx, P6,Out_of_Order);
				vcx=vc;
				free_vc(vcx);
			}
			link->r_upcall=rx_up_ignore;	/* Tear Down Link */
			link->t_upcall=tx_up_flush;
			link->s_upcall=st_up_wait;
		}
		else {
			set_r(link,R0,0);
			queue(&link->t20,rsrtco,rnd(COLLTIME),link);
			/* wait max of COLLTIME */
		}
	}
	else {
		link->coll=0;
		link->am_dce=FALSE;
		set_r(link,R1,0);
	}
}

rsrt44(rec)
struct table_entry *rec;

{
	link->llcn=clear_vcs(link->llcn);
	kill(link->t20);
	link->am_dce=FALSE;
	link->coll=0;		/* Reset Restart Collision Flag */
	set_r(link,R1,0);	/* We are now ready */
}
