/*
   rmm.c : Copyright Paul Healy, EI9GL, 1992.

   911227 : Added this header
   920708 : Version 0.1
   920711 : Profiled and optimised a bit.
            The program spends at least 60% of its time scanning for the
            next message. About the only thing that could be done is to
            block copy the rest of the file after the last message to be
            deleted is found - this is a special case optimisiation which
            will not be any good when a message is to be deleted at or towards
            the end of the mail file.
   920808 : Changed selection for new current message after a message(s) is
            deleted.
   930207 : Updated to use the message walking library.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include "config.h"
#include "rc.h"
#include "lock.h"
#include "misc.h"
#include "help.h"
#include "mailer.h"
#include "pc.h"
#include "buf.h"
#include "mailfile.h"
#include "message.h"
#include "bmh.h"

#ifdef BMH
#define main rmm_main
#endif

#undef DEBUG

static FILE *RMM_fp = NULL;

static MsgResult
dormm(Mailfile *m, char *line)
{
   return mail_msg2file(m, RMM_fp, line);
}

static int   
callrmm(Mailfile *m, int argc, char *argv[])
{
   int result;

   if ( (RMM_fp = fopenbuf(mail_newname(m), "wb", STDOUT_BUFSIZE)) == NULL)
      return -1;

   mail_keeplock(m);

   if (argc == 1)
      result = walknot2msg(m, dormm, mail_curmsg(m));
   else 
      result = walknotselect(m, dormm, argc, argv);

   mail_releaselock(m);
#ifdef DEBUG
   fprintf(stderr, "rmm: about to do a proper close\n");
#endif
   mail_close(m);

   fclose(RMM_fp);

   if (result == -1) {
      unlink(mail_newname(m));
      return -1;
      }
   else
      return mail_update(m, "rmm");
}

int
rmm(Mailfile *m, int argc, char *argv[])
{
   if ( (mail_getfile(m, argc, argv, "rmm") == -1) ||
        (callrmm(m, argc, argv) == -1)
      ) {
      fprintf(stderr, "rmm: failed on %s\n", mail_filename(m));
      return -1;
      }
   else if (mail_size(m)==-1) {
      fprintf(stderr, "rmm: warning check empty failed, %s may still be locked\n", mail_filename(m));
      return -1;
      }
   else {
      mail_setcur(m, mail_prevmsg(m, argc, argv));
#ifdef INDEX
      mapit(m);
#endif
      return 0;
      }
}

int
main(int argc, char *argv[])
{
   Mailfile m;

   if (dohelp(argc, argv, "rmm [+folder] [msglist] (delete mail)"))
      return 0;

   mail_zero(&m);

   rmm(&m, argc, argv);

   return mail_savestate(&m);
}
