/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */

extern unsigned char ndnics;

struct dnic {
	unsigned char adr[3];	/* DNIC in X.121 format */
	unsigned char *rt;		/* Routing table root */
	int len;		/* Length of route */
};

extern struct dnic DNics[];
extern unsigned char NPAroute[][16];

int check_x121(ptrx,adr)
register unsigned char adr[];
unsigned char ptrx[];
{
	static unsigned char i, j, alen, *ptr;

	ptr=ptrx;
	alen=ptr[0] & 0x0f;
	if ( (adr[0] & 0x0f < alen) || (alen == 0) ) return -1;
	j=alen>>1;

	for (i=1; i<=j; i++)
		if (adr[i] != ptr[i]) return -1;

	if (alen & 1 && /* Length is odd, check last nibble */
		 ( ptr[i] &0xf0 != (adr[i] & 0x0f0) )) return -1;
	if (alen == adr[0]&0x0f) return 0;
	return 1;
}

find_route(ptr,x121)
unsigned char **ptr;
unsigned char x121[];
{
	static unsigned char i, disp, lensv;
	static int l, r;
	register unsigned char *ch;

	for (i=0;i<ndnics;i++) {
		if (check_x121(DNics[i].adr,x121) >= 0) break;
	}
	if (i == ndnics) return 0;	/* Invalid DNIC */
	ch = DNics[i].rt;
	lensv = x121[0] & 0x0f;

	i = DNics[i].adr[0];	/* Start with digit after DNIC */
	r = -1;
	do {
		if (*ch != 0) r = *(ch+1);
		if (*ch == 2) /* Modify Address and return */ {
			adr_copy(0,NPAroute[r],0, x121); /* New Called */
			return -1;	/* need to re-route call */
		}

		if (i == lensv) break;
		l = x121[ (i>>1) +1 ];
		if (!(i&1)) /* High */ l = l>>4;
		i++;
		disp = *(ch+2+(l&0x0f));
		ch += disp;		/* Advance to next record */
	} while (disp != 0);

	if (r == -1) return 0;	/* No known route */
	*ptr = &NPAroute[r][0];
	return 1;
}

l2_adr(x121,strd,str)
unsigned char x121[16];
struct ax25_addr *strd;
struct ax25_addr *str;
{
	static int i, j,l,ch;
	register struct ax25_addr *s;

	for (i=0;i<2;i++) {
		if (i=0) s=strd;
		else s=str;
		s->ssid=0;
		for (i=0;i<ALEN;i++) s->call[i]=0x40;
	}
	if ( (l=x121[0]) < 10) return;
	if (l>11) l=11;

	s=strd;
	if (check_x121(DNics[0].adr,x121) != 1) /* Not default DNIC */ {
		j=0;
		for (i=1;i<=4;i++) {
			ch=x121[((i+1) >> 1)];
			if (i & 1) ch= (ch & 0x0ff) >> 4;
			else ch=ch & 0x0f;
			s->call[j++] = (ch | 0x30) << 1;
		}
	}
	else s->call[0]=0;	/* No call here */

	s=str;
	j=0;
	for (i=5;i<=l;i++) {
		ch=x121[((i+1) >> 1)];
		if (i & 1) ch= (ch & 0x0ff) >> 4;
		else ch=ch & 0x0f;
		if (i == 11) str->ssid= ch<<1;
		else s->call[j++] = (ch | 0x30) << 1;
	}
}

l3_adr(str,x121)
struct ax25_addr *str;
unsigned char x121[16];
{
	static int i,len,ch,hi,l;
	static unsigned char x[7];
	register unsigned char *c;

	l = x121[0];
	x121[0] = 0;
	c=str->call;
	for (len=0;len<ALEN;) /* Insure valid [shifted] numbers */ {
		ch = (*c++)>>1;
		if (ch == 0x20) break; /* End of "callsign" */
		if (ch == 0x4f) ch=0x30;	/* O = 0 */
		if (ch == 0x49) ch=0x31;	/* I = 1 */
		if ((ch < 0x30) || (ch > 0x39)) return; /* Non-digit */
		x[len++] = (ch - 0x30);
	}
	ch = ((str->ssid)>>1 & 0x0f);
	if (ch>9) return;	/* Only 0-9 */
	if (ch) x[len++] = ch;	/* Only add if non-zero */

	c = x121;
	*c = l;
	if ( !(*c) ) {
		if (len == 4) *c++ = 0; /* DNIC was specified */
		else /* Use default DNIC */ {
			for (i=0;i<3;i++) *c++ = DNics[0].adr[i];
		}
	} else c += (*c>>1)+1; /* Adding to the address */

	hi = (! (x121[0] & 1) );
	x121[0] += len;

	for (i=0;i<len;i++) {
		ch = x[i];
		if (hi) *c = ch<<4;
		else {
			*c |= ch;
			c++;
		}
		hi = !hi;
	}
}

