/* Mods by SM6RPZ. For history, see rspf.c */
#include "global.h"
#include "mbuf.h"
#include "timer.h"
#include "iface.h"
#include "cmdparse.h"
#include "netuser.h"
#include "internet.h"
#include "ip.h"
#include "socket.h"
#include "rspf.h"

#ifdef AUTOROUTE
extern int Ax25_autoroute;
int RspfActive = 0;
#endif

extern unsigned char Router_horizon;	/* from rspf.c */
extern unsigned char Portable_horizon;	/* from rspf.c */
extern unsigned char Group_horizon;	/* from rspf.c */
extern unsigned char Local_horizon;	/* from rspf.c */
extern unsigned char Non_rspf_horizon;	/* from rspf.c */
extern int Keeprouter;			/* from rspf.c */
extern int Nrspfproc;			/* from rspf.c */
extern unsigned char Rspfsubseq;	/* from rspf.c */
extern short Rspfseq;			/* from rspf.c */

int Rspfownmode = -1;

static int dorspfrefuse __ARGS((int argc,char *argv[],void *p));/* sm6rpz */
static int dohorizon __ARGS((int argc,char *argv[],void *p));	/* sm6rpz */
static int dorspftrace __ARGS((int argc,char *argv[],void *p));	/* sm6rpz */
static int dorspfkick __ARGS((int argc,char *argv[],void *p));	/* sm6rpz */
static int dointerface __ARGS((int argc,char *argv[],void *p));
static int domessage __ARGS((int argc,char *argv[],void *p));
static int domaxping __ARGS((int argc,char *argv[],void *p));
static int domode __ARGS((int argc,char *argv[],void *p));
static int dorrhtimer __ARGS((int argc,char *argv[],void *p));
static int dotimer __ARGS((int argc,char *argv[],void *p));
static int doroutes __ARGS((int argc,char *argv[],void *p));
static int dostatus __ARGS((int argc,char *argv[],void *p));
static int dosuspect __ARGS((int argc,char *argv[],void *p));
static struct timer rrhtimer, rspftimer;

int
dorspf(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    struct cmds rspfcmds[] = {
	"horizon",	dohorizon,	0,	0,	NULLCHAR, /* sm6rpz */
	"interface",    dointerface,    0,	0,	NULLCHAR,
	"kick",		dorspfkick,	0,	0,	NULLCHAR, /* sm6rpz */
	"message",	domessage,	0,	0,	NULLCHAR,
	"maxping",	domaxping,	0,	0,	NULLCHAR,
	"mode",		domode,		0,	0,	NULLCHAR,
	"refuse",	dorspfrefuse,   0,	0,	NULLCHAR, /* sm6rpz */
	"rrhtimer",	dorrhtimer,	0,	0,	NULLCHAR,
	"routes",	doroutes,	0,	0,	NULLCHAR,
	"status",	dostatus,	0,	0,	NULLCHAR,
	"suspecttimer", dosuspect,	0,	0,	NULLCHAR,
	"timer",	dotimer,	0,	0,	NULLCHAR,
	"trace",	dorspftrace,	0,	0,	NULLCHAR, /* sm6rpz */
	NULLCHAR
    };
    return subcmd(rspfcmds,argc,argv,p);
}

/* The suspect timer controls how often old links expire. When a link has
 * expired, we try to renew its entry by various methods.
 */
static int
dosuspect(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    if(Rspfifaces == NULLRIFACE){
	tprintf("RSPF is not active - define interface first.\n");
	return 0;
    }
    if(argc < 2){
	tprintf("Suspect timer: %lu/%lu seconds\n",
	    read_timer(&Susptimer)/1000L,dur_timer(&Susptimer)/1000L);
	return 0;
    }
    stop_timer(&Susptimer);			/* if it's running */
    Susptimer.func = rspfsuspect;		/* what to call on timeout */
    Susptimer.arg = NULL;			/* dummy value */
    set_timer(&Susptimer,atol(argv[1])*1000L);	/* set timer duration */
    start_timer(&Susptimer);			/* and fire it up */
    return 0;
}

/* The RRH timer controls the interval between Router-To-Router Hello
 * messages. These messages announce that your station is live and well
 * and that you are willing to exchange RSPF routing updates.
 */
static int
dorrhtimer(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    if(Rspfifaces == NULLRIFACE){
	tprintf("RSPF is not active - define interface first.\n");
	return 0;
    }
    if(argc < 2){
	tprintf("RRH timer: %lu/%lu seconds\n",
	    read_timer(&rrhtimer)/1000L, dur_timer(&rrhtimer)/1000L);
	    return 0;
    }
    stop_timer(&rrhtimer);			/* if it's running */
    rrhtimer.func = rspfevent;			/* what to call on timeout */
    rrhtimer.arg = (void *) &rrhtimer;
    set_timer(&rrhtimer,atol(argv[1])*1000L);	/* set timer duration */
    start_timer(&rrhtimer);			/* and fire it up */
    return 0;
}

/* This timer controls the interval between the RSPF routing updates. */
static int
dotimer(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    if(Rspfifaces == NULLRIFACE){
	tprintf("RSPF is not active - define interface first.\n");
	return 0;
    }
    if(argc < 2){
	tprintf("RSPF update timer: %lu/%lu seconds\n",
	    read_timer(&rspftimer)/1000L, dur_timer(&rspftimer)/1000L);
	    return 0;
    }
    stop_timer(&rspftimer);			/* if it's running */
    rspftimer.func = rspfevent;			/* what to call on timeout */
    rspftimer.arg = (void *) &rspftimer;
    set_timer(&rspftimer,atol(argv[1])*1000L);	/* set timer duration */
    start_timer(&rspftimer);			/* and fire it up */
    return 0;
}

/* Called when either the RRH timer, the Update timer or the adjacencies
 * suspect timer expires.
 */
void
rspfevent(t)
void *t;
{
    char cmd;
    struct mbuf *bp;
    struct rspfadj *adj = NULLADJ;
    struct timer *tp;

    tp = (struct timer *)t;
    if(tp == &rrhtimer) {
	cmd = RSPFE_RRH;    /* It is time to send a new RRH */
	start_timer(tp);
    } else if(tp == &rspftimer) {
	cmd = RSPFE_UPDATE; /* Time to send a new routing updates */
	start_timer(tp);
    } else {
	for(adj = Adjs; adj != NULLADJ; adj = adj->next)
	    if(&adj->timer == tp)
		break;
	if(adj == NULLADJ)
	    return;
	cmd = RSPFE_CHECK;  /* An adjacency has becomed suspect */
    }
    bp = ambufw(1+sizeof(adj));
    *bp->data = cmd;
    memcpy(bp->data + 1,&adj,sizeof(adj));
    bp->cnt = bp->size;
    enqueue(&Rspfinq,bp);
}

static int
domessage(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    if(argc > 2) {
    	tputs("Usage: rspf message \"<your message>\"\n");
	return 1;
    }

    if(argc < 2) {
	if(Rrh_message != NULLCHAR)
	    tputs(Rrh_message);
    } else {
	if(Rrh_message != NULLCHAR) {
	    free(Rrh_message);
	    Rrh_message = NULLCHAR;		    /* reset the pointer */
	}
	if(!strlen(argv[1]))
	    return 0;				    /* clearing the buffer */
	Rrh_message = callocw(1,strlen(argv[1])+5);   /* allow for EOL */
	strcpy(Rrh_message, argv[1]);
	strcat(Rrh_message, INET_EOL);		    /* add the EOL char */
    }
    return 0;
}

static int
domaxping(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    return setshort(&Rspfpingmax,"Max failed pings before deleting adjacency",
	argc,argv);
}

static int
domode(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    if(argc < 2) {
	tputs("RSPF preferred mode is ");
	if(Rspfownmode == -1)
	    tputs("not set.\n");
	else
	    tprintf("%s.\n",(Rspfownmode & CONNECT_MODE) ? "VC mode" :
		"Datagram mode");
	return 0;
    }
    switch(*argv[1]){
    case 'v':
    case 'c':
    case 'V':
    case 'C':
	Rspfownmode = CONNECT_MODE;
	break;
    case 'd':
    case 'D':
    	Rspfownmode = DATAGRAM_MODE;
	break;
    case 'n':
    case 'N':
	Rspfownmode = -1;
	break;
    default:
	tputs("Usage: rspf mode [vc|datagram|none]\n");
	return 1;
    }
    return 0;
}

static int
dointerface(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    struct rspfiface *riface;
    struct iface *iface;
    struct mbuf *bp;
    int q;

    if(argc < 2){
	tputs("Iface    Quality\n");
	for(riface = Rspfifaces; riface != NULLRIFACE; riface = riface->next)
	    tprintf("%-9s%-11d\n", riface->iface->name, riface->quality);
	return 0;
    }
    if(argc != 3){
	tputs("Usage: rspf interface <name> <quality>\n");
	return 1;
    }
    if((iface = if_lookup(argv[1])) == NULLIF){
	tputs("No such interface.\n");
	return 1;
    }
    if(iface->broadcast == 0){
	tprintf("Broadcast address for interface %s not set.\n",argv[1]);
	return 1;
    }
    q = atoi(argv[2]);
    if(q < 1 || q > 127){
	tputs("Quality must be between 1 and 127\n");
	return 1;
    }
    if((riface = riface_lookup(iface)) == NULLRIFACE) {
	riface = (struct rspfiface *)callocw(1,sizeof(struct rspfiface));
	riface->iface = iface;
	riface->next = Rspfifaces;
	if(Rspfifaces == NULLRIFACE) {
#ifdef AUTOROUTE
	    RspfActive = 1;
	    Ax25_autoroute = 0;
#endif /* AUTOROUTE */
	    newproc("RSPF",1024,rspfmain,0,NULL,NULL,0);
	}
	Rspfifaces = riface;
    }
    riface->quality = q;
    bp = ambufw(1+sizeof(riface));
    *bp->data = RSPFE_RRH;		/* Send an RRH immediately */
    memcpy(bp->data + 1,&riface,sizeof(riface));
    bp->cnt = bp->size;
    enqueue(&Rspfinq,bp);
    return 0;
}

/* Display accumulated routing updates */
static int
doroutes(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    struct mbuf *bp;
    struct rspfrouter *rr;

    if(Rspfifaces == NULLRIFACE){
	tprintf("RSPF is not active - define interface first.\n");
	return 0;
    }
    bp = makeownupdate(Ip_addr);
    if(bp == NULLBUF && Rspfrouters == NULLRROUTER) {
    	tputs("No routing information is available.\n");
	return 0;
    }
    if(bp != NULLBUF) {
    	tputs("      Local routing update:");
	if(Rspfsubseq)
	    tprintf("  Last subseq: %u",Rspfsubseq);
	tputc('\n');
	rspfnodedump(NULLFILE,&bp,0);
	tputc('\n');
    }
    for(rr = Rspfrouters; rr != NULLRROUTER; rr = rr->next) {
	tprintf("      Time since receipt: %s",tformat(secclock() - rr->time));
	if(rr->subseq)
	    tprintf("  Last subseq: %u",rr->subseq);
	if(rr->sent)
	    tputs("  Propagated");
	tputc('\n');
	if(rr->data != NULLBUF) {
	    dup_p(&bp,rr->data,0,len_p(rr->data));
	    rspfnodedump(NULLFILE,&bp,0);
	    tputc('\n');
	}
    }
    return 0;
}

static int
dostatus(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    struct rspfreasm *re;
    struct rspfadj *adj;
    struct mbuf *bp;
    union rspf rspf;

    if(Rspfifaces == NULLRIFACE){
	tprintf("RSPF is not active - define interface first.\n");
	return 0;
    }
    tprintf("Bad checksum %u, Bad version %u, Not RSPF interface %u, Refused %u\n",
	Rspf_stat.badcsum,Rspf_stat.badvers,Rspf_stat.norspfiface,Rspf_stat.refused);
    tprintf("RRH in %u, RRH out %u, Update in %u, Update out %u\n",
	Rspf_stat.rrhin,Rspf_stat.rrhout,Rspf_stat.updatein,Rspf_stat.updateout);
    tprintf("Non-adjacency update %u, Old node report %u, Polls sent %u\n",
	Rspf_stat.noadjupdate,Rspf_stat.oldreport,Rspf_stat.outpolls);
    tprintf("Nrspfproc %d/%d, Keeprouter %d, Sequence %u, Subsequence %u\n",
	Nrspfproc,RSPF_PROCMAX,Keeprouter,(int16)Rspfseq,Rspfsubseq);
    tprintf("Timers: RRH %lu/%lu, update %lu/%lu, suspect %lu/%lu seconds\n",
	read_timer(&rrhtimer)/1000L,  dur_timer(&rrhtimer)/1000L,
	read_timer(&rspftimer)/1000L, dur_timer(&rspftimer)/1000L,
	read_timer(&Susptimer)/1000L, dur_timer(&Susptimer)/1000L);
    if(Adjs == NULLADJ)
	return 0;
    tputs("Address         T Interface Cost Sent  Heard    Timer   TOS State\n");
    for(adj = Adjs; adj != NULLADJ; adj = adj->next) {
	tprintf("%-15s %c %-9s %-4u %-5u %-6lu ", inet_ntoa(adj->addr),
	    (adj->seq || rr_lookup(adj->addr) != NULLRROUTER) ? 'R' : 'E',
	    adj->riface->iface->name,uchar(adj->cost),adj->seq,adj->heard);
	if(run_timer(&adj->timer))
	    tprintf("%4lu/%-4lu ",
		read_timer(&adj->timer)/1000L, dur_timer(&adj->timer)/1000L);
	else
	    if(adj->state & RSPF_SUSPECT)
		tputs(" pinging  ");
	tprintf("%-3u ", uchar(adj->tos));
	switch(adj->state) {
	case RSPF_TENTATIVE:
	    tputs("Tentative\n");
	    break;
	case RSPF_OK:
	    tputs("OK\n");
	    break;
	case RSPF_SUSPECT:
	    tputs("Suspect\n");
	    break;
	case RSPF_BAD:
	    tputs("Bad\n");
	    break;
	default:
	    tprintf("Unknown, %u\n", uchar(adj->state));
	    break;
	}
    }
    if(run_timer(&Rspfreasmt))
	tprintf("Reassembly timer running: %lu/%lu seconds\n",
	    read_timer(&Rspfreasmt)/1000L, dur_timer(&Rspfreasmt)/1000L);
    if(Rspfreasmq != NULLRREASM)
	tputs("Reassembly fragments:\n");
    for(re = Rspfreasmq; re != NULLRREASM; re = re->next) {
	tprintf("Src %s time since last frag %s",inet_ntoa(re->addr),
	    tformat((secclock() - re->time)));
	if(dup_p(&bp,re->data,0,RSPFPKTLEN) == RSPFPKTLEN
	   && ntohrspf(&rspf,&bp) != -1)
	    tprintf(" frag count %u/%u\n",len_q(re->data),rspf.pkthdr.fragtot);
	else {
	    tputc('\n');
	    free_p(bp);
	    continue;
	}
    }
    return 0;
}

/* sm6rpz */
static int
dohorizon(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    unsigned char horizon;

    if((argc > 1 && *argv[1] == '?') || argc > 3) {
	tputs("Usage: rspf horizon router|group|local|portable|non_rspf <horizon>\n");
	return 1;
    }
    if(argc == 1) {
	tprintf("Horizon: router   = %u\n", Router_horizon);
	tprintf("         portable = %u\n", Portable_horizon);
	tprintf("         group    = %u\n", Group_horizon);
	tprintf("         local    = %u\n", Local_horizon);
	tprintf("         non_rpsf = %u\n", Non_rspf_horizon);
	return 0;
    }
    if(argc == 3) {
	horizon = atoi(argv[2]);
	if(horizon > 255) {
	    tputs("Horizon must be between 0 and 255 (if 0, it will not be propagated).\n");
	    return 0;
	}
    }
    switch(*argv[1]) {
    case 'r':
	if(argc == 3)
	    Router_horizon = horizon;
	tprintf("Router horizon = %u\n", Router_horizon);
	break;
    case 'g':
	if(argc == 3)
	    Group_horizon = horizon;
	tprintf("Group horizon = %u\n", Group_horizon);
	break;
    case 'l':
	if(argc == 3)
	    Local_horizon = horizon;
	tprintf("Local horizon = %u\n", Local_horizon);
	break;
    case 'p':
	if(argc == 3)
	    Portable_horizon = horizon;
	tprintf("Portable horizon = %u\n", Portable_horizon);
	break;
    case 'n':
	if(argc == 3)
	    Non_rspf_horizon = horizon;
	tprintf("Non_rspf horizon = %u\n", Non_rspf_horizon);
	break;
    default:
	tputs("Usage: rspf horizon [router|group|local|portable|non_rspf [<horizon>]]\n");
    }
    return 0;
}

/* Add/drop a gateway to the rspf_refuse list which allows us to
 * ignore their advertisements (parts taken from rip.c) -- sm6rpz
 */
static int
dorspfrefuse(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    register struct rspf_refuse *rl;
    int32 addr;

    if((argc > 1 && *argv[1] == '?') || argc == 2 || argc > 3) {
	tputs("Usage: rspf refuse [add|drop <addr>]\n");
	return 1;
    }
    if(argc == 1) {
	tputs("Rspf refuse list:\n");
	for(rl = Rspf_refuse; rl != NULLREFUSE; rl = rl->next)
	    tprintf("%s\n", inet_ntoa(rl->target));
	return 0;
    }
    addr = resolve(argv[2]);
    switch(*argv[1]) {
    case 'a':
	for(rl = Rspf_refuse; rl != NULLREFUSE; rl = rl->next)
	    if(rl->target == addr)
		return 0;	/* Already in table */

	/* get a chunk of memory for the rspf interface descriptor */
	rl = (struct rspf_refuse *)callocw(1,sizeof(struct rspf_refuse));

	/* tack this record on as the first in the list */
	rl->next = Rspf_refuse;
	if(rl->next != NULLREFUSE)
	    rl->next->prev = rl;
	Rspf_refuse = rl;

	/* fill in the gateway to ignore */
	rl->target = addr;
	return 0;
    case 'd':
	for(rl = Rspf_refuse; rl != NULLREFUSE; rl = rl->next)
	    if(rl->target == addr)
		break;
  
	/* leave if we didn't find it */
	if(rl == NULLREFUSE)
	    return 0;

	/* Unlink from list */
	if(rl->next != NULLREFUSE)
	    rl->next->prev = rl->prev;
	if(rl->prev != NULLREFUSE)
	    rl->prev->next = rl->next;
	else
	    Rspf_refuse = rl->next;

	/* and deallocate the structure memory */
	free((char *)rl);
	return 0;
    default:
	tputs("Usage: rspf refuse [add|drop <addr>]\n");
    }
    return 0;
}

/* sm6rpz */
static int
dorspftrace(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    return setshort(&Rspf_trace,"RSPF tracing",argc,argv);
}

/* Called when either the RRH timer, the Update timer or the adjacencies
 * suspect timer expires.
 */
static int
dorspfkick(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    struct mbuf *bp;
    char *notused = NULLCHAR;

    if(argc != 2 || (argc > 1 && *argv[1] == '?')) {
	tputs("Usage: rspf kick rrh|update\n");
	return 1;
    }
    bp = ambufw(5);	    /* 1 byte for cmd + 4 bytes not used */
    switch(*argv[1]) {
    case 'r':
	*bp->data = RSPFE_RRH;
	break;
    case 'u':
	*bp->data = RSPFE_UPDATE;
	break;
    default:
	free_p(bp);	/* clean up */
	tputs("Usage: rspf kick rrh|update\n");
	return 1;
    }
    memcpy(bp->data + 1, &notused, sizeof(notused));
    bp->cnt = bp->size;
    enqueue(&Rspfinq,bp);
    return 0;
}
