/* No Frills or Thrills Satellite Tracking Program */
/* by Charles McGonegal  - N8OKM  12/91 */
/* w/ Ken Friberg - N8PBE */
/* */
/* Computation routine based on AMSAT Orbital Prediction Program */
/* (c) 1980 by Dr. T.A. Clark - W3IWI */
/* See Orbit Magazine, #6, March/April 1981 */
/* */
/* Observer location should be in "location.dat" */
/* latitude - Degrees.fraction North */
/* longitude - Degrees.fraction West */
/* altitude - meters */
/* */
/* Satellite Keplerian elements should be in "elements.dat" */
/* NASA 2-line format */
/* No more than twenty at a time, please */
/* */
/* Output file for 'File' option is 'sat.trk', where sat is name of satellite.*/
/*  Data is appended to file */

#include <stdio.h>
#include <math.h>

#define GM              75369793000000.0 	/* Gravitational Constant */
#define R_earth         6378.135		/* Radius of earth - km   */
#define dR              0.003353		/* inverse of delta radius */
#define SID_SOL         1.0027379093		/* Sidereal : Solar ratio */
#define IN 	        1
#define OUT		0
#define PI		3.14159

struct satellite {
        char name[10];
        int epoch_yr;
        float epoch_time, inclin, raan, eccen, arg_peri, mn_anom, mn_motion,
               sm_axis, range, alt, azi, ecc_anom, phase;
        float  rev_count;

	/* Satellite structure - contains name and elements */
	/* phase is phase units (0-256) */

} sats[21];

struct pt {
        float x, y, z, lat, ln;

	/* This structure holds a geocentric point - lt and ln are options */

};

struct time_s {
        int year;
        float day;

	/* Time used for program - day is day of year and fraction */

};

main()
{
	FILE *printer;			/* Where to print output */
        extern struct satellite sats[];	/* Array of satellites loaded */
        int num_of_sats, oscar, leapy, pass_state, repeat, n;
		/* How many sats were loaded, which one was selected, */
		/* a leapy flag, a pass flag, and a repeat flag */
        struct pt obsloc;		/* Observer's location */
        struct time_s t, stop_t;	/* start and stop times */
        float increment;		/* time increment */
	char buf[3];			/* a little input buffer */
	char fname[13];			/* filename string */

        load_obs(&obsloc);		/* Load Observer's location */
        load_sats(&num_of_sats);	/* Load Satellite elements */

	repeat = 1;
	while (repeat) {
	stop_t.day = t.day = 0;
	t.year = stop_t.year = 0;

        oscar = get_sat(num_of_sats);	/* Ask for the sat of interest */

        get_time(&t, &stop_t);		/* Ask for the start and end dates */
                			/* and times */
	increment = 0.0;
	while(increment <= 0.0)		/* Ask for the time increment */
	        get_inc(&increment);

	if(t.year%4)
		leapy = 0;
	else				/* Set leap year flag */
		leapy = 1;

	printf("\nScreen, Printer, or File? (S/p/f)\n");
	gets(buf);
	if((buf[0] == 'P') || (buf[0] == 'p'))	/* Set Output of calcuations */
		printer = fopen("prn:", "w");
	else if((buf[0] == 'F') || (buf[0] == 'f')){
		n = 0;
		while(sats[oscar].name[n] != '\0') {
			if((sats[oscar].name[n] == '-') ||
			(sats[oscar].name[n]=='/'))
				sats[oscar].name[n] = '_';
			n += 1;
		}
		strcpy(fname, sats[oscar].name);
		strcat(fname, ".trk");
		printer = fopen(fname, "a");
	}
	else
		printer = stdout;

	/* Print a header */

	fprintf(printer, "\nPosition Location for :%10s\n", sats[oscar].name);
	fprintf(printer, " Time \t Azimuth  Altitude   Range \tPhase\n");

	/* Do some iterations ! */

	pass_state = OUT;

        while(((t.year == stop_t.year) && (t.day <= stop_t.day)) ||
		(t.year != stop_t.year)) {
                do_sat(&t, &obsloc, oscar);

		if(sats[oscar].alt < 0.0)
			pass_state = OUT;	/* See if this is a new pass */
		else if(pass_state == OUT) {
			pass_state = IN;
			print_date(t.day, t.year, printer);
		}

		if(sats[oscar].alt >= 0.0){		/* Print data */
			print_data(t.day, oscar, printer);
		}

		t.day +=  increment;
		if(t.day >= 366.0 + leapy) {
			t.day = 1.0 + increment;	/* Increment time */
			t.year += 1;
		}
        }

	if(printer != stdout)
		fclose(printer);			/* Wrap up and exit */

	printf("\nRun another set of passes? (y/N)");
	gets(buf);
	if((buf[0] == 'Y') || (buf[0] == 'y'))
		repeat = 1;
	else
		repeat = 0;
	}

	exit(0);
}

print_date(day, yr, printer)
float day;
int yr;
FILE *printer;
{

	/* Print date that pass occurs */

	int d, m, i;			/* day, month, and counter */
	double ip;			/* int part for modf */
	float dummy;			/* dummy for modf */
	int days[13] = {0,0,31,59,90,120,151,181,212,243,273,304,334};
		/* day of year at beginning of each month */

	dummy = modf(day, &ip);		/* Chop fraction off day */
	day = ip;

	if(!(yr%4)) {
		for(i = 3; i <= 12; i++)	/* Correct for leap year */
			days[i] += 1;
	}

	for(m=1; m <= 12; m++) {
 		if(day <= days[m+1])
			break;			/* Find month */
	}

	d = day - days[m];		/* Find day */

	fprintf(printer, "\n\t\t %2d/%2d/%2d\n", m, d, yr);

}

delta_days(yr1, yr2)
int yr1, yr2;
{

	/* Finds the number of days between New Year's of two different */
	/* years */

	float days;	/* days between dates */
	int n;		/* counter */

	days = 0;
	for(n = yr1; n < yr2; n++) {
		days += 365;
		if(!(n%4))
			days += 1;
	}

	return days;
}

get_time(start, stop)
struct time_s *start, *stop;
{
	/* Gets that starting and ending times and dates, and makes sure one */
 	/* is after the other */

	char startbuf[6] = "start";
	char stopbuf[6] = "stop";	/* strings to send functions */
	char buffer[5];			/* small input buffer */

        while(start->year < 89)		/* Program won't work before 1989 */
                get_d(start, startbuf);

        printf("\nIs stop date different? (y/N)\n");
        gets(buffer);
        if((buffer[0] != 'Y') && (buffer[0] != 'y')) {
                stop->year = start->year;
                stop->day = start->day;
        }
        else {
		get_d(stop, stopbuf);
		while((start->year == stop->year)&&(start->day > stop->day) ||
			(start->year > stop->year))
			get_d(stop, stopbuf);
        }

        get_t(start, startbuf);
        get_t(stop, stopbuf);
        while((start->year == stop->year)&&(start->day >= stop->day))
                get_t(stop, stopbuf);
}

get_d(tm, flag)
struct time_s *tm;
char *flag[];
{
	/* Gets a date and converts it to the day of the year */
	/* Doesn't check to see if you know how many months are in */
	/* a year, or how many days are in a month */

        int yr, month, day, day_of_year, leapy;
        char buffer[10];
        int days[13] = {0,0,31,59,90,120,151,181,212,243,273,304,334};

	/* includes a leap year flag, input buffer, and list of day of year */
	/* at beginning of each month (1-12, not 0-11) */

        printf("\nEnter %s date - mm/dd/yy\n", flag);
        gets(buffer);
        sscanf(buffer, "%d/%d/%d", &month, &day, &yr);

        if(yr%4)
                leapy = 0;
        else				/* Check to see if it's a leapy */
                leapy = 1;

        day_of_year = days[month];
        if(month > 2)
                day_of_year += leapy;

        tm->year = yr;
        tm->day = day_of_year + day;
}

get_t(tm, flag)
struct time_s *tm;
char *flag[];
{
	/* Gets a time and converts it to a fraction of a day, and adds */
	/* to the days of the year given.  Doesn't check to see if you know */
	/* how many hours are in a day, or how many minutes are in an hour. */
	/* Use UTC, and 24 hour format. */

        int hr, min;
        char buffer[10];		/* input buffer */

        printf("\nEnter %s time - hh:mm\n", flag);
        gets(buffer);
        sscanf(buffer, "%d:%d", &hr, &min);

        tm->day += hr/24.0 + min/1440.0;
}
 
print_data(day, i, printer)
float day;
int i;
FILE *printer;
{
	/* Pretty self explanatory if you ask me. */

        extern struct satellite sats[];
	float hr, mn;
	double ip;		/* for modf */

	hr = 24.0 * modf((double)day, &ip);
	mn = 60.0 * modf((double)hr, &ip);
	hr = ip;

	if (mn > 59.5) {
		mn = 0;
		hr += 1;
	}

        fprintf(printer, "%2.0f:%2.0f\t%7.1f\t%8.1f  ",
		hr, mn, sats[i].azi, sats[i].alt);
        fprintf(printer, "  %6.0f\t%5.0f\n", sats[i].range, sats[i].phase);

}

prec_adj(t, w, o, sat)
struct time_s *t;
float *w, *o;
int sat;
{ 
	/* This routine adjusts the argument of perigee and the right */
	/* ascencion of the ascending node */

	float i, e2, e1, k2;	/* inclination and intermediates */
				/* w = argument of perigee, o = raan */

        e2 = 1 - pow(sats[sat].eccen, 2.0);	/* Intermediates */
        e1 = sqrt(e2);

	/* Adjust for precession of orbit */
	
        k2 = 9.95 * pow(R_earth/sats[sat].sm_axis, 3.5) / pow(e2, 2.0);
        k2 = k2 * (t->day - sats[sat].epoch_time);

        i = sats[sat].inclin*PI/180;

        *o = sats[sat].raan - k2 * cos(i);
        *o = *o * PI/180;

        *w = sats[sat].arg_peri + k2*(2.5*pow(cos(i), 2.0) - 0.5);
        *w = *w * PI/180;
}

init_matrix(c, w, o, i)
float c[3][2], w, o, i;
{
	/* This matrix is used in rotation from coordinates in the plane */
	/* of the orbit, to inertial reference frame coordinates */
	/* i = inclination */
	/* w = argument of perigee */
	/* o = raan */

	c[0][0] =  cos(w)*cos(o) - sin(w)*sin(o)*cos(i);
	c[0][1] = -sin(w)*cos(o) - cos(w)*sin(o)*cos(i);
	c[1][0] =  cos(w)*sin(o) + sin(w)*cos(o)*cos(i);
	c[1][1] = -sin(w)*sin(o) + cos(w)*cos(o)*cos(i);
	c[2][0] =  sin(w)*sin(i);
	c[2][1] =  cos(w)*sin(i);
}

float solve_kep(e, ma, sat)
float e, ma;
int sat;
{
	/* Solves Kepler's equation by the Newton-Raphson method - I think */

	float r, m1, m2;   /* some intermediates */
			   /* e, ma = eccentric and mean anomalies */

	/* Trial value for eccentric anomaly */

        if(e == 0)
     e = ma + sats[sat].eccen*sin(ma)+0.5*pow(sats[sat].eccen, 2.0)*sin(2.0*ma);
 
	/* Solve Kepler's equation for Eccentric Anomaly, given mean anomaly */
                               
        for(;;) {
                r = 1 - sats[sat].eccen*cos(e);
                m1 = e - sats[sat].eccen*sin(e);
                m2 = m1 - ma;
                if(fabs(m2) < 1.0e-6)
                        break;
                e = e - m2/r;  
        }

	return e;
}

do_sat(t, obs, sat)
struct time_s *t;
struct pt *obs;
int sat;
{
        extern struct satellite sats[];
        float e, o, w, q, ma, g;
        float x, y, z, x1, y1, z1, s, c2, i, lat, ln;
	float days_eptoyr, days_89toyr;		/* time between various dates */
        double ip;				/* for modf - int part */
	float c[3][2];				/* a rotation matrix */ 

	/* e = eccentric anom.	o = raan	w = argument of perigee */
	/* q = orbital phase	ma = mean anom.	g = rotation variable */
	/* x, y, z, x1, y1, z1,  lat, ln = location variables - geocentric */
	/* s, c2 - intermediates */
	/* i = inclination */

	days_89toyr = delta_days(89, t->year);
	days_eptoyr = delta_days(sats[sat].epoch_yr, t->year);

	/* Adjust outdated elements to current year */

        if(sats[sat].epoch_yr < t->year) {
                sats[sat].epoch_time -= days_eptoyr;
                sats[sat].epoch_yr = t->year;
        }

	i = sats[sat].inclin*PI/180;

	/* Adjust raan, and argument of perigee for precession */

	prec_adj(t, &w,  &o, sat);

	/* Set up a rotation matrix */

	init_matrix(c, w, o, i);

	/* Find the current orbital phase and mean anomaly */ 

        q = sats[sat].mn_motion * (t->day - sats[sat].epoch_time) + 
                sats[sat].mn_anom/360;
        ma = modf((double)q, &ip) * (2*PI);
        
	/* Find Phase units - Useful for transponder schedules */

        sats[sat].phase = ma * 256 / (2*PI);

        e = sats[sat].ecc_anom;

	e = solve_kep(e, ma, sat);

	/* Store result for trial value of next iteration */

        sats[sat].ecc_anom = e;

	/* Find coordinates in the plane of the orbit */

        x = sats[sat].sm_axis*(cos(e) - sats[sat].eccen);
        y = sats[sat].sm_axis*sqrt(1 - pow(sats[sat].eccen, 2.0))*sin(e);

	/* Rotate to Celestial Inertial Reference Frame */

        x1 = x*c[0][0] + y*c[0][1];
        y1 = x*c[1][0] + y*c[1][1];
        z1 = x*c[2][0] + y*c[2][1];

	/* Rotate to Geocentric coordinates */ 

        g = 2*PI*modf(t->day*SID_SOL + (days_89toyr*SID_SOL+0.27676779), &ip);
        x = x1*cos(g) - y1*(-sin(g));
        y = x1*(-sin(g)) + y1*cos(g);
        z = z1;

	/* Find vector from Observer to Sat */

        x1 = x - obs->x;
        y1 = y - obs->y;
        z1 = z - obs->z;

	/* Find Distance to Sat */

        sats[sat].range = sqrt(pow(x1, 2.0) + pow(y1, 2.0) + pow(z1, 2.0));

        lat = obs->lat;
        ln = obs->ln;

	/* Find Local Horizon coordinates */
                
        x = -x1*cos(ln)*sin(lat) - y1*sin(-ln)*sin(lat) + z1*cos(lat);
        y = y1*cos(ln) - x1*sin(-ln);
        z = x1*cos(ln)*cos(lat) + y1*sin(-ln)*cos(lat) + z1*sin(lat);

        s = z/sats[sat].range;
        c2 = sqrt(1-pow(s, 2.0));

	/* Find elevation above horizon */

        sats[sat].alt = atan(s/c2) * 180/PI;

	/* Find azimuth in correct quadrant */

	find_azi(x, y, sat);

}

find_azi(x, y, sat)
float x, y;
int sat;
{
	/* Finds the azimuth in the correct quadrant */

	float d;	/* intermediary */

        if(x < 0)
                d = PI + atan(y/x);
        if(x > 0) {
                if(y >= 0)
                        d = atan(y/x);
                else
                        d = 2*PI + atan(y/x);
        }
        if(x == 0) {
                if(y >= 0)
                        d = PI/2;
                else
                        d = 3*PI/2;
        }

        sats[sat].azi = d*180/PI;
}

get_inc(inc)
float *inc;
{
	/* Self explanatory */

        float i;
        char buffer[5];		/* input buffer */

        printf("\nEnter time increment in minutes.\n");

        gets(buffer);

        i = atof(buffer);

        *inc = i /60.0 / 24.0;
}

get_sat(numsats)
int numsats;
{
	/* Self explanatory */ 

        int sat_of_interest = 0;
        char buffer[10];		/* input buffer */

        printf("\nOscars Loaded:\n\n");

        printsats(numsats);

        printf("\nEnter number of Satellite you are interested in.\n");

        while(sat_of_interest == 0 || sat_of_interest > numsats)        
                sat_of_interest = atoi(gets(buffer));

        return sat_of_interest;
}
        
load_obs(site)
struct pt *site;
{
	/* Loads the Observer's location */

        FILE *datafile;
        char line[10];			/* input buffer */
        float lat, ln, alt, r, geolat;
	/* alt = altitude, r = radius, geolat = intermediate */

		  datafile = fopen("location.dat", "r");

        printf("Loading Observer Location.\n");

        if(!feof(datafile))
                lat = atof(fgets(line, 10, datafile))*PI/180;
        if(!feof(datafile))
                ln = atof(fgets(line, 10, datafile))*PI/180;
        if(!feof(datafile))
                alt = atof(fgets(line, 10, datafile))/1000;

	/* Correct for oblateness of earth */

        r = R_earth *(1 - (dR/2) + (dR/2)*(cos(2*lat))) + alt;
        geolat = atan(pow((1-dR),2.0)*sin(lat)/cos(lat));

	/* Geocentric coodinates */

        site->x = r*cos(geolat)*cos(ln);
        site->y = r*cos(geolat)*sin(-ln);
        site->z = r*sin(geolat);
        site->lat = lat;
        site->ln = ln;

        fclose(datafile);
}

load_sats(numsats)
int *numsats;
{
	/* Loads the satellite elements, and makes some initial calculations */

        extern struct satellite sats[];
        int i = 0;
        int n;				/* counters */
	double ip, dummy;		/* for modf */
        FILE *datafile;

        char line_name[72];
        char line_one[72];		/* input buffers */
        char line_two[72];

		  datafile = fopen("elements.dat", "r");

        printf("Loading Satellite Keps.\n");
        
        while(!feof(datafile)) {
                i += 1;

		/* Loads lines and shuffles until in correct order */
                
                if(!feof(datafile))
                        fgets(line_name, 71, datafile);
                if(!feof(datafile))
                        fgets(line_one, 71, datafile);
                if(!feof(datafile))
                        fgets(line_two, 71, datafile);

                while((line_one[0] != '1') || (line_two[0] != '2')) {
                        strcpy(line_name, line_one);
                        strcpy(line_one, line_two);
                        if(feof(datafile))
                                break;
                        else
                                fgets(line_two, 71, datafile);
                }

	/* Erase unused data at beginning of lines */

        for(n=0; n < 18; n++)
                line_one[n] = ' ';
        for(n=0; n < 8; n++)
                line_two[n] = ' ';

        sscanf(line_name, "%14s", sats[i].name);
        sscanf(line_one, "%2d%12f", &sats[i].epoch_yr, &sats[i].epoch_time);
        sscanf(line_two, "%f %f %f %f %f %11f%f", &sats[i].inclin,
                &sats[i].raan, &sats[i].eccen, &sats[i].arg_peri,
                &sats[i].mn_anom, &sats[i].mn_motion, &sats[i].rev_count);

	/* Set other elements, and calculate semi-major axis */

        sats[i].eccen *= 10e-8;
        sats[i].ecc_anom = 0;
        sats[i].sm_axis = pow((GM/pow(sats[i].mn_motion, 2.0)), 0.33333333333);
        sats[i].range = 0;
        sats[i].alt = 0;
        sats[i].azi = 0;
        sats[i].phase = 0;

	/* Separate Checksum from revolution count and discard */

        dummy = modf((double)sats[i].rev_count/10, &ip);

        sats[i].rev_count = ip;

	printf(". ");
        }
  
        *numsats = i-1;

        fclose(datafile);
}

printsats(n)
int n;
{
	/* Prints out a list of the sats in memory */

        extern struct satellite sats[];
        int i = 0;
        
        for(i = 1;i <= n; i++) {
                printf("%2d\t%10s\n", i, sats[i].name);
        }
}
