/*
   scan.c : Copyright Paul Healy, EI9GL, 1992.

   911227 : Added this header
   920620 : Version 0.1
            - doesn't remember current folder
            - doesn't support listing of a subrange
   920706 : Added check to see if specified folder exists
   920717 : Added ax25 pbbs R: line disposal
   920721 : Fixed bug in listing short messages ("\nF" was not being found)
   920819 : Catered for variation in the Date: header contents
   920830 : Made Date: header processing more rfc822 robust (lots more todo)
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "pc.h"
#include "buf.h"
#include "mailfile.h"
#include "header.h"
#include "rc.h"
#include "misc.h"
#include "help.h"
#include "pager.h"
#include "mailer.h"
#include "buf.h"
#include "mailfile.h"
#include "message.h"

#ifdef BMH
#define main scan_main
#endif

#undef DEBUG

typedef struct {
   char *s;
   int width;
   } tbuf;

static void
tbuf_init(tbuf *tp, char *p, int w)
{
   tp->s = p;
   tp->width = w;
}

static int
tbuf_add_char(tbuf *tp, char ch)
{
   if ((ch == '\n') || (ch == '\r') || (ch == '\t') )
      ch = ' ';
   if (isspace(ch) && isspace(*(tp->s - 1)))
      return 0;

   if (tp->width>0) {
      tp->width--;
      *(tp->s++) = ch;
      *tp->s = 0;
      return 0;
      }
   else
      return -1;
}

static int
tbuf_add_str(tbuf *tp, char *s)
{
   if (tp->width>0) {
      int len = strlen(s);
      int min = tp->width > len ? len : tp->width;
      strncpy(tp->s, s, min);
      tp->width-=min;
      tp->s+=min;
      *(tp->s) = 0;
      return 0;
      }
   else
      return -1;
}

#ifdef AX25BBS
static int
tbuf_del_char(tbuf *tp)
{
   tp->width++;
   *(--tp->s) = 0;

   return 0;
}
#endif

/* -------------------------------------------------------------------- */

static const int EightyColFigure = 48;
static int TopicWidth = 48;

#ifdef notdef
      if ( (lastch=='\n') && (ch=='R') && (mail_curc(m)==':') ) { /* \nR:xxxx */
#endif

static char *
topic(Mailfile *m, char *subject, char *line)
{
   int ch = 0, lastch = 0;
   int width = TopicWidth - 1;
   tbuf tb;

   tbuf_init(&tb, line, width);

   if (tbuf_add_str(&tb, subject) == -1)
      return line;
   else if (tbuf_add_str(&tb, "<<") == -1)
      return line;

   do {
      /*
       * Look for 'From' lines
       */
#ifdef DEBUG
      fprintf(stderr, "scan: %x %x %c %c <%s>\n",
         ch, mail_curc(m), ch, mail_curc(m), line);
#endif
      if ( (mail_curc(m)==EOF) || ((mail_curc(m)=='F') && (ch=='\n')) )
         break;
      else {
         lastch = ch;
         ch = mail_getc(m);
         }

#ifdef AX25BBS
      /*
       * 920717 : Dispose of ax25 pbbs R: lines
       */
      if ( (lastch=='R') && (ch==':') && (mail_curc(m)=='9') ) { /* \nR:9xxx */
         mail_skipline(m);
         ch = '\n';
         tbuf_del_char(&tb);
         continue;
         }
#endif
      } while (tbuf_add_char(&tb, ch) != -1);
      
   tbuf_add_str(&tb, ">>");

   return line;
}

static void
date2mh(char *date, int *day, int *month)
{
   char *cp, mon[4];
   static char *p = NULL;

   *day = 0;
   *month = 0;

   if ((date == NULL) || (strlen(date) < 14))
      return ;

   decomment(date);

   /*
    * Skip over 'Fri, ', 'Friday, ' etc
    *
    * date = 1*2DIGIT month 2*4DIGIT
    *
    * Date: Sun, 06 Jan 1991 17:30:20 UTC
    * Date: 6 Jan 1991 17:30:20 UTC
    */
   for (cp = date; *cp && !isdigit(*cp); cp++)
      ;
   *day = atoi(cp);

   while (*cp && !isalpha(*cp))
      cp++;

   strncpy(mon, cp, 3);
   mon[3] = 0;
   *month = asc2month(mon);

   if (p == NULL)
      p = getrc(dayfirst);

   if (*p == 'n') {
      int t = *day;
      *day = *month;
      *month = t;
      }
}

/* ------------------------------------------------------------------ */
static int
topicwidth(void)
{
   char *p = getrc(maxcol);
   int w;

   if ((p == NULL) || ( (w=atoi(p)) <1) )
      return EightyColFigure;
   else
      return w - 32;
}

static int curmsg = 0;

static MsgResult
dumpmsg(Mailfile *m, char *line)
{
   Header_t hdr;

#ifdef DEBUG
   fflush(stdout);
   fprintf(stderr, "scan: msg %d <%s>\n", mail_curmsg(m), line);
#endif

   if (gethdrb(m, &hdr)  == -1) {
      fflush(stdout);
      fprintf(stderr, "scan: msg %d <%s>\n", mail_curmsg(m), line);
      line[0] = '\0';
      return MSG_NEXT;         /* move to next message */
      }
   else {
      int day = 0, month = 0;
      char *date = HEADER(hdr, DATE);
      char *from = HEADER(hdr, FROM);
      char *subj = HEADER(hdr, SUBJECT);

      if ( date != NULL)
         date2mh(date, &day, &month);

      /*
       * page_printf returns:
       *  0 : => line was put ok, move to the next message
       * -1 : => quit, finish up
       */
      if (page_printf("%4d%c %2d/%02d %-18.18s %s",
         mail_curmsg(m),
         mail_curmsg(m) == curmsg ? '+' : ' ',
         day, month,
         from == NULL ? "" : name(from),
         topic(m, subj == NULL ? "" : subj, line )) == 0
         )
         return MSG_NEXT;
      else
         return MSG_QUIT;
      }
}

/*
          1         2         3         4
01234567890123456789012345678901234567890123456789012345678901234567890123456789
+   6 21/01 Gerard Lawlor      cork<<Hi there wonder if there was owt of intere
CDDDD DD/DD SSSSSSSSSSSSSSSSSS xxx...
                                         1         2         3         4
                               012345678901234567890123456789012345678901234567
*/

int
scan(Mailfile *m, int argc, char *argv[])
{
   page_setup("scan");

   if (mail_getfile(m, argc, argv, "scan") == -1)
      return -1;

   curmsg = mail_curmsg(m);

   TopicWidth = topicwidth();

   if ((argc == 1) || (argv[1][0] == '+') )    
      walkall(m, dumpmsg);
   else 
      walkselect(m, dumpmsg, argc, argv);

   page_fin("scan");

   return 0;
}

int
main(int argc, char *argv[])
{
   Mailfile m;

   if (dohelp(argc, argv, "scan [+folder] [msglist] (list mail)"))
      return 0;

   mail_zero(&m);

   scan(&m, argc, argv);

   return mail_savestate(&m);
}
