/*
   send.c : Taken from bm.

   Copyright 1986 Bdale Garbee, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.
   Copyright 1987 1988 Dave Trulli NN2Z, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.

   920801 : Added dosend.
   920830 : added resent headers
   920904 : Added a little tz intelligence.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include "config.h"
#include "rc.h"
#include "pc.h"
#include "lock.h"
#include "misc.h"
#include "buf.h"
#include "mailfile.h"
#include "send.h"
#include "header.h"
#include "mailer.h"

static struct addr *addrecip(struct addr **head, char *user, char *host);
static struct addr *expandalias(struct addr **head, char *user);
static int recordmsg(FILE *dfile);

/*
 * send a message
 * return -1 on error, 0 otherwise
 * calls dosend
 */
int
dosend(FILE *fp, Header_t *h)
{                                   
   char tf[MAXNAMLEN];
   FILE *tfile;
   struct addr *tolist = NULL;
   struct addr *cclist = NULL;
   int argc;
   char *argv[MAXARGS];
   char *to = HEADERP(h, TO);
   char *cc = HEADERP(h, CC);

   if (to != NULL)      /* hack */
      to = strdup(to);
   if (cc != NULL)
      cc = strdup(cc);

   /*
    * todo: filter out comments here ()
    */
   if ( (to == NULL) ||
        ((argc = parse(to, argv, MAXARGS)) < 1) ||
        ((tolist = make_tolist(argc, argv)) == NULL)
      ) {
      fprintf(stderr, "bmh: no recpients, send aborted\n");
      return -1;
      }

   if ( (cc != NULL) && ((argc = parse(cc, argv, MAXARGS)) > 0))
      cclist = make_tolist(argc, argv);

   if ((tfile = tempfile("bmh", tf, "w+")) == NULL) {
      del_addrlist(tolist);
      return -1;
      }

   if (putrfc822(tfile, h, tolist, cclist) == -1L) {
      fprintf(stderr, "bmh: error writing mail headers\n");
      fclose(tfile);
      unlink(tf);
      return -1;
      }

   if (dupfile(fp, tfile) == -1) {
      fprintf(stderr, "send: error writing message\n");
      return -1;
      }

   queuejob(tfile, tolist, 0L);
   del_addrlist(tolist);

   if (cclist != NULL) {
      queuejob(tfile, cclist, 0L);
      del_addrlist(cclist);
      }

   if (to != NULL)      /* hack */
      free(to);
   if (cc != NULL)
      free(cc);

   if (recordmsg(tfile) == -1)          /* save copy for sender */
      fprintf(stderr, "bmh: error while making sender copy\n");

   fclose(tfile);
   unlink(tf);

   return 0;
}

/*
 * save copy in the record file
 */
static int
recordmsg(FILE *dfile) 
{
   FILE *fp;
   int ret = -1;

   if (getrc(record) == NULL)
      return 0;

   if ((fp = fopen(getrc(record), "a")) == NULL)
      fprintf(stderr, "bmh: error opening %s\n", getrc(record));
   else {
      time_t t = time(NULL);

      fprintf(fp, "From %s@%s %s", getrc(username), getrc(hostname), ctime(&t));
      fseek(dfile, 0L, SEEK_SET);

      if (dupfile(dfile, fp) != -1)
         ret = 0;

      fclose(fp);
      }
      
   return ret;
}

/*
 * place a mail job in the outbound queue
 */
int
queuejob(FILE *dfile, struct addr *tolist, long position)
{
   FILE *fp;
   char tmpstring[255], prefix[255];
   struct addr *tp, *sp;

   for (tp = tolist; tp != NULL; tp = tp->next) {
      long id;

      if (tp->sent)
         continue;

      if ( (id = get_msgid(getrc(mqueue))) == -1L)
         return -1;

      sprintf(prefix, "%s/%ld", getrc(mqueue), id);
      (void) lockit(prefix);

      sprintf(tmpstring,"%s/%ld%s", getrc(mqueue), id, EXT);
      if ( (fp = fopen(tmpstring, "w")) == NULL) {
         fprintf(stderr, "bmh: unable to open %s\n", tmpstring);
         bm_unlock(prefix);
         return -1;
         }

      fseek(dfile, position, SEEK_SET);
      if (dupfile(dfile, fp) == -1) {
         fclose(fp);
         bm_unlock(prefix);
         return -1;
         }
      fclose(fp);

      sprintf(tmpstring, "%s/%ld.wrk", getrc(mqueue), id);
      if((fp = fopen(tmpstring, "w")) == NULL) {
         (void) bm_unlock(prefix);
         return -1;
         }

      fprintf(fp, "%s\n%s@%s\n", tp->host, getrc(username), getrc(hostname));
      fprintf(fp,"%s@%s\n", tp->user, tp->host);
      tp->sent++;

      /*
       * find and add other addresses to the same host
       */
      for (sp = tp->next; sp != NULL; sp = sp->next)
         if (!sp->sent && (strcmp(tp->host, sp->host) == 0)) {
            fprintf(fp,"%s@%s\n", sp->user, sp->host);
            sp->sent++;
            }

      fclose(fp);
      bm_unlock(prefix);
      }

   return 0;
}

#define SKIPWORD(X) while(*X && *X!=' ' && *X!='\t' && *X!='\n' && *X!= ',') X++;
#define SKIPSPACE(X) while(*X ==' ' || *X =='\t' || *X =='\n' || *X == ',') X++;

/*
 * check for and alias and expand alias into a address list
 */
static struct addr *
expandalias(struct addr **head, char *user)
{
   FILE *fp;
   char *s, *p, *h;
   int inalias;
   struct addr *tp;
   char buf[LINELEN];
   
   if ( (fp = fopen(getrc(alias), "r")) == NULL)
      return addrecip(head, user, getrc(hostname) );

   inalias = 0;
   while (fgets(buf, LINELEN, fp) != NULL) {
      p = buf;
      if ( *p == '#' || *p == '\0')
         continue;
      rip(p);

      /* if not in an matching entry skip continuation lines */
      if (!inalias && isspace(*p))
         continue;

      /* when processing an active alias check for a continuation */
      if (inalias) {
         if (!isspace(*p)) 
            break;   /* done */
      } else {
         s = p;
         SKIPWORD(p);
         *p++ = '\0';   /* end the alias name */
         if (strcmp(s,user) != 0)
            continue;   /* no match go on */
         inalias = 1;
      }

      /* process the recipients on the alias line */
      SKIPSPACE(p);
      while(*p != '\0' && *p != '#') {
         s = p;
         SKIPWORD(p);
         if (*p != '\0')
            *p++ = '\0';
         /* find hostname */
         if ((h = strchr(s,'@')) != NULL)
            *h++ = '\0';
         else
            h = getrc(hostname);
         tp = addrecip(head,s,h);
         SKIPSPACE(p);
      }
   }
   (void) fclose(fp);

   if (inalias)   /* found and processed and alias. */
      return tp;

   /* no alias found treat as a local address */
   return addrecip(head, user, getrc(hostname));
}

/*
 * convert a arg list to an list of address structures
 */
struct addr *
make_tolist(int argc, char *argv[])
{
   struct addr *tolist = NULL;
   int i;

   for (i = 0; i<argc; i++) {
      char *host = strchr(argv[i], '@');
      struct addr *tp;

      if (host != NULL) {
         *host++ = '\0';
         if (stricmp(host, getrc(hostname)) == 0)
            host = NULL;
         }
                            
      tp = (host==NULL) ? expandalias(&tolist, argv[i]) :   /* a local address */        
                          addrecip(&tolist, argv[i], host); /* a remote address */

      if (tp == NULL) {
         fprintf(stderr, "bmh: no room for To: list\n");
         del_addrlist(tolist);
         return NULL;
         }
      }

   return tolist;
}

/*
 * delete a list of mail addresses
 */
void
del_addrlist(struct addr *list)
{
   struct addr *tp, *tp1;

   for (tp = list; tp != NULL; tp = tp1) {
      tp1 = tp->next;
      if (tp->user != NULL);
         free(tp->user);
      if (tp->host != NULL);
         free(tp->host);
      free(tp);
   }
}

/*
 * add an address to the front of the list pointed to by head 
 * return NULL if out of memory.
 */
static struct addr *
addrecip(struct addr **head, char *user, char *host)
{
   struct addr *tp = (struct addr *) malloc(sizeof(struct addr));

   if (tp == NULL)
      return NULL;

   /* allocate storage for the user's login */
   if ( (tp->user = strdup(user)) == NULL) {
      free(tp);
      return NULL;
      }

   /* allocate storage for the host name */
   if (host == NULL)
      tp->host = NULL;
   else if ( (tp->host = strdup(host)) == NULL) {
      free(tp->user);
      free(tp);
      return NULL;
      }
   /*
    * add entry to front of list
    */
   tp->next = *head;             /* ok if *head is NULL */
   *head = tp;
   tp->sent = 0;

   return tp;
}

