/* server.c */
/* PA0GRI's server between WA7MBL's mailbox and KA9Q's TCP/IP NETwork */
/*
Input files to server are:
        SERVER.TXT in current directory, containing forwarded msgs via BBS. 
        SERVER.TXT in /SPOOL/MAIL directory containing messages to BBS.
        
Output files from server are:
        SERVER.INP in current directory for messages to BBS.
        .TXT & .WRK files in \SPOOL\MQUEUE directory for messages to NET.

Exit codes: 0 = normal,
            1 = Unable to write SERVER.INP,
            2 = no /BM.RC,

modification history:
        1.2        Tom Leitner (rzmul2@tuvie.uucp)
                   ported to OS-9, use env. variable TCPIP_SPOOL to locate
                   spool files
        1.1        Bdale        Modify to comply with new standards for interface to
                        mail spool directories.
        1.0        PA0GRI        Original release, ala 870526.0 formats.

*/
#include <stdio.h>
#include <ctype.h>
#include <time.h>

#define        LINELEN                128
#define SLINELEN        32
#define MAILQDIR makefname("TCPIP_SPOOL", "/mqueue/")
#define SERVER   "server.txt"

#include "bm.h"
#include "server.h"

extern char serv_ver[];
static char copyright[] = 
"Copyright 1987 Gerard VanDerGrinten, permission granted for non-commercial use.";

FILE        *rcfp;                /* handle for the configuration file */
FILE        *mfile;
FILE        *infile,
                *outfile;
FILE        *tfile,
                *wfile,
                *sfile,
                *lfile;
char        rcline[LINELEN],        /* buffer for config file reading routines */
                hostname[SLINELEN],        /* name of this host from rc file */
                username[SLINELEN],        /* name of this user from rc file */
                gateway[SLINELEN],         /* ip addr of smart mail gateway */
                time_zone[4],              /* users time zone */
                maildir[LINELEN],          /* defined mail directory */
                editor[SLINELEN];          /* user's favorite text editor */

char		*makefname();
static        char
                smtp_to[LINELEN], 
                smtp_subject[LINELEN], 
                tohost[LINELEN],
                thishost[LINELEN],
                tfilename[LINELEN], 
                wfilename[LINELEN],
                sfilename[LINELEN],
                lfilename[LINELEN],
                tstring[LINELEN],                /* temp for integer conv */
                ptimestr[40],
                toipaddr[16];                    /* destination ip addr str */

char        *tch,
                *hostn,
                *tptr;
                
char        callsign[7],
                dest[7],
                line[LINELEN];

int                state = OTHER;

char        area[LINELEN],
                mailbox[LINELEN],
                tmpstring[LINELEN];

main(argc,argv)
int argc;
char *argv[];
{
        int i,j,k;
        char ch,cc,*p;
        char str1[LINELEN], str2[LINELEN];
        char        notename[9];        /* name of notesfile */
        char        notefile[80];        /* temorary for building strings */
        char        *notel, *tch, *index();
        int        notes, msgnum;
        /* announce ourselves */
        if(argc > 1)
                hostn = argv[1];
        else
                hostn = "PA0GRI";
                
        printf("\t\t%s\n%s\n\n",serv_ver,copyright);

        /* initialize globals from config file */
        *editor = 0x00;  /* default no editor? */
        *maildir = 0x00;
        strcpy(time_zone,"MET");
        read_rc();

        if(*maildir == '\0')
                strcpy(maildir, makefname("TCPIP_SPOOL", "/mail"));

        outfile = NULL;  /* preset */
        if((infile = fopen(SERVER,"r")) == NULL) {
                printf(" Cannot open SERVER file \n");
                goto outit;
        }
        
/* we have an input file, so make the output. */
        if((outfile = fopen("server.inp","w+")) == NULL) {
                fclose(infile);                /* close our input file */
                fclose(lfile);                /* close the lock file */
                unlink(lfilename);        /* erase the lock file */
                exit (1);
        }
/*  Start reading the messages in the server file and try to forward them. */
        while(1){
                fgets(line,LINELEN,infile);
                if(feof(infile)) 
                        break; /* end of server part 1 */
/* extract FROM: */
                strncpy(callsign,line,6);
/* make it lower case for NET router */
                for(i = 0 ; i < 7 ; i++)
                        dest[i] = tolower(callsign[i]);
                dest[6] = 0x00;
/* get the supposed destination into line */
                fgets(line,LINELEN,infile);
                rip(line);
                printf(" forwarding from %s to %s\n",dest,line);
/* move the message to AMPRnet or back to PBBS if host unknown */
/*** this will never bounce a message, because we no longer do the
     hostname lookup here... we let net.exe do the translation to an
     IP address....  as of v1.1, Bdale ***/
                if( dosmtpsend() != 0)
                         retsvrmsg();
        }

/* remove lock in mqueue directory */
        fclose(lfile);                /* close the lock file */
        unlink(lfilename);        /* erase the lock file */

outit:
        if (infile != NULL)
                fclose(infile);
        sprintf(line,"%s/%s",maildir,"server.txt");
#ifdef BSLASH
        backslash(line);
#endif BSLASH
        printf(" Opening %s for AMPRnet messages.\n",line);
        if((infile = fopen(line,"r")) != NULL) {
                if (outfile == NULL)
                        if((outfile = fopen("server.inp","w+")) == NULL)
                                exit (1);
                netmsg();
                }        
        printf(" Done processing mail from NET to Mailbox\n");
        if (outfile != NULL)
                fclose(outfile);
        if (infile != NULL) {
                fclose(infile);
                unlink(line);
        }
        exit (0);
}

/* rip      finds cr's or lf's from a string and ends it at that position */
rip(s)
char *s;
{
        char *c;
        c = s;
        while(*c != '\0') {
                switch( *c) {
/*                case '\r': */
                case '\n':
                        *c = '\0';
                        break;
                default:
                        c++;
                        break;
                }
        }
}

/* retsvrmsg  copys the current message to the server.inp file with
              some suggested error messages around it */
retsvrmsg()
{

        fprintf(outfile,"SP %s < SERVER\n",callsign);
        fprintf(outfile,"Unable to deliver mail\n");
        fprintf(outfile,"Unable to deliver following message trough SERVER:\n");
        fprintf(outfile,"Apparently To:      %s\n",line);
        fgets(line,LINELEN,infile);
        rip(line);
        if(strncmp(line,"/EX",3) == 0) {
                fprintf(outfile,"--- No message body found\n/EX\n");
                return(1);
        }
        fprintf(outfile,"Apparently Subject: %s\n \n",line);
        fprintf(outfile," ----- Start of text message body ------ \n");
        while(1) {
                fgets(line,LINELEN,infile);
                if(feof(infile)){
                        return (1);
                }
                rip(line);
                if(strncmp(line,"/EX",3) == 0) {
                        fprintf(outfile," ----- End of text message body ------ \n/EX\n");
                        break;
                }
                fprintf(outfile,"%s\n",line);
        }
        return (0);
}

read_rc()
{
        if((rcfp = fopen(makefname("TCPIP_STARTUP", "/bm.rc"),"r")) == NULL) {/* open config file */
                printf("Cannot open BM.RC in root, check your installation\n");
                exit(2);
        } else
            while (!feof(rcfp)) {
                fgets(rcline,LINELEN,rcfp);
                if (rcline[0] == ';') continue;
                if (strlen(rcline) < 5) continue;
                switch (rc_line_type(rcline)) {
                        case HOST:
                                strcpy(hostname,&rcline[5]);
                                rip(hostname);
                                break;
                        case USER:
                                strcpy(username,&rcline[5]);
                                rip(username);
                                break;
                        case GATE:
                                strcpy(gateway,&rcline[5]);
                                rip(gateway);
                                break;
                        case EDIT:
                                strcpy(editor,&rcline[5]);
                                rip(editor);
                                break;
                        case ZONE:
                                strncpy(time_zone,&rcline[5],3);
                                time_zone[3] = '\0';
                                break;
                        case SMTP: 
                                strcpy(maildir,&rcline[5]);
                                rip(maildir);
                                break;
                        case FIDO:
                        default:
                                /* handle these cases in v2.0! */
                                break;
                }        
    }
        fclose(rcfp);
}

/* return the line_type from a line of the configuration file */
rc_line_type(s)
char *s;
{
        if (strncmp("fido",s,4) == 0) return (FIDO);
        if (strncmp("smtp",s,4) == 0) return (SMTP);
        if (strncmp("host",s,4) == 0) return (HOST);
        if (strncmp("user",s,4) == 0) return (USER);
        if (strncmp("gate",s,4) == 0) return (GATE);
        if (strncmp("edit",s,4) == 0) return (EDIT);
        if (strncmp("zone",s,4) == 0) return (ZONE);
        return (NONE);
}

/* send a message to AMPRnet.  Always returns 0.
 */
int
dosmtpsend()
{
        char        *index(),*rindex();
        int        sequence;

/* open the sequence file */
        strcpy(sfilename,MAILQDIR);
        strcat(sfilename,"sequence.seq");
#ifdef BSLASH
        backslash(sfilename);
#endif BSLASH
        sfile = fopen(sfilename,"r");

/* if sequence file exists, get the value, otherwise fake it */
        if (sfile == NULL) 
                strcpy(tstring,"0");
        else {
                fgets(tstring,LINELEN,sfile);
                fclose(sfile);
        }
        sequence = atoi(tstring);

/* increment sequence number, and write to sequence file */
        sfile = fopen(sfilename,"w+");
        fprintf(sfile,"%d",++sequence);

/* close the sequence file */
        fclose(sfile);

        ptime(ptimestr);

        printf("Processing mail... ");
/* copy the To: line */
        strcpy(smtp_to,line);

/*  get Subject: */
        fgets(smtp_subject,80,infile);
        rip(smtp_subject);
    smtp_subject[79] = '\0';
    
/* create a lock in the mqueue directory */
sprintf(lfilename,"%s%d%s",MAILQDIR,sequence,".lck");
if ((lfile = fopen(lfilename,"w+")) == NULL) {
  printf(" Cannot create lock file\r\n");
  return;
}
fclose(lfile);

    /* open textfile in mqueue directory */
        sprintf(tfilename,"%s%d%s",MAILQDIR,sequence,".txt");
#ifdef BSLASH
        backslash(tfilename);
#endif BSLASH
        tfile = fopen(tfilename,"w+");                /* open textfile for write */
/* write (almost) RFC822-compatible headers using above information */
        fprintf(tfile,"Date:     %s",ptimestr);
        fprintf(tfile,"From:     SERVER@%s\n",hostn);
        fprintf(tfile,"To:       %s\n",smtp_to);
        fprintf(tfile,"Reply-To: %s@%s%%%s\n","SERVER",hostn,dest);
        fprintf(tfile,"Subject:  %s\n",smtp_subject);
        fprintf(tfile,"Message-Id: <%u@%s>\n",sequence,hostname);
        fprintf(tfile,"\n");                    /* add empty line as separator */

        if(strncmp(smtp_subject,"/EX",3) != 0)
                   while(1) {
                        fgets(tstring,LINELEN,infile);
                                if(feof(infile))
                                        break;
                        rip(tstring);
                        if(strncmp(tstring,"/EX",3) == 0)
                                break; /* end of message found */
                        fprintf(tfile,"%s\n",tstring);
                   }
           fclose(tfile);

/* open a workfile in the mqueue directory */
        sprintf(wfilename,"%s%d%s",MAILQDIR,sequence,".wrk");
#ifdef BSLASH
        backslash(wfilename);
#endif BSLASH
        wfile = fopen(wfilename,"w+");                /* open textfile for write */
        fprintf(wfile,"%s\n%s\n%s@%s\n",toipaddr,smtp_to,"SERVER",hostname);
        fclose(wfile);

/* remove the lockfile */
        unlink(lfilename);

        printf("messageid=%u  Done.\n",sequence);
        return 0;
}

/* Return Date/Time in Arpanet format in passed string */
ptime(str)
char *str;
{
        /* Print out the time and date field as
         *                "DAY day MONTH year hh:mm:ss ZONE\n"
         */
        long t;
        struct tm *ltm;
        static char *days[7] = {
                "Sun","Mon","Tue","Wed","Thu","Fri","Sat" };

        static char *months[12] = {
                "Jan","Feb","Mar","Apr","May","Jun",
                "Jul","Aug","Sep","Oct","Nov","Dec" };

        /* Read the system time */
        time(&t);
        ltm = gmtime(&t);

        /* most of the inportent things are decoded , encript into string */
        sprintf(str,"%s, %02d %s 19%02d %02d:%02d:%02d %s\n"
                ,days[ltm->tm_wday],ltm->tm_mday,months[ltm->tm_mon],
                 ltm->tm_year,ltm->tm_hour,ltm->tm_min,ltm->tm_sec,
                 time_zone);
}

/* replace all '/' with '\' in the passed string. */
backslash(s)
char *s;
{
  while (*s != '\0') {
    if (*s == '/')       /* replace slash with backslash */
      *s = '\\';   
    s++;
  }
}
/* netmsg
 *
 *        Program to receive mail that has arrived on PC via KA9Q SMTP
 *        server, and make it available as a WA7MBL message base. 
 *
 *        Copyright 1987 Bdale Garbee N3EUA 
 *        Permission granted for non-commercial distribution and use, so
 *        long as this notice is retained.
 *
 *        Exit Conditions:        none
 */

netmsg()
{
        char from[36], to[36], subj[72];
        char ptimestr[40], *p, *q;
        int i;
        *to = '\0';   /* mark no messages written */
        
        /* process mailbox, writing message files as appropriate */
        do {
            fgets(tmpstring,LINELEN,infile);
#ifdef DEBUG
           printf("\nstate = %u; linetype = %u",state,type(tmpstring));
#endif
           switch (type(tmpstring)) {

            case LT_RCVD:
                        switch (state) {
                            case OTHER:
                            case RCVD:
                                        state = RCVD;
                                    break;
                            case NHDR:
                                        break;  /* ignore received */
                                    case NBODY:
                                        fprintf(outfile,"/EX\n");
                                        *to = '\0';
                                        state = RCVD;
                                          break;
                        }
                break;
                
                   case LT_OTHER:
                        switch (state) {
                            case OTHER:
                                        break;
                            case RCVD:
                                        *to   = '\0';
                                        *from = '\0';
                                        *subj = '\0';
                                    state = NHDR;
                /* fall trough */
                            case NHDR:
                       /* parse for To:, From: and Subject: lines */
                                        if (strncmp(tmpstring, "To:", 3) == 0) {
                                                strncpy(to,tmpstring+10,36);
                                                rip(to);
                                                to[35] = '\0';
                                        }
                                        if (strncmp(tmpstring, "From:", 5) == 0) {
                                                strncpy(from,tmpstring+10,36);
                                                rip(from);
                                                from[35] = '\0';
                                        }
                                        if (strncmp(tmpstring, "Subject:", 8) == 0) {
                                                strncpy(subj,tmpstring+10,72);
                                                rip(subj);
                                                subj[71] = '\0';
                                        }
                                        break;
                                    case NBODY:
                                        rip(tmpstring);
                                        fprintf(outfile,"%s\n",tmpstring);
                                          break;
                    }
                break;
                    
            case LT_BLANK:
                 switch (state) {
                                case OTHER:
                                case RCVD:
                                        break;
                                case NHDR:
/* end of header, now build PBBS header */
                                        if((p = rindex(to,'%')) != NULL) {
                                                q = ++p;
                                                for (i = 0; i < 7 ; i++)
                                                        if( isalnum(q[i]) )
                                                                callsign[i] = toupper(q[i]);
                                                        else
                                                                callsign[i] = '\0';
                                                callsign[6] = '\0'; 
                                                fprintf(outfile,"SP %s < SERVER\n",callsign);
                                        } else
                                                fprintf(outfile,"SB ALL < SERVER\n");

                                        fprintf(outfile,"%s\n",subj);
                                        ptime(ptimestr);
                                        fprintf(outfile
                                                ,"Received: By %s via AMPRnet for %s; %s"
                                                ,hostn,to,ptimestr);
                                        fprintf(outfile,"From:     %s\n",from);
                                        state = NBODY;
                                        break; /* ignore first blank line between header and msg */
                                case NBODY:
                                        rip(tmpstring);
                                        fprintf(outfile,"%s\n",tmpstring);
                                        break;
                        }
                break;

                }
        } while (!feof(infile));

        if (*to != '\0')
                fprintf(outfile,"/EX\n");
        printf("\nNextmsg Done.\n");
        return (0);
}


/* return the type of the passed line */
type(p)
register char *p;
{
#ifdef DEBUG
        char *i;

        i = p;
        printf("\ntest: ");
        while (*i != '\0') printf("%x ",*i++);
#endif
        while (*p == ' ') ++p;
        if (strncmp(p, "Received:", 9) == 0) return (LT_RCVD);
        if (*p == '\0' || *p == '\015' || *p == '\012') return (LT_BLANK);
           else return(LT_OTHER);
}

static char fnamebuffer[100];
char *makefname(env_root, name)
char *env_root, *name;
{
	strncpy(fnamebuffer, getenv(env_root), 100);
	strcat(fnamebuffer, name);
	return fnamebuffer;
}

/* E O F */
