/*
   show.c : Copyright Paul Healy, EI9GL, 1992.

   911227 : Added this header
   920620 : Version 0.1
            - doesn't remember current folder
            - doesn't support listing of a subrange
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "misc.h"
#include "help.h"
#include "pager.h"
#include "mailer.h"
#include "buf.h"
#include "mailfile.h"
#include "message.h"

#ifdef BMH
#define main show_main
#endif

static int Nmsgs = 0;

static int
nextprompt(void)
{
   return Nmsgs++ > 0 ? page_next() : 0;
}

static char *
mta_addr(char *line)
{
   char *start = strchr(line, ' ');

   if (start == NULL)
      return "<>";
   else {
      char *end = strpbrk(start+1, " \t\n");
      
      if (end != NULL)
         *end = '\0';

      return start+1;
      }
}

static MsgResult
showmsg(Mailfile *m, char *line)
{
   int ret = MSG_EOF;

   if (nextprompt() == -1)
      return MSG_QUIT;

   page_setup("show");

   if ( (page_printf("(Message %s:%d)", mail_folder(m), mail_curmsg(m)) == -1) ||
        (page_printf("Return-Path: %s", mta_addr(line)) == -1)
      )
      return MSG_QUIT;

   while ( (mail_getstring(m, line, LINELEN) != NULL) && (ret==MSG_EOF) )
      if (ismsg(line))
         ret = MSG_NEXT;             /* have reached the next message */
      else if (page_puts(line) == -1)
         ret = MSG_QUIT;             /* signal to quit walking */

   page_fin("show");

   return ret;
}

int
show(Mailfile *m, int argc, char *argv[])
{
   Nmsgs = 0;

   if (mail_getfile(m, argc, argv, "show") == -1)
      return -1;

   if (argc == 1)
      walk2msg(m, showmsg);
   else {
      int msg = mail_lastmsg(m, argc, argv);

      walkselect(m, showmsg, argc, argv);
      mail_setcur(m, msg);
      }

   return 0;
}

int
main(int argc, char *argv[])
{
   Mailfile m;

   if (dohelp(argc, argv, "show [+folder] [msglist] (show mail)"))
      return 0;

   mail_zero(&m);

   show(&m, argc, argv);

   return mail_savestate(&m);
}
