/*
   smtp.c

   smtp queue management routines

   Taken from bm.

   Copyright 1986 Bdale Garbee, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.
   Copyright 1987 1988 Dave Trulli NN2Z, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.

   911218 : Added this header
   920802 : Added file locking to get_msgid, bm wasn't locking at all.
   920819 : Allow killing of mail in a mqueue directory whose path has
            '.' in..
   930124 : moved get_msgid to misc.c to avoid makefile awkwardness
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "config.h"
#include "lock.h"
#include "misc.h"
#include "pc.h"
#include "rc.h"
#include "help.h"
#include "pager.h"

#ifdef BMH
#define main smtp_main
#endif

/*
 * list jobs wating to be sent in the mqueue
 */
static void
listqueue(char *mqueue)
{
   char tstring[80];
   char workfile[80];
   char line[20];
   char host[LINELEN];
   char to[LINELEN];
   char from[LINELEN];
   char *p;
   struct stat stbuf;
   struct tm *tminfo;

   page_setup("smtp");

   page_puts("S     Job    Size Date  Time  Host                 From");

   sprintf(workfile,"%s/*.wrk", mqueue); 
   filedir(workfile, 0, line);

   while(line[0] != '\0') {
      FILE *fp;

      sprintf(tstring, "%s/%s", mqueue, line);

      if ((fp = fopen(tstring, "r")) == NULL) {
         perror(tstring);
         continue;
         }

      if ((p = strrchr(line, '.')) != NULL)
         *p = '\0';

      sprintf(tstring, "%s/%s%s", mqueue, line, EXT);
      stat(tstring,&stbuf);
      tminfo = localtime(&stbuf.st_ctime);

      fgets(host, sizeof(host), fp);
      fgets(from, sizeof(from), fp);

      sprintf(tstring, "%s/%s.lck", mqueue, line);

      page_printf("%c %7s %7ld %02d/%02d %02d:%02d %-20s %s",
         access(tstring,0) == 0 ? 'L' : ' ',
         line,
         stbuf.st_size,
         tminfo ->tm_mon+1,
         tminfo->tm_mday,
         tminfo->tm_hour,
         tminfo->tm_min,
         rip(host),
         rip(from) );

      while (fgets(to, sizeof(to), fp) != NULL)
         page_printf("         %s", rip(to));

      fclose(fp);
      filedir(workfile, 1, line);
      }
   page_fin("smtp");
}

static int
killjob(char *mqueue, char *j)
{
   char s[LINELEN];

   sprintf(s, "%s/%s.lck", mqueue, j);

   if (access(s, 0) == 0) {
      char tbuf[20];

      fprintf(stderr, "smtp: warning job %s is locked by SMTP, remove? (y,n)", j);
      fgets(tbuf, sizeof(tbuf), stdin);
      if (tolower(*tbuf) != 'y')
         return -1;
      if (unlink(s) == -1)
         fprintf(stderr, "smtp: error deleting %s\n", s);
      }

   sprintf(s, "%s/%s.wrk", mqueue, j);
   if (unlink(s) == -1)
      fprintf(stderr, "smtp: job %s not found\n", s);

   sprintf(s, "%s/%s.txt", mqueue, j);

   return unlink(s);
}

int
main(int argc, char *argv[])
{
   if (dohelp(argc, argv, "smtp [joblist] (list/kill jobs on the mail queue)"))
      return 0;

   if (argc == 1)
      listqueue(getrc(mqueue));
   else {                                    /* Kill jobs */
      int i;

      for (i=1; i<argc; i++)
         killjob(getrc(mqueue), argv[i]);
      }

   return 0;
}
