/* smtpcli.c
 *        Client routines for Simple Mail Transfer Protocol ala RFC821
 *        A.D. Barksdale Garbee II, aka Bdale, N3EUA
 *        Copyright 1986 Bdale Garbee, All Rights Reserved.
 *        Permission granted for non-commercial copying and use, provided
 *        this notice is retained.
 *         Modified 14 June 1987 by P. Karn for symbolic target addresses,
 *        also rebuilt locking mechanism
 */

#include <stdio.h>
#include "machdep.h"
#include "netuser.h"
#include "mbuf.h"
#include "timer.h"
#include "tcp.h"
#include "smtp.h"
#include "trace.h"

extern int16 lport;                        /* local port placeholder */
int32 resolve();
int32 gateway;
static void sendit();
static struct timer smtpcli_t;
char *index(),*rindex();

/* init routine called when program fired up */
smtpclinit()
{
        int dosmtptick();

        smtpcli_t.func = (void (*)())dosmtptick;/* what to call on timeout */
        smtpcli_t.arg = 0;                        /* dummy value */
        smtpcli_t.start = SMTPCLITIME;                /* set timer duration */
        start_timer(&smtpcli_t);                /* and fire it up */
}

/* this is the routine that gets called every so often to do outgoing mail
   processing */
int
dosmtptick()
{
        char        tmpstring[LINELEN],
                wfilename[13];
        FILE *lfile;
        struct smtp_msg *mp;
        struct socket lsocket, fsocket;
        char *calloc(),*malloc();
        void smtp_rec(), smtp_cts(), smtp_state();

#ifdef        DEBUG
        printf("DOSMTPTICK() entered\n");
#endif
        for(filedir(MAILQUEUE,0,wfilename);wfilename[0] != '\0';
                filedir(MAILQUEUE,1,wfilename)){

                /* Try to create the lock file in an atomic operation
                 * This is unfortunately Aztec-specific
                 */
                sprintf(tmpstring,"%s%s",MAILQDIR,wfilename);
                strcpy(rindex(tmpstring,'.'),".lck");        /* index "can't fail" */
                if((lfile = fopen(tmpstring,"w+")) == NULL)
                        continue;
                fclose(lfile);

                /* if we have work, rebuild the exact (non-wild) filename */
                mp = (struct smtp_msg *)calloc(1,sizeof (struct smtp_msg));
                mp->filename = malloc((unsigned)strlen(MAILQDIR)
                 + strlen(wfilename) + 1);
                strcpy(mp->filename,MAILQDIR);
                strcat(mp->filename,wfilename);
                if(trace & TRACE_SMTP)
                        printf("mail file: %s\r\l",mp->filename);
                mp->wfile = fopen(mp->filename,"r");        /* "can't fail" */

                /* open text file */
                strcpy(tmpstring,mp->filename);
                strcpy(rindex(tmpstring,'.'),".txt");        /* index "can't fail" */
                /* If the text file doesn't exist, a null message is sent */
                mp->tfile = fopen(tmpstring,"r");

                fgets(tmpstring,LINELEN,mp->wfile);        /* read target host */
                rip(tmpstring);

                /* set up the rest of the socket info from what we got */
                if((fsocket.address = resolve(tmpstring)) == 0)
                        fsocket.address = gateway;        /* Use default if unknown */

                fsocket.port = SMTP_PORT;
                lsocket.address = ip_addr;        /* our ip address */
                lsocket.port = lport++;                /* next unused port */

                if(trace & TRACE_SMTP)
                        printf("Connecting to %s\r\l",inet_ntoa(fsocket.address));

                /* open smtp connection */
                mp->state = CLI_OPEN_STATE;                /* init state placeholder */
                mp->tcb = open_tcp(&lsocket,&fsocket,TCP_ACTIVE,tcp_window,
                        smtp_rec,smtp_cts,smtp_state,0,(int *)mp);
                mp->tcb->user = (int *)mp;                /* Upward pointer */
        }
}

/* this is the master state machine that handles a single SMTP transaction */
smtp_transaction(mp)
struct smtp_msg *mp;
{
        char tmpstring[LINELEN];

#ifdef        DEBUG
        printf("SMTP_TRANSACTION() called, state=%u\n",mp->state);
#endif
        if(trace & TRACE_SMTP){
                fputs(mp->buf,stdout);
                printf("\r\l");
                fflush(stdout);
        }
        if(mp->buf[0] == '0' || mp->buf[3] == '-'){
                /* Another line follows; ignore this one */
        } else if (affirmative(mp)) {
                switch(mp->state) {
                case CLI_OPEN_STATE:
                        mp->state = CLI_HELO_STATE;
                        sendit(mp,"HELO %s\r\l",hostname);
                        break;
                case CLI_HELO_STATE:
                        mp->state = CLI_MAIL_STATE;
                        /* issue MAIL command */
                        fgets(tmpstring,LINELEN,mp->wfile);        /* who from */
                        rip(tmpstring);
                        sendit(mp,"MAIL FROM:<%s>\r\l",tmpstring);
                        break;                        
                case CLI_MAIL_STATE:
                        mp->state = CLI_RCPT_STATE;
                        /* issue RCPT command */
                        fgets(tmpstring,LINELEN,mp->wfile);        /* who to */
                        rip(tmpstring);
                        sendit(mp,"RCPT TO:<%s>\r\l",tmpstring);
                        fclose(mp->wfile);
                        break;
                case CLI_RCPT_STATE:
                        mp->state = CLI_DATA_STATE;
                        /* issue DATA command */
                        sendit(mp,"DATA\r\l");
                        break;
                case CLI_DATA_STATE:
                        mp->state = CLI_SEND_STATE;
                        /* Kick the data transfer to get it started */
                        smtp_cts(mp->tcb,mp->tcb->window - mp->tcb->sndcnt);
                        break;
                case CLI_SEND_STATE:
                        /* the transmitter upcall routine will advance the
                           state pointer on end of file, so we do nada... */
                        break;
                case CLI_UNLK_STATE:
                        unlink(mp->filename);        /* unlink workfile */
                        /* close and unlink the textfile */
                        if(mp->tfile != NULLFILE)
                                fclose(mp->tfile);
                        strcpy(tmpstring,mp->filename);
                        strcpy(rindex(tmpstring,'.'),".txt");
                        unlink(tmpstring);
                        strcpy(tmpstring,mp->filename);
                        strcpy(rindex(tmpstring,'.'),".lck");
                        unlink(tmpstring);
                        mp->state = CLI_QUIT_STATE;
                        /* issue a quit command */
                        sendit(mp,"QUIT\r\l");
                        break;
                case CLI_QUIT_STATE:
                        /* either start next transaction, or quit */
                        close_tcp(mp->tcb);        /* close up connection */
                        break;
                }
        } else {
                /* if we get here, means we got a negative reply
                 * for the moment, just let that hose us...
                 */
                mp->state = CLI_QUIT_STATE;
                /* Release the lock file */
                strcpy(tmpstring,mp->filename);
                strcpy(rindex(tmpstring,'.'),".lck");
                unlink(tmpstring);
                /* issue a quit command */
                sendit(mp,"QUIT\r\l");
        }
}

/* return true if the passed string contains a positive response code */
affirmative(mp)
struct smtp_msg *mp;
{
        /* 2 is always good, 3 is ok if we've just sent 'data' command */
        if ((*mp->buf == '2') || 
           ((*mp->buf == '3') && (mp->state == CLI_DATA_STATE)))
                return 1;
        else         return 0;
}

/* smtp receiver upcall routine.  fires up the state machine to parse input */
static
void
smtp_rec(tcb,cnt)
struct tcb *tcb;
int16 cnt;
{
        register struct smtp_msg *mp;
        char *inet_ntoa(), c;
        struct mbuf *bp;
        /* may want a void line here for procedures used */

#ifdef        DEBUG
        printf("SMTP_REC called\n");
#endif
        mp = (struct smtp_msg *)tcb->user;        /* point to our struct */
        recv_tcp(tcb,&bp,cnt);        /* suck up chars from low level routine */

        /* Assemble input line in buffer, return if incomplete */
        while(pullup(&bp,&c,1) == 1) {
                switch(c) {
                case '\r':        /* strip cr's */
                        continue;
#ifndef OSK
                case '\n':        /* line is finished, go do it! */
#else
                case '\l':        /* line is finished, go do it! */
#endif
                        mp->buf[mp->cnt] = '\0';
                        smtp_transaction(mp);
                        mp->cnt = 0;
                        break;
                default:        /* other chars get added to buffer */
                        mp->buf[mp->cnt++] = c;
                        break;
                }
        }
}

/* smtp transmitter ready upcall routine.  twiddles cts flag */
static 
void
smtp_cts(tcb,cnt)
struct tcb *tcb;
int16 cnt;
{
        register struct smtp_msg *mp;
        struct mbuf *bp;
        char *cp;
        int c;

#ifdef        DEBUG
        printf("SMTP_CTS called\n");
#endif
        mp = (struct smtp_msg *)tcb->user;        /* point to our struct */

        /* don't do anything until/unless we're supposed to be sending */
        if(mp->state != CLI_SEND_STATE)
                return;

        if((bp = alloc_mbuf(cnt)) == NULLBUF){
                /* Hard to know what to do here */
                return;
        }
        cp = bp->data;
        while(cnt > 1 && (c = getc(mp->tfile)) != EOF){
#ifdef OSK
                if (c == '\r')  ungetc('\l', mp->tfile);
#endif OSK
                *cp++ = c;
                bp->cnt++;
                cnt--;
        }
        if(bp->cnt != 0)
                send_tcp(tcb,bp);
        else
                free_p(bp);

        if(cnt > 1){        /* EOF seen */
                sendit(mp,"\r\l.\r\l");
                mp->state = CLI_UNLK_STATE;
        }
}

/* smtp state change upcall routine.  cans connection on error */
static
void
smtp_state(tcb,old,new)
struct tcb *tcb;
char old,new;
{
        struct smtp_msg *mp;

#ifdef        DEBUG
        printf("SMTP_STATE called, state=%u\n",new);
#endif
        mp = (struct smtp_msg *)tcb->user;
        switch(new) {
        case ESTABLISHED:
                mp->state = CLI_OPEN_STATE;        /* shouldn't be needed */
                break;
        case CLOSE_WAIT:
                close_tcp(tcb);                        /* shut things down */
                        /* may want to do something here to shut down
                           the rest of the transaction? */
                break;
        case CLOSED:
                del_tcp(tcb);                        /* hosed for good */
                if(mp->filename != NULLCHAR)
                        free(mp->filename);
                free((char *)mp);
                break;
        }
}

/* Send message back to server */
/*VARARGS*/
static
void
sendit(mp,fmt,arg1,arg2)
struct smtp_msg *mp;
char *fmt,*arg1,*arg2;
{
        struct mbuf *bp,*qdata();
        char tmpstring[LINELEN];

        if(trace & TRACE_SMTP){
                printf(">>> ");
                printf(fmt,arg1,arg2);
                fflush(stdout);
        }
        sprintf(tmpstring,fmt,arg1,arg2);
        bp = qdata(tmpstring,(int16)strlen(tmpstring));
        send_tcp(mp->tcb,bp);
}

