/*	SMTP Utilities
 *	used by both client and server.
 *	This material is believed to be in the public domain.
 *
 *	Jan 92	Bill Simpson
 *		These were moved (mostly) from smtpserv.c,
 *		so that the non-server mode could be compiled.
 */
#include <stdio.h>
#include <time.h>
#ifdef UNIX
#include <sys/types.h>
#endif
#include <ctype.h>

#include "global.h"
#ifdef ANSIPROTO
#include <stdarg.h>
#endif
#include "netuser.h"
#include "socket.h"
#include "domain.h"
#include "files.h"
#include "mailutil.h"
#include "smtp.h"

int Smtpmode = 0;

static void smtplog __ARGS((char *fmt,...));
static int router_queue __ARGS((FILE *data,char *from,struct recipient_s *to));
static int mailuser __ARGS((FILE *data,char *from,char *to));

char *Days[7] = {  "Sun","Mon","Tue","Wed","Thu","Fri","Sat" };
char *Months[12] = { "Jan","Feb","Mar","Apr","May","Jun",
		"Jul","Aug","Sep","Oct","Nov","Dec" };

/* Return Date/Time in Arpanet format in passed string */
char *
ptime(t)
long *t;
{
	/* Print out the time and date field as
	 *		"DAY day MONTH year hh:mm:ss ZONE"
	 */
	register struct tm *ltm;
	static char tz[4];
	static char str[40];
	char *p, *getenv();
	/* Read the system time */
	ltm = localtime(t);

	if (*tz == '\0')
		if ((p = getenv("TZ")) == NULL)
			strcpy(tz,"UTC");
		else
			strncpy(tz,p,3);

	/* rfc 822 format */
	sprintf(str,"%s, %.2d %s %02d %02d:%02d:%02d %.3s\n",
		Days[ltm->tm_wday],
		ltm->tm_mday,
		Months[ltm->tm_mon],
		ltm->tm_year,
		ltm->tm_hour,
		ltm->tm_min,
		ltm->tm_sec,
		tz);
	return(str);
}


#if	defined(ANSIPROTO)
static void
smtplog(char *fmt, ...)
{
	va_list ap;
	char *cp;
	long t;
	FILE *fp;

	if ((fp = fopen(Maillog,APPEND_TEXT)) == NULLFILE)
		return;
	time(&t);
	cp = ctime(&t);
	rip(cp);
	fprintf(fp,"%s ",cp);
	va_start(ap,fmt);
	vfprintf(fp,fmt,ap);
	va_end(ap);
	fprintf(fp,"\n");
	fclose(fp);
}

#else

static void
smtplog(fmt,arg1,arg2,arg3,arg4)
char *fmt;
int arg1,arg2,arg3,arg4;
{
	char *cp;
	long t;
	FILE *fp;

	if ((fp = fopen(Maillog,APPEND_TEXT)) == NULLFILE)
		return;
	time(&t);
	cp = ctime(&t);
	rip(cp);
	fprintf(fp,"%s ",cp);
	fprintf(fp,fmt,arg1,arg2,arg3,arg4);
	fprintf(fp,"\n");
	fclose(fp);
}
#endif


/* add an element to the front of the list pointed to by head.
 * (never fails)
 */
struct recipient_s *
add_rcpt(head,val,type)
struct recipient_s **head;
char *val;
int type;
{
	register struct recipient_s *tp;

	tp = (struct recipient_s *)callocw(1,sizeof(struct recipient_s));

	/* allocate storage for the char string */
	tp->val = strdup(val);
	tp->type = type;

	/* add entry to front of existing list */
	tp->next = *head;
	*head = tp;
	return tp;

}

/* delete a list of recipients */
void
del_rcpt(lp)
struct recipient_s *lp;
{
	register struct recipient_s *tp, *tp1;

	for (tp = lp; tp != NULLRCPT; tp = tp1) {
		tp1 = tp->next;
		free(tp->val);
		free((char *)tp);
	}
}


long 
get_msgid()
{
	char sfilename[FILE_PATH_SIZE];
	char s[20];
	register long sequence = 0;
	FILE *sfile;

	sprintf(sfilename,"%s/sequence.seq",Mailqdir);
	sfile = fopen(sfilename,READ_TEXT);

	/* if sequence file exists, get the value, otherwise set it */
	if (sfile != NULL) {
		(void) fgets(s,sizeof(s),sfile);
		sequence = atol(s);
	/* Keep it in range of and 8 digit number to use for dos name prefix. */
		if (sequence < 0L || sequence > 99999999L )
			sequence = 0;
		fclose(sfile);
	}

	/* increment sequence number, and write to sequence file */
	sfile = fopen(sfilename,WRITE_TEXT);
	fprintf(sfile,"%ld",++sequence);
	fclose(sfile);
	return sequence;
}


/* place a mail job in the outbound queue */
int
queuejob(dfile,host,to,from)
FILE *dfile;
char *host;
struct recipient_s *to;
char *from;
{
	FILE *fp;
	struct recipient_s *ap;
	char tmpstring[FILE_PATH_SIZE], prefix[9], buf[512];
	register int cnt;

	sprintf(prefix,"%ld",get_msgid());
	mlock(Mailqdir,prefix);
	sprintf(tmpstring,"%s/%s.txt",Mailqdir,prefix);
	if((fp = fopen(tmpstring,WRITE_TEXT)) == NULLFILE) {
		(void) rmlock(Mailqdir,prefix);
		return 1;
	}
	while((cnt = fread(buf, 1, sizeof(buf), dfile)) > 0)
		if(fwrite(buf, 1, cnt, fp) != cnt)
			break;
	if(ferror(fp)){
		fclose(fp);
		(void) rmlock(Mailqdir,prefix);
		return 1;
	}
	fclose(fp);
	sprintf(tmpstring,"%s/%s.wrk",Mailqdir,prefix);
	if((fp = fopen(tmpstring,WRITE_TEXT)) == NULLFILE) {
		(void) rmlock(Mailqdir,prefix);
		return 1;
	}
	fprintf(fp,"%s\n%s\n",host,from);
	for(ap = to; ap != NULLRCPT; ap = ap->next) {
		fprintf(fp,"%s\n",ap->val);
		smtplog("queue job %s To: %s From: %s",prefix,ap->val,from);
	}
	fclose(fp);
	(void) rmlock(Mailqdir,prefix);
	return 0;
}


/* Deliver mail to the appropriate mail boxes */
static int
router_queue(data,from,to)
FILE *data;
char *from;
struct recipient_s *to;
{
	int c;
	register struct recipient_s *ap;
	FILE *fp;
	char tmpstring[50];
	char prefix[9];

	sprintf(prefix,"%ld",get_msgid());
	mlock(Routeqdir,prefix);
	sprintf(tmpstring,"%s/%s.txt",Routeqdir,prefix);
	if((fp = fopen(tmpstring,WRITE_TEXT)) == NULLFILE) {
		(void) rmlock(Routeqdir,prefix);
		return 1;
	}
	rewind(data);
	/* a character at a time??? {WAS}*/
	while((c = getc(data)) != EOF)
		if(putc(c,fp) == EOF)
			break;
	if(ferror(fp)){
		fclose(fp);
		(void) rmlock(Routeqdir,prefix);
		return 1;
	}
	fclose(fp);
	sprintf(tmpstring,"%s/%s.wrk",Routeqdir,prefix);
	if((fp = fopen(tmpstring,WRITE_TEXT)) == NULLFILE) {
		(void) rmlock(Routeqdir,prefix);
		return 1;
	}
	fprintf(fp,"From: %s\n",from);
	for(ap = to;ap != NULLRCPT;ap = ap->next) {
		fprintf(fp,"To: %s\n",ap->val);
	}
	fclose(fp);
	(void) rmlock(Routeqdir,prefix);
	smtplog("rqueue job %s From: %s",prefix,from);
	return 0;
}


/* General mailit function. It takes a list of addresses which have already
** been verified and expanded for aliases. Based on the current mode,
** the message is placed in an mbox, the outbound smtp queue, or the
** rqueue interface
 */
int
mailit(data,buf,len,from,tolist)
FILE *data;
char *buf;
int len;
char *from;
struct recipient_s *tolist;
{
	struct recipient_s *ap;
	struct recipient_s *qlist = NULLRCPT;
	register FILE *fp;
	char	mailbox[50], *cp;
	char	*qhost;
	int	c, fail = 0;
	time_t	t;

	if ((Smtpmode & QUEUE) != 0)
		return(router_queue(data,from,tolist));

	/* make up new files for outgoing mail */
	do {
		qhost = NULLCHAR;

		for(ap = tolist;ap != NULLRCPT;ap = ap->next) {
			if (ap->type == DOMAIN) {
				char *host;

				if ((host = strrchr(ap->val,'@')) != NULLCHAR)
					host++;
				else
					host = Hostname;
				if(qhost == NULLCHAR)
			     		qhost = host;
				if(stricmp(qhost,host) == 0) {
					ap->type = BADADDR; /* temporary */
					add_rcpt(&qlist,ap->val,0);
				}
			}
		}
		if(qhost != NULLCHAR) {
			rewind(data);
			queuejob(data,qhost,qlist,from);
			del_rcpt(qlist);
			qlist = NULLRCPT;
		}
	} while(qhost != NULLCHAR);

	/* copy local mail to mailboxes */
	for(ap = tolist;ap != NULLRCPT;ap = ap->next) {
		if(ap->type != LOCAL) {
			ap->type = DOMAIN; /* change back */
			continue;
		}
		rewind(data);

		/* strip off host name of LOCAL addresses */
		if ((cp = strchr(ap->val,'@')) != NULLCHAR)
			*cp = '\0';

		/* truncate long user names */
		if (strlen(ap->val) > MBOXLEN)
			ap->val[MBOXLEN] = '\0';

		/* if mail file is busy save it in our smtp queue
		 * and let the smtp daemon try later.
		 */
		if (mlock(Mailspool,ap->val)) {
			add_rcpt(&qlist,ap->val,0);
			fail = queuejob(data,Hostname,qlist,from);
			del_rcpt(qlist);
			qlist = NULLRCPT;
		} else {
			int tocnt = 0;
			char *receipt = NULLCHAR;

			sprintf(mailbox,"%s/%s.txt",Mailspool,ap->val);
#ifndef	AMIGA
			if((fp = fopen(mailbox,APPEND_TEXT)) != NULLFILE) {
#else
			if((fp = fopen(mailbox,"r+")) != NULLFILE) {
				(void) fseek(fp, 0L, 2);
#endif
				time(&t);
				fprintf(fp,"From %s %s",from,ctime(&t));

				/* This copies the header, looking for
				 * To or Cc lines, and Return-Receipt lines.
				 */
				while(fgets(buf,len,data) != NULLCHAR){
					if(buf[0] == '\n'){
						if(tocnt == 0)
							fprintf(fp,"%s%s\n",
								Hdrs[APPARTO],
								ap->val);
						fputc('\n',fp);
						break;
					}
					fputs(buf,fp);

					rip(buf);
					switch(htype(buf)){
					case TO:
					case CC:
						++tocnt;
						break;
					case RRECEIPT:
						if((cp = getaddress(buf,0))
						   != NULLCHAR){
							free(receipt);
							receipt = strdup(cp);
						}
						break;
					}
				}
				pwait(NULL);

				/* do a fast? block copy */
				while((c = fread(buf,1,len,data)) > 0)
					if(fwrite(buf,1,c,fp) != c)
						break;
				if(ferror(fp))
					fail = 1;
				else
					fprintf(fp,"\n");
				/* Leaves an extra blank line between msgs */
				fclose(fp);

				printf("New mail arrived for %s\n",ap->val);
				if(receipt != NULLCHAR){
					rewind(data); /* Send return receipt */
					mdaemon(data,receipt,NULLRCPT,0);
					free(receipt);
				}
			} else 
				fail = 1;
			(void) rmlock(Mailspool,ap->val);
			if (fail)
				break;
			smtplog("deliver: To: %s From: %s",ap->val,from);
		}
	}
	return fail;
}


#define SKIPWORD(X) while(*X && *X!=' ' && *X!='\t' && *X!='\n' && *X!= ',') X++;
#define SKIPSPACE(X) while(*X ==' ' || *X =='\t' || *X =='\n' || *X == ',') X++;

/* check for an alias and expand alias into a address list */
struct recipient_s *
expandalias(head, user)
struct recipient_s **head;
char *user;
{
	FILE *fp;
	register char *s,*p;
	struct rr *rrp, *rrlp;
	int inalias = 0;
	struct recipient_s *tp;
	char buf[PLINELEN];
	
	/* no alias file found */
	if ((fp = fopen(Alias, READ_TEXT)) == NULLFILE) {
		/* Try MB, MG or MR domain name records */
		rrlp = rrp = resolve_mailb(user);
		while(rrp != NULLRR){
			if(rrp->rdlength > 0){
				/* remove the trailing dot */
				rrp->rdata.name[rrp->rdlength-1] = '\0';
				/* replace first dot with @ if there is no @ */
				if(strchr(rrp->rdata.name,'@') == NULLCHAR
				   && (p = strchr(rrp->rdata.name,'.')) !=
				   NULLCHAR)
					*p = '@';
				if(strchr(rrp->rdata.name,'@') != NULLCHAR)
					tp = add_rcpt(head,rrp->rdata.name,
						     DOMAIN);
				else
					tp = add_rcpt(head,rrp->rdata.name,
						     LOCAL);
				++inalias;
			}
			rrp = rrp->next;
		}
		free_rr(rrlp);
		if(inalias)
			return tp;
		else
			return add_rcpt(head, user, LOCAL);
	}

	while (fgets(buf,sizeof(buf),fp) != NULLCHAR) {
		p = buf;
		if ( *p == '#' || *p == '\0')
			continue;
		rip(p);

		/* if not in an matching entry skip continuation lines */
		if (!inalias && isspace(*p))
			continue;

		/* when processing an active alias check for a continuation */
		if (inalias) {
			if (!isspace(*p)) 
				break;	/* done */
		} else {
			s = p;
			SKIPWORD(p);
			*p++ = '\0';	/* end the alias name */
			if (strcmp(s,user) != 0)
				continue;	/* no match go on */
			inalias = 1;
		}

		/* process the recipients on the alias line */
		SKIPSPACE(p);
		while(*p != '\0' && *p != '#') {
			s = p;
			SKIPWORD(p);
			if (*p != '\0')
				*p++ = '\0';

			/* find hostname */
			if (strchr(s,'@') != NULLCHAR)
				tp = add_rcpt(head,s,DOMAIN);
			else
				tp = add_rcpt(head,s,LOCAL);
			SKIPSPACE(p);
		}
	}
	(void) fclose(fp);

	if (inalias)	/* found and processed and alias. */
		return tp;

	/* no alias found treat as a local address */
	return add_rcpt(head, user, LOCAL);
}


#ifdef	MSDOS
/* Illegal characters in a DOS filename */
static char baddoschars[] = "\"[]:|<>+=;,";
#endif

/* test if mail address is valid */
int
validate_address(s)
char *s;
{
	char *cp;
	int32 addr;

	/* if address has @ in it the check dest address */
	if ((cp = strrchr(s,'@')) != NULLCHAR) {
		cp++;
		/* 1st check if its our hostname
		 * if not then check the hosts file and see
		 * if we can resolve their address to a known site
		 * or one of our aliases
		 */
		if (strcmp(cp,Hostname) != 0) {
			if ((addr = mailroute(cp)) == 0
				&& (Smtpmode & QUEUE) == 0)
				return BADADDR;
			if (ismyaddr(addr) == NULLIF)
				return DOMAIN;
		}
		
		/* on a local address remove the host name part */
		*--cp = '\0';
	}

	/* if using an external router leave address alone */
	if ((Smtpmode & QUEUE) != 0)
		return LOCAL;

	/* check for the user%host hack */
	if ((cp = strrchr(s,'%')) != NULLCHAR) {
		*cp = '@';
		cp++;
		/* reroute based on host name following the % seperator */
		if (mailroute(cp) == 0)
			return BADADDR;
		else
			return DOMAIN;
	}

#ifdef MSDOS	/* dos file name checks */
	/* Check for characters illegal in MS-DOS file names */
	for(cp = baddoschars;*cp != '\0';cp++){
		if(strchr(s,*cp) != NULLCHAR)
			return BADADDR;
	}
#endif
	return LOCAL;
}


/* send mail to a single user */
static int
mailuser(data,from,to)
FILE *data;
char *from;
char *to;
{
	char *buf;
	int address_type, ret;
	struct recipient_s *tolist = NULLRCPT;

	/* check if address is ok */
	if ((address_type = validate_address(to)) == BADADDR) {
		return 1;
	}
	/* if a local address check for an alias */
	if (address_type == LOCAL) {
		expandalias(&tolist, to);
	} else {
		/* a remote address is added to the list */
		add_rcpt(&tolist, to, address_type);
	}

	buf = mallocw(TLINELEN);
	ret = mailit(data,buf,TLINELEN,from,tolist);
	free(buf);
	del_rcpt(tolist);
	return ret;

}


/* Mailer daemon return mail mechanism */
int
mdaemon(data,to,lp,bounce)
FILE *data;		/* pointer to rewound data file */
char *to;		/* Overridden by Errors-To: line if bounce is true */
struct recipient_s *lp;        /* error log for failed mail */
int bounce;		/* True for failed mail, otherwise return receipt */
{
	time_t t;
	FILE *tfile;
	char *buf, *cp, *newto = NULLCHAR;
	int cnt;

	buf = mallocw(TLINELEN);

	if ( to == NULLCHAR || *to == '\0' || bounce){
		while(fgets(buf,TLINELEN,data) != NULLCHAR) {
			if(buf[0] == '\n')
				break;
			/* Look for Errors-To: */
			if(htype(buf) == ERRORSTO &&
			   (cp = getaddress(buf,0)) != NULLCHAR){
				free(newto);
				newto = strdup(cp);
				break;
			}
		}
		if ( to == NULLCHAR
		  || newto == NULLCHAR && *to == '\0') {
			free(buf);
			return -1;
		}
		rewind(data);
	}
	if((tfile = tmpfile()) == NULLFILE) {
		free(buf);
		return -1;
	}

	time(&t);
	fprintf(tfile,"%s%s",Hdrs[DATE],ptime(&t));
	fprintf(tfile,"%s<%ld@%s>\n",Hdrs[MSGID],get_msgid(),Hostname);
	fprintf(tfile,"%sMAILER-DAEMON@%s (Mail Delivery Subsystem)\n",
		Hdrs[FROM],Hostname);
	fprintf(tfile,"%s%s\n",Hdrs[TO],newto != NULLCHAR ? newto : to);
	fprintf(tfile,"%s%s\n\n",Hdrs[SUBJECT],
		bounce ? "Failed mail" : "Return receipt");
	if(bounce) {
		fprintf(tfile,"  ===== transcript follows =====\n\n");
		for (; lp != NULLRCPT; lp = lp->next)
			fprintf(tfile,"%s\n",lp->val);
		fprintf(tfile,"\n");
	}
	fprintf(tfile,"  ===== %s follows ====\n",
		bounce ? "Unsent message" : "Message header");

	while(fgets(buf,sizeof(buf),data) != NULLCHAR){
		if(buf[0] == '\n')
			break;
		fputs(buf,tfile);
		pwait(NULL);
	}
	if(bounce){
		fputc('\n',tfile);
		while((cnt = fread(buf,1,sizeof(buf),data)) > 0) {
			fwrite(buf,1,cnt,tfile);
			pwait(NULL);
		}
	}
	free(buf);

	fseek(tfile,0L,0);
	/* A null From<> so no looping replys to MAIL-DAEMONS */
	(void) mailuser(tfile,"",newto != NULLCHAR ? newto : to);
	fclose(tfile);
	free(newto);
	return 0;
}


