/* TCP header tracing routines
 * Copyright 1991 Phil Karn, KA9Q
 */
#include <stdio.h>
#include "global.h"
#include "mbuf.h"
#include "netuser.h"
#include "internet.h"
#include "tcp.h"
#include "ip.h"
#include "trace.h"

/* TCP segment header flags */
static char *Tcpflags[] = {
	"FIN",	/* 0x01 */
	"SYN",	/* 0x02 */
	"RST",	/* 0x04 */
	"PSH",	/* 0x08 */
	"ACK",	/* 0x10 */
	"URG",	/* 0x20 */
	"CE"	/* 0x40 */
};

/* Dump a TCP segment header. Assumed to be in network byte order */
void
tcp_dump(ip,bpp,source,dest,check)
struct iface  *ip;
struct mbuf **bpp;
int32 source,dest;	/* IP source and dest addresses */
int check;		/* 0 if checksum test is to be bypassed */
{
	struct tcp seg;
	struct pseudo_header ph;
	int16 csum;
	int16 dlen;

	if(bpp == NULLBUFP || *bpp == NULLBUF)
		return;

	/* Verify checksum */
	ph.source = source;
	ph.dest = dest;
	ph.protocol = TCP_PTCL;
	ph.length = len_p(*bpp);
	csum = cksum(&ph,*bpp,ph.length);

	ntohtcp(&seg,bpp);

	prtTrace(ip,"TCP: %u->%u Seq x%lx",seg.source,seg.dest,seg.seq,seg.ack);
	if(seg.flags.ack)
		prtTrace(ip," Ack x%lx",seg.ack);
	if(seg.flags.congest)
		prtTrace(ip," %s",Tcpflags[6]);
	if(seg.flags.urg)
		prtTrace(ip," %s",Tcpflags[5]);
	if(seg.flags.ack)
		prtTrace(ip," %s",Tcpflags[4]);
	if(seg.flags.psh)
		prtTrace(ip," %s",Tcpflags[3]);
	if(seg.flags.rst)
		prtTrace(ip," %s",Tcpflags[2]);
	if(seg.flags.syn)
		prtTrace(ip," %s",Tcpflags[1]);
	if(seg.flags.fin)
		prtTrace(ip," %s",Tcpflags[0]);

	prtTrace(ip," Wnd %u",seg.wnd);
	if(seg.flags.urg)
		prtTrace(ip," UP x%x",seg.up);
	/* Print options, if any */
	if(seg.mss != 0)
		prtTrace(ip," MSS %u",seg.mss);
	if((dlen = len_p(*bpp)) != 0)
		prtTrace(ip," Data %u",dlen);
	if(check && csum != 0)
		prtTrace(ip," CHECKSUM ERROR (%u)",csum);
	prtTrace(ip, "\n");
}

