/* Terminal emulator for KA9Q
 * Copyright 1991 Glenn Hamilton McGregor
 * Based on previous code Copyright 1991 Phil Karn
 */

/*
 *  Includes
 */

#include <dos.h>
#include <conio.h>

#include "global.h"
#include "hardware.h"
#include "main.h"
#include "mbuf.h"
#include "session.h"
#include "socket.h"
#include "terminal.h"

/*
 *  Externals
 */

extern FILE *Rawterm;

/*
 *  local variables
 */

/* Keyboard input buffer */
#define	KBSIZE	256
static struct { 			/* kb fifo - queues scan/ascii words */
	int16 buf[KBSIZE];
	int16 *wp;
	int16 *rp;
	int cnt;
} Keyboard;

static int16 kb_biosstat;		/* bios function for kb status */
static int16 kb_biosread;		/* bios function for kb read */

/* temp kb function table */

typedef struct
	{
	int16	code;
	int16	func;
	int16	khar;
	} kb_map_t;

kb_map_t	kb_map[] =
	{
	key_f10,	slc_escape,	0xFFFF,
	key_enter,	slc_forw1,	'\r',
	key_ctl_enter,	slc_forw2,	'\n',
	key_backspace,	slc_ec, 	'\b',
	key_del,	slc_ec, 	'\177',
	key_edt_del,	slc_ec, 	'\177',
	key_ctl_u,	slc_el, 	'\025',
	key_ctl_r,	slc_rp, 	'\022',
	0xFFFF, 	0xFFFF, 	0xFFFF
	};

/*
 *  Forward declations
 */

static int16 kbcode	__ARGS((void));
static int16 kbread	__ARGS((void));
static void kbfunc	__ARGS((int16 code, int16 *func, int16 *khar));
static struct mbuf *ttydriv	__ARGS((struct session *sp, int16 func, int16 khar));

/*
 *  Keyboard code initialization
 */
void
kbinit()
{
	union REGS regs;

	/* determine if bios supports extended keyboard functions */
	regs.x.ax = 0x12FF;
	int86 ( 0x16, &regs, &regs );
	if ( regs.h.al != 0xFF ) {
		kb_biosread = 0x10;	/* extended bios - 101 key */
		kb_biosstat = 0x11;
	} else {
		kb_biosread = 0x00;	/* oldstyle bios - 84 key */
		kb_biosstat = 0x01;
	}

	/* initialize the keyboard fifo */
	Keyboard.rp = Keyboard.wp = Keyboard.buf;
}

/*
 * Keyboard interrupt handler
 */
void
kbint()
{
	int sig = FALSE;
	int16 code;

	while ( ( code = kbcode() ) != 0xFFFF && Keyboard.cnt+1 < KBSIZE ) {
		/* put scan/char code in keyboard fifo */
		*Keyboard.wp++ = code;
		if(Keyboard.wp == &Keyboard.buf[KBSIZE])
			Keyboard.wp = Keyboard.buf;
		Keyboard.cnt++;
		/* remember to tell keyboard task */
		sig = TRUE;
	}
	/* tell keyboard task a key is ready */
	if ( sig ) {
		psignal ( &Keyboard, 0 );
	}
}

/*
 * Keyboard code - reads keystroke from bios returning
 *	scan/ascii code as a word,
 *	or 0xFFFF if no keystroke available.
 */

static int16
kbcode()
{
	union REGS regs;

	/* kludge for WAS - read no chars with scroll lock on */
	if ( *(byte_t *)MK_FP(0x40,0x17) & 0x10 )
		return 0xFFFF;

	/* determine if key stroke is available from bios */
	regs.h.ah = kb_biosstat;
	int86 ( 0x16, &regs, &regs );

	/* char available? */
	if ( ( regs.x.flags & 0x40 ) == 0 ) {
		regs.h.ah = kb_biosread;
		int86 ( 0x16, &regs, &regs );
		return regs.x.ax;
	} else {
		return 0xFFFF;
	}
}

/*
 * kbread -- read a scan/char code from the keyboard fifo.
 *	Suspend until the code is available.
 */
static int16
kbread()
{
	char i_state;
	int16 code;

	i_state = dirps();
	while(Keyboard.cnt == 0)
		pwait(&Keyboard);
	Keyboard.cnt--;
	restore(i_state);
	code = *Keyboard.rp++;
	if(Keyboard.rp == &Keyboard.buf[KBSIZE])
		Keyboard.rp = Keyboard.buf;
	return code;
}

/*
 *  kbfunc - convert a kb code into a function value.
 */

static void
kbfunc ( code, func, khar )
int16 code;
int16 *func, *khar;
{
	kb_map_t *kbm;
	byte_t scan, ascii;

	kbm = kb_map;
	while ( kbm->code != 0xFFFF ) {
		if ( kbm->code == code ) {
			*func = kbm->func;
			*khar = kbm->khar;
			return;
		}
		kbm++;
	}

	scan = hibyte ( code );
	ascii = lobyte ( code );

	if ( scan == 0 ) {
		*func = slc_ascii;
		*khar = ascii;
	} else if ( ascii == 0 || ascii == 0xE0 ) {
		*func = slc_ignore;
		*khar = 0xFFFF;
	} else {
		*func = slc_ascii;
		*khar = ascii;
	}

	return;

}


/* Keyboard input process */
void
keyboard(i,v1,v2)
int i;
void *v1;
void *v2;
{
	int16 code, func, khar;
	struct mbuf *bp;

	/* Keyboard process loop */
	for(;;){
		code = kbread();
		kbfunc( code, &func, &khar );
		if ( func == slc_escape && Current != Command ) {
			/* Save current tty mode and set cooked */
			swapscreen(Current,Command);
			Lastcurr = Current;
			Current = Command;
		}

		Current->row = MOREROWS;
		psignal(&Current->row,1);

		if ( !Current->morewait ) {
			/* If the screen driver was in morewait state, this char
			 * has woken him up. Toss it so it doesn't also get taken
			 * as normal input. If the char was a command escape,
			 * however, it will be accepted; this gives the user
			 * a way out of lengthy output.
			 */
			 if ( (bp = ttydriv(Current,func,khar) ) != NULLBUF){
				send_mbuf(Current->input,bp,0,NULLCHAR,0);
			}
		}
	}
}

#define	LINESIZE	256

/* Accept characters from the incoming tty buffer and process them
 * (if in cooked mode) or just pass them directly (if in raw mode).
 *
 * Echoing (if enabled) is direct to the raw terminal. This requires
 * recording (if enabled) of locally typed info to be done by the session
 * itself so that edited output instead of raw input is recorded.
 */
static struct mbuf *
ttydriv(sp,func,khar)
struct session *sp;
int16 func;
int16 khar;
{
	struct mbuf *bp;
	char *cp,*rp;

	if ( !sp->ttystate.edit ) {
		if ( func == slc_ascii || khar != 0xFFFF ) {
			bp = ambufw(1);
			*bp->data = khar;
			bp->cnt = 1;
			if(sp->ttystate.echo)
				putc(khar,Rawterm);
			return bp;
		} else {
			putc ( '\a', Rawterm );
			return NULLBUF;
		}
	} else {
		if(sp->ttystate.line == NULLBUF)
			sp->ttystate.line = ambufw(LINESIZE);

		bp = sp->ttystate.line;
		cp = bp->data + bp->cnt;

		/* Perform cooked-mode line editing */
		switch ( func ) {

		case slc_ascii:
			if ( bp->cnt < LINESIZE-1 ) {
				*cp = khar;
				bp->cnt++;

				if(sp->ttystate.echo)
					putc(khar,Rawterm);
			} else {
				putc('\a',Rawterm);   /* Beep */
			}
			break;

		case slc_forw1:
		case slc_forw2:    /* assume room for forwarding condition */
			if(sp->ttystate.crnl) {
				*cp = '\n';
				bp->cnt++;
			} else if ( khar != 0xFFFF ) {
				*cp = khar;
				bp->cnt++;
			}
			if(sp->ttystate.echo)
				fputs(Eol,Rawterm);

			sp->ttystate.line = NULLBUF;
			return bp;

		case slc_ec:
			if(bp->cnt != 0){
				bp->cnt--;
				if(sp->ttystate.echo)
					fputs("\b \b",Rawterm);
			}
			break;

		case slc_rp:
			if(sp->ttystate.echo){
				fprintf(Rawterm,"^R%s",Eol) ;
				rp = bp->data;
				while (rp < cp)
					putc(*rp++,Rawterm) ;
			}
			break;

		case slc_el:
			while(bp->cnt != 0){
				bp->cnt--;
				if(sp->ttystate.echo){
					fputs("\b \b",Rawterm);
				}
			}
			break;

		case slc_ignore:
		case slc_escape:
			break;

		default:
			putc('\a', Rawterm);
			break;
		}
	}
	return NULLBUF;
}

void
newscreen(sp)
struct session *sp;
{
	if(sp != NULLSESSION)
		sp->screen = callocw(1,sizeof(struct screen));
}

void
freescreen(sp)
struct session *sp;
{
	if(sp == NULLSESSION || sp->screen == NULLSCREEN)
		return;
	if(sp->screen->save != NULLCHAR)
		free(sp->screen->save);
	free((char *)sp->screen);
}

/* Save specified session screen and resume console screen */
void
swapscreen(old,new)
struct session *old,*new;
{
	struct text_info tr;

	if(old == new)
		return;	/* Nothing to do */

	fflush(Rawterm);
	gettextinfo(&tr);
	if(old != NULLSESSION){
		/* Save old screen */
		if(old->screen->save == NULLCHAR)
			old->screen->save
			 = malloc(2*tr.screenheight*tr.screenwidth);
		if(old->screen->save != NULLCHAR)
			gettext(tr.winleft,tr.wintop,tr.winright,
			 tr.winbottom,old->screen->save);
		old->screen->row = tr.cury;
		old->screen->col = tr.curx;
	}
	if(new != NULLSESSION){
		/* Load new screen */
		if(new->screen->save != NULLCHAR){
			puttext(tr.winleft,tr.wintop,tr.winright,
			 tr.winbottom,new->screen->save);
			gotoxy(new->screen->col,new->screen->row);
			/* Free the memory (saves 4K on a continuous basis) */
			free(new->screen->save);
			new->screen->save = NULLCHAR;
		} else
			clrscr();	/* Start with a fresh slate */
	}
	alert(Display,1);	/* Wake him up */
}
void
display(i,v1,v2)
int i;
void *v1;
void *v2;
{
	int c;
	struct session *sp;

	/* This is very tricky code. Because the value of "Current" can
	 * change any time we do a pwait, we have to be careful to detect
	 * any change and go back and start again.
	 */
	for(;;){
		sp = Current;

		if(sp->morewait){
			pwait(&sp->row);
			if(sp != Current || sp->row <= 0){
				/* Current changed value, or the user
				 * hasn't really hit a key
				 */
				continue;
			}
			/* Erase the prompt */
			fprintf(Rawterm,"\r        \r");
		}
		sp->morewait = 0;
		if((c = rrecvchar(sp->output)) == -1){
			/* the alert() in swapscreen will cause this to
			 * return -1 when current changes
			 */
			pwait(NULL);	/* Prevent a nasty loop */
			continue;
		}
		putc(c,Rawterm);
		if(sp->record != NULLFILE)
			putc(c,sp->record);
		if(sp->flowmode && c == '\n' && --sp->row <= 0){
			fprintf(Rawterm,"--More--");
			sp->morewait = 1;
		}
	}
}
