/* Packet tracing - top level and generic routines, including hex/ascii
 * Copyright 1991 Phil Karn, KA9Q
 *
 * Mods by G1EMM
 *
 * Tracing to session taken from WNOS3, by Johan. K. Reinalda, WG7J
 */
#if defined(PM)
#define INCL_DOSMISC
#define INCL_DOSPROCESS
#define INCL_WIN
#define UM_REFRESH WM_USER + 12
#define TRACE_CACHE 4096
#define MAX_ENTRY 2048
#include <os2.h>
#endif
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include "global.h"
#include "config.h"
#ifdef ANSIPROTO
#include <stdarg.h>
#endif
#include "mbuf.h"
#include "hardware.h"
#include "iface.h"
#include "pktdrvr.h"
#include "commands.h"
#include "session.h"
#include "trace.h"
#ifdef TRACE
static void ascii_dump __ARGS((struct iface *ifp,struct mbuf **bpp));
static void ctohex __ARGS((char *buf,int16 c));
static void fmtline __ARGS((struct iface *ifp,int16 addr,char *buf,int16 len));
static void hex_dump __ARGS((struct iface *ifp,struct mbuf **bpp));
static void showtrace __ARGS((struct iface *ifp));
extern struct session *Current;
extern struct session *Command;
#ifdef MULTITASK
extern int Nokeys;
#endif
#if defined(PM)
void newTrace(struct session *sp);
void freeTrace(struct session *sp);
#endif
extern int Tracesession;
extern struct session *Trace;

int
dostrace(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    if(Trace == NULLSESSION)
        argc = 0; /* No session setup, so don't allow turning it on ! */
    return setbool(&Tracesession,"Trace to session",argc,argv);
}

/* Redefined here so that programs calling dump in the library won't pull
 * in the rest of the package
 */

static char nospace[] = "No space!!\n";

struct tracecmd Tracecmd[] = {
	"input",	IF_TRACE_IN,	IF_TRACE_IN,
	"-input",	0,		IF_TRACE_IN,
	"output",	IF_TRACE_OUT,	IF_TRACE_OUT,
	"-output",	0,		IF_TRACE_OUT,
	"broadcast",	0,		IF_TRACE_NOBC,
	"-broadcast",	IF_TRACE_NOBC,	IF_TRACE_NOBC,
	"raw",		IF_TRACE_RAW,	IF_TRACE_RAW,
	"-raw",		0,		IF_TRACE_RAW,
	"ascii",	IF_TRACE_ASCII,	IF_TRACE_ASCII|IF_TRACE_HEX,
	"-ascii",	0,		IF_TRACE_ASCII|IF_TRACE_HEX,
	"hex",		IF_TRACE_HEX,	IF_TRACE_ASCII|IF_TRACE_HEX,
	"-hex",		IF_TRACE_ASCII,	IF_TRACE_ASCII|IF_TRACE_HEX,
	"off",		0,		0xffff,
	NULLCHAR,	0,		0
};

void
dump(ifp,direction,type,bp)
register struct iface *ifp;
int direction;
unsigned type;
struct mbuf *bp;
{
	struct mbuf *tbp;
	void (*func) __ARGS((struct iface *,struct mbuf **,int));
	int16 size;
	time_t timer;
	char *cp;

    if(ifp == NULL || (ifp->trace & direction) == 0)
        return; /* Nothing to trace */
#if !defined(PM)
    if(Tracesession) {
        /* Disable trace if this is not Trace-sessions,
         * or when shelled out, and not tracing to file */
#ifdef MULTITASK
        if((Current != Trace || Nokeys) && (ifp->trfp == stdout))
#else
        if((Current != Trace) && (ifp->trfp == stdout))
#endif /* MULTITASK */
            return; /* Nothing to trace */
    } else {
        /* Disable trace on non-command sessions or when shelled out */
#ifdef MULTITASK
        if((Current != Command || Nokeys) && (ifp->trfp == stdout))
#else
        if((Current != Command) && (ifp->trfp == stdout))
#endif
            return; /* Nothing to trace */
    }
#endif  /*PM*/
    switch(direction){
	case IF_TRACE_IN:
		if((ifp->trace & IF_TRACE_NOBC)
		 && (Tracef[type].addrtest != NULLFP)
		 && (*Tracef[type].addrtest)(ifp,bp) == 0)
			return;		/* broadcasts are suppressed */
		time(&timer);
		cp = ctime(&timer);
		cp[24] = '\0';
		prtTrace(ifp,"\n%s - %s recv:\n",cp,ifp->name);
		break;
	case IF_TRACE_OUT:
		time(&timer);
		cp = ctime(&timer);
		cp[24] = '\0';
		prtTrace(ifp,"\n%s - %s sent:\n",cp,ifp->name);
		break;
	}
	if(bp == NULLBUF || (size = len_p(bp)) == 0){
		prtTrace(ifp,"empty packet!!\n");
		return;
	}

	if(type < NCLASS)
		func = Tracef[type].tracef;
	else
		func = NULLVFP;

	dup_p(&tbp,bp,0,size);
	if(tbp == NULLBUF){
		prtTrace(ifp,nospace);
		return;
	}
	if(func != NULLVFP)
		(*func)(ifp,&tbp,1);
	if(ifp->trace & IF_TRACE_ASCII){
		/* Dump only data portion of packet in ascii */
		ascii_dump(ifp,&tbp);
	} else if(ifp->trace & IF_TRACE_HEX){
		/* Dump entire packet in hex/ascii */
		free_p(tbp);
		dup_p(&tbp,bp,0,len_p(bp));
		if(tbp != NULLBUF)
			hex_dump(ifp,&tbp);
		else
			prtTrace(ifp,nospace);
	}
	free_p(tbp);
}

/* common trace to device routine */
void prtTrace(struct iface *ip, char *pszBuffer, ...)
{
char * pszXfer;
va_list arg_ptr;

	va_start(arg_ptr, pszBuffer);

#if !defined(PM)
	vfprintf(ip->trfp,pszBuffer, arg_ptr);
        va_end_(arg_ptr);
        return;
#else
	if ((!ip->pszTraceBuffer) && (!ip->trfp))
                {
                va_end(arg_ptr);
		return;
                }
	if (ip->trfp)
		vfprintf(ip->trfp, pszBuffer, arg_ptr);
	else
		{
		pszXfer = malloc(MAX_ENTRY);	/* nice round number */
		if (!pszXfer)	/* no memory */
                        {
                        va_end(arg_ptr);
			return;
                        }
		vsprintf(pszXfer, pszBuffer, arg_ptr);
		if ((strlen(ip->pszTraceBuffer) + strlen(pszXfer)) > (TRACE_CACHE -1))
			{
			while (!WinPostMsg(ip->sp->hwndSession, UM_REFRESH, MPFROMLONG(0L), 0L))
				DosSleep(1);
			while (!WinPostMsg(ip->sp->hwndSession, MLM_INSERT, MPFROMP(ip->pszTraceBuffer), MPFROMLONG(1L)))
				DosSleep(1);
			while (!WinPostMsg(ip->sp->hwndSession, UM_REFRESH, MPFROMLONG(1L), 0L))
				DosSleep(1);
			DosQuerySysInfo(QSV_MS_COUNT, QSV_MS_COUNT, &ip->ulTime, sizeof(ip->ulTime));
			ip->pszTraceBuffer = malloc(TRACE_CACHE);
			if (ip->pszTraceBuffer)
				strcpy(ip->pszTraceBuffer, pszXfer);
			}
		else
			strcat(ip->pszTraceBuffer, pszXfer);
		free(pszXfer);
		}
	va_end(arg_ptr);
#endif
}

/* Dump packet bytes, no interpretation */
void
raw_dump(ifp,direction,bp)
struct iface *ifp;
int direction;
struct mbuf *bp;
{
	struct mbuf *tbp;
	/* Dump entire packet in hex/ascii */
	prtTrace(ifp,"\n******* raw packet dump (%s %s)\n",
		((direction & IF_TRACE_OUT) ? "send" : "recv"),ifp->name);
	dup_p(&tbp,bp,0,len_p(bp));
	if(tbp != NULLBUF)
		hex_dump(ifp,&tbp);
	else
		prtTrace(ifp, nospace);
	prtTrace(ifp,"*******\n");
	free_p(tbp);
	return;
}

/* Dump an mbuf in hex */
static void
hex_dump(ifp,bpp)
struct iface *ifp;
register struct mbuf **bpp;
{
	int16 n;
	int16 address;
	char buf[16];

	if(bpp == NULLBUFP || *bpp == NULLBUF)
		return;

	address = 0;
	while((n = pullup(bpp,buf,sizeof(buf))) != 0){
		fmtline(ifp,address,buf,n);
		address += n;
	}
}
/* Dump an mbuf in ascii */
static void
ascii_dump(ifp,bpp)
struct iface *ifp;
register struct mbuf **bpp;
{
	int c;
	register int16 tot;
	char *pachBuffer;
	char achTmp[6];
	int cnt;
	if(bpp == NULLBUFP || *bpp == NULLBUF)
		return;
	cnt = 0;
	pachBuffer = malloc(MAX_ENTRY);
	*pachBuffer = '\0';
	tot = 0;
	while((c = PULLCHAR(bpp)) != -1){
		if((tot % 64) == 0)
			{
			sprintf(achTmp, "%04x ",tot);
			memcpy(&pachBuffer[cnt], achTmp, strlen(achTmp));
			cnt += strlen(achTmp);
			}
		pachBuffer[cnt++] = isprint(uchar(c)) ? c : '.';
		if((++tot % 64) == 0)
			pachBuffer[cnt++] = '\n';
	}
	if((tot % 64) != 0)
		pachBuffer[cnt++] = '\n';
	if (cnt)
		{
		pachBuffer[cnt] = '\0';
		if (ifp->trfp)
			fprintf(ifp->trfp, "%s", pachBuffer);
		else
			prtTrace(ifp, "%s", pachBuffer); /* added format */
		free(pachBuffer);
		}
	else
		free(pachBuffer);
}
/* Print a buffer up to 16 bytes long in formatted hex with ascii
 * translation, e.g.,
 * 0000: 30 31 32 33 34 35 36 37 38 39 3a 3b 3c 3d 3e 3f  0123456789:;<=>?
 */
static void
fmtline(struct iface *ifp, int16 addr, char *buf, int16 len)
{
	char line[82];
	register char *aptr,*cptr;
	register char c;
	char *pachBuffer;
	memset(line,' ',sizeof(line));
	ctohex(line,(int16)hibyte(addr));
	ctohex(line+2,(int16)lobyte(addr));
	aptr = &line[6];
	cptr = &line[55];
	while(len-- != 0){
		c = *buf++;
		ctohex(aptr,(int16)uchar(c));
		aptr += 3;
		c &= 0x7f;
		*cptr++ = isprint(uchar(c)) ? c : '.';
	}
	*cptr++ = '\n';
#if !defined(PM)
	fwrite(line,1,(unsigned)(cptr-line), ifp->trfp);
#else
	*cptr = '\0';
	if (ifp->trfp)
		fwrite(line,1,(unsigned)(cptr-line), ifp->trfp);
	else
		{
		pachBuffer = malloc(strlen(line) + 1);
		strcpy(pachBuffer, line);
		if ((strlen(ifp->pszTraceBuffer) + strlen(pachBuffer)) > (TRACE_CACHE -1))
			{
			while (!WinPostMsg(ifp->sp->hwndSession, UM_REFRESH, MPFROMLONG(0L), 0L))
				DosSleep(1);
			while (!WinPostMsg(ifp->sp->hwndSession, MLM_INSERT, MPFROMP(ifp->pszTraceBuffer), MPFROMLONG(1L)))
				DosSleep(1);
			while (!WinPostMsg(ifp->sp->hwndSession, UM_REFRESH, MPFROMLONG(1L), 0L))
				DosSleep(1);
			DosQuerySysInfo(QSV_MS_COUNT, QSV_MS_COUNT, &ifp->ulTime, sizeof(ifp->ulTime));
			ifp->pszTraceBuffer = malloc(TRACE_CACHE);
			if (ifp->pszTraceBuffer)
				strcpy(ifp->pszTraceBuffer, pachBuffer);
			}
		else
			strcat(ifp->pszTraceBuffer, pachBuffer);
		free(pachBuffer);
		}
#endif
}
/* Convert byte to two ascii-hex characters */
static void
ctohex(char *buf, int16 c)
{
	static char hex[] = "0123456789abcdef";

	*buf++ = hex[hinibble(c)];
	*buf = hex[lonibble(c)];
}

/* Modify or displace interface trace flags */
int
dotrace(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct iface *ifp;
	struct tracecmd *tp;

	if(argc < 2){
		for(ifp = Ifaces; ifp != NULLIF; ifp = ifp->next)
			showtrace(ifp);
		return 0;
	}
	if((ifp = if_lookup(argv[1])) == NULLIF){
        tprintf(Badinterface,argv[1]);
		return 1;
	}
	if(ifp->port){
		tprintf("No trace on this interface - use master.\n");
		return 1;
	}
	if(argc == 2){
		showtrace(ifp);
		return 0;
	}
	/* MODIFY THIS TO HANDLE MULTIPLE OPTIONS */
	if(argc >= 3){
		for(tp = Tracecmd;tp->name != NULLCHAR;tp++)
			if(strncmp(tp->name,argv[2],strlen(argv[2])) == 0)
				break;
		if(tp->name != NULLCHAR)
			ifp->trace = (ifp->trace & ~tp->mask) | tp->val;
		else
			ifp->trace = htoi(argv[2]);
	}
	/* Always default to stdout unless trace file is given */
#if !defined(PM)
	if(ifp->trfp != NULLFILE && ifp->trfp != stdout)
		fclose(ifp->trfp);
	ifp->trfp = stdout;
#else
	if (ifp->sp)
		{
		freeTrace(ifp->sp);
		free(ifp->sp->name);
		free(ifp->sp);
		free(ifp->pszTraceBuffer);
		ifp->pszTraceBuffer = (char *)NULL;
		ifp->ulTime = 0L;
		ifp->trfp = NULL;
		ifp->sp = NULLSESSION;
		}
	if (ifp->trfp != NULLFILE)
                {
		fclose(ifp->trfp);
                ifp->trfp = NULLFILE;
                }
#endif
	if(ifp->trfile != NULLCHAR)
		free(ifp->trfile);
	ifp->trfile = NULLCHAR;
	if (*argv[2] == '0' || (strcmp("off", argv[2]) == 0))
		return(0);
	if(argc >= 4){
		if((ifp->trfp = fopen(argv[3],APPEND_TEXT)) == NULLFILE){
			tprintf("Can't write to %s\n",argv[3]);
#if !defined(PM)
			ifp->trfp = stdout;
#endif
		} else {
			ifp->trfile = strdup(argv[3]);
		}
	}
#if defined(PM)
	else
		{
		ifp->sp = malloc(sizeof(struct session));
		if (ifp->sp)
			{
			memset(ifp->sp, '\0', sizeof(struct session));
			ifp->pszTraceBuffer = malloc(TRACE_CACHE);
			*(ifp->pszTraceBuffer) = '\0';
			ifp->sp->name = strdup(argv[1]);
			newTrace(ifp->sp);
			}
		}
#endif
	showtrace(ifp);
	return 0;
}

void trtimer(void)
{
clock_t timer;
struct iface *ifp;

	DosQuerySysInfo(QSV_MS_COUNT, QSV_MS_COUNT, &timer, sizeof(timer));
	for(ifp = Ifaces; ifp != NULLIF; ifp = ifp->next)
		{
		if ((ifp->trfp == stdout) || (ifp->trfp == NULL) 
                && ifp->pszTraceBuffer && strlen(ifp->pszTraceBuffer) && ((timer - ifp->ulTime) > 3))
			{
			while (!WinPostMsg(ifp->sp->hwndSession, UM_REFRESH, MPFROMLONG(0L), 0L))
				DosSleep(1);
			while (!WinPostMsg(ifp->sp->hwndSession, MLM_INSERT, MPFROMP(ifp->pszTraceBuffer), MPFROMLONG(1L)))
					DosSleep(1);
			ifp->pszTraceBuffer = malloc(TRACE_CACHE);
			*(ifp->pszTraceBuffer) = '\0';
			ifp->ulTime = timer;
			while (!WinPostMsg(ifp->sp->hwndSession, UM_REFRESH, MPFROMLONG(1L), 0L))
				DosSleep(1);
			}
		}
}
/* Display the trace flags for a particular interface */
static void
showtrace(ifp)
register struct iface *ifp;
{
	if(ifp == NULLIF)
		return;
	tprintf("%s:",ifp->name);
	if(ifp->port){
		tprintf(" trace on master interface only.\n");
		return;
	}
	if(ifp->trace & (IF_TRACE_IN | IF_TRACE_OUT | IF_TRACE_RAW)){
		if(ifp->trace & IF_TRACE_IN)
			tprintf(" input");
		if(ifp->trace & IF_TRACE_OUT)
			tprintf(" output");

		if(ifp->trace & IF_TRACE_NOBC)
			tprintf(" - no broadcasts");

		if(ifp->trace & IF_TRACE_HEX)
			tprintf(" (Hex/ASCII dump)");
		else if(ifp->trace & IF_TRACE_ASCII)
			tprintf(" (ASCII dump)");
		else
			tprintf(" (headers only)");

		if(ifp->trace & IF_TRACE_RAW)
			tprintf(" Raw output");

		if(ifp->trfile != NULLCHAR)
			tprintf(" trace file: %s",ifp->trfile);
		tprintf("\n");
	} else
		tprintf(" tracing off\n");
}

/* shut down all trace files */
void
shuttrace()
{
	struct iface *ifp;

	for(ifp = Ifaces; ifp != NULLIF; ifp = ifp->next){
		if(ifp->trfp != NULLFILE && ifp->trfp != stdout)
			fclose(ifp->trfp);
		if(ifp->trfile != NULLCHAR)
			free(ifp->trfile);
		ifp->trfile = NULLCHAR;
		ifp->trfp = NULLFILE;
		if (ifp->sp)
			{
			freeTrace(ifp->sp);
			free(ifp->sp->name);
			free(ifp->pszTraceBuffer);
			ifp->pszTraceBuffer = (char *)NULL;
			free(ifp->sp);
			ifp->sp = NULLSESSION;
			ifp->trfp = NULL;
			}
	}
}

/* Log messages of the form
 * Tue Jan 31 00:00:00 1987 44.64.0.7:1003 open FTP
 */
#if	defined(ANSIPROTO)
void
trace_log(struct iface *ifp,char *fmt, ...)
{
	va_list ap;
	char *cp;
	long t;

	if(ifp->trfp == NULLFILE)
		return;

	time(&t);
	cp = ctime(&t);
	rip(cp);
	fprintf(ifp->trfp,"%s",cp);

	fprintf(ifp->trfp," - ");
	va_start(ap,fmt);
	vfprintf(ifp->trfp,fmt,ap);
	va_end(ap);
	fprintf(ifp->trfp,"\n");
}
#else
/*VARARGS2*/
void
trace_log(ifp,fmt,arg1,arg2,arg3,arg4,arg5)
struct iface *ifp;
char *fmt;
int arg1,arg2,arg3,arg4,arg5;
{
	char *cp;
	long t;

	if(ifp->trfp == NULLFILE)
		return;

	time(&t);
	cp = ctime(&t);
	rip(cp);
	fprintf(ifp->trfp,"%s",cp);

	fprintf(ifp->trfp," - ");
	fprintf(ifp->trfp,fmt,arg1,arg2,arg3,arg4,arg5);
	fprintf(ifp->trfp,"\n");
}
#endif

#endif /*TRACE*/

