/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */
#include "data.h"
#include "buffer.h"
#include "iface.h"
#include "timer.h"
#include "ax25.h"
#include "ax25l2.h"
#include "l3struc.h"
#include "x25cause.h"
#include "l3calls.h"
#include "tx.h"

extern unsigned char node_addr[16];
extern unsigned char str_work[80];
extern struct datastr *l2_user_info;	/* The message to send dummies */
extern struct ax25_addr L3CALL[];
extern unsigned char L3_MAXVC[];
#define L3SIZ 32

void rx_up_ignore(), tx_up_flush(), level3(), del_ax25();
int st_up_wait();
extern struct ax25_parms *l3parms;
extern unsigned char call_num;
int l2_clear(), l2_reset(), l2_send(), l2_recv(), NULLFCN();
struct datastr *mkpkt();
void rsrtco(), lnk_ign_fcn();

int
l2_st_up(axcbx, oldstate, newstate)
struct axcb *axcbx;		/* Link 		*/
int oldstate, newstate;		/* Old and new states	*/
{
	register struct VCS *vc;
	static struct VCS *lc, *vcp;
	static struct axcb *axcb;
	static int c;
	static unsigned char a;
	struct VCS *new_vc();
	int callcmp();
	void call_request();

	axcb=axcbx;

	puthex2("l2st",axcb,"->",newstate);
	vc=axcb->llcn;
	vcp=vc->peer;
	if ((newstate == 0) && vc) /* Link Reset */ {
		vc->D=D2;
		if (vc->peer) set_d(vc->peer,D2,146);
		set_d(vc,D1,0);
	}
	else {
	   if (newstate == 5) /* Connect occurred */ {
		if (oldstate == 2) /* We made the call, notify peer */ {
			if (vc) set_p(vc,P4,0); /* have a vc */
		} 
		else if (oldstate == 1) /* Connect request */ {
			if (!(vc=new_vc())) return 0; /* No memory */
			if (!(vc->peer=new_vc())) /* No memory */ {
				lc=vc;
				free_vc(lc);
				return 0;
			}
			vcp=vc->peer;
			vcp->peer=vc;
			vc->parent=axcb;
			vc->lcn=0;
			insert_lcn(vc);
			vc->RESET=l2_reset;
			vc->CLEAR=l2_clear;
			vc->SEND=l2_send;
			vc->RECV=l2_recv;
			vc->CONNECTED=NULLFCN;
			str_work[0]=0;
			a=1;
			c=3;
			do {
				if (callcmp(&axcb->path[a],&L3CALL[0])) {
					str_work[0]=0;
					l3_adr(&axcb->path[a-1],str_work);
					if (str_work[0] == 4) /* Was DNIC */ {
						l3_adr(&axcb->path[a-2],str_work);
						c++;
					}
					if (!str_work[0]) /*Failed*/ {
						free_vc(vcp);
						lc=vc;
						free_vc(lc);
						return 0;
					}
					if (!axcb->path[a].ssid & E) /*More*/ {
						memcpy(&vc->digi,&axcb->path[a+1],AXALEN);
					}
					if (a==c) /* Have destination digi */ {
						memcpy(&vcp->digi,&axcb->path[1],AXALEN);
					}
					adr_copy(0,str_work,0,vcp->addr);
					adr_copy(0,node_addr,0,vc->addr);
					break;
				}
			} while ( !(axcb->path[a++].ssid & E));

			mk_nsap(vcp,&axcb->myaddr);
			mk_nsap(vc,axcb->path);
			vc->data[0]=vcp->data[0]=NULL;
			if (call_num == 0) call_num++;
			vc->rand = vcp->rand = (call_num<<8) + rand();
			call_num++;
			vcp->alt = 255;	/* First time through call_request */
			queue(NULL,call_request,0,vcp);
			vc->P = P2;	/* This link is in Call Request */
		   }
		}
		else {
			if (newstate == 1) /* Disconnect occurred */ {
				if (vc=vc->peer) /* Have peer */ {
					vc->peer=NULL;
					c=DTE_Orig;
					if (axcb->parms->retry && !axcb->n2cnt) {
					 	c=Congestion+146;
					 	if (oldstate == 2) c=Ship_Absent;
					}
					set_p(vc,P6,c);
				}
				del_ax25(axcb);
			}
		}
	}
	return 1;
}

#if 0
void
up_restart(lnk)
register struct axcb *lnk;
{
	void rx_up_wait();
	if (lnk->R == R0) set_r(lnk,R2,DTE_Orig); /* Send Restart */
}
#endif
void
tx_l2_msg(lnk)
register struct axcb *lnk;
{
	void rx_up_wait();

	lnk->rpid = PID_TNC;
	rx_up_wait(lnk,0);
}

/* l3_rx_up => see Level3.c, level3() */
/* l3_tx_up => not used, may want to insure no vc's are flow controlled */

l3_st_up(lnk,oldstate,newstate)
register struct axcb *lnk;
int oldstate, newstate;
{
	static struct VCS *vc;
	static struct VCS *lc;
	static int i, c;
	static void (*fcn)();
	void rx_up_wait();
	int route_fail();

	puthex2("l3st",lnk,"->",newstate);
	if (newstate == 0) /* Link Reset */ {
		if (lnk->R == R1) {
			lc=lnk->llcn;
			while (lc) /* More vc's to reset */ {
				vc=lc->peer;
				if (vc) if (vc->P == P4) {
					set_d(vc,D2,146);
				} else set_p(vc,P6,Congestion+146);
				lc=lc->next;
			}
		} else /* Link Reset during L3 Restart */ {
			if (lnk->rpid == PID_X25) rsrtco(lnk); /* Restart again */
			lnk->R = R0;
			lnk->coll=0;
		}
	}
	else {
		if (newstate == 5) /* Connected */ {
			if (lnk->link_num > 0) /* Level 3 Link */ {
				if (i < L3SIZ) lnk->maxvc=L3_MAXVC[i];
				lnk->R=R0;
				lnk->coll=0;
				if (oldstate == 2) c=rnd(COLLTIME);
				else c=0;
				fcn = rsrtco;
			}
			else /* Level 2 User, wait for PID (in I frame) */ {
				lnk->r_upcall=rx_up_wait;
				c=COLLTIME; /* Give L3 User 60 Sec to sent restart */
				fcn = tx_l2_msg;
			}
			queue(&lnk->t20 ,fcn ,c ,lnk);
		}
		else {
			if (newstate == 1) /* Disconnect occurred */ {
				if (route_fail(lnk)) /* Not a network link */ {
					c=DTE_Orig;
					if (lnk->parms->retry != 0
					   && lnk->n2cnt ==0) {
						c=Congestion+146;
					}
					while (lc=lnk->llcn) /* For all VC's */ {
						set_p(lc,P7,c);
						lc=lc->next;
					}
					del_ax25(lnk);
				}
			}
		}
	}
	return 1;
}

void
rx_up_wait(axcb)
register struct axcb *axcb;
{
	static struct datastr *p, *p1;
	struct datastr *frag_pkt();
	void level3();
	int l3_st_up();

	kill(axcb->t20);	/* Timer timed out, or isn't needed anymore */
	if (axcb->rpid != PID_X25) /* Users don't listen to directions!! */ {
		recv_ax25(axcb,&p,0);
		free_pkt(p);
		lnk_ign_fcn(axcb);
		p=dup_pkt(l2_user_info);
		p=frag_pkt(p,128);
		while (p) /* Send each fragment */ {
			p1=p;
			p=p->next;
			p1->next=NULL;
			send_ax25(axcb,p1,PID_TNC);
		}
	}
	else /* Hey! We've got a new Level 3 User!! */ {
		axcb->r_upcall=level3;
		axcb->t_upcall=NULL;
		axcb->s_upcall=l3_st_up;
	}
}

void
rx_up_ignore(axcb)
struct axcb *axcb;
{
	static struct datastr *p;

	recv_ax25(axcb,&p,0); /* Throw away the data */
	free_pkt(p);
}

void
tx_up_flush(axcb)
register struct axcb *axcb;
{
	if ( (axcb->iqueue == NULL) || (axcb->icnt <= 0) ) /* All done! */ {
		close_ax25(axcb);
	}
}

int
st_up_wait(axcb,old,new)
struct axcb *axcb;
int old,new;
{
	puthex2("wait",axcb,"->",new);
	if (new == 1) /* We are done!! */ {
		del_ax25(axcb);
	}
	return 1;
}

int
NULLFCN() {}

int
l3_reset(vc,diag)
struct VCS *vc;
unsigned char diag;
{
	puthex2("l3reset",vc,"",diag);
	tx_rst(vc, DTE_Orig+diag);
	return;
}

int
l2_reset(vcx,diag) 
struct VCS *vcx;
char diag;	/* May want to stuff the diag in the message some day */
{
	static struct datastr *p;

	{
		register struct VCS *vc;
		vc=vcx;		/* VC is ME */
		puthex2("l2reset",vc,"",diag);
		vc=vc->peer;		/* VC is now PEER */
		if (vc->lcn && vc->D == D3) /* Reset came from Level 3 Peer */ {
			tx_crst(vc);
			set_d(vc,D1,0);
		}
	}
	vcx->P = P2;
	{
		register struct axcb *lnk;
		lnk=vcx->parent;
		axconnect(lnk);
		if ((p=mkpkt("\r*** Reset *** ")) == NULL) return;
		putxch(p,diag);
		bappch(p,0x0d);
		send_ax25(lnk,p,PID_TNC);
	}
	return;
}

int
l3_clear(vc,c)
struct VCS *vc;
int c;
{
	tx_clr(vc,c);
	/* Insert code to generate call clear records here */
}

int
l3_connected(vc)
struct VCS *vc;
{
	/* Insert code to generate call records here */
}

int
l2_clear(vc,c)
struct VCS *vc;
unsigned int c;
{
	static struct datastr *p;

	{
		register struct VCS *vcp;
		if ((vcp=vc->peer)) /* Insure we have a peer */ {
			if (vcp->lcn && vcp->P ==P7) /* It's a Level 3 Peer */ {
				tx_cclr(vcp);
				set_p(vcp,P1,0);
			}
		}
	}
	{
		register struct axcb *lnk;
		lnk=vc->parent;
		set_p(vc,P1,0);
		if (!lnk) return;	/* No link exists */
		lnk_ign_fcn(lnk);

		if (lnk->state < 3)
			close_ax25(lnk);
		else {
			if ((p=mkpkt("\r*** Disconnect *** ")) == NULL) return;
			putxch(p,(c>>8));
			putxch(p,(c&0xff));
			bappch(p,0x0d);
			send_ax25(lnk,p,PID_TNC);
		}
	}
}

int
l2_send(vcx, px)
struct VCS *vcx;
struct datastr *px;
{
	register struct axcb *lnk;
	static struct VCS *vc;
	static struct datastr *p;
	static int frsiz, windo;

	vc=vcx;
	puthex2("l2tx",vc,"",px);
	vc_queue_data(vc,(p=px));	/* We were passed data */
	if (vc->P != P4) return 0;	/* Not in data transfer! */
	lnk=vc->parent;

	frsiz = lnk->parms->framesize;
	windo = frsiz * lnk->parms->maxframe;	/* Number of bytes I can send */

	while (p=vc->tx_queue) /* More to send */ {
		if (windo <= lnk->icnt) break;
		vc->tx_queue=p->next;
		p->next=NULL;
		send_ax25(lnk,p,PID_TNC);
	}
	if (lnk->window <= lnk->icnt) return 0;	/* Can't accept any more now */

	return frsiz;	/* Return what we can take now */
}

int
l2_recv(vcx,pkt)
struct VCS *vcx;
struct datastr *pkt;
{
	register struct VCS *vcp;
	static struct VCS *vc;
	static struct axcb *lnk;
	static struct datastr *p;
	static int i;
	int recv_ax25();

	vc=vcx;
	puthex2("l2rx P",vc->P,"busy",vc->RBUSY);
	if (vc->P != P4) return;
	lnk=vc->parent;
	if ((vcp=vc->peer)) /* Have a peer */ {
		i = (*vcp->SEND)(vcp,NULL);
		vc->LBUSY = !i;
		while (!vc->LBUSY) /* And they can take more! */ {
			if (recv_ax25(lnk,&p,i) <= 0) break;
			i = (*vcp->SEND)(vcp,p);
			vc->LBUSY = !i;
		}
	}
}

int
l3_send(vc,px)
register struct VCS *vc;
struct datastr *px;
{
	static struct axcb *lnk;
	static struct datastr *p, *p1;
	static int frsiz, i;

	puthex2("l3tx P",vc->P,"busy",vc->RBUSY);
	vc_queue_data(vc,(p=px));	/* We have data to send */
	if (vc->P != P4) return 0;
	lnk=vc->parent;
	if (vc->RBUSY) return 0;	/* We are flow controlled */
	while ( vc->tx_queue &&		/* We have data to send AND */ 
	   ((vc->LIPR + vc->WI - vc->OPS) % 8) > 0 ) /* Window is open */ {
		i= (vc->OPS << 1) + (vc->OPR << 5);
		if ((p1=new_buffer(3)) == NULL) break;	/* No memory */
		bappch(p1,0x10); /* GFI=1 LCGN=0 */
		bappch(p1,vc->lcn);
		bappch(p1,i); /* p(r) 0 p(s) 0 Packet type */
		p=vc->tx_queue;
		vc->tx_queue=p->next;
		p->next=NULL;
		p1=binsert(p1,p);
		l2_tx(lnk,p1); /* Send the frame */
		vc->OPS = (vc->OPS +1) % 8;
	}
	frsiz=lnk->parms->framesize;
	if (((vc->LIPR + vc->WI - vc->OPS) % 8) > 0 ) /* Window open */ i=frsiz;
	else i=0;	/* Window Closed can't take any more */
	return i;
}

int
l3_recv(vc, pkt)
register struct VCS *vc;
struct datastr *pkt;
{
	static int lbusy;

	puthex2("l3rx",vc,"",pkt);
	if (vc->P != P4) return;
	lbusy=vc->LBUSY;
	vc->LBUSY = !(*vc->peer->SEND)(vc->peer, pkt);
	if (lbusy && !vc->LBUSY) tx_rr(vc);	/* Just got un-busy */
}

void
l2_rx_up(axcb)
struct axcb *axcb;
{
	register struct VCS *vc;

	if (vc=axcb->llcn) (*vc->RECV)(vc,NULL);	/* Have a vc */
}

void
l2_tx_up(axcb)
struct axcb *axcb;
{
	register struct VCS *vc;

	if (vc=axcb->llcn) if (vc=vc->peer) (*vc->RECV)(vc,NULL);
}

void
lnk_l3_pars(lnk)
register struct axcb *lnk;
{
	int l2_window();

	lnk->r_upcall=level3;
	lnk->t_upcall=NULL;
	lnk->s_upcall=l3_st_up;
	lnk->parms=l3parms;
	lnk->window=l2_window(l3parms);
	if ((lnk->link_num = find_call(&lnk->path[0]))>0) lnk->rpid = PID_X25;
}

void
lnk_ign_fcn(lnk)
register struct axcb *lnk;
{
	lnk->r_upcall=rx_up_ignore;
	lnk->t_upcall=tx_up_flush;
	lnk->s_upcall=st_up_wait;
}
