(*===========================================================================*)
(* WA8DED Host mode protocol                                                 *)
(*                                                                           *)
(*   Copyright 1988, 1989 by H. Roy Engehausen.  All rights reserved.        *)
(*   This software may be freely distributed and used, but it may not        *)
(*   under any circumstances be sold by anyone other than the author.        *)
(*   It may be distributed by a commercial company as long as it is          *)
(*   for no cost.                                                            *)
(*                                                                           *)
(*===========================================================================*)

(*---------------------------------------------------------------------------*)
(* HOST to TNC -- Only 1 format                                              *)
(*---------------------------------------------------------------------------*)

TYPE

    host_to_tnc   = RECORD
                      channel     : BYTE;
                      info_cmd    : BYTE;
                      data_count  : BYTE;   (* 1 less than actual data cnt   *)
                      data        : ARRAY[1..256] OF CHAR;
                    END;

(*---------------------------------------------------------------------------*)
(* TNC to HOST -- 3 different formats                                        *)
(*---------------------------------------------------------------------------*)

TYPE

    tnc_to_host   = RECORD

                      channel     : BYTE;

                      CASE tnc_code : BYTE OF

                        0 :     ();  (* Short format.  Only the invariant    *)
                                     (* part                                 *)

                        1,2,3,4,5:   (* Null terminated format.  Data has    *)
                                     (* a null on the end                    *)
                                (data15  : ARRAY[1..256] OF CHAR);

                        6, 7:        (* Byte count format.  Count is 1       *)
                                     (* less than data                       *)

                                (data67_count   : BYTE;
                                 data67         : ARRAY[1..256] OF CHAR);

                    END;

(*---------------------------------------------------------------------------*)
(* Constants for the info/command byte                                       *)
(*---------------------------------------------------------------------------*)

CONST

    info_cmd_info = $00;             (* Transmission is info                 *)
    info_cmd_cmd  = $01;             (* Transmission is command              *)

(*---------------------------------------------------------------------------*)
(* Constants for the TNC to HOST tnc_code byte                               *)
(*---------------------------------------------------------------------------*)

CONST

    t_to_h_ok     = $00;             (* Success                              *)
    t_to_h_okmsg  = $01;             (* Success with message attached        *)
    t_to_h_badmsg = $02;             (* Failure with message attached        *)
    t_to_h_links  = $03;             (* Link status                          *)
    t_to_h_mh_noi = $04;             (* Monitor header.  No I frame follow   *)
    t_to_h_mh_i   = $05;             (* Monitor header.  I frame follows     *)
    t_to_h_mif    = $06;             (* Monitored I frame                    *)
    t_to_h_conn   = $07;             (* Connect information frame            *)

(*---------------------------------------------------------------------------*)
(* Constants for the TNC to HOST messages                                    *)
(*---------------------------------------------------------------------------*)

(*                               123456789012345678901234567890              *)
t_to_h_empty     : STRING[1]  = '';
t_to_h_cnc       : STRING[21] = 'CHANNEL NOT CONNECTED';
t_to_h_badcmd    : STRING[15] = 'INVALID COMMAND';
t_to_h_busy      : STRING[23] = 'TNC BUSY - LINE IGNORED';
t_to_h_cac       : STRING[25] = 'CHANNEL ALREADY CONNECTED';
t_to_h_sac       : STRING[25] = 'STATION ALREADY CONNECTED';
t_to_h_ct        : STRING[13] = 'CONNECTED to ';
t_to_h_lrfm      : STRING[14] = 'LINK RESET fm ';
t_to_h_lrto      : STRING[14] = 'LINK RESET to ';
t_to_h_disc      : STRING[16] = 'DISCONNECTED fm ';
t_to_h_lf        : STRING[18] = 'LINK FAILURE with ';
t_to_h_cr        : STRING[19] = 'CONNECT REQUEST fm ';
t_to_h_fr        : STRING[13] = 'FRAME REJECT ';

