/* NOTE: because of size, the previous 'mailbox.c' has been
 * split in 2 parts:
 * mboxcmd.c, containing the 'mbox' subcommands, and
 * mailbox.c, containing the remote user mailbox commands
 * 920325 - WG7J
 */
/* There are only two functions in this mailbox code that depend on the
 * underlying protocol, namely mbx_getname() and dochat(). All the other
 * functions can hopefully be used without modification on other stream
 * oriented protocols than AX.25 or NET/ROM.
 *
 * SM0RGV 890506, most work done previously by W9NK
 *
 *** Changed 900114 by KA9Q to use newline mapping features in stream socket
 *	interface code; everything here uses C eol convention (\n)
 *
 *	Numerous new commands and other changes by SM0RGV, 900120
 *
 * Gateway function now support outgoing connects with the user's call
 * with inverted ssid. Users can connect to system alias as well...
 * See also several mods in socket.c,ax25.c and others
 * 11/15/91, WG7J/PA3DIS
 *
 * Userlogging, RM,VM and KM commands, and R:-line interpretation
 * added 920307 and later, Johan. K. Reinalda, WG7J/PA3DIS
 *
 * Inactivity timeout-disconnect added 920325 and later - WG7J
 *
 */
#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#if !defined(OS2)
#include <alloc.h>
#endif
#include <stdlib.h>
#include <string.h>
#ifdef  UNIX
#include <sys/types.h>
#include <sys/stat.h>
#endif
#include "global.h"
#include "config.h"
#include "timer.h"
#include "proc.h"
#include "socket.h"
#include "usock.h"
#include "session.h"
#include "smtp.h"
#include "dirutil.h"
#include "telnet.h"
#include "ftp.h"
#include "ftpserv.h"
#include "commands.h"
#include "netuser.h"
#include "files.h"
#include "bm.h"
#include "pktdrvr.h"
#include "ax25.h"
#include "mailbox.h"
#include "ax25mail.h"
#include "nr4mail.h"
#include "cmdparse.h"
#include "mailfor.h"

#ifdef MAILBOX

/* Log all gateway connects to the logfile - WG7J */
#define GWTRACE 1

/* By setting the fp to NULL, we can check in exitbbs()
 * wether a tempfile has been closed or not - WG7J
 */
#define MYFCLOSE(x) { fclose(x); x = (FILE *) 0; }

/*
#define MBDEBUG
*/

extern time_t StartTime;
extern char Myalias[];

struct mbx *Mbox[NUMMBX];
int BbsUsers = 0;
extern char *MMotd;
extern char *Mtmsg;
extern int MAttended;
extern unsigned Maxlet;
extern char Mbpasswd[];

int MbSent = 0;
int MbRead = 0;
int MbRecvd = 0;
#ifdef MBFWD
int MbForwarded = 0;
#endif

extern char Noperm[];
extern char Nosock[];
extern char MboxId[];

static char Loginbanner[] = "\nKA9Q NOS (%s)\n\n";
static char Howtoend[] = "End with /EX or ^Z in first column (^A aborts):\n";
static char CcLine[] = "Cc: ";
static char Mbwelcome[] = "\nWelcome %s,\n";
static char Mbbanner[] = "to the %s TCP/IP Mailbox (%s)\n";
static char Mbwarning[] = "Third Party mail is not permitted.\n";
static char CurUsers[] = "Currently %d user(s)\n";

#ifdef NETROM
static char Mbmenu[] = "Current msg# %d.\n?,A,B,C,D,E,F,H,I,J,K,L,M,N,NR,O,P,R,S,T,U,V,W,X,Z >\n";
#else
static char Mbmenu[] = "Current msg# %d.\n?,A,B,C,D,E,F,H,I,J,K,L,M,O,P,R,S,T,U,V,W,X,Z >\n";
#endif

extern char Mbnrid[];

#ifdef NETROM
#ifdef CALLCLI
static char Longmenu[] = "(?)help  (A)rea     (B)ye   (C)onnect (D)wnload (E)scape\n"
                         "(F)inger (H)elp     (I)nfo  (J)heard  (K)ill    (L)ist   (M)boxusers\n"
                         "(N)odes  (O)perator (P)orts (Q)uery   (R)ead    (S)end   (T)elnet\n"
                         "(U)pload (V)erbose  (W)hat  (X)pert   (Z)ap\n";
#else
static char Longmenu[] = "(?)help    (A)rea  (B)ye    (C)onnect (D)wnload (E)scape    (F)inger\n"
                         "(H)elp     (I)nfo  (J)heard (K)ill    (L)ist    (M)boxusers (N)odes\n"
                         "(O)perator (P)orts (R)ead   (S)end    (T)elnet  (U)pload\n"
                         "(V)erbose  (W)hat  (X)pert  (Z)ap\n";
#endif /*CALLCLI*/
#else /*NETROM*/
#ifdef CALLCLI
static char Longmenu[] = "(?)help (A)rea (B)ye    (C)onnect (D)wnload (E)scape    (F)inger\n"
						 "(H)elp  (I)nfo (J)heard (K)ill    (L)ist    (M)boxusers (O)perator (P)orts\n"
                         "(Q)uery (R)ead (S)end   (T)elnet  (U)pload  (V)erbose   (W)hat     (X)pert\n"
                         "(Z)ap\n";
#else
static char Longmenu[] = "(?)help (A)rea   (B)ye    (C)onnect (D)wnload   (E)scape   (F)inger (H)elp\n"
						 "(I)nfo  (J)heard (K)ill   (L)ist    (M)boxusers (O)perator (P)orts  (R)ead\n"
                         "(S)end  (T)elnet (U)pload (V)erbose (W)hat      (X)pert    (Z)ap\n";
#endif /*CALLCLI*/
#endif /*NETROM*/                                                                                   

extern int dombusers __ARGS((int argc,char *argv[],void *p));
extern int dombpast __ARGS((int argc,char *argv[],void *p));
static int doarea __ARGS((int argc,char *argv[],void *p));
extern int dombstatus __ARGS((int argc,char *argv[],void *p));
extern int dombmailstats __ARGS((int argc,char *argv[],void *p));
static int dombcallbook __ARGS((int argc,char *argv[],void *p));

extern int Usenrid;
extern int MBSecure;
#ifdef MBXTDISC
extern unsigned long Mbtdiscinit;
#endif

/*Enlighten them a bit!
 */
static void
dombconnecthelp() {
#ifdef CONVERS
	tputs("Syntax: 'C                for Conference bridge\n"
#ifdef NETROM
          "        'C <node>'        for NET/ROM connects\n"
#endif
          "        'C <port> <call>' for AX.25 connects\n"
#ifdef CALLBOOK
          "        'CA               to access internet callbook\n"
#endif
          "\n");
#else /* CONVERS */
#ifdef NETROM
    tputs("Syntax: 'C <node>'        for NET/ROM connects\n"
          "        'C <port> <call>' for AX.25 connects\n"
#else
    tputs("Syntax: 'C <port> <call>' for AX.25 connects\n"
#endif /* NETROM */
#ifdef CALLBOOK
          "        'CA               to access internet callbook\n"
#endif
          "\n");
#endif /* CONVERS */
}

int
dombports(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    struct iface *ifp;

    tputs("Available ports:\n");
    for(ifp=Ifaces;ifp!=NULLIF;ifp=ifp->next)
        if(ifp->type == CL_AX25) {
            tprintf("%-7s",ifp->name);
            if(ifp->descr != NULLCHAR)
                tprintf(":  %s",ifp->descr);
            else
                tputc('\n');
        }
    tputc('\n');
    return 0;        
}

/* This is called by the finger-daemon */
void
listusers(s)
int s;
{
	int outsave;
    struct mbx m;

    m.privs = 0;
    m.stype = ' ';

#ifdef notdef
	usprintf(s,"\nCurrent remote users:\n");
#endif
    outsave = Curproc->output;
	Curproc->output = s;
    dombusers(0,NULLCHARP,&m);
	Curproc->output = outsave;
}

struct mbx *
newmbx()
{
	int i;
	struct mbx *m;

	for(i = 0; i < NUMMBX; i++){
		if(Mbox[i] == NULLMBX){
			m = Mbox[i] = (struct mbx *)callocw(1,sizeof(struct mbx));
			m->mbnum = i;
            BbsUsers++;
            return m;
		}
	}
	/* If we get here, there are no free mailbox sessions */
	return NULLMBX;
}

static int
mbx_getname(m)
struct mbx *m;
{
	char *cp;
    FILE *tfp;
	union sp sp;
	char tmp[MAXSOCKSIZE];
	int len = MAXSOCKSIZE;
	int anony = 0;                    
	int oldmode;           
    int founddigit=0;      
    int count=0;             
                                
	sp.p = tmp;
	sp.sa->sa_family = AF_LOCAL;	/* default to AF_LOCAL */
	getpeername(m->user,tmp,&len);
    m->family = sp.sa->sa_family;
    m->path = mallocw(MBXLINE);
	/* This is one of the two parts of the mbox code that depends on the
	 * underlying protocol. We have to figure out the name of the
	 * calling station. This is only practical when AX.25 or NET/ROM is
	 * used. Telnet users have to identify themselves by a login procedure.
	 */
	switch(sp.sa->sa_family){
#ifdef	AX25
	case AF_NETROM:
	case AF_AX25:
		/* NETROM and AX25 socket address structures are "compatible" */
        /*Save user call, in case she/he wants to use gateway function*/
        memcpy(m->call,sp.ax->ax25_addr,AXALEN);
        m->call[ALEN] &= 0xfc;/*Make sure E-bit isn't set !*/
        pax25(m->name,sp.ax->ax25_addr);
        cp = strchr(m->name,'-');
		if(cp != NULLCHAR)			/* get rid of SSID */
			*cp = '\0';
		/* SMTP wants the name to be in lower case */
		cp = m->name; 
		while(*cp){
			if(isupper(*cp))
				*cp = tolower(*cp);
			++cp;
		}
		anony = 1;
		/* Try to find the privileges of this user from the userfile */
        if((m->privs = userlogin(m->name,m->line,&m->path,MBXLINE,&anony)) == -1){
			m->privs = 0;
			free(m->path);
			m->path = NULLCHAR;
		}                                      
        if(m->privs & EXCLUDED_CMD)
			return -1;
		return 0;
#endif
	case AF_LOCAL:
	case AF_INET:
		m->state = MBX_LOGIN;
        tprintf(Loginbanner,Hostname);
        if(Mtmsg != NULLCHAR)
            tputs(Mtmsg);
		for(;;){
            /* Maximum of 3 tries - WG7J */
            if(count++ == 3)
                return -1;
            oldmode = sockmode(m->user,SOCK_ASCII);
            tputs("login: ");
			usflush(m->user);
            if(mbxrecvline(m) == -1)
				return -1;
            if(*m->line == 4) /* Control-d */
                return -1;
            if(*m->line == '\0')
				continue;

            /* add a little test to avoid 'Mailfile busy' syndrome - WG7J */
            if((cp=strchr(m->line,'.')) != NULLCHAR)
                *cp = '_';
            if((cp=strchr(m->line,'/')) != NULLCHAR)
                *cp = '_';
            if((cp=strchr(m->line,'\\')) != NULLCHAR)
                *cp = '_';
            if((tfp=fopen(m->line,"w")) == NULL)  /* Invalid name */
                continue;
            fclose(tfp);
            remove(m->line);
            if(strlen(m->line) < sizeof(m->name))
                strcpy(m->name,m->line);
            else    /* Too long! */
                continue;
			tprintf("Password: %c%c%c",IAC,WILL,TN_ECHO);
			usflush(m->user);
            sockmode(m->user,SOCK_BINARY);
            if(mbxrecvline(m) == -1)
				return -1;
			tprintf("%c%c%c",IAC,WONT,TN_ECHO);
			sockmode(m->user,oldmode);
			tputc('\n');
			usflush(m->user);
			/* This is needed if the password was send before the
			 * telnet no-echo options were receied. We neeed to
			 * flush the eold sequence from the input buffers, sigh
			 */
			if(socklen(m->user,0))/* discard any remaining input */
				recv_mbuf(m->user,NULL,0,NULLCHAR,0);
            if((m->privs = userlogin(m->name,m->line,&m->path,MBXLINE,&anony))
			 != -1){
				if(anony)
                    log(m->user,"MBOX login: %s Password: %s",m->name,m->line);
				else
					log(m->user,"MBOX login: %s",m->name);
				if(m->privs & EXCLUDED_CMD)
					return -1;
#ifdef AX25
                /*try to set the name as the user-call.
                 *this is a very crude test! Be careful...
                 *Login must have at leat 1 digit (0-9) in it,
                 *and it must be possible to convert it to a call.
                 *if this doesn't work, disallow the gateway command,
                 *no matter if this was allowed by priviledges or not.
                 *Be careful, some one with login name '4us' and
                 *permission set to allow gateway/netrom, will
                 *go out as '4us-15' or '4us-0' !!!!!
                 *11/15/91 WG7J/PA3DIS
                 */                    
                for(cp=m->name;*cp != '\0';cp++)
                    if(isdigit((int)*cp))
                        break;
                if(*cp != '\0')
                    founddigit = 1;
                if( (setcall(m->call,m->name) == -1) || (!founddigit) ) {
                    m->privs &= ~AX25_CMD;
                    m->privs &= ~NETROM_CMD;
                }
#else
                m->privs &= ~AX25_CMD;
                m->privs &= ~NETROM_CMD;
                   
#endif /* AX25 */
                /* Set the morerows to MAXLIN for telnet logins - WG7J */
                m->morerows = MAXLIN - 1;
                return 0; 
			}
            tputs("Login incorrect\n");
			*m->name = '\0';	/* wipe any garbage */
		}                                     
	}              
	return 0;
}
                   
/* put up the prompt */                 
void            
putprompt(m)
struct mbx *m;
{                                      
    char area[64];
    char *cp1,*cp2;          

    if(m->sid & MBX_SID)
        tputs(">\n");   
    else {  
        if(m->sid & MBX_NRID)
            tputs(Mbnrid);
        if(m->sid & MBX_AREA) {
            cp1 = m->area;
            cp2 = area;
            /* Convert / and \ into . */
            while(*cp1 != '\0') {
                if(*cp1=='/')
                    *cp2 = '.';                       
                else
                    *cp2 = *cp1;
                cp1++;
                cp2++;
            }
            *cp2 = '\0';
            tprintf("Area: %s ",area);
        }
        tprintf((m->sid & MBX_EXPERT) ? ">\n" : Mbmenu, m->current);
    }
}

#ifdef MBXTDISC
/* Mailbox user has been idle for too long,
 * disconnect the socket - WG7J
 */
void
mboxredundant(m)
struct mbx *m;
{
    struct usock *up;

    /* nasty hack! we may have screwed up reference count */
	/* by invoking newproc("smtp_send",....); Fudge it!   */
    if((up = itop(m->user)) != NULLUSOCK)
        up->refcnt = 1;

    /* Close the socket*/
    close_s(m->user);
    return; /* Keep lint happy */
}
#endif

extern void mbox_converse(struct mbx *);

/* Incoming mailbox session */
void
mbx_incom(s,t,p)
int s;
void *t;
void *p;
{
	struct mbx *m;
	struct usock *up;
	char *buf[3];
	char tmp[AXBUF];
	int rval;

    sockmode(s,SOCK_ASCII);
	sockowner(s,Curproc);	/* We own it now */
	/* Secede from the parent's sockets, and use the network socket that
	 * was passed to us for both input and output. The reference
	 * count on this socket will still be 1; this allows the domboxbye()
	 * command to work by closing that socket with a single call.
	 * If we return, the socket will be closed automatically.
	 */
	close_s(Curproc->output);
	close_s(Curproc->input);
	Curproc->output = Curproc->input = s;

	/* We'll do our own flushing right before we read input */
	setflush(s,-1);

	if((m = newmbx()) == NULLMBX){
        tputs("Too many mailbox sessions\n");
		return;
	}

    m->user = s;
	m->escape = 24;		/* default escape character is Ctrl-X */
	m->type = (int) t;

    /* get the name of the remote station */
	if(mbx_getname(m) == -1) {
		exitbbs(m);
		return;
	}

#ifdef CONVERS
    if((int)t & CONF_LINK) {
        m->state = MBX_CONVERS;
        chname(Curproc,"conference");
        log(s,"open CONF");
        mbox_converse(m);
        Mbox[m->mbnum] = NULLMBX;
        free((char *)m);
        BbsUsers--;
        close_s(Curproc->output);
        return;
    }
#endif

    log(s,"open MBOX");
#ifdef MBXTDISC
    /*Start inactivity timer - WG7J */
    set_timer(&m->tdisc,Mbtdiscinit * 1000L);
    m->tdisc.func = mboxredundant;
    m->tdisc.arg = m;
    start_timer(&m->tdisc);
#endif
    if(m->privs & IS_BBS)
        m->sid = MBX_SID; /*force bbs status*/
    else
        if(m->privs & IS_EXPERT)
            m->sid |= MBX_EXPERT;

    loguser(m);
	m->state = MBX_CMD;	/* start in command state */
    tputs(MboxId);

    /* Say 'hello' only if user is not a bbs - WG7J */
    if(!(m->sid & MBX_SID)) {
        tprintf(Mbwelcome,m->name);
#ifdef AX25
        if(m->family == AF_INET)
#endif
            tprintf(Mbbanner,Hostname,Version);
#ifdef AX25
        else
            tprintf(Mbbanner,pax25(tmp,Mycall),Version);
#endif
        tprintf(CurUsers,BbsUsers);
        if(!ThirdParty)
			tputs(Mbwarning);
		tputc('\n');
		if(MMotd != NULLCHAR)
			tprintf("%s\n",MMotd);
        usflush(m->user);
        /* Enable our local message area,
         * only if we're not a bbs - WG7J
         */
        buf[1] = m->name;
        doarea(2,buf,m);
    }

    /* Send prompt */
    putprompt(m);

    while(mbxrecvline(m) != -1){
        if((rval = mbx_parse(m)) == -2)
			break;
		if(rval == 1)
            tputs("Bad syntax.\n");
#ifdef notdef
        tprintf("last = %ld, new last = %ld\n",m->lastread,m->newlastread);
#endif
        /* Only tell about new mail when in our own area - WG7J */
        if(!(m->sid & MBX_SID) && !strcmp(m->name,m->area) &&
            (isnewprivmail(m) > 0L) )
            tputs("You have new mail. Please Kill when read!\n");
        /* Do not check mailfile if we're bbs, saves a tmpfile- WG7J*/
        if(!(m->sid & MBX_SID))
            scanmail(m);
        putprompt(m);
        m->state = MBX_CMD;
    }
	exitbbs(m);

    /* by invoking newproc("smtp_send",....); Fudge it!   */
    if((up = itop(Curproc->output)) != NULLUSOCK)
        up->refcnt = 1;
    close_s(Curproc->output);

}

void
exitbbs(m)
struct mbx *m;
{
#ifdef MBXTDISC
    stop_timer(&m->tdisc);
#endif
    closenotes(m);
	free(m->to);
	free(m->tofrom);
	free(m->origto);
    free(m->origbbs);
    free(m->subject);
    free(m->date);
    free(m->tomsgid);
	free(m->path);
    /* Close the tempfiles if they are not nullpointers - WG7J */
    if(m->tfile != (FILE *) 0)
        fclose(m->tfile);
    if(m->tfp != (FILE *) 0)
        fclose(m->tfp);
    if(m->stdinbuf != NULLCHAR)
		free(m->stdinbuf);
	if(m->stdoutbuf != NULLCHAR)
		free(m->stdoutbuf);
	free((char *)m->mbox);
	Mbox[m->mbnum] = NULLMBX;
	free((char *)m);
    BbsUsers--;

}

/**********************************************************************/

static int dochat __ARGS((int argc,char *argv[],void *p));
static int dombconnect __ARGS((int argc,char *argv[],void *p));
static int dombcall __ARGS((int argc,char *argv[],void *p));
static int dodownload __ARGS((int argc,char *argv[],void *p));
static int dombupload __ARGS((int argc,char *argv[],void *p));
static int dowhat __ARGS((int argc,char *argv[],void *p));
static int dozap __ARGS((int argc,char *argv[],void *p));
static int dosend __ARGS((int argc,char *argv[],void *p));
static int dosid __ARGS((int argc,char *argv[],void *p));
static int dosysop __ARGS((int argc,char *argv[],void *p));
static int dostars __ARGS((int argc,char *argv[],void *p));
static int dombhelp __ARGS((int argc,char *argv[],void *p));
static int dombtelnet __ARGS((int argc,char *argv[],void *p));
static int dombfinger __ARGS((int argc,char *argv[],void *p));
static int dombexpert __ARGS((int argc,char *argv[],void *p));
int dombmovemail __ARGS((int argc,char *argv[],void *p));
static void gw_alarm __ARGS((void *p));
static void gw_input __ARGS((int s,void *notused,void *p));
static void gw_superv __ARGS((int null,void *proc,void *p));
static int mbx_to __ARGS((int argc,char *argv[],void *p));
static int mbx_data __ARGS((struct mbx *m,struct list *cclist,char *extra));
static int msgidcheck __ARGS((char *string));
static int uuencode __ARGS((FILE *infile,int s,char *infilename));
static int thirdparty __ARGS((struct mbx *m));
#ifdef NETROM
static int dombnrnodes __ARGS((int argc,char *argv[],void *p));
extern int donrneighbour __ARGS((int argc,char *argv[],void *p));
#endif

static struct cmds Mbcmds[] = {
	"",		doreadnext,	0, 0, NULLCHAR,
	"?",		dombhelp,	0, 0, NULLCHAR,
	"area",		doarea,		0, 0, NULLCHAR,
	"bye",		domboxbye,	0, 0, NULLCHAR,
    "connect",  dombconnect,0, 0, NULLCHAR,
#ifdef CALLBOOK
	"call",     dombcall,   0, 0, NULLCHAR,
#endif
    "download", dodownload, 0, 2, "D[U] filename",
	"escape",	dombescape,	0, 0, NULLCHAR,
	"finger",	dombfinger,	0, 0, NULLCHAR,
	"help",		dombhelp,	0, 0, NULLCHAR,
	"info",		dombhelp,	0, 0, NULLCHAR,
#ifdef	AX25
	"jheard",	doaxheard,	0, 0, NULLCHAR,
#endif
    "kill",     dodelmsg,   0, 0, NULLCHAR,
    "list",     dolistnotes,0, 0, NULLCHAR,
	"mboxusers",dombusers,  0, 0, NULLCHAR,
#ifdef NETROM
    "nodes",    dombnrnodes,0, 0, NULLCHAR,
    "nroutes",  donrneighbour, 0, 0, NULLCHAR,
#endif
    "operator", dochat,     0, 0, NULLCHAR,
	"ports",    dombports,  0, 0, NULLCHAR,
#ifdef CALLCLI
    "query", dombcallbook, 0, 2, "Q callsign\nMultiple callsigns allowed per line",
#endif
    "read",     doreadmsg,  0, 0, NULLCHAR,
	"send",		dosend,		0, 0, NULLCHAR,
	"telnet",	dombtelnet,	0, 2, "T hostname",
	"upload",   dombupload, 0, 0, NULLCHAR,
    "verbose",  doreadmsg,  0, 0, NULLCHAR,
    "what",     dowhat,     0, 0, NULLCHAR,
    "xpert",    dombexpert, 0, 0, NULLCHAR,
    "zap",      dozap,      0, 2, "Z filename",
	"[",		dosid,		0, 0, NULLCHAR,
#ifdef	AX25
#ifdef MBFWD
    "f>",       dorevfwd,   0, 0, NULLCHAR,
#endif
#endif
	"@",		dosysop,	0, 0, NULLCHAR,
	"***",		dostars,	0, 0, NULLCHAR,
	NULLCHAR,	NULLFP,		0, 0, "Huh?",
};

/* "twocmds" defines the MBL/RLI two-letter commands, eg. "SB", "SP" and so on.
 * They have to be treated specially since cmdparse() wants a space between
 * the actual command and its arguments.
 * "SP FOO" is converted to "s  foo" and the second command letter is saved
 * in m->stype. Longer commands like "SEND" are unaffected, except for
 * commands starting with "[", i.e. the SID, since we don't know what it will
 * look like.
 */
static char twocmds[] = "aslrd[mvkx";    /* S,L,R,D,M,V,K,X are two-letter commands */

int
mbx_parse(m)
struct mbx *m;
{
	char *cp;
	int i;
	char *newargv[2];
	/* Translate entire buffer to lower case */
	for (cp = m->line; *cp != '\0'; ++cp)
		if(isupper(*cp))
			*cp = tolower(*cp);
	/* Skip any spaces at the begining */
	for(cp = m->line;isspace(*cp);++cp)
		;
	m->stype = ' ';
	if(*cp != '\0' && *(cp+1) != '\0')
	for(i=0; i<strlen(twocmds); ++i){
		if(*cp == twocmds[i] && (isspace(*(cp+2)) || *(cp+2) == '\0'
		 || *cp == '[')){
			if(islower(*(++cp)))
				m->stype = toupper(*cp); /* Save the second character */
			else
				m->stype = *cp;
			*cp = ' ';
			break;
		}
	}
	/* See if the input line consists solely of digits */
	cp = m->line;
	for(cp = m->line;isspace(*cp);++cp)
		;
	newargv[1] = cp;
	for(;*cp != '\0' && isdigit(*cp);++cp)
		;
	if(*cp == '\0' && strlen(newargv[1]) > 0) {
        newargv[0] = "r";
        return doreadmsg(2,newargv,(void *)m);
	}
	else
		return cmdparse(Mbcmds,m->line,(void *)m);
}

/* This works like recvline(), but telnet options are answered and the
 * terminating newline character is not put into the buffer. If the
 * incoming character equals the value of escape, any queued input is
 * flushed and -2 returned.
 */
int
mbxrecvline(m)
struct mbx *m;
{
    int s = m->user;
    int escape = m->escape;
    char *buf = m->line;
	int c, cnt = 0, opt;

	if(buf == NULLCHAR)
		return 0;
	usflush(Curproc->output);
#ifdef MBXTDISC
    start_timer(&m->tdisc);
#endif
	while((c = recvchar(s)) != EOF){
		if(c == IAC){		/* Telnet command escape */
			if((c = recvchar(s)) == EOF)
				break;
			if(c > 250 && c < 255 && (opt = recvchar(s)) != EOF){
#ifdef	foo
				switch(c){
				case WILL:
					tprintf("%c%c%c",IAC,DONT,opt);
					break;
				case WONT:
					tprintf("%c%c%c",IAC,DONT,opt);
					break;
				case DO:
					tprintf("%c%c%c",IAC,WONT,opt);
					break;
				case DONT:
					tprintf("%c%c%c",IAC,WONT,opt);
				}
#endif
/* to be fixed 			usflush(Curproc->output);*/
				continue;
			}
			if(c != IAC && (c = recvchar(s)) == EOF)
				break;
		}
		/* ordinary character */
		if(c == '\r' || c == '\n')
			break;
		if(uchar(c) == escape){
			if(socklen(s,0)) /* discard any remaining input */
				recv_mbuf(s,NULL,0,NULLCHAR,0);
			cnt = -2;
			break;
		}
        /* Handle <del> chars - from wa7tas */
        if(c == 8) {
            if(cnt > 0) {
                *--buf = 0;
                cnt--;
            }
        } else {
            *buf++ = c;
            ++cnt;
        }
        if(cnt == MBXLINE - 1)
			break;
	}
	if(c == EOF && cnt == 0)
		return -1;
	*buf = '\0';
#ifdef MBXTDISC
    /* Restart the timeout-timer - WG7J*/
    start_timer(&m->tdisc);
#endif
    return cnt;
}

extern void updatedefaults(struct mbx *);

int
domboxbye(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;

	m = (struct mbx *)p;

    /* for bbs's, just disconnect */
	if(m->sid & MBX_SID)
		return -2;

#ifdef USERLOG
    setlastread(m);
    if(m->update)
        updatedefaults(m);
#endif
	/* Now say goodbye */
    tprintf("\nThank you %s,\nfor calling the %s NOS Tcp/Ip Mailbox.\n\n",m->name,
		Hostname);
#ifdef ALLSERV
	if(m->type == TIP)
        tputs("Please hang up now.\n");
#endif
	usflush(m->user);
	return -2;	/* signal that exitbbs() should be called */
}

static int
dombhelp(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	char buf[255];
	int i;
	FILE *fp;
	if(*argv[0] == '?') {
        tputs(Longmenu);
		return 0;
	}
	buf[0] = '\0';
	if(argc > 1)
		for(i=0; Mbcmds[i].name != NULLCHAR; ++i)
			if(!strncmp(Mbcmds[i].name,argv[1],strlen(argv[1]))) {
				sprintf(buf,"%s/%s.hlp",Helpdir,Mbcmds[i].name);
				break;
			}
	if(buf[0] == '\0')
        if(*argv[0] == 'i') {
            /* INFO command */
            tprintf(Nosversion,Version);
#if !defined(OS2)
            tprintf("coreleft: %lu bytes.  Uptime: %s.  ",\
                    farcoreleft(),tformat(secclock()),BbsUsers);
#else
            tprintf("Uptime: %s.  ",tformat(secclock()),BbsUsers);
#endif
            tprintf(CurUsers,BbsUsers);
            sprintf(buf,"%s/info.hlp",Helpdir);
        } else
			sprintf(buf,"%s/help.hlp",Helpdir);
	if((fp = fopen(buf,READ_TEXT)) != NULLFILE) {
		sendfile(fp,Curproc->output,ASCII_TYPE,0);
		fclose(fp);
    } else {
        if(*argv[0]!='i')
            tputs("No help available.\n");
    }
    return 0;
}

extern char Badmsg[];
extern char Nomail[];

/* Move messages from current area to another */
int
dombmovemail(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    int num,i;
    int move[NARG];
    char *to;
    struct mbx *m;
    char buf[MBXLINE];
    char *cp;
    FILE * Mfile;           /* file to move to */
    struct let *cmsg;
    long size;
    int thisone;

    m = (struct mbx *)p;

    if(argc == 1) {
        tputs("Syntax: MM area - moves current message\n"
                "        MM n1 [n2...] area - move message n1 (n2...)\n");
        return 0;
    }
    if(m->mfile == NULLFILE) {
        tputs(Nomail);
        return 0;
    }
    if(argc == 2) {
        /* NO message #, use current message */
        num = 1;
        to = argv[1];
        move[0] = m->current;
    } else {
        /* Scan all message # to move */
        num=0;
        for(i=1;i<argc-1;i++) {
            if( (move[num]=atoi(argv[i])) != 0 )
                if(move[num] <= m->nmsgs)
                    num++;
            else
                tprintf(Badmsg,argv[i]);
        }
        to = argv[argc-1];
    }
#ifdef notdef
    /* translate '.' and '\' to '/' */
    for(cp=to;*cp!='\0';cp++)
        if((*cp == '.') || (*cp == '\\'))
            *cp = '/';
#endif

    /* Now try to lock the destination file */
    if(mlock(Mailspool,to) == -1) {
        tprintf("Can't lock '%s', please try later\n",to);
        return 0;
    }
    /* open the destination file for appending */
    sprintf(buf,"%s/%s.txt",Mailspool,to);
    if((Mfile=fopen(buf,"a")) == NULLFILE) {
        tprintf("Can't open '%s'\n",buf);
        rmlock(Mailspool,to);
        return 0;
    }
    /* Okay, let's do the work */
    for(i=0;i<num;i++) {
        thisone = move[i];
        cmsg = &m->mbox[thisone];
        /* find start of this message */
        fseek(m->mfile,cmsg->start,0);
        size = cmsg->size;
        /* now read this message */
        while(size > 0 && fgets(buf,MBXLINE,m->mfile)!= NULLCHAR) {
            fputs(buf,Mfile);
            size -= strlen(buf);
        }
        /* delete this message */
        cmsg->status |= BM_DELETE;
        m->change = 1;
        tprintf("Message %d moved...\n",thisone);
    }
    fclose(Mfile);
    rmlock(Mailspool,to);
    return 0;
}

static int
dombexpert(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    struct mbx *m;

    m = (struct mbx *)p;

    switch(m->stype) {
    case 'M':
        if(argc == 1)
            tprintf("-more- after %d lines\n",m->morerows);
        else {
            m->morerows = atoi(argv[1]);
            m->update = 1;
        }
        break;
    case 'A':
        m->sid ^= MBX_AREA;
        m->update = 1;
        break;
    case 'N':
        m->sid ^= MBX_NRID;
        m->update = 1;
        break;
    default:
        m->sid ^= MBX_EXPERT;
        m->update = 1;
        break;
    }
    return 0;
}


static int
dochat(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	char buf[8], *newargv[3];
    struct mbx *m;

    m = (struct mbx *)p;

	if(MAttended){
        m->state = MBX_CHAT;
        newargv[0] = "C";
		newargv[1] = Hostname;
		sprintf(buf,"%d",IPPORT_TTYLINK);
		newargv[2] = buf;
		return dombtelnet(3,newargv,p);
	}
	else {
        tputs("Sorry - the system is unattended.\007\n");
	}
	/* It returns only after a disconnect or refusal */
	return 0;
}

#ifdef CALLBOOK
extern char *InetCallserver;
extern char *InetCallserverport;

/*This is a simple way of estiblishing a connection
 *to a callbook-server via tcp - WG7J
 */
static int
dombcall(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	char buf[8], *newargv[3];

    if(InetCallserver!=NULLCHAR){
        newargv[0] = "C";
        newargv[1] = InetCallserver;
        newargv[2] = InetCallserverport;
		return dombtelnet(3,newargv,p);
	}
	else {
        tputs("Internet callserver not configured\n");
	}
	/* It returns only after a disconnect or refusal */
	return 0;
}

#endif /*CALLBOOK*/

#ifdef CALLCLI
static int
dombcallbook(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
  char buf[8], *newargv[3];
  extern char *Callserver;
  int req, ret = 0;

  m = (struct mbx *) p;
/*   tprintf("Looking up - %s - in the callbook at %s\n",argv[1],Callserver);*/
  sprintf(buf,"%d",IPPORT_CALLDB);
  newargv[0] = "C";
  newargv[1] = Callserver;
  newargv[2] = buf;

  for (req = 1; req < argc; req++)  {
  if (argv[req] == NULLCHAR) return ret;
  m->startmsg = mallocw(80);  /* is freed each time by gw_connect()         */
  sprintf(m->startmsg,"%s\n", argv[req]);
  log(m->user, "%s checking callbook: %s",m->name,argv[req]);
  tprintf("Looking for \" %s \" in the callbook at %s\n",argv[req],Callserver);
  ret = dombtelnet(3,newargv,p);
  }
  return ret; /* It looks like all possible returns are zero anyway!    */
}
#endif

/*Some additional security - WG7J
 *NO_3PARTY =  disallow all 3rd party mail
 *NO_SENDCMD = only allow mail to sysop
 */
static int
dosend(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	int cccnt = 0, fail = 0;
	char *host, *cp, fullfrom[MBXLINE], sigwork[LINELEN], *rhdr = NULLCHAR;
	struct list *ap, *cclist = NULLLIST;
	struct mbx *m;
	FILE *fp;
    time_t now;
    int done = 0;
    char *cp2;

#ifdef RLINE
    struct tm t;
#define ODLEN   16
#define OBLEN   32
    char tmpline[MBXLINE];
    char origdate[ODLEN] ="";
    char origbbs[OBLEN] = "";
    char fwdbbs[NUMFWDBBS][FWDBBSLEN+1];
    int myfwds = 0;
    int i;
    int zulu;
    int check_r = 0;
    int found_r = 0;
    int put_r = 1;
#endif

	m = (struct mbx *)p;

    if((m->stype != 'R' || (m->sid & MBX_SID)) && mbx_to(argc,argv,m)
	   == -1){
        tputs((m->sid & MBX_SID) ? "NO - syntax error\n" : \
                    "S command syntax error - format is:\n" \
                    "S[C|F] name [@ host] [< from_addr] [$bulletin_id]\n" \
                    "SR [number]\n");
		return 0;
	}

    /*Check for send permission */
    if(m->privs & NO_SENDCMD) { /*is this to 'SYSOP' or 'sysop' ?*/
        if(stricmp(m->to,"sysop")) {
            tputs((m->sid & MBX_SID) ? "NO - permission denied\n" : \
                    "Sorry, only mail to 'sysop' allowed!\n");
            return 0;
        }
    }

    if(m->stype != 'R' && msgidcheck(m->tomsgid)) {
		if(m->sid & MBX_SID)
			tputs("NO - ");
		tprintf("Already have %s\n",m->tomsgid);
		return 0;
	}
    if(m->stype == 'R' && !(m->sid & MBX_SID) &&
	   mbx_reply(argc,argv,m,&cclist,&rhdr) == -1)
		return 0;
	if((cp = rewrite_address(m->to)) != NULLCHAR)
	     if(strcmp(m->to,cp) != 0){
		  m->origto = m->to;
		  m->to = cp;
	     }
	     else
		  free(cp);

    /* refuse any mail that gets rewritten into 'refuse' - WG7J */
    if(!strcmp(m->to,"refuse")) {
        tputs((m->sid & MBX_SID) ? "NO - refused\n" : \
                "Bad user or hostname,  please mail 'sysop' for help\n");
        return 0;
    }

    if( (!ThirdParty && !(m->privs & SYSOP_CMD)) || (m->privs & NO_3PARTY) )
		if(thirdparty(m) == 0){
            tputs(Mbwarning);
			return 0;
		}

    /* Send the new 'To:' line to sysops only - WG7J */
    if((m->privs&SYSOP_CMD) && (m->origto != NULLCHAR || m->stype == 'R') \
         && !(m->sid & MBX_SID))
            tprintf("To: %s\n", m->to);
	if(validate_address(m->to) == 0){
        tputs((m->sid & MBX_SID) ? "NO - bad address\n" : \
                "Bad user or hostname,  please mail 'sysop' for help\n");
        free(rhdr);
		del_list(cclist);
		/* We don't free any more buffers here. They are freed upon
		 * the next call to mbx_to() or to domboxbye()
		 */
		return 0;
	}
	/* Display the Cc: line (during SR command) */
	for(ap = cclist; ap != NULLLIST; ap = ap->next) {
		if(cccnt == 0){
			tprintf("%s",Hdrs[CC]);
			cccnt = 4;
		}
		else {
			tputs(", ");
			cccnt += 2;
		}
		if(cccnt + strlen(ap->val) > 80 - 3) {
			tputs("\n    ");
			cccnt = 4;
		}
		tputs(ap->val);
		cccnt += strlen(ap->val);
	}
	if(cccnt)
		tputc('\n');

    /* If the the command was 'SC' then read the Cc: list now - WG7J */
    if((m->stype == 'C') && !(m->sid & MBX_SID)) {
        m->stype = 'P'; /* make everything private */
        tputs(CcLine);
        if(mbxrecvline(m) != -1) {
            if(strlen(m->line)) {
                if(*m->line == 0x01) { /* CTRL-A, abort */
                    free(rhdr);
                    del_list(cclist);
                    tputs("Aborted.\n");
                    return 0;
                }
                cp = m->line;
                /* get all the Cc addresses, separated by comma's */
                while((cp2=strchr(cp,',')) != NULLCHAR) {
                    *cp2 = '\0';
                    /*get rid of leading spaces or tabs*/
                    while(*cp == ' ' || *cp == '\t')
                        cp++;
                    if(strlen(cp))
                        addlist(&cclist,cp,0);
                    cp = cp2 + 1;
                }
                /* Do the last or only one */
                /* get rid of leading spaces or tabs*/
                while(*cp == ' ' || *cp == '\t')
                    cp++;
                if(strlen(cp))
                    addlist(&cclist,cp,0);
            }
        } else {
            free(rhdr);
            del_list(cclist);
            return 0;
        }
    }

    /* Now check to make sure we can create the needed tempfiles - WG7J */
    if((m->tfile = tmpfile()) == NULLFILE) {
        free(rhdr);
        del_list(cclist);
        tputs((m->sid & MBX_SID) ? "NO - no temp file\n" : \
                "Can't create temp file for mail\n");
        return 0;
    }
#ifdef RLINE
    /* Only accept R: lines from bbs's */
    if((Rdate || Rreturn || Rfwdcheck) && (m->sid & MBX_SID)){
        /* Going to interpret R:headers,
         * we need another tempfile !
         */
        if((m->tfp = tmpfile()) == NULLFILE) {
            free(rhdr);
            del_list(cclist);
            tputs("NO - no temp file\n");
            return 0;
        }
        /* Now we got enough :-) */
        check_r = 1;
        Checklock++;
    }
#endif

	m->state = MBX_SUBJ;
    if(m->stype != 'R' || (m->sid & MBX_SID)) {
        tputs((m->sid & MBX_SID) ? "OK\n" : "Subject:\n");
        if(mbxrecvline(m) == -1) {
#ifdef RLINE
            if(check_r) {
                MYFCLOSE(m->tfp);
                Checklock--;
            }
#endif
            return 0;
        }
	}
	else				/* Replying to a message */
		tprintf("Subject: %s\n",m->line);

    m->subject = strdup(m->line);

#ifdef RLINE
    if(!check_r) {
#endif
        mbx_data(m,cclist,rhdr);
        /*Finish smtp headers*/
        fprintf(m->tfile,"\n");
#ifdef RLINE
    }
#endif
    free(rhdr);
	m->state = MBX_DATA;
    if(!(m->sid & MBX_SID) && m->stype != 'F')
		tprintf("Enter message.  %s",Howtoend);

    if(m->stype != 'F' || (m->sid & MBX_SID) != 0) {
        while(mbxrecvline(m) != -1){
			if(m->line[0] == 0x01){  /* CTRL-A */
                MYFCLOSE(m->tfile);
#ifdef RLINE
                if(check_r)
                    MYFCLOSE(m->tfp);
#endif
                tputs("Aborted.\n");
				del_list(cclist);
				return 0;
			}
            if(m->line[0] != CTLZ && stricmp(m->line, "/ex")) {
#ifdef RLINE
                if(check_r) {
                    /* Check for R: lines to start with */
                    if(!strncmp(m->line,"R:",2)) { /*found one*/
                        found_r = 1;
                        /*Write this line to the second tempfile
                         *for later rewriting to the real one
                         */
                        fprintf(m->tfp,"%s\n",m->line);
                        /* Find the '@[:]CALL.STATE.COUNTRY'or
                         * or the '?[:]CALL.STATE.COUNTRY' string
                         * The : is optional.
                         */
                        if( ((cp=strchr(m->line,'@')) != NULLCHAR) ||
                            ((cp=strchr(m->line,'?')) != NULLCHAR) ) {
                            if((cp2=strchr(cp,' ')) != NULLCHAR)
                                *cp2 = '\0';
                            if((cp2=strchr(cp,'\n')) != NULLCHAR)
                                *cp2 = '\0';
                            if((cp2=strchr(cp,'\t')) != NULLCHAR)
                                *cp2 = '\0';
                            /* Some bbs's send @bbs instead of @:bbs*/
                            if (*++cp == ':')
                                cp++;
                            /* if we use 'return addres'
                             * copy whole 'domain' name
                             */
                            if(Rreturn)
                                if(strlen(cp) <= OBLEN)
                                    strcpy(origbbs,cp);
                            /* Optimize forwarding ? */
                            if(Rfwdcheck) {
                                /*if there is a HADDRESS, cut off after '.'*/
                                if((cp2=strchr(cp,'.')) != NULLCHAR)
                                    *cp2 = '\0';
                                /*cross-check with MyFwds list*/
                                for(i=0;i<Numfwds;i++) {
                                    if(!strcmp(MyFwds[i],cp)) {
                                        /*Found one !*/
                                        strcpy(fwdbbs[myfwds++],cp);
                                        break;
                                    }
                                }
                            }
                        }
                        if(Rdate) {
                            /* Find the 'R:yymmdd/hhmmz' string */
                            if((cp=strchr(m->line,' ')) != NULLCHAR) {
                                *cp = '\0';
                                if(strlen(m->line+2) <= ODLEN)
                                    strcpy(origdate,m->line+2);
                            }
                        }
                    } else {
                        /* The previous line was last R: line
                         * so we're done checking
                         * now write the smtp headers and
                         * all saved R: lines to the right tempfile
                         */
                        check_r = 0;
                        Checklock--;
                        /*Did we actually find one ?*/
                        if(found_r) {
                            if(Rreturn)
#if !defined(OS2)
                                m->origbbs = strdup(strlwr(origbbs));
#else
                                m->origbbs = strdup((origbbs));
#endif
                            if(Rdate) {
                                if((cp=strchr(origdate,'/')) != NULLCHAR) {
                                    if((*(cp+5) == 'z') || (*(cp+5) == 'Z')) {
                                        *(cp+5) = '\0';
                                        zulu = 1;
                                    }
                                    t.tm_min = atoi(cp+3);
                                    *(cp+3) = '\0';
                                    t.tm_hour = atoi(cp+1);
                                    *cp = '\0';
                                    t.tm_mday = atoi(&origdate[4]);
                                    origdate[4] = '\0';
                                    t.tm_mon = (atoi(&origdate[2]) - 1);
                                    origdate[2] = '\0';
                                    t.tm_year = atoi(origdate);
                                    /* Set the date in rfc 822 format */
                                    m->date = mallocw(40);
                                    sprintf(m->date,"%.2d %s %02d %02d:%02d:00 %.3s\n",
                                        t.tm_mday,
                                        Months[t.tm_mon],
                                        t.tm_year,
                                        t.tm_hour,
                                        t.tm_min,
                                        zulu ? "UTC" : "");
                                }
                            }
                        }
                        /* Now write the headers,
                         * possibly adding Xforwarded lines for bulletins,
                         * or anything that has a BID.
                         * Add the X-Forwarded lines FIRST,
                         * this speeds up forwarding...
                         */
                        if(Rfwdcheck && found_r && \
                           ((m->stype == 'B') || (m->tomsgid)) ){
                            /*write Xforwarded headers*/
                            for(i=0;i<myfwds;i++) {
                                fprintf(m->tfile,"%s%s\n",Hdrs[XFORWARD],fwdbbs[i]);
                            }
                        }
                        /*write regular headers*/
                        mbx_data(m,cclist,rhdr);
                        /* Finish smtp headers */
                        fprintf(m->tfile,"\n");

                        /* Now copy the R: lines back */
                        if(found_r) {
                            rewind(m->tfp);
                            while(fgets(tmpline,sizeof(tmpline),m->tfp)!=NULLCHAR)
                                fputs(tmpline,m->tfile);
                        }
                        MYFCLOSE(m->tfp);

                        /* And add this first non-R: line */
                        fprintf(m->tfile,"%s\n",m->line);
                    }
                } else
#endif
				fprintf(m->tfile,"%s\n",m->line);
            } else {
#ifdef RLINE
                if(check_r) {
                    /* Hmm, this means we never finished the R: headers
                     * tmp file still open !
                     */
                    MYFCLOSE(m->tfp);
                }
#endif
                done = 1; /* To indicate the difference between
                           * mbxrecvline() returning -1 and /ex ! - WG7J
                           */
                break;  /* all done */
            }
		}
        if(!done) {
            /* We did NOT get ^Z or /EX, but mbxrecvline returned -1 !!!
             * This means the connection is gone ! - WG7J
             */
            MYFCLOSE(m->tfile);
#ifdef RLINE
            if(check_r)
                MYFCLOSE(m->tfp);
#endif
            del_list(cclist);
            free(rhdr); /*Just in case*/
            return 0;
        }
    } else {
		fprintf(m->tfile,"----- Forwarded message -----\n\n");
		msgtofile(m,m->current,m->tfile,0);
		fprintf(m->tfile,"----- End of forwarded message -----\n");
	}

	/* Insert customised signature if one is found */
	if(!(m->sid & MBX_SID)) {	/* not a forwarding BBS */
	     sprintf(sigwork,"%s/%s.sig",Signature,
		     m->tofrom ? m->tofrom : m->name);
	     if((fp = fopen(sigwork,READ_TEXT)) != NULLFILE){
		  while(fgets(sigwork,LINELEN,fp) != NULLCHAR)
			fputs(sigwork,m->tfile);
		  fclose(fp);
	     }
	}

	if((host = strrchr(m->to,'@')) == NULLCHAR) {
		host = Hostname;	/* use our hostname */
		if(m->origto != NULLCHAR) {
			/* rewrite_address() will be called again by our
			 * SMTP server, so revert to the original address.
			 */
		 	free(m->to);
			m->to = m->origto;
			m->origto = NULLCHAR;
		}
	}
	else
		host++;	/* use the host part of address */

	/* make up full from name for work file */
	if(m->tofrom != NULLCHAR)
		sprintf(fullfrom,"%s%%%s@%s",m->tofrom, m->name, Hostname);
	else
		sprintf(fullfrom,"%s@%s",m->name,Hostname);
	if(cclist != NULLLIST && stricmp(host,Hostname) != 0) {
		fseek(m->tfile,0L,0);	/* reset to beginning */
		fail = queuejob(m->tfile,Hostname,cclist,fullfrom);
		del_list(cclist);
		cclist = NULLLIST;
	}
	addlist(&cclist,m->to,0);
	fseek(m->tfile,0L,0);
	fail += queuejob(m->tfile,host,cclist,fullfrom);
	del_list(cclist);
    MYFCLOSE(m->tfile);
    if(fail){
        if(!(m->sid & MBX_SID)) /* only when we're not a bbs */
            tputs("Couldn't queue message for delivery\n");
    } else {
        if(!(m->sid & MBX_SID)) /* only when we're not a bbs */
            tputs("Message queued\n");
        if(m->sid & MBX_SID)
            MbRecvd++;
        else
            MbSent++;
        /* Timestamp added to allow automatic expiry of bid file - WG7J */
        time(&now);
        if(m->tomsgid != NULLCHAR && \
          (fp = fopen(Historyfile,APPEND_TEXT)) != NULLFILE) {
          fprintf(fp,"%s %ld\n",m->tomsgid,now); /* Save BID in history file */
		  fclose(fp);
	     }
    }
    smtptick(NULL);     /* wake SMTP to send that mail */
	return 0;
}

#ifdef NETROM
static int
dombnrnodes(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;

	m = (struct mbx *)p;

	if(argc < 2)
		return doroutedump();
    if(*argv[1] == '*')
        argc = 1;
    return dorouteinfo(argc,argv,p);
}
#endif

static int
dosid(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
    char *cp;

	m = (struct mbx *)p;
	if(argc == 1)
		return 1;
    if(argv[1][strlen(argv[1]) - 1] != ']') /* must be an SID */
		return 1;
#ifdef notdef
    if(m->stype == 'Z' && strncmp(argv[1],"cz",2) == 0) {
        /* LAN-LINK's [ZCZ] */
        m->sid |= MBX_LL;
        return 0;
    }
#endif
    /* Other bbs's */
    m->sid = MBX_SID;

    /* Now check to see if this is an RLI board.
	 * As usual, Hank does it a bit differently from
	 * the rest of the world.
	 */
	if(m->stype == 'R' && strncmp(argv[1],"li",2) == 0)/* [RLI] at a minimum */
		m->sid |= MBX_RLI_SID;
	/* Check to see if the BBS supports a kludge called "hierarchical
	 * routing designators."
	 *
	 * No need to check for ']' -- it must be there or this is not
	 * a valid mbox id -- it is checked earlier (fix de OH3LKU)
     *
     * Sid format is [BBSTYPE-VERSION-OPTIONS]
     * check for LAST -, to allow for version portion. - WG7J
     */
    if((cp = strrchr(argv[1],'-')) != NULLCHAR
        && (cp=strchr(cp+1,'h')) != NULLCHAR
        && strchr(cp+1,'$'))
            m->sid |= MBX_HIER_SID;
	return 0;
}

int
dombescape(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;

	m = (struct mbx *)p;
	if(argc < 2){
        tputs("The escape character is: ");
		if(m->escape < 32)
			tprintf("CTRL-%c\n",m->escape+'A'-1);
		else
			tprintf("'%c'\n",m->escape);
		return 0;
	}
	if(strlen(argv[1]) > 1)
		if(isdigit(*argv[1]))
			m->escape = (char) atoi(argv[1]);
		else
			return 1;
	else
		m->escape = *argv[1];
	return 0;
}

static int
dodownload(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	FILE *fp;
	char *file;

	m = (struct mbx *)p;
	file = pathname(m->path,argv[1]);
	if(!permcheck(m->path,m->privs,RETR_CMD,file)){
        tputs(Noperm);
		return 0;
	}
    m->state = MBX_DOWNLOAD;
    if((fp = fopen(file,READ_TEXT)) == NULLFILE)
		tprintf("Can't open \"%s\": %s\n",file,strerror(errno));
	else
		if(m->stype == 'U'){			/* uuencode ? */
			fclose(fp);
			fp = fopen(file,READ_BINARY);	/* assume non-ascii */
			uuencode(fp,m->user,file);
		} else
			sendfile(fp,m->user,ASCII_TYPE,0);
	free(file);
	fclose(fp);
	return 0;
}

static int
dombupload(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	FILE *fp;
    char *file;

	m = (struct mbx *)p;
	file = pathname(m->path,argv[1]);
	if(!permcheck(m->path,m->privs,STOR_CMD,file)){
        tputs(Noperm);
		return 0;
	}
	if((fp = fopen(file,WRITE_TEXT)) == NULLFILE){
		tprintf("Can't create \"%s\": %s\n",file,strerror(errno));
		free(file);
		return 0;
	}
	log(m->user,"MBOX upload: %s",file);
    m->state = MBX_UPLOAD;
    tprintf("Send file,  %s",Howtoend);
	for(;;){
        if(mbxrecvline(m) == -1){
			remove(file);
			break;
		}
        if(*m->line == 0x01){  /* CTRL-A */
			remove(file);
            tputs("Aborted.\n");
			break;
		}
        if(*m->line == CTLZ || !stricmp("/ex",m->line))
			break;
        fputs(m->line,fp);
#if !defined(UNIX) && !defined(__TURBOC__) && !defined(AMIGA)
		/* Needed only if the OS uses a CR/LF
		 * convention and putc doesn't do
		 * an automatic translation
		 */
		if(putc('\r',fp) == EOF)
			break;
#endif
		if(putc('\n',fp) == EOF)
			break;
	}
	free(file);
	fclose(fp);
	return 0;
}

static int
dowhat(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	FILE *fp;
	char *file;

	m = (struct mbx *)p;
	if(argc < 2)
		file = strdup(m->path);
	else
		file = pathname(m->path,argv[1]);
	if(!permcheck(m->path,m->privs,RETR_CMD,file)){
        tputs(Noperm);
		return 0;
	}
    m->state = MBX_WHAT;
	if((fp = dir(file,1)) == NULLFILE)
		tprintf("Can't read directory: \"%s\": %s\n",file,strerror(errno));
	else
		sendfile(fp,m->user,ASCII_TYPE,0);
	free(file);
	fclose(fp);
	return 0;
}

static int
dozap(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	char *file;

	m = (struct mbx *)p;
	file = pathname(m->path,argv[1]);
	if(!permcheck(m->path,m->privs,DELE_CMD,file)){
        tputs(Noperm);
		return 0;
	}
	if(remove(file))
		tprintf("Zap failed: %s\n",strerror(errno));
	log(m->user,"MBOX Zap: %s",file);
	free(file);
	return 0;
}

/*Password protection added - 920118, WG7J */
static int
dosysop(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	int c;
    int len,pwdc[5],i,valid=0;
    char *cp;
    extern struct cmds Cmds[];

	m = (struct mbx *) p;
    log(m->user,"MBOX: %s attempting SYSOP",m->name);

    /*If you want anyone with the password to go sysop-mode
     *comment out the next 4 line ! -WG7J
     */
    if(!(m->privs & SYSOP_CMD)){
        tputs(Noperm);
		return 0;
	}

    /*only if set,
     *check for the password before letting users proceed
     */
    m->state = MBX_SYSOPTRY;
    if((len = strlen(Mbpasswd)) != 0) {;
        for (i=0;i<5;i++)
#if !defined(OS2)
            tprintf("%d ",(pwdc[i]=random(len))); /*print the random chars*/
#else
            tprintf("%d ",(pwdc[i]=rand())); /*print the random chars*/
#endif
        tputc('\n');
        while(1) {
            c = mbxrecvline(m);
            if(c == EOF || c == -2)
                return 0;
            if(*m->line == '\0')
                break;
            cp = m->line;
            for(i=0;i<5;i++)
                if(*cp++ != Mbpasswd[pwdc[i]])
                    break;
            if (i == 5)
                valid = 1;
        }
    if(!valid)
        return 0;
    }

    log(m->user,"MBOX: %s is now SYSOP",m->name);
    m->state = MBX_SYSOP;
    tputs("\n\aType 'exit' to return\n\nBe VERY carefull!!\n");

	for(;;){
        tputs("Net> ");
		usflush(Curproc->output);
        if(mbxrecvline(m) < 0)
			break;
		log(m->user,"MBOX sysop: %s",m->line);
        if(cmdparse(Cmds,m->line,NULL) == -2)
            break;
	}
	return 0;
}

/* Handle the "*** Done" command when reverse forwarding ends or the
 * "*** LINKED to" command.
 */
static int
dostars(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	int anony = 1;
	int founddigit = 0;
	int oldprivs;
	char *cp;

	m = (struct mbx *)p;

	/* The "*** LINKED to" command is only allowed to stations with
	 * SYSOP privileges to prevent others from obtaining the same.
	 */
#ifdef notdef
	if((m->privs & SYSOP_CMD) && argc == 4 && !strcmp(argv[1],"linked")) {
#endif
	/* Allow 'linked to' from anyone, but reset SYSOP priviledges
	 * when the sysop-password is not set !
	 * Also try to set the new call !
	 * Inspired by Kurt, wb5bbw
	 * Check for the strange TEXNET linked message !
	 * 920220 - WG7J
	 */
	if( ((argc == 4) || (argc == 9)) && !strcmp(argv[1],"linked")) {
		if(argc == 9)
			/*TEXNET ?*/
			if(strcmp(argv[8],"texnet."))
				return 0;
#ifdef GWTRACE
		log(m->user,"MBOX LINKED: %s changed to %s",m->name,argv[3]);
#endif
        strcpy(m->name,argv[3]);
		oldprivs = m->privs; /*Save this !*/
		/* Try to find the privileges of this user from the userfile */
		if((m->privs = userlogin(m->name,NULLCHAR,&m->path,MBXLINE,
					 &anony)) == -1)
		     if((m->privs = userlogin("bbs",NULLCHAR,&m->path,
				      MBXLINE,&anony)) == -1)
			  if((m->privs = userlogin("anonymous",NULLCHAR,
					   &m->path,MBXLINE,&anony)) == -1){
						m->privs = 0;
						free(m->path);
						m->path = NULLCHAR;
			  }
		if(m->privs & EXCLUDED_CMD)
			return domboxbye(0,NULLCHARP,p);
#ifdef AX25
        /* Set the call */
		for(cp=m->name;*cp != '\0';cp++)
			if(isdigit((int)*cp))
				break;
		if(*cp != '\0')
			founddigit = 1;
		if( (setcall(m->call,m->name) == -1) || (!founddigit) ) {
			m->privs &= ~AX25_CMD;
			m->privs &= ~NETROM_CMD;
		}
#else
        m->privs &= ~AX25_CMD;
        m->privs &= ~NETROM_CMD;
#endif
        /*Kill ssid in name, if any*/
        if((cp=strchr(m->name,'-')) != NULLCHAR)
            *cp = '\0';
        /*Check if sysop password is set,
		 *if not, disallow sysop privs no matter what !
		 */
		if(*Mbpasswd == '\0')
			m->privs &= ~SYSOP_CMD;
		/*Check to see if any of NO_READ,NO_SEND or NO_3PARTY were set,
		 *if so, dis-allow those no matter what
		 *(so that users cannot get priviledges by issuing a ***linked)
		 * 920220 - WG7J
		 */
		if(oldprivs & NO_SENDCMD)
			m->privs |= NO_SENDCMD;
		if(oldprivs & NO_READCMD)
			m->privs |= NO_READCMD;
		if(oldprivs & NO_3PARTY)
			m->privs |= NO_3PARTY;

		tprintf("Oh, hello %s.\n",m->name);
		changearea(m,m->name);
		return 0;
	}
	if(argc > 1 && (m->sid & MBX_SID))	/* "*** Done" or similar */
		return 2;
	return -1;
}

static int
doarea(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	char *cp, *area;
	FILE *fp;
    char buf[MBXLINE];

	m = (struct mbx *) p;
	if(argc < 2){
		area = strdup(m->area);
		while((cp = strchr(area,'/')) != NULLCHAR)
			*cp = '.';
		tprintf("Current message area is: %s\n\n",area);
		free(area);

        tprintf("Available areas are:\n%-15s",m->name);
        if(m->stype == 'F')
            tputs("  Your private mail area\n");
        else
            tputc('\n');
		if((fp = fopen(Arealist,READ_TEXT)) == NULLFILE)
			return 0;
        if(m->stype == 'F')
            sendfile(fp,m->user,ASCII_TYPE,0);
        else {
            /* send only the area names, not the description */
            while(fgets(buf,MBXLINE,fp) != NULLCHAR) {
                if(buf[0] != '#') { /* skip comments */
                    if((cp=strchr(buf,' ')) != NULLCHAR) {
                        *cp++ = '\n';
                        *cp = '\0';
                    }
                    if((cp=strchr(buf,'\t')) != NULLCHAR) {
                        *cp++ = '\n';
                        *cp = '\0';
                    }
                    tputs(buf);
                }
            }
            tputs("\nType AF to get description of areas\n\n");
        }
        fclose(fp);
		return 0;
	}
	if((m->privs & SYSOP_CMD) || strcmp(m->name,argv[1]) == 0){
		/*Private mail-areas*/
		changearea(m,argv[1]);
		if(m->nmsgs){
			if(!strcmp(m->name,m->area))
                tputs("You have ");
			else {
				area = strdup(m->area);
				while((cp = strchr(area,'/')) != NULLCHAR)
					*cp = '.';
				tprintf("%s: ",area);
				free(area);
			}
			tprintf("%d message%s -  %d new.\n", m->nmsgs,
			  m->nmsgs == 1 ? " " : "s ", m->newmsgs);
		}
		return 0;
	}
	if(isarea(argv[1])) {
		/*Public mail areas*/
		changearea(m,argv[1]);
		area = strdup(m->area);
		while((cp = strchr(area,'/')) != NULLCHAR)
			*cp = '.';
#ifdef USERLOG
		tprintf("%s: %d message%s -  %d new\n\n",area,m->nmsgs,
			m->nmsgs == 1 ? " " : "s ", m->newmsgs);
#else
        tprintf("%s: %d message%s.\n", area, m->nmsgs,
          m->nmsgs == 1 ? "" : "s");
#endif
        free(area);
	}
	else
		tprintf("No such message area: %s\n",argv[1]);
	return 0;
}

/* subroutine to do the actual switch from one area to another */
/* USERLOGGING added by WG7J */
void
changearea(m,area)
struct mbx *m;
char *area;
{
	char *cp;

#ifdef USERLOG
    setlastread(m);
#endif
	closenotes(m);
	m->nmsgs = m->newmsgs = m->current = 0;
    strcpy(m->area,area);
	while((cp = strchr(m->area,'.')) != NULLCHAR)
		*cp = '/';
	while((cp = strchr(m->area,'\\')) != NULLCHAR)
		*cp = '/';

#ifdef USERLOG
    /* only read last read message-id if this is not a bbs,
     * current area is a public area and not 'help'
     * or area starts with 'sys'
     */
    if(!(m->sid & MBX_SID))
        if( (strcmp(area,"help") && isarea(area)) || !strncmp(m->area,"sys",3) )
            getlastread(m);
#endif

    scanmail(m);
}

static int
dombtelnet(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	int s, len, i;
	char dsocket[MAXSOCKSIZE];
	struct sockaddr_in fsocket;

	m = (struct mbx *) p;
	fsocket.sin_family = AF_INET;
	if(argc < 3)
		fsocket.sin_port = IPPORT_TELNET;
	else
		fsocket.sin_port = atoip(argv[2]); /* was atoi */

	if((fsocket.sin_addr.s_addr = resolve(argv[1])) == 0){
		tprintf(Badhost,argv[1]);
        /* Free m->starmsg if set ! - WG7J */
        if(m->startmsg != NULLCHAR) {
            free(m->startmsg);
            m->startmsg = NULLCHAR;
        }
        return 0;
	}
	/* Only local telnets are are allowed to the unprivileged user */
    /* If the first letter of the command is 'C', then it was
     * the CALL,QUERY,OPERATOR or CONVERS command !
     * Allow these, no matter what.
     * This way, you can give access to the Internet callserver,
     * but disable all other telnets - WG7J
     */
    if(*argv[0] != 'C') {
        if(!(m->privs & TELNET_CMD) /* && !ismyaddr(fsocket.sin_addr.s_addr)*/ ){
            tputs(Noperm);
            /* Free m->starmsg if set ! - WG7J */
            /* Shouldn't happen here, but just in case */
            if(m->startmsg != NULLCHAR) {
                free(m->startmsg);
                m->startmsg = NULLCHAR;
            }
            return 0;
        }
    }
	if((s = socket(AF_INET,SOCK_STREAM,0)) == -1){
        tputs(Nosock);
        /* Free m->starmsg if set ! - WG7J */
        if(m->startmsg != NULLCHAR) {
            free(m->startmsg);
            m->startmsg = NULLCHAR;
        }
        return 0;
	}
#ifdef GWTRACE
	log(m->user,"MBOX TELNET: %s to %s:%d",m->name,argv[1],fsocket.sin_port);
#endif
	if(fsocket.sin_port == IPPORT_TTYLINK) {
		m->startmsg = mallocw(80);
		len = MAXSOCKSIZE;
		i = getpeername(m->user,dsocket,&len);
		sprintf(m->startmsg,"*** Incoming call from %s@%s ***\n",
			m->name,i != -1 ? psocket(dsocket): Hostname);
	}
    m->state = MBX_GATEWAY;
    return gw_connect(m,s,(struct sockaddr *)&fsocket,SOCKSIZE);
}

static int
dombfinger(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	char *host, *user = NULLCHAR, buf[8], *newargv[3];

	if(argc > 2){
        tputs("Usage: F user@host  or  F @host  or  F user.\n");
		return 0;
	}
	host = Hostname;
	if(argc == 2){
		if((host = strchr(argv[1], '@')) != NULLCHAR){
			*host = '\0';
			host++;
		} else
			host = Hostname;
		user = argv[1];
	}
	m = (struct mbx *) p;
	m->startmsg = mallocw(80);
	if(user != NULLCHAR)
		sprintf(m->startmsg,"%s\n",user);
	else
		strcpy(m->startmsg,"\n");
    newargv[0] = "t";
    newargv[1] = host;
	sprintf(buf,"%d",IPPORT_FINGER);
	newargv[2] = buf;
	return dombtelnet(3,newargv,p);
}

/* Generic mbox gateway code. It sends and frees the contents of m->startmsg
 * when the connection has been established unless it a null pointer.
 */
static int
gw_connect(m,s,fsocket,len)
struct mbx *m;
int s;
struct sockaddr *fsocket;
int len;
{
	int c;
	char *cp, *cp1;
	struct proc *child;
	struct gwalarm *gwa;
    char *node, *tocall, whereto[128], buf[80];
    char temp[AXBUF];
    struct nrroute_tab *rp;

    sockmode(s,SOCK_ASCII);
    child = newproc("gw supervisor",256,gw_superv,0,Curproc,m,0);
    tputs("Trying...  The escape character is: ");
    if(m->escape < 32)
        tprintf("CTRL-%c\n",m->escape+'A'-1);
    else
        tprintf("'%c'\n",m->escape);
	usflush(Curproc->output);

    /*find out where we're going to*/
    tocall = strdup(psocket(fsocket));
    if((cp1 = strchr(tocall,' ')) != NULLCHAR)
		*cp1 = '\0';
#ifdef NETROM
    if(fsocket->sa_family == AF_NETROM) {
        /*find the node alias*/
        setcall(temp,tocall);
        rp = find_nrroute(temp);
        node = strdup(rp->alias);
        if((cp1 = strchr(node,' ')) != NULLCHAR)
                    *cp1 = '\0';
        sprintf(whereto,"%s:%s",node,tocall);
        free(node);
    } else
#endif
	 strcpy(whereto,tocall);
    free(tocall);

    if(connect(s,(char *)fsocket,len) == -1){
		if((cp = sockerr(s)) != NULLCHAR) {
			switch(cp[0]) {
			case 'R':
				sprintf(buf,"%susy from",
					(m->family == AF_NETROM)?"B":"*** b");
				break;
			case 'T':
				if(m->family != AF_NETROM) {
					sprintf(buf,"*** timeout with");
					break;
				}
			default:
				sprintf(buf,"%sailure with",
					(m->family == AF_NETROM)?"F":"*** f");
				break;
			}
            tprintf("%s%s %s\n\n",
                (m->family == AF_NETROM) ? Mbnrid : "",buf,whereto);
		}
        shutdown(s,2);  /* HB9RWM suggestion */
        close_s(s);
        alert(child, EABORT);
//		killproc(child);
        /* Free m->starmsg if set ! - WG7J */
        if(m->startmsg != NULLCHAR) {
            free(m->startmsg);
            m->startmsg = NULLCHAR;
        }

        return 0;
	}
    /* The user did not type the escape character */
//	killproc(child); not yet

	tprintf("%s%sonnected to %s\n",
        (m->family == AF_NETROM) ? Mbnrid : "",
		(m->family == AF_NETROM) ? "C" : "*** c",
        whereto);

	if(m->startmsg != NULLCHAR){
		usputs(s,m->startmsg);
		free(m->startmsg);
		m->startmsg = NULLCHAR;
	}

	/* Since NOS does not flush the output socket after a certain
	 * period of time, we have to arrange that ourselves.
	 */
	gwa = (struct gwalarm *) mallocw(sizeof(struct gwalarm));
	gwa->s1 = Curproc->output;
	gwa->s2 = s;
    set_timer(&gwa->t,2*1000L);
	gwa->t.func = gw_alarm;
	gwa->t.arg = (void *) gwa;
	start_timer(&gwa->t);
	/* Fork off the receive process */
    child = newproc("gw in",1024,gw_input,s,m,Curproc,0);

	for(;;){
		if((c = recvchar(Curproc->input)) == EOF)
			break;
        if (errno == EABORT)
           break;
        if(c == m->escape){
            if(socklen(Curproc->input,0))
                recv_mbuf(Curproc->input,NULL,0,NULLCHAR,0);
			break;
        }
#ifdef MBXTDISC
        if(c == '\n')
            start_timer(&m->tdisc);
#endif
        if(usputc(s,c) == EOF)
			break;
	}
	stop_timer(&gwa->t);
	free((char *)gwa);
	close_s(s);
    alert(child, EABORT);
//	killproc(child); /* get rid of the receive process */
    if(m->family == AF_INET)
		tprintf("%c%c%c\n",IAC,WONT,TN_ECHO);
	return 0;
}

static void
gw_input(s,notused,p)
int s;
void *notused;
void *p;
{
	int c;
	struct proc *parent;
	struct mbx *m;
	char *cp, *cp1;
	char response[4];

	parent = (struct proc *) p;
	m = (struct mbx *) notused;

    cp1 = strdup(Mbnrid);
    if((cp = strchr(cp1,'}')) != NULLCHAR)
		*cp = '\0';
	strupr(cp1);

#ifdef notdef
    while((c = recvchar(s)) != EOF)
		tputc(c);
#endif
    while((c = recvchar(s)) != EOF){

		if(c != IAC){
			tputc((char)c);
			continue;
		}
		/* IAC received, get command sequence */
		c = recvchar(s);
		switch(c){
		case WILL:
			response[0] = IAC;
			response[1] = DONT;
			response[2] = recvchar(s);
			response[3] = '\0';
			usputs(s,response);
			break;
		case WONT:
		case DONT:
			c = recvchar(s);
			break;
		case DO:
			response[0] = IAC;
			response[1] = WONT;
			response[2] = recvchar(s);
			response[3] = '\0';
			usputs(s,response);
			break;
		case IAC:	/* Escaped IAC */
			usputc(s,IAC);
			break;
		}
    }

	if((cp = sockerr(s)) != NULLCHAR && m->family != AF_NETROM) {
		switch(cp[0]) {
		case 'T':
            usprintf(m->user,"\n*** %s: Link failure",cp1);
			break;
		case 'R':
			usprintf(m->user,"*** DM received");
			break;
		}
	}
    usprintf(m->user,"\n%s%seconnected to %s\n\n",
        (m->family == AF_NETROM) ? Mbnrid : "",
		(m->family == AF_NETROM) ? "R" : "*** r",
		cp1);

	free(cp1);
	cp1 = NULLCHAR;

	/* Tell the parent that we are no longer connected */
	alert(parent,ENOTCONN);
//	pwait(Curproc); /* Now wait to be killed */
}

/* Check if the escape character is typed while the parent process is busy
 * doing other things. 
 */
static void
gw_superv(null,proc,p)
int null;
void *proc;
void *p;
{
	struct proc *parent;
	struct mbx *m;
	int c;
	parent = (struct proc *) proc;
	m = (struct mbx *) p;
	while((c = recvchar(Curproc->input)) != EOF)
		if(c == m->escape){
			/* flush anything in the input queue */
			if(socklen(Curproc->input,0))
				recv_mbuf(Curproc->input,NULL,0,NULLCHAR,0);
			break;
		}
	alert(parent,EINTR);	 /* Tell the parent to quit */
//        killself(); not necessary
//	pwait(Curproc);		 /* Please kill me */
}

static void
gw_alarm(p)
void *p;
{
	struct gwalarm *gwa = (struct gwalarm *)p;
	char oldbl;
	struct usock *up;

	/* Flush sockets s1 and s2, but first make sure that the socket
	 * is set to non-blocking mode, to prevent the flush from blocking
	 * if the high water mark has been reached.
	 */
	if((up = itop(gwa->s1)) != NULLUSOCK) {
		oldbl = up->noblock;
		up->noblock = 1;
		usflush(gwa->s1);
		up->noblock = oldbl;
	}
	if((up = itop(gwa->s2)) != NULLUSOCK) {
		oldbl = up->noblock;
		up->noblock = 1;
		usflush(gwa->s2);
		up->noblock = oldbl;
	}
	start_timer(&gwa->t);
}

/* States for send line parser state machine */
#define		LOOK_FOR_USER		2
#define		IN_USER			3
#define		AFTER_USER		4
#define		LOOK_FOR_HOST		5
#define		IN_HOST			6
#define		AFTER_HOST		7
#define		LOOK_FOR_FROM		8
#define		IN_FROM			9
#define		AFTER_FROM		10
#define		LOOK_FOR_MSGID		11
#define		IN_MSGID		12
#define		FINAL_STATE		13
#define		ERROR_STATE		14

/* Prepare the addressee.  If the address is bad, return -1, otherwise
 * return 0
 */
static int
mbx_to(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	register char *cp;
	int state, i;
	char *user, *host, *from, *msgid;
	int userlen = 0, hostlen = 0, fromlen = 0, msgidlen = 0;
	struct mbx *m;

	m = (struct mbx *)p;
	/* Free anything that might be allocated
	 * since the last call to mbx_to() or mbx_reply()
	 */
	free(m->to);
	m->to = NULLCHAR;
	free(m->tofrom);
	m->tofrom = NULLCHAR;
	free(m->tomsgid);
	m->tomsgid = NULLCHAR;
	free(m->origto);
	m->origto = NULLCHAR;
    free(m->origbbs);
    m->origbbs = NULLCHAR;
    free(m->subject);
    m->subject = NULLCHAR;
    free(m->date);
    m->date = NULLCHAR;

	if(argc == 1)
		return -1;
	i = 1;
	cp = argv[i];
	state = LOOK_FOR_USER;
	while(state < FINAL_STATE){
#ifdef MBDEBUG
		tprintf("State is %d, char is %c\n", state, *cp);
#endif
		switch(state){
		case LOOK_FOR_USER:
			if(*cp == '@' || *cp == '<' || *cp == '$'){
				state = ERROR_STATE;		/* no user */
			} else {
				user = cp;			/* point at start */
				userlen++;			/* start counting */
				state = IN_USER;
			}
			break;
		case IN_USER:
			switch(*cp){
			case '\0':
				state = AFTER_USER;		/* done with username */
				break;
			case '@':
				state = LOOK_FOR_HOST;		/* hostname should follow */
				break;
			case '<':
				state = LOOK_FOR_FROM;		/* from name should follow */
				break;
			case '$':
				state = LOOK_FOR_MSGID;	/* message id should follow */
				break;
			default:
				userlen++;			/* part of username */
			}
			break;
		case AFTER_USER:
			switch(*cp){
			case '@':
				state = LOOK_FOR_HOST;		/* hostname follows */
				break;
			case '<':
				state = LOOK_FOR_FROM;		/* fromname follows */
				break;
			case '$':
			state = LOOK_FOR_MSGID;	/* message id follows */
				break;
			default:
				state = ERROR_STATE;
			}
			break;
		case LOOK_FOR_HOST:
			if(*cp == '@' || *cp == '<' || *cp == '$'){
				state = ERROR_STATE;
				break;
			}
			if(*cp == '\0')
				break;
			host = cp;
			hostlen++;
			state = IN_HOST;
			break;
		case IN_HOST:
			switch(*cp){
			case '\0':
				state = AFTER_HOST;		/* found user@host */
				break;
			case '@':
				state = ERROR_STATE;		/* user@host@? */
				break;
			case '<':
				state = LOOK_FOR_FROM;		/* fromname follows */
				break;
			case '$':
				state = LOOK_FOR_MSGID;	/* message id follows */
				break;
			default:
				hostlen++;
			}
			break;
		case AFTER_HOST:
			switch(*cp){
			case '@':
				state = ERROR_STATE;		/* user@host @ */
				break;
			case '<':
				state = LOOK_FOR_FROM;		/* user@host < */
				break;
			case '$':
				state = LOOK_FOR_MSGID;	/* user@host $ */
				break;
			default:
				state = ERROR_STATE;		/* user@host foo */
			}
			break;
		case LOOK_FOR_FROM:
			if(*cp == '@' || *cp == '<' || *cp == '$'){
				state = ERROR_STATE;
				break;
			}
			if(*cp == '\0')
				break;
			from = cp;
			fromlen++;
			state = IN_FROM;
			break;
		case IN_FROM:
			switch(*cp){
			case '\0':
				state = AFTER_FROM;		/* user@host <foo */
				break;
			case '<':
				state = ERROR_STATE;		/* user@host <foo< */
				break;
			case '$':
				state = LOOK_FOR_MSGID;	/* message id follows */
				break;
			default:
				fromlen++;
			}
			break;
		case AFTER_FROM:
			switch(*cp){
			case '@':				/* user@host <foo @ */
			case '<':				/* user@host <foo < */
				state = ERROR_STATE;
				break;
			case '$':
				state = LOOK_FOR_MSGID;	/* user@host <foo $ */
				break;
			default:
				state = ERROR_STATE;		/* user@host foo */
			}
			break;
		case LOOK_FOR_MSGID:
			if(*cp == '\0')
				break;
			msgid = cp;
			msgidlen++;
			state = IN_MSGID;
			break;
		case IN_MSGID:
			if(*cp == '\0')
				state = FINAL_STATE;
			else
				msgidlen++;
			break;
		default:
			/* what are we doing in this state? */
			state = ERROR_STATE;
		}
		if(*(cp) == '\0'){
			++i;
			if(i < argc)
			cp = argv[i];
			else break;
		} else
			++cp;
	}
	if(state == ERROR_STATE || state == LOOK_FOR_HOST
	 || state == LOOK_FOR_FROM || state == LOOK_FOR_MSGID)
		return -1;		/* syntax error */

	m->to = mallocw(userlen + hostlen + 2);

	strncpy(m->to, user, userlen);
	m->to[userlen] = '\0';

	if(hostlen){
		m->to[userlen] = '@';
		strncpy(m->to + userlen + 1, host, hostlen);
		m->to[userlen + hostlen + 1] = '\0';
	}
	if(fromlen){
		m->tofrom = mallocw(fromlen + 1);
		strncpy(m->tofrom, from, fromlen);
		m->tofrom[fromlen] = '\0';
	}
	if(msgidlen){
		m->tomsgid = mallocw(msgidlen + 1);
		strncpy(m->tomsgid, msgid, msgidlen);
		m->tomsgid[msgidlen] = '\0';
	}
	return 0;
}

/* This opens the data file and writes the mail header into it.
 * Returns 0 if OK, and -1 if not.
 */
static int
mbx_data(m,cclist,extra)
struct mbx *m;
struct list *cclist;	/* list of carbon copy recipients */
char *extra;        /* optional extra header lines */
{
    time_t t;
    struct list *ap;
	int cccnt = 0;
	
#ifdef notdef
    time(&t);
/*These 2 lines get added again when the smtp-server handles the mail
 *not really needed - WG7J
 */
    fprintf(m->tfile,Hdrs[RECEIVED]);
	if(m->tofrom != NULLCHAR)
		fprintf(m->tfile,"from %s ",m->name);
	fprintf(m->tfile,"by %s (%s)\n\tid AA%ld ; %s",
        Hostname, Version, get_msgid(),ptime(&t));
#endif

    /* If m->date is set, use this one (comes from bbs-forwarded mail) */
    if(m->date != NULLCHAR)
        fprintf(m->tfile,"%s%s",Hdrs[DATE],m->date);
    else {
        time(&t);
        fprintf(m->tfile,"%s%s",Hdrs[DATE],ptime(&t));
    }

    /* Bulletin ID, if any */
	fprintf(m->tfile,Hdrs[MSGID]);
	if(m->tomsgid)
		fprintf(m->tfile,"<%s@%s.bbs>\n", m->tomsgid, m->name);
	else
		fprintf(m->tfile,"<%ld@%s>\n",get_msgid(), Hostname);

    /* From : , could use 'real bbs address', if origbbs is set */
    fprintf(m->tfile,Hdrs[FROM]);
    if(m->tofrom) {  /* BBS style '< call' */
        if(m->origbbs != NULLCHAR)
            fprintf(m->tfile,"%s@%s\n",m->tofrom,m->origbbs);
        else
            fprintf(m->tfile,"%s%%%s@%s\n",m->tofrom, m->name, Hostname);
    } else {
        if(m->origbbs != NULLCHAR)
            fprintf(m->tfile,"%s@%s\n",m->name,m->origbbs);
        else
            fprintf(m->tfile,"%s@%s\n",m->name,Hostname);
    }

    fprintf(m->tfile,"%s%s\n",Hdrs[TO],m->origto != NULLCHAR ? m->origto : m->to);

    /* Write Cc: line */
	for(ap = cclist; ap != NULLLIST; ap = ap->next) {
		if(cccnt == 0){
			fprintf(m->tfile,"%s",Hdrs[CC]);
			cccnt = 4;
		}
		else {
		       fprintf(m->tfile,", ");
		       cccnt += 2;
		}
		if(cccnt + strlen(ap->val) > 80 - 3) {
		       fprintf(m->tfile,"\n    ");
		       cccnt = 4;
		}
		fputs(ap->val,m->tfile);
		cccnt += strlen(ap->val);
	}
	if(cccnt)
		fputc('\n',m->tfile);
    fprintf(m->tfile,"%s%s\n",Hdrs[SUBJECT],m->subject);
	if(!isspace(m->stype) && ((m->stype != 'R' && m->stype != 'F') ||
	  (m->sid & MBX_SID) !=0))
		  fprintf(m->tfile,"%s%c\n", Hdrs[BBSTYPE],m->stype);

#ifdef notdef
    /*not really needed, the 'From <user%fwdbbs@host>' shows this too!*/
    /* Also store the 'real smtp from' address */
    if((m->tofrom != NULLCHAR) && (m->origbbs != NULLCHAR))
        fprintf(m->tfile,"%s%s%%%s@%s\n",Hdrs[XFROM],m->tofrom,m->name, Hostname);
#endif

    if(extra != NULLCHAR)
		fprintf(m->tfile,extra);

	return 0;
}

/* Returns true if string is in history file or if string appears to be a
 * message id generated by our system.
 */
static int
msgidcheck(string)
char *string;
{
     FILE *fp;
     char buf[LINELEN], *cp;
     if(string == NULLCHAR)
	  return 0;
     /* BID's that we have generated ourselves are not kept in the history
      * file. Such BID's are in the nnnn_hhhh form, where hhhh is a part of
      * our hostname, truncated so that the BID is no longer than 11
      * characters.
      */
     if((cp = strchr(string,'_')) != NULLCHAR && *(cp+1) != '\0' && 
	strnicmp(cp+1,Hostname,strlen(cp+1)) == 0)
	  return 1;

     if((fp = fopen(Historyfile,READ_TEXT)) == NULLFILE)
	  return 0;
     while(fgets(buf,LINELEN,fp) != NULLCHAR) {
        rip(buf);
        /* Get rid of following spaces or tabs, to separate the
         * bid and it's time stamp entry - WG7J
         */
        if( ((cp=strchr(buf,' ')) != NULLCHAR) || \
            ((cp=strchr(buf,'\t')) != NULLCHAR) )
            *cp = '\0';
        if(stricmp(string,buf) == 0) {    /* found */
            fclose(fp);
            return 1;
        }
    }
    fclose(fp);
    return 0;
}
     

/* uuencode a file -- translated from C++; both versions copyright 1990
   by David R. Evans, G4AMJ/NQ0I
*/

static int
uuencode(infile,s,infilename)
FILE *infile;
int s;			/* output socket */
char *infilename;
{
  int n_read_so_far = 0, n_written_so_far = 0, in_chars, n, mode = 0755;
  unsigned long cnt = 0;
  unsigned char in[3], out[4], line[100];
#ifdef UNIX
  struct stat stb;
  
  if(stat(infilename,&stb) != -1)
       mode = stb.st_mode & 0777;	/* get real file protection mode */
#endif
  usprintf(s, "begin %03o %s\n", mode, infilename);

  /* do the encode */
 for(;;){	
    in_chars = fread(in, 1, 3, infile);
    out[0] = in[0] >> 2;
    out[1] = in[0] << 6;
    out[1] = out[1] >> 2;
    out[1] = out[1] | (in[1] >> 4);
    out[2] = in[1] << 4;
    out[2] = out[2] >> 2;
    out[2] = out[2] | (in[2] >> 6);
    out[3] = in[2] << 2;
    out[3] = out[3] >> 2;
    for (n = 0; n < 4; n++)
      out[n] += ' ';
    n_read_so_far += in_chars;
    for (n = 0; n < 4; n++)
      line[n_written_so_far++] = out[n];
    if (((in_chars != 3) || (n_written_so_far == 60)) && n_read_so_far > 0) {
      line[(n_read_so_far + 2) / 3 * 4] = '\0';
      
      usprintf(s,"%c%s\n",n_read_so_far + ' ', line);
      cnt += n_read_so_far;
      n_read_so_far = 0;
      n_written_so_far = 0;
    }
    if (in_chars == 0)
      break;
  }
  if (usprintf(s," \nend\nsize %lu\n", cnt) == EOF)
    return 1;
  return 0;
}

/* Attempt to determine if this is third-pary mail. */
static int 
thirdparty(m)
struct mbx *m;
{
	char buf[MBXLINE], *cp, *rp;
	FILE *fp;

	if(strchr(m->to,'@') != NULLCHAR || strchr(m->to,'%') != NULLCHAR
			|| strchr(m->to,'!') != NULLCHAR)
		return 0;

	rp = strdup(Hostname);

	if((cp = strchr(rp, '.')) != NULLCHAR)
		*cp = '\0';

	if(stricmp(m->to,rp) == 0){
		free(rp);
		return -1;
	}
	free(rp);

	if(stricmp(m->to,"sysop") == 0)
		return -1;

	if((fp = fopen(Arealist,READ_TEXT)) == NULLFILE)
		return 0;

	while(fgets(buf,MBXLINE,fp) != NULLCHAR){
		/* The first word on each line is all that matters */
		if((cp = strchr(buf, ' ')) != NULLCHAR)
			*cp = '\0';
		if((cp = strchr(buf,'\t')) != NULLCHAR)
			*cp = '\0';
		if(stricmp(m->to,buf) == 0){
			fclose(fp);
			return -1;
		}
	}
	fclose(fp);
	return 0;
}

#ifdef NETROM
extern struct nrroute_tab *find_nrboth(char *alias,char *call);
#endif

extern int Mbconverse;

static int
dombconnect(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
    char buf[7];
    struct nrroute_tab *np;
    int ndigis,i,s;
    struct sockaddr_nr lsocket,fsocket;
    struct sockaddr_ax alsocket;    /*the local socket*/
    struct sockaddr_ax afsocket;    /*the remote socket*/
    struct iface *ifp;
    char alias[AXBUF];
    char local_call[AXALEN];
    char digis[MAXDIGIS][AXALEN];
	char target[AXALEN];

    m = (struct mbx *) p;

	if(argc == 1){
#ifdef  CONVERS
        if(!Mbconverse) {
            tputs("Convers server not enabled\n");
            return 0;
        }
#ifdef GWTRACE
        log(m->user,"MBOX CONVERS: %s",m->name);
#endif
        m->state = MBX_CONVERS;
        mbox_converse(m);
#else
        dombconnecthelp();
#endif
		return 0;
	}

    if(MBSecure)
#ifdef NETROM
        if((m->family != AF_AX25) && (m->family != AF_NETROM)) {
#else
		if(m->family != AF_AX25) {
#endif
            tputs(Noperm);
            return 0;
        }

    if (argc == 2) {
#ifndef NETROM
		dombconnecthelp();
		return 0;
	}
#else
        /*NETROM connection wanted*/
        if(!(m->privs & NETROM_CMD)) {
            tputs(Noperm);
            return 0;
        }

        if(Nrifaces[0].iface == NULLIF){
            tputs("NET/ROM not activated.\n\n");
            return 0;
        }
        /* See if the requested destination is a known alias or call,
         * use it if it is.  Otherwize give an error message.
         */
        putalias(alias,argv[1],0);
        strupr(argv[1]);    /*make sure it's upper case*/
        if((np = find_nrboth(alias,argv[1])) == NULLNRRTAB){
            /*no such call or node alias*/
            tputs("no such node\n\n");
            dombconnecthelp();
            dombports(0,NULL,NULL);
            return 0;
        }

        if((s = socket(AF_NETROM,SOCK_SEQPACKET,0)) == -1){
            tputs(Nosock);
            return 0;
        }
#ifdef GWTRACE
		log(m->user,"MBOX NETROM: %s to %s",m->name,argv[1]);
#endif
		lsocket.nr_family = AF_NETROM;

        /* Set up our local username, bind would use Mycall instead */
        memcpy(lsocket.nr_addr.user,m->call,AXALEN);
        /* Set up our source address */
        memcpy(lsocket.nr_addr.node,Nr_iface->hwaddr,AXALEN);

        bind(s,(char *)&lsocket,sizeof(struct sockaddr_nr));

        memcpy(fsocket.nr_addr.user,np->call,AXALEN);
        memcpy(fsocket.nr_addr.node,np->call,AXALEN);
        fsocket.nr_family = AF_NETROM;
        m->state = MBX_GATEWAY;
        return gw_connect(m,s,(struct sockaddr *)&fsocket, sizeof(struct sockaddr_nr));
    }
#endif /*NETROM*/

#ifdef AX25
    if(argc > 2) {
        /*AX25 gateway connection wanted*/
        if(!(m->privs & AX25_CMD)) {
            tputs(Noperm);
            return 0;
        }

        if((ifp = if_lookup(argv[1])) == NULLIF) {
            tprintf("Unknown port %s\n",argv[1]);
            dombports(0,NULL,NULL);
            return 0;
        }
        if(ifp->type != CL_AX25){
            tprintf("Port %s not usable for AX.25 connects\n",argv[1]);
            dombports(0,NULL,NULL);
            return 0;
        }
        if(setcall(target,argv[2]) == -1){
            tprintf("Bad call %s\n",argv[2]);
            return 0;
        }
        /* If digipeaters are given, put them in the routing table */
        if(argc > 3){
            ndigis = argc - 3;
            if(ndigis > MAXDIGIS){
                tputs("Too many digipeaters\n");
                return 0;
            }
            for(i=0;i<ndigis;i++){
                if(setcall(digis[i],argv[i+3]) == -1){
                    tprintf("Bad digipeater %s\n",argv[i+3]);
                    return 0;
                }
            }
			if(ax_add(target,AX_AUTO,digis,ndigis) == NULLAXR){
                tputs("AX25 route add failed\n");
                return 0;
            }
        }
        if((s = socket(AF_AX25,SOCK_STREAM,0)) == -1){
            tputs(Nosock);
            return 0;
        }
#ifdef GWTRACE
		log(m->user,"MBOX AX25: %s to %s on %s",m->name,argv[2],argv[1]);
#endif

        /*fill in the known stuff*/
        alsocket.sax_family = afsocket.sax_family= AF_AX25;
        
        /*the remote call to connect to*/
        setcall(afsocket.ax25_addr,argv[2]);

        /*the outgoing interface*/
        strncpy(afsocket.iface,argv[1],ILEN);

        /*now set local user call, invert ssid*/
        memcpy(local_call,m->call,AXALEN);
        local_call[AXALEN-1] ^= 0x1e;
        memcpy(alsocket.ax25_addr,local_call,AXALEN);
        /*and bind it (otherwize Mycall will be used!)*/
        bind(s,(char *)&alsocket,sizeof(struct sockaddr_ax));
        m->state = MBX_GATEWAY;
        return gw_connect(m,s,(struct sockaddr *)&afsocket, sizeof(struct sockaddr_ax));
    }
#endif /* AX25 */

    return 0;
}

#endif /* MAILBOX */

